package dev.pfaff.glass.apple.gcd;

import dev.pfaff.glass.posix.PThread;
import dev.pfaff.log4truth.Logger;

import java.lang.foreign.*;
import java.lang.invoke.MethodHandle;
import java.lang.invoke.MethodHandles;
import java.util.Arrays;
import java.util.List;
import java.util.concurrent.ConcurrentLinkedQueue;

import static dev.pfaff.glass.Interop.ADDRESS_INT;
import static dev.pfaff.glass.apple.SystemFramework.SYSTEM_FRAMEWORK;
import static java.lang.invoke.MethodType.methodType;

/**
 * Bindings to Apple's Grand Central Dispatch.
 */
public final class GrandCentralDispatch {
	static final MethodHandle MH_dispatch_async_f;
	static final MethodHandle MH_dispatch_async_and_wait_f;
	private static final List<DispatchQueue> globalQueues;
	private static final DispatchQueue mainQueue;

	static final FunctionDescriptor FD_runTaskBound = FunctionDescriptor.ofVoid();
	static final MethodHandle MH_runNextTask;

	static {
		var l = MethodHandles.lookup();

		try {
			MH_runNextTask = l.findStatic(GrandCentralDispatch.class,
					"runTask",
					methodType(void.class, ConcurrentLinkedQueue.class));
		} catch (NoSuchMethodException | IllegalAccessException e) {
			throw new RuntimeException(e);
		}

		var linker = Linker.nativeLinker();
		MH_dispatch_async_f = linker.downcallHandle(SYSTEM_FRAMEWORK.find("dispatch_async_f").get(),
				FunctionDescriptor.ofVoid(ValueLayout.ADDRESS, ValueLayout.ADDRESS, ValueLayout.ADDRESS));
		MH_dispatch_async_and_wait_f = linker.downcallHandle(SYSTEM_FRAMEWORK.find("dispatch_async_and_wait_f").get(),
				FunctionDescriptor.ofVoid(ValueLayout.ADDRESS, ValueLayout.ADDRESS, ValueLayout.ADDRESS));
		var MH_dispatch_get_global_queue = linker.downcallHandle(SYSTEM_FRAMEWORK.find("dispatch_get_global_queue").get(),
				FunctionDescriptor.of(ValueLayout.ADDRESS, ADDRESS_INT, ADDRESS_INT));
		globalQueues = Arrays.stream(QoSClass.values())
				.map(qosClass -> {
					MemorySegment MS_queue;
					try {
						MS_queue = (MemorySegment) MH_dispatch_get_global_queue.invokeExact(qosClass.value, 0L);
					} catch (Throwable e) {
						throw new RuntimeException(e);
					}
					return new DispatchQueue(MS_queue, Arena.ofAuto());
				})
				.toList();
		mainQueue = new DispatchQueue(SYSTEM_FRAMEWORK.find("_dispatch_main_q").get(), Arena.ofAuto(), PThread::pthread_main_np);
	}

	/**
	 * Runs a single task from the queue on the current thread.
	 */
	private static void runTask(ConcurrentLinkedQueue<Runnable> queue) {
		try {
			var task = queue.poll();
			if (task != null) {
				task.run();
			} else {
				handleEmptyQueue(queue);
			}
		} catch (Throwable e) {
			handleUncaughtExceptionInRunTask(e);
		}
	}

	private static void handleEmptyQueue(ConcurrentLinkedQueue<Runnable> queue) {
		Logger.log(() -> "runTask called on empty queue 0x" + Long.toHexString(System.identityHashCode(queue)));
	}

	private static void handleUncaughtExceptionInRunTask(Throwable e) {
		Thread.getDefaultUncaughtExceptionHandler().uncaughtException(Thread.currentThread(), e);
	}

	public static DispatchQueue getMainQueue() {
		return mainQueue;
	}

	public static DispatchQueue getGlobalQueue(QoSClass qosClass) {
		return globalQueues.get(qosClass.ordinal());
	}
}
