package dev.pfaff.glass.opengl;

import java.lang.foreign.*;
import java.lang.invoke.MethodHandle;
import java.util.List;

import dev.pfaff.log4truth.StringSupplier;
import dev.pfaff.glass.opengl.enums.*;

import static java.lang.foreign.ValueLayout.*;

import static dev.pfaff.glass.opengl.Util.*;

/**
 * OpenGL bindings for Java.
 */
public final class GL {
	/**
	 * Whether correctness checks are enabled.
	 */
	public static final boolean CHECKS = Util.CHECKS;
	/**
	 * Whether error checks after each native call are enabled.
	 */
	public static final boolean ERROR_CHECKS = Util.ERROR_CHECKS;
	/**
	 * Whether logging commands is enabled.
	 */
	public static final boolean LOG_COMMANDS = Util.LOG_COMMANDS;

	private final SymbolLookup symbolLookup;
	private final GLFunctions functions;

	public GL(SymbolLookup symbolLookup) {
		this.symbolLookup = symbolLookup;
		this.functions = GLFunctions.make(symbolLookup);
	}

	public SymbolLookup symbolLookup() {
		return symbolLookup;
	}

	public GLFunctions functions() {
		return functions;
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glAccum">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glAccum.xhtml">registry.khronos.org</a>
	 */
	public void glAccum(GLAccumOp op, float value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glAccum(\{op}, \{value})");
		try {
		    functions.glAccum().invokeExact(op, value);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were op=\{op}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glAccumxOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glAccumxOES.xhtml">registry.khronos.org</a>
	 */
	public void glAccumxOES(GLEnum op, int value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glAccumxOES(\{op}, \{value})");
		try {
		    functions.glAccumxOES().invokeExact(op, value);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were op=\{op}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glActiveProgramEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glActiveProgramEXT.xhtml">registry.khronos.org</a>
	 */
	public void glActiveProgramEXT(int program) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glActiveProgramEXT(\{program})");
		try {
		    functions.glActiveProgramEXT().invokeExact(program);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glActiveShaderProgram">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glActiveShaderProgram.xhtml">registry.khronos.org</a>
	 */
	public void glActiveShaderProgram(int pipeline, int program) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glActiveShaderProgram(\{pipeline}, \{program})");
		try {
		    functions.glActiveShaderProgram().invokeExact(pipeline, program);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glActiveShaderProgramEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glActiveShaderProgramEXT.xhtml">registry.khronos.org</a>
	 */
	public void glActiveShaderProgramEXT(int pipeline, int program) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glActiveShaderProgramEXT(\{pipeline}, \{program})");
		try {
		    functions.glActiveShaderProgramEXT().invokeExact(pipeline, program);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glActiveStencilFaceEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glActiveStencilFaceEXT.xhtml">registry.khronos.org</a>
	 */
	public void glActiveStencilFaceEXT(GLTriangleFace face) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glActiveStencilFaceEXT(\{face})");
		try {
		    functions.glActiveStencilFaceEXT().invokeExact(face);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were face=\{face}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glActiveTexture">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glActiveTexture.xhtml">registry.khronos.org</a>
	 */
	public void glActiveTexture(GLTextureUnit texture) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glActiveTexture(\{texture})");
		try {
		    functions.glActiveTexture().invokeExact(texture);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were texture=\{texture}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glActiveTextureARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glActiveTextureARB.xhtml">registry.khronos.org</a>
	 */
	public void glActiveTextureARB(GLTextureUnit texture) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glActiveTextureARB(\{texture})");
		try {
		    functions.glActiveTextureARB().invokeExact(texture);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were texture=\{texture}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glActiveVaryingNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glActiveVaryingNV.xhtml">registry.khronos.org</a>
	 */
	public void glActiveVaryingNV(int program, ConstPointer<Byte> name) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glActiveVaryingNV(\{program}, \{name})");
		try {
		    functions.glActiveVaryingNV().invokeExact(program, name);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glAlphaFragmentOp1ATI">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glAlphaFragmentOp1ATI.xhtml">registry.khronos.org</a>
	 */
	public void glAlphaFragmentOp1ATI(GLFragmentOp1ATI op, int dst, int dstMod, int arg1, int arg1Rep, int arg1Mod) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glAlphaFragmentOp1ATI(\{op}, \{dst}, \{dstMod}, \{arg1}, \{arg1Rep}, \{arg1Mod})");
		try {
		    functions.glAlphaFragmentOp1ATI().invokeExact(op, dst, dstMod, arg1, arg1Rep, arg1Mod);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were op=\{op}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glAlphaFragmentOp2ATI">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glAlphaFragmentOp2ATI.xhtml">registry.khronos.org</a>
	 */
	public void glAlphaFragmentOp2ATI(GLFragmentOp2ATI op, int dst, int dstMod, int arg1, int arg1Rep, int arg1Mod, int arg2, int arg2Rep, int arg2Mod) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glAlphaFragmentOp2ATI(\{op}, \{dst}, \{dstMod}, \{arg1}, \{arg1Rep}, \{arg1Mod}, \{arg2}, \{arg2Rep}, \{arg2Mod})");
		try {
		    functions.glAlphaFragmentOp2ATI().invokeExact(op, dst, dstMod, arg1, arg1Rep, arg1Mod, arg2, arg2Rep, arg2Mod);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were op=\{op}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glAlphaFragmentOp3ATI">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glAlphaFragmentOp3ATI.xhtml">registry.khronos.org</a>
	 */
	public void glAlphaFragmentOp3ATI(GLFragmentOp3ATI op, int dst, int dstMod, int arg1, int arg1Rep, int arg1Mod, int arg2, int arg2Rep, int arg2Mod, int arg3, int arg3Rep, int arg3Mod) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glAlphaFragmentOp3ATI(\{op}, \{dst}, \{dstMod}, \{arg1}, \{arg1Rep}, \{arg1Mod}, \{arg2}, \{arg2Rep}, \{arg2Mod}, \{arg3}, \{arg3Rep}, \{arg3Mod})");
		try {
		    functions.glAlphaFragmentOp3ATI().invokeExact(op, dst, dstMod, arg1, arg1Rep, arg1Mod, arg2, arg2Rep, arg2Mod, arg3, arg3Rep, arg3Mod);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were op=\{op}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glAlphaFunc">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glAlphaFunc.xhtml">registry.khronos.org</a>
	 */
	public void glAlphaFunc(GLAlphaFunction func, float ref) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glAlphaFunc(\{func}, \{ref})");
		try {
		    functions.glAlphaFunc().invokeExact(func, ref);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were func=\{func}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glAlphaFuncQCOM">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glAlphaFuncQCOM.xhtml">registry.khronos.org</a>
	 */
	public void glAlphaFuncQCOM(GLEnum func, float ref) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glAlphaFuncQCOM(\{func}, \{ref})");
		try {
		    functions.glAlphaFuncQCOM().invokeExact(func, ref);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were func=\{func}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glAlphaFuncx">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glAlphaFuncx.xhtml">registry.khronos.org</a>
	 */
	public void glAlphaFuncx(GLAlphaFunction func, int ref) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glAlphaFuncx(\{func}, \{ref})");
		try {
		    functions.glAlphaFuncx().invokeExact(func, ref);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were func=\{func}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glAlphaFuncxOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glAlphaFuncxOES.xhtml">registry.khronos.org</a>
	 */
	public void glAlphaFuncxOES(GLAlphaFunction func, int ref) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glAlphaFuncxOES(\{func}, \{ref})");
		try {
		    functions.glAlphaFuncxOES().invokeExact(func, ref);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were func=\{func}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glAlphaToCoverageDitherControlNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glAlphaToCoverageDitherControlNV.xhtml">registry.khronos.org</a>
	 */
	public void glAlphaToCoverageDitherControlNV(GLEnum mode) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glAlphaToCoverageDitherControlNV(\{mode})");
		try {
		    functions.glAlphaToCoverageDitherControlNV().invokeExact(mode);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glApplyFramebufferAttachmentCMAAINTEL">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glApplyFramebufferAttachmentCMAAINTEL.xhtml">registry.khronos.org</a>
	 */
	public void glApplyFramebufferAttachmentCMAAINTEL() {
	    if (LOG_COMMANDS) logCmd(() -> STR."glApplyFramebufferAttachmentCMAAINTEL()");
		try {
		    functions.glApplyFramebufferAttachmentCMAAINTEL().invokeExact();
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glApplyTextureEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glApplyTextureEXT.xhtml">registry.khronos.org</a>
	 */
	public void glApplyTextureEXT(GLLightTextureModeEXT mode) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glApplyTextureEXT(\{mode})");
		try {
		    functions.glApplyTextureEXT().invokeExact(mode);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glAcquireKeyedMutexWin32EXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glAcquireKeyedMutexWin32EXT.xhtml">registry.khronos.org</a>
	 */
	public boolean glAcquireKeyedMutexWin32EXT(int memory, long key, int timeout) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glAcquireKeyedMutexWin32EXT(\{memory}, \{key}, \{timeout})");
		try {
		    var r = (boolean)functions.glAcquireKeyedMutexWin32EXT().invokeExact(memory, key, timeout);
			checkError();
			return r;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glAreProgramsResidentNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glAreProgramsResidentNV.xhtml">registry.khronos.org</a>
	 */
	public boolean glAreProgramsResidentNV(int n, ConstPointer<Integer> programs, MutPointer<Boolean> residences) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glAreProgramsResidentNV(\{n}, \{programs}, \{residences})");
		try {
		    var r = (boolean)functions.glAreProgramsResidentNV().invokeExact(n, programs, residences);
			checkError();
			return r;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glAreTexturesResident">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glAreTexturesResident.xhtml">registry.khronos.org</a>
	 */
	public boolean glAreTexturesResident(int n, ConstPointer<Integer> textures, MutPointer<Boolean> residences) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glAreTexturesResident(\{n}, \{textures}, \{residences})");
		try {
		    var r = (boolean)functions.glAreTexturesResident().invokeExact(n, textures, residences);
			checkError();
			return r;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glAreTexturesResidentEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glAreTexturesResidentEXT.xhtml">registry.khronos.org</a>
	 */
	public boolean glAreTexturesResidentEXT(int n, ConstPointer<Integer> textures, MutPointer<Boolean> residences) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glAreTexturesResidentEXT(\{n}, \{textures}, \{residences})");
		try {
		    var r = (boolean)functions.glAreTexturesResidentEXT().invokeExact(n, textures, residences);
			checkError();
			return r;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glArrayElement">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glArrayElement.xhtml">registry.khronos.org</a>
	 */
	public void glArrayElement(int i) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glArrayElement(\{i})");
		try {
		    functions.glArrayElement().invokeExact(i);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glArrayElementEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glArrayElementEXT.xhtml">registry.khronos.org</a>
	 */
	public void glArrayElementEXT(int i) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glArrayElementEXT(\{i})");
		try {
		    functions.glArrayElementEXT().invokeExact(i);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glArrayObjectATI">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glArrayObjectATI.xhtml">registry.khronos.org</a>
	 */
	public void glArrayObjectATI(GLEnableCap array, int size, GLScalarType type, int stride, int buffer, int offset) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glArrayObjectATI(\{array}, \{size}, \{type}, \{stride}, \{buffer}, \{offset})");
		try {
		    functions.glArrayObjectATI().invokeExact(array, size, type, stride, buffer, offset);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were array=\{array}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glAsyncCopyBufferSubDataNVX">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glAsyncCopyBufferSubDataNVX.xhtml">registry.khronos.org</a>
	 */
	public int glAsyncCopyBufferSubDataNVX(int waitSemaphoreCount, ConstPointer<Integer> waitSemaphoreArray, ConstPointer<Long> fenceValueArray, int readGpu, int writeGpuMask, int readBuffer, int writeBuffer, long readOffset, long writeOffset, long size, int signalSemaphoreCount, ConstPointer<Integer> signalSemaphoreArray, ConstPointer<Long> signalValueArray) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glAsyncCopyBufferSubDataNVX(\{waitSemaphoreCount}, \{waitSemaphoreArray}, \{fenceValueArray}, \{readGpu}, \{writeGpuMask}, \{readBuffer}, \{writeBuffer}, \{readOffset}, \{writeOffset}, \{size}, \{signalSemaphoreCount}, \{signalSemaphoreArray}, \{signalValueArray})");
		try {
		    var r = (int)functions.glAsyncCopyBufferSubDataNVX().invokeExact(waitSemaphoreCount, waitSemaphoreArray, fenceValueArray, readGpu, writeGpuMask, readBuffer, writeBuffer, readOffset, writeOffset, size, signalSemaphoreCount, signalSemaphoreArray, signalValueArray);
			checkError();
			return r;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glAsyncCopyImageSubDataNVX">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glAsyncCopyImageSubDataNVX.xhtml">registry.khronos.org</a>
	 */
	public int glAsyncCopyImageSubDataNVX(int waitSemaphoreCount, ConstPointer<Integer> waitSemaphoreArray, ConstPointer<Long> waitValueArray, int srcGpu, int dstGpuMask, int srcName, GLEnum srcTarget, int srcLevel, int srcX, int srcY, int srcZ, int dstName, GLEnum dstTarget, int dstLevel, int dstX, int dstY, int dstZ, int srcWidth, int srcHeight, int srcDepth, int signalSemaphoreCount, ConstPointer<Integer> signalSemaphoreArray, ConstPointer<Long> signalValueArray) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glAsyncCopyImageSubDataNVX(\{waitSemaphoreCount}, \{waitSemaphoreArray}, \{waitValueArray}, \{srcGpu}, \{dstGpuMask}, \{srcName}, \{srcTarget}, \{srcLevel}, \{srcX}, \{srcY}, \{srcZ}, \{dstName}, \{dstTarget}, \{dstLevel}, \{dstX}, \{dstY}, \{dstZ}, \{srcWidth}, \{srcHeight}, \{srcDepth}, \{signalSemaphoreCount}, \{signalSemaphoreArray}, \{signalValueArray})");
		try {
		    var r = (int)functions.glAsyncCopyImageSubDataNVX().invokeExact(waitSemaphoreCount, waitSemaphoreArray, waitValueArray, srcGpu, dstGpuMask, srcName, srcTarget, srcLevel, srcX, srcY, srcZ, dstName, dstTarget, dstLevel, dstX, dstY, dstZ, srcWidth, srcHeight, srcDepth, signalSemaphoreCount, signalSemaphoreArray, signalValueArray);
			checkError();
			return r;
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were srcTarget=\{srcTarget}, dstTarget=\{dstTarget}.");
			}
			throw e;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glAsyncMarkerSGIX">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glAsyncMarkerSGIX.xhtml">registry.khronos.org</a>
	 */
	public void glAsyncMarkerSGIX(int marker) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glAsyncMarkerSGIX(\{marker})");
		try {
		    functions.glAsyncMarkerSGIX().invokeExact(marker);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glAttachObjectARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glAttachObjectARB.xhtml">registry.khronos.org</a>
	 */
	public void glAttachObjectARB(int containerObj, int obj) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glAttachObjectARB(\{containerObj}, \{obj})");
		try {
		    functions.glAttachObjectARB().invokeExact(containerObj, obj);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glAttachShader">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glAttachShader.xhtml">registry.khronos.org</a>
	 */
	public void glAttachShader(int program, int shader) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glAttachShader(\{program}, \{shader})");
		try {
		    functions.glAttachShader().invokeExact(program, shader);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBegin">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBegin.xhtml">registry.khronos.org</a>
	 */
	public void glBegin(GLPrimitiveType mode) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBegin(\{mode})");
		try {
		    functions.glBegin().invokeExact(mode);
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBeginConditionalRender">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBeginConditionalRender.xhtml">registry.khronos.org</a>
	 */
	public void glBeginConditionalRender(int id, GLConditionalRenderMode mode) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBeginConditionalRender(\{id}, \{mode})");
		try {
		    functions.glBeginConditionalRender().invokeExact(id, mode);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBeginConditionalRenderNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBeginConditionalRenderNV.xhtml">registry.khronos.org</a>
	 */
	public void glBeginConditionalRenderNV(int id, GLConditionalRenderMode mode) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBeginConditionalRenderNV(\{id}, \{mode})");
		try {
		    functions.glBeginConditionalRenderNV().invokeExact(id, mode);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBeginConditionalRenderNVX">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBeginConditionalRenderNVX.xhtml">registry.khronos.org</a>
	 */
	public void glBeginConditionalRenderNVX(int id) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBeginConditionalRenderNVX(\{id})");
		try {
		    functions.glBeginConditionalRenderNVX().invokeExact(id);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBeginFragmentShaderATI">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBeginFragmentShaderATI.xhtml">registry.khronos.org</a>
	 */
	public void glBeginFragmentShaderATI() {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBeginFragmentShaderATI()");
		try {
		    functions.glBeginFragmentShaderATI().invokeExact();
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBeginOcclusionQueryNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBeginOcclusionQueryNV.xhtml">registry.khronos.org</a>
	 */
	public void glBeginOcclusionQueryNV(int id) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBeginOcclusionQueryNV(\{id})");
		try {
		    functions.glBeginOcclusionQueryNV().invokeExact(id);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBeginPerfMonitorAMD">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBeginPerfMonitorAMD.xhtml">registry.khronos.org</a>
	 */
	public void glBeginPerfMonitorAMD(int monitor) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBeginPerfMonitorAMD(\{monitor})");
		try {
		    functions.glBeginPerfMonitorAMD().invokeExact(monitor);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBeginPerfQueryINTEL">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBeginPerfQueryINTEL.xhtml">registry.khronos.org</a>
	 */
	public void glBeginPerfQueryINTEL(int queryHandle) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBeginPerfQueryINTEL(\{queryHandle})");
		try {
		    functions.glBeginPerfQueryINTEL().invokeExact(queryHandle);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBeginQuery">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBeginQuery.xhtml">registry.khronos.org</a>
	 */
	public void glBeginQuery(GLQueryTarget target, int id) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBeginQuery(\{target}, \{id})");
		try {
		    functions.glBeginQuery().invokeExact(target, id);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBeginQueryARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBeginQueryARB.xhtml">registry.khronos.org</a>
	 */
	public void glBeginQueryARB(GLQueryTarget target, int id) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBeginQueryARB(\{target}, \{id})");
		try {
		    functions.glBeginQueryARB().invokeExact(target, id);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBeginQueryEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBeginQueryEXT.xhtml">registry.khronos.org</a>
	 */
	public void glBeginQueryEXT(GLQueryTarget target, int id) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBeginQueryEXT(\{target}, \{id})");
		try {
		    functions.glBeginQueryEXT().invokeExact(target, id);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBeginQueryIndexed">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBeginQueryIndexed.xhtml">registry.khronos.org</a>
	 */
	public void glBeginQueryIndexed(GLQueryTarget target, int index, int id) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBeginQueryIndexed(\{target}, \{index}, \{id})");
		try {
		    functions.glBeginQueryIndexed().invokeExact(target, index, id);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBeginTransformFeedback">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBeginTransformFeedback.xhtml">registry.khronos.org</a>
	 */
	public void glBeginTransformFeedback(GLPrimitiveType primitiveMode) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBeginTransformFeedback(\{primitiveMode})");
		try {
		    functions.glBeginTransformFeedback().invokeExact(primitiveMode);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were primitiveMode=\{primitiveMode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBeginTransformFeedbackEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBeginTransformFeedbackEXT.xhtml">registry.khronos.org</a>
	 */
	public void glBeginTransformFeedbackEXT(GLPrimitiveType primitiveMode) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBeginTransformFeedbackEXT(\{primitiveMode})");
		try {
		    functions.glBeginTransformFeedbackEXT().invokeExact(primitiveMode);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were primitiveMode=\{primitiveMode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBeginTransformFeedbackNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBeginTransformFeedbackNV.xhtml">registry.khronos.org</a>
	 */
	public void glBeginTransformFeedbackNV(GLPrimitiveType primitiveMode) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBeginTransformFeedbackNV(\{primitiveMode})");
		try {
		    functions.glBeginTransformFeedbackNV().invokeExact(primitiveMode);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were primitiveMode=\{primitiveMode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBeginVertexShaderEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBeginVertexShaderEXT.xhtml">registry.khronos.org</a>
	 */
	public void glBeginVertexShaderEXT() {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBeginVertexShaderEXT()");
		try {
		    functions.glBeginVertexShaderEXT().invokeExact();
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBeginVideoCaptureNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBeginVideoCaptureNV.xhtml">registry.khronos.org</a>
	 */
	public void glBeginVideoCaptureNV(int video_capture_slot) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBeginVideoCaptureNV(\{video_capture_slot})");
		try {
		    functions.glBeginVideoCaptureNV().invokeExact(video_capture_slot);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBindAttribLocation">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBindAttribLocation.xhtml">registry.khronos.org</a>
	 */
	public void glBindAttribLocation(int program, int index, ConstPointer<Byte> name) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBindAttribLocation(\{program}, \{index}, \{name})");
		try {
		    functions.glBindAttribLocation().invokeExact(program, index, name);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBindAttribLocationARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBindAttribLocationARB.xhtml">registry.khronos.org</a>
	 */
	public void glBindAttribLocationARB(int programObj, int index, ConstPointer<Byte> name) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBindAttribLocationARB(\{programObj}, \{index}, \{name})");
		try {
		    functions.glBindAttribLocationARB().invokeExact(programObj, index, name);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBindBuffer">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBindBuffer.xhtml">registry.khronos.org</a>
	 */
	public void glBindBuffer(GLBufferTarget target, int buffer) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBindBuffer(\{target}, \{buffer})");
		try {
		    functions.glBindBuffer().invokeExact(target, buffer);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBindBufferARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBindBufferARB.xhtml">registry.khronos.org</a>
	 */
	public void glBindBufferARB(GLBufferTarget target, int buffer) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBindBufferARB(\{target}, \{buffer})");
		try {
		    functions.glBindBufferARB().invokeExact(target, buffer);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBindBufferBase">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBindBufferBase.xhtml">registry.khronos.org</a>
	 */
	public void glBindBufferBase(GLBufferTarget target, int index, int buffer) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBindBufferBase(\{target}, \{index}, \{buffer})");
		try {
		    functions.glBindBufferBase().invokeExact(target, index, buffer);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBindBufferBaseEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBindBufferBaseEXT.xhtml">registry.khronos.org</a>
	 */
	public void glBindBufferBaseEXT(GLBufferTarget target, int index, int buffer) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBindBufferBaseEXT(\{target}, \{index}, \{buffer})");
		try {
		    functions.glBindBufferBaseEXT().invokeExact(target, index, buffer);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBindBufferBaseNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBindBufferBaseNV.xhtml">registry.khronos.org</a>
	 */
	public void glBindBufferBaseNV(GLBufferTarget target, int index, int buffer) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBindBufferBaseNV(\{target}, \{index}, \{buffer})");
		try {
		    functions.glBindBufferBaseNV().invokeExact(target, index, buffer);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBindBufferOffsetEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBindBufferOffsetEXT.xhtml">registry.khronos.org</a>
	 */
	public void glBindBufferOffsetEXT(GLBufferTarget target, int index, int buffer, long offset) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBindBufferOffsetEXT(\{target}, \{index}, \{buffer}, \{offset})");
		try {
		    functions.glBindBufferOffsetEXT().invokeExact(target, index, buffer, offset);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBindBufferOffsetNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBindBufferOffsetNV.xhtml">registry.khronos.org</a>
	 */
	public void glBindBufferOffsetNV(GLBufferTarget target, int index, int buffer, long offset) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBindBufferOffsetNV(\{target}, \{index}, \{buffer}, \{offset})");
		try {
		    functions.glBindBufferOffsetNV().invokeExact(target, index, buffer, offset);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBindBufferRange">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBindBufferRange.xhtml">registry.khronos.org</a>
	 */
	public void glBindBufferRange(GLBufferTarget target, int index, int buffer, long offset, long size) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBindBufferRange(\{target}, \{index}, \{buffer}, \{offset}, \{size})");
		try {
		    functions.glBindBufferRange().invokeExact(target, index, buffer, offset, size);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBindBufferRangeEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBindBufferRangeEXT.xhtml">registry.khronos.org</a>
	 */
	public void glBindBufferRangeEXT(GLBufferTarget target, int index, int buffer, long offset, long size) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBindBufferRangeEXT(\{target}, \{index}, \{buffer}, \{offset}, \{size})");
		try {
		    functions.glBindBufferRangeEXT().invokeExact(target, index, buffer, offset, size);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBindBufferRangeNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBindBufferRangeNV.xhtml">registry.khronos.org</a>
	 */
	public void glBindBufferRangeNV(GLBufferTarget target, int index, int buffer, long offset, long size) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBindBufferRangeNV(\{target}, \{index}, \{buffer}, \{offset}, \{size})");
		try {
		    functions.glBindBufferRangeNV().invokeExact(target, index, buffer, offset, size);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBindBuffersBase">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBindBuffersBase.xhtml">registry.khronos.org</a>
	 */
	public void glBindBuffersBase(GLBufferTarget target, int first, int count, ConstPointer<Integer> buffers) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBindBuffersBase(\{target}, \{first}, \{count}, \{buffers})");
		try {
		    functions.glBindBuffersBase().invokeExact(target, first, count, buffers);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBindBuffersRange">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBindBuffersRange.xhtml">registry.khronos.org</a>
	 */
	public void glBindBuffersRange(GLBufferTarget target, int first, int count, ConstPointer<Integer> buffers, ConstPointer<Long> offsets, ConstPointer<Long> sizes) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBindBuffersRange(\{target}, \{first}, \{count}, \{buffers}, \{offsets}, \{sizes})");
		try {
		    functions.glBindBuffersRange().invokeExact(target, first, count, buffers, offsets, sizes);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBindFragDataLocation">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBindFragDataLocation.xhtml">registry.khronos.org</a>
	 */
	public void glBindFragDataLocation(int program, int color, ConstPointer<Byte> name) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBindFragDataLocation(\{program}, \{color}, \{name})");
		try {
		    functions.glBindFragDataLocation().invokeExact(program, color, name);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBindFragDataLocationEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBindFragDataLocationEXT.xhtml">registry.khronos.org</a>
	 */
	public void glBindFragDataLocationEXT(int program, int color, ConstPointer<Byte> name) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBindFragDataLocationEXT(\{program}, \{color}, \{name})");
		try {
		    functions.glBindFragDataLocationEXT().invokeExact(program, color, name);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBindFragDataLocationIndexed">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBindFragDataLocationIndexed.xhtml">registry.khronos.org</a>
	 */
	public void glBindFragDataLocationIndexed(int program, int colorNumber, int index, ConstPointer<Byte> name) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBindFragDataLocationIndexed(\{program}, \{colorNumber}, \{index}, \{name})");
		try {
		    functions.glBindFragDataLocationIndexed().invokeExact(program, colorNumber, index, name);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBindFragDataLocationIndexedEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBindFragDataLocationIndexedEXT.xhtml">registry.khronos.org</a>
	 */
	public void glBindFragDataLocationIndexedEXT(int program, int colorNumber, int index, ConstPointer<Byte> name) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBindFragDataLocationIndexedEXT(\{program}, \{colorNumber}, \{index}, \{name})");
		try {
		    functions.glBindFragDataLocationIndexedEXT().invokeExact(program, colorNumber, index, name);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBindFragmentShaderATI">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBindFragmentShaderATI.xhtml">registry.khronos.org</a>
	 */
	public void glBindFragmentShaderATI(int id) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBindFragmentShaderATI(\{id})");
		try {
		    functions.glBindFragmentShaderATI().invokeExact(id);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBindFramebuffer">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBindFramebuffer.xhtml">registry.khronos.org</a>
	 */
	public void glBindFramebuffer(GLFramebufferTarget target, int framebuffer) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBindFramebuffer(\{target}, \{framebuffer})");
		try {
		    functions.glBindFramebuffer().invokeExact(target, framebuffer);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBindFramebufferEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBindFramebufferEXT.xhtml">registry.khronos.org</a>
	 */
	public void glBindFramebufferEXT(GLFramebufferTarget target, int framebuffer) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBindFramebufferEXT(\{target}, \{framebuffer})");
		try {
		    functions.glBindFramebufferEXT().invokeExact(target, framebuffer);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBindFramebufferOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBindFramebufferOES.xhtml">registry.khronos.org</a>
	 */
	public void glBindFramebufferOES(GLFramebufferTarget target, int framebuffer) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBindFramebufferOES(\{target}, \{framebuffer})");
		try {
		    functions.glBindFramebufferOES().invokeExact(target, framebuffer);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBindImageTexture">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBindImageTexture.xhtml">registry.khronos.org</a>
	 */
	public void glBindImageTexture(int unit, int texture, int level, boolean layered, int layer, GLBufferAccess access, GLInternalFormat format) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBindImageTexture(\{unit}, \{texture}, \{level}, \{layered}, \{layer}, \{access}, \{format})");
		try {
		    functions.glBindImageTexture().invokeExact(unit, texture, level, layered, layer, access, format);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were access=\{access}, format=\{format}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBindImageTextureEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBindImageTextureEXT.xhtml">registry.khronos.org</a>
	 */
	public void glBindImageTextureEXT(int index, int texture, int level, boolean layered, int layer, GLBufferAccess access, int format) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBindImageTextureEXT(\{index}, \{texture}, \{level}, \{layered}, \{layer}, \{access}, \{format})");
		try {
		    functions.glBindImageTextureEXT().invokeExact(index, texture, level, layered, layer, access, format);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were access=\{access}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBindImageTextures">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBindImageTextures.xhtml">registry.khronos.org</a>
	 */
	public void glBindImageTextures(int first, int count, ConstPointer<Integer> textures) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBindImageTextures(\{first}, \{count}, \{textures})");
		try {
		    functions.glBindImageTextures().invokeExact(first, count, textures);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBindLightParameterEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBindLightParameterEXT.xhtml">registry.khronos.org</a>
	 */
	public int glBindLightParameterEXT(GLLightName light, GLLightParameter value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBindLightParameterEXT(\{light}, \{value})");
		try {
		    var r = (int)functions.glBindLightParameterEXT().invokeExact(light, value);
			checkError();
			return r;
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were light=\{light}, value=\{value}.");
			}
			throw e;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBindMaterialParameterEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBindMaterialParameterEXT.xhtml">registry.khronos.org</a>
	 */
	public int glBindMaterialParameterEXT(GLTriangleFace face, GLMaterialParameter value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBindMaterialParameterEXT(\{face}, \{value})");
		try {
		    var r = (int)functions.glBindMaterialParameterEXT().invokeExact(face, value);
			checkError();
			return r;
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were face=\{face}, value=\{value}.");
			}
			throw e;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBindMultiTextureEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBindMultiTextureEXT.xhtml">registry.khronos.org</a>
	 */
	public void glBindMultiTextureEXT(GLTextureUnit texunit, GLTextureTarget target, int texture) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBindMultiTextureEXT(\{texunit}, \{target}, \{texture})");
		try {
		    functions.glBindMultiTextureEXT().invokeExact(texunit, target, texture);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were texunit=\{texunit}, target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBindParameterEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBindParameterEXT.xhtml">registry.khronos.org</a>
	 */
	public int glBindParameterEXT(GLVertexShaderParameterEXT value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBindParameterEXT(\{value})");
		try {
		    var r = (int)functions.glBindParameterEXT().invokeExact(value);
			checkError();
			return r;
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were value=\{value}.");
			}
			throw e;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBindProgramARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBindProgramARB.xhtml">registry.khronos.org</a>
	 */
	public void glBindProgramARB(GLProgramTarget target, int program) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBindProgramARB(\{target}, \{program})");
		try {
		    functions.glBindProgramARB().invokeExact(target, program);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBindProgramNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBindProgramNV.xhtml">registry.khronos.org</a>
	 */
	public void glBindProgramNV(GLVertexAttribEnumNV target, int id) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBindProgramNV(\{target}, \{id})");
		try {
		    functions.glBindProgramNV().invokeExact(target, id);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBindProgramPipeline">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBindProgramPipeline.xhtml">registry.khronos.org</a>
	 */
	public void glBindProgramPipeline(int pipeline) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBindProgramPipeline(\{pipeline})");
		try {
		    functions.glBindProgramPipeline().invokeExact(pipeline);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBindProgramPipelineEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBindProgramPipelineEXT.xhtml">registry.khronos.org</a>
	 */
	public void glBindProgramPipelineEXT(int pipeline) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBindProgramPipelineEXT(\{pipeline})");
		try {
		    functions.glBindProgramPipelineEXT().invokeExact(pipeline);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBindRenderbuffer">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBindRenderbuffer.xhtml">registry.khronos.org</a>
	 */
	public void glBindRenderbuffer(GLRenderbufferTarget target, int renderbuffer) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBindRenderbuffer(\{target}, \{renderbuffer})");
		try {
		    functions.glBindRenderbuffer().invokeExact(target, renderbuffer);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBindRenderbufferEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBindRenderbufferEXT.xhtml">registry.khronos.org</a>
	 */
	public void glBindRenderbufferEXT(GLRenderbufferTarget target, int renderbuffer) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBindRenderbufferEXT(\{target}, \{renderbuffer})");
		try {
		    functions.glBindRenderbufferEXT().invokeExact(target, renderbuffer);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBindRenderbufferOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBindRenderbufferOES.xhtml">registry.khronos.org</a>
	 */
	public void glBindRenderbufferOES(GLRenderbufferTarget target, int renderbuffer) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBindRenderbufferOES(\{target}, \{renderbuffer})");
		try {
		    functions.glBindRenderbufferOES().invokeExact(target, renderbuffer);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBindSampler">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBindSampler.xhtml">registry.khronos.org</a>
	 */
	public void glBindSampler(int unit, int sampler) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBindSampler(\{unit}, \{sampler})");
		try {
		    functions.glBindSampler().invokeExact(unit, sampler);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBindSamplers">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBindSamplers.xhtml">registry.khronos.org</a>
	 */
	public void glBindSamplers(int first, int count, ConstPointer<Integer> samplers) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBindSamplers(\{first}, \{count}, \{samplers})");
		try {
		    functions.glBindSamplers().invokeExact(first, count, samplers);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBindShadingRateImageNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBindShadingRateImageNV.xhtml">registry.khronos.org</a>
	 */
	public void glBindShadingRateImageNV(int texture) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBindShadingRateImageNV(\{texture})");
		try {
		    functions.glBindShadingRateImageNV().invokeExact(texture);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBindTexGenParameterEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBindTexGenParameterEXT.xhtml">registry.khronos.org</a>
	 */
	public int glBindTexGenParameterEXT(GLTextureUnit unit, GLTextureCoordName coord, GLTextureGenParameter value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBindTexGenParameterEXT(\{unit}, \{coord}, \{value})");
		try {
		    var r = (int)functions.glBindTexGenParameterEXT().invokeExact(unit, coord, value);
			checkError();
			return r;
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were unit=\{unit}, coord=\{coord}, value=\{value}.");
			}
			throw e;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBindTexture">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBindTexture.xhtml">registry.khronos.org</a>
	 */
	public void glBindTexture(GLTextureTarget target, int texture) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBindTexture(\{target}, \{texture})");
		try {
		    functions.glBindTexture().invokeExact(target, texture);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBindTextureEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBindTextureEXT.xhtml">registry.khronos.org</a>
	 */
	public void glBindTextureEXT(GLTextureTarget target, int texture) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBindTextureEXT(\{target}, \{texture})");
		try {
		    functions.glBindTextureEXT().invokeExact(target, texture);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBindTextureUnit">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBindTextureUnit.xhtml">registry.khronos.org</a>
	 */
	public void glBindTextureUnit(int unit, int texture) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBindTextureUnit(\{unit}, \{texture})");
		try {
		    functions.glBindTextureUnit().invokeExact(unit, texture);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBindTextureUnitParameterEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBindTextureUnitParameterEXT.xhtml">registry.khronos.org</a>
	 */
	public int glBindTextureUnitParameterEXT(GLTextureUnit unit, GLVertexShaderTextureUnitParameter value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBindTextureUnitParameterEXT(\{unit}, \{value})");
		try {
		    var r = (int)functions.glBindTextureUnitParameterEXT().invokeExact(unit, value);
			checkError();
			return r;
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were unit=\{unit}, value=\{value}.");
			}
			throw e;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBindTextures">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBindTextures.xhtml">registry.khronos.org</a>
	 */
	public void glBindTextures(int first, int count, ConstPointer<Integer> textures) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBindTextures(\{first}, \{count}, \{textures})");
		try {
		    functions.glBindTextures().invokeExact(first, count, textures);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBindTransformFeedback">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBindTransformFeedback.xhtml">registry.khronos.org</a>
	 */
	public void glBindTransformFeedback(GLBindTransformFeedbackTarget target, int id) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBindTransformFeedback(\{target}, \{id})");
		try {
		    functions.glBindTransformFeedback().invokeExact(target, id);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBindTransformFeedbackNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBindTransformFeedbackNV.xhtml">registry.khronos.org</a>
	 */
	public void glBindTransformFeedbackNV(GLBufferTarget target, int id) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBindTransformFeedbackNV(\{target}, \{id})");
		try {
		    functions.glBindTransformFeedbackNV().invokeExact(target, id);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBindVertexArray">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBindVertexArray.xhtml">registry.khronos.org</a>
	 */
	public void glBindVertexArray(int array) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBindVertexArray(\{array})");
		try {
		    functions.glBindVertexArray().invokeExact(array);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBindVertexArrayAPPLE">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBindVertexArrayAPPLE.xhtml">registry.khronos.org</a>
	 */
	public void glBindVertexArrayAPPLE(int array) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBindVertexArrayAPPLE(\{array})");
		try {
		    functions.glBindVertexArrayAPPLE().invokeExact(array);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBindVertexArrayOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBindVertexArrayOES.xhtml">registry.khronos.org</a>
	 */
	public void glBindVertexArrayOES(int array) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBindVertexArrayOES(\{array})");
		try {
		    functions.glBindVertexArrayOES().invokeExact(array);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBindVertexBuffer">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBindVertexBuffer.xhtml">registry.khronos.org</a>
	 */
	public void glBindVertexBuffer(int bindingindex, int buffer, long offset, int stride) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBindVertexBuffer(\{bindingindex}, \{buffer}, \{offset}, \{stride})");
		try {
		    functions.glBindVertexBuffer().invokeExact(bindingindex, buffer, offset, stride);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBindVertexBuffers">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBindVertexBuffers.xhtml">registry.khronos.org</a>
	 */
	public void glBindVertexBuffers(int first, int count, ConstPointer<Integer> buffers, ConstPointer<Long> offsets, ConstPointer<Integer> strides) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBindVertexBuffers(\{first}, \{count}, \{buffers}, \{offsets}, \{strides})");
		try {
		    functions.glBindVertexBuffers().invokeExact(first, count, buffers, offsets, strides);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBindVertexShaderEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBindVertexShaderEXT.xhtml">registry.khronos.org</a>
	 */
	public void glBindVertexShaderEXT(int id) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBindVertexShaderEXT(\{id})");
		try {
		    functions.glBindVertexShaderEXT().invokeExact(id);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBindVideoCaptureStreamBufferNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBindVideoCaptureStreamBufferNV.xhtml">registry.khronos.org</a>
	 */
	public void glBindVideoCaptureStreamBufferNV(int video_capture_slot, int stream, GLEnum frame_region, long offset) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBindVideoCaptureStreamBufferNV(\{video_capture_slot}, \{stream}, \{frame_region}, \{offset})");
		try {
		    functions.glBindVideoCaptureStreamBufferNV().invokeExact(video_capture_slot, stream, frame_region, offset);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were frame_region=\{frame_region}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBindVideoCaptureStreamTextureNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBindVideoCaptureStreamTextureNV.xhtml">registry.khronos.org</a>
	 */
	public void glBindVideoCaptureStreamTextureNV(int video_capture_slot, int stream, GLEnum frame_region, GLEnum target, int texture) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBindVideoCaptureStreamTextureNV(\{video_capture_slot}, \{stream}, \{frame_region}, \{target}, \{texture})");
		try {
		    functions.glBindVideoCaptureStreamTextureNV().invokeExact(video_capture_slot, stream, frame_region, target, texture);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were frame_region=\{frame_region}, target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBinormal3bEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBinormal3bEXT.xhtml">registry.khronos.org</a>
	 */
	public void glBinormal3bEXT(byte bx, byte by, byte bz) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBinormal3bEXT(\{bx}, \{by}, \{bz})");
		try {
		    functions.glBinormal3bEXT().invokeExact(bx, by, bz);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBinormal3bvEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBinormal3bvEXT.xhtml">registry.khronos.org</a>
	 */
	public void glBinormal3bvEXT(ConstPointer<Byte> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBinormal3bvEXT(\{v})");
		try {
		    functions.glBinormal3bvEXT().invokeExact(v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBinormal3dEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBinormal3dEXT.xhtml">registry.khronos.org</a>
	 */
	public void glBinormal3dEXT(double bx, double by, double bz) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBinormal3dEXT(\{bx}, \{by}, \{bz})");
		try {
		    functions.glBinormal3dEXT().invokeExact(bx, by, bz);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBinormal3dvEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBinormal3dvEXT.xhtml">registry.khronos.org</a>
	 */
	public void glBinormal3dvEXT(ConstPointer<Double> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBinormal3dvEXT(\{v})");
		try {
		    functions.glBinormal3dvEXT().invokeExact(v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBinormal3fEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBinormal3fEXT.xhtml">registry.khronos.org</a>
	 */
	public void glBinormal3fEXT(float bx, float by, float bz) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBinormal3fEXT(\{bx}, \{by}, \{bz})");
		try {
		    functions.glBinormal3fEXT().invokeExact(bx, by, bz);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBinormal3fvEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBinormal3fvEXT.xhtml">registry.khronos.org</a>
	 */
	public void glBinormal3fvEXT(ConstPointer<Float> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBinormal3fvEXT(\{v})");
		try {
		    functions.glBinormal3fvEXT().invokeExact(v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBinormal3iEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBinormal3iEXT.xhtml">registry.khronos.org</a>
	 */
	public void glBinormal3iEXT(int bx, int by, int bz) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBinormal3iEXT(\{bx}, \{by}, \{bz})");
		try {
		    functions.glBinormal3iEXT().invokeExact(bx, by, bz);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBinormal3ivEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBinormal3ivEXT.xhtml">registry.khronos.org</a>
	 */
	public void glBinormal3ivEXT(ConstPointer<Integer> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBinormal3ivEXT(\{v})");
		try {
		    functions.glBinormal3ivEXT().invokeExact(v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBinormal3sEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBinormal3sEXT.xhtml">registry.khronos.org</a>
	 */
	public void glBinormal3sEXT(short bx, short by, short bz) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBinormal3sEXT(\{bx}, \{by}, \{bz})");
		try {
		    functions.glBinormal3sEXT().invokeExact(bx, by, bz);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBinormal3svEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBinormal3svEXT.xhtml">registry.khronos.org</a>
	 */
	public void glBinormal3svEXT(ConstPointer<Short> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBinormal3svEXT(\{v})");
		try {
		    functions.glBinormal3svEXT().invokeExact(v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBinormalPointerEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBinormalPointerEXT.xhtml">registry.khronos.org</a>
	 */
	public void glBinormalPointerEXT(GLBinormalPointerTypeEXT type, int stride, ConstPointer<Void> pointer) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBinormalPointerEXT(\{type}, \{stride}, \{pointer})");
		try {
		    functions.glBinormalPointerEXT().invokeExact(type, stride, pointer);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBitmap">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBitmap.xhtml">registry.khronos.org</a>
	 */
	public void glBitmap(int width, int height, float xorig, float yorig, float xmove, float ymove, ConstPointer<Byte> bitmap) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBitmap(\{width}, \{height}, \{xorig}, \{yorig}, \{xmove}, \{ymove}, \{bitmap})");
		try {
		    functions.glBitmap().invokeExact(width, height, xorig, yorig, xmove, ymove, bitmap);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBitmapxOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBitmapxOES.xhtml">registry.khronos.org</a>
	 */
	public void glBitmapxOES(int width, int height, int xorig, int yorig, int xmove, int ymove, ConstPointer<Byte> bitmap) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBitmapxOES(\{width}, \{height}, \{xorig}, \{yorig}, \{xmove}, \{ymove}, \{bitmap})");
		try {
		    functions.glBitmapxOES().invokeExact(width, height, xorig, yorig, xmove, ymove, bitmap);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBlendBarrier">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBlendBarrier.xhtml">registry.khronos.org</a>
	 */
	public void glBlendBarrier() {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBlendBarrier()");
		try {
		    functions.glBlendBarrier().invokeExact();
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBlendBarrierKHR">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBlendBarrierKHR.xhtml">registry.khronos.org</a>
	 */
	public void glBlendBarrierKHR() {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBlendBarrierKHR()");
		try {
		    functions.glBlendBarrierKHR().invokeExact();
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBlendBarrierNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBlendBarrierNV.xhtml">registry.khronos.org</a>
	 */
	public void glBlendBarrierNV() {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBlendBarrierNV()");
		try {
		    functions.glBlendBarrierNV().invokeExact();
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBlendColor">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBlendColor.xhtml">registry.khronos.org</a>
	 */
	public void glBlendColor(float red, float green, float blue, float alpha) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBlendColor(\{red}, \{green}, \{blue}, \{alpha})");
		try {
		    functions.glBlendColor().invokeExact(red, green, blue, alpha);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBlendColorEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBlendColorEXT.xhtml">registry.khronos.org</a>
	 */
	public void glBlendColorEXT(float red, float green, float blue, float alpha) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBlendColorEXT(\{red}, \{green}, \{blue}, \{alpha})");
		try {
		    functions.glBlendColorEXT().invokeExact(red, green, blue, alpha);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBlendColorxOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBlendColorxOES.xhtml">registry.khronos.org</a>
	 */
	public void glBlendColorxOES(int red, int green, int blue, int alpha) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBlendColorxOES(\{red}, \{green}, \{blue}, \{alpha})");
		try {
		    functions.glBlendColorxOES().invokeExact(red, green, blue, alpha);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBlendEquation">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBlendEquation.xhtml">registry.khronos.org</a>
	 */
	public void glBlendEquation(GLBlendEquationModeEXT mode) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBlendEquation(\{mode})");
		try {
		    functions.glBlendEquation().invokeExact(mode);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBlendEquationEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBlendEquationEXT.xhtml">registry.khronos.org</a>
	 */
	public void glBlendEquationEXT(GLBlendEquationModeEXT mode) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBlendEquationEXT(\{mode})");
		try {
		    functions.glBlendEquationEXT().invokeExact(mode);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBlendEquationIndexedAMD">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBlendEquationIndexedAMD.xhtml">registry.khronos.org</a>
	 */
	public void glBlendEquationIndexedAMD(int buf, GLBlendEquationModeEXT mode) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBlendEquationIndexedAMD(\{buf}, \{mode})");
		try {
		    functions.glBlendEquationIndexedAMD().invokeExact(buf, mode);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBlendEquationOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBlendEquationOES.xhtml">registry.khronos.org</a>
	 */
	public void glBlendEquationOES(GLBlendEquationModeEXT mode) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBlendEquationOES(\{mode})");
		try {
		    functions.glBlendEquationOES().invokeExact(mode);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBlendEquationSeparate">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBlendEquationSeparate.xhtml">registry.khronos.org</a>
	 */
	public void glBlendEquationSeparate(GLBlendEquationModeEXT modeRGB, GLBlendEquationModeEXT modeAlpha) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBlendEquationSeparate(\{modeRGB}, \{modeAlpha})");
		try {
		    functions.glBlendEquationSeparate().invokeExact(modeRGB, modeAlpha);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were modeRGB=\{modeRGB}, modeAlpha=\{modeAlpha}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBlendEquationSeparateEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBlendEquationSeparateEXT.xhtml">registry.khronos.org</a>
	 */
	public void glBlendEquationSeparateEXT(GLBlendEquationModeEXT modeRGB, GLBlendEquationModeEXT modeAlpha) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBlendEquationSeparateEXT(\{modeRGB}, \{modeAlpha})");
		try {
		    functions.glBlendEquationSeparateEXT().invokeExact(modeRGB, modeAlpha);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were modeRGB=\{modeRGB}, modeAlpha=\{modeAlpha}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBlendEquationSeparateIndexedAMD">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBlendEquationSeparateIndexedAMD.xhtml">registry.khronos.org</a>
	 */
	public void glBlendEquationSeparateIndexedAMD(int buf, GLBlendEquationModeEXT modeRGB, GLBlendEquationModeEXT modeAlpha) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBlendEquationSeparateIndexedAMD(\{buf}, \{modeRGB}, \{modeAlpha})");
		try {
		    functions.glBlendEquationSeparateIndexedAMD().invokeExact(buf, modeRGB, modeAlpha);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were modeRGB=\{modeRGB}, modeAlpha=\{modeAlpha}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBlendEquationSeparateOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBlendEquationSeparateOES.xhtml">registry.khronos.org</a>
	 */
	public void glBlendEquationSeparateOES(GLBlendEquationModeEXT modeRGB, GLBlendEquationModeEXT modeAlpha) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBlendEquationSeparateOES(\{modeRGB}, \{modeAlpha})");
		try {
		    functions.glBlendEquationSeparateOES().invokeExact(modeRGB, modeAlpha);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were modeRGB=\{modeRGB}, modeAlpha=\{modeAlpha}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBlendEquationSeparatei">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBlendEquationSeparatei.xhtml">registry.khronos.org</a>
	 */
	public void glBlendEquationSeparatei(int buf, GLBlendEquationModeEXT modeRGB, GLBlendEquationModeEXT modeAlpha) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBlendEquationSeparatei(\{buf}, \{modeRGB}, \{modeAlpha})");
		try {
		    functions.glBlendEquationSeparatei().invokeExact(buf, modeRGB, modeAlpha);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were modeRGB=\{modeRGB}, modeAlpha=\{modeAlpha}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBlendEquationSeparateiARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBlendEquationSeparateiARB.xhtml">registry.khronos.org</a>
	 */
	public void glBlendEquationSeparateiARB(int buf, GLBlendEquationModeEXT modeRGB, GLBlendEquationModeEXT modeAlpha) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBlendEquationSeparateiARB(\{buf}, \{modeRGB}, \{modeAlpha})");
		try {
		    functions.glBlendEquationSeparateiARB().invokeExact(buf, modeRGB, modeAlpha);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were modeRGB=\{modeRGB}, modeAlpha=\{modeAlpha}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBlendEquationSeparateiEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBlendEquationSeparateiEXT.xhtml">registry.khronos.org</a>
	 */
	public void glBlendEquationSeparateiEXT(int buf, GLBlendEquationModeEXT modeRGB, GLBlendEquationModeEXT modeAlpha) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBlendEquationSeparateiEXT(\{buf}, \{modeRGB}, \{modeAlpha})");
		try {
		    functions.glBlendEquationSeparateiEXT().invokeExact(buf, modeRGB, modeAlpha);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were modeRGB=\{modeRGB}, modeAlpha=\{modeAlpha}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBlendEquationSeparateiOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBlendEquationSeparateiOES.xhtml">registry.khronos.org</a>
	 */
	public void glBlendEquationSeparateiOES(int buf, GLBlendEquationModeEXT modeRGB, GLBlendEquationModeEXT modeAlpha) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBlendEquationSeparateiOES(\{buf}, \{modeRGB}, \{modeAlpha})");
		try {
		    functions.glBlendEquationSeparateiOES().invokeExact(buf, modeRGB, modeAlpha);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were modeRGB=\{modeRGB}, modeAlpha=\{modeAlpha}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBlendEquationi">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBlendEquationi.xhtml">registry.khronos.org</a>
	 */
	public void glBlendEquationi(int buf, GLBlendEquationModeEXT mode) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBlendEquationi(\{buf}, \{mode})");
		try {
		    functions.glBlendEquationi().invokeExact(buf, mode);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBlendEquationiARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBlendEquationiARB.xhtml">registry.khronos.org</a>
	 */
	public void glBlendEquationiARB(int buf, GLBlendEquationModeEXT mode) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBlendEquationiARB(\{buf}, \{mode})");
		try {
		    functions.glBlendEquationiARB().invokeExact(buf, mode);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBlendEquationiEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBlendEquationiEXT.xhtml">registry.khronos.org</a>
	 */
	public void glBlendEquationiEXT(int buf, GLBlendEquationModeEXT mode) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBlendEquationiEXT(\{buf}, \{mode})");
		try {
		    functions.glBlendEquationiEXT().invokeExact(buf, mode);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBlendEquationiOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBlendEquationiOES.xhtml">registry.khronos.org</a>
	 */
	public void glBlendEquationiOES(int buf, GLBlendEquationModeEXT mode) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBlendEquationiOES(\{buf}, \{mode})");
		try {
		    functions.glBlendEquationiOES().invokeExact(buf, mode);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBlendFunc">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBlendFunc.xhtml">registry.khronos.org</a>
	 */
	public void glBlendFunc(GLBlendingFactor sfactor, GLBlendingFactor dfactor) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBlendFunc(\{sfactor}, \{dfactor})");
		try {
		    functions.glBlendFunc().invokeExact(sfactor, dfactor);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were sfactor=\{sfactor}, dfactor=\{dfactor}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBlendFuncIndexedAMD">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBlendFuncIndexedAMD.xhtml">registry.khronos.org</a>
	 */
	public void glBlendFuncIndexedAMD(int buf, GLEnum src, GLEnum dst) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBlendFuncIndexedAMD(\{buf}, \{src}, \{dst})");
		try {
		    functions.glBlendFuncIndexedAMD().invokeExact(buf, src, dst);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were src=\{src}, dst=\{dst}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBlendFuncSeparate">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBlendFuncSeparate.xhtml">registry.khronos.org</a>
	 */
	public void glBlendFuncSeparate(GLBlendingFactor sfactorRGB, GLBlendingFactor dfactorRGB, GLBlendingFactor sfactorAlpha, GLBlendingFactor dfactorAlpha) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBlendFuncSeparate(\{sfactorRGB}, \{dfactorRGB}, \{sfactorAlpha}, \{dfactorAlpha})");
		try {
		    functions.glBlendFuncSeparate().invokeExact(sfactorRGB, dfactorRGB, sfactorAlpha, dfactorAlpha);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were sfactorRGB=\{sfactorRGB}, dfactorRGB=\{dfactorRGB}, sfactorAlpha=\{sfactorAlpha}, dfactorAlpha=\{dfactorAlpha}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBlendFuncSeparateEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBlendFuncSeparateEXT.xhtml">registry.khronos.org</a>
	 */
	public void glBlendFuncSeparateEXT(GLBlendingFactor sfactorRGB, GLBlendingFactor dfactorRGB, GLBlendingFactor sfactorAlpha, GLBlendingFactor dfactorAlpha) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBlendFuncSeparateEXT(\{sfactorRGB}, \{dfactorRGB}, \{sfactorAlpha}, \{dfactorAlpha})");
		try {
		    functions.glBlendFuncSeparateEXT().invokeExact(sfactorRGB, dfactorRGB, sfactorAlpha, dfactorAlpha);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were sfactorRGB=\{sfactorRGB}, dfactorRGB=\{dfactorRGB}, sfactorAlpha=\{sfactorAlpha}, dfactorAlpha=\{dfactorAlpha}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBlendFuncSeparateINGR">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBlendFuncSeparateINGR.xhtml">registry.khronos.org</a>
	 */
	public void glBlendFuncSeparateINGR(GLBlendingFactor sfactorRGB, GLBlendingFactor dfactorRGB, GLBlendingFactor sfactorAlpha, GLBlendingFactor dfactorAlpha) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBlendFuncSeparateINGR(\{sfactorRGB}, \{dfactorRGB}, \{sfactorAlpha}, \{dfactorAlpha})");
		try {
		    functions.glBlendFuncSeparateINGR().invokeExact(sfactorRGB, dfactorRGB, sfactorAlpha, dfactorAlpha);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were sfactorRGB=\{sfactorRGB}, dfactorRGB=\{dfactorRGB}, sfactorAlpha=\{sfactorAlpha}, dfactorAlpha=\{dfactorAlpha}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBlendFuncSeparateIndexedAMD">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBlendFuncSeparateIndexedAMD.xhtml">registry.khronos.org</a>
	 */
	public void glBlendFuncSeparateIndexedAMD(int buf, GLBlendingFactor srcRGB, GLBlendingFactor dstRGB, GLBlendingFactor srcAlpha, GLBlendingFactor dstAlpha) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBlendFuncSeparateIndexedAMD(\{buf}, \{srcRGB}, \{dstRGB}, \{srcAlpha}, \{dstAlpha})");
		try {
		    functions.glBlendFuncSeparateIndexedAMD().invokeExact(buf, srcRGB, dstRGB, srcAlpha, dstAlpha);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were srcRGB=\{srcRGB}, dstRGB=\{dstRGB}, srcAlpha=\{srcAlpha}, dstAlpha=\{dstAlpha}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBlendFuncSeparateOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBlendFuncSeparateOES.xhtml">registry.khronos.org</a>
	 */
	public void glBlendFuncSeparateOES(GLBlendingFactor srcRGB, GLBlendingFactor dstRGB, GLBlendingFactor srcAlpha, GLBlendingFactor dstAlpha) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBlendFuncSeparateOES(\{srcRGB}, \{dstRGB}, \{srcAlpha}, \{dstAlpha})");
		try {
		    functions.glBlendFuncSeparateOES().invokeExact(srcRGB, dstRGB, srcAlpha, dstAlpha);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were srcRGB=\{srcRGB}, dstRGB=\{dstRGB}, srcAlpha=\{srcAlpha}, dstAlpha=\{dstAlpha}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBlendFuncSeparatei">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBlendFuncSeparatei.xhtml">registry.khronos.org</a>
	 */
	public void glBlendFuncSeparatei(int buf, GLBlendingFactor srcRGB, GLBlendingFactor dstRGB, GLBlendingFactor srcAlpha, GLBlendingFactor dstAlpha) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBlendFuncSeparatei(\{buf}, \{srcRGB}, \{dstRGB}, \{srcAlpha}, \{dstAlpha})");
		try {
		    functions.glBlendFuncSeparatei().invokeExact(buf, srcRGB, dstRGB, srcAlpha, dstAlpha);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were srcRGB=\{srcRGB}, dstRGB=\{dstRGB}, srcAlpha=\{srcAlpha}, dstAlpha=\{dstAlpha}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBlendFuncSeparateiARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBlendFuncSeparateiARB.xhtml">registry.khronos.org</a>
	 */
	public void glBlendFuncSeparateiARB(int buf, GLBlendingFactor srcRGB, GLBlendingFactor dstRGB, GLBlendingFactor srcAlpha, GLBlendingFactor dstAlpha) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBlendFuncSeparateiARB(\{buf}, \{srcRGB}, \{dstRGB}, \{srcAlpha}, \{dstAlpha})");
		try {
		    functions.glBlendFuncSeparateiARB().invokeExact(buf, srcRGB, dstRGB, srcAlpha, dstAlpha);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were srcRGB=\{srcRGB}, dstRGB=\{dstRGB}, srcAlpha=\{srcAlpha}, dstAlpha=\{dstAlpha}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBlendFuncSeparateiEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBlendFuncSeparateiEXT.xhtml">registry.khronos.org</a>
	 */
	public void glBlendFuncSeparateiEXT(int buf, GLBlendingFactor srcRGB, GLBlendingFactor dstRGB, GLBlendingFactor srcAlpha, GLBlendingFactor dstAlpha) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBlendFuncSeparateiEXT(\{buf}, \{srcRGB}, \{dstRGB}, \{srcAlpha}, \{dstAlpha})");
		try {
		    functions.glBlendFuncSeparateiEXT().invokeExact(buf, srcRGB, dstRGB, srcAlpha, dstAlpha);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were srcRGB=\{srcRGB}, dstRGB=\{dstRGB}, srcAlpha=\{srcAlpha}, dstAlpha=\{dstAlpha}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBlendFuncSeparateiOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBlendFuncSeparateiOES.xhtml">registry.khronos.org</a>
	 */
	public void glBlendFuncSeparateiOES(int buf, GLBlendingFactor srcRGB, GLBlendingFactor dstRGB, GLBlendingFactor srcAlpha, GLBlendingFactor dstAlpha) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBlendFuncSeparateiOES(\{buf}, \{srcRGB}, \{dstRGB}, \{srcAlpha}, \{dstAlpha})");
		try {
		    functions.glBlendFuncSeparateiOES().invokeExact(buf, srcRGB, dstRGB, srcAlpha, dstAlpha);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were srcRGB=\{srcRGB}, dstRGB=\{dstRGB}, srcAlpha=\{srcAlpha}, dstAlpha=\{dstAlpha}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBlendFunci">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBlendFunci.xhtml">registry.khronos.org</a>
	 */
	public void glBlendFunci(int buf, GLBlendingFactor src, GLBlendingFactor dst) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBlendFunci(\{buf}, \{src}, \{dst})");
		try {
		    functions.glBlendFunci().invokeExact(buf, src, dst);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were src=\{src}, dst=\{dst}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBlendFunciARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBlendFunciARB.xhtml">registry.khronos.org</a>
	 */
	public void glBlendFunciARB(int buf, GLBlendingFactor src, GLBlendingFactor dst) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBlendFunciARB(\{buf}, \{src}, \{dst})");
		try {
		    functions.glBlendFunciARB().invokeExact(buf, src, dst);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were src=\{src}, dst=\{dst}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBlendFunciEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBlendFunciEXT.xhtml">registry.khronos.org</a>
	 */
	public void glBlendFunciEXT(int buf, GLBlendingFactor src, GLBlendingFactor dst) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBlendFunciEXT(\{buf}, \{src}, \{dst})");
		try {
		    functions.glBlendFunciEXT().invokeExact(buf, src, dst);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were src=\{src}, dst=\{dst}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBlendFunciOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBlendFunciOES.xhtml">registry.khronos.org</a>
	 */
	public void glBlendFunciOES(int buf, GLBlendingFactor src, GLBlendingFactor dst) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBlendFunciOES(\{buf}, \{src}, \{dst})");
		try {
		    functions.glBlendFunciOES().invokeExact(buf, src, dst);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were src=\{src}, dst=\{dst}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBlendParameteriNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBlendParameteriNV.xhtml">registry.khronos.org</a>
	 */
	public void glBlendParameteriNV(GLEnum pname, int value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBlendParameteriNV(\{pname}, \{value})");
		try {
		    functions.glBlendParameteriNV().invokeExact(pname, value);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBlitFramebuffer">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBlitFramebuffer.xhtml">registry.khronos.org</a>
	 */
	public void glBlitFramebuffer(int srcX0, int srcY0, int srcX1, int srcY1, int dstX0, int dstY0, int dstX1, int dstY1, int mask, GLBlitFramebufferFilter filter) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBlitFramebuffer(\{srcX0}, \{srcY0}, \{srcX1}, \{srcY1}, \{dstX0}, \{dstY0}, \{dstX1}, \{dstY1}, \{mask}, \{filter})");
		try {
		    functions.glBlitFramebuffer().invokeExact(srcX0, srcY0, srcX1, srcY1, dstX0, dstY0, dstX1, dstY1, mask, filter);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were filter=\{filter}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBlitFramebufferANGLE">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBlitFramebufferANGLE.xhtml">registry.khronos.org</a>
	 */
	public void glBlitFramebufferANGLE(int srcX0, int srcY0, int srcX1, int srcY1, int dstX0, int dstY0, int dstX1, int dstY1, int mask, GLBlitFramebufferFilter filter) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBlitFramebufferANGLE(\{srcX0}, \{srcY0}, \{srcX1}, \{srcY1}, \{dstX0}, \{dstY0}, \{dstX1}, \{dstY1}, \{mask}, \{filter})");
		try {
		    functions.glBlitFramebufferANGLE().invokeExact(srcX0, srcY0, srcX1, srcY1, dstX0, dstY0, dstX1, dstY1, mask, filter);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were filter=\{filter}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBlitFramebufferEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBlitFramebufferEXT.xhtml">registry.khronos.org</a>
	 */
	public void glBlitFramebufferEXT(int srcX0, int srcY0, int srcX1, int srcY1, int dstX0, int dstY0, int dstX1, int dstY1, int mask, GLBlitFramebufferFilter filter) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBlitFramebufferEXT(\{srcX0}, \{srcY0}, \{srcX1}, \{srcY1}, \{dstX0}, \{dstY0}, \{dstX1}, \{dstY1}, \{mask}, \{filter})");
		try {
		    functions.glBlitFramebufferEXT().invokeExact(srcX0, srcY0, srcX1, srcY1, dstX0, dstY0, dstX1, dstY1, mask, filter);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were filter=\{filter}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBlitFramebufferLayerEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBlitFramebufferLayerEXT.xhtml">registry.khronos.org</a>
	 */
	public void glBlitFramebufferLayerEXT(int srcX0, int srcY0, int srcX1, int srcY1, int srcLayer, int dstX0, int dstY0, int dstX1, int dstY1, int dstLayer, int mask, GLBlitFramebufferFilter filter) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBlitFramebufferLayerEXT(\{srcX0}, \{srcY0}, \{srcX1}, \{srcY1}, \{srcLayer}, \{dstX0}, \{dstY0}, \{dstX1}, \{dstY1}, \{dstLayer}, \{mask}, \{filter})");
		try {
		    functions.glBlitFramebufferLayerEXT().invokeExact(srcX0, srcY0, srcX1, srcY1, srcLayer, dstX0, dstY0, dstX1, dstY1, dstLayer, mask, filter);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were filter=\{filter}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBlitFramebufferLayersEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBlitFramebufferLayersEXT.xhtml">registry.khronos.org</a>
	 */
	public void glBlitFramebufferLayersEXT(int srcX0, int srcY0, int srcX1, int srcY1, int dstX0, int dstY0, int dstX1, int dstY1, int mask, GLBlitFramebufferFilter filter) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBlitFramebufferLayersEXT(\{srcX0}, \{srcY0}, \{srcX1}, \{srcY1}, \{dstX0}, \{dstY0}, \{dstX1}, \{dstY1}, \{mask}, \{filter})");
		try {
		    functions.glBlitFramebufferLayersEXT().invokeExact(srcX0, srcY0, srcX1, srcY1, dstX0, dstY0, dstX1, dstY1, mask, filter);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were filter=\{filter}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBlitFramebufferNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBlitFramebufferNV.xhtml">registry.khronos.org</a>
	 */
	public void glBlitFramebufferNV(int srcX0, int srcY0, int srcX1, int srcY1, int dstX0, int dstY0, int dstX1, int dstY1, int mask, GLBlitFramebufferFilter filter) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBlitFramebufferNV(\{srcX0}, \{srcY0}, \{srcX1}, \{srcY1}, \{dstX0}, \{dstY0}, \{dstX1}, \{dstY1}, \{mask}, \{filter})");
		try {
		    functions.glBlitFramebufferNV().invokeExact(srcX0, srcY0, srcX1, srcY1, dstX0, dstY0, dstX1, dstY1, mask, filter);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were filter=\{filter}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBlitNamedFramebuffer">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBlitNamedFramebuffer.xhtml">registry.khronos.org</a>
	 */
	public void glBlitNamedFramebuffer(int readFramebuffer, int drawFramebuffer, int srcX0, int srcY0, int srcX1, int srcY1, int dstX0, int dstY0, int dstX1, int dstY1, int mask, GLBlitFramebufferFilter filter) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBlitNamedFramebuffer(\{readFramebuffer}, \{drawFramebuffer}, \{srcX0}, \{srcY0}, \{srcX1}, \{srcY1}, \{dstX0}, \{dstY0}, \{dstX1}, \{dstY1}, \{mask}, \{filter})");
		try {
		    functions.glBlitNamedFramebuffer().invokeExact(readFramebuffer, drawFramebuffer, srcX0, srcY0, srcX1, srcY1, dstX0, dstY0, dstX1, dstY1, mask, filter);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were filter=\{filter}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBufferAddressRangeNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBufferAddressRangeNV.xhtml">registry.khronos.org</a>
	 */
	public void glBufferAddressRangeNV(GLEnum pname, int index, long address, long length) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBufferAddressRangeNV(\{pname}, \{index}, \{address}, \{length})");
		try {
		    functions.glBufferAddressRangeNV().invokeExact(pname, index, address, length);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBufferAttachMemoryNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBufferAttachMemoryNV.xhtml">registry.khronos.org</a>
	 */
	public void glBufferAttachMemoryNV(GLBufferTarget target, int memory, long offset) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBufferAttachMemoryNV(\{target}, \{memory}, \{offset})");
		try {
		    functions.glBufferAttachMemoryNV().invokeExact(target, memory, offset);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBufferData">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBufferData.xhtml">registry.khronos.org</a>
	 */
	public void glBufferData(GLBufferTarget target, long size, ConstPointer<Void> data, GLBufferUsage usage) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBufferData(\{target}, \{size}, \{data}, \{usage})");
		try {
		    functions.glBufferData().invokeExact(target, size, data, usage);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, usage=\{usage}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBufferDataARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBufferDataARB.xhtml">registry.khronos.org</a>
	 */
	public void glBufferDataARB(GLBufferTarget target, long size, ConstPointer<Void> data, GLBufferUsage usage) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBufferDataARB(\{target}, \{size}, \{data}, \{usage})");
		try {
		    functions.glBufferDataARB().invokeExact(target, size, data, usage);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, usage=\{usage}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBufferPageCommitmentARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBufferPageCommitmentARB.xhtml">registry.khronos.org</a>
	 */
	public void glBufferPageCommitmentARB(GLEnum target, long offset, long size, boolean commit) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBufferPageCommitmentARB(\{target}, \{offset}, \{size}, \{commit})");
		try {
		    functions.glBufferPageCommitmentARB().invokeExact(target, offset, size, commit);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBufferPageCommitmentMemNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBufferPageCommitmentMemNV.xhtml">registry.khronos.org</a>
	 */
	public void glBufferPageCommitmentMemNV(GLBufferStorageTarget target, long offset, long size, int memory, long memOffset, boolean commit) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBufferPageCommitmentMemNV(\{target}, \{offset}, \{size}, \{memory}, \{memOffset}, \{commit})");
		try {
		    functions.glBufferPageCommitmentMemNV().invokeExact(target, offset, size, memory, memOffset, commit);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBufferParameteriAPPLE">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBufferParameteriAPPLE.xhtml">registry.khronos.org</a>
	 */
	public void glBufferParameteriAPPLE(GLEnum target, GLEnum pname, int param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBufferParameteriAPPLE(\{target}, \{pname}, \{param})");
		try {
		    functions.glBufferParameteriAPPLE().invokeExact(target, pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBufferStorage">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBufferStorage.xhtml">registry.khronos.org</a>
	 */
	public void glBufferStorage(GLBufferStorageTarget target, long size, ConstPointer<Void> data, int flags) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBufferStorage(\{target}, \{size}, \{data}, \{flags})");
		try {
		    functions.glBufferStorage().invokeExact(target, size, data, flags);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBufferStorageEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBufferStorageEXT.xhtml">registry.khronos.org</a>
	 */
	public void glBufferStorageEXT(GLBufferStorageTarget target, long size, ConstPointer<Void> data, int flags) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBufferStorageEXT(\{target}, \{size}, \{data}, \{flags})");
		try {
		    functions.glBufferStorageEXT().invokeExact(target, size, data, flags);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBufferStorageExternalEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBufferStorageExternalEXT.xhtml">registry.khronos.org</a>
	 */
	public void glBufferStorageExternalEXT(GLEnum target, long offset, long size, int clientBuffer, int flags) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBufferStorageExternalEXT(\{target}, \{offset}, \{size}, \{clientBuffer}, \{flags})");
		try {
		    functions.glBufferStorageExternalEXT().invokeExact(target, offset, size, clientBuffer, flags);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBufferStorageMemEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBufferStorageMemEXT.xhtml">registry.khronos.org</a>
	 */
	public void glBufferStorageMemEXT(GLBufferTarget target, long size, int memory, long offset) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBufferStorageMemEXT(\{target}, \{size}, \{memory}, \{offset})");
		try {
		    functions.glBufferStorageMemEXT().invokeExact(target, size, memory, offset);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBufferSubData">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBufferSubData.xhtml">registry.khronos.org</a>
	 */
	public void glBufferSubData(GLBufferTarget target, long offset, long size, ConstPointer<Void> data) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBufferSubData(\{target}, \{offset}, \{size}, \{data})");
		try {
		    functions.glBufferSubData().invokeExact(target, offset, size, data);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glBufferSubDataARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glBufferSubDataARB.xhtml">registry.khronos.org</a>
	 */
	public void glBufferSubDataARB(GLBufferTarget target, long offset, long size, ConstPointer<Void> data) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glBufferSubDataARB(\{target}, \{offset}, \{size}, \{data})");
		try {
		    functions.glBufferSubDataARB().invokeExact(target, offset, size, data);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCallCommandListNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCallCommandListNV.xhtml">registry.khronos.org</a>
	 */
	public void glCallCommandListNV(int list) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCallCommandListNV(\{list})");
		try {
		    functions.glCallCommandListNV().invokeExact(list);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCallList">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCallList.xhtml">registry.khronos.org</a>
	 */
	public void glCallList(int list) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCallList(\{list})");
		try {
		    functions.glCallList().invokeExact(list);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCallLists">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCallLists.xhtml">registry.khronos.org</a>
	 */
	public void glCallLists(int n, GLListNameType type, ConstPointer<Void> lists) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCallLists(\{n}, \{type}, \{lists})");
		try {
		    functions.glCallLists().invokeExact(n, type, lists);
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCheckFramebufferStatus">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCheckFramebufferStatus.xhtml">registry.khronos.org</a>
	 */
	public GLFramebufferStatus glCheckFramebufferStatus(GLFramebufferTarget target) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCheckFramebufferStatus(\{target})");
		try {
		    var r = (GLFramebufferStatus)functions.glCheckFramebufferStatus().invokeExact(target);
			checkError();
			return r;
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCheckFramebufferStatusEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCheckFramebufferStatusEXT.xhtml">registry.khronos.org</a>
	 */
	public GLFramebufferStatus glCheckFramebufferStatusEXT(GLFramebufferTarget target) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCheckFramebufferStatusEXT(\{target})");
		try {
		    var r = (GLFramebufferStatus)functions.glCheckFramebufferStatusEXT().invokeExact(target);
			checkError();
			return r;
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCheckFramebufferStatusOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCheckFramebufferStatusOES.xhtml">registry.khronos.org</a>
	 */
	public GLFramebufferStatus glCheckFramebufferStatusOES(GLFramebufferTarget target) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCheckFramebufferStatusOES(\{target})");
		try {
		    var r = (GLFramebufferStatus)functions.glCheckFramebufferStatusOES().invokeExact(target);
			checkError();
			return r;
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCheckNamedFramebufferStatus">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCheckNamedFramebufferStatus.xhtml">registry.khronos.org</a>
	 */
	public GLFramebufferStatus glCheckNamedFramebufferStatus(int framebuffer, GLFramebufferTarget target) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCheckNamedFramebufferStatus(\{framebuffer}, \{target})");
		try {
		    var r = (GLFramebufferStatus)functions.glCheckNamedFramebufferStatus().invokeExact(framebuffer, target);
			checkError();
			return r;
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCheckNamedFramebufferStatusEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCheckNamedFramebufferStatusEXT.xhtml">registry.khronos.org</a>
	 */
	public GLFramebufferStatus glCheckNamedFramebufferStatusEXT(int framebuffer, GLFramebufferTarget target) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCheckNamedFramebufferStatusEXT(\{framebuffer}, \{target})");
		try {
		    var r = (GLFramebufferStatus)functions.glCheckNamedFramebufferStatusEXT().invokeExact(framebuffer, target);
			checkError();
			return r;
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glClampColor">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glClampColor.xhtml">registry.khronos.org</a>
	 */
	public void glClampColor(GLClampColorTargetARB target, GLClampColorModeARB clamp) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glClampColor(\{target}, \{clamp})");
		try {
		    functions.glClampColor().invokeExact(target, clamp);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, clamp=\{clamp}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glClampColorARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glClampColorARB.xhtml">registry.khronos.org</a>
	 */
	public void glClampColorARB(GLClampColorTargetARB target, GLClampColorModeARB clamp) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glClampColorARB(\{target}, \{clamp})");
		try {
		    functions.glClampColorARB().invokeExact(target, clamp);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, clamp=\{clamp}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glClear">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glClear.xhtml">registry.khronos.org</a>
	 */
	public void glClear(int mask) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glClear(\{mask})");
		try {
		    functions.glClear().invokeExact(mask);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glClearAccum">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glClearAccum.xhtml">registry.khronos.org</a>
	 */
	public void glClearAccum(float red, float green, float blue, float alpha) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glClearAccum(\{red}, \{green}, \{blue}, \{alpha})");
		try {
		    functions.glClearAccum().invokeExact(red, green, blue, alpha);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glClearAccumxOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glClearAccumxOES.xhtml">registry.khronos.org</a>
	 */
	public void glClearAccumxOES(int red, int green, int blue, int alpha) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glClearAccumxOES(\{red}, \{green}, \{blue}, \{alpha})");
		try {
		    functions.glClearAccumxOES().invokeExact(red, green, blue, alpha);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glClearBufferData">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glClearBufferData.xhtml">registry.khronos.org</a>
	 */
	public void glClearBufferData(GLBufferStorageTarget target, GLSizedInternalFormat internalformat, GLPixelFormat format, GLPixelType type, ConstPointer<Void> data) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glClearBufferData(\{target}, \{internalformat}, \{format}, \{type}, \{data})");
		try {
		    functions.glClearBufferData().invokeExact(target, internalformat, format, type, data);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, internalformat=\{internalformat}, format=\{format}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glClearBufferSubData">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glClearBufferSubData.xhtml">registry.khronos.org</a>
	 */
	public void glClearBufferSubData(GLBufferTarget target, GLSizedInternalFormat internalformat, long offset, long size, GLPixelFormat format, GLPixelType type, ConstPointer<Void> data) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glClearBufferSubData(\{target}, \{internalformat}, \{offset}, \{size}, \{format}, \{type}, \{data})");
		try {
		    functions.glClearBufferSubData().invokeExact(target, internalformat, offset, size, format, type, data);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, internalformat=\{internalformat}, format=\{format}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glClearBufferfi">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glClearBufferfi.xhtml">registry.khronos.org</a>
	 */
	public void glClearBufferfi(GLBuffer buffer, int drawbuffer, float depth, int stencil) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glClearBufferfi(\{buffer}, \{drawbuffer}, \{depth}, \{stencil})");
		try {
		    functions.glClearBufferfi().invokeExact(buffer, drawbuffer, depth, stencil);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were buffer=\{buffer}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glClearBufferfv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glClearBufferfv.xhtml">registry.khronos.org</a>
	 */
	public void glClearBufferfv(GLBuffer buffer, int drawbuffer, ConstPointer<Float> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glClearBufferfv(\{buffer}, \{drawbuffer}, \{value})");
		try {
		    functions.glClearBufferfv().invokeExact(buffer, drawbuffer, value);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were buffer=\{buffer}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glClearBufferiv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glClearBufferiv.xhtml">registry.khronos.org</a>
	 */
	public void glClearBufferiv(GLBuffer buffer, int drawbuffer, ConstPointer<Integer> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glClearBufferiv(\{buffer}, \{drawbuffer}, \{value})");
		try {
		    functions.glClearBufferiv().invokeExact(buffer, drawbuffer, value);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were buffer=\{buffer}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glClearBufferuiv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glClearBufferuiv.xhtml">registry.khronos.org</a>
	 */
	public void glClearBufferuiv(GLBuffer buffer, int drawbuffer, ConstPointer<Integer> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glClearBufferuiv(\{buffer}, \{drawbuffer}, \{value})");
		try {
		    functions.glClearBufferuiv().invokeExact(buffer, drawbuffer, value);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were buffer=\{buffer}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glClearColor">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glClearColor.xhtml">registry.khronos.org</a>
	 */
	public void glClearColor(float red, float green, float blue, float alpha) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glClearColor(\{red}, \{green}, \{blue}, \{alpha})");
		try {
		    functions.glClearColor().invokeExact(red, green, blue, alpha);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glClearColorIiEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glClearColorIiEXT.xhtml">registry.khronos.org</a>
	 */
	public void glClearColorIiEXT(int red, int green, int blue, int alpha) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glClearColorIiEXT(\{red}, \{green}, \{blue}, \{alpha})");
		try {
		    functions.glClearColorIiEXT().invokeExact(red, green, blue, alpha);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glClearColorIuiEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glClearColorIuiEXT.xhtml">registry.khronos.org</a>
	 */
	public void glClearColorIuiEXT(int red, int green, int blue, int alpha) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glClearColorIuiEXT(\{red}, \{green}, \{blue}, \{alpha})");
		try {
		    functions.glClearColorIuiEXT().invokeExact(red, green, blue, alpha);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glClearColorx">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glClearColorx.xhtml">registry.khronos.org</a>
	 */
	public void glClearColorx(int red, int green, int blue, int alpha) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glClearColorx(\{red}, \{green}, \{blue}, \{alpha})");
		try {
		    functions.glClearColorx().invokeExact(red, green, blue, alpha);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glClearColorxOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glClearColorxOES.xhtml">registry.khronos.org</a>
	 */
	public void glClearColorxOES(int red, int green, int blue, int alpha) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glClearColorxOES(\{red}, \{green}, \{blue}, \{alpha})");
		try {
		    functions.glClearColorxOES().invokeExact(red, green, blue, alpha);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glClearDepth">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glClearDepth.xhtml">registry.khronos.org</a>
	 */
	public void glClearDepth(double depth) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glClearDepth(\{depth})");
		try {
		    functions.glClearDepth().invokeExact(depth);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glClearDepthdNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glClearDepthdNV.xhtml">registry.khronos.org</a>
	 */
	public void glClearDepthdNV(double depth) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glClearDepthdNV(\{depth})");
		try {
		    functions.glClearDepthdNV().invokeExact(depth);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glClearDepthf">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glClearDepthf.xhtml">registry.khronos.org</a>
	 */
	public void glClearDepthf(float d) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glClearDepthf(\{d})");
		try {
		    functions.glClearDepthf().invokeExact(d);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glClearDepthfOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glClearDepthfOES.xhtml">registry.khronos.org</a>
	 */
	public void glClearDepthfOES(float depth) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glClearDepthfOES(\{depth})");
		try {
		    functions.glClearDepthfOES().invokeExact(depth);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glClearDepthx">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glClearDepthx.xhtml">registry.khronos.org</a>
	 */
	public void glClearDepthx(int depth) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glClearDepthx(\{depth})");
		try {
		    functions.glClearDepthx().invokeExact(depth);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glClearDepthxOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glClearDepthxOES.xhtml">registry.khronos.org</a>
	 */
	public void glClearDepthxOES(int depth) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glClearDepthxOES(\{depth})");
		try {
		    functions.glClearDepthxOES().invokeExact(depth);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glClearIndex">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glClearIndex.xhtml">registry.khronos.org</a>
	 */
	public void glClearIndex(float c) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glClearIndex(\{c})");
		try {
		    functions.glClearIndex().invokeExact(c);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glClearNamedBufferData">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glClearNamedBufferData.xhtml">registry.khronos.org</a>
	 */
	public void glClearNamedBufferData(int buffer, GLSizedInternalFormat internalformat, GLPixelFormat format, GLPixelType type, ConstPointer<Void> data) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glClearNamedBufferData(\{buffer}, \{internalformat}, \{format}, \{type}, \{data})");
		try {
		    functions.glClearNamedBufferData().invokeExact(buffer, internalformat, format, type, data);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were internalformat=\{internalformat}, format=\{format}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glClearNamedBufferDataEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glClearNamedBufferDataEXT.xhtml">registry.khronos.org</a>
	 */
	public void glClearNamedBufferDataEXT(int buffer, GLSizedInternalFormat internalformat, GLPixelFormat format, GLPixelType type, ConstPointer<Void> data) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glClearNamedBufferDataEXT(\{buffer}, \{internalformat}, \{format}, \{type}, \{data})");
		try {
		    functions.glClearNamedBufferDataEXT().invokeExact(buffer, internalformat, format, type, data);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were internalformat=\{internalformat}, format=\{format}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glClearNamedBufferSubData">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glClearNamedBufferSubData.xhtml">registry.khronos.org</a>
	 */
	public void glClearNamedBufferSubData(int buffer, GLSizedInternalFormat internalformat, long offset, long size, GLPixelFormat format, GLPixelType type, ConstPointer<Void> data) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glClearNamedBufferSubData(\{buffer}, \{internalformat}, \{offset}, \{size}, \{format}, \{type}, \{data})");
		try {
		    functions.glClearNamedBufferSubData().invokeExact(buffer, internalformat, offset, size, format, type, data);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were internalformat=\{internalformat}, format=\{format}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glClearNamedBufferSubDataEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glClearNamedBufferSubDataEXT.xhtml">registry.khronos.org</a>
	 */
	public void glClearNamedBufferSubDataEXT(int buffer, GLSizedInternalFormat internalformat, long offset, long size, GLPixelFormat format, GLPixelType type, ConstPointer<Void> data) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glClearNamedBufferSubDataEXT(\{buffer}, \{internalformat}, \{offset}, \{size}, \{format}, \{type}, \{data})");
		try {
		    functions.glClearNamedBufferSubDataEXT().invokeExact(buffer, internalformat, offset, size, format, type, data);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were internalformat=\{internalformat}, format=\{format}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glClearNamedFramebufferfi">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glClearNamedFramebufferfi.xhtml">registry.khronos.org</a>
	 */
	public void glClearNamedFramebufferfi(int framebuffer, GLBuffer buffer, int drawbuffer, float depth, int stencil) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glClearNamedFramebufferfi(\{framebuffer}, \{buffer}, \{drawbuffer}, \{depth}, \{stencil})");
		try {
		    functions.glClearNamedFramebufferfi().invokeExact(framebuffer, buffer, drawbuffer, depth, stencil);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were buffer=\{buffer}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glClearNamedFramebufferfv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glClearNamedFramebufferfv.xhtml">registry.khronos.org</a>
	 */
	public void glClearNamedFramebufferfv(int framebuffer, GLBuffer buffer, int drawbuffer, ConstPointer<Float> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glClearNamedFramebufferfv(\{framebuffer}, \{buffer}, \{drawbuffer}, \{value})");
		try {
		    functions.glClearNamedFramebufferfv().invokeExact(framebuffer, buffer, drawbuffer, value);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were buffer=\{buffer}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glClearNamedFramebufferiv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glClearNamedFramebufferiv.xhtml">registry.khronos.org</a>
	 */
	public void glClearNamedFramebufferiv(int framebuffer, GLBuffer buffer, int drawbuffer, ConstPointer<Integer> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glClearNamedFramebufferiv(\{framebuffer}, \{buffer}, \{drawbuffer}, \{value})");
		try {
		    functions.glClearNamedFramebufferiv().invokeExact(framebuffer, buffer, drawbuffer, value);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were buffer=\{buffer}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glClearNamedFramebufferuiv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glClearNamedFramebufferuiv.xhtml">registry.khronos.org</a>
	 */
	public void glClearNamedFramebufferuiv(int framebuffer, GLBuffer buffer, int drawbuffer, ConstPointer<Integer> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glClearNamedFramebufferuiv(\{framebuffer}, \{buffer}, \{drawbuffer}, \{value})");
		try {
		    functions.glClearNamedFramebufferuiv().invokeExact(framebuffer, buffer, drawbuffer, value);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were buffer=\{buffer}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glClearPixelLocalStorageuiEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glClearPixelLocalStorageuiEXT.xhtml">registry.khronos.org</a>
	 */
	public void glClearPixelLocalStorageuiEXT(int offset, int n, ConstPointer<Integer> values) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glClearPixelLocalStorageuiEXT(\{offset}, \{n}, \{values})");
		try {
		    functions.glClearPixelLocalStorageuiEXT().invokeExact(offset, n, values);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glClearStencil">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glClearStencil.xhtml">registry.khronos.org</a>
	 */
	public void glClearStencil(int s) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glClearStencil(\{s})");
		try {
		    functions.glClearStencil().invokeExact(s);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glClearTexImage">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glClearTexImage.xhtml">registry.khronos.org</a>
	 */
	public void glClearTexImage(int texture, int level, GLPixelFormat format, GLPixelType type, ConstPointer<Void> data) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glClearTexImage(\{texture}, \{level}, \{format}, \{type}, \{data})");
		try {
		    functions.glClearTexImage().invokeExact(texture, level, format, type, data);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were format=\{format}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glClearTexImageEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glClearTexImageEXT.xhtml">registry.khronos.org</a>
	 */
	public void glClearTexImageEXT(int texture, int level, GLPixelFormat format, GLPixelType type, ConstPointer<Void> data) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glClearTexImageEXT(\{texture}, \{level}, \{format}, \{type}, \{data})");
		try {
		    functions.glClearTexImageEXT().invokeExact(texture, level, format, type, data);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were format=\{format}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glClearTexSubImage">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glClearTexSubImage.xhtml">registry.khronos.org</a>
	 */
	public void glClearTexSubImage(int texture, int level, int xoffset, int yoffset, int zoffset, int width, int height, int depth, GLPixelFormat format, GLPixelType type, ConstPointer<Void> data) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glClearTexSubImage(\{texture}, \{level}, \{xoffset}, \{yoffset}, \{zoffset}, \{width}, \{height}, \{depth}, \{format}, \{type}, \{data})");
		try {
		    functions.glClearTexSubImage().invokeExact(texture, level, xoffset, yoffset, zoffset, width, height, depth, format, type, data);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were format=\{format}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glClearTexSubImageEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glClearTexSubImageEXT.xhtml">registry.khronos.org</a>
	 */
	public void glClearTexSubImageEXT(int texture, int level, int xoffset, int yoffset, int zoffset, int width, int height, int depth, GLPixelFormat format, GLPixelType type, ConstPointer<Void> data) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glClearTexSubImageEXT(\{texture}, \{level}, \{xoffset}, \{yoffset}, \{zoffset}, \{width}, \{height}, \{depth}, \{format}, \{type}, \{data})");
		try {
		    functions.glClearTexSubImageEXT().invokeExact(texture, level, xoffset, yoffset, zoffset, width, height, depth, format, type, data);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were format=\{format}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glClientActiveTexture">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glClientActiveTexture.xhtml">registry.khronos.org</a>
	 */
	public void glClientActiveTexture(GLTextureUnit texture) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glClientActiveTexture(\{texture})");
		try {
		    functions.glClientActiveTexture().invokeExact(texture);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were texture=\{texture}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glClientActiveTextureARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glClientActiveTextureARB.xhtml">registry.khronos.org</a>
	 */
	public void glClientActiveTextureARB(GLTextureUnit texture) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glClientActiveTextureARB(\{texture})");
		try {
		    functions.glClientActiveTextureARB().invokeExact(texture);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were texture=\{texture}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glClientActiveVertexStreamATI">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glClientActiveVertexStreamATI.xhtml">registry.khronos.org</a>
	 */
	public void glClientActiveVertexStreamATI(GLVertexStreamATI stream) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glClientActiveVertexStreamATI(\{stream})");
		try {
		    functions.glClientActiveVertexStreamATI().invokeExact(stream);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were stream=\{stream}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glClientAttribDefaultEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glClientAttribDefaultEXT.xhtml">registry.khronos.org</a>
	 */
	public void glClientAttribDefaultEXT(int mask) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glClientAttribDefaultEXT(\{mask})");
		try {
		    functions.glClientAttribDefaultEXT().invokeExact(mask);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glClientWaitSemaphoreui64NVX">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glClientWaitSemaphoreui64NVX.xhtml">registry.khronos.org</a>
	 */
	public void glClientWaitSemaphoreui64NVX(int fenceObjectCount, ConstPointer<Integer> semaphoreArray, ConstPointer<Long> fenceValueArray) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glClientWaitSemaphoreui64NVX(\{fenceObjectCount}, \{semaphoreArray}, \{fenceValueArray})");
		try {
		    functions.glClientWaitSemaphoreui64NVX().invokeExact(fenceObjectCount, semaphoreArray, fenceValueArray);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glClientWaitSync">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glClientWaitSync.xhtml">registry.khronos.org</a>
	 */
	public GLSyncStatus glClientWaitSync(ConstPointer<Void> sync, int flags, long timeout) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glClientWaitSync(\{sync}, \{flags}, \{timeout})");
		try {
		    var r = (GLSyncStatus)functions.glClientWaitSync().invokeExact(sync, flags, timeout);
			checkError();
			return r;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glClientWaitSyncAPPLE">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glClientWaitSyncAPPLE.xhtml">registry.khronos.org</a>
	 */
	public GLSyncStatus glClientWaitSyncAPPLE(ConstPointer<Void> sync, int flags, long timeout) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glClientWaitSyncAPPLE(\{sync}, \{flags}, \{timeout})");
		try {
		    var r = (GLSyncStatus)functions.glClientWaitSyncAPPLE().invokeExact(sync, flags, timeout);
			checkError();
			return r;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glClipControl">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glClipControl.xhtml">registry.khronos.org</a>
	 */
	public void glClipControl(GLClipControlOrigin origin, GLClipControlDepth depth) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glClipControl(\{origin}, \{depth})");
		try {
		    functions.glClipControl().invokeExact(origin, depth);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were origin=\{origin}, depth=\{depth}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glClipControlEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glClipControlEXT.xhtml">registry.khronos.org</a>
	 */
	public void glClipControlEXT(GLEnum origin, GLEnum depth) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glClipControlEXT(\{origin}, \{depth})");
		try {
		    functions.glClipControlEXT().invokeExact(origin, depth);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were origin=\{origin}, depth=\{depth}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glClipPlane">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glClipPlane.xhtml">registry.khronos.org</a>
	 */
	public void glClipPlane(GLClipPlaneName plane, ConstPointer<Double> equation) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glClipPlane(\{plane}, \{equation})");
		try {
		    functions.glClipPlane().invokeExact(plane, equation);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were plane=\{plane}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glClipPlanef">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glClipPlanef.xhtml">registry.khronos.org</a>
	 */
	public void glClipPlanef(GLClipPlaneName p, ConstPointer<Float> eqn) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glClipPlanef(\{p}, \{eqn})");
		try {
		    functions.glClipPlanef().invokeExact(p, eqn);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were p=\{p}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glClipPlanefIMG">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glClipPlanefIMG.xhtml">registry.khronos.org</a>
	 */
	public void glClipPlanefIMG(GLClipPlaneName p, ConstPointer<Float> eqn) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glClipPlanefIMG(\{p}, \{eqn})");
		try {
		    functions.glClipPlanefIMG().invokeExact(p, eqn);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were p=\{p}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glClipPlanefOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glClipPlanefOES.xhtml">registry.khronos.org</a>
	 */
	public void glClipPlanefOES(GLClipPlaneName plane, ConstPointer<Float> equation) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glClipPlanefOES(\{plane}, \{equation})");
		try {
		    functions.glClipPlanefOES().invokeExact(plane, equation);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were plane=\{plane}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glClipPlanex">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glClipPlanex.xhtml">registry.khronos.org</a>
	 */
	public void glClipPlanex(GLClipPlaneName plane, ConstPointer<Integer> equation) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glClipPlanex(\{plane}, \{equation})");
		try {
		    functions.glClipPlanex().invokeExact(plane, equation);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were plane=\{plane}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glClipPlanexIMG">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glClipPlanexIMG.xhtml">registry.khronos.org</a>
	 */
	public void glClipPlanexIMG(GLClipPlaneName p, ConstPointer<Integer> eqn) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glClipPlanexIMG(\{p}, \{eqn})");
		try {
		    functions.glClipPlanexIMG().invokeExact(p, eqn);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were p=\{p}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glClipPlanexOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glClipPlanexOES.xhtml">registry.khronos.org</a>
	 */
	public void glClipPlanexOES(GLClipPlaneName plane, ConstPointer<Integer> equation) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glClipPlanexOES(\{plane}, \{equation})");
		try {
		    functions.glClipPlanexOES().invokeExact(plane, equation);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were plane=\{plane}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glColor3b">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glColor3b.xhtml">registry.khronos.org</a>
	 */
	public void glColor3b(byte red, byte green, byte blue) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glColor3b(\{red}, \{green}, \{blue})");
		try {
		    functions.glColor3b().invokeExact(red, green, blue);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glColor3bv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glColor3bv.xhtml">registry.khronos.org</a>
	 */
	public void glColor3bv(ConstPointer<Byte> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glColor3bv(\{v})");
		try {
		    functions.glColor3bv().invokeExact(v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glColor3d">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glColor3d.xhtml">registry.khronos.org</a>
	 */
	public void glColor3d(double red, double green, double blue) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glColor3d(\{red}, \{green}, \{blue})");
		try {
		    functions.glColor3d().invokeExact(red, green, blue);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glColor3dv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glColor3dv.xhtml">registry.khronos.org</a>
	 */
	public void glColor3dv(ConstPointer<Double> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glColor3dv(\{v})");
		try {
		    functions.glColor3dv().invokeExact(v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glColor3f">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glColor3f.xhtml">registry.khronos.org</a>
	 */
	public void glColor3f(float red, float green, float blue) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glColor3f(\{red}, \{green}, \{blue})");
		try {
		    functions.glColor3f().invokeExact(red, green, blue);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glColor3fVertex3fSUN">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glColor3fVertex3fSUN.xhtml">registry.khronos.org</a>
	 */
	public void glColor3fVertex3fSUN(float r, float g, float b, float x, float y, float z) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glColor3fVertex3fSUN(\{r}, \{g}, \{b}, \{x}, \{y}, \{z})");
		try {
		    functions.glColor3fVertex3fSUN().invokeExact(r, g, b, x, y, z);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glColor3fVertex3fvSUN">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glColor3fVertex3fvSUN.xhtml">registry.khronos.org</a>
	 */
	public void glColor3fVertex3fvSUN(ConstPointer<Float> c, ConstPointer<Float> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glColor3fVertex3fvSUN(\{c}, \{v})");
		try {
		    functions.glColor3fVertex3fvSUN().invokeExact(c, v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glColor3fv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glColor3fv.xhtml">registry.khronos.org</a>
	 */
	public void glColor3fv(ConstPointer<Float> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glColor3fv(\{v})");
		try {
		    functions.glColor3fv().invokeExact(v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glColor3hNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glColor3hNV.xhtml">registry.khronos.org</a>
	 */
	public void glColor3hNV(char red, char green, char blue) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glColor3hNV(\{red}, \{green}, \{blue})");
		try {
		    functions.glColor3hNV().invokeExact(red, green, blue);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glColor3hvNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glColor3hvNV.xhtml">registry.khronos.org</a>
	 */
	public void glColor3hvNV(ConstPointer<Character> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glColor3hvNV(\{v})");
		try {
		    functions.glColor3hvNV().invokeExact(v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glColor3i">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glColor3i.xhtml">registry.khronos.org</a>
	 */
	public void glColor3i(int red, int green, int blue) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glColor3i(\{red}, \{green}, \{blue})");
		try {
		    functions.glColor3i().invokeExact(red, green, blue);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glColor3iv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glColor3iv.xhtml">registry.khronos.org</a>
	 */
	public void glColor3iv(ConstPointer<Integer> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glColor3iv(\{v})");
		try {
		    functions.glColor3iv().invokeExact(v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glColor3s">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glColor3s.xhtml">registry.khronos.org</a>
	 */
	public void glColor3s(short red, short green, short blue) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glColor3s(\{red}, \{green}, \{blue})");
		try {
		    functions.glColor3s().invokeExact(red, green, blue);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glColor3sv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glColor3sv.xhtml">registry.khronos.org</a>
	 */
	public void glColor3sv(ConstPointer<Short> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glColor3sv(\{v})");
		try {
		    functions.glColor3sv().invokeExact(v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glColor3ub">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glColor3ub.xhtml">registry.khronos.org</a>
	 */
	public void glColor3ub(byte red, byte green, byte blue) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glColor3ub(\{red}, \{green}, \{blue})");
		try {
		    functions.glColor3ub().invokeExact(red, green, blue);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glColor3ubv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glColor3ubv.xhtml">registry.khronos.org</a>
	 */
	public void glColor3ubv(ConstPointer<Byte> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glColor3ubv(\{v})");
		try {
		    functions.glColor3ubv().invokeExact(v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glColor3ui">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glColor3ui.xhtml">registry.khronos.org</a>
	 */
	public void glColor3ui(int red, int green, int blue) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glColor3ui(\{red}, \{green}, \{blue})");
		try {
		    functions.glColor3ui().invokeExact(red, green, blue);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glColor3uiv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glColor3uiv.xhtml">registry.khronos.org</a>
	 */
	public void glColor3uiv(ConstPointer<Integer> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glColor3uiv(\{v})");
		try {
		    functions.glColor3uiv().invokeExact(v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glColor3us">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glColor3us.xhtml">registry.khronos.org</a>
	 */
	public void glColor3us(char red, char green, char blue) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glColor3us(\{red}, \{green}, \{blue})");
		try {
		    functions.glColor3us().invokeExact(red, green, blue);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glColor3usv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glColor3usv.xhtml">registry.khronos.org</a>
	 */
	public void glColor3usv(ConstPointer<Character> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glColor3usv(\{v})");
		try {
		    functions.glColor3usv().invokeExact(v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glColor3xOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glColor3xOES.xhtml">registry.khronos.org</a>
	 */
	public void glColor3xOES(int red, int green, int blue) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glColor3xOES(\{red}, \{green}, \{blue})");
		try {
		    functions.glColor3xOES().invokeExact(red, green, blue);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glColor3xvOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glColor3xvOES.xhtml">registry.khronos.org</a>
	 */
	public void glColor3xvOES(ConstPointer<Integer> components) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glColor3xvOES(\{components})");
		try {
		    functions.glColor3xvOES().invokeExact(components);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glColor4b">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glColor4b.xhtml">registry.khronos.org</a>
	 */
	public void glColor4b(byte red, byte green, byte blue, byte alpha) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glColor4b(\{red}, \{green}, \{blue}, \{alpha})");
		try {
		    functions.glColor4b().invokeExact(red, green, blue, alpha);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glColor4bv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glColor4bv.xhtml">registry.khronos.org</a>
	 */
	public void glColor4bv(ConstPointer<Byte> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glColor4bv(\{v})");
		try {
		    functions.glColor4bv().invokeExact(v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glColor4d">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glColor4d.xhtml">registry.khronos.org</a>
	 */
	public void glColor4d(double red, double green, double blue, double alpha) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glColor4d(\{red}, \{green}, \{blue}, \{alpha})");
		try {
		    functions.glColor4d().invokeExact(red, green, blue, alpha);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glColor4dv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glColor4dv.xhtml">registry.khronos.org</a>
	 */
	public void glColor4dv(ConstPointer<Double> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glColor4dv(\{v})");
		try {
		    functions.glColor4dv().invokeExact(v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glColor4f">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glColor4f.xhtml">registry.khronos.org</a>
	 */
	public void glColor4f(float red, float green, float blue, float alpha) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glColor4f(\{red}, \{green}, \{blue}, \{alpha})");
		try {
		    functions.glColor4f().invokeExact(red, green, blue, alpha);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glColor4fNormal3fVertex3fSUN">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glColor4fNormal3fVertex3fSUN.xhtml">registry.khronos.org</a>
	 */
	public void glColor4fNormal3fVertex3fSUN(float r, float g, float b, float a, float nx, float ny, float nz, float x, float y, float z) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glColor4fNormal3fVertex3fSUN(\{r}, \{g}, \{b}, \{a}, \{nx}, \{ny}, \{nz}, \{x}, \{y}, \{z})");
		try {
		    functions.glColor4fNormal3fVertex3fSUN().invokeExact(r, g, b, a, nx, ny, nz, x, y, z);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glColor4fNormal3fVertex3fvSUN">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glColor4fNormal3fVertex3fvSUN.xhtml">registry.khronos.org</a>
	 */
	public void glColor4fNormal3fVertex3fvSUN(ConstPointer<Float> c, ConstPointer<Float> n, ConstPointer<Float> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glColor4fNormal3fVertex3fvSUN(\{c}, \{n}, \{v})");
		try {
		    functions.glColor4fNormal3fVertex3fvSUN().invokeExact(c, n, v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glColor4fv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glColor4fv.xhtml">registry.khronos.org</a>
	 */
	public void glColor4fv(ConstPointer<Float> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glColor4fv(\{v})");
		try {
		    functions.glColor4fv().invokeExact(v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glColor4hNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glColor4hNV.xhtml">registry.khronos.org</a>
	 */
	public void glColor4hNV(char red, char green, char blue, char alpha) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glColor4hNV(\{red}, \{green}, \{blue}, \{alpha})");
		try {
		    functions.glColor4hNV().invokeExact(red, green, blue, alpha);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glColor4hvNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glColor4hvNV.xhtml">registry.khronos.org</a>
	 */
	public void glColor4hvNV(ConstPointer<Character> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glColor4hvNV(\{v})");
		try {
		    functions.glColor4hvNV().invokeExact(v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glColor4i">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glColor4i.xhtml">registry.khronos.org</a>
	 */
	public void glColor4i(int red, int green, int blue, int alpha) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glColor4i(\{red}, \{green}, \{blue}, \{alpha})");
		try {
		    functions.glColor4i().invokeExact(red, green, blue, alpha);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glColor4iv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glColor4iv.xhtml">registry.khronos.org</a>
	 */
	public void glColor4iv(ConstPointer<Integer> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glColor4iv(\{v})");
		try {
		    functions.glColor4iv().invokeExact(v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glColor4s">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glColor4s.xhtml">registry.khronos.org</a>
	 */
	public void glColor4s(short red, short green, short blue, short alpha) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glColor4s(\{red}, \{green}, \{blue}, \{alpha})");
		try {
		    functions.glColor4s().invokeExact(red, green, blue, alpha);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glColor4sv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glColor4sv.xhtml">registry.khronos.org</a>
	 */
	public void glColor4sv(ConstPointer<Short> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glColor4sv(\{v})");
		try {
		    functions.glColor4sv().invokeExact(v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glColor4ub">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glColor4ub.xhtml">registry.khronos.org</a>
	 */
	public void glColor4ub(byte red, byte green, byte blue, byte alpha) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glColor4ub(\{red}, \{green}, \{blue}, \{alpha})");
		try {
		    functions.glColor4ub().invokeExact(red, green, blue, alpha);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glColor4ubVertex2fSUN">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glColor4ubVertex2fSUN.xhtml">registry.khronos.org</a>
	 */
	public void glColor4ubVertex2fSUN(byte r, byte g, byte b, byte a, float x, float y) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glColor4ubVertex2fSUN(\{r}, \{g}, \{b}, \{a}, \{x}, \{y})");
		try {
		    functions.glColor4ubVertex2fSUN().invokeExact(r, g, b, a, x, y);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glColor4ubVertex2fvSUN">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glColor4ubVertex2fvSUN.xhtml">registry.khronos.org</a>
	 */
	public void glColor4ubVertex2fvSUN(ConstPointer<Byte> c, ConstPointer<Float> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glColor4ubVertex2fvSUN(\{c}, \{v})");
		try {
		    functions.glColor4ubVertex2fvSUN().invokeExact(c, v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glColor4ubVertex3fSUN">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glColor4ubVertex3fSUN.xhtml">registry.khronos.org</a>
	 */
	public void glColor4ubVertex3fSUN(byte r, byte g, byte b, byte a, float x, float y, float z) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glColor4ubVertex3fSUN(\{r}, \{g}, \{b}, \{a}, \{x}, \{y}, \{z})");
		try {
		    functions.glColor4ubVertex3fSUN().invokeExact(r, g, b, a, x, y, z);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glColor4ubVertex3fvSUN">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glColor4ubVertex3fvSUN.xhtml">registry.khronos.org</a>
	 */
	public void glColor4ubVertex3fvSUN(ConstPointer<Byte> c, ConstPointer<Float> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glColor4ubVertex3fvSUN(\{c}, \{v})");
		try {
		    functions.glColor4ubVertex3fvSUN().invokeExact(c, v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glColor4ubv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glColor4ubv.xhtml">registry.khronos.org</a>
	 */
	public void glColor4ubv(ConstPointer<Byte> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glColor4ubv(\{v})");
		try {
		    functions.glColor4ubv().invokeExact(v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glColor4ui">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glColor4ui.xhtml">registry.khronos.org</a>
	 */
	public void glColor4ui(int red, int green, int blue, int alpha) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glColor4ui(\{red}, \{green}, \{blue}, \{alpha})");
		try {
		    functions.glColor4ui().invokeExact(red, green, blue, alpha);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glColor4uiv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glColor4uiv.xhtml">registry.khronos.org</a>
	 */
	public void glColor4uiv(ConstPointer<Integer> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glColor4uiv(\{v})");
		try {
		    functions.glColor4uiv().invokeExact(v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glColor4us">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glColor4us.xhtml">registry.khronos.org</a>
	 */
	public void glColor4us(char red, char green, char blue, char alpha) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glColor4us(\{red}, \{green}, \{blue}, \{alpha})");
		try {
		    functions.glColor4us().invokeExact(red, green, blue, alpha);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glColor4usv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glColor4usv.xhtml">registry.khronos.org</a>
	 */
	public void glColor4usv(ConstPointer<Character> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glColor4usv(\{v})");
		try {
		    functions.glColor4usv().invokeExact(v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glColor4x">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glColor4x.xhtml">registry.khronos.org</a>
	 */
	public void glColor4x(int red, int green, int blue, int alpha) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glColor4x(\{red}, \{green}, \{blue}, \{alpha})");
		try {
		    functions.glColor4x().invokeExact(red, green, blue, alpha);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glColor4xOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glColor4xOES.xhtml">registry.khronos.org</a>
	 */
	public void glColor4xOES(int red, int green, int blue, int alpha) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glColor4xOES(\{red}, \{green}, \{blue}, \{alpha})");
		try {
		    functions.glColor4xOES().invokeExact(red, green, blue, alpha);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glColor4xvOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glColor4xvOES.xhtml">registry.khronos.org</a>
	 */
	public void glColor4xvOES(ConstPointer<Integer> components) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glColor4xvOES(\{components})");
		try {
		    functions.glColor4xvOES().invokeExact(components);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glColorFormatNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glColorFormatNV.xhtml">registry.khronos.org</a>
	 */
	public void glColorFormatNV(int size, GLColorPointerType type, int stride) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glColorFormatNV(\{size}, \{type}, \{stride})");
		try {
		    functions.glColorFormatNV().invokeExact(size, type, stride);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glColorFragmentOp1ATI">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glColorFragmentOp1ATI.xhtml">registry.khronos.org</a>
	 */
	public void glColorFragmentOp1ATI(GLFragmentOp1ATI op, int dst, int dstMask, int dstMod, int arg1, int arg1Rep, int arg1Mod) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glColorFragmentOp1ATI(\{op}, \{dst}, \{dstMask}, \{dstMod}, \{arg1}, \{arg1Rep}, \{arg1Mod})");
		try {
		    functions.glColorFragmentOp1ATI().invokeExact(op, dst, dstMask, dstMod, arg1, arg1Rep, arg1Mod);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were op=\{op}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glColorFragmentOp2ATI">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glColorFragmentOp2ATI.xhtml">registry.khronos.org</a>
	 */
	public void glColorFragmentOp2ATI(GLFragmentOp2ATI op, int dst, int dstMask, int dstMod, int arg1, int arg1Rep, int arg1Mod, int arg2, int arg2Rep, int arg2Mod) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glColorFragmentOp2ATI(\{op}, \{dst}, \{dstMask}, \{dstMod}, \{arg1}, \{arg1Rep}, \{arg1Mod}, \{arg2}, \{arg2Rep}, \{arg2Mod})");
		try {
		    functions.glColorFragmentOp2ATI().invokeExact(op, dst, dstMask, dstMod, arg1, arg1Rep, arg1Mod, arg2, arg2Rep, arg2Mod);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were op=\{op}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glColorFragmentOp3ATI">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glColorFragmentOp3ATI.xhtml">registry.khronos.org</a>
	 */
	public void glColorFragmentOp3ATI(GLFragmentOp3ATI op, int dst, int dstMask, int dstMod, int arg1, int arg1Rep, int arg1Mod, int arg2, int arg2Rep, int arg2Mod, int arg3, int arg3Rep, int arg3Mod) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glColorFragmentOp3ATI(\{op}, \{dst}, \{dstMask}, \{dstMod}, \{arg1}, \{arg1Rep}, \{arg1Mod}, \{arg2}, \{arg2Rep}, \{arg2Mod}, \{arg3}, \{arg3Rep}, \{arg3Mod})");
		try {
		    functions.glColorFragmentOp3ATI().invokeExact(op, dst, dstMask, dstMod, arg1, arg1Rep, arg1Mod, arg2, arg2Rep, arg2Mod, arg3, arg3Rep, arg3Mod);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were op=\{op}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glColorMask">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glColorMask.xhtml">registry.khronos.org</a>
	 */
	public void glColorMask(boolean red, boolean green, boolean blue, boolean alpha) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glColorMask(\{red}, \{green}, \{blue}, \{alpha})");
		try {
		    functions.glColorMask().invokeExact(red, green, blue, alpha);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glColorMaskIndexedEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glColorMaskIndexedEXT.xhtml">registry.khronos.org</a>
	 */
	public void glColorMaskIndexedEXT(int index, boolean r, boolean g, boolean b, boolean a) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glColorMaskIndexedEXT(\{index}, \{r}, \{g}, \{b}, \{a})");
		try {
		    functions.glColorMaskIndexedEXT().invokeExact(index, r, g, b, a);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glColorMaski">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glColorMaski.xhtml">registry.khronos.org</a>
	 */
	public void glColorMaski(int index, boolean r, boolean g, boolean b, boolean a) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glColorMaski(\{index}, \{r}, \{g}, \{b}, \{a})");
		try {
		    functions.glColorMaski().invokeExact(index, r, g, b, a);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glColorMaskiEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glColorMaskiEXT.xhtml">registry.khronos.org</a>
	 */
	public void glColorMaskiEXT(int index, boolean r, boolean g, boolean b, boolean a) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glColorMaskiEXT(\{index}, \{r}, \{g}, \{b}, \{a})");
		try {
		    functions.glColorMaskiEXT().invokeExact(index, r, g, b, a);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glColorMaskiOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glColorMaskiOES.xhtml">registry.khronos.org</a>
	 */
	public void glColorMaskiOES(int index, boolean r, boolean g, boolean b, boolean a) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glColorMaskiOES(\{index}, \{r}, \{g}, \{b}, \{a})");
		try {
		    functions.glColorMaskiOES().invokeExact(index, r, g, b, a);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glColorMaterial">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glColorMaterial.xhtml">registry.khronos.org</a>
	 */
	public void glColorMaterial(GLTriangleFace face, GLColorMaterialParameter mode) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glColorMaterial(\{face}, \{mode})");
		try {
		    functions.glColorMaterial().invokeExact(face, mode);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were face=\{face}, mode=\{mode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glColorP3ui">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glColorP3ui.xhtml">registry.khronos.org</a>
	 */
	public void glColorP3ui(GLColorPointerType type, int color) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glColorP3ui(\{type}, \{color})");
		try {
		    functions.glColorP3ui().invokeExact(type, color);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glColorP3uiv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glColorP3uiv.xhtml">registry.khronos.org</a>
	 */
	public void glColorP3uiv(GLColorPointerType type, ConstPointer<Integer> color) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glColorP3uiv(\{type}, \{color})");
		try {
		    functions.glColorP3uiv().invokeExact(type, color);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glColorP4ui">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glColorP4ui.xhtml">registry.khronos.org</a>
	 */
	public void glColorP4ui(GLColorPointerType type, int color) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glColorP4ui(\{type}, \{color})");
		try {
		    functions.glColorP4ui().invokeExact(type, color);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glColorP4uiv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glColorP4uiv.xhtml">registry.khronos.org</a>
	 */
	public void glColorP4uiv(GLColorPointerType type, ConstPointer<Integer> color) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glColorP4uiv(\{type}, \{color})");
		try {
		    functions.glColorP4uiv().invokeExact(type, color);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glColorPointer">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glColorPointer.xhtml">registry.khronos.org</a>
	 */
	public void glColorPointer(int size, GLColorPointerType type, int stride, ConstPointer<Void> pointer) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glColorPointer(\{size}, \{type}, \{stride}, \{pointer})");
		try {
		    functions.glColorPointer().invokeExact(size, type, stride, pointer);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glColorPointerEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glColorPointerEXT.xhtml">registry.khronos.org</a>
	 */
	public void glColorPointerEXT(int size, GLColorPointerType type, int stride, int count, ConstPointer<Void> pointer) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glColorPointerEXT(\{size}, \{type}, \{stride}, \{count}, \{pointer})");
		try {
		    functions.glColorPointerEXT().invokeExact(size, type, stride, count, pointer);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glColorPointerListIBM">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glColorPointerListIBM.xhtml">registry.khronos.org</a>
	 */
	public void glColorPointerListIBM(int size, GLColorPointerType type, int stride, MutPointer<ConstPointer<Void>> pointer, int ptrstride) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glColorPointerListIBM(\{size}, \{type}, \{stride}, \{pointer}, \{ptrstride})");
		try {
		    functions.glColorPointerListIBM().invokeExact(size, type, stride, pointer, ptrstride);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glColorPointervINTEL">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glColorPointervINTEL.xhtml">registry.khronos.org</a>
	 */
	public void glColorPointervINTEL(int size, GLVertexPointerType type, MutPointer<ConstPointer<Void>> pointer) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glColorPointervINTEL(\{size}, \{type}, \{pointer})");
		try {
		    functions.glColorPointervINTEL().invokeExact(size, type, pointer);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glColorSubTable">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glColorSubTable.xhtml">registry.khronos.org</a>
	 */
	public void glColorSubTable(GLColorTableTarget target, int start, int count, GLPixelFormat format, GLPixelType type, ConstPointer<Void> data) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glColorSubTable(\{target}, \{start}, \{count}, \{format}, \{type}, \{data})");
		try {
		    functions.glColorSubTable().invokeExact(target, start, count, format, type, data);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, format=\{format}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glColorSubTableEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glColorSubTableEXT.xhtml">registry.khronos.org</a>
	 */
	public void glColorSubTableEXT(GLColorTableTarget target, int start, int count, GLPixelFormat format, GLPixelType type, ConstPointer<Void> data) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glColorSubTableEXT(\{target}, \{start}, \{count}, \{format}, \{type}, \{data})");
		try {
		    functions.glColorSubTableEXT().invokeExact(target, start, count, format, type, data);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, format=\{format}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glColorTable">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glColorTable.xhtml">registry.khronos.org</a>
	 */
	public void glColorTable(GLColorTableTarget target, GLInternalFormat internalformat, int width, GLPixelFormat format, GLPixelType type, ConstPointer<Void> table) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glColorTable(\{target}, \{internalformat}, \{width}, \{format}, \{type}, \{table})");
		try {
		    functions.glColorTable().invokeExact(target, internalformat, width, format, type, table);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, internalformat=\{internalformat}, format=\{format}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glColorTableEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glColorTableEXT.xhtml">registry.khronos.org</a>
	 */
	public void glColorTableEXT(GLColorTableTarget target, GLInternalFormat internalFormat, int width, GLPixelFormat format, GLPixelType type, ConstPointer<Void> table) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glColorTableEXT(\{target}, \{internalFormat}, \{width}, \{format}, \{type}, \{table})");
		try {
		    functions.glColorTableEXT().invokeExact(target, internalFormat, width, format, type, table);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, internalFormat=\{internalFormat}, format=\{format}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glColorTableParameterfv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glColorTableParameterfv.xhtml">registry.khronos.org</a>
	 */
	public void glColorTableParameterfv(GLColorTableTarget target, GLColorTableParameterPName pname, ConstPointer<Float> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glColorTableParameterfv(\{target}, \{pname}, \{params})");
		try {
		    functions.glColorTableParameterfv().invokeExact(target, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glColorTableParameterfvSGI">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glColorTableParameterfvSGI.xhtml">registry.khronos.org</a>
	 */
	public void glColorTableParameterfvSGI(GLColorTableTargetSGI target, GLColorTableParameterPName pname, ConstPointer<Float> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glColorTableParameterfvSGI(\{target}, \{pname}, \{params})");
		try {
		    functions.glColorTableParameterfvSGI().invokeExact(target, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glColorTableParameteriv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glColorTableParameteriv.xhtml">registry.khronos.org</a>
	 */
	public void glColorTableParameteriv(GLColorTableTarget target, GLColorTableParameterPName pname, ConstPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glColorTableParameteriv(\{target}, \{pname}, \{params})");
		try {
		    functions.glColorTableParameteriv().invokeExact(target, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glColorTableParameterivSGI">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glColorTableParameterivSGI.xhtml">registry.khronos.org</a>
	 */
	public void glColorTableParameterivSGI(GLColorTableTargetSGI target, GLColorTableParameterPName pname, ConstPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glColorTableParameterivSGI(\{target}, \{pname}, \{params})");
		try {
		    functions.glColorTableParameterivSGI().invokeExact(target, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glColorTableSGI">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glColorTableSGI.xhtml">registry.khronos.org</a>
	 */
	public void glColorTableSGI(GLColorTableTargetSGI target, GLInternalFormat internalformat, int width, GLPixelFormat format, GLPixelType type, ConstPointer<Void> table) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glColorTableSGI(\{target}, \{internalformat}, \{width}, \{format}, \{type}, \{table})");
		try {
		    functions.glColorTableSGI().invokeExact(target, internalformat, width, format, type, table);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, internalformat=\{internalformat}, format=\{format}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCombinerInputNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCombinerInputNV.xhtml">registry.khronos.org</a>
	 */
	public void glCombinerInputNV(GLCombinerStageNV stage, GLCombinerPortionNV portion, GLCombinerVariableNV variable, GLCombinerRegisterNV input, GLCombinerMappingNV mapping, GLCombinerComponentUsageNV componentUsage) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCombinerInputNV(\{stage}, \{portion}, \{variable}, \{input}, \{mapping}, \{componentUsage})");
		try {
		    functions.glCombinerInputNV().invokeExact(stage, portion, variable, input, mapping, componentUsage);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were stage=\{stage}, portion=\{portion}, variable=\{variable}, input=\{input}, mapping=\{mapping}, componentUsage=\{componentUsage}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCombinerOutputNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCombinerOutputNV.xhtml">registry.khronos.org</a>
	 */
	public void glCombinerOutputNV(GLCombinerStageNV stage, GLCombinerPortionNV portion, GLCombinerRegisterNV abOutput, GLCombinerRegisterNV cdOutput, GLCombinerRegisterNV sumOutput, GLCombinerScaleNV scale, GLCombinerBiasNV bias, boolean abDotProduct, boolean cdDotProduct, boolean muxSum) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCombinerOutputNV(\{stage}, \{portion}, \{abOutput}, \{cdOutput}, \{sumOutput}, \{scale}, \{bias}, \{abDotProduct}, \{cdDotProduct}, \{muxSum})");
		try {
		    functions.glCombinerOutputNV().invokeExact(stage, portion, abOutput, cdOutput, sumOutput, scale, bias, abDotProduct, cdDotProduct, muxSum);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were stage=\{stage}, portion=\{portion}, abOutput=\{abOutput}, cdOutput=\{cdOutput}, sumOutput=\{sumOutput}, scale=\{scale}, bias=\{bias}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCombinerParameterfNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCombinerParameterfNV.xhtml">registry.khronos.org</a>
	 */
	public void glCombinerParameterfNV(GLCombinerParameterNV pname, float param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCombinerParameterfNV(\{pname}, \{param})");
		try {
		    functions.glCombinerParameterfNV().invokeExact(pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCombinerParameterfvNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCombinerParameterfvNV.xhtml">registry.khronos.org</a>
	 */
	public void glCombinerParameterfvNV(GLCombinerParameterNV pname, ConstPointer<Float> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCombinerParameterfvNV(\{pname}, \{params})");
		try {
		    functions.glCombinerParameterfvNV().invokeExact(pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCombinerParameteriNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCombinerParameteriNV.xhtml">registry.khronos.org</a>
	 */
	public void glCombinerParameteriNV(GLCombinerParameterNV pname, int param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCombinerParameteriNV(\{pname}, \{param})");
		try {
		    functions.glCombinerParameteriNV().invokeExact(pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCombinerParameterivNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCombinerParameterivNV.xhtml">registry.khronos.org</a>
	 */
	public void glCombinerParameterivNV(GLCombinerParameterNV pname, ConstPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCombinerParameterivNV(\{pname}, \{params})");
		try {
		    functions.glCombinerParameterivNV().invokeExact(pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCombinerStageParameterfvNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCombinerStageParameterfvNV.xhtml">registry.khronos.org</a>
	 */
	public void glCombinerStageParameterfvNV(GLCombinerStageNV stage, GLCombinerParameterNV pname, ConstPointer<Float> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCombinerStageParameterfvNV(\{stage}, \{pname}, \{params})");
		try {
		    functions.glCombinerStageParameterfvNV().invokeExact(stage, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were stage=\{stage}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCommandListSegmentsNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCommandListSegmentsNV.xhtml">registry.khronos.org</a>
	 */
	public void glCommandListSegmentsNV(int list, int segments) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCommandListSegmentsNV(\{list}, \{segments})");
		try {
		    functions.glCommandListSegmentsNV().invokeExact(list, segments);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCompileCommandListNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCompileCommandListNV.xhtml">registry.khronos.org</a>
	 */
	public void glCompileCommandListNV(int list) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCompileCommandListNV(\{list})");
		try {
		    functions.glCompileCommandListNV().invokeExact(list);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCompileShader">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCompileShader.xhtml">registry.khronos.org</a>
	 */
	public void glCompileShader(int shader) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCompileShader(\{shader})");
		try {
		    functions.glCompileShader().invokeExact(shader);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCompileShaderARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCompileShaderARB.xhtml">registry.khronos.org</a>
	 */
	public void glCompileShaderARB(int shaderObj) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCompileShaderARB(\{shaderObj})");
		try {
		    functions.glCompileShaderARB().invokeExact(shaderObj);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCompileShaderIncludeARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCompileShaderIncludeARB.xhtml">registry.khronos.org</a>
	 */
	public void glCompileShaderIncludeARB(int shader, int count, ConstPointer<ConstPointer<Byte>> path, ConstPointer<Integer> length) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCompileShaderIncludeARB(\{shader}, \{count}, \{path}, \{length})");
		try {
		    functions.glCompileShaderIncludeARB().invokeExact(shader, count, path, length);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCompressedMultiTexImage1DEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCompressedMultiTexImage1DEXT.xhtml">registry.khronos.org</a>
	 */
	public void glCompressedMultiTexImage1DEXT(GLTextureUnit texunit, GLTextureTarget target, int level, GLInternalFormat internalformat, int width, int border, int imageSize, ConstPointer<Void> bits) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCompressedMultiTexImage1DEXT(\{texunit}, \{target}, \{level}, \{internalformat}, \{width}, \{border}, \{imageSize}, \{bits})");
		try {
		    functions.glCompressedMultiTexImage1DEXT().invokeExact(texunit, target, level, internalformat, width, border, imageSize, bits);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were texunit=\{texunit}, target=\{target}, internalformat=\{internalformat}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCompressedMultiTexImage2DEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCompressedMultiTexImage2DEXT.xhtml">registry.khronos.org</a>
	 */
	public void glCompressedMultiTexImage2DEXT(GLTextureUnit texunit, GLTextureTarget target, int level, GLInternalFormat internalformat, int width, int height, int border, int imageSize, ConstPointer<Void> bits) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCompressedMultiTexImage2DEXT(\{texunit}, \{target}, \{level}, \{internalformat}, \{width}, \{height}, \{border}, \{imageSize}, \{bits})");
		try {
		    functions.glCompressedMultiTexImage2DEXT().invokeExact(texunit, target, level, internalformat, width, height, border, imageSize, bits);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were texunit=\{texunit}, target=\{target}, internalformat=\{internalformat}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCompressedMultiTexImage3DEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCompressedMultiTexImage3DEXT.xhtml">registry.khronos.org</a>
	 */
	public void glCompressedMultiTexImage3DEXT(GLTextureUnit texunit, GLTextureTarget target, int level, GLInternalFormat internalformat, int width, int height, int depth, int border, int imageSize, ConstPointer<Void> bits) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCompressedMultiTexImage3DEXT(\{texunit}, \{target}, \{level}, \{internalformat}, \{width}, \{height}, \{depth}, \{border}, \{imageSize}, \{bits})");
		try {
		    functions.glCompressedMultiTexImage3DEXT().invokeExact(texunit, target, level, internalformat, width, height, depth, border, imageSize, bits);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were texunit=\{texunit}, target=\{target}, internalformat=\{internalformat}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCompressedMultiTexSubImage1DEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCompressedMultiTexSubImage1DEXT.xhtml">registry.khronos.org</a>
	 */
	public void glCompressedMultiTexSubImage1DEXT(GLTextureUnit texunit, GLTextureTarget target, int level, int xoffset, int width, GLInternalFormat format, int imageSize, ConstPointer<Void> bits) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCompressedMultiTexSubImage1DEXT(\{texunit}, \{target}, \{level}, \{xoffset}, \{width}, \{format}, \{imageSize}, \{bits})");
		try {
		    functions.glCompressedMultiTexSubImage1DEXT().invokeExact(texunit, target, level, xoffset, width, format, imageSize, bits);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were texunit=\{texunit}, target=\{target}, format=\{format}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCompressedMultiTexSubImage2DEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCompressedMultiTexSubImage2DEXT.xhtml">registry.khronos.org</a>
	 */
	public void glCompressedMultiTexSubImage2DEXT(GLTextureUnit texunit, GLTextureTarget target, int level, int xoffset, int yoffset, int width, int height, GLInternalFormat format, int imageSize, ConstPointer<Void> bits) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCompressedMultiTexSubImage2DEXT(\{texunit}, \{target}, \{level}, \{xoffset}, \{yoffset}, \{width}, \{height}, \{format}, \{imageSize}, \{bits})");
		try {
		    functions.glCompressedMultiTexSubImage2DEXT().invokeExact(texunit, target, level, xoffset, yoffset, width, height, format, imageSize, bits);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were texunit=\{texunit}, target=\{target}, format=\{format}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCompressedMultiTexSubImage3DEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCompressedMultiTexSubImage3DEXT.xhtml">registry.khronos.org</a>
	 */
	public void glCompressedMultiTexSubImage3DEXT(GLTextureUnit texunit, GLTextureTarget target, int level, int xoffset, int yoffset, int zoffset, int width, int height, int depth, GLInternalFormat format, int imageSize, ConstPointer<Void> bits) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCompressedMultiTexSubImage3DEXT(\{texunit}, \{target}, \{level}, \{xoffset}, \{yoffset}, \{zoffset}, \{width}, \{height}, \{depth}, \{format}, \{imageSize}, \{bits})");
		try {
		    functions.glCompressedMultiTexSubImage3DEXT().invokeExact(texunit, target, level, xoffset, yoffset, zoffset, width, height, depth, format, imageSize, bits);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were texunit=\{texunit}, target=\{target}, format=\{format}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCompressedTexImage1D">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCompressedTexImage1D.xhtml">registry.khronos.org</a>
	 */
	public void glCompressedTexImage1D(GLTextureTarget target, int level, GLInternalFormat internalformat, int width, int border, int imageSize, ConstPointer<Void> data) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCompressedTexImage1D(\{target}, \{level}, \{internalformat}, \{width}, \{border}, \{imageSize}, \{data})");
		try {
		    functions.glCompressedTexImage1D().invokeExact(target, level, internalformat, width, border, imageSize, data);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, internalformat=\{internalformat}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCompressedTexImage1DARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCompressedTexImage1DARB.xhtml">registry.khronos.org</a>
	 */
	public void glCompressedTexImage1DARB(GLTextureTarget target, int level, GLInternalFormat internalformat, int width, int border, int imageSize, ConstPointer<Void> data) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCompressedTexImage1DARB(\{target}, \{level}, \{internalformat}, \{width}, \{border}, \{imageSize}, \{data})");
		try {
		    functions.glCompressedTexImage1DARB().invokeExact(target, level, internalformat, width, border, imageSize, data);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, internalformat=\{internalformat}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCompressedTexImage2D">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCompressedTexImage2D.xhtml">registry.khronos.org</a>
	 */
	public void glCompressedTexImage2D(GLTextureTarget target, int level, GLInternalFormat internalformat, int width, int height, int border, int imageSize, ConstPointer<Void> data) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCompressedTexImage2D(\{target}, \{level}, \{internalformat}, \{width}, \{height}, \{border}, \{imageSize}, \{data})");
		try {
		    functions.glCompressedTexImage2D().invokeExact(target, level, internalformat, width, height, border, imageSize, data);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, internalformat=\{internalformat}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCompressedTexImage2DARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCompressedTexImage2DARB.xhtml">registry.khronos.org</a>
	 */
	public void glCompressedTexImage2DARB(GLTextureTarget target, int level, GLInternalFormat internalformat, int width, int height, int border, int imageSize, ConstPointer<Void> data) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCompressedTexImage2DARB(\{target}, \{level}, \{internalformat}, \{width}, \{height}, \{border}, \{imageSize}, \{data})");
		try {
		    functions.glCompressedTexImage2DARB().invokeExact(target, level, internalformat, width, height, border, imageSize, data);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, internalformat=\{internalformat}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCompressedTexImage3D">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCompressedTexImage3D.xhtml">registry.khronos.org</a>
	 */
	public void glCompressedTexImage3D(GLTextureTarget target, int level, GLInternalFormat internalformat, int width, int height, int depth, int border, int imageSize, ConstPointer<Void> data) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCompressedTexImage3D(\{target}, \{level}, \{internalformat}, \{width}, \{height}, \{depth}, \{border}, \{imageSize}, \{data})");
		try {
		    functions.glCompressedTexImage3D().invokeExact(target, level, internalformat, width, height, depth, border, imageSize, data);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, internalformat=\{internalformat}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCompressedTexImage3DARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCompressedTexImage3DARB.xhtml">registry.khronos.org</a>
	 */
	public void glCompressedTexImage3DARB(GLTextureTarget target, int level, GLInternalFormat internalformat, int width, int height, int depth, int border, int imageSize, ConstPointer<Void> data) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCompressedTexImage3DARB(\{target}, \{level}, \{internalformat}, \{width}, \{height}, \{depth}, \{border}, \{imageSize}, \{data})");
		try {
		    functions.glCompressedTexImage3DARB().invokeExact(target, level, internalformat, width, height, depth, border, imageSize, data);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, internalformat=\{internalformat}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCompressedTexImage3DOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCompressedTexImage3DOES.xhtml">registry.khronos.org</a>
	 */
	public void glCompressedTexImage3DOES(GLTextureTarget target, int level, GLInternalFormat internalformat, int width, int height, int depth, int border, int imageSize, ConstPointer<Void> data) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCompressedTexImage3DOES(\{target}, \{level}, \{internalformat}, \{width}, \{height}, \{depth}, \{border}, \{imageSize}, \{data})");
		try {
		    functions.glCompressedTexImage3DOES().invokeExact(target, level, internalformat, width, height, depth, border, imageSize, data);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, internalformat=\{internalformat}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCompressedTexSubImage1D">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCompressedTexSubImage1D.xhtml">registry.khronos.org</a>
	 */
	public void glCompressedTexSubImage1D(GLTextureTarget target, int level, int xoffset, int width, GLInternalFormat format, int imageSize, ConstPointer<Void> data) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCompressedTexSubImage1D(\{target}, \{level}, \{xoffset}, \{width}, \{format}, \{imageSize}, \{data})");
		try {
		    functions.glCompressedTexSubImage1D().invokeExact(target, level, xoffset, width, format, imageSize, data);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, format=\{format}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCompressedTexSubImage1DARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCompressedTexSubImage1DARB.xhtml">registry.khronos.org</a>
	 */
	public void glCompressedTexSubImage1DARB(GLTextureTarget target, int level, int xoffset, int width, GLInternalFormat format, int imageSize, ConstPointer<Void> data) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCompressedTexSubImage1DARB(\{target}, \{level}, \{xoffset}, \{width}, \{format}, \{imageSize}, \{data})");
		try {
		    functions.glCompressedTexSubImage1DARB().invokeExact(target, level, xoffset, width, format, imageSize, data);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, format=\{format}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCompressedTexSubImage2D">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCompressedTexSubImage2D.xhtml">registry.khronos.org</a>
	 */
	public void glCompressedTexSubImage2D(GLTextureTarget target, int level, int xoffset, int yoffset, int width, int height, GLInternalFormat format, int imageSize, ConstPointer<Void> data) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCompressedTexSubImage2D(\{target}, \{level}, \{xoffset}, \{yoffset}, \{width}, \{height}, \{format}, \{imageSize}, \{data})");
		try {
		    functions.glCompressedTexSubImage2D().invokeExact(target, level, xoffset, yoffset, width, height, format, imageSize, data);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, format=\{format}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCompressedTexSubImage2DARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCompressedTexSubImage2DARB.xhtml">registry.khronos.org</a>
	 */
	public void glCompressedTexSubImage2DARB(GLTextureTarget target, int level, int xoffset, int yoffset, int width, int height, GLInternalFormat format, int imageSize, ConstPointer<Void> data) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCompressedTexSubImage2DARB(\{target}, \{level}, \{xoffset}, \{yoffset}, \{width}, \{height}, \{format}, \{imageSize}, \{data})");
		try {
		    functions.glCompressedTexSubImage2DARB().invokeExact(target, level, xoffset, yoffset, width, height, format, imageSize, data);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, format=\{format}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCompressedTexSubImage3D">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCompressedTexSubImage3D.xhtml">registry.khronos.org</a>
	 */
	public void glCompressedTexSubImage3D(GLTextureTarget target, int level, int xoffset, int yoffset, int zoffset, int width, int height, int depth, GLInternalFormat format, int imageSize, ConstPointer<Void> data) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCompressedTexSubImage3D(\{target}, \{level}, \{xoffset}, \{yoffset}, \{zoffset}, \{width}, \{height}, \{depth}, \{format}, \{imageSize}, \{data})");
		try {
		    functions.glCompressedTexSubImage3D().invokeExact(target, level, xoffset, yoffset, zoffset, width, height, depth, format, imageSize, data);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, format=\{format}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCompressedTexSubImage3DARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCompressedTexSubImage3DARB.xhtml">registry.khronos.org</a>
	 */
	public void glCompressedTexSubImage3DARB(GLTextureTarget target, int level, int xoffset, int yoffset, int zoffset, int width, int height, int depth, GLInternalFormat format, int imageSize, ConstPointer<Void> data) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCompressedTexSubImage3DARB(\{target}, \{level}, \{xoffset}, \{yoffset}, \{zoffset}, \{width}, \{height}, \{depth}, \{format}, \{imageSize}, \{data})");
		try {
		    functions.glCompressedTexSubImage3DARB().invokeExact(target, level, xoffset, yoffset, zoffset, width, height, depth, format, imageSize, data);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, format=\{format}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCompressedTexSubImage3DOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCompressedTexSubImage3DOES.xhtml">registry.khronos.org</a>
	 */
	public void glCompressedTexSubImage3DOES(GLTextureTarget target, int level, int xoffset, int yoffset, int zoffset, int width, int height, int depth, GLInternalFormat format, int imageSize, ConstPointer<Void> data) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCompressedTexSubImage3DOES(\{target}, \{level}, \{xoffset}, \{yoffset}, \{zoffset}, \{width}, \{height}, \{depth}, \{format}, \{imageSize}, \{data})");
		try {
		    functions.glCompressedTexSubImage3DOES().invokeExact(target, level, xoffset, yoffset, zoffset, width, height, depth, format, imageSize, data);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, format=\{format}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCompressedTextureImage1DEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCompressedTextureImage1DEXT.xhtml">registry.khronos.org</a>
	 */
	public void glCompressedTextureImage1DEXT(int texture, GLTextureTarget target, int level, GLInternalFormat internalformat, int width, int border, int imageSize, ConstPointer<Void> bits) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCompressedTextureImage1DEXT(\{texture}, \{target}, \{level}, \{internalformat}, \{width}, \{border}, \{imageSize}, \{bits})");
		try {
		    functions.glCompressedTextureImage1DEXT().invokeExact(texture, target, level, internalformat, width, border, imageSize, bits);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, internalformat=\{internalformat}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCompressedTextureImage2DEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCompressedTextureImage2DEXT.xhtml">registry.khronos.org</a>
	 */
	public void glCompressedTextureImage2DEXT(int texture, GLTextureTarget target, int level, GLInternalFormat internalformat, int width, int height, int border, int imageSize, ConstPointer<Void> bits) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCompressedTextureImage2DEXT(\{texture}, \{target}, \{level}, \{internalformat}, \{width}, \{height}, \{border}, \{imageSize}, \{bits})");
		try {
		    functions.glCompressedTextureImage2DEXT().invokeExact(texture, target, level, internalformat, width, height, border, imageSize, bits);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, internalformat=\{internalformat}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCompressedTextureImage3DEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCompressedTextureImage3DEXT.xhtml">registry.khronos.org</a>
	 */
	public void glCompressedTextureImage3DEXT(int texture, GLTextureTarget target, int level, GLInternalFormat internalformat, int width, int height, int depth, int border, int imageSize, ConstPointer<Void> bits) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCompressedTextureImage3DEXT(\{texture}, \{target}, \{level}, \{internalformat}, \{width}, \{height}, \{depth}, \{border}, \{imageSize}, \{bits})");
		try {
		    functions.glCompressedTextureImage3DEXT().invokeExact(texture, target, level, internalformat, width, height, depth, border, imageSize, bits);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, internalformat=\{internalformat}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCompressedTextureSubImage1D">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCompressedTextureSubImage1D.xhtml">registry.khronos.org</a>
	 */
	public void glCompressedTextureSubImage1D(int texture, int level, int xoffset, int width, GLInternalFormat format, int imageSize, ConstPointer<Void> data) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCompressedTextureSubImage1D(\{texture}, \{level}, \{xoffset}, \{width}, \{format}, \{imageSize}, \{data})");
		try {
		    functions.glCompressedTextureSubImage1D().invokeExact(texture, level, xoffset, width, format, imageSize, data);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were format=\{format}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCompressedTextureSubImage1DEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCompressedTextureSubImage1DEXT.xhtml">registry.khronos.org</a>
	 */
	public void glCompressedTextureSubImage1DEXT(int texture, GLTextureTarget target, int level, int xoffset, int width, GLInternalFormat format, int imageSize, ConstPointer<Void> bits) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCompressedTextureSubImage1DEXT(\{texture}, \{target}, \{level}, \{xoffset}, \{width}, \{format}, \{imageSize}, \{bits})");
		try {
		    functions.glCompressedTextureSubImage1DEXT().invokeExact(texture, target, level, xoffset, width, format, imageSize, bits);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, format=\{format}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCompressedTextureSubImage2D">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCompressedTextureSubImage2D.xhtml">registry.khronos.org</a>
	 */
	public void glCompressedTextureSubImage2D(int texture, int level, int xoffset, int yoffset, int width, int height, GLInternalFormat format, int imageSize, ConstPointer<Void> data) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCompressedTextureSubImage2D(\{texture}, \{level}, \{xoffset}, \{yoffset}, \{width}, \{height}, \{format}, \{imageSize}, \{data})");
		try {
		    functions.glCompressedTextureSubImage2D().invokeExact(texture, level, xoffset, yoffset, width, height, format, imageSize, data);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were format=\{format}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCompressedTextureSubImage2DEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCompressedTextureSubImage2DEXT.xhtml">registry.khronos.org</a>
	 */
	public void glCompressedTextureSubImage2DEXT(int texture, GLTextureTarget target, int level, int xoffset, int yoffset, int width, int height, GLInternalFormat format, int imageSize, ConstPointer<Void> bits) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCompressedTextureSubImage2DEXT(\{texture}, \{target}, \{level}, \{xoffset}, \{yoffset}, \{width}, \{height}, \{format}, \{imageSize}, \{bits})");
		try {
		    functions.glCompressedTextureSubImage2DEXT().invokeExact(texture, target, level, xoffset, yoffset, width, height, format, imageSize, bits);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, format=\{format}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCompressedTextureSubImage3D">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCompressedTextureSubImage3D.xhtml">registry.khronos.org</a>
	 */
	public void glCompressedTextureSubImage3D(int texture, int level, int xoffset, int yoffset, int zoffset, int width, int height, int depth, GLInternalFormat format, int imageSize, ConstPointer<Void> data) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCompressedTextureSubImage3D(\{texture}, \{level}, \{xoffset}, \{yoffset}, \{zoffset}, \{width}, \{height}, \{depth}, \{format}, \{imageSize}, \{data})");
		try {
		    functions.glCompressedTextureSubImage3D().invokeExact(texture, level, xoffset, yoffset, zoffset, width, height, depth, format, imageSize, data);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were format=\{format}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCompressedTextureSubImage3DEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCompressedTextureSubImage3DEXT.xhtml">registry.khronos.org</a>
	 */
	public void glCompressedTextureSubImage3DEXT(int texture, GLTextureTarget target, int level, int xoffset, int yoffset, int zoffset, int width, int height, int depth, GLInternalFormat format, int imageSize, ConstPointer<Void> bits) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCompressedTextureSubImage3DEXT(\{texture}, \{target}, \{level}, \{xoffset}, \{yoffset}, \{zoffset}, \{width}, \{height}, \{depth}, \{format}, \{imageSize}, \{bits})");
		try {
		    functions.glCompressedTextureSubImage3DEXT().invokeExact(texture, target, level, xoffset, yoffset, zoffset, width, height, depth, format, imageSize, bits);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, format=\{format}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glConservativeRasterParameterfNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glConservativeRasterParameterfNV.xhtml">registry.khronos.org</a>
	 */
	public void glConservativeRasterParameterfNV(GLEnum pname, float value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glConservativeRasterParameterfNV(\{pname}, \{value})");
		try {
		    functions.glConservativeRasterParameterfNV().invokeExact(pname, value);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glConservativeRasterParameteriNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glConservativeRasterParameteriNV.xhtml">registry.khronos.org</a>
	 */
	public void glConservativeRasterParameteriNV(GLEnum pname, int param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glConservativeRasterParameteriNV(\{pname}, \{param})");
		try {
		    functions.glConservativeRasterParameteriNV().invokeExact(pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glConvolutionFilter1D">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glConvolutionFilter1D.xhtml">registry.khronos.org</a>
	 */
	public void glConvolutionFilter1D(GLConvolutionTarget target, GLInternalFormat internalformat, int width, GLPixelFormat format, GLPixelType type, ConstPointer<Void> image) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glConvolutionFilter1D(\{target}, \{internalformat}, \{width}, \{format}, \{type}, \{image})");
		try {
		    functions.glConvolutionFilter1D().invokeExact(target, internalformat, width, format, type, image);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, internalformat=\{internalformat}, format=\{format}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glConvolutionFilter1DEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glConvolutionFilter1DEXT.xhtml">registry.khronos.org</a>
	 */
	public void glConvolutionFilter1DEXT(GLConvolutionTargetEXT target, GLInternalFormat internalformat, int width, GLPixelFormat format, GLPixelType type, ConstPointer<Void> image) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glConvolutionFilter1DEXT(\{target}, \{internalformat}, \{width}, \{format}, \{type}, \{image})");
		try {
		    functions.glConvolutionFilter1DEXT().invokeExact(target, internalformat, width, format, type, image);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, internalformat=\{internalformat}, format=\{format}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glConvolutionFilter2D">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glConvolutionFilter2D.xhtml">registry.khronos.org</a>
	 */
	public void glConvolutionFilter2D(GLConvolutionTarget target, GLInternalFormat internalformat, int width, int height, GLPixelFormat format, GLPixelType type, ConstPointer<Void> image) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glConvolutionFilter2D(\{target}, \{internalformat}, \{width}, \{height}, \{format}, \{type}, \{image})");
		try {
		    functions.glConvolutionFilter2D().invokeExact(target, internalformat, width, height, format, type, image);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, internalformat=\{internalformat}, format=\{format}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glConvolutionFilter2DEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glConvolutionFilter2DEXT.xhtml">registry.khronos.org</a>
	 */
	public void glConvolutionFilter2DEXT(GLConvolutionTargetEXT target, GLInternalFormat internalformat, int width, int height, GLPixelFormat format, GLPixelType type, ConstPointer<Void> image) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glConvolutionFilter2DEXT(\{target}, \{internalformat}, \{width}, \{height}, \{format}, \{type}, \{image})");
		try {
		    functions.glConvolutionFilter2DEXT().invokeExact(target, internalformat, width, height, format, type, image);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, internalformat=\{internalformat}, format=\{format}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glConvolutionParameterf">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glConvolutionParameterf.xhtml">registry.khronos.org</a>
	 */
	public void glConvolutionParameterf(GLConvolutionTarget target, GLConvolutionParameter pname, float params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glConvolutionParameterf(\{target}, \{pname}, \{params})");
		try {
		    functions.glConvolutionParameterf().invokeExact(target, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glConvolutionParameterfEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glConvolutionParameterfEXT.xhtml">registry.khronos.org</a>
	 */
	public void glConvolutionParameterfEXT(GLConvolutionTargetEXT target, GLConvolutionParameter pname, float params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glConvolutionParameterfEXT(\{target}, \{pname}, \{params})");
		try {
		    functions.glConvolutionParameterfEXT().invokeExact(target, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glConvolutionParameterfv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glConvolutionParameterfv.xhtml">registry.khronos.org</a>
	 */
	public void glConvolutionParameterfv(GLConvolutionTarget target, GLConvolutionParameter pname, ConstPointer<Float> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glConvolutionParameterfv(\{target}, \{pname}, \{params})");
		try {
		    functions.glConvolutionParameterfv().invokeExact(target, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glConvolutionParameterfvEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glConvolutionParameterfvEXT.xhtml">registry.khronos.org</a>
	 */
	public void glConvolutionParameterfvEXT(GLConvolutionTargetEXT target, GLConvolutionParameter pname, ConstPointer<Float> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glConvolutionParameterfvEXT(\{target}, \{pname}, \{params})");
		try {
		    functions.glConvolutionParameterfvEXT().invokeExact(target, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glConvolutionParameteri">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glConvolutionParameteri.xhtml">registry.khronos.org</a>
	 */
	public void glConvolutionParameteri(GLConvolutionTarget target, GLConvolutionParameter pname, int params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glConvolutionParameteri(\{target}, \{pname}, \{params})");
		try {
		    functions.glConvolutionParameteri().invokeExact(target, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glConvolutionParameteriEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glConvolutionParameteriEXT.xhtml">registry.khronos.org</a>
	 */
	public void glConvolutionParameteriEXT(GLConvolutionTargetEXT target, GLConvolutionParameter pname, int params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glConvolutionParameteriEXT(\{target}, \{pname}, \{params})");
		try {
		    functions.glConvolutionParameteriEXT().invokeExact(target, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glConvolutionParameteriv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glConvolutionParameteriv.xhtml">registry.khronos.org</a>
	 */
	public void glConvolutionParameteriv(GLConvolutionTarget target, GLConvolutionParameter pname, ConstPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glConvolutionParameteriv(\{target}, \{pname}, \{params})");
		try {
		    functions.glConvolutionParameteriv().invokeExact(target, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glConvolutionParameterivEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glConvolutionParameterivEXT.xhtml">registry.khronos.org</a>
	 */
	public void glConvolutionParameterivEXT(GLConvolutionTargetEXT target, GLConvolutionParameter pname, ConstPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glConvolutionParameterivEXT(\{target}, \{pname}, \{params})");
		try {
		    functions.glConvolutionParameterivEXT().invokeExact(target, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glConvolutionParameterxOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glConvolutionParameterxOES.xhtml">registry.khronos.org</a>
	 */
	public void glConvolutionParameterxOES(GLConvolutionTargetEXT target, GLConvolutionParameter pname, int param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glConvolutionParameterxOES(\{target}, \{pname}, \{param})");
		try {
		    functions.glConvolutionParameterxOES().invokeExact(target, pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glConvolutionParameterxvOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glConvolutionParameterxvOES.xhtml">registry.khronos.org</a>
	 */
	public void glConvolutionParameterxvOES(GLConvolutionTargetEXT target, GLConvolutionParameter pname, ConstPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glConvolutionParameterxvOES(\{target}, \{pname}, \{params})");
		try {
		    functions.glConvolutionParameterxvOES().invokeExact(target, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCopyBufferSubData">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCopyBufferSubData.xhtml">registry.khronos.org</a>
	 */
	public void glCopyBufferSubData(GLCopyBufferSubDataTarget readTarget, GLCopyBufferSubDataTarget writeTarget, long readOffset, long writeOffset, long size) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCopyBufferSubData(\{readTarget}, \{writeTarget}, \{readOffset}, \{writeOffset}, \{size})");
		try {
		    functions.glCopyBufferSubData().invokeExact(readTarget, writeTarget, readOffset, writeOffset, size);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were readTarget=\{readTarget}, writeTarget=\{writeTarget}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCopyBufferSubDataNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCopyBufferSubDataNV.xhtml">registry.khronos.org</a>
	 */
	public void glCopyBufferSubDataNV(GLCopyBufferSubDataTarget readTarget, GLCopyBufferSubDataTarget writeTarget, long readOffset, long writeOffset, long size) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCopyBufferSubDataNV(\{readTarget}, \{writeTarget}, \{readOffset}, \{writeOffset}, \{size})");
		try {
		    functions.glCopyBufferSubDataNV().invokeExact(readTarget, writeTarget, readOffset, writeOffset, size);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were readTarget=\{readTarget}, writeTarget=\{writeTarget}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCopyColorSubTable">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCopyColorSubTable.xhtml">registry.khronos.org</a>
	 */
	public void glCopyColorSubTable(GLColorTableTarget target, int start, int x, int y, int width) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCopyColorSubTable(\{target}, \{start}, \{x}, \{y}, \{width})");
		try {
		    functions.glCopyColorSubTable().invokeExact(target, start, x, y, width);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCopyColorSubTableEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCopyColorSubTableEXT.xhtml">registry.khronos.org</a>
	 */
	public void glCopyColorSubTableEXT(GLColorTableTarget target, int start, int x, int y, int width) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCopyColorSubTableEXT(\{target}, \{start}, \{x}, \{y}, \{width})");
		try {
		    functions.glCopyColorSubTableEXT().invokeExact(target, start, x, y, width);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCopyColorTable">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCopyColorTable.xhtml">registry.khronos.org</a>
	 */
	public void glCopyColorTable(GLColorTableTarget target, GLInternalFormat internalformat, int x, int y, int width) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCopyColorTable(\{target}, \{internalformat}, \{x}, \{y}, \{width})");
		try {
		    functions.glCopyColorTable().invokeExact(target, internalformat, x, y, width);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, internalformat=\{internalformat}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCopyColorTableSGI">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCopyColorTableSGI.xhtml">registry.khronos.org</a>
	 */
	public void glCopyColorTableSGI(GLColorTableTargetSGI target, GLInternalFormat internalformat, int x, int y, int width) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCopyColorTableSGI(\{target}, \{internalformat}, \{x}, \{y}, \{width})");
		try {
		    functions.glCopyColorTableSGI().invokeExact(target, internalformat, x, y, width);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, internalformat=\{internalformat}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCopyConvolutionFilter1D">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCopyConvolutionFilter1D.xhtml">registry.khronos.org</a>
	 */
	public void glCopyConvolutionFilter1D(GLConvolutionTarget target, GLInternalFormat internalformat, int x, int y, int width) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCopyConvolutionFilter1D(\{target}, \{internalformat}, \{x}, \{y}, \{width})");
		try {
		    functions.glCopyConvolutionFilter1D().invokeExact(target, internalformat, x, y, width);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, internalformat=\{internalformat}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCopyConvolutionFilter1DEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCopyConvolutionFilter1DEXT.xhtml">registry.khronos.org</a>
	 */
	public void glCopyConvolutionFilter1DEXT(GLConvolutionTargetEXT target, GLInternalFormat internalformat, int x, int y, int width) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCopyConvolutionFilter1DEXT(\{target}, \{internalformat}, \{x}, \{y}, \{width})");
		try {
		    functions.glCopyConvolutionFilter1DEXT().invokeExact(target, internalformat, x, y, width);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, internalformat=\{internalformat}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCopyConvolutionFilter2D">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCopyConvolutionFilter2D.xhtml">registry.khronos.org</a>
	 */
	public void glCopyConvolutionFilter2D(GLConvolutionTarget target, GLInternalFormat internalformat, int x, int y, int width, int height) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCopyConvolutionFilter2D(\{target}, \{internalformat}, \{x}, \{y}, \{width}, \{height})");
		try {
		    functions.glCopyConvolutionFilter2D().invokeExact(target, internalformat, x, y, width, height);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, internalformat=\{internalformat}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCopyConvolutionFilter2DEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCopyConvolutionFilter2DEXT.xhtml">registry.khronos.org</a>
	 */
	public void glCopyConvolutionFilter2DEXT(GLConvolutionTargetEXT target, GLInternalFormat internalformat, int x, int y, int width, int height) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCopyConvolutionFilter2DEXT(\{target}, \{internalformat}, \{x}, \{y}, \{width}, \{height})");
		try {
		    functions.glCopyConvolutionFilter2DEXT().invokeExact(target, internalformat, x, y, width, height);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, internalformat=\{internalformat}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCopyImageSubData">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCopyImageSubData.xhtml">registry.khronos.org</a>
	 */
	public void glCopyImageSubData(int srcName, GLCopyImageSubDataTarget srcTarget, int srcLevel, int srcX, int srcY, int srcZ, int dstName, GLCopyImageSubDataTarget dstTarget, int dstLevel, int dstX, int dstY, int dstZ, int srcWidth, int srcHeight, int srcDepth) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCopyImageSubData(\{srcName}, \{srcTarget}, \{srcLevel}, \{srcX}, \{srcY}, \{srcZ}, \{dstName}, \{dstTarget}, \{dstLevel}, \{dstX}, \{dstY}, \{dstZ}, \{srcWidth}, \{srcHeight}, \{srcDepth})");
		try {
		    functions.glCopyImageSubData().invokeExact(srcName, srcTarget, srcLevel, srcX, srcY, srcZ, dstName, dstTarget, dstLevel, dstX, dstY, dstZ, srcWidth, srcHeight, srcDepth);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were srcTarget=\{srcTarget}, dstTarget=\{dstTarget}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCopyImageSubDataEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCopyImageSubDataEXT.xhtml">registry.khronos.org</a>
	 */
	public void glCopyImageSubDataEXT(int srcName, GLCopyBufferSubDataTarget srcTarget, int srcLevel, int srcX, int srcY, int srcZ, int dstName, GLCopyBufferSubDataTarget dstTarget, int dstLevel, int dstX, int dstY, int dstZ, int srcWidth, int srcHeight, int srcDepth) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCopyImageSubDataEXT(\{srcName}, \{srcTarget}, \{srcLevel}, \{srcX}, \{srcY}, \{srcZ}, \{dstName}, \{dstTarget}, \{dstLevel}, \{dstX}, \{dstY}, \{dstZ}, \{srcWidth}, \{srcHeight}, \{srcDepth})");
		try {
		    functions.glCopyImageSubDataEXT().invokeExact(srcName, srcTarget, srcLevel, srcX, srcY, srcZ, dstName, dstTarget, dstLevel, dstX, dstY, dstZ, srcWidth, srcHeight, srcDepth);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were srcTarget=\{srcTarget}, dstTarget=\{dstTarget}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCopyImageSubDataNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCopyImageSubDataNV.xhtml">registry.khronos.org</a>
	 */
	public void glCopyImageSubDataNV(int srcName, GLCopyBufferSubDataTarget srcTarget, int srcLevel, int srcX, int srcY, int srcZ, int dstName, GLCopyBufferSubDataTarget dstTarget, int dstLevel, int dstX, int dstY, int dstZ, int width, int height, int depth) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCopyImageSubDataNV(\{srcName}, \{srcTarget}, \{srcLevel}, \{srcX}, \{srcY}, \{srcZ}, \{dstName}, \{dstTarget}, \{dstLevel}, \{dstX}, \{dstY}, \{dstZ}, \{width}, \{height}, \{depth})");
		try {
		    functions.glCopyImageSubDataNV().invokeExact(srcName, srcTarget, srcLevel, srcX, srcY, srcZ, dstName, dstTarget, dstLevel, dstX, dstY, dstZ, width, height, depth);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were srcTarget=\{srcTarget}, dstTarget=\{dstTarget}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCopyImageSubDataOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCopyImageSubDataOES.xhtml">registry.khronos.org</a>
	 */
	public void glCopyImageSubDataOES(int srcName, GLCopyBufferSubDataTarget srcTarget, int srcLevel, int srcX, int srcY, int srcZ, int dstName, GLCopyBufferSubDataTarget dstTarget, int dstLevel, int dstX, int dstY, int dstZ, int srcWidth, int srcHeight, int srcDepth) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCopyImageSubDataOES(\{srcName}, \{srcTarget}, \{srcLevel}, \{srcX}, \{srcY}, \{srcZ}, \{dstName}, \{dstTarget}, \{dstLevel}, \{dstX}, \{dstY}, \{dstZ}, \{srcWidth}, \{srcHeight}, \{srcDepth})");
		try {
		    functions.glCopyImageSubDataOES().invokeExact(srcName, srcTarget, srcLevel, srcX, srcY, srcZ, dstName, dstTarget, dstLevel, dstX, dstY, dstZ, srcWidth, srcHeight, srcDepth);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were srcTarget=\{srcTarget}, dstTarget=\{dstTarget}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCopyMultiTexImage1DEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCopyMultiTexImage1DEXT.xhtml">registry.khronos.org</a>
	 */
	public void glCopyMultiTexImage1DEXT(GLTextureUnit texunit, GLTextureTarget target, int level, GLInternalFormat internalformat, int x, int y, int width, int border) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCopyMultiTexImage1DEXT(\{texunit}, \{target}, \{level}, \{internalformat}, \{x}, \{y}, \{width}, \{border})");
		try {
		    functions.glCopyMultiTexImage1DEXT().invokeExact(texunit, target, level, internalformat, x, y, width, border);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were texunit=\{texunit}, target=\{target}, internalformat=\{internalformat}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCopyMultiTexImage2DEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCopyMultiTexImage2DEXT.xhtml">registry.khronos.org</a>
	 */
	public void glCopyMultiTexImage2DEXT(GLTextureUnit texunit, GLTextureTarget target, int level, GLInternalFormat internalformat, int x, int y, int width, int height, int border) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCopyMultiTexImage2DEXT(\{texunit}, \{target}, \{level}, \{internalformat}, \{x}, \{y}, \{width}, \{height}, \{border})");
		try {
		    functions.glCopyMultiTexImage2DEXT().invokeExact(texunit, target, level, internalformat, x, y, width, height, border);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were texunit=\{texunit}, target=\{target}, internalformat=\{internalformat}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCopyMultiTexSubImage1DEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCopyMultiTexSubImage1DEXT.xhtml">registry.khronos.org</a>
	 */
	public void glCopyMultiTexSubImage1DEXT(GLTextureUnit texunit, GLTextureTarget target, int level, int xoffset, int x, int y, int width) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCopyMultiTexSubImage1DEXT(\{texunit}, \{target}, \{level}, \{xoffset}, \{x}, \{y}, \{width})");
		try {
		    functions.glCopyMultiTexSubImage1DEXT().invokeExact(texunit, target, level, xoffset, x, y, width);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were texunit=\{texunit}, target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCopyMultiTexSubImage2DEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCopyMultiTexSubImage2DEXT.xhtml">registry.khronos.org</a>
	 */
	public void glCopyMultiTexSubImage2DEXT(GLTextureUnit texunit, GLTextureTarget target, int level, int xoffset, int yoffset, int x, int y, int width, int height) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCopyMultiTexSubImage2DEXT(\{texunit}, \{target}, \{level}, \{xoffset}, \{yoffset}, \{x}, \{y}, \{width}, \{height})");
		try {
		    functions.glCopyMultiTexSubImage2DEXT().invokeExact(texunit, target, level, xoffset, yoffset, x, y, width, height);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were texunit=\{texunit}, target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCopyMultiTexSubImage3DEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCopyMultiTexSubImage3DEXT.xhtml">registry.khronos.org</a>
	 */
	public void glCopyMultiTexSubImage3DEXT(GLTextureUnit texunit, GLTextureTarget target, int level, int xoffset, int yoffset, int zoffset, int x, int y, int width, int height) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCopyMultiTexSubImage3DEXT(\{texunit}, \{target}, \{level}, \{xoffset}, \{yoffset}, \{zoffset}, \{x}, \{y}, \{width}, \{height})");
		try {
		    functions.glCopyMultiTexSubImage3DEXT().invokeExact(texunit, target, level, xoffset, yoffset, zoffset, x, y, width, height);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were texunit=\{texunit}, target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCopyNamedBufferSubData">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCopyNamedBufferSubData.xhtml">registry.khronos.org</a>
	 */
	public void glCopyNamedBufferSubData(int readBuffer, int writeBuffer, long readOffset, long writeOffset, long size) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCopyNamedBufferSubData(\{readBuffer}, \{writeBuffer}, \{readOffset}, \{writeOffset}, \{size})");
		try {
		    functions.glCopyNamedBufferSubData().invokeExact(readBuffer, writeBuffer, readOffset, writeOffset, size);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCopyPathNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCopyPathNV.xhtml">registry.khronos.org</a>
	 */
	public void glCopyPathNV(int resultPath, int srcPath) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCopyPathNV(\{resultPath}, \{srcPath})");
		try {
		    functions.glCopyPathNV().invokeExact(resultPath, srcPath);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCopyPixels">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCopyPixels.xhtml">registry.khronos.org</a>
	 */
	public void glCopyPixels(int x, int y, int width, int height, GLPixelCopyType type) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCopyPixels(\{x}, \{y}, \{width}, \{height}, \{type})");
		try {
		    functions.glCopyPixels().invokeExact(x, y, width, height, type);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCopyTexImage1D">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCopyTexImage1D.xhtml">registry.khronos.org</a>
	 */
	public void glCopyTexImage1D(GLTextureTarget target, int level, GLInternalFormat internalformat, int x, int y, int width, int border) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCopyTexImage1D(\{target}, \{level}, \{internalformat}, \{x}, \{y}, \{width}, \{border})");
		try {
		    functions.glCopyTexImage1D().invokeExact(target, level, internalformat, x, y, width, border);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, internalformat=\{internalformat}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCopyTexImage1DEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCopyTexImage1DEXT.xhtml">registry.khronos.org</a>
	 */
	public void glCopyTexImage1DEXT(GLTextureTarget target, int level, GLInternalFormat internalformat, int x, int y, int width, int border) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCopyTexImage1DEXT(\{target}, \{level}, \{internalformat}, \{x}, \{y}, \{width}, \{border})");
		try {
		    functions.glCopyTexImage1DEXT().invokeExact(target, level, internalformat, x, y, width, border);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, internalformat=\{internalformat}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCopyTexImage2D">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCopyTexImage2D.xhtml">registry.khronos.org</a>
	 */
	public void glCopyTexImage2D(GLTextureTarget target, int level, GLInternalFormat internalformat, int x, int y, int width, int height, int border) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCopyTexImage2D(\{target}, \{level}, \{internalformat}, \{x}, \{y}, \{width}, \{height}, \{border})");
		try {
		    functions.glCopyTexImage2D().invokeExact(target, level, internalformat, x, y, width, height, border);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, internalformat=\{internalformat}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCopyTexImage2DEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCopyTexImage2DEXT.xhtml">registry.khronos.org</a>
	 */
	public void glCopyTexImage2DEXT(GLTextureTarget target, int level, GLInternalFormat internalformat, int x, int y, int width, int height, int border) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCopyTexImage2DEXT(\{target}, \{level}, \{internalformat}, \{x}, \{y}, \{width}, \{height}, \{border})");
		try {
		    functions.glCopyTexImage2DEXT().invokeExact(target, level, internalformat, x, y, width, height, border);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, internalformat=\{internalformat}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCopyTexSubImage1D">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCopyTexSubImage1D.xhtml">registry.khronos.org</a>
	 */
	public void glCopyTexSubImage1D(GLTextureTarget target, int level, int xoffset, int x, int y, int width) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCopyTexSubImage1D(\{target}, \{level}, \{xoffset}, \{x}, \{y}, \{width})");
		try {
		    functions.glCopyTexSubImage1D().invokeExact(target, level, xoffset, x, y, width);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCopyTexSubImage1DEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCopyTexSubImage1DEXT.xhtml">registry.khronos.org</a>
	 */
	public void glCopyTexSubImage1DEXT(GLTextureTarget target, int level, int xoffset, int x, int y, int width) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCopyTexSubImage1DEXT(\{target}, \{level}, \{xoffset}, \{x}, \{y}, \{width})");
		try {
		    functions.glCopyTexSubImage1DEXT().invokeExact(target, level, xoffset, x, y, width);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCopyTexSubImage2D">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCopyTexSubImage2D.xhtml">registry.khronos.org</a>
	 */
	public void glCopyTexSubImage2D(GLTextureTarget target, int level, int xoffset, int yoffset, int x, int y, int width, int height) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCopyTexSubImage2D(\{target}, \{level}, \{xoffset}, \{yoffset}, \{x}, \{y}, \{width}, \{height})");
		try {
		    functions.glCopyTexSubImage2D().invokeExact(target, level, xoffset, yoffset, x, y, width, height);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCopyTexSubImage2DEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCopyTexSubImage2DEXT.xhtml">registry.khronos.org</a>
	 */
	public void glCopyTexSubImage2DEXT(GLTextureTarget target, int level, int xoffset, int yoffset, int x, int y, int width, int height) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCopyTexSubImage2DEXT(\{target}, \{level}, \{xoffset}, \{yoffset}, \{x}, \{y}, \{width}, \{height})");
		try {
		    functions.glCopyTexSubImage2DEXT().invokeExact(target, level, xoffset, yoffset, x, y, width, height);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCopyTexSubImage3D">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCopyTexSubImage3D.xhtml">registry.khronos.org</a>
	 */
	public void glCopyTexSubImage3D(GLTextureTarget target, int level, int xoffset, int yoffset, int zoffset, int x, int y, int width, int height) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCopyTexSubImage3D(\{target}, \{level}, \{xoffset}, \{yoffset}, \{zoffset}, \{x}, \{y}, \{width}, \{height})");
		try {
		    functions.glCopyTexSubImage3D().invokeExact(target, level, xoffset, yoffset, zoffset, x, y, width, height);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCopyTexSubImage3DEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCopyTexSubImage3DEXT.xhtml">registry.khronos.org</a>
	 */
	public void glCopyTexSubImage3DEXT(GLTextureTarget target, int level, int xoffset, int yoffset, int zoffset, int x, int y, int width, int height) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCopyTexSubImage3DEXT(\{target}, \{level}, \{xoffset}, \{yoffset}, \{zoffset}, \{x}, \{y}, \{width}, \{height})");
		try {
		    functions.glCopyTexSubImage3DEXT().invokeExact(target, level, xoffset, yoffset, zoffset, x, y, width, height);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCopyTexSubImage3DOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCopyTexSubImage3DOES.xhtml">registry.khronos.org</a>
	 */
	public void glCopyTexSubImage3DOES(GLEnum target, int level, int xoffset, int yoffset, int zoffset, int x, int y, int width, int height) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCopyTexSubImage3DOES(\{target}, \{level}, \{xoffset}, \{yoffset}, \{zoffset}, \{x}, \{y}, \{width}, \{height})");
		try {
		    functions.glCopyTexSubImage3DOES().invokeExact(target, level, xoffset, yoffset, zoffset, x, y, width, height);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCopyTextureImage1DEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCopyTextureImage1DEXT.xhtml">registry.khronos.org</a>
	 */
	public void glCopyTextureImage1DEXT(int texture, GLTextureTarget target, int level, GLInternalFormat internalformat, int x, int y, int width, int border) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCopyTextureImage1DEXT(\{texture}, \{target}, \{level}, \{internalformat}, \{x}, \{y}, \{width}, \{border})");
		try {
		    functions.glCopyTextureImage1DEXT().invokeExact(texture, target, level, internalformat, x, y, width, border);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, internalformat=\{internalformat}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCopyTextureImage2DEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCopyTextureImage2DEXT.xhtml">registry.khronos.org</a>
	 */
	public void glCopyTextureImage2DEXT(int texture, GLTextureTarget target, int level, GLInternalFormat internalformat, int x, int y, int width, int height, int border) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCopyTextureImage2DEXT(\{texture}, \{target}, \{level}, \{internalformat}, \{x}, \{y}, \{width}, \{height}, \{border})");
		try {
		    functions.glCopyTextureImage2DEXT().invokeExact(texture, target, level, internalformat, x, y, width, height, border);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, internalformat=\{internalformat}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCopyTextureLevelsAPPLE">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCopyTextureLevelsAPPLE.xhtml">registry.khronos.org</a>
	 */
	public void glCopyTextureLevelsAPPLE(int destinationTexture, int sourceTexture, int sourceBaseLevel, int sourceLevelCount) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCopyTextureLevelsAPPLE(\{destinationTexture}, \{sourceTexture}, \{sourceBaseLevel}, \{sourceLevelCount})");
		try {
		    functions.glCopyTextureLevelsAPPLE().invokeExact(destinationTexture, sourceTexture, sourceBaseLevel, sourceLevelCount);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCopyTextureSubImage1D">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCopyTextureSubImage1D.xhtml">registry.khronos.org</a>
	 */
	public void glCopyTextureSubImage1D(int texture, int level, int xoffset, int x, int y, int width) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCopyTextureSubImage1D(\{texture}, \{level}, \{xoffset}, \{x}, \{y}, \{width})");
		try {
		    functions.glCopyTextureSubImage1D().invokeExact(texture, level, xoffset, x, y, width);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCopyTextureSubImage1DEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCopyTextureSubImage1DEXT.xhtml">registry.khronos.org</a>
	 */
	public void glCopyTextureSubImage1DEXT(int texture, GLTextureTarget target, int level, int xoffset, int x, int y, int width) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCopyTextureSubImage1DEXT(\{texture}, \{target}, \{level}, \{xoffset}, \{x}, \{y}, \{width})");
		try {
		    functions.glCopyTextureSubImage1DEXT().invokeExact(texture, target, level, xoffset, x, y, width);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCopyTextureSubImage2D">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCopyTextureSubImage2D.xhtml">registry.khronos.org</a>
	 */
	public void glCopyTextureSubImage2D(int texture, int level, int xoffset, int yoffset, int x, int y, int width, int height) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCopyTextureSubImage2D(\{texture}, \{level}, \{xoffset}, \{yoffset}, \{x}, \{y}, \{width}, \{height})");
		try {
		    functions.glCopyTextureSubImage2D().invokeExact(texture, level, xoffset, yoffset, x, y, width, height);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCopyTextureSubImage2DEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCopyTextureSubImage2DEXT.xhtml">registry.khronos.org</a>
	 */
	public void glCopyTextureSubImage2DEXT(int texture, GLTextureTarget target, int level, int xoffset, int yoffset, int x, int y, int width, int height) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCopyTextureSubImage2DEXT(\{texture}, \{target}, \{level}, \{xoffset}, \{yoffset}, \{x}, \{y}, \{width}, \{height})");
		try {
		    functions.glCopyTextureSubImage2DEXT().invokeExact(texture, target, level, xoffset, yoffset, x, y, width, height);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCopyTextureSubImage3D">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCopyTextureSubImage3D.xhtml">registry.khronos.org</a>
	 */
	public void glCopyTextureSubImage3D(int texture, int level, int xoffset, int yoffset, int zoffset, int x, int y, int width, int height) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCopyTextureSubImage3D(\{texture}, \{level}, \{xoffset}, \{yoffset}, \{zoffset}, \{x}, \{y}, \{width}, \{height})");
		try {
		    functions.glCopyTextureSubImage3D().invokeExact(texture, level, xoffset, yoffset, zoffset, x, y, width, height);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCopyTextureSubImage3DEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCopyTextureSubImage3DEXT.xhtml">registry.khronos.org</a>
	 */
	public void glCopyTextureSubImage3DEXT(int texture, GLTextureTarget target, int level, int xoffset, int yoffset, int zoffset, int x, int y, int width, int height) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCopyTextureSubImage3DEXT(\{texture}, \{target}, \{level}, \{xoffset}, \{yoffset}, \{zoffset}, \{x}, \{y}, \{width}, \{height})");
		try {
		    functions.glCopyTextureSubImage3DEXT().invokeExact(texture, target, level, xoffset, yoffset, zoffset, x, y, width, height);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCoverFillPathInstancedNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCoverFillPathInstancedNV.xhtml">registry.khronos.org</a>
	 */
	public void glCoverFillPathInstancedNV(int numPaths, GLPathElementType pathNameType, ConstPointer<Void> paths, int pathBase, GLInstancedPathCoverMode coverMode, GLPathTransformType transformType, ConstPointer<Float> transformValues) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCoverFillPathInstancedNV(\{numPaths}, \{pathNameType}, \{paths}, \{pathBase}, \{coverMode}, \{transformType}, \{transformValues})");
		try {
		    functions.glCoverFillPathInstancedNV().invokeExact(numPaths, pathNameType, paths, pathBase, coverMode, transformType, transformValues);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pathNameType=\{pathNameType}, coverMode=\{coverMode}, transformType=\{transformType}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCoverFillPathNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCoverFillPathNV.xhtml">registry.khronos.org</a>
	 */
	public void glCoverFillPathNV(int path, GLPathCoverMode coverMode) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCoverFillPathNV(\{path}, \{coverMode})");
		try {
		    functions.glCoverFillPathNV().invokeExact(path, coverMode);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were coverMode=\{coverMode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCoverStrokePathInstancedNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCoverStrokePathInstancedNV.xhtml">registry.khronos.org</a>
	 */
	public void glCoverStrokePathInstancedNV(int numPaths, GLPathElementType pathNameType, ConstPointer<Void> paths, int pathBase, GLInstancedPathCoverMode coverMode, GLPathTransformType transformType, ConstPointer<Float> transformValues) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCoverStrokePathInstancedNV(\{numPaths}, \{pathNameType}, \{paths}, \{pathBase}, \{coverMode}, \{transformType}, \{transformValues})");
		try {
		    functions.glCoverStrokePathInstancedNV().invokeExact(numPaths, pathNameType, paths, pathBase, coverMode, transformType, transformValues);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pathNameType=\{pathNameType}, coverMode=\{coverMode}, transformType=\{transformType}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCoverStrokePathNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCoverStrokePathNV.xhtml">registry.khronos.org</a>
	 */
	public void glCoverStrokePathNV(int path, GLPathCoverMode coverMode) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCoverStrokePathNV(\{path}, \{coverMode})");
		try {
		    functions.glCoverStrokePathNV().invokeExact(path, coverMode);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were coverMode=\{coverMode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCoverageMaskNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCoverageMaskNV.xhtml">registry.khronos.org</a>
	 */
	public void glCoverageMaskNV(boolean mask) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCoverageMaskNV(\{mask})");
		try {
		    functions.glCoverageMaskNV().invokeExact(mask);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCoverageModulationNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCoverageModulationNV.xhtml">registry.khronos.org</a>
	 */
	public void glCoverageModulationNV(GLEnum components) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCoverageModulationNV(\{components})");
		try {
		    functions.glCoverageModulationNV().invokeExact(components);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were components=\{components}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCoverageModulationTableNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCoverageModulationTableNV.xhtml">registry.khronos.org</a>
	 */
	public void glCoverageModulationTableNV(int n, ConstPointer<Float> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCoverageModulationTableNV(\{n}, \{v})");
		try {
		    functions.glCoverageModulationTableNV().invokeExact(n, v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCoverageOperationNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCoverageOperationNV.xhtml">registry.khronos.org</a>
	 */
	public void glCoverageOperationNV(GLEnum operation) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCoverageOperationNV(\{operation})");
		try {
		    functions.glCoverageOperationNV().invokeExact(operation);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were operation=\{operation}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCreateBuffers">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCreateBuffers.xhtml">registry.khronos.org</a>
	 */
	public void glCreateBuffers(int n, MutPointer<Integer> buffers) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCreateBuffers(\{n}, \{buffers})");
		try {
		    functions.glCreateBuffers().invokeExact(n, buffers);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCreateCommandListsNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCreateCommandListsNV.xhtml">registry.khronos.org</a>
	 */
	public void glCreateCommandListsNV(int n, MutPointer<Integer> lists) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCreateCommandListsNV(\{n}, \{lists})");
		try {
		    functions.glCreateCommandListsNV().invokeExact(n, lists);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCreateFramebuffers">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCreateFramebuffers.xhtml">registry.khronos.org</a>
	 */
	public void glCreateFramebuffers(int n, MutPointer<Integer> framebuffers) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCreateFramebuffers(\{n}, \{framebuffers})");
		try {
		    functions.glCreateFramebuffers().invokeExact(n, framebuffers);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCreateMemoryObjectsEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCreateMemoryObjectsEXT.xhtml">registry.khronos.org</a>
	 */
	public void glCreateMemoryObjectsEXT(int n, MutPointer<Integer> memoryObjects) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCreateMemoryObjectsEXT(\{n}, \{memoryObjects})");
		try {
		    functions.glCreateMemoryObjectsEXT().invokeExact(n, memoryObjects);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCreatePerfQueryINTEL">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCreatePerfQueryINTEL.xhtml">registry.khronos.org</a>
	 */
	public void glCreatePerfQueryINTEL(int queryId, MutPointer<Integer> queryHandle) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCreatePerfQueryINTEL(\{queryId}, \{queryHandle})");
		try {
		    functions.glCreatePerfQueryINTEL().invokeExact(queryId, queryHandle);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCreateProgram">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCreateProgram.xhtml">registry.khronos.org</a>
	 */
	public int glCreateProgram() {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCreateProgram()");
		try {
		    var r = (int)functions.glCreateProgram().invokeExact();
			checkError();
			return r;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCreateProgramObjectARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCreateProgramObjectARB.xhtml">registry.khronos.org</a>
	 */
	public int glCreateProgramObjectARB() {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCreateProgramObjectARB()");
		try {
		    var r = (int)functions.glCreateProgramObjectARB().invokeExact();
			checkError();
			return r;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCreateProgramPipelines">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCreateProgramPipelines.xhtml">registry.khronos.org</a>
	 */
	public void glCreateProgramPipelines(int n, MutPointer<Integer> pipelines) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCreateProgramPipelines(\{n}, \{pipelines})");
		try {
		    functions.glCreateProgramPipelines().invokeExact(n, pipelines);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCreateProgressFenceNVX">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCreateProgressFenceNVX.xhtml">registry.khronos.org</a>
	 */
	public int glCreateProgressFenceNVX() {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCreateProgressFenceNVX()");
		try {
		    var r = (int)functions.glCreateProgressFenceNVX().invokeExact();
			checkError();
			return r;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCreateQueries">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCreateQueries.xhtml">registry.khronos.org</a>
	 */
	public void glCreateQueries(GLQueryTarget target, int n, MutPointer<Integer> ids) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCreateQueries(\{target}, \{n}, \{ids})");
		try {
		    functions.glCreateQueries().invokeExact(target, n, ids);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCreateRenderbuffers">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCreateRenderbuffers.xhtml">registry.khronos.org</a>
	 */
	public void glCreateRenderbuffers(int n, MutPointer<Integer> renderbuffers) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCreateRenderbuffers(\{n}, \{renderbuffers})");
		try {
		    functions.glCreateRenderbuffers().invokeExact(n, renderbuffers);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCreateSamplers">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCreateSamplers.xhtml">registry.khronos.org</a>
	 */
	public void glCreateSamplers(int n, MutPointer<Integer> samplers) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCreateSamplers(\{n}, \{samplers})");
		try {
		    functions.glCreateSamplers().invokeExact(n, samplers);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCreateSemaphoresNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCreateSemaphoresNV.xhtml">registry.khronos.org</a>
	 */
	public void glCreateSemaphoresNV(int n, MutPointer<Integer> semaphores) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCreateSemaphoresNV(\{n}, \{semaphores})");
		try {
		    functions.glCreateSemaphoresNV().invokeExact(n, semaphores);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCreateShader">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCreateShader.xhtml">registry.khronos.org</a>
	 */
	public int glCreateShader(GLShaderType type) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCreateShader(\{type})");
		try {
		    var r = (int)functions.glCreateShader().invokeExact(type);
			checkError();
			return r;
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCreateShaderObjectARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCreateShaderObjectARB.xhtml">registry.khronos.org</a>
	 */
	public int glCreateShaderObjectARB(GLShaderType shaderType) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCreateShaderObjectARB(\{shaderType})");
		try {
		    var r = (int)functions.glCreateShaderObjectARB().invokeExact(shaderType);
			checkError();
			return r;
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were shaderType=\{shaderType}.");
			}
			throw e;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCreateShaderProgramEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCreateShaderProgramEXT.xhtml">registry.khronos.org</a>
	 */
	public int glCreateShaderProgramEXT(GLShaderType type, ConstPointer<Byte> string) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCreateShaderProgramEXT(\{type}, \{string})");
		try {
		    var r = (int)functions.glCreateShaderProgramEXT().invokeExact(type, string);
			checkError();
			return r;
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCreateShaderProgramv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCreateShaderProgramv.xhtml">registry.khronos.org</a>
	 */
	public int glCreateShaderProgramv(GLShaderType type, int count, ConstPointer<ConstPointer<Byte>> strings) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCreateShaderProgramv(\{type}, \{count}, \{strings})");
		try {
		    var r = (int)functions.glCreateShaderProgramv().invokeExact(type, count, strings);
			checkError();
			return r;
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCreateShaderProgramvEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCreateShaderProgramvEXT.xhtml">registry.khronos.org</a>
	 */
	public int glCreateShaderProgramvEXT(GLShaderType type, int count, ConstPointer<ConstPointer<Byte>> strings) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCreateShaderProgramvEXT(\{type}, \{count}, \{strings})");
		try {
		    var r = (int)functions.glCreateShaderProgramvEXT().invokeExact(type, count, strings);
			checkError();
			return r;
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCreateStatesNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCreateStatesNV.xhtml">registry.khronos.org</a>
	 */
	public void glCreateStatesNV(int n, MutPointer<Integer> states) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCreateStatesNV(\{n}, \{states})");
		try {
		    functions.glCreateStatesNV().invokeExact(n, states);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCreateSyncFromCLeventARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCreateSyncFromCLeventARB.xhtml">registry.khronos.org</a>
	 */
	public ConstPointer<Void> glCreateSyncFromCLeventARB(MutPointer<Void> context, MutPointer<Void> event, int flags) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCreateSyncFromCLeventARB(\{context}, \{event}, \{flags})");
		try {
		    var r = (ConstPointer<Void>)functions.glCreateSyncFromCLeventARB().invokeExact(context, event, flags);
			checkError();
			return r;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCreateTextures">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCreateTextures.xhtml">registry.khronos.org</a>
	 */
	public void glCreateTextures(GLTextureTarget target, int n, MutPointer<Integer> textures) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCreateTextures(\{target}, \{n}, \{textures})");
		try {
		    functions.glCreateTextures().invokeExact(target, n, textures);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCreateTransformFeedbacks">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCreateTransformFeedbacks.xhtml">registry.khronos.org</a>
	 */
	public void glCreateTransformFeedbacks(int n, MutPointer<Integer> ids) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCreateTransformFeedbacks(\{n}, \{ids})");
		try {
		    functions.glCreateTransformFeedbacks().invokeExact(n, ids);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCreateVertexArrays">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCreateVertexArrays.xhtml">registry.khronos.org</a>
	 */
	public void glCreateVertexArrays(int n, MutPointer<Integer> arrays) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCreateVertexArrays(\{n}, \{arrays})");
		try {
		    functions.glCreateVertexArrays().invokeExact(n, arrays);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCullFace">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCullFace.xhtml">registry.khronos.org</a>
	 */
	public void glCullFace(GLTriangleFace mode) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCullFace(\{mode})");
		try {
		    functions.glCullFace().invokeExact(mode);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCullParameterdvEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCullParameterdvEXT.xhtml">registry.khronos.org</a>
	 */
	public void glCullParameterdvEXT(GLCullParameterEXT pname, MutPointer<Double> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCullParameterdvEXT(\{pname}, \{params})");
		try {
		    functions.glCullParameterdvEXT().invokeExact(pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCullParameterfvEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCullParameterfvEXT.xhtml">registry.khronos.org</a>
	 */
	public void glCullParameterfvEXT(GLCullParameterEXT pname, MutPointer<Float> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCullParameterfvEXT(\{pname}, \{params})");
		try {
		    functions.glCullParameterfvEXT().invokeExact(pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCurrentPaletteMatrixARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCurrentPaletteMatrixARB.xhtml">registry.khronos.org</a>
	 */
	public void glCurrentPaletteMatrixARB(int index) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCurrentPaletteMatrixARB(\{index})");
		try {
		    functions.glCurrentPaletteMatrixARB().invokeExact(index);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glCurrentPaletteMatrixOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glCurrentPaletteMatrixOES.xhtml">registry.khronos.org</a>
	 */
	public void glCurrentPaletteMatrixOES(int matrixpaletteindex) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glCurrentPaletteMatrixOES(\{matrixpaletteindex})");
		try {
		    functions.glCurrentPaletteMatrixOES().invokeExact(matrixpaletteindex);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDebugMessageCallback">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDebugMessageCallback.xhtml">registry.khronos.org</a>
	 */
	public void glDebugMessageCallback(ConstPointer<Void> callback, ConstPointer<Void> userParam) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDebugMessageCallback(\{callback}, \{userParam})");
		try {
		    functions.glDebugMessageCallback().invokeExact(callback, userParam);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDebugMessageCallbackAMD">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDebugMessageCallbackAMD.xhtml">registry.khronos.org</a>
	 */
	public void glDebugMessageCallbackAMD(ConstPointer<Void> callback, MutPointer<Void> userParam) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDebugMessageCallbackAMD(\{callback}, \{userParam})");
		try {
		    functions.glDebugMessageCallbackAMD().invokeExact(callback, userParam);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDebugMessageCallbackARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDebugMessageCallbackARB.xhtml">registry.khronos.org</a>
	 */
	public void glDebugMessageCallbackARB(ConstPointer<Void> callback, ConstPointer<Void> userParam) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDebugMessageCallbackARB(\{callback}, \{userParam})");
		try {
		    functions.glDebugMessageCallbackARB().invokeExact(callback, userParam);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDebugMessageCallbackKHR">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDebugMessageCallbackKHR.xhtml">registry.khronos.org</a>
	 */
	public void glDebugMessageCallbackKHR(ConstPointer<Void> callback, ConstPointer<Void> userParam) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDebugMessageCallbackKHR(\{callback}, \{userParam})");
		try {
		    functions.glDebugMessageCallbackKHR().invokeExact(callback, userParam);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDebugMessageControl">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDebugMessageControl.xhtml">registry.khronos.org</a>
	 */
	public void glDebugMessageControl(GLDebugSource source, GLDebugType type, GLDebugSeverity severity, int count, ConstPointer<Integer> ids, boolean enabled) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDebugMessageControl(\{source}, \{type}, \{severity}, \{count}, \{ids}, \{enabled})");
		try {
		    functions.glDebugMessageControl().invokeExact(source, type, severity, count, ids, enabled);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were source=\{source}, type=\{type}, severity=\{severity}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDebugMessageControlARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDebugMessageControlARB.xhtml">registry.khronos.org</a>
	 */
	public void glDebugMessageControlARB(GLDebugSource source, GLDebugType type, GLDebugSeverity severity, int count, ConstPointer<Integer> ids, boolean enabled) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDebugMessageControlARB(\{source}, \{type}, \{severity}, \{count}, \{ids}, \{enabled})");
		try {
		    functions.glDebugMessageControlARB().invokeExact(source, type, severity, count, ids, enabled);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were source=\{source}, type=\{type}, severity=\{severity}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDebugMessageControlKHR">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDebugMessageControlKHR.xhtml">registry.khronos.org</a>
	 */
	public void glDebugMessageControlKHR(GLDebugSource source, GLDebugType type, GLDebugSeverity severity, int count, ConstPointer<Integer> ids, boolean enabled) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDebugMessageControlKHR(\{source}, \{type}, \{severity}, \{count}, \{ids}, \{enabled})");
		try {
		    functions.glDebugMessageControlKHR().invokeExact(source, type, severity, count, ids, enabled);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were source=\{source}, type=\{type}, severity=\{severity}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDebugMessageEnableAMD">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDebugMessageEnableAMD.xhtml">registry.khronos.org</a>
	 */
	public void glDebugMessageEnableAMD(GLEnum category, GLDebugSeverity severity, int count, ConstPointer<Integer> ids, boolean enabled) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDebugMessageEnableAMD(\{category}, \{severity}, \{count}, \{ids}, \{enabled})");
		try {
		    functions.glDebugMessageEnableAMD().invokeExact(category, severity, count, ids, enabled);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were category=\{category}, severity=\{severity}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDebugMessageInsert">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDebugMessageInsert.xhtml">registry.khronos.org</a>
	 */
	public void glDebugMessageInsert(GLDebugSource source, GLDebugType type, int id, GLDebugSeverity severity, int length, ConstPointer<Byte> buf) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDebugMessageInsert(\{source}, \{type}, \{id}, \{severity}, \{length}, \{buf})");
		try {
		    functions.glDebugMessageInsert().invokeExact(source, type, id, severity, length, buf);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were source=\{source}, type=\{type}, severity=\{severity}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDebugMessageInsertAMD">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDebugMessageInsertAMD.xhtml">registry.khronos.org</a>
	 */
	public void glDebugMessageInsertAMD(GLEnum category, GLDebugSeverity severity, int id, int length, ConstPointer<Byte> buf) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDebugMessageInsertAMD(\{category}, \{severity}, \{id}, \{length}, \{buf})");
		try {
		    functions.glDebugMessageInsertAMD().invokeExact(category, severity, id, length, buf);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were category=\{category}, severity=\{severity}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDebugMessageInsertARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDebugMessageInsertARB.xhtml">registry.khronos.org</a>
	 */
	public void glDebugMessageInsertARB(GLDebugSource source, GLDebugType type, int id, GLDebugSeverity severity, int length, ConstPointer<Byte> buf) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDebugMessageInsertARB(\{source}, \{type}, \{id}, \{severity}, \{length}, \{buf})");
		try {
		    functions.glDebugMessageInsertARB().invokeExact(source, type, id, severity, length, buf);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were source=\{source}, type=\{type}, severity=\{severity}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDebugMessageInsertKHR">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDebugMessageInsertKHR.xhtml">registry.khronos.org</a>
	 */
	public void glDebugMessageInsertKHR(GLDebugSource source, GLDebugType type, int id, GLDebugSeverity severity, int length, ConstPointer<Byte> buf) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDebugMessageInsertKHR(\{source}, \{type}, \{id}, \{severity}, \{length}, \{buf})");
		try {
		    functions.glDebugMessageInsertKHR().invokeExact(source, type, id, severity, length, buf);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were source=\{source}, type=\{type}, severity=\{severity}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDeformSGIX">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDeformSGIX.xhtml">registry.khronos.org</a>
	 */
	public void glDeformSGIX(int mask) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDeformSGIX(\{mask})");
		try {
		    functions.glDeformSGIX().invokeExact(mask);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDeformationMap3dSGIX">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDeformationMap3dSGIX.xhtml">registry.khronos.org</a>
	 */
	public void glDeformationMap3dSGIX(GLFfdTargetSGIX target, double u1, double u2, int ustride, int uorder, double v1, double v2, int vstride, int vorder, double w1, double w2, int wstride, int worder, ConstPointer<Double> points) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDeformationMap3dSGIX(\{target}, \{u1}, \{u2}, \{ustride}, \{uorder}, \{v1}, \{v2}, \{vstride}, \{vorder}, \{w1}, \{w2}, \{wstride}, \{worder}, \{points})");
		try {
		    functions.glDeformationMap3dSGIX().invokeExact(target, u1, u2, ustride, uorder, v1, v2, vstride, vorder, w1, w2, wstride, worder, points);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDeformationMap3fSGIX">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDeformationMap3fSGIX.xhtml">registry.khronos.org</a>
	 */
	public void glDeformationMap3fSGIX(GLFfdTargetSGIX target, float u1, float u2, int ustride, int uorder, float v1, float v2, int vstride, int vorder, float w1, float w2, int wstride, int worder, ConstPointer<Float> points) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDeformationMap3fSGIX(\{target}, \{u1}, \{u2}, \{ustride}, \{uorder}, \{v1}, \{v2}, \{vstride}, \{vorder}, \{w1}, \{w2}, \{wstride}, \{worder}, \{points})");
		try {
		    functions.glDeformationMap3fSGIX().invokeExact(target, u1, u2, ustride, uorder, v1, v2, vstride, vorder, w1, w2, wstride, worder, points);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDeleteAsyncMarkersSGIX">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDeleteAsyncMarkersSGIX.xhtml">registry.khronos.org</a>
	 */
	public void glDeleteAsyncMarkersSGIX(int marker, int range) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDeleteAsyncMarkersSGIX(\{marker}, \{range})");
		try {
		    functions.glDeleteAsyncMarkersSGIX().invokeExact(marker, range);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDeleteBuffers">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDeleteBuffers.xhtml">registry.khronos.org</a>
	 */
	public void glDeleteBuffers(int n, ConstPointer<Integer> buffers) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDeleteBuffers(\{n}, \{buffers})");
		try {
		    functions.glDeleteBuffers().invokeExact(n, buffers);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	public void glDeleteBuffer(int buffer) {
        try (var arena = Arena.ofConfined()) {
            var buf = arena.allocate(GLTypeCache.ML_I(), buffer);
            glDeleteBuffers(1, new ConstPointer<>(buf));
        }
    }

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDeleteBuffersARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDeleteBuffersARB.xhtml">registry.khronos.org</a>
	 */
	public void glDeleteBuffersARB(int n, ConstPointer<Integer> buffers) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDeleteBuffersARB(\{n}, \{buffers})");
		try {
		    functions.glDeleteBuffersARB().invokeExact(n, buffers);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDeleteCommandListsNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDeleteCommandListsNV.xhtml">registry.khronos.org</a>
	 */
	public void glDeleteCommandListsNV(int n, ConstPointer<Integer> lists) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDeleteCommandListsNV(\{n}, \{lists})");
		try {
		    functions.glDeleteCommandListsNV().invokeExact(n, lists);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDeleteFencesAPPLE">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDeleteFencesAPPLE.xhtml">registry.khronos.org</a>
	 */
	public void glDeleteFencesAPPLE(int n, ConstPointer<Integer> fences) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDeleteFencesAPPLE(\{n}, \{fences})");
		try {
		    functions.glDeleteFencesAPPLE().invokeExact(n, fences);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDeleteFencesNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDeleteFencesNV.xhtml">registry.khronos.org</a>
	 */
	public void glDeleteFencesNV(int n, ConstPointer<Integer> fences) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDeleteFencesNV(\{n}, \{fences})");
		try {
		    functions.glDeleteFencesNV().invokeExact(n, fences);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDeleteFragmentShaderATI">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDeleteFragmentShaderATI.xhtml">registry.khronos.org</a>
	 */
	public void glDeleteFragmentShaderATI(int id) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDeleteFragmentShaderATI(\{id})");
		try {
		    functions.glDeleteFragmentShaderATI().invokeExact(id);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDeleteFramebuffers">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDeleteFramebuffers.xhtml">registry.khronos.org</a>
	 */
	public void glDeleteFramebuffers(int n, ConstPointer<Integer> framebuffers) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDeleteFramebuffers(\{n}, \{framebuffers})");
		try {
		    functions.glDeleteFramebuffers().invokeExact(n, framebuffers);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDeleteFramebuffersEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDeleteFramebuffersEXT.xhtml">registry.khronos.org</a>
	 */
	public void glDeleteFramebuffersEXT(int n, ConstPointer<Integer> framebuffers) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDeleteFramebuffersEXT(\{n}, \{framebuffers})");
		try {
		    functions.glDeleteFramebuffersEXT().invokeExact(n, framebuffers);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDeleteFramebuffersOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDeleteFramebuffersOES.xhtml">registry.khronos.org</a>
	 */
	public void glDeleteFramebuffersOES(int n, ConstPointer<Integer> framebuffers) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDeleteFramebuffersOES(\{n}, \{framebuffers})");
		try {
		    functions.glDeleteFramebuffersOES().invokeExact(n, framebuffers);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDeleteLists">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDeleteLists.xhtml">registry.khronos.org</a>
	 */
	public void glDeleteLists(int list, int range) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDeleteLists(\{list}, \{range})");
		try {
		    functions.glDeleteLists().invokeExact(list, range);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDeleteMemoryObjectsEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDeleteMemoryObjectsEXT.xhtml">registry.khronos.org</a>
	 */
	public void glDeleteMemoryObjectsEXT(int n, ConstPointer<Integer> memoryObjects) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDeleteMemoryObjectsEXT(\{n}, \{memoryObjects})");
		try {
		    functions.glDeleteMemoryObjectsEXT().invokeExact(n, memoryObjects);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDeleteNamedStringARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDeleteNamedStringARB.xhtml">registry.khronos.org</a>
	 */
	public void glDeleteNamedStringARB(int namelen, ConstPointer<Byte> name) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDeleteNamedStringARB(\{namelen}, \{name})");
		try {
		    functions.glDeleteNamedStringARB().invokeExact(namelen, name);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDeleteNamesAMD">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDeleteNamesAMD.xhtml">registry.khronos.org</a>
	 */
	public void glDeleteNamesAMD(GLEnum identifier, int num, ConstPointer<Integer> names) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDeleteNamesAMD(\{identifier}, \{num}, \{names})");
		try {
		    functions.glDeleteNamesAMD().invokeExact(identifier, num, names);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were identifier=\{identifier}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDeleteObjectARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDeleteObjectARB.xhtml">registry.khronos.org</a>
	 */
	public void glDeleteObjectARB(int obj) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDeleteObjectARB(\{obj})");
		try {
		    functions.glDeleteObjectARB().invokeExact(obj);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDeleteOcclusionQueriesNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDeleteOcclusionQueriesNV.xhtml">registry.khronos.org</a>
	 */
	public void glDeleteOcclusionQueriesNV(int n, ConstPointer<Integer> ids) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDeleteOcclusionQueriesNV(\{n}, \{ids})");
		try {
		    functions.glDeleteOcclusionQueriesNV().invokeExact(n, ids);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDeletePathsNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDeletePathsNV.xhtml">registry.khronos.org</a>
	 */
	public void glDeletePathsNV(int path, int range) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDeletePathsNV(\{path}, \{range})");
		try {
		    functions.glDeletePathsNV().invokeExact(path, range);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDeletePerfMonitorsAMD">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDeletePerfMonitorsAMD.xhtml">registry.khronos.org</a>
	 */
	public void glDeletePerfMonitorsAMD(int n, MutPointer<Integer> monitors) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDeletePerfMonitorsAMD(\{n}, \{monitors})");
		try {
		    functions.glDeletePerfMonitorsAMD().invokeExact(n, monitors);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDeletePerfQueryINTEL">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDeletePerfQueryINTEL.xhtml">registry.khronos.org</a>
	 */
	public void glDeletePerfQueryINTEL(int queryHandle) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDeletePerfQueryINTEL(\{queryHandle})");
		try {
		    functions.glDeletePerfQueryINTEL().invokeExact(queryHandle);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDeleteProgram">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDeleteProgram.xhtml">registry.khronos.org</a>
	 */
	public void glDeleteProgram(int program) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDeleteProgram(\{program})");
		try {
		    functions.glDeleteProgram().invokeExact(program);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDeleteProgramPipelines">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDeleteProgramPipelines.xhtml">registry.khronos.org</a>
	 */
	public void glDeleteProgramPipelines(int n, ConstPointer<Integer> pipelines) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDeleteProgramPipelines(\{n}, \{pipelines})");
		try {
		    functions.glDeleteProgramPipelines().invokeExact(n, pipelines);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDeleteProgramPipelinesEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDeleteProgramPipelinesEXT.xhtml">registry.khronos.org</a>
	 */
	public void glDeleteProgramPipelinesEXT(int n, ConstPointer<Integer> pipelines) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDeleteProgramPipelinesEXT(\{n}, \{pipelines})");
		try {
		    functions.glDeleteProgramPipelinesEXT().invokeExact(n, pipelines);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDeleteProgramsARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDeleteProgramsARB.xhtml">registry.khronos.org</a>
	 */
	public void glDeleteProgramsARB(int n, ConstPointer<Integer> programs) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDeleteProgramsARB(\{n}, \{programs})");
		try {
		    functions.glDeleteProgramsARB().invokeExact(n, programs);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDeleteProgramsNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDeleteProgramsNV.xhtml">registry.khronos.org</a>
	 */
	public void glDeleteProgramsNV(int n, ConstPointer<Integer> programs) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDeleteProgramsNV(\{n}, \{programs})");
		try {
		    functions.glDeleteProgramsNV().invokeExact(n, programs);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDeleteQueries">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDeleteQueries.xhtml">registry.khronos.org</a>
	 */
	public void glDeleteQueries(int n, ConstPointer<Integer> ids) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDeleteQueries(\{n}, \{ids})");
		try {
		    functions.glDeleteQueries().invokeExact(n, ids);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDeleteQueriesARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDeleteQueriesARB.xhtml">registry.khronos.org</a>
	 */
	public void glDeleteQueriesARB(int n, ConstPointer<Integer> ids) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDeleteQueriesARB(\{n}, \{ids})");
		try {
		    functions.glDeleteQueriesARB().invokeExact(n, ids);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDeleteQueriesEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDeleteQueriesEXT.xhtml">registry.khronos.org</a>
	 */
	public void glDeleteQueriesEXT(int n, ConstPointer<Integer> ids) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDeleteQueriesEXT(\{n}, \{ids})");
		try {
		    functions.glDeleteQueriesEXT().invokeExact(n, ids);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDeleteQueryResourceTagNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDeleteQueryResourceTagNV.xhtml">registry.khronos.org</a>
	 */
	public void glDeleteQueryResourceTagNV(int n, ConstPointer<Integer> tagIds) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDeleteQueryResourceTagNV(\{n}, \{tagIds})");
		try {
		    functions.glDeleteQueryResourceTagNV().invokeExact(n, tagIds);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDeleteRenderbuffers">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDeleteRenderbuffers.xhtml">registry.khronos.org</a>
	 */
	public void glDeleteRenderbuffers(int n, ConstPointer<Integer> renderbuffers) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDeleteRenderbuffers(\{n}, \{renderbuffers})");
		try {
		    functions.glDeleteRenderbuffers().invokeExact(n, renderbuffers);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDeleteRenderbuffersEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDeleteRenderbuffersEXT.xhtml">registry.khronos.org</a>
	 */
	public void glDeleteRenderbuffersEXT(int n, ConstPointer<Integer> renderbuffers) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDeleteRenderbuffersEXT(\{n}, \{renderbuffers})");
		try {
		    functions.glDeleteRenderbuffersEXT().invokeExact(n, renderbuffers);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDeleteRenderbuffersOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDeleteRenderbuffersOES.xhtml">registry.khronos.org</a>
	 */
	public void glDeleteRenderbuffersOES(int n, ConstPointer<Integer> renderbuffers) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDeleteRenderbuffersOES(\{n}, \{renderbuffers})");
		try {
		    functions.glDeleteRenderbuffersOES().invokeExact(n, renderbuffers);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDeleteSamplers">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDeleteSamplers.xhtml">registry.khronos.org</a>
	 */
	public void glDeleteSamplers(int count, ConstPointer<Integer> samplers) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDeleteSamplers(\{count}, \{samplers})");
		try {
		    functions.glDeleteSamplers().invokeExact(count, samplers);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDeleteSemaphoresEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDeleteSemaphoresEXT.xhtml">registry.khronos.org</a>
	 */
	public void glDeleteSemaphoresEXT(int n, ConstPointer<Integer> semaphores) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDeleteSemaphoresEXT(\{n}, \{semaphores})");
		try {
		    functions.glDeleteSemaphoresEXT().invokeExact(n, semaphores);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDeleteShader">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDeleteShader.xhtml">registry.khronos.org</a>
	 */
	public void glDeleteShader(int shader) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDeleteShader(\{shader})");
		try {
		    functions.glDeleteShader().invokeExact(shader);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDeleteStatesNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDeleteStatesNV.xhtml">registry.khronos.org</a>
	 */
	public void glDeleteStatesNV(int n, ConstPointer<Integer> states) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDeleteStatesNV(\{n}, \{states})");
		try {
		    functions.glDeleteStatesNV().invokeExact(n, states);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDeleteSync">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDeleteSync.xhtml">registry.khronos.org</a>
	 */
	public void glDeleteSync(ConstPointer<Void> sync) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDeleteSync(\{sync})");
		try {
		    functions.glDeleteSync().invokeExact(sync);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDeleteSyncAPPLE">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDeleteSyncAPPLE.xhtml">registry.khronos.org</a>
	 */
	public void glDeleteSyncAPPLE(ConstPointer<Void> sync) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDeleteSyncAPPLE(\{sync})");
		try {
		    functions.glDeleteSyncAPPLE().invokeExact(sync);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDeleteTextures">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDeleteTextures.xhtml">registry.khronos.org</a>
	 */
	public void glDeleteTextures(int n, ConstPointer<Integer> textures) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDeleteTextures(\{n}, \{textures})");
		try {
		    functions.glDeleteTextures().invokeExact(n, textures);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	public void glDeleteTexture(int texture) {
        try (var arena = Arena.ofConfined()) {
            var buf = arena.allocate(GLTypeCache.ML_I(), texture);
            glDeleteTextures(1, new ConstPointer<>(buf));
        }
    }

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDeleteTexturesEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDeleteTexturesEXT.xhtml">registry.khronos.org</a>
	 */
	public void glDeleteTexturesEXT(int n, ConstPointer<Integer> textures) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDeleteTexturesEXT(\{n}, \{textures})");
		try {
		    functions.glDeleteTexturesEXT().invokeExact(n, textures);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDeleteTransformFeedbacks">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDeleteTransformFeedbacks.xhtml">registry.khronos.org</a>
	 */
	public void glDeleteTransformFeedbacks(int n, ConstPointer<Integer> ids) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDeleteTransformFeedbacks(\{n}, \{ids})");
		try {
		    functions.glDeleteTransformFeedbacks().invokeExact(n, ids);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDeleteTransformFeedbacksNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDeleteTransformFeedbacksNV.xhtml">registry.khronos.org</a>
	 */
	public void glDeleteTransformFeedbacksNV(int n, ConstPointer<Integer> ids) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDeleteTransformFeedbacksNV(\{n}, \{ids})");
		try {
		    functions.glDeleteTransformFeedbacksNV().invokeExact(n, ids);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDeleteVertexArrays">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDeleteVertexArrays.xhtml">registry.khronos.org</a>
	 */
	public void glDeleteVertexArrays(int n, ConstPointer<Integer> arrays) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDeleteVertexArrays(\{n}, \{arrays})");
		try {
		    functions.glDeleteVertexArrays().invokeExact(n, arrays);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	public void glDeleteVertexArray(int array) {
        try (var arena = Arena.ofConfined()) {
            var buf = arena.allocate(GLTypeCache.ML_I(), array);
            glDeleteVertexArrays(1, new ConstPointer<>(buf));
        }
    }

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDeleteVertexArraysAPPLE">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDeleteVertexArraysAPPLE.xhtml">registry.khronos.org</a>
	 */
	public void glDeleteVertexArraysAPPLE(int n, ConstPointer<Integer> arrays) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDeleteVertexArraysAPPLE(\{n}, \{arrays})");
		try {
		    functions.glDeleteVertexArraysAPPLE().invokeExact(n, arrays);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDeleteVertexArraysOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDeleteVertexArraysOES.xhtml">registry.khronos.org</a>
	 */
	public void glDeleteVertexArraysOES(int n, ConstPointer<Integer> arrays) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDeleteVertexArraysOES(\{n}, \{arrays})");
		try {
		    functions.glDeleteVertexArraysOES().invokeExact(n, arrays);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDeleteVertexShaderEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDeleteVertexShaderEXT.xhtml">registry.khronos.org</a>
	 */
	public void glDeleteVertexShaderEXT(int id) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDeleteVertexShaderEXT(\{id})");
		try {
		    functions.glDeleteVertexShaderEXT().invokeExact(id);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDepthBoundsEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDepthBoundsEXT.xhtml">registry.khronos.org</a>
	 */
	public void glDepthBoundsEXT(double zmin, double zmax) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDepthBoundsEXT(\{zmin}, \{zmax})");
		try {
		    functions.glDepthBoundsEXT().invokeExact(zmin, zmax);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDepthBoundsdNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDepthBoundsdNV.xhtml">registry.khronos.org</a>
	 */
	public void glDepthBoundsdNV(double zmin, double zmax) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDepthBoundsdNV(\{zmin}, \{zmax})");
		try {
		    functions.glDepthBoundsdNV().invokeExact(zmin, zmax);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDepthFunc">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDepthFunc.xhtml">registry.khronos.org</a>
	 */
	public void glDepthFunc(GLDepthFunction func) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDepthFunc(\{func})");
		try {
		    functions.glDepthFunc().invokeExact(func);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were func=\{func}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDepthMask">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDepthMask.xhtml">registry.khronos.org</a>
	 */
	public void glDepthMask(boolean flag) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDepthMask(\{flag})");
		try {
		    functions.glDepthMask().invokeExact(flag);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDepthRange">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDepthRange.xhtml">registry.khronos.org</a>
	 */
	public void glDepthRange(double n, double f) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDepthRange(\{n}, \{f})");
		try {
		    functions.glDepthRange().invokeExact(n, f);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDepthRangeArraydvNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDepthRangeArraydvNV.xhtml">registry.khronos.org</a>
	 */
	public void glDepthRangeArraydvNV(int first, int count, ConstPointer<Double> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDepthRangeArraydvNV(\{first}, \{count}, \{v})");
		try {
		    functions.glDepthRangeArraydvNV().invokeExact(first, count, v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDepthRangeArrayfvNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDepthRangeArrayfvNV.xhtml">registry.khronos.org</a>
	 */
	public void glDepthRangeArrayfvNV(int first, int count, ConstPointer<Float> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDepthRangeArrayfvNV(\{first}, \{count}, \{v})");
		try {
		    functions.glDepthRangeArrayfvNV().invokeExact(first, count, v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDepthRangeArrayfvOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDepthRangeArrayfvOES.xhtml">registry.khronos.org</a>
	 */
	public void glDepthRangeArrayfvOES(int first, int count, ConstPointer<Float> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDepthRangeArrayfvOES(\{first}, \{count}, \{v})");
		try {
		    functions.glDepthRangeArrayfvOES().invokeExact(first, count, v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDepthRangeArrayv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDepthRangeArrayv.xhtml">registry.khronos.org</a>
	 */
	public void glDepthRangeArrayv(int first, int count, ConstPointer<Double> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDepthRangeArrayv(\{first}, \{count}, \{v})");
		try {
		    functions.glDepthRangeArrayv().invokeExact(first, count, v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDepthRangeIndexed">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDepthRangeIndexed.xhtml">registry.khronos.org</a>
	 */
	public void glDepthRangeIndexed(int index, double n, double f) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDepthRangeIndexed(\{index}, \{n}, \{f})");
		try {
		    functions.glDepthRangeIndexed().invokeExact(index, n, f);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDepthRangeIndexeddNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDepthRangeIndexeddNV.xhtml">registry.khronos.org</a>
	 */
	public void glDepthRangeIndexeddNV(int index, double n, double f) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDepthRangeIndexeddNV(\{index}, \{n}, \{f})");
		try {
		    functions.glDepthRangeIndexeddNV().invokeExact(index, n, f);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDepthRangeIndexedfNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDepthRangeIndexedfNV.xhtml">registry.khronos.org</a>
	 */
	public void glDepthRangeIndexedfNV(int index, float n, float f) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDepthRangeIndexedfNV(\{index}, \{n}, \{f})");
		try {
		    functions.glDepthRangeIndexedfNV().invokeExact(index, n, f);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDepthRangeIndexedfOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDepthRangeIndexedfOES.xhtml">registry.khronos.org</a>
	 */
	public void glDepthRangeIndexedfOES(int index, float n, float f) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDepthRangeIndexedfOES(\{index}, \{n}, \{f})");
		try {
		    functions.glDepthRangeIndexedfOES().invokeExact(index, n, f);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDepthRangedNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDepthRangedNV.xhtml">registry.khronos.org</a>
	 */
	public void glDepthRangedNV(double zNear, double zFar) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDepthRangedNV(\{zNear}, \{zFar})");
		try {
		    functions.glDepthRangedNV().invokeExact(zNear, zFar);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDepthRangef">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDepthRangef.xhtml">registry.khronos.org</a>
	 */
	public void glDepthRangef(float n, float f) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDepthRangef(\{n}, \{f})");
		try {
		    functions.glDepthRangef().invokeExact(n, f);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDepthRangefOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDepthRangefOES.xhtml">registry.khronos.org</a>
	 */
	public void glDepthRangefOES(float n, float f) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDepthRangefOES(\{n}, \{f})");
		try {
		    functions.glDepthRangefOES().invokeExact(n, f);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDepthRangex">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDepthRangex.xhtml">registry.khronos.org</a>
	 */
	public void glDepthRangex(int n, int f) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDepthRangex(\{n}, \{f})");
		try {
		    functions.glDepthRangex().invokeExact(n, f);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDepthRangexOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDepthRangexOES.xhtml">registry.khronos.org</a>
	 */
	public void glDepthRangexOES(int n, int f) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDepthRangexOES(\{n}, \{f})");
		try {
		    functions.glDepthRangexOES().invokeExact(n, f);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDetachObjectARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDetachObjectARB.xhtml">registry.khronos.org</a>
	 */
	public void glDetachObjectARB(int containerObj, int attachedObj) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDetachObjectARB(\{containerObj}, \{attachedObj})");
		try {
		    functions.glDetachObjectARB().invokeExact(containerObj, attachedObj);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDetachShader">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDetachShader.xhtml">registry.khronos.org</a>
	 */
	public void glDetachShader(int program, int shader) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDetachShader(\{program}, \{shader})");
		try {
		    functions.glDetachShader().invokeExact(program, shader);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDetailTexFuncSGIS">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDetailTexFuncSGIS.xhtml">registry.khronos.org</a>
	 */
	public void glDetailTexFuncSGIS(GLTextureTarget target, int n, ConstPointer<Float> points) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDetailTexFuncSGIS(\{target}, \{n}, \{points})");
		try {
		    functions.glDetailTexFuncSGIS().invokeExact(target, n, points);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDisable">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDisable.xhtml">registry.khronos.org</a>
	 */
	public void glDisable(GLEnableCap cap) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDisable(\{cap})");
		try {
		    functions.glDisable().invokeExact(cap);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were cap=\{cap}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDisableClientState">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDisableClientState.xhtml">registry.khronos.org</a>
	 */
	public void glDisableClientState(GLEnableCap array) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDisableClientState(\{array})");
		try {
		    functions.glDisableClientState().invokeExact(array);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were array=\{array}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDisableClientStateIndexedEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDisableClientStateIndexedEXT.xhtml">registry.khronos.org</a>
	 */
	public void glDisableClientStateIndexedEXT(GLEnableCap array, int index) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDisableClientStateIndexedEXT(\{array}, \{index})");
		try {
		    functions.glDisableClientStateIndexedEXT().invokeExact(array, index);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were array=\{array}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDisableClientStateiEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDisableClientStateiEXT.xhtml">registry.khronos.org</a>
	 */
	public void glDisableClientStateiEXT(GLEnableCap array, int index) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDisableClientStateiEXT(\{array}, \{index})");
		try {
		    functions.glDisableClientStateiEXT().invokeExact(array, index);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were array=\{array}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDisableDriverControlQCOM">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDisableDriverControlQCOM.xhtml">registry.khronos.org</a>
	 */
	public void glDisableDriverControlQCOM(int driverControl) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDisableDriverControlQCOM(\{driverControl})");
		try {
		    functions.glDisableDriverControlQCOM().invokeExact(driverControl);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDisableIndexedEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDisableIndexedEXT.xhtml">registry.khronos.org</a>
	 */
	public void glDisableIndexedEXT(GLEnableCap target, int index) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDisableIndexedEXT(\{target}, \{index})");
		try {
		    functions.glDisableIndexedEXT().invokeExact(target, index);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDisableVariantClientStateEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDisableVariantClientStateEXT.xhtml">registry.khronos.org</a>
	 */
	public void glDisableVariantClientStateEXT(int id) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDisableVariantClientStateEXT(\{id})");
		try {
		    functions.glDisableVariantClientStateEXT().invokeExact(id);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDisableVertexArrayAttrib">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDisableVertexArrayAttrib.xhtml">registry.khronos.org</a>
	 */
	public void glDisableVertexArrayAttrib(int vaobj, int index) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDisableVertexArrayAttrib(\{vaobj}, \{index})");
		try {
		    functions.glDisableVertexArrayAttrib().invokeExact(vaobj, index);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDisableVertexArrayAttribEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDisableVertexArrayAttribEXT.xhtml">registry.khronos.org</a>
	 */
	public void glDisableVertexArrayAttribEXT(int vaobj, int index) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDisableVertexArrayAttribEXT(\{vaobj}, \{index})");
		try {
		    functions.glDisableVertexArrayAttribEXT().invokeExact(vaobj, index);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDisableVertexArrayEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDisableVertexArrayEXT.xhtml">registry.khronos.org</a>
	 */
	public void glDisableVertexArrayEXT(int vaobj, GLEnableCap array) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDisableVertexArrayEXT(\{vaobj}, \{array})");
		try {
		    functions.glDisableVertexArrayEXT().invokeExact(vaobj, array);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were array=\{array}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDisableVertexAttribAPPLE">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDisableVertexAttribAPPLE.xhtml">registry.khronos.org</a>
	 */
	public void glDisableVertexAttribAPPLE(int index, GLEnum pname) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDisableVertexAttribAPPLE(\{index}, \{pname})");
		try {
		    functions.glDisableVertexAttribAPPLE().invokeExact(index, pname);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDisableVertexAttribArray">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDisableVertexAttribArray.xhtml">registry.khronos.org</a>
	 */
	public void glDisableVertexAttribArray(int index) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDisableVertexAttribArray(\{index})");
		try {
		    functions.glDisableVertexAttribArray().invokeExact(index);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDisableVertexAttribArrayARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDisableVertexAttribArrayARB.xhtml">registry.khronos.org</a>
	 */
	public void glDisableVertexAttribArrayARB(int index) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDisableVertexAttribArrayARB(\{index})");
		try {
		    functions.glDisableVertexAttribArrayARB().invokeExact(index);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDisablei">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDisablei.xhtml">registry.khronos.org</a>
	 */
	public void glDisablei(GLEnableCap target, int index) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDisablei(\{target}, \{index})");
		try {
		    functions.glDisablei().invokeExact(target, index);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDisableiEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDisableiEXT.xhtml">registry.khronos.org</a>
	 */
	public void glDisableiEXT(GLEnableCap target, int index) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDisableiEXT(\{target}, \{index})");
		try {
		    functions.glDisableiEXT().invokeExact(target, index);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDisableiNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDisableiNV.xhtml">registry.khronos.org</a>
	 */
	public void glDisableiNV(GLEnableCap target, int index) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDisableiNV(\{target}, \{index})");
		try {
		    functions.glDisableiNV().invokeExact(target, index);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDisableiOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDisableiOES.xhtml">registry.khronos.org</a>
	 */
	public void glDisableiOES(GLEnableCap target, int index) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDisableiOES(\{target}, \{index})");
		try {
		    functions.glDisableiOES().invokeExact(target, index);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDiscardFramebufferEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDiscardFramebufferEXT.xhtml">registry.khronos.org</a>
	 */
	public void glDiscardFramebufferEXT(GLFramebufferTarget target, int numAttachments, ConstPointer<GLInvalidateFramebufferAttachment> attachments) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDiscardFramebufferEXT(\{target}, \{numAttachments}, \{attachments})");
		try {
		    functions.glDiscardFramebufferEXT().invokeExact(target, numAttachments, attachments);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDispatchCompute">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDispatchCompute.xhtml">registry.khronos.org</a>
	 */
	public void glDispatchCompute(int num_groups_x, int num_groups_y, int num_groups_z) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDispatchCompute(\{num_groups_x}, \{num_groups_y}, \{num_groups_z})");
		try {
		    functions.glDispatchCompute().invokeExact(num_groups_x, num_groups_y, num_groups_z);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDispatchComputeGroupSizeARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDispatchComputeGroupSizeARB.xhtml">registry.khronos.org</a>
	 */
	public void glDispatchComputeGroupSizeARB(int num_groups_x, int num_groups_y, int num_groups_z, int group_size_x, int group_size_y, int group_size_z) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDispatchComputeGroupSizeARB(\{num_groups_x}, \{num_groups_y}, \{num_groups_z}, \{group_size_x}, \{group_size_y}, \{group_size_z})");
		try {
		    functions.glDispatchComputeGroupSizeARB().invokeExact(num_groups_x, num_groups_y, num_groups_z, group_size_x, group_size_y, group_size_z);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDispatchComputeIndirect">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDispatchComputeIndirect.xhtml">registry.khronos.org</a>
	 */
	public void glDispatchComputeIndirect(long indirect) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDispatchComputeIndirect(\{indirect})");
		try {
		    functions.glDispatchComputeIndirect().invokeExact(indirect);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDrawArrays">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDrawArrays.xhtml">registry.khronos.org</a>
	 */
	public void glDrawArrays(GLPrimitiveType mode, int first, int count) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDrawArrays(\{mode}, \{first}, \{count})");
		try {
		    functions.glDrawArrays().invokeExact(mode, first, count);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDrawArraysEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDrawArraysEXT.xhtml">registry.khronos.org</a>
	 */
	public void glDrawArraysEXT(GLPrimitiveType mode, int first, int count) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDrawArraysEXT(\{mode}, \{first}, \{count})");
		try {
		    functions.glDrawArraysEXT().invokeExact(mode, first, count);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDrawArraysIndirect">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDrawArraysIndirect.xhtml">registry.khronos.org</a>
	 */
	public void glDrawArraysIndirect(GLPrimitiveType mode, ConstPointer<Void> indirect) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDrawArraysIndirect(\{mode}, \{indirect})");
		try {
		    functions.glDrawArraysIndirect().invokeExact(mode, indirect);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDrawArraysInstanced">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDrawArraysInstanced.xhtml">registry.khronos.org</a>
	 */
	public void glDrawArraysInstanced(GLPrimitiveType mode, int first, int count, int instancecount) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDrawArraysInstanced(\{mode}, \{first}, \{count}, \{instancecount})");
		try {
		    functions.glDrawArraysInstanced().invokeExact(mode, first, count, instancecount);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDrawArraysInstancedANGLE">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDrawArraysInstancedANGLE.xhtml">registry.khronos.org</a>
	 */
	public void glDrawArraysInstancedANGLE(GLPrimitiveType mode, int first, int count, int primcount) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDrawArraysInstancedANGLE(\{mode}, \{first}, \{count}, \{primcount})");
		try {
		    functions.glDrawArraysInstancedANGLE().invokeExact(mode, first, count, primcount);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDrawArraysInstancedARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDrawArraysInstancedARB.xhtml">registry.khronos.org</a>
	 */
	public void glDrawArraysInstancedARB(GLPrimitiveType mode, int first, int count, int primcount) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDrawArraysInstancedARB(\{mode}, \{first}, \{count}, \{primcount})");
		try {
		    functions.glDrawArraysInstancedARB().invokeExact(mode, first, count, primcount);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 * @param mode          the draw mode.
	 * @param first         the index of the first element in the enabled arrays.
	 * @param count         the number of elements per instance.
	 * @param instancecount the number of instances.
	 * @param baseinstance  the index of the first instance.
	 *
	 * @see <a href="https://docs.gl/gl4/glDrawArraysInstancedBaseInstance">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDrawArraysInstancedBaseInstance.xhtml">registry.khronos.org</a>
	 */
	public void glDrawArraysInstancedBaseInstance(GLPrimitiveType mode, int first, int count, int instancecount, int baseinstance) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDrawArraysInstancedBaseInstance(\{mode}, \{first}, \{count}, \{instancecount}, \{baseinstance})");
		try {
		    functions.glDrawArraysInstancedBaseInstance().invokeExact(mode, first, count, instancecount, baseinstance);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 * @param mode          the draw mode.
	 * @param first         the index of the first element in the enabled arrays.
	 * @param count         the number of elements per instance.
	 * @param instancecount the number of instances.
	 * @param baseinstance  the index of the first instance.
	 *
	 * @see <a href="https://docs.gl/gl4/glDrawArraysInstancedBaseInstanceEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDrawArraysInstancedBaseInstanceEXT.xhtml">registry.khronos.org</a>
	 */
	public void glDrawArraysInstancedBaseInstanceEXT(GLPrimitiveType mode, int first, int count, int instancecount, int baseinstance) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDrawArraysInstancedBaseInstanceEXT(\{mode}, \{first}, \{count}, \{instancecount}, \{baseinstance})");
		try {
		    functions.glDrawArraysInstancedBaseInstanceEXT().invokeExact(mode, first, count, instancecount, baseinstance);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 * primcount should be renamed to instanceCount for OpenGL ES
	 *
	 * @see <a href="https://docs.gl/gl4/glDrawArraysInstancedEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDrawArraysInstancedEXT.xhtml">registry.khronos.org</a>
	 */
	public void glDrawArraysInstancedEXT(GLPrimitiveType mode, int start, int count, int primcount) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDrawArraysInstancedEXT(\{mode}, \{start}, \{count}, \{primcount})");
		try {
		    functions.glDrawArraysInstancedEXT().invokeExact(mode, start, count, primcount);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDrawArraysInstancedNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDrawArraysInstancedNV.xhtml">registry.khronos.org</a>
	 */
	public void glDrawArraysInstancedNV(GLPrimitiveType mode, int first, int count, int primcount) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDrawArraysInstancedNV(\{mode}, \{first}, \{count}, \{primcount})");
		try {
		    functions.glDrawArraysInstancedNV().invokeExact(mode, first, count, primcount);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDrawBuffer">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDrawBuffer.xhtml">registry.khronos.org</a>
	 */
	public void glDrawBuffer(GLDrawBufferMode buf) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDrawBuffer(\{buf})");
		try {
		    functions.glDrawBuffer().invokeExact(buf);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were buf=\{buf}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDrawBuffers">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDrawBuffers.xhtml">registry.khronos.org</a>
	 */
	public void glDrawBuffers(int n, ConstPointer<GLDrawBufferMode> bufs) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDrawBuffers(\{n}, \{bufs})");
		try {
		    functions.glDrawBuffers().invokeExact(n, bufs);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDrawBuffersARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDrawBuffersARB.xhtml">registry.khronos.org</a>
	 */
	public void glDrawBuffersARB(int n, ConstPointer<GLDrawBufferMode> bufs) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDrawBuffersARB(\{n}, \{bufs})");
		try {
		    functions.glDrawBuffersARB().invokeExact(n, bufs);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDrawBuffersATI">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDrawBuffersATI.xhtml">registry.khronos.org</a>
	 */
	public void glDrawBuffersATI(int n, ConstPointer<GLDrawBufferMode> bufs) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDrawBuffersATI(\{n}, \{bufs})");
		try {
		    functions.glDrawBuffersATI().invokeExact(n, bufs);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDrawBuffersEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDrawBuffersEXT.xhtml">registry.khronos.org</a>
	 */
	public void glDrawBuffersEXT(int n, ConstPointer<GLEnum> bufs) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDrawBuffersEXT(\{n}, \{bufs})");
		try {
		    functions.glDrawBuffersEXT().invokeExact(n, bufs);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDrawBuffersIndexedEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDrawBuffersIndexedEXT.xhtml">registry.khronos.org</a>
	 */
	public void glDrawBuffersIndexedEXT(int n, ConstPointer<GLEnum> location, ConstPointer<Integer> indices) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDrawBuffersIndexedEXT(\{n}, \{location}, \{indices})");
		try {
		    functions.glDrawBuffersIndexedEXT().invokeExact(n, location, indices);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDrawBuffersNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDrawBuffersNV.xhtml">registry.khronos.org</a>
	 */
	public void glDrawBuffersNV(int n, ConstPointer<GLEnum> bufs) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDrawBuffersNV(\{n}, \{bufs})");
		try {
		    functions.glDrawBuffersNV().invokeExact(n, bufs);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDrawCommandsAddressNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDrawCommandsAddressNV.xhtml">registry.khronos.org</a>
	 */
	public void glDrawCommandsAddressNV(GLEnum primitiveMode, ConstPointer<Long> indirects, ConstPointer<Integer> sizes, int count) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDrawCommandsAddressNV(\{primitiveMode}, \{indirects}, \{sizes}, \{count})");
		try {
		    functions.glDrawCommandsAddressNV().invokeExact(primitiveMode, indirects, sizes, count);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were primitiveMode=\{primitiveMode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDrawCommandsNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDrawCommandsNV.xhtml">registry.khronos.org</a>
	 */
	public void glDrawCommandsNV(GLEnum primitiveMode, int buffer, ConstPointer<Long> indirects, ConstPointer<Integer> sizes, int count) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDrawCommandsNV(\{primitiveMode}, \{buffer}, \{indirects}, \{sizes}, \{count})");
		try {
		    functions.glDrawCommandsNV().invokeExact(primitiveMode, buffer, indirects, sizes, count);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were primitiveMode=\{primitiveMode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDrawCommandsStatesAddressNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDrawCommandsStatesAddressNV.xhtml">registry.khronos.org</a>
	 */
	public void glDrawCommandsStatesAddressNV(ConstPointer<Long> indirects, ConstPointer<Integer> sizes, ConstPointer<Integer> states, ConstPointer<Integer> fbos, int count) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDrawCommandsStatesAddressNV(\{indirects}, \{sizes}, \{states}, \{fbos}, \{count})");
		try {
		    functions.glDrawCommandsStatesAddressNV().invokeExact(indirects, sizes, states, fbos, count);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDrawCommandsStatesNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDrawCommandsStatesNV.xhtml">registry.khronos.org</a>
	 */
	public void glDrawCommandsStatesNV(int buffer, ConstPointer<Long> indirects, ConstPointer<Integer> sizes, ConstPointer<Integer> states, ConstPointer<Integer> fbos, int count) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDrawCommandsStatesNV(\{buffer}, \{indirects}, \{sizes}, \{states}, \{fbos}, \{count})");
		try {
		    functions.glDrawCommandsStatesNV().invokeExact(buffer, indirects, sizes, states, fbos, count);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDrawElementArrayAPPLE">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDrawElementArrayAPPLE.xhtml">registry.khronos.org</a>
	 */
	public void glDrawElementArrayAPPLE(GLPrimitiveType mode, int first, int count) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDrawElementArrayAPPLE(\{mode}, \{first}, \{count})");
		try {
		    functions.glDrawElementArrayAPPLE().invokeExact(mode, first, count);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDrawElementArrayATI">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDrawElementArrayATI.xhtml">registry.khronos.org</a>
	 */
	public void glDrawElementArrayATI(GLPrimitiveType mode, int count) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDrawElementArrayATI(\{mode}, \{count})");
		try {
		    functions.glDrawElementArrayATI().invokeExact(mode, count);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDrawElements">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDrawElements.xhtml">registry.khronos.org</a>
	 */
	public void glDrawElements(GLPrimitiveType mode, int count, GLDrawElementsType type, ConstPointer<Void> indices) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDrawElements(\{mode}, \{count}, \{type}, \{indices})");
		try {
		    functions.glDrawElements().invokeExact(mode, count, type, indices);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDrawElementsBaseVertex">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDrawElementsBaseVertex.xhtml">registry.khronos.org</a>
	 */
	public void glDrawElementsBaseVertex(GLPrimitiveType mode, int count, GLDrawElementsType type, ConstPointer<Void> indices, int basevertex) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDrawElementsBaseVertex(\{mode}, \{count}, \{type}, \{indices}, \{basevertex})");
		try {
		    functions.glDrawElementsBaseVertex().invokeExact(mode, count, type, indices, basevertex);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDrawElementsBaseVertexEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDrawElementsBaseVertexEXT.xhtml">registry.khronos.org</a>
	 */
	public void glDrawElementsBaseVertexEXT(GLPrimitiveType mode, int count, GLDrawElementsType type, ConstPointer<Void> indices, int basevertex) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDrawElementsBaseVertexEXT(\{mode}, \{count}, \{type}, \{indices}, \{basevertex})");
		try {
		    functions.glDrawElementsBaseVertexEXT().invokeExact(mode, count, type, indices, basevertex);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDrawElementsBaseVertexOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDrawElementsBaseVertexOES.xhtml">registry.khronos.org</a>
	 */
	public void glDrawElementsBaseVertexOES(GLPrimitiveType mode, int count, GLDrawElementsType type, ConstPointer<Void> indices, int basevertex) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDrawElementsBaseVertexOES(\{mode}, \{count}, \{type}, \{indices}, \{basevertex})");
		try {
		    functions.glDrawElementsBaseVertexOES().invokeExact(mode, count, type, indices, basevertex);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDrawElementsIndirect">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDrawElementsIndirect.xhtml">registry.khronos.org</a>
	 */
	public void glDrawElementsIndirect(GLPrimitiveType mode, GLDrawElementsType type, ConstPointer<Void> indirect) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDrawElementsIndirect(\{mode}, \{type}, \{indirect})");
		try {
		    functions.glDrawElementsIndirect().invokeExact(mode, type, indirect);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDrawElementsInstanced">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDrawElementsInstanced.xhtml">registry.khronos.org</a>
	 */
	public void glDrawElementsInstanced(GLPrimitiveType mode, int count, GLDrawElementsType type, ConstPointer<Void> indices, int instancecount) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDrawElementsInstanced(\{mode}, \{count}, \{type}, \{indices}, \{instancecount})");
		try {
		    functions.glDrawElementsInstanced().invokeExact(mode, count, type, indices, instancecount);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDrawElementsInstancedANGLE">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDrawElementsInstancedANGLE.xhtml">registry.khronos.org</a>
	 */
	public void glDrawElementsInstancedANGLE(GLPrimitiveType mode, int count, GLDrawElementsType type, ConstPointer<Void> indices, int primcount) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDrawElementsInstancedANGLE(\{mode}, \{count}, \{type}, \{indices}, \{primcount})");
		try {
		    functions.glDrawElementsInstancedANGLE().invokeExact(mode, count, type, indices, primcount);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDrawElementsInstancedARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDrawElementsInstancedARB.xhtml">registry.khronos.org</a>
	 */
	public void glDrawElementsInstancedARB(GLPrimitiveType mode, int count, GLDrawElementsType type, ConstPointer<Void> indices, int primcount) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDrawElementsInstancedARB(\{mode}, \{count}, \{type}, \{indices}, \{primcount})");
		try {
		    functions.glDrawElementsInstancedARB().invokeExact(mode, count, type, indices, primcount);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDrawElementsInstancedBaseInstance">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDrawElementsInstancedBaseInstance.xhtml">registry.khronos.org</a>
	 */
	public void glDrawElementsInstancedBaseInstance(GLPrimitiveType mode, int count, GLDrawElementsType type, ConstPointer<Void> indices, int instancecount, int baseinstance) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDrawElementsInstancedBaseInstance(\{mode}, \{count}, \{type}, \{indices}, \{instancecount}, \{baseinstance})");
		try {
		    functions.glDrawElementsInstancedBaseInstance().invokeExact(mode, count, type, indices, instancecount, baseinstance);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDrawElementsInstancedBaseInstanceEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDrawElementsInstancedBaseInstanceEXT.xhtml">registry.khronos.org</a>
	 */
	public void glDrawElementsInstancedBaseInstanceEXT(GLPrimitiveType mode, int count, GLDrawElementsType type, ConstPointer<Void> indices, int instancecount, int baseinstance) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDrawElementsInstancedBaseInstanceEXT(\{mode}, \{count}, \{type}, \{indices}, \{instancecount}, \{baseinstance})");
		try {
		    functions.glDrawElementsInstancedBaseInstanceEXT().invokeExact(mode, count, type, indices, instancecount, baseinstance);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDrawElementsInstancedBaseVertex">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDrawElementsInstancedBaseVertex.xhtml">registry.khronos.org</a>
	 */
	public void glDrawElementsInstancedBaseVertex(GLPrimitiveType mode, int count, GLDrawElementsType type, ConstPointer<Void> indices, int instancecount, int basevertex) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDrawElementsInstancedBaseVertex(\{mode}, \{count}, \{type}, \{indices}, \{instancecount}, \{basevertex})");
		try {
		    functions.glDrawElementsInstancedBaseVertex().invokeExact(mode, count, type, indices, instancecount, basevertex);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDrawElementsInstancedBaseVertexBaseInstance">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDrawElementsInstancedBaseVertexBaseInstance.xhtml">registry.khronos.org</a>
	 */
	public void glDrawElementsInstancedBaseVertexBaseInstance(GLPrimitiveType mode, int count, GLDrawElementsType type, ConstPointer<Void> indices, int instancecount, int basevertex, int baseinstance) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDrawElementsInstancedBaseVertexBaseInstance(\{mode}, \{count}, \{type}, \{indices}, \{instancecount}, \{basevertex}, \{baseinstance})");
		try {
		    functions.glDrawElementsInstancedBaseVertexBaseInstance().invokeExact(mode, count, type, indices, instancecount, basevertex, baseinstance);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDrawElementsInstancedBaseVertexBaseInstanceEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDrawElementsInstancedBaseVertexBaseInstanceEXT.xhtml">registry.khronos.org</a>
	 */
	public void glDrawElementsInstancedBaseVertexBaseInstanceEXT(GLPrimitiveType mode, int count, GLDrawElementsType type, ConstPointer<Void> indices, int instancecount, int basevertex, int baseinstance) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDrawElementsInstancedBaseVertexBaseInstanceEXT(\{mode}, \{count}, \{type}, \{indices}, \{instancecount}, \{basevertex}, \{baseinstance})");
		try {
		    functions.glDrawElementsInstancedBaseVertexBaseInstanceEXT().invokeExact(mode, count, type, indices, instancecount, basevertex, baseinstance);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDrawElementsInstancedBaseVertexEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDrawElementsInstancedBaseVertexEXT.xhtml">registry.khronos.org</a>
	 */
	public void glDrawElementsInstancedBaseVertexEXT(GLPrimitiveType mode, int count, GLDrawElementsType type, ConstPointer<Void> indices, int instancecount, int basevertex) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDrawElementsInstancedBaseVertexEXT(\{mode}, \{count}, \{type}, \{indices}, \{instancecount}, \{basevertex})");
		try {
		    functions.glDrawElementsInstancedBaseVertexEXT().invokeExact(mode, count, type, indices, instancecount, basevertex);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDrawElementsInstancedBaseVertexOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDrawElementsInstancedBaseVertexOES.xhtml">registry.khronos.org</a>
	 */
	public void glDrawElementsInstancedBaseVertexOES(GLPrimitiveType mode, int count, GLDrawElementsType type, ConstPointer<Void> indices, int instancecount, int basevertex) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDrawElementsInstancedBaseVertexOES(\{mode}, \{count}, \{type}, \{indices}, \{instancecount}, \{basevertex})");
		try {
		    functions.glDrawElementsInstancedBaseVertexOES().invokeExact(mode, count, type, indices, instancecount, basevertex);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 * primcount should be renamed to instanceCount for OpenGL ES
	 *
	 * @see <a href="https://docs.gl/gl4/glDrawElementsInstancedEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDrawElementsInstancedEXT.xhtml">registry.khronos.org</a>
	 */
	public void glDrawElementsInstancedEXT(GLPrimitiveType mode, int count, GLDrawElementsType type, ConstPointer<Void> indices, int primcount) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDrawElementsInstancedEXT(\{mode}, \{count}, \{type}, \{indices}, \{primcount})");
		try {
		    functions.glDrawElementsInstancedEXT().invokeExact(mode, count, type, indices, primcount);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDrawElementsInstancedNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDrawElementsInstancedNV.xhtml">registry.khronos.org</a>
	 */
	public void glDrawElementsInstancedNV(GLPrimitiveType mode, int count, GLDrawElementsType type, ConstPointer<Void> indices, int primcount) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDrawElementsInstancedNV(\{mode}, \{count}, \{type}, \{indices}, \{primcount})");
		try {
		    functions.glDrawElementsInstancedNV().invokeExact(mode, count, type, indices, primcount);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDrawMeshArraysSUN">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDrawMeshArraysSUN.xhtml">registry.khronos.org</a>
	 */
	public void glDrawMeshArraysSUN(GLPrimitiveType mode, int first, int count, int width) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDrawMeshArraysSUN(\{mode}, \{first}, \{count}, \{width})");
		try {
		    functions.glDrawMeshArraysSUN().invokeExact(mode, first, count, width);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDrawMeshTasksNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDrawMeshTasksNV.xhtml">registry.khronos.org</a>
	 */
	public void glDrawMeshTasksNV(int first, int count) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDrawMeshTasksNV(\{first}, \{count})");
		try {
		    functions.glDrawMeshTasksNV().invokeExact(first, count);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDrawMeshTasksIndirectNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDrawMeshTasksIndirectNV.xhtml">registry.khronos.org</a>
	 */
	public void glDrawMeshTasksIndirectNV(long indirect) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDrawMeshTasksIndirectNV(\{indirect})");
		try {
		    functions.glDrawMeshTasksIndirectNV().invokeExact(indirect);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDrawPixels">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDrawPixels.xhtml">registry.khronos.org</a>
	 */
	public void glDrawPixels(int width, int height, GLPixelFormat format, GLPixelType type, ConstPointer<Void> pixels) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDrawPixels(\{width}, \{height}, \{format}, \{type}, \{pixels})");
		try {
		    functions.glDrawPixels().invokeExact(width, height, format, type, pixels);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were format=\{format}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDrawRangeElementArrayAPPLE">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDrawRangeElementArrayAPPLE.xhtml">registry.khronos.org</a>
	 */
	public void glDrawRangeElementArrayAPPLE(GLPrimitiveType mode, int start, int end, int first, int count) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDrawRangeElementArrayAPPLE(\{mode}, \{start}, \{end}, \{first}, \{count})");
		try {
		    functions.glDrawRangeElementArrayAPPLE().invokeExact(mode, start, end, first, count);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDrawRangeElementArrayATI">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDrawRangeElementArrayATI.xhtml">registry.khronos.org</a>
	 */
	public void glDrawRangeElementArrayATI(GLPrimitiveType mode, int start, int end, int count) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDrawRangeElementArrayATI(\{mode}, \{start}, \{end}, \{count})");
		try {
		    functions.glDrawRangeElementArrayATI().invokeExact(mode, start, end, count);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDrawRangeElements">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDrawRangeElements.xhtml">registry.khronos.org</a>
	 */
	public void glDrawRangeElements(GLPrimitiveType mode, int start, int end, int count, GLDrawElementsType type, ConstPointer<Void> indices) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDrawRangeElements(\{mode}, \{start}, \{end}, \{count}, \{type}, \{indices})");
		try {
		    functions.glDrawRangeElements().invokeExact(mode, start, end, count, type, indices);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDrawRangeElementsBaseVertex">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDrawRangeElementsBaseVertex.xhtml">registry.khronos.org</a>
	 */
	public void glDrawRangeElementsBaseVertex(GLPrimitiveType mode, int start, int end, int count, GLDrawElementsType type, ConstPointer<Void> indices, int basevertex) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDrawRangeElementsBaseVertex(\{mode}, \{start}, \{end}, \{count}, \{type}, \{indices}, \{basevertex})");
		try {
		    functions.glDrawRangeElementsBaseVertex().invokeExact(mode, start, end, count, type, indices, basevertex);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDrawRangeElementsBaseVertexEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDrawRangeElementsBaseVertexEXT.xhtml">registry.khronos.org</a>
	 */
	public void glDrawRangeElementsBaseVertexEXT(GLPrimitiveType mode, int start, int end, int count, GLDrawElementsType type, ConstPointer<Void> indices, int basevertex) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDrawRangeElementsBaseVertexEXT(\{mode}, \{start}, \{end}, \{count}, \{type}, \{indices}, \{basevertex})");
		try {
		    functions.glDrawRangeElementsBaseVertexEXT().invokeExact(mode, start, end, count, type, indices, basevertex);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDrawRangeElementsBaseVertexOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDrawRangeElementsBaseVertexOES.xhtml">registry.khronos.org</a>
	 */
	public void glDrawRangeElementsBaseVertexOES(GLPrimitiveType mode, int start, int end, int count, GLDrawElementsType type, ConstPointer<Void> indices, int basevertex) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDrawRangeElementsBaseVertexOES(\{mode}, \{start}, \{end}, \{count}, \{type}, \{indices}, \{basevertex})");
		try {
		    functions.glDrawRangeElementsBaseVertexOES().invokeExact(mode, start, end, count, type, indices, basevertex);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDrawRangeElementsEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDrawRangeElementsEXT.xhtml">registry.khronos.org</a>
	 */
	public void glDrawRangeElementsEXT(GLPrimitiveType mode, int start, int end, int count, GLDrawElementsType type, ConstPointer<Void> indices) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDrawRangeElementsEXT(\{mode}, \{start}, \{end}, \{count}, \{type}, \{indices})");
		try {
		    functions.glDrawRangeElementsEXT().invokeExact(mode, start, end, count, type, indices);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDrawTexfOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDrawTexfOES.xhtml">registry.khronos.org</a>
	 */
	public void glDrawTexfOES(float x, float y, float z, float width, float height) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDrawTexfOES(\{x}, \{y}, \{z}, \{width}, \{height})");
		try {
		    functions.glDrawTexfOES().invokeExact(x, y, z, width, height);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDrawTexfvOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDrawTexfvOES.xhtml">registry.khronos.org</a>
	 */
	public void glDrawTexfvOES(ConstPointer<Float> coords) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDrawTexfvOES(\{coords})");
		try {
		    functions.glDrawTexfvOES().invokeExact(coords);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDrawTexiOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDrawTexiOES.xhtml">registry.khronos.org</a>
	 */
	public void glDrawTexiOES(int x, int y, int z, int width, int height) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDrawTexiOES(\{x}, \{y}, \{z}, \{width}, \{height})");
		try {
		    functions.glDrawTexiOES().invokeExact(x, y, z, width, height);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDrawTexivOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDrawTexivOES.xhtml">registry.khronos.org</a>
	 */
	public void glDrawTexivOES(ConstPointer<Integer> coords) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDrawTexivOES(\{coords})");
		try {
		    functions.glDrawTexivOES().invokeExact(coords);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDrawTexsOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDrawTexsOES.xhtml">registry.khronos.org</a>
	 */
	public void glDrawTexsOES(short x, short y, short z, short width, short height) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDrawTexsOES(\{x}, \{y}, \{z}, \{width}, \{height})");
		try {
		    functions.glDrawTexsOES().invokeExact(x, y, z, width, height);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDrawTexsvOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDrawTexsvOES.xhtml">registry.khronos.org</a>
	 */
	public void glDrawTexsvOES(ConstPointer<Short> coords) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDrawTexsvOES(\{coords})");
		try {
		    functions.glDrawTexsvOES().invokeExact(coords);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDrawTextureNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDrawTextureNV.xhtml">registry.khronos.org</a>
	 */
	public void glDrawTextureNV(int texture, int sampler, float x0, float y0, float x1, float y1, float z, float s0, float t0, float s1, float t1) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDrawTextureNV(\{texture}, \{sampler}, \{x0}, \{y0}, \{x1}, \{y1}, \{z}, \{s0}, \{t0}, \{s1}, \{t1})");
		try {
		    functions.glDrawTextureNV().invokeExact(texture, sampler, x0, y0, x1, y1, z, s0, t0, s1, t1);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDrawTexxOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDrawTexxOES.xhtml">registry.khronos.org</a>
	 */
	public void glDrawTexxOES(int x, int y, int z, int width, int height) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDrawTexxOES(\{x}, \{y}, \{z}, \{width}, \{height})");
		try {
		    functions.glDrawTexxOES().invokeExact(x, y, z, width, height);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDrawTexxvOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDrawTexxvOES.xhtml">registry.khronos.org</a>
	 */
	public void glDrawTexxvOES(ConstPointer<Integer> coords) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDrawTexxvOES(\{coords})");
		try {
		    functions.glDrawTexxvOES().invokeExact(coords);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDrawTransformFeedback">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDrawTransformFeedback.xhtml">registry.khronos.org</a>
	 */
	public void glDrawTransformFeedback(GLPrimitiveType mode, int id) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDrawTransformFeedback(\{mode}, \{id})");
		try {
		    functions.glDrawTransformFeedback().invokeExact(mode, id);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDrawTransformFeedbackEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDrawTransformFeedbackEXT.xhtml">registry.khronos.org</a>
	 */
	public void glDrawTransformFeedbackEXT(GLPrimitiveType mode, int id) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDrawTransformFeedbackEXT(\{mode}, \{id})");
		try {
		    functions.glDrawTransformFeedbackEXT().invokeExact(mode, id);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDrawTransformFeedbackInstanced">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDrawTransformFeedbackInstanced.xhtml">registry.khronos.org</a>
	 */
	public void glDrawTransformFeedbackInstanced(GLPrimitiveType mode, int id, int instancecount) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDrawTransformFeedbackInstanced(\{mode}, \{id}, \{instancecount})");
		try {
		    functions.glDrawTransformFeedbackInstanced().invokeExact(mode, id, instancecount);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDrawTransformFeedbackInstancedEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDrawTransformFeedbackInstancedEXT.xhtml">registry.khronos.org</a>
	 */
	public void glDrawTransformFeedbackInstancedEXT(GLPrimitiveType mode, int id, int instancecount) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDrawTransformFeedbackInstancedEXT(\{mode}, \{id}, \{instancecount})");
		try {
		    functions.glDrawTransformFeedbackInstancedEXT().invokeExact(mode, id, instancecount);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDrawTransformFeedbackNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDrawTransformFeedbackNV.xhtml">registry.khronos.org</a>
	 */
	public void glDrawTransformFeedbackNV(GLPrimitiveType mode, int id) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDrawTransformFeedbackNV(\{mode}, \{id})");
		try {
		    functions.glDrawTransformFeedbackNV().invokeExact(mode, id);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDrawTransformFeedbackStream">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDrawTransformFeedbackStream.xhtml">registry.khronos.org</a>
	 */
	public void glDrawTransformFeedbackStream(GLPrimitiveType mode, int id, int stream) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDrawTransformFeedbackStream(\{mode}, \{id}, \{stream})");
		try {
		    functions.glDrawTransformFeedbackStream().invokeExact(mode, id, stream);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDrawTransformFeedbackStreamInstanced">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDrawTransformFeedbackStreamInstanced.xhtml">registry.khronos.org</a>
	 */
	public void glDrawTransformFeedbackStreamInstanced(GLPrimitiveType mode, int id, int stream, int instancecount) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDrawTransformFeedbackStreamInstanced(\{mode}, \{id}, \{stream}, \{instancecount})");
		try {
		    functions.glDrawTransformFeedbackStreamInstanced().invokeExact(mode, id, stream, instancecount);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glEGLImageTargetRenderbufferStorageOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glEGLImageTargetRenderbufferStorageOES.xhtml">registry.khronos.org</a>
	 */
	public void glEGLImageTargetRenderbufferStorageOES(GLEnum target, int image) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glEGLImageTargetRenderbufferStorageOES(\{target}, \{image})");
		try {
		    functions.glEGLImageTargetRenderbufferStorageOES().invokeExact(target, image);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glEGLImageTargetTexStorageEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glEGLImageTargetTexStorageEXT.xhtml">registry.khronos.org</a>
	 */
	public void glEGLImageTargetTexStorageEXT(GLEnum target, int image, ConstPointer<Integer> attrib_list) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glEGLImageTargetTexStorageEXT(\{target}, \{image}, \{attrib_list})");
		try {
		    functions.glEGLImageTargetTexStorageEXT().invokeExact(target, image, attrib_list);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glEGLImageTargetTexture2DOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glEGLImageTargetTexture2DOES.xhtml">registry.khronos.org</a>
	 */
	public void glEGLImageTargetTexture2DOES(GLEnum target, int image) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glEGLImageTargetTexture2DOES(\{target}, \{image})");
		try {
		    functions.glEGLImageTargetTexture2DOES().invokeExact(target, image);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glEGLImageTargetTextureStorageEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glEGLImageTargetTextureStorageEXT.xhtml">registry.khronos.org</a>
	 */
	public void glEGLImageTargetTextureStorageEXT(int texture, int image, ConstPointer<Integer> attrib_list) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glEGLImageTargetTextureStorageEXT(\{texture}, \{image}, \{attrib_list})");
		try {
		    functions.glEGLImageTargetTextureStorageEXT().invokeExact(texture, image, attrib_list);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glEdgeFlag">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glEdgeFlag.xhtml">registry.khronos.org</a>
	 */
	public void glEdgeFlag(boolean flag) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glEdgeFlag(\{flag})");
		try {
		    functions.glEdgeFlag().invokeExact(flag);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glEdgeFlagFormatNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glEdgeFlagFormatNV.xhtml">registry.khronos.org</a>
	 */
	public void glEdgeFlagFormatNV(int stride) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glEdgeFlagFormatNV(\{stride})");
		try {
		    functions.glEdgeFlagFormatNV().invokeExact(stride);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glEdgeFlagPointer">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glEdgeFlagPointer.xhtml">registry.khronos.org</a>
	 */
	public void glEdgeFlagPointer(int stride, ConstPointer<Void> pointer) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glEdgeFlagPointer(\{stride}, \{pointer})");
		try {
		    functions.glEdgeFlagPointer().invokeExact(stride, pointer);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glEdgeFlagPointerEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glEdgeFlagPointerEXT.xhtml">registry.khronos.org</a>
	 */
	public void glEdgeFlagPointerEXT(int stride, int count, ConstPointer<Boolean> pointer) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glEdgeFlagPointerEXT(\{stride}, \{count}, \{pointer})");
		try {
		    functions.glEdgeFlagPointerEXT().invokeExact(stride, count, pointer);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glEdgeFlagPointerListIBM">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glEdgeFlagPointerListIBM.xhtml">registry.khronos.org</a>
	 */
	public void glEdgeFlagPointerListIBM(int stride, MutPointer<ConstPointer<Boolean>> pointer, int ptrstride) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glEdgeFlagPointerListIBM(\{stride}, \{pointer}, \{ptrstride})");
		try {
		    functions.glEdgeFlagPointerListIBM().invokeExact(stride, pointer, ptrstride);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glEdgeFlagv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glEdgeFlagv.xhtml">registry.khronos.org</a>
	 */
	public void glEdgeFlagv(ConstPointer<Boolean> flag) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glEdgeFlagv(\{flag})");
		try {
		    functions.glEdgeFlagv().invokeExact(flag);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glElementPointerAPPLE">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glElementPointerAPPLE.xhtml">registry.khronos.org</a>
	 */
	public void glElementPointerAPPLE(GLElementPointerTypeATI type, ConstPointer<Void> pointer) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glElementPointerAPPLE(\{type}, \{pointer})");
		try {
		    functions.glElementPointerAPPLE().invokeExact(type, pointer);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glElementPointerATI">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glElementPointerATI.xhtml">registry.khronos.org</a>
	 */
	public void glElementPointerATI(GLElementPointerTypeATI type, ConstPointer<Void> pointer) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glElementPointerATI(\{type}, \{pointer})");
		try {
		    functions.glElementPointerATI().invokeExact(type, pointer);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glEnable">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glEnable.xhtml">registry.khronos.org</a>
	 */
	public void glEnable(GLEnableCap cap) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glEnable(\{cap})");
		try {
		    functions.glEnable().invokeExact(cap);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were cap=\{cap}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glEnableClientState">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glEnableClientState.xhtml">registry.khronos.org</a>
	 */
	public void glEnableClientState(GLEnableCap array) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glEnableClientState(\{array})");
		try {
		    functions.glEnableClientState().invokeExact(array);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were array=\{array}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glEnableClientStateIndexedEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glEnableClientStateIndexedEXT.xhtml">registry.khronos.org</a>
	 */
	public void glEnableClientStateIndexedEXT(GLEnableCap array, int index) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glEnableClientStateIndexedEXT(\{array}, \{index})");
		try {
		    functions.glEnableClientStateIndexedEXT().invokeExact(array, index);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were array=\{array}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glEnableClientStateiEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glEnableClientStateiEXT.xhtml">registry.khronos.org</a>
	 */
	public void glEnableClientStateiEXT(GLEnableCap array, int index) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glEnableClientStateiEXT(\{array}, \{index})");
		try {
		    functions.glEnableClientStateiEXT().invokeExact(array, index);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were array=\{array}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glEnableDriverControlQCOM">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glEnableDriverControlQCOM.xhtml">registry.khronos.org</a>
	 */
	public void glEnableDriverControlQCOM(int driverControl) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glEnableDriverControlQCOM(\{driverControl})");
		try {
		    functions.glEnableDriverControlQCOM().invokeExact(driverControl);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glEnableIndexedEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glEnableIndexedEXT.xhtml">registry.khronos.org</a>
	 */
	public void glEnableIndexedEXT(GLEnableCap target, int index) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glEnableIndexedEXT(\{target}, \{index})");
		try {
		    functions.glEnableIndexedEXT().invokeExact(target, index);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glEnableVariantClientStateEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glEnableVariantClientStateEXT.xhtml">registry.khronos.org</a>
	 */
	public void glEnableVariantClientStateEXT(int id) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glEnableVariantClientStateEXT(\{id})");
		try {
		    functions.glEnableVariantClientStateEXT().invokeExact(id);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glEnableVertexArrayAttrib">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glEnableVertexArrayAttrib.xhtml">registry.khronos.org</a>
	 */
	public void glEnableVertexArrayAttrib(int vaobj, int index) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glEnableVertexArrayAttrib(\{vaobj}, \{index})");
		try {
		    functions.glEnableVertexArrayAttrib().invokeExact(vaobj, index);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glEnableVertexArrayAttribEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glEnableVertexArrayAttribEXT.xhtml">registry.khronos.org</a>
	 */
	public void glEnableVertexArrayAttribEXT(int vaobj, int index) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glEnableVertexArrayAttribEXT(\{vaobj}, \{index})");
		try {
		    functions.glEnableVertexArrayAttribEXT().invokeExact(vaobj, index);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glEnableVertexArrayEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glEnableVertexArrayEXT.xhtml">registry.khronos.org</a>
	 */
	public void glEnableVertexArrayEXT(int vaobj, GLEnableCap array) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glEnableVertexArrayEXT(\{vaobj}, \{array})");
		try {
		    functions.glEnableVertexArrayEXT().invokeExact(vaobj, array);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were array=\{array}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glEnableVertexAttribAPPLE">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glEnableVertexAttribAPPLE.xhtml">registry.khronos.org</a>
	 */
	public void glEnableVertexAttribAPPLE(int index, GLEnum pname) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glEnableVertexAttribAPPLE(\{index}, \{pname})");
		try {
		    functions.glEnableVertexAttribAPPLE().invokeExact(index, pname);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glEnableVertexAttribArray">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glEnableVertexAttribArray.xhtml">registry.khronos.org</a>
	 */
	public void glEnableVertexAttribArray(int index) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glEnableVertexAttribArray(\{index})");
		try {
		    functions.glEnableVertexAttribArray().invokeExact(index);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glEnableVertexAttribArrayARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glEnableVertexAttribArrayARB.xhtml">registry.khronos.org</a>
	 */
	public void glEnableVertexAttribArrayARB(int index) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glEnableVertexAttribArrayARB(\{index})");
		try {
		    functions.glEnableVertexAttribArrayARB().invokeExact(index);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glEnablei">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glEnablei.xhtml">registry.khronos.org</a>
	 */
	public void glEnablei(GLEnableCap target, int index) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glEnablei(\{target}, \{index})");
		try {
		    functions.glEnablei().invokeExact(target, index);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glEnableiEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glEnableiEXT.xhtml">registry.khronos.org</a>
	 */
	public void glEnableiEXT(GLEnableCap target, int index) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glEnableiEXT(\{target}, \{index})");
		try {
		    functions.glEnableiEXT().invokeExact(target, index);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glEnableiNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glEnableiNV.xhtml">registry.khronos.org</a>
	 */
	public void glEnableiNV(GLEnableCap target, int index) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glEnableiNV(\{target}, \{index})");
		try {
		    functions.glEnableiNV().invokeExact(target, index);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glEnableiOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glEnableiOES.xhtml">registry.khronos.org</a>
	 */
	public void glEnableiOES(GLEnableCap target, int index) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glEnableiOES(\{target}, \{index})");
		try {
		    functions.glEnableiOES().invokeExact(target, index);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glEnd">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glEnd.xhtml">registry.khronos.org</a>
	 */
	public void glEnd() {
	    if (LOG_COMMANDS) logCmd(() -> STR."glEnd()");
		try {
		    functions.glEnd().invokeExact();
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glEndConditionalRender">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glEndConditionalRender.xhtml">registry.khronos.org</a>
	 */
	public void glEndConditionalRender() {
	    if (LOG_COMMANDS) logCmd(() -> STR."glEndConditionalRender()");
		try {
		    functions.glEndConditionalRender().invokeExact();
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glEndConditionalRenderNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glEndConditionalRenderNV.xhtml">registry.khronos.org</a>
	 */
	public void glEndConditionalRenderNV() {
	    if (LOG_COMMANDS) logCmd(() -> STR."glEndConditionalRenderNV()");
		try {
		    functions.glEndConditionalRenderNV().invokeExact();
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glEndConditionalRenderNVX">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glEndConditionalRenderNVX.xhtml">registry.khronos.org</a>
	 */
	public void glEndConditionalRenderNVX() {
	    if (LOG_COMMANDS) logCmd(() -> STR."glEndConditionalRenderNVX()");
		try {
		    functions.glEndConditionalRenderNVX().invokeExact();
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glEndFragmentShaderATI">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glEndFragmentShaderATI.xhtml">registry.khronos.org</a>
	 */
	public void glEndFragmentShaderATI() {
	    if (LOG_COMMANDS) logCmd(() -> STR."glEndFragmentShaderATI()");
		try {
		    functions.glEndFragmentShaderATI().invokeExact();
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glEndList">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glEndList.xhtml">registry.khronos.org</a>
	 */
	public void glEndList() {
	    if (LOG_COMMANDS) logCmd(() -> STR."glEndList()");
		try {
		    functions.glEndList().invokeExact();
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glEndOcclusionQueryNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glEndOcclusionQueryNV.xhtml">registry.khronos.org</a>
	 */
	public void glEndOcclusionQueryNV() {
	    if (LOG_COMMANDS) logCmd(() -> STR."glEndOcclusionQueryNV()");
		try {
		    functions.glEndOcclusionQueryNV().invokeExact();
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glEndPerfMonitorAMD">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glEndPerfMonitorAMD.xhtml">registry.khronos.org</a>
	 */
	public void glEndPerfMonitorAMD(int monitor) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glEndPerfMonitorAMD(\{monitor})");
		try {
		    functions.glEndPerfMonitorAMD().invokeExact(monitor);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glEndPerfQueryINTEL">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glEndPerfQueryINTEL.xhtml">registry.khronos.org</a>
	 */
	public void glEndPerfQueryINTEL(int queryHandle) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glEndPerfQueryINTEL(\{queryHandle})");
		try {
		    functions.glEndPerfQueryINTEL().invokeExact(queryHandle);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glEndQuery">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glEndQuery.xhtml">registry.khronos.org</a>
	 */
	public void glEndQuery(GLQueryTarget target) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glEndQuery(\{target})");
		try {
		    functions.glEndQuery().invokeExact(target);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glEndQueryARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glEndQueryARB.xhtml">registry.khronos.org</a>
	 */
	public void glEndQueryARB(GLQueryTarget target) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glEndQueryARB(\{target})");
		try {
		    functions.glEndQueryARB().invokeExact(target);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glEndQueryEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glEndQueryEXT.xhtml">registry.khronos.org</a>
	 */
	public void glEndQueryEXT(GLQueryTarget target) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glEndQueryEXT(\{target})");
		try {
		    functions.glEndQueryEXT().invokeExact(target);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glEndQueryIndexed">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glEndQueryIndexed.xhtml">registry.khronos.org</a>
	 */
	public void glEndQueryIndexed(GLQueryTarget target, int index) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glEndQueryIndexed(\{target}, \{index})");
		try {
		    functions.glEndQueryIndexed().invokeExact(target, index);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glEndTilingQCOM">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glEndTilingQCOM.xhtml">registry.khronos.org</a>
	 */
	public void glEndTilingQCOM(int preserveMask) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glEndTilingQCOM(\{preserveMask})");
		try {
		    functions.glEndTilingQCOM().invokeExact(preserveMask);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glEndTransformFeedback">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glEndTransformFeedback.xhtml">registry.khronos.org</a>
	 */
	public void glEndTransformFeedback() {
	    if (LOG_COMMANDS) logCmd(() -> STR."glEndTransformFeedback()");
		try {
		    functions.glEndTransformFeedback().invokeExact();
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glEndTransformFeedbackEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glEndTransformFeedbackEXT.xhtml">registry.khronos.org</a>
	 */
	public void glEndTransformFeedbackEXT() {
	    if (LOG_COMMANDS) logCmd(() -> STR."glEndTransformFeedbackEXT()");
		try {
		    functions.glEndTransformFeedbackEXT().invokeExact();
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glEndTransformFeedbackNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glEndTransformFeedbackNV.xhtml">registry.khronos.org</a>
	 */
	public void glEndTransformFeedbackNV() {
	    if (LOG_COMMANDS) logCmd(() -> STR."glEndTransformFeedbackNV()");
		try {
		    functions.glEndTransformFeedbackNV().invokeExact();
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glEndVertexShaderEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glEndVertexShaderEXT.xhtml">registry.khronos.org</a>
	 */
	public void glEndVertexShaderEXT() {
	    if (LOG_COMMANDS) logCmd(() -> STR."glEndVertexShaderEXT()");
		try {
		    functions.glEndVertexShaderEXT().invokeExact();
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glEndVideoCaptureNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glEndVideoCaptureNV.xhtml">registry.khronos.org</a>
	 */
	public void glEndVideoCaptureNV(int video_capture_slot) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glEndVideoCaptureNV(\{video_capture_slot})");
		try {
		    functions.glEndVideoCaptureNV().invokeExact(video_capture_slot);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glEvalCoord1d">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glEvalCoord1d.xhtml">registry.khronos.org</a>
	 */
	public void glEvalCoord1d(double u) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glEvalCoord1d(\{u})");
		try {
		    functions.glEvalCoord1d().invokeExact(u);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glEvalCoord1dv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glEvalCoord1dv.xhtml">registry.khronos.org</a>
	 */
	public void glEvalCoord1dv(ConstPointer<Double> u) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glEvalCoord1dv(\{u})");
		try {
		    functions.glEvalCoord1dv().invokeExact(u);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glEvalCoord1f">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glEvalCoord1f.xhtml">registry.khronos.org</a>
	 */
	public void glEvalCoord1f(float u) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glEvalCoord1f(\{u})");
		try {
		    functions.glEvalCoord1f().invokeExact(u);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glEvalCoord1fv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glEvalCoord1fv.xhtml">registry.khronos.org</a>
	 */
	public void glEvalCoord1fv(ConstPointer<Float> u) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glEvalCoord1fv(\{u})");
		try {
		    functions.glEvalCoord1fv().invokeExact(u);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glEvalCoord1xOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glEvalCoord1xOES.xhtml">registry.khronos.org</a>
	 */
	public void glEvalCoord1xOES(int u) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glEvalCoord1xOES(\{u})");
		try {
		    functions.glEvalCoord1xOES().invokeExact(u);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glEvalCoord1xvOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glEvalCoord1xvOES.xhtml">registry.khronos.org</a>
	 */
	public void glEvalCoord1xvOES(ConstPointer<Integer> coords) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glEvalCoord1xvOES(\{coords})");
		try {
		    functions.glEvalCoord1xvOES().invokeExact(coords);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glEvalCoord2d">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glEvalCoord2d.xhtml">registry.khronos.org</a>
	 */
	public void glEvalCoord2d(double u, double v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glEvalCoord2d(\{u}, \{v})");
		try {
		    functions.glEvalCoord2d().invokeExact(u, v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glEvalCoord2dv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glEvalCoord2dv.xhtml">registry.khronos.org</a>
	 */
	public void glEvalCoord2dv(ConstPointer<Double> u) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glEvalCoord2dv(\{u})");
		try {
		    functions.glEvalCoord2dv().invokeExact(u);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glEvalCoord2f">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glEvalCoord2f.xhtml">registry.khronos.org</a>
	 */
	public void glEvalCoord2f(float u, float v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glEvalCoord2f(\{u}, \{v})");
		try {
		    functions.glEvalCoord2f().invokeExact(u, v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glEvalCoord2fv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glEvalCoord2fv.xhtml">registry.khronos.org</a>
	 */
	public void glEvalCoord2fv(ConstPointer<Float> u) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glEvalCoord2fv(\{u})");
		try {
		    functions.glEvalCoord2fv().invokeExact(u);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glEvalCoord2xOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glEvalCoord2xOES.xhtml">registry.khronos.org</a>
	 */
	public void glEvalCoord2xOES(int u, int v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glEvalCoord2xOES(\{u}, \{v})");
		try {
		    functions.glEvalCoord2xOES().invokeExact(u, v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glEvalCoord2xvOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glEvalCoord2xvOES.xhtml">registry.khronos.org</a>
	 */
	public void glEvalCoord2xvOES(ConstPointer<Integer> coords) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glEvalCoord2xvOES(\{coords})");
		try {
		    functions.glEvalCoord2xvOES().invokeExact(coords);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glEvalMapsNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glEvalMapsNV.xhtml">registry.khronos.org</a>
	 */
	public void glEvalMapsNV(GLEvalTargetNV target, GLEvalMapsModeNV mode) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glEvalMapsNV(\{target}, \{mode})");
		try {
		    functions.glEvalMapsNV().invokeExact(target, mode);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, mode=\{mode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glEvalMesh1">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glEvalMesh1.xhtml">registry.khronos.org</a>
	 */
	public void glEvalMesh1(GLMeshMode1 mode, int i1, int i2) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glEvalMesh1(\{mode}, \{i1}, \{i2})");
		try {
		    functions.glEvalMesh1().invokeExact(mode, i1, i2);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glEvalMesh2">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glEvalMesh2.xhtml">registry.khronos.org</a>
	 */
	public void glEvalMesh2(GLMeshMode2 mode, int i1, int i2, int j1, int j2) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glEvalMesh2(\{mode}, \{i1}, \{i2}, \{j1}, \{j2})");
		try {
		    functions.glEvalMesh2().invokeExact(mode, i1, i2, j1, j2);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glEvalPoint1">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glEvalPoint1.xhtml">registry.khronos.org</a>
	 */
	public void glEvalPoint1(int i) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glEvalPoint1(\{i})");
		try {
		    functions.glEvalPoint1().invokeExact(i);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glEvalPoint2">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glEvalPoint2.xhtml">registry.khronos.org</a>
	 */
	public void glEvalPoint2(int i, int j) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glEvalPoint2(\{i}, \{j})");
		try {
		    functions.glEvalPoint2().invokeExact(i, j);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glEvaluateDepthValuesARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glEvaluateDepthValuesARB.xhtml">registry.khronos.org</a>
	 */
	public void glEvaluateDepthValuesARB() {
	    if (LOG_COMMANDS) logCmd(() -> STR."glEvaluateDepthValuesARB()");
		try {
		    functions.glEvaluateDepthValuesARB().invokeExact();
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glExecuteProgramNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glExecuteProgramNV.xhtml">registry.khronos.org</a>
	 */
	public void glExecuteProgramNV(GLVertexAttribEnumNV target, int id, ConstPointer<Float> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glExecuteProgramNV(\{target}, \{id}, \{params})");
		try {
		    functions.glExecuteProgramNV().invokeExact(target, id, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glExtGetBufferPointervQCOM">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glExtGetBufferPointervQCOM.xhtml">registry.khronos.org</a>
	 */
	public void glExtGetBufferPointervQCOM(GLEnum target, MutPointer<MutPointer<Void>> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glExtGetBufferPointervQCOM(\{target}, \{params})");
		try {
		    functions.glExtGetBufferPointervQCOM().invokeExact(target, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glExtGetBuffersQCOM">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glExtGetBuffersQCOM.xhtml">registry.khronos.org</a>
	 */
	public void glExtGetBuffersQCOM(MutPointer<Integer> buffers, int maxBuffers, MutPointer<Integer> numBuffers) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glExtGetBuffersQCOM(\{buffers}, \{maxBuffers}, \{numBuffers})");
		try {
		    functions.glExtGetBuffersQCOM().invokeExact(buffers, maxBuffers, numBuffers);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glExtGetFramebuffersQCOM">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glExtGetFramebuffersQCOM.xhtml">registry.khronos.org</a>
	 */
	public void glExtGetFramebuffersQCOM(MutPointer<Integer> framebuffers, int maxFramebuffers, MutPointer<Integer> numFramebuffers) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glExtGetFramebuffersQCOM(\{framebuffers}, \{maxFramebuffers}, \{numFramebuffers})");
		try {
		    functions.glExtGetFramebuffersQCOM().invokeExact(framebuffers, maxFramebuffers, numFramebuffers);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glExtGetProgramBinarySourceQCOM">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glExtGetProgramBinarySourceQCOM.xhtml">registry.khronos.org</a>
	 */
	public void glExtGetProgramBinarySourceQCOM(int program, GLShaderType shadertype, MutPointer<Byte> source, MutPointer<Integer> length) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glExtGetProgramBinarySourceQCOM(\{program}, \{shadertype}, \{source}, \{length})");
		try {
		    functions.glExtGetProgramBinarySourceQCOM().invokeExact(program, shadertype, source, length);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were shadertype=\{shadertype}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glExtGetProgramsQCOM">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glExtGetProgramsQCOM.xhtml">registry.khronos.org</a>
	 */
	public void glExtGetProgramsQCOM(MutPointer<Integer> programs, int maxPrograms, MutPointer<Integer> numPrograms) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glExtGetProgramsQCOM(\{programs}, \{maxPrograms}, \{numPrograms})");
		try {
		    functions.glExtGetProgramsQCOM().invokeExact(programs, maxPrograms, numPrograms);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glExtGetRenderbuffersQCOM">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glExtGetRenderbuffersQCOM.xhtml">registry.khronos.org</a>
	 */
	public void glExtGetRenderbuffersQCOM(MutPointer<Integer> renderbuffers, int maxRenderbuffers, MutPointer<Integer> numRenderbuffers) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glExtGetRenderbuffersQCOM(\{renderbuffers}, \{maxRenderbuffers}, \{numRenderbuffers})");
		try {
		    functions.glExtGetRenderbuffersQCOM().invokeExact(renderbuffers, maxRenderbuffers, numRenderbuffers);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glExtGetShadersQCOM">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glExtGetShadersQCOM.xhtml">registry.khronos.org</a>
	 */
	public void glExtGetShadersQCOM(MutPointer<Integer> shaders, int maxShaders, MutPointer<Integer> numShaders) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glExtGetShadersQCOM(\{shaders}, \{maxShaders}, \{numShaders})");
		try {
		    functions.glExtGetShadersQCOM().invokeExact(shaders, maxShaders, numShaders);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glExtGetTexLevelParameterivQCOM">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glExtGetTexLevelParameterivQCOM.xhtml">registry.khronos.org</a>
	 */
	public void glExtGetTexLevelParameterivQCOM(int texture, GLEnum face, int level, GLEnum pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glExtGetTexLevelParameterivQCOM(\{texture}, \{face}, \{level}, \{pname}, \{params})");
		try {
		    functions.glExtGetTexLevelParameterivQCOM().invokeExact(texture, face, level, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were face=\{face}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glExtGetTexSubImageQCOM">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glExtGetTexSubImageQCOM.xhtml">registry.khronos.org</a>
	 */
	public void glExtGetTexSubImageQCOM(GLEnum target, int level, int xoffset, int yoffset, int zoffset, int width, int height, int depth, GLPixelFormat format, GLPixelType type, MutPointer<Void> texels) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glExtGetTexSubImageQCOM(\{target}, \{level}, \{xoffset}, \{yoffset}, \{zoffset}, \{width}, \{height}, \{depth}, \{format}, \{type}, \{texels})");
		try {
		    functions.glExtGetTexSubImageQCOM().invokeExact(target, level, xoffset, yoffset, zoffset, width, height, depth, format, type, texels);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, format=\{format}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glExtGetTexturesQCOM">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glExtGetTexturesQCOM.xhtml">registry.khronos.org</a>
	 */
	public void glExtGetTexturesQCOM(MutPointer<Integer> textures, int maxTextures, MutPointer<Integer> numTextures) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glExtGetTexturesQCOM(\{textures}, \{maxTextures}, \{numTextures})");
		try {
		    functions.glExtGetTexturesQCOM().invokeExact(textures, maxTextures, numTextures);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glExtIsProgramBinaryQCOM">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glExtIsProgramBinaryQCOM.xhtml">registry.khronos.org</a>
	 */
	public boolean glExtIsProgramBinaryQCOM(int program) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glExtIsProgramBinaryQCOM(\{program})");
		try {
		    var r = (boolean)functions.glExtIsProgramBinaryQCOM().invokeExact(program);
			checkError();
			return r;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glExtTexObjectStateOverrideiQCOM">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glExtTexObjectStateOverrideiQCOM.xhtml">registry.khronos.org</a>
	 */
	public void glExtTexObjectStateOverrideiQCOM(GLEnum target, GLEnum pname, int param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glExtTexObjectStateOverrideiQCOM(\{target}, \{pname}, \{param})");
		try {
		    functions.glExtTexObjectStateOverrideiQCOM().invokeExact(target, pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glExtractComponentEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glExtractComponentEXT.xhtml">registry.khronos.org</a>
	 */
	public void glExtractComponentEXT(int res, int src, int num) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glExtractComponentEXT(\{res}, \{src}, \{num})");
		try {
		    functions.glExtractComponentEXT().invokeExact(res, src, num);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glFeedbackBuffer">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glFeedbackBuffer.xhtml">registry.khronos.org</a>
	 */
	public void glFeedbackBuffer(int size, GLFeedbackType type, MutPointer<Float> buffer) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glFeedbackBuffer(\{size}, \{type}, \{buffer})");
		try {
		    functions.glFeedbackBuffer().invokeExact(size, type, buffer);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glFeedbackBufferxOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glFeedbackBufferxOES.xhtml">registry.khronos.org</a>
	 */
	public void glFeedbackBufferxOES(int n, GLEnum type, ConstPointer<Integer> buffer) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glFeedbackBufferxOES(\{n}, \{type}, \{buffer})");
		try {
		    functions.glFeedbackBufferxOES().invokeExact(n, type, buffer);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glFenceSync">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glFenceSync.xhtml">registry.khronos.org</a>
	 */
	public ConstPointer<Void> glFenceSync(GLSyncCondition condition, int flags) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glFenceSync(\{condition}, \{flags})");
		try {
		    var r = (ConstPointer<Void>)functions.glFenceSync().invokeExact(condition, flags);
			checkError();
			return r;
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were condition=\{condition}.");
			}
			throw e;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glFenceSyncAPPLE">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glFenceSyncAPPLE.xhtml">registry.khronos.org</a>
	 */
	public ConstPointer<Void> glFenceSyncAPPLE(GLSyncCondition condition, int flags) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glFenceSyncAPPLE(\{condition}, \{flags})");
		try {
		    var r = (ConstPointer<Void>)functions.glFenceSyncAPPLE().invokeExact(condition, flags);
			checkError();
			return r;
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were condition=\{condition}.");
			}
			throw e;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glFinalCombinerInputNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glFinalCombinerInputNV.xhtml">registry.khronos.org</a>
	 */
	public void glFinalCombinerInputNV(GLCombinerVariableNV variable, GLCombinerRegisterNV input, GLCombinerMappingNV mapping, GLCombinerComponentUsageNV componentUsage) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glFinalCombinerInputNV(\{variable}, \{input}, \{mapping}, \{componentUsage})");
		try {
		    functions.glFinalCombinerInputNV().invokeExact(variable, input, mapping, componentUsage);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were variable=\{variable}, input=\{input}, mapping=\{mapping}, componentUsage=\{componentUsage}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glFinish">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glFinish.xhtml">registry.khronos.org</a>
	 */
	public void glFinish() {
	    if (LOG_COMMANDS) logCmd(() -> STR."glFinish()");
		try {
		    functions.glFinish().invokeExact();
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glFinishAsyncSGIX">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glFinishAsyncSGIX.xhtml">registry.khronos.org</a>
	 */
	public int glFinishAsyncSGIX(MutPointer<Integer> markerp) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glFinishAsyncSGIX(\{markerp})");
		try {
		    var r = (int)functions.glFinishAsyncSGIX().invokeExact(markerp);
			checkError();
			return r;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glFinishFenceAPPLE">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glFinishFenceAPPLE.xhtml">registry.khronos.org</a>
	 */
	public void glFinishFenceAPPLE(int fence) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glFinishFenceAPPLE(\{fence})");
		try {
		    functions.glFinishFenceAPPLE().invokeExact(fence);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glFinishFenceNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glFinishFenceNV.xhtml">registry.khronos.org</a>
	 */
	public void glFinishFenceNV(int fence) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glFinishFenceNV(\{fence})");
		try {
		    functions.glFinishFenceNV().invokeExact(fence);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glFinishObjectAPPLE">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glFinishObjectAPPLE.xhtml">registry.khronos.org</a>
	 */
	public void glFinishObjectAPPLE(GLObjectTypeAPPLE object, int name) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glFinishObjectAPPLE(\{object}, \{name})");
		try {
		    functions.glFinishObjectAPPLE().invokeExact(object, name);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were object=\{object}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glFinishTextureSUNX">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glFinishTextureSUNX.xhtml">registry.khronos.org</a>
	 */
	public void glFinishTextureSUNX() {
	    if (LOG_COMMANDS) logCmd(() -> STR."glFinishTextureSUNX()");
		try {
		    functions.glFinishTextureSUNX().invokeExact();
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glFlush">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glFlush.xhtml">registry.khronos.org</a>
	 */
	public void glFlush() {
	    if (LOG_COMMANDS) logCmd(() -> STR."glFlush()");
		try {
		    functions.glFlush().invokeExact();
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glFlushMappedBufferRange">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glFlushMappedBufferRange.xhtml">registry.khronos.org</a>
	 */
	public void glFlushMappedBufferRange(GLBufferTarget target, long offset, long length) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glFlushMappedBufferRange(\{target}, \{offset}, \{length})");
		try {
		    functions.glFlushMappedBufferRange().invokeExact(target, offset, length);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glFlushMappedBufferRangeAPPLE">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glFlushMappedBufferRangeAPPLE.xhtml">registry.khronos.org</a>
	 */
	public void glFlushMappedBufferRangeAPPLE(GLBufferTarget target, long offset, long size) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glFlushMappedBufferRangeAPPLE(\{target}, \{offset}, \{size})");
		try {
		    functions.glFlushMappedBufferRangeAPPLE().invokeExact(target, offset, size);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glFlushMappedBufferRangeEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glFlushMappedBufferRangeEXT.xhtml">registry.khronos.org</a>
	 */
	public void glFlushMappedBufferRangeEXT(GLBufferTarget target, long offset, long length) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glFlushMappedBufferRangeEXT(\{target}, \{offset}, \{length})");
		try {
		    functions.glFlushMappedBufferRangeEXT().invokeExact(target, offset, length);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glFlushMappedNamedBufferRange">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glFlushMappedNamedBufferRange.xhtml">registry.khronos.org</a>
	 */
	public void glFlushMappedNamedBufferRange(int buffer, long offset, long length) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glFlushMappedNamedBufferRange(\{buffer}, \{offset}, \{length})");
		try {
		    functions.glFlushMappedNamedBufferRange().invokeExact(buffer, offset, length);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glFlushMappedNamedBufferRangeEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glFlushMappedNamedBufferRangeEXT.xhtml">registry.khronos.org</a>
	 */
	public void glFlushMappedNamedBufferRangeEXT(int buffer, long offset, long length) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glFlushMappedNamedBufferRangeEXT(\{buffer}, \{offset}, \{length})");
		try {
		    functions.glFlushMappedNamedBufferRangeEXT().invokeExact(buffer, offset, length);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glFlushPixelDataRangeNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glFlushPixelDataRangeNV.xhtml">registry.khronos.org</a>
	 */
	public void glFlushPixelDataRangeNV(GLPixelDataRangeTargetNV target) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glFlushPixelDataRangeNV(\{target})");
		try {
		    functions.glFlushPixelDataRangeNV().invokeExact(target);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glFlushRasterSGIX">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glFlushRasterSGIX.xhtml">registry.khronos.org</a>
	 */
	public void glFlushRasterSGIX() {
	    if (LOG_COMMANDS) logCmd(() -> STR."glFlushRasterSGIX()");
		try {
		    functions.glFlushRasterSGIX().invokeExact();
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glFlushStaticDataIBM">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glFlushStaticDataIBM.xhtml">registry.khronos.org</a>
	 */
	public void glFlushStaticDataIBM(GLEnum target) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glFlushStaticDataIBM(\{target})");
		try {
		    functions.glFlushStaticDataIBM().invokeExact(target);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glFlushVertexArrayRangeAPPLE">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glFlushVertexArrayRangeAPPLE.xhtml">registry.khronos.org</a>
	 */
	public void glFlushVertexArrayRangeAPPLE(int length, MutPointer<Void> pointer) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glFlushVertexArrayRangeAPPLE(\{length}, \{pointer})");
		try {
		    functions.glFlushVertexArrayRangeAPPLE().invokeExact(length, pointer);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glFlushVertexArrayRangeNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glFlushVertexArrayRangeNV.xhtml">registry.khronos.org</a>
	 */
	public void glFlushVertexArrayRangeNV() {
	    if (LOG_COMMANDS) logCmd(() -> STR."glFlushVertexArrayRangeNV()");
		try {
		    functions.glFlushVertexArrayRangeNV().invokeExact();
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glFogCoordFormatNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glFogCoordFormatNV.xhtml">registry.khronos.org</a>
	 */
	public void glFogCoordFormatNV(GLEnum type, int stride) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glFogCoordFormatNV(\{type}, \{stride})");
		try {
		    functions.glFogCoordFormatNV().invokeExact(type, stride);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glFogCoordPointer">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glFogCoordPointer.xhtml">registry.khronos.org</a>
	 */
	public void glFogCoordPointer(GLFogPointerTypeEXT type, int stride, ConstPointer<Void> pointer) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glFogCoordPointer(\{type}, \{stride}, \{pointer})");
		try {
		    functions.glFogCoordPointer().invokeExact(type, stride, pointer);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glFogCoordPointerEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glFogCoordPointerEXT.xhtml">registry.khronos.org</a>
	 */
	public void glFogCoordPointerEXT(GLFogPointerTypeEXT type, int stride, ConstPointer<Void> pointer) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glFogCoordPointerEXT(\{type}, \{stride}, \{pointer})");
		try {
		    functions.glFogCoordPointerEXT().invokeExact(type, stride, pointer);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glFogCoordPointerListIBM">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glFogCoordPointerListIBM.xhtml">registry.khronos.org</a>
	 */
	public void glFogCoordPointerListIBM(GLFogPointerTypeIBM type, int stride, MutPointer<ConstPointer<Void>> pointer, int ptrstride) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glFogCoordPointerListIBM(\{type}, \{stride}, \{pointer}, \{ptrstride})");
		try {
		    functions.glFogCoordPointerListIBM().invokeExact(type, stride, pointer, ptrstride);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glFogCoordd">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glFogCoordd.xhtml">registry.khronos.org</a>
	 */
	public void glFogCoordd(double coord) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glFogCoordd(\{coord})");
		try {
		    functions.glFogCoordd().invokeExact(coord);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glFogCoorddEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glFogCoorddEXT.xhtml">registry.khronos.org</a>
	 */
	public void glFogCoorddEXT(double coord) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glFogCoorddEXT(\{coord})");
		try {
		    functions.glFogCoorddEXT().invokeExact(coord);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glFogCoorddv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glFogCoorddv.xhtml">registry.khronos.org</a>
	 */
	public void glFogCoorddv(ConstPointer<Double> coord) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glFogCoorddv(\{coord})");
		try {
		    functions.glFogCoorddv().invokeExact(coord);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glFogCoorddvEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glFogCoorddvEXT.xhtml">registry.khronos.org</a>
	 */
	public void glFogCoorddvEXT(ConstPointer<Double> coord) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glFogCoorddvEXT(\{coord})");
		try {
		    functions.glFogCoorddvEXT().invokeExact(coord);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glFogCoordf">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glFogCoordf.xhtml">registry.khronos.org</a>
	 */
	public void glFogCoordf(float coord) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glFogCoordf(\{coord})");
		try {
		    functions.glFogCoordf().invokeExact(coord);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glFogCoordfEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glFogCoordfEXT.xhtml">registry.khronos.org</a>
	 */
	public void glFogCoordfEXT(float coord) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glFogCoordfEXT(\{coord})");
		try {
		    functions.glFogCoordfEXT().invokeExact(coord);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glFogCoordfv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glFogCoordfv.xhtml">registry.khronos.org</a>
	 */
	public void glFogCoordfv(ConstPointer<Float> coord) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glFogCoordfv(\{coord})");
		try {
		    functions.glFogCoordfv().invokeExact(coord);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glFogCoordfvEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glFogCoordfvEXT.xhtml">registry.khronos.org</a>
	 */
	public void glFogCoordfvEXT(ConstPointer<Float> coord) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glFogCoordfvEXT(\{coord})");
		try {
		    functions.glFogCoordfvEXT().invokeExact(coord);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glFogCoordhNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glFogCoordhNV.xhtml">registry.khronos.org</a>
	 */
	public void glFogCoordhNV(char fog) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glFogCoordhNV(\{fog})");
		try {
		    functions.glFogCoordhNV().invokeExact(fog);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glFogCoordhvNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glFogCoordhvNV.xhtml">registry.khronos.org</a>
	 */
	public void glFogCoordhvNV(ConstPointer<Character> fog) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glFogCoordhvNV(\{fog})");
		try {
		    functions.glFogCoordhvNV().invokeExact(fog);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glFogFuncSGIS">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glFogFuncSGIS.xhtml">registry.khronos.org</a>
	 */
	public void glFogFuncSGIS(int n, ConstPointer<Float> points) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glFogFuncSGIS(\{n}, \{points})");
		try {
		    functions.glFogFuncSGIS().invokeExact(n, points);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glFogf">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glFogf.xhtml">registry.khronos.org</a>
	 */
	public void glFogf(GLFogParameter pname, float param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glFogf(\{pname}, \{param})");
		try {
		    functions.glFogf().invokeExact(pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glFogfv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glFogfv.xhtml">registry.khronos.org</a>
	 */
	public void glFogfv(GLFogParameter pname, ConstPointer<Float> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glFogfv(\{pname}, \{params})");
		try {
		    functions.glFogfv().invokeExact(pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glFogi">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glFogi.xhtml">registry.khronos.org</a>
	 */
	public void glFogi(GLFogParameter pname, int param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glFogi(\{pname}, \{param})");
		try {
		    functions.glFogi().invokeExact(pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glFogiv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glFogiv.xhtml">registry.khronos.org</a>
	 */
	public void glFogiv(GLFogParameter pname, ConstPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glFogiv(\{pname}, \{params})");
		try {
		    functions.glFogiv().invokeExact(pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glFogx">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glFogx.xhtml">registry.khronos.org</a>
	 */
	public void glFogx(GLFogPName pname, int param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glFogx(\{pname}, \{param})");
		try {
		    functions.glFogx().invokeExact(pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glFogxOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glFogxOES.xhtml">registry.khronos.org</a>
	 */
	public void glFogxOES(GLFogPName pname, int param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glFogxOES(\{pname}, \{param})");
		try {
		    functions.glFogxOES().invokeExact(pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glFogxv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glFogxv.xhtml">registry.khronos.org</a>
	 */
	public void glFogxv(GLFogPName pname, ConstPointer<Integer> param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glFogxv(\{pname}, \{param})");
		try {
		    functions.glFogxv().invokeExact(pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glFogxvOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glFogxvOES.xhtml">registry.khronos.org</a>
	 */
	public void glFogxvOES(GLFogPName pname, ConstPointer<Integer> param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glFogxvOES(\{pname}, \{param})");
		try {
		    functions.glFogxvOES().invokeExact(pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glFragmentColorMaterialSGIX">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glFragmentColorMaterialSGIX.xhtml">registry.khronos.org</a>
	 */
	public void glFragmentColorMaterialSGIX(GLTriangleFace face, GLMaterialParameter mode) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glFragmentColorMaterialSGIX(\{face}, \{mode})");
		try {
		    functions.glFragmentColorMaterialSGIX().invokeExact(face, mode);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were face=\{face}, mode=\{mode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glFragmentCoverageColorNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glFragmentCoverageColorNV.xhtml">registry.khronos.org</a>
	 */
	public void glFragmentCoverageColorNV(int color) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glFragmentCoverageColorNV(\{color})");
		try {
		    functions.glFragmentCoverageColorNV().invokeExact(color);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glFragmentLightModelfSGIX">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glFragmentLightModelfSGIX.xhtml">registry.khronos.org</a>
	 */
	public void glFragmentLightModelfSGIX(GLFragmentLightModelParameterSGIX pname, float param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glFragmentLightModelfSGIX(\{pname}, \{param})");
		try {
		    functions.glFragmentLightModelfSGIX().invokeExact(pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glFragmentLightModelfvSGIX">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glFragmentLightModelfvSGIX.xhtml">registry.khronos.org</a>
	 */
	public void glFragmentLightModelfvSGIX(GLFragmentLightModelParameterSGIX pname, ConstPointer<Float> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glFragmentLightModelfvSGIX(\{pname}, \{params})");
		try {
		    functions.glFragmentLightModelfvSGIX().invokeExact(pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glFragmentLightModeliSGIX">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glFragmentLightModeliSGIX.xhtml">registry.khronos.org</a>
	 */
	public void glFragmentLightModeliSGIX(GLFragmentLightModelParameterSGIX pname, int param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glFragmentLightModeliSGIX(\{pname}, \{param})");
		try {
		    functions.glFragmentLightModeliSGIX().invokeExact(pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glFragmentLightModelivSGIX">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glFragmentLightModelivSGIX.xhtml">registry.khronos.org</a>
	 */
	public void glFragmentLightModelivSGIX(GLFragmentLightModelParameterSGIX pname, ConstPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glFragmentLightModelivSGIX(\{pname}, \{params})");
		try {
		    functions.glFragmentLightModelivSGIX().invokeExact(pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glFragmentLightfSGIX">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glFragmentLightfSGIX.xhtml">registry.khronos.org</a>
	 */
	public void glFragmentLightfSGIX(GLFragmentLightNameSGIX light, GLFragmentLightParameterSGIX pname, float param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glFragmentLightfSGIX(\{light}, \{pname}, \{param})");
		try {
		    functions.glFragmentLightfSGIX().invokeExact(light, pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were light=\{light}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glFragmentLightfvSGIX">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glFragmentLightfvSGIX.xhtml">registry.khronos.org</a>
	 */
	public void glFragmentLightfvSGIX(GLFragmentLightNameSGIX light, GLFragmentLightParameterSGIX pname, ConstPointer<Float> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glFragmentLightfvSGIX(\{light}, \{pname}, \{params})");
		try {
		    functions.glFragmentLightfvSGIX().invokeExact(light, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were light=\{light}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glFragmentLightiSGIX">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glFragmentLightiSGIX.xhtml">registry.khronos.org</a>
	 */
	public void glFragmentLightiSGIX(GLFragmentLightNameSGIX light, GLFragmentLightParameterSGIX pname, int param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glFragmentLightiSGIX(\{light}, \{pname}, \{param})");
		try {
		    functions.glFragmentLightiSGIX().invokeExact(light, pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were light=\{light}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glFragmentLightivSGIX">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glFragmentLightivSGIX.xhtml">registry.khronos.org</a>
	 */
	public void glFragmentLightivSGIX(GLFragmentLightNameSGIX light, GLFragmentLightParameterSGIX pname, ConstPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glFragmentLightivSGIX(\{light}, \{pname}, \{params})");
		try {
		    functions.glFragmentLightivSGIX().invokeExact(light, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were light=\{light}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glFragmentMaterialfSGIX">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glFragmentMaterialfSGIX.xhtml">registry.khronos.org</a>
	 */
	public void glFragmentMaterialfSGIX(GLTriangleFace face, GLMaterialParameter pname, float param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glFragmentMaterialfSGIX(\{face}, \{pname}, \{param})");
		try {
		    functions.glFragmentMaterialfSGIX().invokeExact(face, pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were face=\{face}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glFragmentMaterialfvSGIX">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glFragmentMaterialfvSGIX.xhtml">registry.khronos.org</a>
	 */
	public void glFragmentMaterialfvSGIX(GLTriangleFace face, GLMaterialParameter pname, ConstPointer<Float> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glFragmentMaterialfvSGIX(\{face}, \{pname}, \{params})");
		try {
		    functions.glFragmentMaterialfvSGIX().invokeExact(face, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were face=\{face}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glFragmentMaterialiSGIX">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glFragmentMaterialiSGIX.xhtml">registry.khronos.org</a>
	 */
	public void glFragmentMaterialiSGIX(GLTriangleFace face, GLMaterialParameter pname, int param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glFragmentMaterialiSGIX(\{face}, \{pname}, \{param})");
		try {
		    functions.glFragmentMaterialiSGIX().invokeExact(face, pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were face=\{face}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glFragmentMaterialivSGIX">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glFragmentMaterialivSGIX.xhtml">registry.khronos.org</a>
	 */
	public void glFragmentMaterialivSGIX(GLTriangleFace face, GLMaterialParameter pname, ConstPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glFragmentMaterialivSGIX(\{face}, \{pname}, \{params})");
		try {
		    functions.glFragmentMaterialivSGIX().invokeExact(face, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were face=\{face}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glFrameTerminatorGREMEDY">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glFrameTerminatorGREMEDY.xhtml">registry.khronos.org</a>
	 */
	public void glFrameTerminatorGREMEDY() {
	    if (LOG_COMMANDS) logCmd(() -> STR."glFrameTerminatorGREMEDY()");
		try {
		    functions.glFrameTerminatorGREMEDY().invokeExact();
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glFrameZoomSGIX">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glFrameZoomSGIX.xhtml">registry.khronos.org</a>
	 */
	public void glFrameZoomSGIX(int factor) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glFrameZoomSGIX(\{factor})");
		try {
		    functions.glFrameZoomSGIX().invokeExact(factor);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glFramebufferDrawBufferEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glFramebufferDrawBufferEXT.xhtml">registry.khronos.org</a>
	 */
	public void glFramebufferDrawBufferEXT(int framebuffer, GLDrawBufferMode mode) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glFramebufferDrawBufferEXT(\{framebuffer}, \{mode})");
		try {
		    functions.glFramebufferDrawBufferEXT().invokeExact(framebuffer, mode);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glFramebufferDrawBuffersEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glFramebufferDrawBuffersEXT.xhtml">registry.khronos.org</a>
	 */
	public void glFramebufferDrawBuffersEXT(int framebuffer, int n, ConstPointer<GLDrawBufferMode> bufs) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glFramebufferDrawBuffersEXT(\{framebuffer}, \{n}, \{bufs})");
		try {
		    functions.glFramebufferDrawBuffersEXT().invokeExact(framebuffer, n, bufs);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glFramebufferFetchBarrierEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glFramebufferFetchBarrierEXT.xhtml">registry.khronos.org</a>
	 */
	public void glFramebufferFetchBarrierEXT() {
	    if (LOG_COMMANDS) logCmd(() -> STR."glFramebufferFetchBarrierEXT()");
		try {
		    functions.glFramebufferFetchBarrierEXT().invokeExact();
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glFramebufferFetchBarrierQCOM">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glFramebufferFetchBarrierQCOM.xhtml">registry.khronos.org</a>
	 */
	public void glFramebufferFetchBarrierQCOM() {
	    if (LOG_COMMANDS) logCmd(() -> STR."glFramebufferFetchBarrierQCOM()");
		try {
		    functions.glFramebufferFetchBarrierQCOM().invokeExact();
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glFramebufferFoveationConfigQCOM">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glFramebufferFoveationConfigQCOM.xhtml">registry.khronos.org</a>
	 */
	public void glFramebufferFoveationConfigQCOM(int framebuffer, int numLayers, int focalPointsPerLayer, int requestedFeatures, MutPointer<Integer> providedFeatures) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glFramebufferFoveationConfigQCOM(\{framebuffer}, \{numLayers}, \{focalPointsPerLayer}, \{requestedFeatures}, \{providedFeatures})");
		try {
		    functions.glFramebufferFoveationConfigQCOM().invokeExact(framebuffer, numLayers, focalPointsPerLayer, requestedFeatures, providedFeatures);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glFramebufferFoveationParametersQCOM">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glFramebufferFoveationParametersQCOM.xhtml">registry.khronos.org</a>
	 */
	public void glFramebufferFoveationParametersQCOM(int framebuffer, int layer, int focalPoint, float focalX, float focalY, float gainX, float gainY, float foveaArea) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glFramebufferFoveationParametersQCOM(\{framebuffer}, \{layer}, \{focalPoint}, \{focalX}, \{focalY}, \{gainX}, \{gainY}, \{foveaArea})");
		try {
		    functions.glFramebufferFoveationParametersQCOM().invokeExact(framebuffer, layer, focalPoint, focalX, focalY, gainX, gainY, foveaArea);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glFramebufferParameteri">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glFramebufferParameteri.xhtml">registry.khronos.org</a>
	 */
	public void glFramebufferParameteri(GLFramebufferTarget target, GLFramebufferParameterName pname, int param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glFramebufferParameteri(\{target}, \{pname}, \{param})");
		try {
		    functions.glFramebufferParameteri().invokeExact(target, pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glFramebufferPixelLocalStorageSizeEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glFramebufferPixelLocalStorageSizeEXT.xhtml">registry.khronos.org</a>
	 */
	public void glFramebufferPixelLocalStorageSizeEXT(int target, int size) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glFramebufferPixelLocalStorageSizeEXT(\{target}, \{size})");
		try {
		    functions.glFramebufferPixelLocalStorageSizeEXT().invokeExact(target, size);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glFramebufferReadBufferEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glFramebufferReadBufferEXT.xhtml">registry.khronos.org</a>
	 */
	public void glFramebufferReadBufferEXT(int framebuffer, GLReadBufferMode mode) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glFramebufferReadBufferEXT(\{framebuffer}, \{mode})");
		try {
		    functions.glFramebufferReadBufferEXT().invokeExact(framebuffer, mode);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glFramebufferRenderbuffer">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glFramebufferRenderbuffer.xhtml">registry.khronos.org</a>
	 */
	public void glFramebufferRenderbuffer(GLFramebufferTarget target, GLFramebufferAttachment attachment, GLRenderbufferTarget renderbuffertarget, int renderbuffer) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glFramebufferRenderbuffer(\{target}, \{attachment}, \{renderbuffertarget}, \{renderbuffer})");
		try {
		    functions.glFramebufferRenderbuffer().invokeExact(target, attachment, renderbuffertarget, renderbuffer);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, attachment=\{attachment}, renderbuffertarget=\{renderbuffertarget}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glFramebufferRenderbufferEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glFramebufferRenderbufferEXT.xhtml">registry.khronos.org</a>
	 */
	public void glFramebufferRenderbufferEXT(GLFramebufferTarget target, GLFramebufferAttachment attachment, GLRenderbufferTarget renderbuffertarget, int renderbuffer) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glFramebufferRenderbufferEXT(\{target}, \{attachment}, \{renderbuffertarget}, \{renderbuffer})");
		try {
		    functions.glFramebufferRenderbufferEXT().invokeExact(target, attachment, renderbuffertarget, renderbuffer);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, attachment=\{attachment}, renderbuffertarget=\{renderbuffertarget}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glFramebufferRenderbufferOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glFramebufferRenderbufferOES.xhtml">registry.khronos.org</a>
	 */
	public void glFramebufferRenderbufferOES(GLFramebufferTarget target, GLFramebufferAttachment attachment, GLRenderbufferTarget renderbuffertarget, int renderbuffer) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glFramebufferRenderbufferOES(\{target}, \{attachment}, \{renderbuffertarget}, \{renderbuffer})");
		try {
		    functions.glFramebufferRenderbufferOES().invokeExact(target, attachment, renderbuffertarget, renderbuffer);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, attachment=\{attachment}, renderbuffertarget=\{renderbuffertarget}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glFramebufferSampleLocationsfvARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glFramebufferSampleLocationsfvARB.xhtml">registry.khronos.org</a>
	 */
	public void glFramebufferSampleLocationsfvARB(GLFramebufferTarget target, int start, int count, ConstPointer<Float> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glFramebufferSampleLocationsfvARB(\{target}, \{start}, \{count}, \{v})");
		try {
		    functions.glFramebufferSampleLocationsfvARB().invokeExact(target, start, count, v);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glFramebufferSampleLocationsfvNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glFramebufferSampleLocationsfvNV.xhtml">registry.khronos.org</a>
	 */
	public void glFramebufferSampleLocationsfvNV(GLFramebufferTarget target, int start, int count, ConstPointer<Float> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glFramebufferSampleLocationsfvNV(\{target}, \{start}, \{count}, \{v})");
		try {
		    functions.glFramebufferSampleLocationsfvNV().invokeExact(target, start, count, v);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glFramebufferSamplePositionsfvAMD">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glFramebufferSamplePositionsfvAMD.xhtml">registry.khronos.org</a>
	 */
	public void glFramebufferSamplePositionsfvAMD(GLFramebufferTarget target, int numsamples, int pixelindex, ConstPointer<Float> values) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glFramebufferSamplePositionsfvAMD(\{target}, \{numsamples}, \{pixelindex}, \{values})");
		try {
		    functions.glFramebufferSamplePositionsfvAMD().invokeExact(target, numsamples, pixelindex, values);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glFramebufferShadingRateEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glFramebufferShadingRateEXT.xhtml">registry.khronos.org</a>
	 */
	public void glFramebufferShadingRateEXT(GLFramebufferTarget target, GLFramebufferAttachment attachment, int texture, int baseLayer, int numLayers, int texelWidth, int texelHeight) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glFramebufferShadingRateEXT(\{target}, \{attachment}, \{texture}, \{baseLayer}, \{numLayers}, \{texelWidth}, \{texelHeight})");
		try {
		    functions.glFramebufferShadingRateEXT().invokeExact(target, attachment, texture, baseLayer, numLayers, texelWidth, texelHeight);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, attachment=\{attachment}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glFramebufferTexture">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glFramebufferTexture.xhtml">registry.khronos.org</a>
	 */
	public void glFramebufferTexture(GLFramebufferTarget target, GLFramebufferAttachment attachment, int texture, int level) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glFramebufferTexture(\{target}, \{attachment}, \{texture}, \{level})");
		try {
		    functions.glFramebufferTexture().invokeExact(target, attachment, texture, level);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, attachment=\{attachment}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glFramebufferTexture1D">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glFramebufferTexture1D.xhtml">registry.khronos.org</a>
	 */
	public void glFramebufferTexture1D(GLFramebufferTarget target, GLFramebufferAttachment attachment, GLTextureTarget textarget, int texture, int level) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glFramebufferTexture1D(\{target}, \{attachment}, \{textarget}, \{texture}, \{level})");
		try {
		    functions.glFramebufferTexture1D().invokeExact(target, attachment, textarget, texture, level);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, attachment=\{attachment}, textarget=\{textarget}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glFramebufferTexture1DEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glFramebufferTexture1DEXT.xhtml">registry.khronos.org</a>
	 */
	public void glFramebufferTexture1DEXT(GLFramebufferTarget target, GLFramebufferAttachment attachment, GLTextureTarget textarget, int texture, int level) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glFramebufferTexture1DEXT(\{target}, \{attachment}, \{textarget}, \{texture}, \{level})");
		try {
		    functions.glFramebufferTexture1DEXT().invokeExact(target, attachment, textarget, texture, level);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, attachment=\{attachment}, textarget=\{textarget}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glFramebufferTexture2D">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glFramebufferTexture2D.xhtml">registry.khronos.org</a>
	 */
	public void glFramebufferTexture2D(GLFramebufferTarget target, GLFramebufferAttachment attachment, GLTextureTarget textarget, int texture, int level) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glFramebufferTexture2D(\{target}, \{attachment}, \{textarget}, \{texture}, \{level})");
		try {
		    functions.glFramebufferTexture2D().invokeExact(target, attachment, textarget, texture, level);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, attachment=\{attachment}, textarget=\{textarget}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glFramebufferTexture2DEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glFramebufferTexture2DEXT.xhtml">registry.khronos.org</a>
	 */
	public void glFramebufferTexture2DEXT(GLFramebufferTarget target, GLFramebufferAttachment attachment, GLTextureTarget textarget, int texture, int level) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glFramebufferTexture2DEXT(\{target}, \{attachment}, \{textarget}, \{texture}, \{level})");
		try {
		    functions.glFramebufferTexture2DEXT().invokeExact(target, attachment, textarget, texture, level);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, attachment=\{attachment}, textarget=\{textarget}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glFramebufferTexture2DDownsampleIMG">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glFramebufferTexture2DDownsampleIMG.xhtml">registry.khronos.org</a>
	 */
	public void glFramebufferTexture2DDownsampleIMG(GLFramebufferTarget target, GLFramebufferAttachment attachment, GLTextureTarget textarget, int texture, int level, int xscale, int yscale) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glFramebufferTexture2DDownsampleIMG(\{target}, \{attachment}, \{textarget}, \{texture}, \{level}, \{xscale}, \{yscale})");
		try {
		    functions.glFramebufferTexture2DDownsampleIMG().invokeExact(target, attachment, textarget, texture, level, xscale, yscale);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, attachment=\{attachment}, textarget=\{textarget}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glFramebufferTexture2DMultisampleEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glFramebufferTexture2DMultisampleEXT.xhtml">registry.khronos.org</a>
	 */
	public void glFramebufferTexture2DMultisampleEXT(GLFramebufferTarget target, GLFramebufferAttachment attachment, GLTextureTarget textarget, int texture, int level, int samples) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glFramebufferTexture2DMultisampleEXT(\{target}, \{attachment}, \{textarget}, \{texture}, \{level}, \{samples})");
		try {
		    functions.glFramebufferTexture2DMultisampleEXT().invokeExact(target, attachment, textarget, texture, level, samples);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, attachment=\{attachment}, textarget=\{textarget}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glFramebufferTexture2DMultisampleIMG">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glFramebufferTexture2DMultisampleIMG.xhtml">registry.khronos.org</a>
	 */
	public void glFramebufferTexture2DMultisampleIMG(GLFramebufferTarget target, GLFramebufferAttachment attachment, GLTextureTarget textarget, int texture, int level, int samples) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glFramebufferTexture2DMultisampleIMG(\{target}, \{attachment}, \{textarget}, \{texture}, \{level}, \{samples})");
		try {
		    functions.glFramebufferTexture2DMultisampleIMG().invokeExact(target, attachment, textarget, texture, level, samples);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, attachment=\{attachment}, textarget=\{textarget}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glFramebufferTexture2DOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glFramebufferTexture2DOES.xhtml">registry.khronos.org</a>
	 */
	public void glFramebufferTexture2DOES(GLFramebufferTarget target, GLFramebufferAttachment attachment, GLTextureTarget textarget, int texture, int level) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glFramebufferTexture2DOES(\{target}, \{attachment}, \{textarget}, \{texture}, \{level})");
		try {
		    functions.glFramebufferTexture2DOES().invokeExact(target, attachment, textarget, texture, level);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, attachment=\{attachment}, textarget=\{textarget}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glFramebufferTexture3D">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glFramebufferTexture3D.xhtml">registry.khronos.org</a>
	 */
	public void glFramebufferTexture3D(GLFramebufferTarget target, GLFramebufferAttachment attachment, GLTextureTarget textarget, int texture, int level, int zoffset) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glFramebufferTexture3D(\{target}, \{attachment}, \{textarget}, \{texture}, \{level}, \{zoffset})");
		try {
		    functions.glFramebufferTexture3D().invokeExact(target, attachment, textarget, texture, level, zoffset);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, attachment=\{attachment}, textarget=\{textarget}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glFramebufferTexture3DEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glFramebufferTexture3DEXT.xhtml">registry.khronos.org</a>
	 */
	public void glFramebufferTexture3DEXT(GLFramebufferTarget target, GLFramebufferAttachment attachment, GLTextureTarget textarget, int texture, int level, int zoffset) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glFramebufferTexture3DEXT(\{target}, \{attachment}, \{textarget}, \{texture}, \{level}, \{zoffset})");
		try {
		    functions.glFramebufferTexture3DEXT().invokeExact(target, attachment, textarget, texture, level, zoffset);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, attachment=\{attachment}, textarget=\{textarget}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glFramebufferTexture3DOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glFramebufferTexture3DOES.xhtml">registry.khronos.org</a>
	 */
	public void glFramebufferTexture3DOES(GLFramebufferTarget target, GLFramebufferAttachment attachment, GLTextureTarget textarget, int texture, int level, int zoffset) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glFramebufferTexture3DOES(\{target}, \{attachment}, \{textarget}, \{texture}, \{level}, \{zoffset})");
		try {
		    functions.glFramebufferTexture3DOES().invokeExact(target, attachment, textarget, texture, level, zoffset);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, attachment=\{attachment}, textarget=\{textarget}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glFramebufferTextureARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glFramebufferTextureARB.xhtml">registry.khronos.org</a>
	 */
	public void glFramebufferTextureARB(GLFramebufferTarget target, GLFramebufferAttachment attachment, int texture, int level) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glFramebufferTextureARB(\{target}, \{attachment}, \{texture}, \{level})");
		try {
		    functions.glFramebufferTextureARB().invokeExact(target, attachment, texture, level);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, attachment=\{attachment}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glFramebufferTextureEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glFramebufferTextureEXT.xhtml">registry.khronos.org</a>
	 */
	public void glFramebufferTextureEXT(GLFramebufferTarget target, GLFramebufferAttachment attachment, int texture, int level) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glFramebufferTextureEXT(\{target}, \{attachment}, \{texture}, \{level})");
		try {
		    functions.glFramebufferTextureEXT().invokeExact(target, attachment, texture, level);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, attachment=\{attachment}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glFramebufferTextureFaceARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glFramebufferTextureFaceARB.xhtml">registry.khronos.org</a>
	 */
	public void glFramebufferTextureFaceARB(GLFramebufferTarget target, GLFramebufferAttachment attachment, int texture, int level, GLTextureTarget face) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glFramebufferTextureFaceARB(\{target}, \{attachment}, \{texture}, \{level}, \{face})");
		try {
		    functions.glFramebufferTextureFaceARB().invokeExact(target, attachment, texture, level, face);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, attachment=\{attachment}, face=\{face}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glFramebufferTextureFaceEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glFramebufferTextureFaceEXT.xhtml">registry.khronos.org</a>
	 */
	public void glFramebufferTextureFaceEXT(GLFramebufferTarget target, GLFramebufferAttachment attachment, int texture, int level, GLTextureTarget face) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glFramebufferTextureFaceEXT(\{target}, \{attachment}, \{texture}, \{level}, \{face})");
		try {
		    functions.glFramebufferTextureFaceEXT().invokeExact(target, attachment, texture, level, face);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, attachment=\{attachment}, face=\{face}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glFramebufferTextureLayer">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glFramebufferTextureLayer.xhtml">registry.khronos.org</a>
	 */
	public void glFramebufferTextureLayer(GLFramebufferTarget target, GLFramebufferAttachment attachment, int texture, int level, int layer) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glFramebufferTextureLayer(\{target}, \{attachment}, \{texture}, \{level}, \{layer})");
		try {
		    functions.glFramebufferTextureLayer().invokeExact(target, attachment, texture, level, layer);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, attachment=\{attachment}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glFramebufferTextureLayerARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glFramebufferTextureLayerARB.xhtml">registry.khronos.org</a>
	 */
	public void glFramebufferTextureLayerARB(GLFramebufferTarget target, GLFramebufferAttachment attachment, int texture, int level, int layer) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glFramebufferTextureLayerARB(\{target}, \{attachment}, \{texture}, \{level}, \{layer})");
		try {
		    functions.glFramebufferTextureLayerARB().invokeExact(target, attachment, texture, level, layer);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, attachment=\{attachment}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glFramebufferTextureLayerEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glFramebufferTextureLayerEXT.xhtml">registry.khronos.org</a>
	 */
	public void glFramebufferTextureLayerEXT(GLFramebufferTarget target, GLFramebufferAttachment attachment, int texture, int level, int layer) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glFramebufferTextureLayerEXT(\{target}, \{attachment}, \{texture}, \{level}, \{layer})");
		try {
		    functions.glFramebufferTextureLayerEXT().invokeExact(target, attachment, texture, level, layer);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, attachment=\{attachment}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glFramebufferTextureLayerDownsampleIMG">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glFramebufferTextureLayerDownsampleIMG.xhtml">registry.khronos.org</a>
	 */
	public void glFramebufferTextureLayerDownsampleIMG(GLFramebufferTarget target, GLFramebufferAttachment attachment, int texture, int level, int layer, int xscale, int yscale) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glFramebufferTextureLayerDownsampleIMG(\{target}, \{attachment}, \{texture}, \{level}, \{layer}, \{xscale}, \{yscale})");
		try {
		    functions.glFramebufferTextureLayerDownsampleIMG().invokeExact(target, attachment, texture, level, layer, xscale, yscale);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, attachment=\{attachment}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glFramebufferTextureMultisampleMultiviewOVR">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glFramebufferTextureMultisampleMultiviewOVR.xhtml">registry.khronos.org</a>
	 */
	public void glFramebufferTextureMultisampleMultiviewOVR(GLFramebufferTarget target, GLFramebufferAttachment attachment, int texture, int level, int samples, int baseViewIndex, int numViews) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glFramebufferTextureMultisampleMultiviewOVR(\{target}, \{attachment}, \{texture}, \{level}, \{samples}, \{baseViewIndex}, \{numViews})");
		try {
		    functions.glFramebufferTextureMultisampleMultiviewOVR().invokeExact(target, attachment, texture, level, samples, baseViewIndex, numViews);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, attachment=\{attachment}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glFramebufferTextureMultiviewOVR">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glFramebufferTextureMultiviewOVR.xhtml">registry.khronos.org</a>
	 */
	public void glFramebufferTextureMultiviewOVR(GLFramebufferTarget target, GLFramebufferAttachment attachment, int texture, int level, int baseViewIndex, int numViews) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glFramebufferTextureMultiviewOVR(\{target}, \{attachment}, \{texture}, \{level}, \{baseViewIndex}, \{numViews})");
		try {
		    functions.glFramebufferTextureMultiviewOVR().invokeExact(target, attachment, texture, level, baseViewIndex, numViews);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, attachment=\{attachment}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glFramebufferTextureOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glFramebufferTextureOES.xhtml">registry.khronos.org</a>
	 */
	public void glFramebufferTextureOES(GLFramebufferTarget target, GLFramebufferAttachment attachment, int texture, int level) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glFramebufferTextureOES(\{target}, \{attachment}, \{texture}, \{level})");
		try {
		    functions.glFramebufferTextureOES().invokeExact(target, attachment, texture, level);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, attachment=\{attachment}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glFreeObjectBufferATI">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glFreeObjectBufferATI.xhtml">registry.khronos.org</a>
	 */
	public void glFreeObjectBufferATI(int buffer) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glFreeObjectBufferATI(\{buffer})");
		try {
		    functions.glFreeObjectBufferATI().invokeExact(buffer);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glFrontFace">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glFrontFace.xhtml">registry.khronos.org</a>
	 */
	public void glFrontFace(GLFrontFaceDirection mode) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glFrontFace(\{mode})");
		try {
		    functions.glFrontFace().invokeExact(mode);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glFrustum">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glFrustum.xhtml">registry.khronos.org</a>
	 */
	public void glFrustum(double left, double right, double bottom, double top, double zNear, double zFar) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glFrustum(\{left}, \{right}, \{bottom}, \{top}, \{zNear}, \{zFar})");
		try {
		    functions.glFrustum().invokeExact(left, right, bottom, top, zNear, zFar);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glFrustumf">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glFrustumf.xhtml">registry.khronos.org</a>
	 */
	public void glFrustumf(float l, float r, float b, float t, float n, float f) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glFrustumf(\{l}, \{r}, \{b}, \{t}, \{n}, \{f})");
		try {
		    functions.glFrustumf().invokeExact(l, r, b, t, n, f);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glFrustumfOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glFrustumfOES.xhtml">registry.khronos.org</a>
	 */
	public void glFrustumfOES(float l, float r, float b, float t, float n, float f) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glFrustumfOES(\{l}, \{r}, \{b}, \{t}, \{n}, \{f})");
		try {
		    functions.glFrustumfOES().invokeExact(l, r, b, t, n, f);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glFrustumx">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glFrustumx.xhtml">registry.khronos.org</a>
	 */
	public void glFrustumx(int l, int r, int b, int t, int n, int f) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glFrustumx(\{l}, \{r}, \{b}, \{t}, \{n}, \{f})");
		try {
		    functions.glFrustumx().invokeExact(l, r, b, t, n, f);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glFrustumxOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glFrustumxOES.xhtml">registry.khronos.org</a>
	 */
	public void glFrustumxOES(int l, int r, int b, int t, int n, int f) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glFrustumxOES(\{l}, \{r}, \{b}, \{t}, \{n}, \{f})");
		try {
		    functions.glFrustumxOES().invokeExact(l, r, b, t, n, f);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGenAsyncMarkersSGIX">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGenAsyncMarkersSGIX.xhtml">registry.khronos.org</a>
	 */
	public int glGenAsyncMarkersSGIX(int range) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGenAsyncMarkersSGIX(\{range})");
		try {
		    var r = (int)functions.glGenAsyncMarkersSGIX().invokeExact(range);
			checkError();
			return r;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGenBuffers">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGenBuffers.xhtml">registry.khronos.org</a>
	 */
	public void glGenBuffers(int n, MutPointer<Integer> buffers) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGenBuffers(\{n}, \{buffers})");
		try {
		    functions.glGenBuffers().invokeExact(n, buffers);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	public int glGenBuffer() {
        try (var arena = Arena.ofConfined()) {
            var buf = arena.allocate(GLTypeCache.ML_I());
            glGenBuffers(1, new MutPointer<>(buf));
			var rvalue = buf.getAtIndex(GLTypeCache.ML_I(), 0);
            return rvalue;
        }
    }

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGenBuffersARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGenBuffersARB.xhtml">registry.khronos.org</a>
	 */
	public void glGenBuffersARB(int n, MutPointer<Integer> buffers) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGenBuffersARB(\{n}, \{buffers})");
		try {
		    functions.glGenBuffersARB().invokeExact(n, buffers);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGenFencesAPPLE">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGenFencesAPPLE.xhtml">registry.khronos.org</a>
	 */
	public void glGenFencesAPPLE(int n, MutPointer<Integer> fences) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGenFencesAPPLE(\{n}, \{fences})");
		try {
		    functions.glGenFencesAPPLE().invokeExact(n, fences);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGenFencesNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGenFencesNV.xhtml">registry.khronos.org</a>
	 */
	public void glGenFencesNV(int n, MutPointer<Integer> fences) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGenFencesNV(\{n}, \{fences})");
		try {
		    functions.glGenFencesNV().invokeExact(n, fences);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGenFragmentShadersATI">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGenFragmentShadersATI.xhtml">registry.khronos.org</a>
	 */
	public int glGenFragmentShadersATI(int range) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGenFragmentShadersATI(\{range})");
		try {
		    var r = (int)functions.glGenFragmentShadersATI().invokeExact(range);
			checkError();
			return r;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGenFramebuffers">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGenFramebuffers.xhtml">registry.khronos.org</a>
	 */
	public void glGenFramebuffers(int n, MutPointer<Integer> framebuffers) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGenFramebuffers(\{n}, \{framebuffers})");
		try {
		    functions.glGenFramebuffers().invokeExact(n, framebuffers);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGenFramebuffersEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGenFramebuffersEXT.xhtml">registry.khronos.org</a>
	 */
	public void glGenFramebuffersEXT(int n, MutPointer<Integer> framebuffers) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGenFramebuffersEXT(\{n}, \{framebuffers})");
		try {
		    functions.glGenFramebuffersEXT().invokeExact(n, framebuffers);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGenFramebuffersOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGenFramebuffersOES.xhtml">registry.khronos.org</a>
	 */
	public void glGenFramebuffersOES(int n, MutPointer<Integer> framebuffers) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGenFramebuffersOES(\{n}, \{framebuffers})");
		try {
		    functions.glGenFramebuffersOES().invokeExact(n, framebuffers);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGenLists">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGenLists.xhtml">registry.khronos.org</a>
	 */
	public int glGenLists(int range) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGenLists(\{range})");
		try {
		    var r = (int)functions.glGenLists().invokeExact(range);
			checkError();
			return r;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGenNamesAMD">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGenNamesAMD.xhtml">registry.khronos.org</a>
	 */
	public void glGenNamesAMD(GLEnum identifier, int num, MutPointer<Integer> names) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGenNamesAMD(\{identifier}, \{num}, \{names})");
		try {
		    functions.glGenNamesAMD().invokeExact(identifier, num, names);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were identifier=\{identifier}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGenOcclusionQueriesNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGenOcclusionQueriesNV.xhtml">registry.khronos.org</a>
	 */
	public void glGenOcclusionQueriesNV(int n, MutPointer<Integer> ids) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGenOcclusionQueriesNV(\{n}, \{ids})");
		try {
		    functions.glGenOcclusionQueriesNV().invokeExact(n, ids);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGenPathsNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGenPathsNV.xhtml">registry.khronos.org</a>
	 */
	public int glGenPathsNV(int range) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGenPathsNV(\{range})");
		try {
		    var r = (int)functions.glGenPathsNV().invokeExact(range);
			checkError();
			return r;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGenPerfMonitorsAMD">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGenPerfMonitorsAMD.xhtml">registry.khronos.org</a>
	 */
	public void glGenPerfMonitorsAMD(int n, MutPointer<Integer> monitors) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGenPerfMonitorsAMD(\{n}, \{monitors})");
		try {
		    functions.glGenPerfMonitorsAMD().invokeExact(n, monitors);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGenProgramPipelines">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGenProgramPipelines.xhtml">registry.khronos.org</a>
	 */
	public void glGenProgramPipelines(int n, MutPointer<Integer> pipelines) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGenProgramPipelines(\{n}, \{pipelines})");
		try {
		    functions.glGenProgramPipelines().invokeExact(n, pipelines);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGenProgramPipelinesEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGenProgramPipelinesEXT.xhtml">registry.khronos.org</a>
	 */
	public void glGenProgramPipelinesEXT(int n, MutPointer<Integer> pipelines) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGenProgramPipelinesEXT(\{n}, \{pipelines})");
		try {
		    functions.glGenProgramPipelinesEXT().invokeExact(n, pipelines);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGenProgramsARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGenProgramsARB.xhtml">registry.khronos.org</a>
	 */
	public void glGenProgramsARB(int n, MutPointer<Integer> programs) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGenProgramsARB(\{n}, \{programs})");
		try {
		    functions.glGenProgramsARB().invokeExact(n, programs);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGenProgramsNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGenProgramsNV.xhtml">registry.khronos.org</a>
	 */
	public void glGenProgramsNV(int n, MutPointer<Integer> programs) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGenProgramsNV(\{n}, \{programs})");
		try {
		    functions.glGenProgramsNV().invokeExact(n, programs);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGenQueries">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGenQueries.xhtml">registry.khronos.org</a>
	 */
	public void glGenQueries(int n, MutPointer<Integer> ids) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGenQueries(\{n}, \{ids})");
		try {
		    functions.glGenQueries().invokeExact(n, ids);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGenQueriesARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGenQueriesARB.xhtml">registry.khronos.org</a>
	 */
	public void glGenQueriesARB(int n, MutPointer<Integer> ids) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGenQueriesARB(\{n}, \{ids})");
		try {
		    functions.glGenQueriesARB().invokeExact(n, ids);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGenQueriesEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGenQueriesEXT.xhtml">registry.khronos.org</a>
	 */
	public void glGenQueriesEXT(int n, MutPointer<Integer> ids) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGenQueriesEXT(\{n}, \{ids})");
		try {
		    functions.glGenQueriesEXT().invokeExact(n, ids);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGenQueryResourceTagNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGenQueryResourceTagNV.xhtml">registry.khronos.org</a>
	 */
	public void glGenQueryResourceTagNV(int n, MutPointer<Integer> tagIds) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGenQueryResourceTagNV(\{n}, \{tagIds})");
		try {
		    functions.glGenQueryResourceTagNV().invokeExact(n, tagIds);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGenRenderbuffers">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGenRenderbuffers.xhtml">registry.khronos.org</a>
	 */
	public void glGenRenderbuffers(int n, MutPointer<Integer> renderbuffers) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGenRenderbuffers(\{n}, \{renderbuffers})");
		try {
		    functions.glGenRenderbuffers().invokeExact(n, renderbuffers);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGenRenderbuffersEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGenRenderbuffersEXT.xhtml">registry.khronos.org</a>
	 */
	public void glGenRenderbuffersEXT(int n, MutPointer<Integer> renderbuffers) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGenRenderbuffersEXT(\{n}, \{renderbuffers})");
		try {
		    functions.glGenRenderbuffersEXT().invokeExact(n, renderbuffers);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGenRenderbuffersOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGenRenderbuffersOES.xhtml">registry.khronos.org</a>
	 */
	public void glGenRenderbuffersOES(int n, MutPointer<Integer> renderbuffers) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGenRenderbuffersOES(\{n}, \{renderbuffers})");
		try {
		    functions.glGenRenderbuffersOES().invokeExact(n, renderbuffers);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGenSamplers">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGenSamplers.xhtml">registry.khronos.org</a>
	 */
	public void glGenSamplers(int count, MutPointer<Integer> samplers) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGenSamplers(\{count}, \{samplers})");
		try {
		    functions.glGenSamplers().invokeExact(count, samplers);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGenSemaphoresEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGenSemaphoresEXT.xhtml">registry.khronos.org</a>
	 */
	public void glGenSemaphoresEXT(int n, MutPointer<Integer> semaphores) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGenSemaphoresEXT(\{n}, \{semaphores})");
		try {
		    functions.glGenSemaphoresEXT().invokeExact(n, semaphores);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGenSymbolsEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGenSymbolsEXT.xhtml">registry.khronos.org</a>
	 */
	public int glGenSymbolsEXT(GLDataTypeEXT datatype, GLVertexShaderStorageTypeEXT storagetype, GLParameterRangeEXT range, int components) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGenSymbolsEXT(\{datatype}, \{storagetype}, \{range}, \{components})");
		try {
		    var r = (int)functions.glGenSymbolsEXT().invokeExact(datatype, storagetype, range, components);
			checkError();
			return r;
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were datatype=\{datatype}, storagetype=\{storagetype}, range=\{range}.");
			}
			throw e;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGenTextures">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGenTextures.xhtml">registry.khronos.org</a>
	 */
	public void glGenTextures(int n, MutPointer<Integer> textures) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGenTextures(\{n}, \{textures})");
		try {
		    functions.glGenTextures().invokeExact(n, textures);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	public int glGenTexture() {
        try (var arena = Arena.ofConfined()) {
            var buf = arena.allocate(GLTypeCache.ML_I());
            glGenTextures(1, new MutPointer<>(buf));
			var rvalue = buf.getAtIndex(GLTypeCache.ML_I(), 0);
            return rvalue;
        }
    }

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGenTexturesEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGenTexturesEXT.xhtml">registry.khronos.org</a>
	 */
	public void glGenTexturesEXT(int n, MutPointer<Integer> textures) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGenTexturesEXT(\{n}, \{textures})");
		try {
		    functions.glGenTexturesEXT().invokeExact(n, textures);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGenTransformFeedbacks">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGenTransformFeedbacks.xhtml">registry.khronos.org</a>
	 */
	public void glGenTransformFeedbacks(int n, MutPointer<Integer> ids) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGenTransformFeedbacks(\{n}, \{ids})");
		try {
		    functions.glGenTransformFeedbacks().invokeExact(n, ids);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGenTransformFeedbacksNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGenTransformFeedbacksNV.xhtml">registry.khronos.org</a>
	 */
	public void glGenTransformFeedbacksNV(int n, MutPointer<Integer> ids) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGenTransformFeedbacksNV(\{n}, \{ids})");
		try {
		    functions.glGenTransformFeedbacksNV().invokeExact(n, ids);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGenVertexArrays">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGenVertexArrays.xhtml">registry.khronos.org</a>
	 */
	public void glGenVertexArrays(int n, MutPointer<Integer> arrays) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGenVertexArrays(\{n}, \{arrays})");
		try {
		    functions.glGenVertexArrays().invokeExact(n, arrays);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	public int glGenVertexArray() {
        try (var arena = Arena.ofConfined()) {
            var buf = arena.allocate(GLTypeCache.ML_I());
            glGenVertexArrays(1, new MutPointer<>(buf));
			var rvalue = buf.getAtIndex(GLTypeCache.ML_I(), 0);
            return rvalue;
        }
    }

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGenVertexArraysAPPLE">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGenVertexArraysAPPLE.xhtml">registry.khronos.org</a>
	 */
	public void glGenVertexArraysAPPLE(int n, MutPointer<Integer> arrays) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGenVertexArraysAPPLE(\{n}, \{arrays})");
		try {
		    functions.glGenVertexArraysAPPLE().invokeExact(n, arrays);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGenVertexArraysOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGenVertexArraysOES.xhtml">registry.khronos.org</a>
	 */
	public void glGenVertexArraysOES(int n, MutPointer<Integer> arrays) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGenVertexArraysOES(\{n}, \{arrays})");
		try {
		    functions.glGenVertexArraysOES().invokeExact(n, arrays);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGenVertexShadersEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGenVertexShadersEXT.xhtml">registry.khronos.org</a>
	 */
	public int glGenVertexShadersEXT(int range) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGenVertexShadersEXT(\{range})");
		try {
		    var r = (int)functions.glGenVertexShadersEXT().invokeExact(range);
			checkError();
			return r;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGenerateMipmap">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGenerateMipmap.xhtml">registry.khronos.org</a>
	 */
	public void glGenerateMipmap(GLTextureTarget target) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGenerateMipmap(\{target})");
		try {
		    functions.glGenerateMipmap().invokeExact(target);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGenerateMipmapEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGenerateMipmapEXT.xhtml">registry.khronos.org</a>
	 */
	public void glGenerateMipmapEXT(GLTextureTarget target) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGenerateMipmapEXT(\{target})");
		try {
		    functions.glGenerateMipmapEXT().invokeExact(target);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGenerateMipmapOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGenerateMipmapOES.xhtml">registry.khronos.org</a>
	 */
	public void glGenerateMipmapOES(GLTextureTarget target) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGenerateMipmapOES(\{target})");
		try {
		    functions.glGenerateMipmapOES().invokeExact(target);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGenerateMultiTexMipmapEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGenerateMultiTexMipmapEXT.xhtml">registry.khronos.org</a>
	 */
	public void glGenerateMultiTexMipmapEXT(GLTextureUnit texunit, GLTextureTarget target) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGenerateMultiTexMipmapEXT(\{texunit}, \{target})");
		try {
		    functions.glGenerateMultiTexMipmapEXT().invokeExact(texunit, target);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were texunit=\{texunit}, target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGenerateTextureMipmap">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGenerateTextureMipmap.xhtml">registry.khronos.org</a>
	 */
	public void glGenerateTextureMipmap(int texture) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGenerateTextureMipmap(\{texture})");
		try {
		    functions.glGenerateTextureMipmap().invokeExact(texture);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGenerateTextureMipmapEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGenerateTextureMipmapEXT.xhtml">registry.khronos.org</a>
	 */
	public void glGenerateTextureMipmapEXT(int texture, GLTextureTarget target) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGenerateTextureMipmapEXT(\{texture}, \{target})");
		try {
		    functions.glGenerateTextureMipmapEXT().invokeExact(texture, target);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetActiveAtomicCounterBufferiv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetActiveAtomicCounterBufferiv.xhtml">registry.khronos.org</a>
	 */
	public void glGetActiveAtomicCounterBufferiv(int program, int bufferIndex, GLAtomicCounterBufferPName pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetActiveAtomicCounterBufferiv(\{program}, \{bufferIndex}, \{pname}, \{params})");
		try {
		    functions.glGetActiveAtomicCounterBufferiv().invokeExact(program, bufferIndex, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetActiveAttrib">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetActiveAttrib.xhtml">registry.khronos.org</a>
	 */
	public void glGetActiveAttrib(int program, int index, int bufSize, MutPointer<Integer> length, MutPointer<Integer> size, MutPointer<GLAttributeType> type, MutPointer<Byte> name) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetActiveAttrib(\{program}, \{index}, \{bufSize}, \{length}, \{size}, \{type}, \{name})");
		try {
		    functions.glGetActiveAttrib().invokeExact(program, index, bufSize, length, size, type, name);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetActiveAttribARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetActiveAttribARB.xhtml">registry.khronos.org</a>
	 */
	public void glGetActiveAttribARB(int programObj, int index, int maxLength, MutPointer<Integer> length, MutPointer<Integer> size, MutPointer<GLAttributeType> type, MutPointer<Byte> name) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetActiveAttribARB(\{programObj}, \{index}, \{maxLength}, \{length}, \{size}, \{type}, \{name})");
		try {
		    functions.glGetActiveAttribARB().invokeExact(programObj, index, maxLength, length, size, type, name);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetActiveSubroutineName">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetActiveSubroutineName.xhtml">registry.khronos.org</a>
	 */
	public void glGetActiveSubroutineName(int program, GLShaderType shadertype, int index, int bufSize, MutPointer<Integer> length, MutPointer<Byte> name) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetActiveSubroutineName(\{program}, \{shadertype}, \{index}, \{bufSize}, \{length}, \{name})");
		try {
		    functions.glGetActiveSubroutineName().invokeExact(program, shadertype, index, bufSize, length, name);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were shadertype=\{shadertype}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetActiveSubroutineUniformName">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetActiveSubroutineUniformName.xhtml">registry.khronos.org</a>
	 */
	public void glGetActiveSubroutineUniformName(int program, GLShaderType shadertype, int index, int bufSize, MutPointer<Integer> length, MutPointer<Byte> name) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetActiveSubroutineUniformName(\{program}, \{shadertype}, \{index}, \{bufSize}, \{length}, \{name})");
		try {
		    functions.glGetActiveSubroutineUniformName().invokeExact(program, shadertype, index, bufSize, length, name);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were shadertype=\{shadertype}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetActiveSubroutineUniformiv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetActiveSubroutineUniformiv.xhtml">registry.khronos.org</a>
	 */
	public void glGetActiveSubroutineUniformiv(int program, GLShaderType shadertype, int index, GLSubroutineParameterName pname, MutPointer<Integer> values) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetActiveSubroutineUniformiv(\{program}, \{shadertype}, \{index}, \{pname}, \{values})");
		try {
		    functions.glGetActiveSubroutineUniformiv().invokeExact(program, shadertype, index, pname, values);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were shadertype=\{shadertype}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetActiveUniform">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetActiveUniform.xhtml">registry.khronos.org</a>
	 */
	public void glGetActiveUniform(int program, int index, int bufSize, MutPointer<Integer> length, MutPointer<Integer> size, MutPointer<GLUniformType> type, MutPointer<Byte> name) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetActiveUniform(\{program}, \{index}, \{bufSize}, \{length}, \{size}, \{type}, \{name})");
		try {
		    functions.glGetActiveUniform().invokeExact(program, index, bufSize, length, size, type, name);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetActiveUniformARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetActiveUniformARB.xhtml">registry.khronos.org</a>
	 */
	public void glGetActiveUniformARB(int programObj, int index, int maxLength, MutPointer<Integer> length, MutPointer<Integer> size, MutPointer<GLUniformType> type, MutPointer<Byte> name) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetActiveUniformARB(\{programObj}, \{index}, \{maxLength}, \{length}, \{size}, \{type}, \{name})");
		try {
		    functions.glGetActiveUniformARB().invokeExact(programObj, index, maxLength, length, size, type, name);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetActiveUniformBlockName">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetActiveUniformBlockName.xhtml">registry.khronos.org</a>
	 */
	public void glGetActiveUniformBlockName(int program, int uniformBlockIndex, int bufSize, MutPointer<Integer> length, MutPointer<Byte> uniformBlockName) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetActiveUniformBlockName(\{program}, \{uniformBlockIndex}, \{bufSize}, \{length}, \{uniformBlockName})");
		try {
		    functions.glGetActiveUniformBlockName().invokeExact(program, uniformBlockIndex, bufSize, length, uniformBlockName);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetActiveUniformBlockiv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetActiveUniformBlockiv.xhtml">registry.khronos.org</a>
	 */
	public void glGetActiveUniformBlockiv(int program, int uniformBlockIndex, GLUniformBlockPName pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetActiveUniformBlockiv(\{program}, \{uniformBlockIndex}, \{pname}, \{params})");
		try {
		    functions.glGetActiveUniformBlockiv().invokeExact(program, uniformBlockIndex, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetActiveUniformName">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetActiveUniformName.xhtml">registry.khronos.org</a>
	 */
	public void glGetActiveUniformName(int program, int uniformIndex, int bufSize, MutPointer<Integer> length, MutPointer<Byte> uniformName) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetActiveUniformName(\{program}, \{uniformIndex}, \{bufSize}, \{length}, \{uniformName})");
		try {
		    functions.glGetActiveUniformName().invokeExact(program, uniformIndex, bufSize, length, uniformName);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetActiveUniformsiv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetActiveUniformsiv.xhtml">registry.khronos.org</a>
	 */
	public void glGetActiveUniformsiv(int program, int uniformCount, ConstPointer<Integer> uniformIndices, GLUniformPName pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetActiveUniformsiv(\{program}, \{uniformCount}, \{uniformIndices}, \{pname}, \{params})");
		try {
		    functions.glGetActiveUniformsiv().invokeExact(program, uniformCount, uniformIndices, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetActiveVaryingNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetActiveVaryingNV.xhtml">registry.khronos.org</a>
	 */
	public void glGetActiveVaryingNV(int program, int index, int bufSize, MutPointer<Integer> length, MutPointer<Integer> size, MutPointer<GLEnum> type, MutPointer<Byte> name) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetActiveVaryingNV(\{program}, \{index}, \{bufSize}, \{length}, \{size}, \{type}, \{name})");
		try {
		    functions.glGetActiveVaryingNV().invokeExact(program, index, bufSize, length, size, type, name);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetArrayObjectfvATI">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetArrayObjectfvATI.xhtml">registry.khronos.org</a>
	 */
	public void glGetArrayObjectfvATI(GLEnableCap array, GLArrayObjectPNameATI pname, MutPointer<Float> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetArrayObjectfvATI(\{array}, \{pname}, \{params})");
		try {
		    functions.glGetArrayObjectfvATI().invokeExact(array, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were array=\{array}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetArrayObjectivATI">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetArrayObjectivATI.xhtml">registry.khronos.org</a>
	 */
	public void glGetArrayObjectivATI(GLEnableCap array, GLArrayObjectPNameATI pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetArrayObjectivATI(\{array}, \{pname}, \{params})");
		try {
		    functions.glGetArrayObjectivATI().invokeExact(array, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were array=\{array}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 * Could be an alias of glGetAttachedShaders except that GLhandleARB is different on MacOS X
	 *
	 * @see <a href="https://docs.gl/gl4/glGetAttachedObjectsARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetAttachedObjectsARB.xhtml">registry.khronos.org</a>
	 */
	public void glGetAttachedObjectsARB(int containerObj, int maxCount, MutPointer<Integer> count, MutPointer<Integer> obj) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetAttachedObjectsARB(\{containerObj}, \{maxCount}, \{count}, \{obj})");
		try {
		    functions.glGetAttachedObjectsARB().invokeExact(containerObj, maxCount, count, obj);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetAttachedShaders">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetAttachedShaders.xhtml">registry.khronos.org</a>
	 */
	public void glGetAttachedShaders(int program, int maxCount, MutPointer<Integer> count, MutPointer<Integer> shaders) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetAttachedShaders(\{program}, \{maxCount}, \{count}, \{shaders})");
		try {
		    functions.glGetAttachedShaders().invokeExact(program, maxCount, count, shaders);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetAttribLocation">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetAttribLocation.xhtml">registry.khronos.org</a>
	 */
	public int glGetAttribLocation(int program, ConstPointer<Byte> name) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetAttribLocation(\{program}, \{name})");
		try {
		    var r = (int)functions.glGetAttribLocation().invokeExact(program, name);
			checkError();
			return r;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetAttribLocationARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetAttribLocationARB.xhtml">registry.khronos.org</a>
	 */
	public int glGetAttribLocationARB(int programObj, ConstPointer<Byte> name) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetAttribLocationARB(\{programObj}, \{name})");
		try {
		    var r = (int)functions.glGetAttribLocationARB().invokeExact(programObj, name);
			checkError();
			return r;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetBooleanIndexedvEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetBooleanIndexedvEXT.xhtml">registry.khronos.org</a>
	 */
	public void glGetBooleanIndexedvEXT(GLBufferTarget target, int index, MutPointer<Boolean> data) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetBooleanIndexedvEXT(\{target}, \{index}, \{data})");
		try {
		    functions.glGetBooleanIndexedvEXT().invokeExact(target, index, data);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetBooleani_v">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetBooleani_v.xhtml">registry.khronos.org</a>
	 */
	public void glGetBooleani_v(GLBufferTarget target, int index, MutPointer<Boolean> data) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetBooleani_v(\{target}, \{index}, \{data})");
		try {
		    functions.glGetBooleani_v().invokeExact(target, index, data);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetBooleanv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetBooleanv.xhtml">registry.khronos.org</a>
	 */
	public void glGetBooleanv(GLGetPName pname, MutPointer<Boolean> data) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetBooleanv(\{pname}, \{data})");
		try {
		    functions.glGetBooleanv().invokeExact(pname, data);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetBufferParameteri64v">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetBufferParameteri64v.xhtml">registry.khronos.org</a>
	 */
	public void glGetBufferParameteri64v(GLBufferTarget target, GLBufferPNameARB pname, MutPointer<Long> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetBufferParameteri64v(\{target}, \{pname}, \{params})");
		try {
		    functions.glGetBufferParameteri64v().invokeExact(target, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetBufferParameteriv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetBufferParameteriv.xhtml">registry.khronos.org</a>
	 */
	public void glGetBufferParameteriv(GLBufferTarget target, GLBufferPNameARB pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetBufferParameteriv(\{target}, \{pname}, \{params})");
		try {
		    functions.glGetBufferParameteriv().invokeExact(target, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetBufferParameterivARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetBufferParameterivARB.xhtml">registry.khronos.org</a>
	 */
	public void glGetBufferParameterivARB(GLBufferTarget target, GLBufferPNameARB pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetBufferParameterivARB(\{target}, \{pname}, \{params})");
		try {
		    functions.glGetBufferParameterivARB().invokeExact(target, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetBufferParameterui64vNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetBufferParameterui64vNV.xhtml">registry.khronos.org</a>
	 */
	public void glGetBufferParameterui64vNV(GLBufferTarget target, GLEnum pname, MutPointer<Long> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetBufferParameterui64vNV(\{target}, \{pname}, \{params})");
		try {
		    functions.glGetBufferParameterui64vNV().invokeExact(target, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetBufferPointerv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetBufferPointerv.xhtml">registry.khronos.org</a>
	 */
	public void glGetBufferPointerv(GLBufferTarget target, GLBufferPointerNameARB pname, MutPointer<MutPointer<Void>> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetBufferPointerv(\{target}, \{pname}, \{params})");
		try {
		    functions.glGetBufferPointerv().invokeExact(target, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetBufferPointervARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetBufferPointervARB.xhtml">registry.khronos.org</a>
	 */
	public void glGetBufferPointervARB(GLBufferTarget target, GLBufferPointerNameARB pname, MutPointer<MutPointer<Void>> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetBufferPointervARB(\{target}, \{pname}, \{params})");
		try {
		    functions.glGetBufferPointervARB().invokeExact(target, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetBufferPointervOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetBufferPointervOES.xhtml">registry.khronos.org</a>
	 */
	public void glGetBufferPointervOES(GLBufferTarget target, GLBufferPointerNameARB pname, MutPointer<MutPointer<Void>> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetBufferPointervOES(\{target}, \{pname}, \{params})");
		try {
		    functions.glGetBufferPointervOES().invokeExact(target, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetBufferSubData">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetBufferSubData.xhtml">registry.khronos.org</a>
	 */
	public void glGetBufferSubData(GLBufferTarget target, long offset, long size, MutPointer<Void> data) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetBufferSubData(\{target}, \{offset}, \{size}, \{data})");
		try {
		    functions.glGetBufferSubData().invokeExact(target, offset, size, data);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetBufferSubDataARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetBufferSubDataARB.xhtml">registry.khronos.org</a>
	 */
	public void glGetBufferSubDataARB(GLBufferTarget target, long offset, long size, MutPointer<Void> data) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetBufferSubDataARB(\{target}, \{offset}, \{size}, \{data})");
		try {
		    functions.glGetBufferSubDataARB().invokeExact(target, offset, size, data);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetClipPlane">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetClipPlane.xhtml">registry.khronos.org</a>
	 */
	public void glGetClipPlane(GLClipPlaneName plane, MutPointer<Double> equation) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetClipPlane(\{plane}, \{equation})");
		try {
		    functions.glGetClipPlane().invokeExact(plane, equation);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were plane=\{plane}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetClipPlanef">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetClipPlanef.xhtml">registry.khronos.org</a>
	 */
	public void glGetClipPlanef(GLClipPlaneName plane, MutPointer<Float> equation) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetClipPlanef(\{plane}, \{equation})");
		try {
		    functions.glGetClipPlanef().invokeExact(plane, equation);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were plane=\{plane}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetClipPlanefOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetClipPlanefOES.xhtml">registry.khronos.org</a>
	 */
	public void glGetClipPlanefOES(GLClipPlaneName plane, MutPointer<Float> equation) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetClipPlanefOES(\{plane}, \{equation})");
		try {
		    functions.glGetClipPlanefOES().invokeExact(plane, equation);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were plane=\{plane}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetClipPlanex">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetClipPlanex.xhtml">registry.khronos.org</a>
	 */
	public void glGetClipPlanex(GLClipPlaneName plane, MutPointer<Integer> equation) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetClipPlanex(\{plane}, \{equation})");
		try {
		    functions.glGetClipPlanex().invokeExact(plane, equation);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were plane=\{plane}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetClipPlanexOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetClipPlanexOES.xhtml">registry.khronos.org</a>
	 */
	public void glGetClipPlanexOES(GLClipPlaneName plane, MutPointer<Integer> equation) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetClipPlanexOES(\{plane}, \{equation})");
		try {
		    functions.glGetClipPlanexOES().invokeExact(plane, equation);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were plane=\{plane}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetColorTable">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetColorTable.xhtml">registry.khronos.org</a>
	 */
	public void glGetColorTable(GLColorTableTarget target, GLPixelFormat format, GLPixelType type, MutPointer<Void> table) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetColorTable(\{target}, \{format}, \{type}, \{table})");
		try {
		    functions.glGetColorTable().invokeExact(target, format, type, table);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, format=\{format}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetColorTableEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetColorTableEXT.xhtml">registry.khronos.org</a>
	 */
	public void glGetColorTableEXT(GLColorTableTarget target, GLPixelFormat format, GLPixelType type, MutPointer<Void> data) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetColorTableEXT(\{target}, \{format}, \{type}, \{data})");
		try {
		    functions.glGetColorTableEXT().invokeExact(target, format, type, data);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, format=\{format}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetColorTableParameterfv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetColorTableParameterfv.xhtml">registry.khronos.org</a>
	 */
	public void glGetColorTableParameterfv(GLColorTableTarget target, GLColorTableParameterPName pname, MutPointer<Float> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetColorTableParameterfv(\{target}, \{pname}, \{params})");
		try {
		    functions.glGetColorTableParameterfv().invokeExact(target, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetColorTableParameterfvEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetColorTableParameterfvEXT.xhtml">registry.khronos.org</a>
	 */
	public void glGetColorTableParameterfvEXT(GLColorTableTarget target, GLColorTableParameterPName pname, MutPointer<Float> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetColorTableParameterfvEXT(\{target}, \{pname}, \{params})");
		try {
		    functions.glGetColorTableParameterfvEXT().invokeExact(target, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetColorTableParameterfvSGI">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetColorTableParameterfvSGI.xhtml">registry.khronos.org</a>
	 */
	public void glGetColorTableParameterfvSGI(GLColorTableTargetSGI target, GLColorTableParameterPName pname, MutPointer<Float> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetColorTableParameterfvSGI(\{target}, \{pname}, \{params})");
		try {
		    functions.glGetColorTableParameterfvSGI().invokeExact(target, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetColorTableParameteriv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetColorTableParameteriv.xhtml">registry.khronos.org</a>
	 */
	public void glGetColorTableParameteriv(GLColorTableTarget target, GLColorTableParameterPName pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetColorTableParameteriv(\{target}, \{pname}, \{params})");
		try {
		    functions.glGetColorTableParameteriv().invokeExact(target, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetColorTableParameterivEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetColorTableParameterivEXT.xhtml">registry.khronos.org</a>
	 */
	public void glGetColorTableParameterivEXT(GLColorTableTarget target, GLColorTableParameterPName pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetColorTableParameterivEXT(\{target}, \{pname}, \{params})");
		try {
		    functions.glGetColorTableParameterivEXT().invokeExact(target, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetColorTableParameterivSGI">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetColorTableParameterivSGI.xhtml">registry.khronos.org</a>
	 */
	public void glGetColorTableParameterivSGI(GLColorTableTargetSGI target, GLColorTableParameterPName pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetColorTableParameterivSGI(\{target}, \{pname}, \{params})");
		try {
		    functions.glGetColorTableParameterivSGI().invokeExact(target, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetColorTableSGI">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetColorTableSGI.xhtml">registry.khronos.org</a>
	 */
	public void glGetColorTableSGI(GLColorTableTargetSGI target, GLPixelFormat format, GLPixelType type, MutPointer<Void> table) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetColorTableSGI(\{target}, \{format}, \{type}, \{table})");
		try {
		    functions.glGetColorTableSGI().invokeExact(target, format, type, table);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, format=\{format}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetCombinerInputParameterfvNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetCombinerInputParameterfvNV.xhtml">registry.khronos.org</a>
	 */
	public void glGetCombinerInputParameterfvNV(GLCombinerStageNV stage, GLCombinerPortionNV portion, GLCombinerVariableNV variable, GLCombinerParameterNV pname, MutPointer<Float> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetCombinerInputParameterfvNV(\{stage}, \{portion}, \{variable}, \{pname}, \{params})");
		try {
		    functions.glGetCombinerInputParameterfvNV().invokeExact(stage, portion, variable, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were stage=\{stage}, portion=\{portion}, variable=\{variable}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetCombinerInputParameterivNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetCombinerInputParameterivNV.xhtml">registry.khronos.org</a>
	 */
	public void glGetCombinerInputParameterivNV(GLCombinerStageNV stage, GLCombinerPortionNV portion, GLCombinerVariableNV variable, GLCombinerParameterNV pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetCombinerInputParameterivNV(\{stage}, \{portion}, \{variable}, \{pname}, \{params})");
		try {
		    functions.glGetCombinerInputParameterivNV().invokeExact(stage, portion, variable, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were stage=\{stage}, portion=\{portion}, variable=\{variable}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetCombinerOutputParameterfvNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetCombinerOutputParameterfvNV.xhtml">registry.khronos.org</a>
	 */
	public void glGetCombinerOutputParameterfvNV(GLCombinerStageNV stage, GLCombinerPortionNV portion, GLCombinerParameterNV pname, MutPointer<Float> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetCombinerOutputParameterfvNV(\{stage}, \{portion}, \{pname}, \{params})");
		try {
		    functions.glGetCombinerOutputParameterfvNV().invokeExact(stage, portion, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were stage=\{stage}, portion=\{portion}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetCombinerOutputParameterivNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetCombinerOutputParameterivNV.xhtml">registry.khronos.org</a>
	 */
	public void glGetCombinerOutputParameterivNV(GLCombinerStageNV stage, GLCombinerPortionNV portion, GLCombinerParameterNV pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetCombinerOutputParameterivNV(\{stage}, \{portion}, \{pname}, \{params})");
		try {
		    functions.glGetCombinerOutputParameterivNV().invokeExact(stage, portion, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were stage=\{stage}, portion=\{portion}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetCombinerStageParameterfvNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetCombinerStageParameterfvNV.xhtml">registry.khronos.org</a>
	 */
	public void glGetCombinerStageParameterfvNV(GLCombinerStageNV stage, GLCombinerParameterNV pname, MutPointer<Float> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetCombinerStageParameterfvNV(\{stage}, \{pname}, \{params})");
		try {
		    functions.glGetCombinerStageParameterfvNV().invokeExact(stage, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were stage=\{stage}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetCommandHeaderNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetCommandHeaderNV.xhtml">registry.khronos.org</a>
	 */
	public int glGetCommandHeaderNV(GLCommandOpcodesNV tokenID, int size) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetCommandHeaderNV(\{tokenID}, \{size})");
		try {
		    var r = (int)functions.glGetCommandHeaderNV().invokeExact(tokenID, size);
			checkError();
			return r;
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were tokenID=\{tokenID}.");
			}
			throw e;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetCompressedMultiTexImageEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetCompressedMultiTexImageEXT.xhtml">registry.khronos.org</a>
	 */
	public void glGetCompressedMultiTexImageEXT(GLTextureUnit texunit, GLTextureTarget target, int lod, MutPointer<Void> img) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetCompressedMultiTexImageEXT(\{texunit}, \{target}, \{lod}, \{img})");
		try {
		    functions.glGetCompressedMultiTexImageEXT().invokeExact(texunit, target, lod, img);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were texunit=\{texunit}, target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetCompressedTexImage">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetCompressedTexImage.xhtml">registry.khronos.org</a>
	 */
	public void glGetCompressedTexImage(GLTextureTarget target, int level, MutPointer<Void> img) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetCompressedTexImage(\{target}, \{level}, \{img})");
		try {
		    functions.glGetCompressedTexImage().invokeExact(target, level, img);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetCompressedTexImageARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetCompressedTexImageARB.xhtml">registry.khronos.org</a>
	 */
	public void glGetCompressedTexImageARB(GLTextureTarget target, int level, MutPointer<Void> img) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetCompressedTexImageARB(\{target}, \{level}, \{img})");
		try {
		    functions.glGetCompressedTexImageARB().invokeExact(target, level, img);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetCompressedTextureImage">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetCompressedTextureImage.xhtml">registry.khronos.org</a>
	 */
	public void glGetCompressedTextureImage(int texture, int level, int bufSize, MutPointer<Void> pixels) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetCompressedTextureImage(\{texture}, \{level}, \{bufSize}, \{pixels})");
		try {
		    functions.glGetCompressedTextureImage().invokeExact(texture, level, bufSize, pixels);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetCompressedTextureImageEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetCompressedTextureImageEXT.xhtml">registry.khronos.org</a>
	 */
	public void glGetCompressedTextureImageEXT(int texture, GLTextureTarget target, int lod, MutPointer<Void> img) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetCompressedTextureImageEXT(\{texture}, \{target}, \{lod}, \{img})");
		try {
		    functions.glGetCompressedTextureImageEXT().invokeExact(texture, target, lod, img);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetCompressedTextureSubImage">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetCompressedTextureSubImage.xhtml">registry.khronos.org</a>
	 */
	public void glGetCompressedTextureSubImage(int texture, int level, int xoffset, int yoffset, int zoffset, int width, int height, int depth, int bufSize, MutPointer<Void> pixels) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetCompressedTextureSubImage(\{texture}, \{level}, \{xoffset}, \{yoffset}, \{zoffset}, \{width}, \{height}, \{depth}, \{bufSize}, \{pixels})");
		try {
		    functions.glGetCompressedTextureSubImage().invokeExact(texture, level, xoffset, yoffset, zoffset, width, height, depth, bufSize, pixels);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetConvolutionFilter">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetConvolutionFilter.xhtml">registry.khronos.org</a>
	 */
	public void glGetConvolutionFilter(GLConvolutionTarget target, GLPixelFormat format, GLPixelType type, MutPointer<Void> image) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetConvolutionFilter(\{target}, \{format}, \{type}, \{image})");
		try {
		    functions.glGetConvolutionFilter().invokeExact(target, format, type, image);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, format=\{format}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetConvolutionFilterEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetConvolutionFilterEXT.xhtml">registry.khronos.org</a>
	 */
	public void glGetConvolutionFilterEXT(GLConvolutionTargetEXT target, GLPixelFormat format, GLPixelType type, MutPointer<Void> image) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetConvolutionFilterEXT(\{target}, \{format}, \{type}, \{image})");
		try {
		    functions.glGetConvolutionFilterEXT().invokeExact(target, format, type, image);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, format=\{format}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetConvolutionParameterfv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetConvolutionParameterfv.xhtml">registry.khronos.org</a>
	 */
	public void glGetConvolutionParameterfv(GLConvolutionTarget target, GLConvolutionParameter pname, MutPointer<Float> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetConvolutionParameterfv(\{target}, \{pname}, \{params})");
		try {
		    functions.glGetConvolutionParameterfv().invokeExact(target, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetConvolutionParameterfvEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetConvolutionParameterfvEXT.xhtml">registry.khronos.org</a>
	 */
	public void glGetConvolutionParameterfvEXT(GLConvolutionTargetEXT target, GLConvolutionParameter pname, MutPointer<Float> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetConvolutionParameterfvEXT(\{target}, \{pname}, \{params})");
		try {
		    functions.glGetConvolutionParameterfvEXT().invokeExact(target, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetConvolutionParameteriv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetConvolutionParameteriv.xhtml">registry.khronos.org</a>
	 */
	public void glGetConvolutionParameteriv(GLConvolutionTarget target, GLConvolutionParameter pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetConvolutionParameteriv(\{target}, \{pname}, \{params})");
		try {
		    functions.glGetConvolutionParameteriv().invokeExact(target, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetConvolutionParameterivEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetConvolutionParameterivEXT.xhtml">registry.khronos.org</a>
	 */
	public void glGetConvolutionParameterivEXT(GLConvolutionTargetEXT target, GLConvolutionParameter pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetConvolutionParameterivEXT(\{target}, \{pname}, \{params})");
		try {
		    functions.glGetConvolutionParameterivEXT().invokeExact(target, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetConvolutionParameterxvOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetConvolutionParameterxvOES.xhtml">registry.khronos.org</a>
	 */
	public void glGetConvolutionParameterxvOES(GLEnum target, GLEnum pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetConvolutionParameterxvOES(\{target}, \{pname}, \{params})");
		try {
		    functions.glGetConvolutionParameterxvOES().invokeExact(target, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetCoverageModulationTableNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetCoverageModulationTableNV.xhtml">registry.khronos.org</a>
	 */
	public void glGetCoverageModulationTableNV(int bufSize, MutPointer<Float> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetCoverageModulationTableNV(\{bufSize}, \{v})");
		try {
		    functions.glGetCoverageModulationTableNV().invokeExact(bufSize, v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetDebugMessageLog">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetDebugMessageLog.xhtml">registry.khronos.org</a>
	 */
	public int glGetDebugMessageLog(int count, int bufSize, MutPointer<GLDebugSource> sources, MutPointer<GLDebugType> types, MutPointer<Integer> ids, MutPointer<GLDebugSeverity> severities, MutPointer<Integer> lengths, MutPointer<Byte> messageLog) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetDebugMessageLog(\{count}, \{bufSize}, \{sources}, \{types}, \{ids}, \{severities}, \{lengths}, \{messageLog})");
		try {
		    var r = (int)functions.glGetDebugMessageLog().invokeExact(count, bufSize, sources, types, ids, severities, lengths, messageLog);
			checkError();
			return r;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetDebugMessageLogAMD">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetDebugMessageLogAMD.xhtml">registry.khronos.org</a>
	 */
	public int glGetDebugMessageLogAMD(int count, int bufSize, MutPointer<GLEnum> categories, MutPointer<GLDebugSeverity> severities, MutPointer<Integer> ids, MutPointer<Integer> lengths, MutPointer<Byte> message) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetDebugMessageLogAMD(\{count}, \{bufSize}, \{categories}, \{severities}, \{ids}, \{lengths}, \{message})");
		try {
		    var r = (int)functions.glGetDebugMessageLogAMD().invokeExact(count, bufSize, categories, severities, ids, lengths, message);
			checkError();
			return r;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetDebugMessageLogARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetDebugMessageLogARB.xhtml">registry.khronos.org</a>
	 */
	public int glGetDebugMessageLogARB(int count, int bufSize, MutPointer<GLDebugSource> sources, MutPointer<GLDebugType> types, MutPointer<Integer> ids, MutPointer<GLDebugSeverity> severities, MutPointer<Integer> lengths, MutPointer<Byte> messageLog) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetDebugMessageLogARB(\{count}, \{bufSize}, \{sources}, \{types}, \{ids}, \{severities}, \{lengths}, \{messageLog})");
		try {
		    var r = (int)functions.glGetDebugMessageLogARB().invokeExact(count, bufSize, sources, types, ids, severities, lengths, messageLog);
			checkError();
			return r;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetDebugMessageLogKHR">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetDebugMessageLogKHR.xhtml">registry.khronos.org</a>
	 */
	public int glGetDebugMessageLogKHR(int count, int bufSize, MutPointer<GLDebugSource> sources, MutPointer<GLDebugType> types, MutPointer<Integer> ids, MutPointer<GLDebugSeverity> severities, MutPointer<Integer> lengths, MutPointer<Byte> messageLog) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetDebugMessageLogKHR(\{count}, \{bufSize}, \{sources}, \{types}, \{ids}, \{severities}, \{lengths}, \{messageLog})");
		try {
		    var r = (int)functions.glGetDebugMessageLogKHR().invokeExact(count, bufSize, sources, types, ids, severities, lengths, messageLog);
			checkError();
			return r;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetDetailTexFuncSGIS">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetDetailTexFuncSGIS.xhtml">registry.khronos.org</a>
	 */
	public void glGetDetailTexFuncSGIS(GLTextureTarget target, MutPointer<Float> points) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetDetailTexFuncSGIS(\{target}, \{points})");
		try {
		    functions.glGetDetailTexFuncSGIS().invokeExact(target, points);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetDoubleIndexedvEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetDoubleIndexedvEXT.xhtml">registry.khronos.org</a>
	 */
	public void glGetDoubleIndexedvEXT(GLGetPName target, int index, MutPointer<Double> data) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetDoubleIndexedvEXT(\{target}, \{index}, \{data})");
		try {
		    functions.glGetDoubleIndexedvEXT().invokeExact(target, index, data);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetDoublei_v">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetDoublei_v.xhtml">registry.khronos.org</a>
	 */
	public void glGetDoublei_v(GLGetPName target, int index, MutPointer<Double> data) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetDoublei_v(\{target}, \{index}, \{data})");
		try {
		    functions.glGetDoublei_v().invokeExact(target, index, data);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetDoublei_vEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetDoublei_vEXT.xhtml">registry.khronos.org</a>
	 */
	public void glGetDoublei_vEXT(GLGetPName pname, int index, MutPointer<Double> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetDoublei_vEXT(\{pname}, \{index}, \{params})");
		try {
		    functions.glGetDoublei_vEXT().invokeExact(pname, index, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetDoublev">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetDoublev.xhtml">registry.khronos.org</a>
	 */
	public void glGetDoublev(GLGetPName pname, MutPointer<Double> data) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetDoublev(\{pname}, \{data})");
		try {
		    functions.glGetDoublev().invokeExact(pname, data);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetDriverControlStringQCOM">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetDriverControlStringQCOM.xhtml">registry.khronos.org</a>
	 */
	public void glGetDriverControlStringQCOM(int driverControl, int bufSize, MutPointer<Integer> length, MutPointer<Byte> driverControlString) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetDriverControlStringQCOM(\{driverControl}, \{bufSize}, \{length}, \{driverControlString})");
		try {
		    functions.glGetDriverControlStringQCOM().invokeExact(driverControl, bufSize, length, driverControlString);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetDriverControlsQCOM">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetDriverControlsQCOM.xhtml">registry.khronos.org</a>
	 */
	public void glGetDriverControlsQCOM(MutPointer<Integer> num, int size, MutPointer<Integer> driverControls) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetDriverControlsQCOM(\{num}, \{size}, \{driverControls})");
		try {
		    functions.glGetDriverControlsQCOM().invokeExact(num, size, driverControls);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetError">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetError.xhtml">registry.khronos.org</a>
	 */
	public GLErrorCode glGetError() {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetError()");
		try {
		    var r = (GLErrorCode)functions.glGetError().invokeExact();
			return r;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetFenceivNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetFenceivNV.xhtml">registry.khronos.org</a>
	 */
	public void glGetFenceivNV(int fence, GLFenceParameterNameNV pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetFenceivNV(\{fence}, \{pname}, \{params})");
		try {
		    functions.glGetFenceivNV().invokeExact(fence, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetFinalCombinerInputParameterfvNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetFinalCombinerInputParameterfvNV.xhtml">registry.khronos.org</a>
	 */
	public void glGetFinalCombinerInputParameterfvNV(GLCombinerVariableNV variable, GLCombinerParameterNV pname, MutPointer<Float> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetFinalCombinerInputParameterfvNV(\{variable}, \{pname}, \{params})");
		try {
		    functions.glGetFinalCombinerInputParameterfvNV().invokeExact(variable, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were variable=\{variable}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetFinalCombinerInputParameterivNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetFinalCombinerInputParameterivNV.xhtml">registry.khronos.org</a>
	 */
	public void glGetFinalCombinerInputParameterivNV(GLCombinerVariableNV variable, GLCombinerParameterNV pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetFinalCombinerInputParameterivNV(\{variable}, \{pname}, \{params})");
		try {
		    functions.glGetFinalCombinerInputParameterivNV().invokeExact(variable, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were variable=\{variable}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetFirstPerfQueryIdINTEL">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetFirstPerfQueryIdINTEL.xhtml">registry.khronos.org</a>
	 */
	public void glGetFirstPerfQueryIdINTEL(MutPointer<Integer> queryId) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetFirstPerfQueryIdINTEL(\{queryId})");
		try {
		    functions.glGetFirstPerfQueryIdINTEL().invokeExact(queryId);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetFixedv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetFixedv.xhtml">registry.khronos.org</a>
	 */
	public void glGetFixedv(GLGetPName pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetFixedv(\{pname}, \{params})");
		try {
		    functions.glGetFixedv().invokeExact(pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetFixedvOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetFixedvOES.xhtml">registry.khronos.org</a>
	 */
	public void glGetFixedvOES(GLGetPName pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetFixedvOES(\{pname}, \{params})");
		try {
		    functions.glGetFixedvOES().invokeExact(pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetFloatIndexedvEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetFloatIndexedvEXT.xhtml">registry.khronos.org</a>
	 */
	public void glGetFloatIndexedvEXT(GLGetPName target, int index, MutPointer<Float> data) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetFloatIndexedvEXT(\{target}, \{index}, \{data})");
		try {
		    functions.glGetFloatIndexedvEXT().invokeExact(target, index, data);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetFloati_v">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetFloati_v.xhtml">registry.khronos.org</a>
	 */
	public void glGetFloati_v(GLGetPName target, int index, MutPointer<Float> data) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetFloati_v(\{target}, \{index}, \{data})");
		try {
		    functions.glGetFloati_v().invokeExact(target, index, data);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetFloati_vEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetFloati_vEXT.xhtml">registry.khronos.org</a>
	 */
	public void glGetFloati_vEXT(GLGetPName pname, int index, MutPointer<Float> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetFloati_vEXT(\{pname}, \{index}, \{params})");
		try {
		    functions.glGetFloati_vEXT().invokeExact(pname, index, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetFloati_vNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetFloati_vNV.xhtml">registry.khronos.org</a>
	 */
	public void glGetFloati_vNV(GLGetPName target, int index, MutPointer<Float> data) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetFloati_vNV(\{target}, \{index}, \{data})");
		try {
		    functions.glGetFloati_vNV().invokeExact(target, index, data);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetFloati_vOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetFloati_vOES.xhtml">registry.khronos.org</a>
	 */
	public void glGetFloati_vOES(GLGetPName target, int index, MutPointer<Float> data) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetFloati_vOES(\{target}, \{index}, \{data})");
		try {
		    functions.glGetFloati_vOES().invokeExact(target, index, data);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetFloatv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetFloatv.xhtml">registry.khronos.org</a>
	 */
	public void glGetFloatv(GLGetPName pname, MutPointer<Float> data) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetFloatv(\{pname}, \{data})");
		try {
		    functions.glGetFloatv().invokeExact(pname, data);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetFogFuncSGIS">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetFogFuncSGIS.xhtml">registry.khronos.org</a>
	 */
	public void glGetFogFuncSGIS(MutPointer<Float> points) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetFogFuncSGIS(\{points})");
		try {
		    functions.glGetFogFuncSGIS().invokeExact(points);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetFragDataIndex">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetFragDataIndex.xhtml">registry.khronos.org</a>
	 */
	public int glGetFragDataIndex(int program, ConstPointer<Byte> name) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetFragDataIndex(\{program}, \{name})");
		try {
		    var r = (int)functions.glGetFragDataIndex().invokeExact(program, name);
			checkError();
			return r;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetFragDataIndexEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetFragDataIndexEXT.xhtml">registry.khronos.org</a>
	 */
	public int glGetFragDataIndexEXT(int program, ConstPointer<Byte> name) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetFragDataIndexEXT(\{program}, \{name})");
		try {
		    var r = (int)functions.glGetFragDataIndexEXT().invokeExact(program, name);
			checkError();
			return r;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetFragDataLocation">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetFragDataLocation.xhtml">registry.khronos.org</a>
	 */
	public int glGetFragDataLocation(int program, ConstPointer<Byte> name) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetFragDataLocation(\{program}, \{name})");
		try {
		    var r = (int)functions.glGetFragDataLocation().invokeExact(program, name);
			checkError();
			return r;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetFragDataLocationEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetFragDataLocationEXT.xhtml">registry.khronos.org</a>
	 */
	public int glGetFragDataLocationEXT(int program, ConstPointer<Byte> name) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetFragDataLocationEXT(\{program}, \{name})");
		try {
		    var r = (int)functions.glGetFragDataLocationEXT().invokeExact(program, name);
			checkError();
			return r;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetFragmentLightfvSGIX">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetFragmentLightfvSGIX.xhtml">registry.khronos.org</a>
	 */
	public void glGetFragmentLightfvSGIX(GLFragmentLightNameSGIX light, GLFragmentLightParameterSGIX pname, MutPointer<Float> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetFragmentLightfvSGIX(\{light}, \{pname}, \{params})");
		try {
		    functions.glGetFragmentLightfvSGIX().invokeExact(light, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were light=\{light}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetFragmentLightivSGIX">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetFragmentLightivSGIX.xhtml">registry.khronos.org</a>
	 */
	public void glGetFragmentLightivSGIX(GLFragmentLightNameSGIX light, GLFragmentLightParameterSGIX pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetFragmentLightivSGIX(\{light}, \{pname}, \{params})");
		try {
		    functions.glGetFragmentLightivSGIX().invokeExact(light, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were light=\{light}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetFragmentMaterialfvSGIX">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetFragmentMaterialfvSGIX.xhtml">registry.khronos.org</a>
	 */
	public void glGetFragmentMaterialfvSGIX(GLTriangleFace face, GLMaterialParameter pname, MutPointer<Float> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetFragmentMaterialfvSGIX(\{face}, \{pname}, \{params})");
		try {
		    functions.glGetFragmentMaterialfvSGIX().invokeExact(face, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were face=\{face}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetFragmentMaterialivSGIX">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetFragmentMaterialivSGIX.xhtml">registry.khronos.org</a>
	 */
	public void glGetFragmentMaterialivSGIX(GLTriangleFace face, GLMaterialParameter pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetFragmentMaterialivSGIX(\{face}, \{pname}, \{params})");
		try {
		    functions.glGetFragmentMaterialivSGIX().invokeExact(face, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were face=\{face}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetFragmentShadingRatesEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetFragmentShadingRatesEXT.xhtml">registry.khronos.org</a>
	 */
	public void glGetFragmentShadingRatesEXT(int samples, int maxCount, MutPointer<Integer> count, MutPointer<GLShadingRate> shadingRates) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetFragmentShadingRatesEXT(\{samples}, \{maxCount}, \{count}, \{shadingRates})");
		try {
		    functions.glGetFragmentShadingRatesEXT().invokeExact(samples, maxCount, count, shadingRates);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetFramebufferAttachmentParameteriv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetFramebufferAttachmentParameteriv.xhtml">registry.khronos.org</a>
	 */
	public void glGetFramebufferAttachmentParameteriv(GLFramebufferTarget target, GLFramebufferAttachment attachment, GLFramebufferAttachmentParameterName pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetFramebufferAttachmentParameteriv(\{target}, \{attachment}, \{pname}, \{params})");
		try {
		    functions.glGetFramebufferAttachmentParameteriv().invokeExact(target, attachment, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, attachment=\{attachment}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetFramebufferAttachmentParameterivEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetFramebufferAttachmentParameterivEXT.xhtml">registry.khronos.org</a>
	 */
	public void glGetFramebufferAttachmentParameterivEXT(GLFramebufferTarget target, GLFramebufferAttachment attachment, GLFramebufferAttachmentParameterName pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetFramebufferAttachmentParameterivEXT(\{target}, \{attachment}, \{pname}, \{params})");
		try {
		    functions.glGetFramebufferAttachmentParameterivEXT().invokeExact(target, attachment, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, attachment=\{attachment}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetFramebufferAttachmentParameterivOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetFramebufferAttachmentParameterivOES.xhtml">registry.khronos.org</a>
	 */
	public void glGetFramebufferAttachmentParameterivOES(GLFramebufferTarget target, GLFramebufferAttachment attachment, GLFramebufferAttachmentParameterName pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetFramebufferAttachmentParameterivOES(\{target}, \{attachment}, \{pname}, \{params})");
		try {
		    functions.glGetFramebufferAttachmentParameterivOES().invokeExact(target, attachment, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, attachment=\{attachment}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetFramebufferParameterfvAMD">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetFramebufferParameterfvAMD.xhtml">registry.khronos.org</a>
	 */
	public void glGetFramebufferParameterfvAMD(GLFramebufferTarget target, GLFramebufferAttachmentParameterName pname, int numsamples, int pixelindex, int size, MutPointer<Float> values) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetFramebufferParameterfvAMD(\{target}, \{pname}, \{numsamples}, \{pixelindex}, \{size}, \{values})");
		try {
		    functions.glGetFramebufferParameterfvAMD().invokeExact(target, pname, numsamples, pixelindex, size, values);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetFramebufferParameteriv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetFramebufferParameteriv.xhtml">registry.khronos.org</a>
	 */
	public void glGetFramebufferParameteriv(GLFramebufferTarget target, GLFramebufferAttachmentParameterName pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetFramebufferParameteriv(\{target}, \{pname}, \{params})");
		try {
		    functions.glGetFramebufferParameteriv().invokeExact(target, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetFramebufferParameterivEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetFramebufferParameterivEXT.xhtml">registry.khronos.org</a>
	 */
	public void glGetFramebufferParameterivEXT(int framebuffer, GLGetFramebufferParameter pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetFramebufferParameterivEXT(\{framebuffer}, \{pname}, \{params})");
		try {
		    functions.glGetFramebufferParameterivEXT().invokeExact(framebuffer, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetFramebufferPixelLocalStorageSizeEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetFramebufferPixelLocalStorageSizeEXT.xhtml">registry.khronos.org</a>
	 */
	public int glGetFramebufferPixelLocalStorageSizeEXT(int target) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetFramebufferPixelLocalStorageSizeEXT(\{target})");
		try {
		    var r = (int)functions.glGetFramebufferPixelLocalStorageSizeEXT().invokeExact(target);
			checkError();
			return r;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetGraphicsResetStatus">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetGraphicsResetStatus.xhtml">registry.khronos.org</a>
	 */
	public GLGraphicsResetStatus glGetGraphicsResetStatus() {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetGraphicsResetStatus()");
		try {
		    var r = (GLGraphicsResetStatus)functions.glGetGraphicsResetStatus().invokeExact();
			checkError();
			return r;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetGraphicsResetStatusARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetGraphicsResetStatusARB.xhtml">registry.khronos.org</a>
	 */
	public GLGraphicsResetStatus glGetGraphicsResetStatusARB() {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetGraphicsResetStatusARB()");
		try {
		    var r = (GLGraphicsResetStatus)functions.glGetGraphicsResetStatusARB().invokeExact();
			checkError();
			return r;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetGraphicsResetStatusEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetGraphicsResetStatusEXT.xhtml">registry.khronos.org</a>
	 */
	public GLGraphicsResetStatus glGetGraphicsResetStatusEXT() {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetGraphicsResetStatusEXT()");
		try {
		    var r = (GLGraphicsResetStatus)functions.glGetGraphicsResetStatusEXT().invokeExact();
			checkError();
			return r;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetGraphicsResetStatusKHR">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetGraphicsResetStatusKHR.xhtml">registry.khronos.org</a>
	 */
	public GLGraphicsResetStatus glGetGraphicsResetStatusKHR() {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetGraphicsResetStatusKHR()");
		try {
		    var r = (GLGraphicsResetStatus)functions.glGetGraphicsResetStatusKHR().invokeExact();
			checkError();
			return r;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetHandleARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetHandleARB.xhtml">registry.khronos.org</a>
	 */
	public int glGetHandleARB(GLContainerType pname) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetHandleARB(\{pname})");
		try {
		    var r = (int)functions.glGetHandleARB().invokeExact(pname);
			checkError();
			return r;
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetHistogram">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetHistogram.xhtml">registry.khronos.org</a>
	 */
	public void glGetHistogram(GLHistogramTargetEXT target, boolean reset, GLPixelFormat format, GLPixelType type, MutPointer<Void> values) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetHistogram(\{target}, \{reset}, \{format}, \{type}, \{values})");
		try {
		    functions.glGetHistogram().invokeExact(target, reset, format, type, values);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, format=\{format}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetHistogramEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetHistogramEXT.xhtml">registry.khronos.org</a>
	 */
	public void glGetHistogramEXT(GLHistogramTargetEXT target, boolean reset, GLPixelFormat format, GLPixelType type, MutPointer<Void> values) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetHistogramEXT(\{target}, \{reset}, \{format}, \{type}, \{values})");
		try {
		    functions.glGetHistogramEXT().invokeExact(target, reset, format, type, values);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, format=\{format}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetHistogramParameterfv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetHistogramParameterfv.xhtml">registry.khronos.org</a>
	 */
	public void glGetHistogramParameterfv(GLHistogramTargetEXT target, GLGetHistogramParameterPNameEXT pname, MutPointer<Float> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetHistogramParameterfv(\{target}, \{pname}, \{params})");
		try {
		    functions.glGetHistogramParameterfv().invokeExact(target, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetHistogramParameterfvEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetHistogramParameterfvEXT.xhtml">registry.khronos.org</a>
	 */
	public void glGetHistogramParameterfvEXT(GLHistogramTargetEXT target, GLGetHistogramParameterPNameEXT pname, MutPointer<Float> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetHistogramParameterfvEXT(\{target}, \{pname}, \{params})");
		try {
		    functions.glGetHistogramParameterfvEXT().invokeExact(target, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetHistogramParameteriv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetHistogramParameteriv.xhtml">registry.khronos.org</a>
	 */
	public void glGetHistogramParameteriv(GLHistogramTargetEXT target, GLGetHistogramParameterPNameEXT pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetHistogramParameteriv(\{target}, \{pname}, \{params})");
		try {
		    functions.glGetHistogramParameteriv().invokeExact(target, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetHistogramParameterivEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetHistogramParameterivEXT.xhtml">registry.khronos.org</a>
	 */
	public void glGetHistogramParameterivEXT(GLHistogramTargetEXT target, GLGetHistogramParameterPNameEXT pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetHistogramParameterivEXT(\{target}, \{pname}, \{params})");
		try {
		    functions.glGetHistogramParameterivEXT().invokeExact(target, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetHistogramParameterxvOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetHistogramParameterxvOES.xhtml">registry.khronos.org</a>
	 */
	public void glGetHistogramParameterxvOES(GLHistogramTargetEXT target, GLGetHistogramParameterPNameEXT pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetHistogramParameterxvOES(\{target}, \{pname}, \{params})");
		try {
		    functions.glGetHistogramParameterxvOES().invokeExact(target, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetImageHandleARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetImageHandleARB.xhtml">registry.khronos.org</a>
	 */
	public long glGetImageHandleARB(int texture, int level, boolean layered, int layer, GLPixelFormat format) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetImageHandleARB(\{texture}, \{level}, \{layered}, \{layer}, \{format})");
		try {
		    var r = (long)functions.glGetImageHandleARB().invokeExact(texture, level, layered, layer, format);
			checkError();
			return r;
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were format=\{format}.");
			}
			throw e;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetImageHandleNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetImageHandleNV.xhtml">registry.khronos.org</a>
	 */
	public long glGetImageHandleNV(int texture, int level, boolean layered, int layer, GLPixelFormat format) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetImageHandleNV(\{texture}, \{level}, \{layered}, \{layer}, \{format})");
		try {
		    var r = (long)functions.glGetImageHandleNV().invokeExact(texture, level, layered, layer, format);
			checkError();
			return r;
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were format=\{format}.");
			}
			throw e;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetImageTransformParameterfvHP">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetImageTransformParameterfvHP.xhtml">registry.khronos.org</a>
	 */
	public void glGetImageTransformParameterfvHP(GLImageTransformTargetHP target, GLImageTransformPNameHP pname, MutPointer<Float> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetImageTransformParameterfvHP(\{target}, \{pname}, \{params})");
		try {
		    functions.glGetImageTransformParameterfvHP().invokeExact(target, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetImageTransformParameterivHP">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetImageTransformParameterivHP.xhtml">registry.khronos.org</a>
	 */
	public void glGetImageTransformParameterivHP(GLImageTransformTargetHP target, GLImageTransformPNameHP pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetImageTransformParameterivHP(\{target}, \{pname}, \{params})");
		try {
		    functions.glGetImageTransformParameterivHP().invokeExact(target, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetInfoLogARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetInfoLogARB.xhtml">registry.khronos.org</a>
	 */
	public void glGetInfoLogARB(int obj, int maxLength, MutPointer<Integer> length, MutPointer<Byte> infoLog) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetInfoLogARB(\{obj}, \{maxLength}, \{length}, \{infoLog})");
		try {
		    functions.glGetInfoLogARB().invokeExact(obj, maxLength, length, infoLog);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetInstrumentsSGIX">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetInstrumentsSGIX.xhtml">registry.khronos.org</a>
	 */
	public int glGetInstrumentsSGIX() {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetInstrumentsSGIX()");
		try {
		    var r = (int)functions.glGetInstrumentsSGIX().invokeExact();
			checkError();
			return r;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetInteger64i_v">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetInteger64i_v.xhtml">registry.khronos.org</a>
	 */
	public void glGetInteger64i_v(GLGetPName target, int index, MutPointer<Long> data) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetInteger64i_v(\{target}, \{index}, \{data})");
		try {
		    functions.glGetInteger64i_v().invokeExact(target, index, data);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetInteger64v">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetInteger64v.xhtml">registry.khronos.org</a>
	 */
	public void glGetInteger64v(GLGetPName pname, MutPointer<Long> data) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetInteger64v(\{pname}, \{data})");
		try {
		    functions.glGetInteger64v().invokeExact(pname, data);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetInteger64vAPPLE">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetInteger64vAPPLE.xhtml">registry.khronos.org</a>
	 */
	public void glGetInteger64vAPPLE(GLGetPName pname, MutPointer<Long> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetInteger64vAPPLE(\{pname}, \{params})");
		try {
		    functions.glGetInteger64vAPPLE().invokeExact(pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetInteger64vEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetInteger64vEXT.xhtml">registry.khronos.org</a>
	 */
	public void glGetInteger64vEXT(GLGetPName pname, MutPointer<Long> data) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetInteger64vEXT(\{pname}, \{data})");
		try {
		    functions.glGetInteger64vEXT().invokeExact(pname, data);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetIntegerIndexedvEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetIntegerIndexedvEXT.xhtml">registry.khronos.org</a>
	 */
	public void glGetIntegerIndexedvEXT(GLGetPName target, int index, MutPointer<Integer> data) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetIntegerIndexedvEXT(\{target}, \{index}, \{data})");
		try {
		    functions.glGetIntegerIndexedvEXT().invokeExact(target, index, data);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetIntegeri_v">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetIntegeri_v.xhtml">registry.khronos.org</a>
	 */
	public void glGetIntegeri_v(GLGetPName target, int index, MutPointer<Integer> data) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetIntegeri_v(\{target}, \{index}, \{data})");
		try {
		    functions.glGetIntegeri_v().invokeExact(target, index, data);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetIntegeri_vEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetIntegeri_vEXT.xhtml">registry.khronos.org</a>
	 */
	public void glGetIntegeri_vEXT(GLGetPName target, int index, MutPointer<Integer> data) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetIntegeri_vEXT(\{target}, \{index}, \{data})");
		try {
		    functions.glGetIntegeri_vEXT().invokeExact(target, index, data);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetIntegerui64i_vNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetIntegerui64i_vNV.xhtml">registry.khronos.org</a>
	 */
	public void glGetIntegerui64i_vNV(GLEnum value, int index, MutPointer<Long> result) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetIntegerui64i_vNV(\{value}, \{index}, \{result})");
		try {
		    functions.glGetIntegerui64i_vNV().invokeExact(value, index, result);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were value=\{value}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetIntegerui64vNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetIntegerui64vNV.xhtml">registry.khronos.org</a>
	 */
	public void glGetIntegerui64vNV(GLEnum value, MutPointer<Long> result) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetIntegerui64vNV(\{value}, \{result})");
		try {
		    functions.glGetIntegerui64vNV().invokeExact(value, result);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were value=\{value}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetIntegerv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetIntegerv.xhtml">registry.khronos.org</a>
	 */
	public void glGetIntegerv(GLGetPName pname, MutPointer<Integer> data) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetIntegerv(\{pname}, \{data})");
		try {
		    functions.glGetIntegerv().invokeExact(pname, data);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	public int glGetInteger(GLGetPName pname) {
        try (var arena = Arena.ofConfined()) {
            var buf = arena.allocate(GLTypeCache.ML_I());
            glGetIntegerv(pname, new MutPointer<>(buf));
			var rvalue = buf.getAtIndex(GLTypeCache.ML_I(), 0);
            return rvalue;
        }
    }

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetInternalformatSampleivNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetInternalformatSampleivNV.xhtml">registry.khronos.org</a>
	 */
	public void glGetInternalformatSampleivNV(GLTextureTarget target, GLInternalFormat internalformat, int samples, GLInternalFormatPName pname, int count, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetInternalformatSampleivNV(\{target}, \{internalformat}, \{samples}, \{pname}, \{count}, \{params})");
		try {
		    functions.glGetInternalformatSampleivNV().invokeExact(target, internalformat, samples, pname, count, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, internalformat=\{internalformat}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetInternalformati64v">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetInternalformati64v.xhtml">registry.khronos.org</a>
	 */
	public void glGetInternalformati64v(GLTextureTarget target, GLInternalFormat internalformat, GLInternalFormatPName pname, int count, MutPointer<Long> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetInternalformati64v(\{target}, \{internalformat}, \{pname}, \{count}, \{params})");
		try {
		    functions.glGetInternalformati64v().invokeExact(target, internalformat, pname, count, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, internalformat=\{internalformat}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetInternalformativ">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetInternalformativ.xhtml">registry.khronos.org</a>
	 */
	public void glGetInternalformativ(GLTextureTarget target, GLInternalFormat internalformat, GLInternalFormatPName pname, int count, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetInternalformativ(\{target}, \{internalformat}, \{pname}, \{count}, \{params})");
		try {
		    functions.glGetInternalformativ().invokeExact(target, internalformat, pname, count, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, internalformat=\{internalformat}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetInvariantBooleanvEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetInvariantBooleanvEXT.xhtml">registry.khronos.org</a>
	 */
	public void glGetInvariantBooleanvEXT(int id, GLGetVariantValueEXT value, MutPointer<Boolean> data) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetInvariantBooleanvEXT(\{id}, \{value}, \{data})");
		try {
		    functions.glGetInvariantBooleanvEXT().invokeExact(id, value, data);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were value=\{value}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetInvariantFloatvEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetInvariantFloatvEXT.xhtml">registry.khronos.org</a>
	 */
	public void glGetInvariantFloatvEXT(int id, GLGetVariantValueEXT value, MutPointer<Float> data) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetInvariantFloatvEXT(\{id}, \{value}, \{data})");
		try {
		    functions.glGetInvariantFloatvEXT().invokeExact(id, value, data);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were value=\{value}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetInvariantIntegervEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetInvariantIntegervEXT.xhtml">registry.khronos.org</a>
	 */
	public void glGetInvariantIntegervEXT(int id, GLGetVariantValueEXT value, MutPointer<Integer> data) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetInvariantIntegervEXT(\{id}, \{value}, \{data})");
		try {
		    functions.glGetInvariantIntegervEXT().invokeExact(id, value, data);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were value=\{value}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetLightfv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetLightfv.xhtml">registry.khronos.org</a>
	 */
	public void glGetLightfv(GLLightName light, GLLightParameter pname, MutPointer<Float> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetLightfv(\{light}, \{pname}, \{params})");
		try {
		    functions.glGetLightfv().invokeExact(light, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were light=\{light}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetLightiv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetLightiv.xhtml">registry.khronos.org</a>
	 */
	public void glGetLightiv(GLLightName light, GLLightParameter pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetLightiv(\{light}, \{pname}, \{params})");
		try {
		    functions.glGetLightiv().invokeExact(light, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were light=\{light}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetLightxOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetLightxOES.xhtml">registry.khronos.org</a>
	 */
	public void glGetLightxOES(GLLightName light, GLLightParameter pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetLightxOES(\{light}, \{pname}, \{params})");
		try {
		    functions.glGetLightxOES().invokeExact(light, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were light=\{light}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetLightxv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetLightxv.xhtml">registry.khronos.org</a>
	 */
	public void glGetLightxv(GLLightName light, GLLightParameter pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetLightxv(\{light}, \{pname}, \{params})");
		try {
		    functions.glGetLightxv().invokeExact(light, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were light=\{light}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetLightxvOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetLightxvOES.xhtml">registry.khronos.org</a>
	 */
	public void glGetLightxvOES(GLLightName light, GLLightParameter pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetLightxvOES(\{light}, \{pname}, \{params})");
		try {
		    functions.glGetLightxvOES().invokeExact(light, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were light=\{light}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetListParameterfvSGIX">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetListParameterfvSGIX.xhtml">registry.khronos.org</a>
	 */
	public void glGetListParameterfvSGIX(int list, GLListParameterName pname, MutPointer<Float> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetListParameterfvSGIX(\{list}, \{pname}, \{params})");
		try {
		    functions.glGetListParameterfvSGIX().invokeExact(list, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetListParameterivSGIX">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetListParameterivSGIX.xhtml">registry.khronos.org</a>
	 */
	public void glGetListParameterivSGIX(int list, GLListParameterName pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetListParameterivSGIX(\{list}, \{pname}, \{params})");
		try {
		    functions.glGetListParameterivSGIX().invokeExact(list, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetLocalConstantBooleanvEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetLocalConstantBooleanvEXT.xhtml">registry.khronos.org</a>
	 */
	public void glGetLocalConstantBooleanvEXT(int id, GLGetVariantValueEXT value, MutPointer<Boolean> data) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetLocalConstantBooleanvEXT(\{id}, \{value}, \{data})");
		try {
		    functions.glGetLocalConstantBooleanvEXT().invokeExact(id, value, data);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were value=\{value}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetLocalConstantFloatvEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetLocalConstantFloatvEXT.xhtml">registry.khronos.org</a>
	 */
	public void glGetLocalConstantFloatvEXT(int id, GLGetVariantValueEXT value, MutPointer<Float> data) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetLocalConstantFloatvEXT(\{id}, \{value}, \{data})");
		try {
		    functions.glGetLocalConstantFloatvEXT().invokeExact(id, value, data);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were value=\{value}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetLocalConstantIntegervEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetLocalConstantIntegervEXT.xhtml">registry.khronos.org</a>
	 */
	public void glGetLocalConstantIntegervEXT(int id, GLGetVariantValueEXT value, MutPointer<Integer> data) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetLocalConstantIntegervEXT(\{id}, \{value}, \{data})");
		try {
		    functions.glGetLocalConstantIntegervEXT().invokeExact(id, value, data);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were value=\{value}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetMapAttribParameterfvNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetMapAttribParameterfvNV.xhtml">registry.khronos.org</a>
	 */
	public void glGetMapAttribParameterfvNV(GLEvalTargetNV target, int index, GLMapAttribParameterNV pname, MutPointer<Float> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetMapAttribParameterfvNV(\{target}, \{index}, \{pname}, \{params})");
		try {
		    functions.glGetMapAttribParameterfvNV().invokeExact(target, index, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetMapAttribParameterivNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetMapAttribParameterivNV.xhtml">registry.khronos.org</a>
	 */
	public void glGetMapAttribParameterivNV(GLEvalTargetNV target, int index, GLMapAttribParameterNV pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetMapAttribParameterivNV(\{target}, \{index}, \{pname}, \{params})");
		try {
		    functions.glGetMapAttribParameterivNV().invokeExact(target, index, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetMapControlPointsNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetMapControlPointsNV.xhtml">registry.khronos.org</a>
	 */
	public void glGetMapControlPointsNV(GLEvalTargetNV target, int index, GLMapTypeNV type, int ustride, int vstride, boolean packed, MutPointer<Void> points) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetMapControlPointsNV(\{target}, \{index}, \{type}, \{ustride}, \{vstride}, \{packed}, \{points})");
		try {
		    functions.glGetMapControlPointsNV().invokeExact(target, index, type, ustride, vstride, packed, points);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetMapParameterfvNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetMapParameterfvNV.xhtml">registry.khronos.org</a>
	 */
	public void glGetMapParameterfvNV(GLEvalTargetNV target, GLMapParameterNV pname, MutPointer<Float> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetMapParameterfvNV(\{target}, \{pname}, \{params})");
		try {
		    functions.glGetMapParameterfvNV().invokeExact(target, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetMapParameterivNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetMapParameterivNV.xhtml">registry.khronos.org</a>
	 */
	public void glGetMapParameterivNV(GLEvalTargetNV target, GLMapParameterNV pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetMapParameterivNV(\{target}, \{pname}, \{params})");
		try {
		    functions.glGetMapParameterivNV().invokeExact(target, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetMapdv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetMapdv.xhtml">registry.khronos.org</a>
	 */
	public void glGetMapdv(GLMapTarget target, GLGetMapQuery query, MutPointer<Double> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetMapdv(\{target}, \{query}, \{v})");
		try {
		    functions.glGetMapdv().invokeExact(target, query, v);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, query=\{query}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetMapfv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetMapfv.xhtml">registry.khronos.org</a>
	 */
	public void glGetMapfv(GLMapTarget target, GLGetMapQuery query, MutPointer<Float> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetMapfv(\{target}, \{query}, \{v})");
		try {
		    functions.glGetMapfv().invokeExact(target, query, v);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, query=\{query}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetMapiv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetMapiv.xhtml">registry.khronos.org</a>
	 */
	public void glGetMapiv(GLMapTarget target, GLGetMapQuery query, MutPointer<Integer> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetMapiv(\{target}, \{query}, \{v})");
		try {
		    functions.glGetMapiv().invokeExact(target, query, v);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, query=\{query}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetMapxvOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetMapxvOES.xhtml">registry.khronos.org</a>
	 */
	public void glGetMapxvOES(GLMapTarget target, GLGetMapQuery query, MutPointer<Integer> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetMapxvOES(\{target}, \{query}, \{v})");
		try {
		    functions.glGetMapxvOES().invokeExact(target, query, v);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, query=\{query}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetMaterialfv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetMaterialfv.xhtml">registry.khronos.org</a>
	 */
	public void glGetMaterialfv(GLTriangleFace face, GLMaterialParameter pname, MutPointer<Float> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetMaterialfv(\{face}, \{pname}, \{params})");
		try {
		    functions.glGetMaterialfv().invokeExact(face, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were face=\{face}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetMaterialiv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetMaterialiv.xhtml">registry.khronos.org</a>
	 */
	public void glGetMaterialiv(GLTriangleFace face, GLMaterialParameter pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetMaterialiv(\{face}, \{pname}, \{params})");
		try {
		    functions.glGetMaterialiv().invokeExact(face, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were face=\{face}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetMaterialxOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetMaterialxOES.xhtml">registry.khronos.org</a>
	 */
	public void glGetMaterialxOES(GLTriangleFace face, GLMaterialParameter pname, int param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetMaterialxOES(\{face}, \{pname}, \{param})");
		try {
		    functions.glGetMaterialxOES().invokeExact(face, pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were face=\{face}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetMaterialxv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetMaterialxv.xhtml">registry.khronos.org</a>
	 */
	public void glGetMaterialxv(GLTriangleFace face, GLMaterialParameter pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetMaterialxv(\{face}, \{pname}, \{params})");
		try {
		    functions.glGetMaterialxv().invokeExact(face, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were face=\{face}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetMaterialxvOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetMaterialxvOES.xhtml">registry.khronos.org</a>
	 */
	public void glGetMaterialxvOES(GLTriangleFace face, GLMaterialParameter pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetMaterialxvOES(\{face}, \{pname}, \{params})");
		try {
		    functions.glGetMaterialxvOES().invokeExact(face, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were face=\{face}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetMemoryObjectDetachedResourcesuivNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetMemoryObjectDetachedResourcesuivNV.xhtml">registry.khronos.org</a>
	 */
	public void glGetMemoryObjectDetachedResourcesuivNV(int memory, GLEnum pname, int first, int count, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetMemoryObjectDetachedResourcesuivNV(\{memory}, \{pname}, \{first}, \{count}, \{params})");
		try {
		    functions.glGetMemoryObjectDetachedResourcesuivNV().invokeExact(memory, pname, first, count, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetMemoryObjectParameterivEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetMemoryObjectParameterivEXT.xhtml">registry.khronos.org</a>
	 */
	public void glGetMemoryObjectParameterivEXT(int memoryObject, GLMemoryObjectParameterName pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetMemoryObjectParameterivEXT(\{memoryObject}, \{pname}, \{params})");
		try {
		    functions.glGetMemoryObjectParameterivEXT().invokeExact(memoryObject, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetMinmax">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetMinmax.xhtml">registry.khronos.org</a>
	 */
	public void glGetMinmax(GLMinmaxTargetEXT target, boolean reset, GLPixelFormat format, GLPixelType type, MutPointer<Void> values) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetMinmax(\{target}, \{reset}, \{format}, \{type}, \{values})");
		try {
		    functions.glGetMinmax().invokeExact(target, reset, format, type, values);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, format=\{format}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetMinmaxEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetMinmaxEXT.xhtml">registry.khronos.org</a>
	 */
	public void glGetMinmaxEXT(GLMinmaxTargetEXT target, boolean reset, GLPixelFormat format, GLPixelType type, MutPointer<Void> values) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetMinmaxEXT(\{target}, \{reset}, \{format}, \{type}, \{values})");
		try {
		    functions.glGetMinmaxEXT().invokeExact(target, reset, format, type, values);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, format=\{format}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetMinmaxParameterfv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetMinmaxParameterfv.xhtml">registry.khronos.org</a>
	 */
	public void glGetMinmaxParameterfv(GLMinmaxTargetEXT target, GLGetMinmaxParameterPNameEXT pname, MutPointer<Float> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetMinmaxParameterfv(\{target}, \{pname}, \{params})");
		try {
		    functions.glGetMinmaxParameterfv().invokeExact(target, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetMinmaxParameterfvEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetMinmaxParameterfvEXT.xhtml">registry.khronos.org</a>
	 */
	public void glGetMinmaxParameterfvEXT(GLMinmaxTargetEXT target, GLGetMinmaxParameterPNameEXT pname, MutPointer<Float> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetMinmaxParameterfvEXT(\{target}, \{pname}, \{params})");
		try {
		    functions.glGetMinmaxParameterfvEXT().invokeExact(target, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetMinmaxParameteriv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetMinmaxParameteriv.xhtml">registry.khronos.org</a>
	 */
	public void glGetMinmaxParameteriv(GLMinmaxTargetEXT target, GLGetMinmaxParameterPNameEXT pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetMinmaxParameteriv(\{target}, \{pname}, \{params})");
		try {
		    functions.glGetMinmaxParameteriv().invokeExact(target, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetMinmaxParameterivEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetMinmaxParameterivEXT.xhtml">registry.khronos.org</a>
	 */
	public void glGetMinmaxParameterivEXT(GLMinmaxTargetEXT target, GLGetMinmaxParameterPNameEXT pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetMinmaxParameterivEXT(\{target}, \{pname}, \{params})");
		try {
		    functions.glGetMinmaxParameterivEXT().invokeExact(target, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetMultiTexEnvfvEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetMultiTexEnvfvEXT.xhtml">registry.khronos.org</a>
	 */
	public void glGetMultiTexEnvfvEXT(GLTextureUnit texunit, GLTextureEnvTarget target, GLTextureEnvParameter pname, MutPointer<Float> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetMultiTexEnvfvEXT(\{texunit}, \{target}, \{pname}, \{params})");
		try {
		    functions.glGetMultiTexEnvfvEXT().invokeExact(texunit, target, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were texunit=\{texunit}, target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetMultiTexEnvivEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetMultiTexEnvivEXT.xhtml">registry.khronos.org</a>
	 */
	public void glGetMultiTexEnvivEXT(GLTextureUnit texunit, GLTextureEnvTarget target, GLTextureEnvParameter pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetMultiTexEnvivEXT(\{texunit}, \{target}, \{pname}, \{params})");
		try {
		    functions.glGetMultiTexEnvivEXT().invokeExact(texunit, target, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were texunit=\{texunit}, target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetMultiTexGendvEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetMultiTexGendvEXT.xhtml">registry.khronos.org</a>
	 */
	public void glGetMultiTexGendvEXT(GLTextureUnit texunit, GLTextureCoordName coord, GLTextureGenParameter pname, MutPointer<Double> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetMultiTexGendvEXT(\{texunit}, \{coord}, \{pname}, \{params})");
		try {
		    functions.glGetMultiTexGendvEXT().invokeExact(texunit, coord, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were texunit=\{texunit}, coord=\{coord}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetMultiTexGenfvEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetMultiTexGenfvEXT.xhtml">registry.khronos.org</a>
	 */
	public void glGetMultiTexGenfvEXT(GLTextureUnit texunit, GLTextureCoordName coord, GLTextureGenParameter pname, MutPointer<Float> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetMultiTexGenfvEXT(\{texunit}, \{coord}, \{pname}, \{params})");
		try {
		    functions.glGetMultiTexGenfvEXT().invokeExact(texunit, coord, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were texunit=\{texunit}, coord=\{coord}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetMultiTexGenivEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetMultiTexGenivEXT.xhtml">registry.khronos.org</a>
	 */
	public void glGetMultiTexGenivEXT(GLTextureUnit texunit, GLTextureCoordName coord, GLTextureGenParameter pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetMultiTexGenivEXT(\{texunit}, \{coord}, \{pname}, \{params})");
		try {
		    functions.glGetMultiTexGenivEXT().invokeExact(texunit, coord, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were texunit=\{texunit}, coord=\{coord}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetMultiTexImageEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetMultiTexImageEXT.xhtml">registry.khronos.org</a>
	 */
	public void glGetMultiTexImageEXT(GLTextureUnit texunit, GLTextureTarget target, int level, GLPixelFormat format, GLPixelType type, MutPointer<Void> pixels) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetMultiTexImageEXT(\{texunit}, \{target}, \{level}, \{format}, \{type}, \{pixels})");
		try {
		    functions.glGetMultiTexImageEXT().invokeExact(texunit, target, level, format, type, pixels);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were texunit=\{texunit}, target=\{target}, format=\{format}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetMultiTexLevelParameterfvEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetMultiTexLevelParameterfvEXT.xhtml">registry.khronos.org</a>
	 */
	public void glGetMultiTexLevelParameterfvEXT(GLTextureUnit texunit, GLTextureTarget target, int level, GLGetTextureParameter pname, MutPointer<Float> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetMultiTexLevelParameterfvEXT(\{texunit}, \{target}, \{level}, \{pname}, \{params})");
		try {
		    functions.glGetMultiTexLevelParameterfvEXT().invokeExact(texunit, target, level, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were texunit=\{texunit}, target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetMultiTexLevelParameterivEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetMultiTexLevelParameterivEXT.xhtml">registry.khronos.org</a>
	 */
	public void glGetMultiTexLevelParameterivEXT(GLTextureUnit texunit, GLTextureTarget target, int level, GLGetTextureParameter pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetMultiTexLevelParameterivEXT(\{texunit}, \{target}, \{level}, \{pname}, \{params})");
		try {
		    functions.glGetMultiTexLevelParameterivEXT().invokeExact(texunit, target, level, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were texunit=\{texunit}, target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetMultiTexParameterIivEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetMultiTexParameterIivEXT.xhtml">registry.khronos.org</a>
	 */
	public void glGetMultiTexParameterIivEXT(GLTextureUnit texunit, GLTextureTarget target, GLGetTextureParameter pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetMultiTexParameterIivEXT(\{texunit}, \{target}, \{pname}, \{params})");
		try {
		    functions.glGetMultiTexParameterIivEXT().invokeExact(texunit, target, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were texunit=\{texunit}, target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetMultiTexParameterIuivEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetMultiTexParameterIuivEXT.xhtml">registry.khronos.org</a>
	 */
	public void glGetMultiTexParameterIuivEXT(GLTextureUnit texunit, GLTextureTarget target, GLGetTextureParameter pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetMultiTexParameterIuivEXT(\{texunit}, \{target}, \{pname}, \{params})");
		try {
		    functions.glGetMultiTexParameterIuivEXT().invokeExact(texunit, target, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were texunit=\{texunit}, target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetMultiTexParameterfvEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetMultiTexParameterfvEXT.xhtml">registry.khronos.org</a>
	 */
	public void glGetMultiTexParameterfvEXT(GLTextureUnit texunit, GLTextureTarget target, GLGetTextureParameter pname, MutPointer<Float> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetMultiTexParameterfvEXT(\{texunit}, \{target}, \{pname}, \{params})");
		try {
		    functions.glGetMultiTexParameterfvEXT().invokeExact(texunit, target, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were texunit=\{texunit}, target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetMultiTexParameterivEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetMultiTexParameterivEXT.xhtml">registry.khronos.org</a>
	 */
	public void glGetMultiTexParameterivEXT(GLTextureUnit texunit, GLTextureTarget target, GLGetTextureParameter pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetMultiTexParameterivEXT(\{texunit}, \{target}, \{pname}, \{params})");
		try {
		    functions.glGetMultiTexParameterivEXT().invokeExact(texunit, target, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were texunit=\{texunit}, target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetMultisamplefv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetMultisamplefv.xhtml">registry.khronos.org</a>
	 */
	public void glGetMultisamplefv(GLGetMultisamplePNameNV pname, int index, MutPointer<Float> val) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetMultisamplefv(\{pname}, \{index}, \{val})");
		try {
		    functions.glGetMultisamplefv().invokeExact(pname, index, val);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetMultisamplefvNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetMultisamplefvNV.xhtml">registry.khronos.org</a>
	 */
	public void glGetMultisamplefvNV(GLGetMultisamplePNameNV pname, int index, MutPointer<Float> val) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetMultisamplefvNV(\{pname}, \{index}, \{val})");
		try {
		    functions.glGetMultisamplefvNV().invokeExact(pname, index, val);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetNamedBufferParameteri64v">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetNamedBufferParameteri64v.xhtml">registry.khronos.org</a>
	 */
	public void glGetNamedBufferParameteri64v(int buffer, GLBufferPNameARB pname, MutPointer<Long> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetNamedBufferParameteri64v(\{buffer}, \{pname}, \{params})");
		try {
		    functions.glGetNamedBufferParameteri64v().invokeExact(buffer, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetNamedBufferParameteriv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetNamedBufferParameteriv.xhtml">registry.khronos.org</a>
	 */
	public void glGetNamedBufferParameteriv(int buffer, GLBufferPNameARB pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetNamedBufferParameteriv(\{buffer}, \{pname}, \{params})");
		try {
		    functions.glGetNamedBufferParameteriv().invokeExact(buffer, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetNamedBufferParameterivEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetNamedBufferParameterivEXT.xhtml">registry.khronos.org</a>
	 */
	public void glGetNamedBufferParameterivEXT(int buffer, GLBufferPNameARB pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetNamedBufferParameterivEXT(\{buffer}, \{pname}, \{params})");
		try {
		    functions.glGetNamedBufferParameterivEXT().invokeExact(buffer, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetNamedBufferParameterui64vNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetNamedBufferParameterui64vNV.xhtml">registry.khronos.org</a>
	 */
	public void glGetNamedBufferParameterui64vNV(int buffer, GLBufferPNameARB pname, MutPointer<Long> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetNamedBufferParameterui64vNV(\{buffer}, \{pname}, \{params})");
		try {
		    functions.glGetNamedBufferParameterui64vNV().invokeExact(buffer, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetNamedBufferPointerv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetNamedBufferPointerv.xhtml">registry.khronos.org</a>
	 */
	public void glGetNamedBufferPointerv(int buffer, GLBufferPointerNameARB pname, MutPointer<MutPointer<Void>> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetNamedBufferPointerv(\{buffer}, \{pname}, \{params})");
		try {
		    functions.glGetNamedBufferPointerv().invokeExact(buffer, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetNamedBufferPointervEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetNamedBufferPointervEXT.xhtml">registry.khronos.org</a>
	 */
	public void glGetNamedBufferPointervEXT(int buffer, GLBufferPointerNameARB pname, MutPointer<MutPointer<Void>> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetNamedBufferPointervEXT(\{buffer}, \{pname}, \{params})");
		try {
		    functions.glGetNamedBufferPointervEXT().invokeExact(buffer, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetNamedBufferSubData">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetNamedBufferSubData.xhtml">registry.khronos.org</a>
	 */
	public void glGetNamedBufferSubData(int buffer, long offset, long size, MutPointer<Void> data) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetNamedBufferSubData(\{buffer}, \{offset}, \{size}, \{data})");
		try {
		    functions.glGetNamedBufferSubData().invokeExact(buffer, offset, size, data);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetNamedBufferSubDataEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetNamedBufferSubDataEXT.xhtml">registry.khronos.org</a>
	 */
	public void glGetNamedBufferSubDataEXT(int buffer, long offset, long size, MutPointer<Void> data) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetNamedBufferSubDataEXT(\{buffer}, \{offset}, \{size}, \{data})");
		try {
		    functions.glGetNamedBufferSubDataEXT().invokeExact(buffer, offset, size, data);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetNamedFramebufferParameterfvAMD">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetNamedFramebufferParameterfvAMD.xhtml">registry.khronos.org</a>
	 */
	public void glGetNamedFramebufferParameterfvAMD(int framebuffer, GLEnum pname, int numsamples, int pixelindex, int size, MutPointer<Float> values) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetNamedFramebufferParameterfvAMD(\{framebuffer}, \{pname}, \{numsamples}, \{pixelindex}, \{size}, \{values})");
		try {
		    functions.glGetNamedFramebufferParameterfvAMD().invokeExact(framebuffer, pname, numsamples, pixelindex, size, values);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetNamedFramebufferAttachmentParameteriv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetNamedFramebufferAttachmentParameteriv.xhtml">registry.khronos.org</a>
	 */
	public void glGetNamedFramebufferAttachmentParameteriv(int framebuffer, GLFramebufferAttachment attachment, GLFramebufferAttachmentParameterName pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetNamedFramebufferAttachmentParameteriv(\{framebuffer}, \{attachment}, \{pname}, \{params})");
		try {
		    functions.glGetNamedFramebufferAttachmentParameteriv().invokeExact(framebuffer, attachment, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were attachment=\{attachment}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetNamedFramebufferAttachmentParameterivEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetNamedFramebufferAttachmentParameterivEXT.xhtml">registry.khronos.org</a>
	 */
	public void glGetNamedFramebufferAttachmentParameterivEXT(int framebuffer, GLFramebufferAttachment attachment, GLFramebufferAttachmentParameterName pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetNamedFramebufferAttachmentParameterivEXT(\{framebuffer}, \{attachment}, \{pname}, \{params})");
		try {
		    functions.glGetNamedFramebufferAttachmentParameterivEXT().invokeExact(framebuffer, attachment, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were attachment=\{attachment}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetNamedFramebufferParameteriv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetNamedFramebufferParameteriv.xhtml">registry.khronos.org</a>
	 */
	public void glGetNamedFramebufferParameteriv(int framebuffer, GLGetFramebufferParameter pname, MutPointer<Integer> param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetNamedFramebufferParameteriv(\{framebuffer}, \{pname}, \{param})");
		try {
		    functions.glGetNamedFramebufferParameteriv().invokeExact(framebuffer, pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetNamedFramebufferParameterivEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetNamedFramebufferParameterivEXT.xhtml">registry.khronos.org</a>
	 */
	public void glGetNamedFramebufferParameterivEXT(int framebuffer, GLGetFramebufferParameter pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetNamedFramebufferParameterivEXT(\{framebuffer}, \{pname}, \{params})");
		try {
		    functions.glGetNamedFramebufferParameterivEXT().invokeExact(framebuffer, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetNamedProgramLocalParameterIivEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetNamedProgramLocalParameterIivEXT.xhtml">registry.khronos.org</a>
	 */
	public void glGetNamedProgramLocalParameterIivEXT(int program, GLProgramTarget target, int index, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetNamedProgramLocalParameterIivEXT(\{program}, \{target}, \{index}, \{params})");
		try {
		    functions.glGetNamedProgramLocalParameterIivEXT().invokeExact(program, target, index, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetNamedProgramLocalParameterIuivEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetNamedProgramLocalParameterIuivEXT.xhtml">registry.khronos.org</a>
	 */
	public void glGetNamedProgramLocalParameterIuivEXT(int program, GLProgramTarget target, int index, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetNamedProgramLocalParameterIuivEXT(\{program}, \{target}, \{index}, \{params})");
		try {
		    functions.glGetNamedProgramLocalParameterIuivEXT().invokeExact(program, target, index, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetNamedProgramLocalParameterdvEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetNamedProgramLocalParameterdvEXT.xhtml">registry.khronos.org</a>
	 */
	public void glGetNamedProgramLocalParameterdvEXT(int program, GLProgramTarget target, int index, MutPointer<Double> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetNamedProgramLocalParameterdvEXT(\{program}, \{target}, \{index}, \{params})");
		try {
		    functions.glGetNamedProgramLocalParameterdvEXT().invokeExact(program, target, index, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetNamedProgramLocalParameterfvEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetNamedProgramLocalParameterfvEXT.xhtml">registry.khronos.org</a>
	 */
	public void glGetNamedProgramLocalParameterfvEXT(int program, GLProgramTarget target, int index, MutPointer<Float> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetNamedProgramLocalParameterfvEXT(\{program}, \{target}, \{index}, \{params})");
		try {
		    functions.glGetNamedProgramLocalParameterfvEXT().invokeExact(program, target, index, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetNamedProgramStringEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetNamedProgramStringEXT.xhtml">registry.khronos.org</a>
	 */
	public void glGetNamedProgramStringEXT(int program, GLProgramTarget target, GLProgramStringProperty pname, MutPointer<Void> string) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetNamedProgramStringEXT(\{program}, \{target}, \{pname}, \{string})");
		try {
		    functions.glGetNamedProgramStringEXT().invokeExact(program, target, pname, string);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetNamedProgramivEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetNamedProgramivEXT.xhtml">registry.khronos.org</a>
	 */
	public void glGetNamedProgramivEXT(int program, GLProgramTarget target, GLProgramPropertyARB pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetNamedProgramivEXT(\{program}, \{target}, \{pname}, \{params})");
		try {
		    functions.glGetNamedProgramivEXT().invokeExact(program, target, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetNamedRenderbufferParameteriv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetNamedRenderbufferParameteriv.xhtml">registry.khronos.org</a>
	 */
	public void glGetNamedRenderbufferParameteriv(int renderbuffer, GLRenderbufferParameterName pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetNamedRenderbufferParameteriv(\{renderbuffer}, \{pname}, \{params})");
		try {
		    functions.glGetNamedRenderbufferParameteriv().invokeExact(renderbuffer, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetNamedRenderbufferParameterivEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetNamedRenderbufferParameterivEXT.xhtml">registry.khronos.org</a>
	 */
	public void glGetNamedRenderbufferParameterivEXT(int renderbuffer, GLRenderbufferParameterName pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetNamedRenderbufferParameterivEXT(\{renderbuffer}, \{pname}, \{params})");
		try {
		    functions.glGetNamedRenderbufferParameterivEXT().invokeExact(renderbuffer, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetNamedStringARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetNamedStringARB.xhtml">registry.khronos.org</a>
	 */
	public void glGetNamedStringARB(int namelen, ConstPointer<Byte> name, int bufSize, MutPointer<Integer> stringlen, MutPointer<Byte> string) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetNamedStringARB(\{namelen}, \{name}, \{bufSize}, \{stringlen}, \{string})");
		try {
		    functions.glGetNamedStringARB().invokeExact(namelen, name, bufSize, stringlen, string);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetNamedStringivARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetNamedStringivARB.xhtml">registry.khronos.org</a>
	 */
	public void glGetNamedStringivARB(int namelen, ConstPointer<Byte> name, GLEnum pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetNamedStringivARB(\{namelen}, \{name}, \{pname}, \{params})");
		try {
		    functions.glGetNamedStringivARB().invokeExact(namelen, name, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetNextPerfQueryIdINTEL">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetNextPerfQueryIdINTEL.xhtml">registry.khronos.org</a>
	 */
	public void glGetNextPerfQueryIdINTEL(int queryId, MutPointer<Integer> nextQueryId) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetNextPerfQueryIdINTEL(\{queryId}, \{nextQueryId})");
		try {
		    functions.glGetNextPerfQueryIdINTEL().invokeExact(queryId, nextQueryId);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetObjectBufferfvATI">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetObjectBufferfvATI.xhtml">registry.khronos.org</a>
	 */
	public void glGetObjectBufferfvATI(int buffer, GLArrayObjectPNameATI pname, MutPointer<Float> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetObjectBufferfvATI(\{buffer}, \{pname}, \{params})");
		try {
		    functions.glGetObjectBufferfvATI().invokeExact(buffer, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetObjectBufferivATI">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetObjectBufferivATI.xhtml">registry.khronos.org</a>
	 */
	public void glGetObjectBufferivATI(int buffer, GLArrayObjectPNameATI pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetObjectBufferivATI(\{buffer}, \{pname}, \{params})");
		try {
		    functions.glGetObjectBufferivATI().invokeExact(buffer, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetObjectLabel">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetObjectLabel.xhtml">registry.khronos.org</a>
	 */
	public void glGetObjectLabel(GLObjectIdentifier identifier, int name, int bufSize, MutPointer<Integer> length, MutPointer<Byte> label) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetObjectLabel(\{identifier}, \{name}, \{bufSize}, \{length}, \{label})");
		try {
		    functions.glGetObjectLabel().invokeExact(identifier, name, bufSize, length, label);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were identifier=\{identifier}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetObjectLabelEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetObjectLabelEXT.xhtml">registry.khronos.org</a>
	 */
	public void glGetObjectLabelEXT(GLEnum type, int object, int bufSize, MutPointer<Integer> length, MutPointer<Byte> label) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetObjectLabelEXT(\{type}, \{object}, \{bufSize}, \{length}, \{label})");
		try {
		    functions.glGetObjectLabelEXT().invokeExact(type, object, bufSize, length, label);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetObjectLabelKHR">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetObjectLabelKHR.xhtml">registry.khronos.org</a>
	 */
	public void glGetObjectLabelKHR(GLEnum identifier, int name, int bufSize, MutPointer<Integer> length, MutPointer<Byte> label) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetObjectLabelKHR(\{identifier}, \{name}, \{bufSize}, \{length}, \{label})");
		try {
		    functions.glGetObjectLabelKHR().invokeExact(identifier, name, bufSize, length, label);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were identifier=\{identifier}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetObjectParameterfvARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetObjectParameterfvARB.xhtml">registry.khronos.org</a>
	 */
	public void glGetObjectParameterfvARB(int obj, GLEnum pname, MutPointer<Float> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetObjectParameterfvARB(\{obj}, \{pname}, \{params})");
		try {
		    functions.glGetObjectParameterfvARB().invokeExact(obj, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetObjectParameterivAPPLE">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetObjectParameterivAPPLE.xhtml">registry.khronos.org</a>
	 */
	public void glGetObjectParameterivAPPLE(GLEnum objectType, int name, GLEnum pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetObjectParameterivAPPLE(\{objectType}, \{name}, \{pname}, \{params})");
		try {
		    functions.glGetObjectParameterivAPPLE().invokeExact(objectType, name, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were objectType=\{objectType}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetObjectParameterivARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetObjectParameterivARB.xhtml">registry.khronos.org</a>
	 */
	public void glGetObjectParameterivARB(int obj, GLEnum pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetObjectParameterivARB(\{obj}, \{pname}, \{params})");
		try {
		    functions.glGetObjectParameterivARB().invokeExact(obj, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetObjectPtrLabel">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetObjectPtrLabel.xhtml">registry.khronos.org</a>
	 */
	public void glGetObjectPtrLabel(ConstPointer<Void> ptr, int bufSize, MutPointer<Integer> length, MutPointer<Byte> label) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetObjectPtrLabel(\{ptr}, \{bufSize}, \{length}, \{label})");
		try {
		    functions.glGetObjectPtrLabel().invokeExact(ptr, bufSize, length, label);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetObjectPtrLabelKHR">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetObjectPtrLabelKHR.xhtml">registry.khronos.org</a>
	 */
	public void glGetObjectPtrLabelKHR(ConstPointer<Void> ptr, int bufSize, MutPointer<Integer> length, MutPointer<Byte> label) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetObjectPtrLabelKHR(\{ptr}, \{bufSize}, \{length}, \{label})");
		try {
		    functions.glGetObjectPtrLabelKHR().invokeExact(ptr, bufSize, length, label);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetOcclusionQueryivNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetOcclusionQueryivNV.xhtml">registry.khronos.org</a>
	 */
	public void glGetOcclusionQueryivNV(int id, GLOcclusionQueryParameterNameNV pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetOcclusionQueryivNV(\{id}, \{pname}, \{params})");
		try {
		    functions.glGetOcclusionQueryivNV().invokeExact(id, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetOcclusionQueryuivNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetOcclusionQueryuivNV.xhtml">registry.khronos.org</a>
	 */
	public void glGetOcclusionQueryuivNV(int id, GLOcclusionQueryParameterNameNV pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetOcclusionQueryuivNV(\{id}, \{pname}, \{params})");
		try {
		    functions.glGetOcclusionQueryuivNV().invokeExact(id, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetPathColorGenfvNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetPathColorGenfvNV.xhtml">registry.khronos.org</a>
	 */
	public void glGetPathColorGenfvNV(GLPathColor color, GLPathGenMode pname, MutPointer<Float> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetPathColorGenfvNV(\{color}, \{pname}, \{value})");
		try {
		    functions.glGetPathColorGenfvNV().invokeExact(color, pname, value);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were color=\{color}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetPathColorGenivNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetPathColorGenivNV.xhtml">registry.khronos.org</a>
	 */
	public void glGetPathColorGenivNV(GLPathColor color, GLPathGenMode pname, MutPointer<Integer> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetPathColorGenivNV(\{color}, \{pname}, \{value})");
		try {
		    functions.glGetPathColorGenivNV().invokeExact(color, pname, value);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were color=\{color}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetPathCommandsNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetPathCommandsNV.xhtml">registry.khronos.org</a>
	 */
	public void glGetPathCommandsNV(int path, MutPointer<Byte> commands) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetPathCommandsNV(\{path}, \{commands})");
		try {
		    functions.glGetPathCommandsNV().invokeExact(path, commands);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetPathCoordsNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetPathCoordsNV.xhtml">registry.khronos.org</a>
	 */
	public void glGetPathCoordsNV(int path, MutPointer<Float> coords) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetPathCoordsNV(\{path}, \{coords})");
		try {
		    functions.glGetPathCoordsNV().invokeExact(path, coords);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetPathDashArrayNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetPathDashArrayNV.xhtml">registry.khronos.org</a>
	 */
	public void glGetPathDashArrayNV(int path, MutPointer<Float> dashArray) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetPathDashArrayNV(\{path}, \{dashArray})");
		try {
		    functions.glGetPathDashArrayNV().invokeExact(path, dashArray);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetPathLengthNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetPathLengthNV.xhtml">registry.khronos.org</a>
	 */
	public float glGetPathLengthNV(int path, int startSegment, int numSegments) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetPathLengthNV(\{path}, \{startSegment}, \{numSegments})");
		try {
		    var r = (float)functions.glGetPathLengthNV().invokeExact(path, startSegment, numSegments);
			checkError();
			return r;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetPathMetricRangeNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetPathMetricRangeNV.xhtml">registry.khronos.org</a>
	 */
	public void glGetPathMetricRangeNV(int metricQueryMask, int firstPathName, int numPaths, int stride, MutPointer<Float> metrics) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetPathMetricRangeNV(\{metricQueryMask}, \{firstPathName}, \{numPaths}, \{stride}, \{metrics})");
		try {
		    functions.glGetPathMetricRangeNV().invokeExact(metricQueryMask, firstPathName, numPaths, stride, metrics);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetPathMetricsNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetPathMetricsNV.xhtml">registry.khronos.org</a>
	 */
	public void glGetPathMetricsNV(int metricQueryMask, int numPaths, GLPathElementType pathNameType, ConstPointer<Void> paths, int pathBase, int stride, MutPointer<Float> metrics) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetPathMetricsNV(\{metricQueryMask}, \{numPaths}, \{pathNameType}, \{paths}, \{pathBase}, \{stride}, \{metrics})");
		try {
		    functions.glGetPathMetricsNV().invokeExact(metricQueryMask, numPaths, pathNameType, paths, pathBase, stride, metrics);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pathNameType=\{pathNameType}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetPathParameterfvNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetPathParameterfvNV.xhtml">registry.khronos.org</a>
	 */
	public void glGetPathParameterfvNV(int path, GLPathParameter pname, MutPointer<Float> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetPathParameterfvNV(\{path}, \{pname}, \{value})");
		try {
		    functions.glGetPathParameterfvNV().invokeExact(path, pname, value);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetPathParameterivNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetPathParameterivNV.xhtml">registry.khronos.org</a>
	 */
	public void glGetPathParameterivNV(int path, GLPathParameter pname, MutPointer<Integer> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetPathParameterivNV(\{path}, \{pname}, \{value})");
		try {
		    functions.glGetPathParameterivNV().invokeExact(path, pname, value);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetPathSpacingNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetPathSpacingNV.xhtml">registry.khronos.org</a>
	 */
	public void glGetPathSpacingNV(GLPathListMode pathListMode, int numPaths, GLPathElementType pathNameType, ConstPointer<Void> paths, int pathBase, float advanceScale, float kerningScale, GLPathTransformType transformType, MutPointer<Float> returnedSpacing) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetPathSpacingNV(\{pathListMode}, \{numPaths}, \{pathNameType}, \{paths}, \{pathBase}, \{advanceScale}, \{kerningScale}, \{transformType}, \{returnedSpacing})");
		try {
		    functions.glGetPathSpacingNV().invokeExact(pathListMode, numPaths, pathNameType, paths, pathBase, advanceScale, kerningScale, transformType, returnedSpacing);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pathListMode=\{pathListMode}, pathNameType=\{pathNameType}, transformType=\{transformType}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetPathTexGenfvNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetPathTexGenfvNV.xhtml">registry.khronos.org</a>
	 */
	public void glGetPathTexGenfvNV(GLTextureUnit texCoordSet, GLPathGenMode pname, MutPointer<Float> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetPathTexGenfvNV(\{texCoordSet}, \{pname}, \{value})");
		try {
		    functions.glGetPathTexGenfvNV().invokeExact(texCoordSet, pname, value);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were texCoordSet=\{texCoordSet}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetPathTexGenivNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetPathTexGenivNV.xhtml">registry.khronos.org</a>
	 */
	public void glGetPathTexGenivNV(GLTextureUnit texCoordSet, GLPathGenMode pname, MutPointer<Integer> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetPathTexGenivNV(\{texCoordSet}, \{pname}, \{value})");
		try {
		    functions.glGetPathTexGenivNV().invokeExact(texCoordSet, pname, value);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were texCoordSet=\{texCoordSet}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetPerfCounterInfoINTEL">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetPerfCounterInfoINTEL.xhtml">registry.khronos.org</a>
	 */
	public void glGetPerfCounterInfoINTEL(int queryId, int counterId, int counterNameLength, MutPointer<Byte> counterName, int counterDescLength, MutPointer<Byte> counterDesc, MutPointer<Integer> counterOffset, MutPointer<Integer> counterDataSize, MutPointer<Integer> counterTypeEnum, MutPointer<Integer> counterDataTypeEnum, MutPointer<Long> rawCounterMaxValue) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetPerfCounterInfoINTEL(\{queryId}, \{counterId}, \{counterNameLength}, \{counterName}, \{counterDescLength}, \{counterDesc}, \{counterOffset}, \{counterDataSize}, \{counterTypeEnum}, \{counterDataTypeEnum}, \{rawCounterMaxValue})");
		try {
		    functions.glGetPerfCounterInfoINTEL().invokeExact(queryId, counterId, counterNameLength, counterName, counterDescLength, counterDesc, counterOffset, counterDataSize, counterTypeEnum, counterDataTypeEnum, rawCounterMaxValue);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetPerfMonitorCounterDataAMD">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetPerfMonitorCounterDataAMD.xhtml">registry.khronos.org</a>
	 */
	public void glGetPerfMonitorCounterDataAMD(int monitor, GLEnum pname, int dataSize, MutPointer<Integer> data, MutPointer<Integer> bytesWritten) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetPerfMonitorCounterDataAMD(\{monitor}, \{pname}, \{dataSize}, \{data}, \{bytesWritten})");
		try {
		    functions.glGetPerfMonitorCounterDataAMD().invokeExact(monitor, pname, dataSize, data, bytesWritten);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetPerfMonitorCounterInfoAMD">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetPerfMonitorCounterInfoAMD.xhtml">registry.khronos.org</a>
	 */
	public void glGetPerfMonitorCounterInfoAMD(int group, int counter, GLEnum pname, MutPointer<Void> data) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetPerfMonitorCounterInfoAMD(\{group}, \{counter}, \{pname}, \{data})");
		try {
		    functions.glGetPerfMonitorCounterInfoAMD().invokeExact(group, counter, pname, data);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetPerfMonitorCounterStringAMD">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetPerfMonitorCounterStringAMD.xhtml">registry.khronos.org</a>
	 */
	public void glGetPerfMonitorCounterStringAMD(int group, int counter, int bufSize, MutPointer<Integer> length, MutPointer<Byte> counterString) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetPerfMonitorCounterStringAMD(\{group}, \{counter}, \{bufSize}, \{length}, \{counterString})");
		try {
		    functions.glGetPerfMonitorCounterStringAMD().invokeExact(group, counter, bufSize, length, counterString);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetPerfMonitorCountersAMD">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetPerfMonitorCountersAMD.xhtml">registry.khronos.org</a>
	 */
	public void glGetPerfMonitorCountersAMD(int group, MutPointer<Integer> numCounters, MutPointer<Integer> maxActiveCounters, int counterSize, MutPointer<Integer> counters) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetPerfMonitorCountersAMD(\{group}, \{numCounters}, \{maxActiveCounters}, \{counterSize}, \{counters})");
		try {
		    functions.glGetPerfMonitorCountersAMD().invokeExact(group, numCounters, maxActiveCounters, counterSize, counters);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetPerfMonitorGroupStringAMD">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetPerfMonitorGroupStringAMD.xhtml">registry.khronos.org</a>
	 */
	public void glGetPerfMonitorGroupStringAMD(int group, int bufSize, MutPointer<Integer> length, MutPointer<Byte> groupString) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetPerfMonitorGroupStringAMD(\{group}, \{bufSize}, \{length}, \{groupString})");
		try {
		    functions.glGetPerfMonitorGroupStringAMD().invokeExact(group, bufSize, length, groupString);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetPerfMonitorGroupsAMD">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetPerfMonitorGroupsAMD.xhtml">registry.khronos.org</a>
	 */
	public void glGetPerfMonitorGroupsAMD(MutPointer<Integer> numGroups, int groupsSize, MutPointer<Integer> groups) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetPerfMonitorGroupsAMD(\{numGroups}, \{groupsSize}, \{groups})");
		try {
		    functions.glGetPerfMonitorGroupsAMD().invokeExact(numGroups, groupsSize, groups);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetPerfQueryDataINTEL">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetPerfQueryDataINTEL.xhtml">registry.khronos.org</a>
	 */
	public void glGetPerfQueryDataINTEL(int queryHandle, int flags, int dataSize, MutPointer<Void> data, MutPointer<Integer> bytesWritten) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetPerfQueryDataINTEL(\{queryHandle}, \{flags}, \{dataSize}, \{data}, \{bytesWritten})");
		try {
		    functions.glGetPerfQueryDataINTEL().invokeExact(queryHandle, flags, dataSize, data, bytesWritten);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetPerfQueryIdByNameINTEL">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetPerfQueryIdByNameINTEL.xhtml">registry.khronos.org</a>
	 */
	public void glGetPerfQueryIdByNameINTEL(MutPointer<Byte> queryName, MutPointer<Integer> queryId) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetPerfQueryIdByNameINTEL(\{queryName}, \{queryId})");
		try {
		    functions.glGetPerfQueryIdByNameINTEL().invokeExact(queryName, queryId);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetPerfQueryInfoINTEL">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetPerfQueryInfoINTEL.xhtml">registry.khronos.org</a>
	 */
	public void glGetPerfQueryInfoINTEL(int queryId, int queryNameLength, MutPointer<Byte> queryName, MutPointer<Integer> dataSize, MutPointer<Integer> noCounters, MutPointer<Integer> noInstances, MutPointer<Integer> capsMask) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetPerfQueryInfoINTEL(\{queryId}, \{queryNameLength}, \{queryName}, \{dataSize}, \{noCounters}, \{noInstances}, \{capsMask})");
		try {
		    functions.glGetPerfQueryInfoINTEL().invokeExact(queryId, queryNameLength, queryName, dataSize, noCounters, noInstances, capsMask);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetPixelMapfv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetPixelMapfv.xhtml">registry.khronos.org</a>
	 */
	public void glGetPixelMapfv(GLPixelMap map, MutPointer<Float> values) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetPixelMapfv(\{map}, \{values})");
		try {
		    functions.glGetPixelMapfv().invokeExact(map, values);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were map=\{map}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetPixelMapuiv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetPixelMapuiv.xhtml">registry.khronos.org</a>
	 */
	public void glGetPixelMapuiv(GLPixelMap map, MutPointer<Integer> values) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetPixelMapuiv(\{map}, \{values})");
		try {
		    functions.glGetPixelMapuiv().invokeExact(map, values);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were map=\{map}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetPixelMapusv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetPixelMapusv.xhtml">registry.khronos.org</a>
	 */
	public void glGetPixelMapusv(GLPixelMap map, MutPointer<Character> values) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetPixelMapusv(\{map}, \{values})");
		try {
		    functions.glGetPixelMapusv().invokeExact(map, values);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were map=\{map}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetPixelMapxv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetPixelMapxv.xhtml">registry.khronos.org</a>
	 */
	public void glGetPixelMapxv(GLPixelMap map, int size, MutPointer<Integer> values) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetPixelMapxv(\{map}, \{size}, \{values})");
		try {
		    functions.glGetPixelMapxv().invokeExact(map, size, values);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were map=\{map}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetPixelTexGenParameterfvSGIS">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetPixelTexGenParameterfvSGIS.xhtml">registry.khronos.org</a>
	 */
	public void glGetPixelTexGenParameterfvSGIS(GLPixelTexGenParameterNameSGIS pname, MutPointer<Float> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetPixelTexGenParameterfvSGIS(\{pname}, \{params})");
		try {
		    functions.glGetPixelTexGenParameterfvSGIS().invokeExact(pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetPixelTexGenParameterivSGIS">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetPixelTexGenParameterivSGIS.xhtml">registry.khronos.org</a>
	 */
	public void glGetPixelTexGenParameterivSGIS(GLPixelTexGenParameterNameSGIS pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetPixelTexGenParameterivSGIS(\{pname}, \{params})");
		try {
		    functions.glGetPixelTexGenParameterivSGIS().invokeExact(pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetPixelTransformParameterfvEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetPixelTransformParameterfvEXT.xhtml">registry.khronos.org</a>
	 */
	public void glGetPixelTransformParameterfvEXT(GLEnum target, GLEnum pname, MutPointer<Float> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetPixelTransformParameterfvEXT(\{target}, \{pname}, \{params})");
		try {
		    functions.glGetPixelTransformParameterfvEXT().invokeExact(target, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetPixelTransformParameterivEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetPixelTransformParameterivEXT.xhtml">registry.khronos.org</a>
	 */
	public void glGetPixelTransformParameterivEXT(GLEnum target, GLEnum pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetPixelTransformParameterivEXT(\{target}, \{pname}, \{params})");
		try {
		    functions.glGetPixelTransformParameterivEXT().invokeExact(target, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetPointerIndexedvEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetPointerIndexedvEXT.xhtml">registry.khronos.org</a>
	 */
	public void glGetPointerIndexedvEXT(GLEnum target, int index, MutPointer<MutPointer<Void>> data) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetPointerIndexedvEXT(\{target}, \{index}, \{data})");
		try {
		    functions.glGetPointerIndexedvEXT().invokeExact(target, index, data);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetPointeri_vEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetPointeri_vEXT.xhtml">registry.khronos.org</a>
	 */
	public void glGetPointeri_vEXT(GLEnum pname, int index, MutPointer<MutPointer<Void>> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetPointeri_vEXT(\{pname}, \{index}, \{params})");
		try {
		    functions.glGetPointeri_vEXT().invokeExact(pname, index, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetPointerv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetPointerv.xhtml">registry.khronos.org</a>
	 */
	public void glGetPointerv(GLGetPointervPName pname, MutPointer<MutPointer<Void>> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetPointerv(\{pname}, \{params})");
		try {
		    functions.glGetPointerv().invokeExact(pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	public MutPointer<Void> glGetPointer(GLGetPointervPName pname) {
        try (var arena = Arena.ofConfined()) {
            var buf = arena.allocate(GLTypeCache.ML_P());
            glGetPointerv(pname, new MutPointer<>(buf));
			var rvalue = buf.getAtIndex(GLTypeCache.ML_P(), 0);
            return MutPointer.newNullable(rvalue);
        }
    }

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetPointervEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetPointervEXT.xhtml">registry.khronos.org</a>
	 */
	public void glGetPointervEXT(GLGetPointervPName pname, MutPointer<MutPointer<Void>> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetPointervEXT(\{pname}, \{params})");
		try {
		    functions.glGetPointervEXT().invokeExact(pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetPointervKHR">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetPointervKHR.xhtml">registry.khronos.org</a>
	 */
	public void glGetPointervKHR(GLEnum pname, MutPointer<MutPointer<Void>> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetPointervKHR(\{pname}, \{params})");
		try {
		    functions.glGetPointervKHR().invokeExact(pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetPolygonStipple">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetPolygonStipple.xhtml">registry.khronos.org</a>
	 */
	public void glGetPolygonStipple(MutPointer<Byte> mask) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetPolygonStipple(\{mask})");
		try {
		    functions.glGetPolygonStipple().invokeExact(mask);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetProgramBinary">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetProgramBinary.xhtml">registry.khronos.org</a>
	 */
	public void glGetProgramBinary(int program, int bufSize, MutPointer<Integer> length, MutPointer<GLEnum> binaryFormat, MutPointer<Void> binary) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetProgramBinary(\{program}, \{bufSize}, \{length}, \{binaryFormat}, \{binary})");
		try {
		    functions.glGetProgramBinary().invokeExact(program, bufSize, length, binaryFormat, binary);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetProgramBinaryOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetProgramBinaryOES.xhtml">registry.khronos.org</a>
	 */
	public void glGetProgramBinaryOES(int program, int bufSize, MutPointer<Integer> length, MutPointer<GLEnum> binaryFormat, MutPointer<Void> binary) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetProgramBinaryOES(\{program}, \{bufSize}, \{length}, \{binaryFormat}, \{binary})");
		try {
		    functions.glGetProgramBinaryOES().invokeExact(program, bufSize, length, binaryFormat, binary);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetProgramEnvParameterIivNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetProgramEnvParameterIivNV.xhtml">registry.khronos.org</a>
	 */
	public void glGetProgramEnvParameterIivNV(GLProgramTarget target, int index, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetProgramEnvParameterIivNV(\{target}, \{index}, \{params})");
		try {
		    functions.glGetProgramEnvParameterIivNV().invokeExact(target, index, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetProgramEnvParameterIuivNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetProgramEnvParameterIuivNV.xhtml">registry.khronos.org</a>
	 */
	public void glGetProgramEnvParameterIuivNV(GLProgramTarget target, int index, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetProgramEnvParameterIuivNV(\{target}, \{index}, \{params})");
		try {
		    functions.glGetProgramEnvParameterIuivNV().invokeExact(target, index, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetProgramEnvParameterdvARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetProgramEnvParameterdvARB.xhtml">registry.khronos.org</a>
	 */
	public void glGetProgramEnvParameterdvARB(GLProgramTarget target, int index, MutPointer<Double> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetProgramEnvParameterdvARB(\{target}, \{index}, \{params})");
		try {
		    functions.glGetProgramEnvParameterdvARB().invokeExact(target, index, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetProgramEnvParameterfvARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetProgramEnvParameterfvARB.xhtml">registry.khronos.org</a>
	 */
	public void glGetProgramEnvParameterfvARB(GLProgramTarget target, int index, MutPointer<Float> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetProgramEnvParameterfvARB(\{target}, \{index}, \{params})");
		try {
		    functions.glGetProgramEnvParameterfvARB().invokeExact(target, index, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetProgramInfoLog">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetProgramInfoLog.xhtml">registry.khronos.org</a>
	 */
	public void glGetProgramInfoLog(int program, int bufSize, MutPointer<Integer> length, MutPointer<Byte> infoLog) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetProgramInfoLog(\{program}, \{bufSize}, \{length}, \{infoLog})");
		try {
		    functions.glGetProgramInfoLog().invokeExact(program, bufSize, length, infoLog);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetProgramInterfaceiv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetProgramInterfaceiv.xhtml">registry.khronos.org</a>
	 */
	public void glGetProgramInterfaceiv(int program, GLProgramInterface programInterface, GLProgramInterfacePName pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetProgramInterfaceiv(\{program}, \{programInterface}, \{pname}, \{params})");
		try {
		    functions.glGetProgramInterfaceiv().invokeExact(program, programInterface, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were programInterface=\{programInterface}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetProgramLocalParameterIivNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetProgramLocalParameterIivNV.xhtml">registry.khronos.org</a>
	 */
	public void glGetProgramLocalParameterIivNV(GLProgramTarget target, int index, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetProgramLocalParameterIivNV(\{target}, \{index}, \{params})");
		try {
		    functions.glGetProgramLocalParameterIivNV().invokeExact(target, index, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetProgramLocalParameterIuivNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetProgramLocalParameterIuivNV.xhtml">registry.khronos.org</a>
	 */
	public void glGetProgramLocalParameterIuivNV(GLProgramTarget target, int index, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetProgramLocalParameterIuivNV(\{target}, \{index}, \{params})");
		try {
		    functions.glGetProgramLocalParameterIuivNV().invokeExact(target, index, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetProgramLocalParameterdvARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetProgramLocalParameterdvARB.xhtml">registry.khronos.org</a>
	 */
	public void glGetProgramLocalParameterdvARB(GLProgramTarget target, int index, MutPointer<Double> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetProgramLocalParameterdvARB(\{target}, \{index}, \{params})");
		try {
		    functions.glGetProgramLocalParameterdvARB().invokeExact(target, index, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetProgramLocalParameterfvARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetProgramLocalParameterfvARB.xhtml">registry.khronos.org</a>
	 */
	public void glGetProgramLocalParameterfvARB(GLProgramTarget target, int index, MutPointer<Float> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetProgramLocalParameterfvARB(\{target}, \{index}, \{params})");
		try {
		    functions.glGetProgramLocalParameterfvARB().invokeExact(target, index, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetProgramNamedParameterdvNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetProgramNamedParameterdvNV.xhtml">registry.khronos.org</a>
	 */
	public void glGetProgramNamedParameterdvNV(int id, int len, ConstPointer<Byte> name, MutPointer<Double> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetProgramNamedParameterdvNV(\{id}, \{len}, \{name}, \{params})");
		try {
		    functions.glGetProgramNamedParameterdvNV().invokeExact(id, len, name, params);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetProgramNamedParameterfvNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetProgramNamedParameterfvNV.xhtml">registry.khronos.org</a>
	 */
	public void glGetProgramNamedParameterfvNV(int id, int len, ConstPointer<Byte> name, MutPointer<Float> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetProgramNamedParameterfvNV(\{id}, \{len}, \{name}, \{params})");
		try {
		    functions.glGetProgramNamedParameterfvNV().invokeExact(id, len, name, params);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetProgramParameterdvNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetProgramParameterdvNV.xhtml">registry.khronos.org</a>
	 */
	public void glGetProgramParameterdvNV(GLVertexAttribEnumNV target, int index, GLVertexAttribEnumNV pname, MutPointer<Double> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetProgramParameterdvNV(\{target}, \{index}, \{pname}, \{params})");
		try {
		    functions.glGetProgramParameterdvNV().invokeExact(target, index, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetProgramParameterfvNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetProgramParameterfvNV.xhtml">registry.khronos.org</a>
	 */
	public void glGetProgramParameterfvNV(GLVertexAttribEnumNV target, int index, GLVertexAttribEnumNV pname, MutPointer<Float> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetProgramParameterfvNV(\{target}, \{index}, \{pname}, \{params})");
		try {
		    functions.glGetProgramParameterfvNV().invokeExact(target, index, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetProgramPipelineInfoLog">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetProgramPipelineInfoLog.xhtml">registry.khronos.org</a>
	 */
	public void glGetProgramPipelineInfoLog(int pipeline, int bufSize, MutPointer<Integer> length, MutPointer<Byte> infoLog) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetProgramPipelineInfoLog(\{pipeline}, \{bufSize}, \{length}, \{infoLog})");
		try {
		    functions.glGetProgramPipelineInfoLog().invokeExact(pipeline, bufSize, length, infoLog);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetProgramPipelineInfoLogEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetProgramPipelineInfoLogEXT.xhtml">registry.khronos.org</a>
	 */
	public void glGetProgramPipelineInfoLogEXT(int pipeline, int bufSize, MutPointer<Integer> length, MutPointer<Byte> infoLog) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetProgramPipelineInfoLogEXT(\{pipeline}, \{bufSize}, \{length}, \{infoLog})");
		try {
		    functions.glGetProgramPipelineInfoLogEXT().invokeExact(pipeline, bufSize, length, infoLog);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetProgramPipelineiv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetProgramPipelineiv.xhtml">registry.khronos.org</a>
	 */
	public void glGetProgramPipelineiv(int pipeline, GLPipelineParameterName pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetProgramPipelineiv(\{pipeline}, \{pname}, \{params})");
		try {
		    functions.glGetProgramPipelineiv().invokeExact(pipeline, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetProgramPipelineivEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetProgramPipelineivEXT.xhtml">registry.khronos.org</a>
	 */
	public void glGetProgramPipelineivEXT(int pipeline, GLPipelineParameterName pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetProgramPipelineivEXT(\{pipeline}, \{pname}, \{params})");
		try {
		    functions.glGetProgramPipelineivEXT().invokeExact(pipeline, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetProgramResourceIndex">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetProgramResourceIndex.xhtml">registry.khronos.org</a>
	 */
	public int glGetProgramResourceIndex(int program, GLProgramInterface programInterface, ConstPointer<Byte> name) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetProgramResourceIndex(\{program}, \{programInterface}, \{name})");
		try {
		    var r = (int)functions.glGetProgramResourceIndex().invokeExact(program, programInterface, name);
			checkError();
			return r;
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were programInterface=\{programInterface}.");
			}
			throw e;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetProgramResourceLocation">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetProgramResourceLocation.xhtml">registry.khronos.org</a>
	 */
	public int glGetProgramResourceLocation(int program, GLProgramInterface programInterface, ConstPointer<Byte> name) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetProgramResourceLocation(\{program}, \{programInterface}, \{name})");
		try {
		    var r = (int)functions.glGetProgramResourceLocation().invokeExact(program, programInterface, name);
			checkError();
			return r;
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were programInterface=\{programInterface}.");
			}
			throw e;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetProgramResourceLocationIndex">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetProgramResourceLocationIndex.xhtml">registry.khronos.org</a>
	 */
	public int glGetProgramResourceLocationIndex(int program, GLProgramInterface programInterface, ConstPointer<Byte> name) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetProgramResourceLocationIndex(\{program}, \{programInterface}, \{name})");
		try {
		    var r = (int)functions.glGetProgramResourceLocationIndex().invokeExact(program, programInterface, name);
			checkError();
			return r;
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were programInterface=\{programInterface}.");
			}
			throw e;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetProgramResourceLocationIndexEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetProgramResourceLocationIndexEXT.xhtml">registry.khronos.org</a>
	 */
	public int glGetProgramResourceLocationIndexEXT(int program, GLProgramInterface programInterface, ConstPointer<Byte> name) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetProgramResourceLocationIndexEXT(\{program}, \{programInterface}, \{name})");
		try {
		    var r = (int)functions.glGetProgramResourceLocationIndexEXT().invokeExact(program, programInterface, name);
			checkError();
			return r;
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were programInterface=\{programInterface}.");
			}
			throw e;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetProgramResourceName">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetProgramResourceName.xhtml">registry.khronos.org</a>
	 */
	public void glGetProgramResourceName(int program, GLProgramInterface programInterface, int index, int bufSize, MutPointer<Integer> length, MutPointer<Byte> name) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetProgramResourceName(\{program}, \{programInterface}, \{index}, \{bufSize}, \{length}, \{name})");
		try {
		    functions.glGetProgramResourceName().invokeExact(program, programInterface, index, bufSize, length, name);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were programInterface=\{programInterface}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetProgramResourcefvNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetProgramResourcefvNV.xhtml">registry.khronos.org</a>
	 */
	public void glGetProgramResourcefvNV(int program, GLProgramInterface programInterface, int index, int propCount, ConstPointer<GLEnum> props, int count, MutPointer<Integer> length, MutPointer<Float> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetProgramResourcefvNV(\{program}, \{programInterface}, \{index}, \{propCount}, \{props}, \{count}, \{length}, \{params})");
		try {
		    functions.glGetProgramResourcefvNV().invokeExact(program, programInterface, index, propCount, props, count, length, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were programInterface=\{programInterface}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetProgramResourceiv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetProgramResourceiv.xhtml">registry.khronos.org</a>
	 */
	public void glGetProgramResourceiv(int program, GLProgramInterface programInterface, int index, int propCount, ConstPointer<GLProgramResourceProperty> props, int count, MutPointer<Integer> length, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetProgramResourceiv(\{program}, \{programInterface}, \{index}, \{propCount}, \{props}, \{count}, \{length}, \{params})");
		try {
		    functions.glGetProgramResourceiv().invokeExact(program, programInterface, index, propCount, props, count, length, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were programInterface=\{programInterface}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetProgramStageiv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetProgramStageiv.xhtml">registry.khronos.org</a>
	 */
	public void glGetProgramStageiv(int program, GLShaderType shadertype, GLProgramStagePName pname, MutPointer<Integer> values) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetProgramStageiv(\{program}, \{shadertype}, \{pname}, \{values})");
		try {
		    functions.glGetProgramStageiv().invokeExact(program, shadertype, pname, values);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were shadertype=\{shadertype}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetProgramStringARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetProgramStringARB.xhtml">registry.khronos.org</a>
	 */
	public void glGetProgramStringARB(GLProgramTarget target, GLProgramStringProperty pname, MutPointer<Void> string) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetProgramStringARB(\{target}, \{pname}, \{string})");
		try {
		    functions.glGetProgramStringARB().invokeExact(target, pname, string);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetProgramStringNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetProgramStringNV.xhtml">registry.khronos.org</a>
	 */
	public void glGetProgramStringNV(int id, GLVertexAttribEnumNV pname, MutPointer<Byte> program) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetProgramStringNV(\{id}, \{pname}, \{program})");
		try {
		    functions.glGetProgramStringNV().invokeExact(id, pname, program);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetProgramSubroutineParameteruivNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetProgramSubroutineParameteruivNV.xhtml">registry.khronos.org</a>
	 */
	public void glGetProgramSubroutineParameteruivNV(GLEnum target, int index, MutPointer<Integer> param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetProgramSubroutineParameteruivNV(\{target}, \{index}, \{param})");
		try {
		    functions.glGetProgramSubroutineParameteruivNV().invokeExact(target, index, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetProgramiv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetProgramiv.xhtml">registry.khronos.org</a>
	 */
	public void glGetProgramiv(int program, GLProgramPropertyARB pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetProgramiv(\{program}, \{pname}, \{params})");
		try {
		    functions.glGetProgramiv().invokeExact(program, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetProgramivARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetProgramivARB.xhtml">registry.khronos.org</a>
	 */
	public void glGetProgramivARB(GLProgramTarget target, GLProgramPropertyARB pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetProgramivARB(\{target}, \{pname}, \{params})");
		try {
		    functions.glGetProgramivARB().invokeExact(target, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetProgramivNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetProgramivNV.xhtml">registry.khronos.org</a>
	 */
	public void glGetProgramivNV(int id, GLVertexAttribEnumNV pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetProgramivNV(\{id}, \{pname}, \{params})");
		try {
		    functions.glGetProgramivNV().invokeExact(id, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetQueryBufferObjecti64v">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetQueryBufferObjecti64v.xhtml">registry.khronos.org</a>
	 */
	public void glGetQueryBufferObjecti64v(int id, int buffer, GLQueryObjectParameterName pname, long offset) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetQueryBufferObjecti64v(\{id}, \{buffer}, \{pname}, \{offset})");
		try {
		    functions.glGetQueryBufferObjecti64v().invokeExact(id, buffer, pname, offset);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetQueryBufferObjectiv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetQueryBufferObjectiv.xhtml">registry.khronos.org</a>
	 */
	public void glGetQueryBufferObjectiv(int id, int buffer, GLQueryObjectParameterName pname, long offset) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetQueryBufferObjectiv(\{id}, \{buffer}, \{pname}, \{offset})");
		try {
		    functions.glGetQueryBufferObjectiv().invokeExact(id, buffer, pname, offset);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetQueryBufferObjectui64v">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetQueryBufferObjectui64v.xhtml">registry.khronos.org</a>
	 */
	public void glGetQueryBufferObjectui64v(int id, int buffer, GLQueryObjectParameterName pname, long offset) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetQueryBufferObjectui64v(\{id}, \{buffer}, \{pname}, \{offset})");
		try {
		    functions.glGetQueryBufferObjectui64v().invokeExact(id, buffer, pname, offset);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetQueryBufferObjectuiv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetQueryBufferObjectuiv.xhtml">registry.khronos.org</a>
	 */
	public void glGetQueryBufferObjectuiv(int id, int buffer, GLQueryObjectParameterName pname, long offset) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetQueryBufferObjectuiv(\{id}, \{buffer}, \{pname}, \{offset})");
		try {
		    functions.glGetQueryBufferObjectuiv().invokeExact(id, buffer, pname, offset);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetQueryIndexediv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetQueryIndexediv.xhtml">registry.khronos.org</a>
	 */
	public void glGetQueryIndexediv(GLQueryTarget target, int index, GLQueryParameterName pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetQueryIndexediv(\{target}, \{index}, \{pname}, \{params})");
		try {
		    functions.glGetQueryIndexediv().invokeExact(target, index, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetQueryObjecti64v">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetQueryObjecti64v.xhtml">registry.khronos.org</a>
	 */
	public void glGetQueryObjecti64v(int id, GLQueryObjectParameterName pname, MutPointer<Long> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetQueryObjecti64v(\{id}, \{pname}, \{params})");
		try {
		    functions.glGetQueryObjecti64v().invokeExact(id, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetQueryObjecti64vEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetQueryObjecti64vEXT.xhtml">registry.khronos.org</a>
	 */
	public void glGetQueryObjecti64vEXT(int id, GLQueryObjectParameterName pname, MutPointer<Long> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetQueryObjecti64vEXT(\{id}, \{pname}, \{params})");
		try {
		    functions.glGetQueryObjecti64vEXT().invokeExact(id, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetQueryObjectiv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetQueryObjectiv.xhtml">registry.khronos.org</a>
	 */
	public void glGetQueryObjectiv(int id, GLQueryObjectParameterName pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetQueryObjectiv(\{id}, \{pname}, \{params})");
		try {
		    functions.glGetQueryObjectiv().invokeExact(id, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetQueryObjectivARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetQueryObjectivARB.xhtml">registry.khronos.org</a>
	 */
	public void glGetQueryObjectivARB(int id, GLQueryObjectParameterName pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetQueryObjectivARB(\{id}, \{pname}, \{params})");
		try {
		    functions.glGetQueryObjectivARB().invokeExact(id, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetQueryObjectivEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetQueryObjectivEXT.xhtml">registry.khronos.org</a>
	 */
	public void glGetQueryObjectivEXT(int id, GLQueryObjectParameterName pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetQueryObjectivEXT(\{id}, \{pname}, \{params})");
		try {
		    functions.glGetQueryObjectivEXT().invokeExact(id, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetQueryObjectui64v">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetQueryObjectui64v.xhtml">registry.khronos.org</a>
	 */
	public void glGetQueryObjectui64v(int id, GLQueryObjectParameterName pname, MutPointer<Long> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetQueryObjectui64v(\{id}, \{pname}, \{params})");
		try {
		    functions.glGetQueryObjectui64v().invokeExact(id, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetQueryObjectui64vEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetQueryObjectui64vEXT.xhtml">registry.khronos.org</a>
	 */
	public void glGetQueryObjectui64vEXT(int id, GLQueryObjectParameterName pname, MutPointer<Long> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetQueryObjectui64vEXT(\{id}, \{pname}, \{params})");
		try {
		    functions.glGetQueryObjectui64vEXT().invokeExact(id, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetQueryObjectuiv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetQueryObjectuiv.xhtml">registry.khronos.org</a>
	 */
	public void glGetQueryObjectuiv(int id, GLQueryObjectParameterName pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetQueryObjectuiv(\{id}, \{pname}, \{params})");
		try {
		    functions.glGetQueryObjectuiv().invokeExact(id, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetQueryObjectuivARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetQueryObjectuivARB.xhtml">registry.khronos.org</a>
	 */
	public void glGetQueryObjectuivARB(int id, GLQueryObjectParameterName pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetQueryObjectuivARB(\{id}, \{pname}, \{params})");
		try {
		    functions.glGetQueryObjectuivARB().invokeExact(id, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetQueryObjectuivEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetQueryObjectuivEXT.xhtml">registry.khronos.org</a>
	 */
	public void glGetQueryObjectuivEXT(int id, GLQueryObjectParameterName pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetQueryObjectuivEXT(\{id}, \{pname}, \{params})");
		try {
		    functions.glGetQueryObjectuivEXT().invokeExact(id, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetQueryiv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetQueryiv.xhtml">registry.khronos.org</a>
	 */
	public void glGetQueryiv(GLQueryTarget target, GLQueryParameterName pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetQueryiv(\{target}, \{pname}, \{params})");
		try {
		    functions.glGetQueryiv().invokeExact(target, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetQueryivARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetQueryivARB.xhtml">registry.khronos.org</a>
	 */
	public void glGetQueryivARB(GLQueryTarget target, GLQueryParameterName pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetQueryivARB(\{target}, \{pname}, \{params})");
		try {
		    functions.glGetQueryivARB().invokeExact(target, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetQueryivEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetQueryivEXT.xhtml">registry.khronos.org</a>
	 */
	public void glGetQueryivEXT(GLQueryTarget target, GLQueryParameterName pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetQueryivEXT(\{target}, \{pname}, \{params})");
		try {
		    functions.glGetQueryivEXT().invokeExact(target, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetRenderbufferParameteriv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetRenderbufferParameteriv.xhtml">registry.khronos.org</a>
	 */
	public void glGetRenderbufferParameteriv(GLRenderbufferTarget target, GLRenderbufferParameterName pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetRenderbufferParameteriv(\{target}, \{pname}, \{params})");
		try {
		    functions.glGetRenderbufferParameteriv().invokeExact(target, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetRenderbufferParameterivEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetRenderbufferParameterivEXT.xhtml">registry.khronos.org</a>
	 */
	public void glGetRenderbufferParameterivEXT(GLRenderbufferTarget target, GLRenderbufferParameterName pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetRenderbufferParameterivEXT(\{target}, \{pname}, \{params})");
		try {
		    functions.glGetRenderbufferParameterivEXT().invokeExact(target, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetRenderbufferParameterivOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetRenderbufferParameterivOES.xhtml">registry.khronos.org</a>
	 */
	public void glGetRenderbufferParameterivOES(GLRenderbufferTarget target, GLRenderbufferParameterName pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetRenderbufferParameterivOES(\{target}, \{pname}, \{params})");
		try {
		    functions.glGetRenderbufferParameterivOES().invokeExact(target, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetSamplerParameterIiv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetSamplerParameterIiv.xhtml">registry.khronos.org</a>
	 */
	public void glGetSamplerParameterIiv(int sampler, GLSamplerParameterI pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetSamplerParameterIiv(\{sampler}, \{pname}, \{params})");
		try {
		    functions.glGetSamplerParameterIiv().invokeExact(sampler, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetSamplerParameterIivEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetSamplerParameterIivEXT.xhtml">registry.khronos.org</a>
	 */
	public void glGetSamplerParameterIivEXT(int sampler, GLSamplerParameterI pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetSamplerParameterIivEXT(\{sampler}, \{pname}, \{params})");
		try {
		    functions.glGetSamplerParameterIivEXT().invokeExact(sampler, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetSamplerParameterIivOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetSamplerParameterIivOES.xhtml">registry.khronos.org</a>
	 */
	public void glGetSamplerParameterIivOES(int sampler, GLSamplerParameterI pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetSamplerParameterIivOES(\{sampler}, \{pname}, \{params})");
		try {
		    functions.glGetSamplerParameterIivOES().invokeExact(sampler, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetSamplerParameterIuiv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetSamplerParameterIuiv.xhtml">registry.khronos.org</a>
	 */
	public void glGetSamplerParameterIuiv(int sampler, GLSamplerParameterI pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetSamplerParameterIuiv(\{sampler}, \{pname}, \{params})");
		try {
		    functions.glGetSamplerParameterIuiv().invokeExact(sampler, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetSamplerParameterIuivEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetSamplerParameterIuivEXT.xhtml">registry.khronos.org</a>
	 */
	public void glGetSamplerParameterIuivEXT(int sampler, GLSamplerParameterI pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetSamplerParameterIuivEXT(\{sampler}, \{pname}, \{params})");
		try {
		    functions.glGetSamplerParameterIuivEXT().invokeExact(sampler, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetSamplerParameterIuivOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetSamplerParameterIuivOES.xhtml">registry.khronos.org</a>
	 */
	public void glGetSamplerParameterIuivOES(int sampler, GLSamplerParameterI pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetSamplerParameterIuivOES(\{sampler}, \{pname}, \{params})");
		try {
		    functions.glGetSamplerParameterIuivOES().invokeExact(sampler, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetSamplerParameterfv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetSamplerParameterfv.xhtml">registry.khronos.org</a>
	 */
	public void glGetSamplerParameterfv(int sampler, GLSamplerParameterF pname, MutPointer<Float> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetSamplerParameterfv(\{sampler}, \{pname}, \{params})");
		try {
		    functions.glGetSamplerParameterfv().invokeExact(sampler, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetSamplerParameteriv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetSamplerParameteriv.xhtml">registry.khronos.org</a>
	 */
	public void glGetSamplerParameteriv(int sampler, GLSamplerParameterI pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetSamplerParameteriv(\{sampler}, \{pname}, \{params})");
		try {
		    functions.glGetSamplerParameteriv().invokeExact(sampler, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetSemaphoreParameterivNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetSemaphoreParameterivNV.xhtml">registry.khronos.org</a>
	 */
	public void glGetSemaphoreParameterivNV(int semaphore, GLSemaphoreParameterName pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetSemaphoreParameterivNV(\{semaphore}, \{pname}, \{params})");
		try {
		    functions.glGetSemaphoreParameterivNV().invokeExact(semaphore, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetSemaphoreParameterui64vEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetSemaphoreParameterui64vEXT.xhtml">registry.khronos.org</a>
	 */
	public void glGetSemaphoreParameterui64vEXT(int semaphore, GLSemaphoreParameterName pname, MutPointer<Long> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetSemaphoreParameterui64vEXT(\{semaphore}, \{pname}, \{params})");
		try {
		    functions.glGetSemaphoreParameterui64vEXT().invokeExact(semaphore, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetSeparableFilter">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetSeparableFilter.xhtml">registry.khronos.org</a>
	 */
	public void glGetSeparableFilter(GLSeparableTargetEXT target, GLPixelFormat format, GLPixelType type, MutPointer<Void> row, MutPointer<Void> column, MutPointer<Void> span) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetSeparableFilter(\{target}, \{format}, \{type}, \{row}, \{column}, \{span})");
		try {
		    functions.glGetSeparableFilter().invokeExact(target, format, type, row, column, span);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, format=\{format}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetSeparableFilterEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetSeparableFilterEXT.xhtml">registry.khronos.org</a>
	 */
	public void glGetSeparableFilterEXT(GLSeparableTargetEXT target, GLPixelFormat format, GLPixelType type, MutPointer<Void> row, MutPointer<Void> column, MutPointer<Void> span) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetSeparableFilterEXT(\{target}, \{format}, \{type}, \{row}, \{column}, \{span})");
		try {
		    functions.glGetSeparableFilterEXT().invokeExact(target, format, type, row, column, span);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, format=\{format}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetShaderInfoLog">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetShaderInfoLog.xhtml">registry.khronos.org</a>
	 */
	public void glGetShaderInfoLog(int shader, int bufSize, MutPointer<Integer> length, MutPointer<Byte> infoLog) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetShaderInfoLog(\{shader}, \{bufSize}, \{length}, \{infoLog})");
		try {
		    functions.glGetShaderInfoLog().invokeExact(shader, bufSize, length, infoLog);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetShaderPrecisionFormat">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetShaderPrecisionFormat.xhtml">registry.khronos.org</a>
	 */
	public void glGetShaderPrecisionFormat(GLShaderType shadertype, GLPrecisionType precisiontype, MutPointer<Integer> range, MutPointer<Integer> precision) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetShaderPrecisionFormat(\{shadertype}, \{precisiontype}, \{range}, \{precision})");
		try {
		    functions.glGetShaderPrecisionFormat().invokeExact(shadertype, precisiontype, range, precision);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were shadertype=\{shadertype}, precisiontype=\{precisiontype}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetShaderSource">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetShaderSource.xhtml">registry.khronos.org</a>
	 */
	public void glGetShaderSource(int shader, int bufSize, MutPointer<Integer> length, MutPointer<Byte> source) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetShaderSource(\{shader}, \{bufSize}, \{length}, \{source})");
		try {
		    functions.glGetShaderSource().invokeExact(shader, bufSize, length, source);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetShaderSourceARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetShaderSourceARB.xhtml">registry.khronos.org</a>
	 */
	public void glGetShaderSourceARB(int obj, int maxLength, MutPointer<Integer> length, MutPointer<Byte> source) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetShaderSourceARB(\{obj}, \{maxLength}, \{length}, \{source})");
		try {
		    functions.glGetShaderSourceARB().invokeExact(obj, maxLength, length, source);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetShaderiv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetShaderiv.xhtml">registry.khronos.org</a>
	 */
	public void glGetShaderiv(int shader, GLShaderParameterName pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetShaderiv(\{shader}, \{pname}, \{params})");
		try {
		    functions.glGetShaderiv().invokeExact(shader, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetShadingRateImagePaletteNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetShadingRateImagePaletteNV.xhtml">registry.khronos.org</a>
	 */
	public void glGetShadingRateImagePaletteNV(int viewport, int entry, MutPointer<GLEnum> rate) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetShadingRateImagePaletteNV(\{viewport}, \{entry}, \{rate})");
		try {
		    functions.glGetShadingRateImagePaletteNV().invokeExact(viewport, entry, rate);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetShadingRateSampleLocationivNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetShadingRateSampleLocationivNV.xhtml">registry.khronos.org</a>
	 */
	public void glGetShadingRateSampleLocationivNV(GLEnum rate, int samples, int index, MutPointer<Integer> location) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetShadingRateSampleLocationivNV(\{rate}, \{samples}, \{index}, \{location})");
		try {
		    functions.glGetShadingRateSampleLocationivNV().invokeExact(rate, samples, index, location);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were rate=\{rate}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetSharpenTexFuncSGIS">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetSharpenTexFuncSGIS.xhtml">registry.khronos.org</a>
	 */
	public void glGetSharpenTexFuncSGIS(GLTextureTarget target, MutPointer<Float> points) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetSharpenTexFuncSGIS(\{target}, \{points})");
		try {
		    functions.glGetSharpenTexFuncSGIS().invokeExact(target, points);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetStageIndexNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetStageIndexNV.xhtml">registry.khronos.org</a>
	 */
	public char glGetStageIndexNV(GLShaderType shadertype) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetStageIndexNV(\{shadertype})");
		try {
		    var r = (char)functions.glGetStageIndexNV().invokeExact(shadertype);
			checkError();
			return r;
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were shadertype=\{shadertype}.");
			}
			throw e;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetString">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetString.xhtml">registry.khronos.org</a>
	 */
	public ConstPointer<Byte> glGetString(GLStringName name) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetString(\{name})");
		try {
		    var r = (ConstPointer<Byte>)functions.glGetString().invokeExact(name);
			checkError();
			return r;
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were name=\{name}.");
			}
			throw e;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetStringi">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetStringi.xhtml">registry.khronos.org</a>
	 */
	public ConstPointer<Byte> glGetStringi(GLStringName name, int index) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetStringi(\{name}, \{index})");
		try {
		    var r = (ConstPointer<Byte>)functions.glGetStringi().invokeExact(name, index);
			checkError();
			return r;
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were name=\{name}.");
			}
			throw e;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetSubroutineIndex">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetSubroutineIndex.xhtml">registry.khronos.org</a>
	 */
	public int glGetSubroutineIndex(int program, GLShaderType shadertype, ConstPointer<Byte> name) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetSubroutineIndex(\{program}, \{shadertype}, \{name})");
		try {
		    var r = (int)functions.glGetSubroutineIndex().invokeExact(program, shadertype, name);
			checkError();
			return r;
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were shadertype=\{shadertype}.");
			}
			throw e;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetSubroutineUniformLocation">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetSubroutineUniformLocation.xhtml">registry.khronos.org</a>
	 */
	public int glGetSubroutineUniformLocation(int program, GLShaderType shadertype, ConstPointer<Byte> name) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetSubroutineUniformLocation(\{program}, \{shadertype}, \{name})");
		try {
		    var r = (int)functions.glGetSubroutineUniformLocation().invokeExact(program, shadertype, name);
			checkError();
			return r;
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were shadertype=\{shadertype}.");
			}
			throw e;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetSynciv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetSynciv.xhtml">registry.khronos.org</a>
	 */
	public void glGetSynciv(ConstPointer<Void> sync, GLSyncParameterName pname, int count, MutPointer<Integer> length, MutPointer<Integer> values) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetSynciv(\{sync}, \{pname}, \{count}, \{length}, \{values})");
		try {
		    functions.glGetSynciv().invokeExact(sync, pname, count, length, values);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetSyncivAPPLE">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetSyncivAPPLE.xhtml">registry.khronos.org</a>
	 */
	public void glGetSyncivAPPLE(ConstPointer<Void> sync, GLSyncParameterName pname, int count, MutPointer<Integer> length, MutPointer<Integer> values) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetSyncivAPPLE(\{sync}, \{pname}, \{count}, \{length}, \{values})");
		try {
		    functions.glGetSyncivAPPLE().invokeExact(sync, pname, count, length, values);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetTexBumpParameterfvATI">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetTexBumpParameterfvATI.xhtml">registry.khronos.org</a>
	 */
	public void glGetTexBumpParameterfvATI(GLGetTexBumpParameterATI pname, MutPointer<Float> param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetTexBumpParameterfvATI(\{pname}, \{param})");
		try {
		    functions.glGetTexBumpParameterfvATI().invokeExact(pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetTexBumpParameterivATI">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetTexBumpParameterivATI.xhtml">registry.khronos.org</a>
	 */
	public void glGetTexBumpParameterivATI(GLGetTexBumpParameterATI pname, MutPointer<Integer> param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetTexBumpParameterivATI(\{pname}, \{param})");
		try {
		    functions.glGetTexBumpParameterivATI().invokeExact(pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetTexEnvfv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetTexEnvfv.xhtml">registry.khronos.org</a>
	 */
	public void glGetTexEnvfv(GLTextureEnvTarget target, GLTextureEnvParameter pname, MutPointer<Float> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetTexEnvfv(\{target}, \{pname}, \{params})");
		try {
		    functions.glGetTexEnvfv().invokeExact(target, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetTexEnviv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetTexEnviv.xhtml">registry.khronos.org</a>
	 */
	public void glGetTexEnviv(GLTextureEnvTarget target, GLTextureEnvParameter pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetTexEnviv(\{target}, \{pname}, \{params})");
		try {
		    functions.glGetTexEnviv().invokeExact(target, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetTexEnvxv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetTexEnvxv.xhtml">registry.khronos.org</a>
	 */
	public void glGetTexEnvxv(GLTextureEnvTarget target, GLTextureEnvParameter pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetTexEnvxv(\{target}, \{pname}, \{params})");
		try {
		    functions.glGetTexEnvxv().invokeExact(target, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetTexEnvxvOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetTexEnvxvOES.xhtml">registry.khronos.org</a>
	 */
	public void glGetTexEnvxvOES(GLTextureEnvTarget target, GLTextureEnvParameter pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetTexEnvxvOES(\{target}, \{pname}, \{params})");
		try {
		    functions.glGetTexEnvxvOES().invokeExact(target, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetTexFilterFuncSGIS">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetTexFilterFuncSGIS.xhtml">registry.khronos.org</a>
	 */
	public void glGetTexFilterFuncSGIS(GLTextureTarget target, GLTextureFilterSGIS filter, MutPointer<Float> weights) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetTexFilterFuncSGIS(\{target}, \{filter}, \{weights})");
		try {
		    functions.glGetTexFilterFuncSGIS().invokeExact(target, filter, weights);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, filter=\{filter}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetTexGendv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetTexGendv.xhtml">registry.khronos.org</a>
	 */
	public void glGetTexGendv(GLTextureCoordName coord, GLTextureGenParameter pname, MutPointer<Double> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetTexGendv(\{coord}, \{pname}, \{params})");
		try {
		    functions.glGetTexGendv().invokeExact(coord, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were coord=\{coord}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetTexGenfv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetTexGenfv.xhtml">registry.khronos.org</a>
	 */
	public void glGetTexGenfv(GLTextureCoordName coord, GLTextureGenParameter pname, MutPointer<Float> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetTexGenfv(\{coord}, \{pname}, \{params})");
		try {
		    functions.glGetTexGenfv().invokeExact(coord, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were coord=\{coord}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetTexGenfvOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetTexGenfvOES.xhtml">registry.khronos.org</a>
	 */
	public void glGetTexGenfvOES(GLTextureCoordName coord, GLTextureGenParameter pname, MutPointer<Float> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetTexGenfvOES(\{coord}, \{pname}, \{params})");
		try {
		    functions.glGetTexGenfvOES().invokeExact(coord, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were coord=\{coord}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetTexGeniv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetTexGeniv.xhtml">registry.khronos.org</a>
	 */
	public void glGetTexGeniv(GLTextureCoordName coord, GLTextureGenParameter pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetTexGeniv(\{coord}, \{pname}, \{params})");
		try {
		    functions.glGetTexGeniv().invokeExact(coord, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were coord=\{coord}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetTexGenivOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetTexGenivOES.xhtml">registry.khronos.org</a>
	 */
	public void glGetTexGenivOES(GLTextureCoordName coord, GLTextureGenParameter pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetTexGenivOES(\{coord}, \{pname}, \{params})");
		try {
		    functions.glGetTexGenivOES().invokeExact(coord, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were coord=\{coord}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetTexGenxvOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetTexGenxvOES.xhtml">registry.khronos.org</a>
	 */
	public void glGetTexGenxvOES(GLTextureCoordName coord, GLTextureGenParameter pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetTexGenxvOES(\{coord}, \{pname}, \{params})");
		try {
		    functions.glGetTexGenxvOES().invokeExact(coord, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were coord=\{coord}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetTexImage">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetTexImage.xhtml">registry.khronos.org</a>
	 */
	public void glGetTexImage(GLTextureTarget target, int level, GLPixelFormat format, GLPixelType type, MutPointer<Void> pixels) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetTexImage(\{target}, \{level}, \{format}, \{type}, \{pixels})");
		try {
		    functions.glGetTexImage().invokeExact(target, level, format, type, pixels);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, format=\{format}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetTexLevelParameterfv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetTexLevelParameterfv.xhtml">registry.khronos.org</a>
	 */
	public void glGetTexLevelParameterfv(GLTextureTarget target, int level, GLGetTextureParameter pname, MutPointer<Float> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetTexLevelParameterfv(\{target}, \{level}, \{pname}, \{params})");
		try {
		    functions.glGetTexLevelParameterfv().invokeExact(target, level, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetTexLevelParameteriv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetTexLevelParameteriv.xhtml">registry.khronos.org</a>
	 */
	public void glGetTexLevelParameteriv(GLTextureTarget target, int level, GLGetTextureParameter pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetTexLevelParameteriv(\{target}, \{level}, \{pname}, \{params})");
		try {
		    functions.glGetTexLevelParameteriv().invokeExact(target, level, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetTexLevelParameterxvOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetTexLevelParameterxvOES.xhtml">registry.khronos.org</a>
	 */
	public void glGetTexLevelParameterxvOES(GLTextureTarget target, int level, GLGetTextureParameter pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetTexLevelParameterxvOES(\{target}, \{level}, \{pname}, \{params})");
		try {
		    functions.glGetTexLevelParameterxvOES().invokeExact(target, level, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetTexParameterIiv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetTexParameterIiv.xhtml">registry.khronos.org</a>
	 */
	public void glGetTexParameterIiv(GLTextureTarget target, GLGetTextureParameter pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetTexParameterIiv(\{target}, \{pname}, \{params})");
		try {
		    functions.glGetTexParameterIiv().invokeExact(target, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetTexParameterIivEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetTexParameterIivEXT.xhtml">registry.khronos.org</a>
	 */
	public void glGetTexParameterIivEXT(GLTextureTarget target, GLGetTextureParameter pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetTexParameterIivEXT(\{target}, \{pname}, \{params})");
		try {
		    functions.glGetTexParameterIivEXT().invokeExact(target, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetTexParameterIivOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetTexParameterIivOES.xhtml">registry.khronos.org</a>
	 */
	public void glGetTexParameterIivOES(GLTextureTarget target, GLGetTextureParameter pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetTexParameterIivOES(\{target}, \{pname}, \{params})");
		try {
		    functions.glGetTexParameterIivOES().invokeExact(target, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetTexParameterIuiv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetTexParameterIuiv.xhtml">registry.khronos.org</a>
	 */
	public void glGetTexParameterIuiv(GLTextureTarget target, GLGetTextureParameter pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetTexParameterIuiv(\{target}, \{pname}, \{params})");
		try {
		    functions.glGetTexParameterIuiv().invokeExact(target, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetTexParameterIuivEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetTexParameterIuivEXT.xhtml">registry.khronos.org</a>
	 */
	public void glGetTexParameterIuivEXT(GLTextureTarget target, GLGetTextureParameter pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetTexParameterIuivEXT(\{target}, \{pname}, \{params})");
		try {
		    functions.glGetTexParameterIuivEXT().invokeExact(target, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetTexParameterIuivOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetTexParameterIuivOES.xhtml">registry.khronos.org</a>
	 */
	public void glGetTexParameterIuivOES(GLTextureTarget target, GLGetTextureParameter pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetTexParameterIuivOES(\{target}, \{pname}, \{params})");
		try {
		    functions.glGetTexParameterIuivOES().invokeExact(target, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetTexParameterPointervAPPLE">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetTexParameterPointervAPPLE.xhtml">registry.khronos.org</a>
	 */
	public void glGetTexParameterPointervAPPLE(GLEnum target, GLEnum pname, MutPointer<MutPointer<Void>> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetTexParameterPointervAPPLE(\{target}, \{pname}, \{params})");
		try {
		    functions.glGetTexParameterPointervAPPLE().invokeExact(target, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetTexParameterfv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetTexParameterfv.xhtml">registry.khronos.org</a>
	 */
	public void glGetTexParameterfv(GLTextureTarget target, GLGetTextureParameter pname, MutPointer<Float> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetTexParameterfv(\{target}, \{pname}, \{params})");
		try {
		    functions.glGetTexParameterfv().invokeExact(target, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetTexParameteriv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetTexParameteriv.xhtml">registry.khronos.org</a>
	 */
	public void glGetTexParameteriv(GLTextureTarget target, GLGetTextureParameter pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetTexParameteriv(\{target}, \{pname}, \{params})");
		try {
		    functions.glGetTexParameteriv().invokeExact(target, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetTexParameterxv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetTexParameterxv.xhtml">registry.khronos.org</a>
	 */
	public void glGetTexParameterxv(GLTextureTarget target, GLGetTextureParameter pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetTexParameterxv(\{target}, \{pname}, \{params})");
		try {
		    functions.glGetTexParameterxv().invokeExact(target, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetTexParameterxvOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetTexParameterxvOES.xhtml">registry.khronos.org</a>
	 */
	public void glGetTexParameterxvOES(GLTextureTarget target, GLGetTextureParameter pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetTexParameterxvOES(\{target}, \{pname}, \{params})");
		try {
		    functions.glGetTexParameterxvOES().invokeExact(target, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetTextureHandleARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetTextureHandleARB.xhtml">registry.khronos.org</a>
	 */
	public long glGetTextureHandleARB(int texture) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetTextureHandleARB(\{texture})");
		try {
		    var r = (long)functions.glGetTextureHandleARB().invokeExact(texture);
			checkError();
			return r;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetTextureHandleIMG">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetTextureHandleIMG.xhtml">registry.khronos.org</a>
	 */
	public long glGetTextureHandleIMG(int texture) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetTextureHandleIMG(\{texture})");
		try {
		    var r = (long)functions.glGetTextureHandleIMG().invokeExact(texture);
			checkError();
			return r;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetTextureHandleNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetTextureHandleNV.xhtml">registry.khronos.org</a>
	 */
	public long glGetTextureHandleNV(int texture) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetTextureHandleNV(\{texture})");
		try {
		    var r = (long)functions.glGetTextureHandleNV().invokeExact(texture);
			checkError();
			return r;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetTextureImage">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetTextureImage.xhtml">registry.khronos.org</a>
	 */
	public void glGetTextureImage(int texture, int level, GLPixelFormat format, GLPixelType type, int bufSize, MutPointer<Void> pixels) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetTextureImage(\{texture}, \{level}, \{format}, \{type}, \{bufSize}, \{pixels})");
		try {
		    functions.glGetTextureImage().invokeExact(texture, level, format, type, bufSize, pixels);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were format=\{format}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetTextureImageEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetTextureImageEXT.xhtml">registry.khronos.org</a>
	 */
	public void glGetTextureImageEXT(int texture, GLTextureTarget target, int level, GLPixelFormat format, GLPixelType type, MutPointer<Void> pixels) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetTextureImageEXT(\{texture}, \{target}, \{level}, \{format}, \{type}, \{pixels})");
		try {
		    functions.glGetTextureImageEXT().invokeExact(texture, target, level, format, type, pixels);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, format=\{format}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetTextureLevelParameterfv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetTextureLevelParameterfv.xhtml">registry.khronos.org</a>
	 */
	public void glGetTextureLevelParameterfv(int texture, int level, GLGetTextureParameter pname, MutPointer<Float> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetTextureLevelParameterfv(\{texture}, \{level}, \{pname}, \{params})");
		try {
		    functions.glGetTextureLevelParameterfv().invokeExact(texture, level, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetTextureLevelParameterfvEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetTextureLevelParameterfvEXT.xhtml">registry.khronos.org</a>
	 */
	public void glGetTextureLevelParameterfvEXT(int texture, GLTextureTarget target, int level, GLGetTextureParameter pname, MutPointer<Float> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetTextureLevelParameterfvEXT(\{texture}, \{target}, \{level}, \{pname}, \{params})");
		try {
		    functions.glGetTextureLevelParameterfvEXT().invokeExact(texture, target, level, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetTextureLevelParameteriv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetTextureLevelParameteriv.xhtml">registry.khronos.org</a>
	 */
	public void glGetTextureLevelParameteriv(int texture, int level, GLGetTextureParameter pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetTextureLevelParameteriv(\{texture}, \{level}, \{pname}, \{params})");
		try {
		    functions.glGetTextureLevelParameteriv().invokeExact(texture, level, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetTextureLevelParameterivEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetTextureLevelParameterivEXT.xhtml">registry.khronos.org</a>
	 */
	public void glGetTextureLevelParameterivEXT(int texture, GLTextureTarget target, int level, GLGetTextureParameter pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetTextureLevelParameterivEXT(\{texture}, \{target}, \{level}, \{pname}, \{params})");
		try {
		    functions.glGetTextureLevelParameterivEXT().invokeExact(texture, target, level, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetTextureParameterIiv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetTextureParameterIiv.xhtml">registry.khronos.org</a>
	 */
	public void glGetTextureParameterIiv(int texture, GLGetTextureParameter pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetTextureParameterIiv(\{texture}, \{pname}, \{params})");
		try {
		    functions.glGetTextureParameterIiv().invokeExact(texture, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetTextureParameterIivEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetTextureParameterIivEXT.xhtml">registry.khronos.org</a>
	 */
	public void glGetTextureParameterIivEXT(int texture, GLTextureTarget target, GLGetTextureParameter pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetTextureParameterIivEXT(\{texture}, \{target}, \{pname}, \{params})");
		try {
		    functions.glGetTextureParameterIivEXT().invokeExact(texture, target, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetTextureParameterIuiv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetTextureParameterIuiv.xhtml">registry.khronos.org</a>
	 */
	public void glGetTextureParameterIuiv(int texture, GLGetTextureParameter pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetTextureParameterIuiv(\{texture}, \{pname}, \{params})");
		try {
		    functions.glGetTextureParameterIuiv().invokeExact(texture, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetTextureParameterIuivEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetTextureParameterIuivEXT.xhtml">registry.khronos.org</a>
	 */
	public void glGetTextureParameterIuivEXT(int texture, GLTextureTarget target, GLGetTextureParameter pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetTextureParameterIuivEXT(\{texture}, \{target}, \{pname}, \{params})");
		try {
		    functions.glGetTextureParameterIuivEXT().invokeExact(texture, target, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetTextureParameterfv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetTextureParameterfv.xhtml">registry.khronos.org</a>
	 */
	public void glGetTextureParameterfv(int texture, GLGetTextureParameter pname, MutPointer<Float> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetTextureParameterfv(\{texture}, \{pname}, \{params})");
		try {
		    functions.glGetTextureParameterfv().invokeExact(texture, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetTextureParameterfvEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetTextureParameterfvEXT.xhtml">registry.khronos.org</a>
	 */
	public void glGetTextureParameterfvEXT(int texture, GLTextureTarget target, GLGetTextureParameter pname, MutPointer<Float> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetTextureParameterfvEXT(\{texture}, \{target}, \{pname}, \{params})");
		try {
		    functions.glGetTextureParameterfvEXT().invokeExact(texture, target, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetTextureParameteriv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetTextureParameteriv.xhtml">registry.khronos.org</a>
	 */
	public void glGetTextureParameteriv(int texture, GLGetTextureParameter pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetTextureParameteriv(\{texture}, \{pname}, \{params})");
		try {
		    functions.glGetTextureParameteriv().invokeExact(texture, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetTextureParameterivEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetTextureParameterivEXT.xhtml">registry.khronos.org</a>
	 */
	public void glGetTextureParameterivEXT(int texture, GLTextureTarget target, GLGetTextureParameter pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetTextureParameterivEXT(\{texture}, \{target}, \{pname}, \{params})");
		try {
		    functions.glGetTextureParameterivEXT().invokeExact(texture, target, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetTextureSamplerHandleARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetTextureSamplerHandleARB.xhtml">registry.khronos.org</a>
	 */
	public long glGetTextureSamplerHandleARB(int texture, int sampler) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetTextureSamplerHandleARB(\{texture}, \{sampler})");
		try {
		    var r = (long)functions.glGetTextureSamplerHandleARB().invokeExact(texture, sampler);
			checkError();
			return r;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetTextureSamplerHandleIMG">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetTextureSamplerHandleIMG.xhtml">registry.khronos.org</a>
	 */
	public long glGetTextureSamplerHandleIMG(int texture, int sampler) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetTextureSamplerHandleIMG(\{texture}, \{sampler})");
		try {
		    var r = (long)functions.glGetTextureSamplerHandleIMG().invokeExact(texture, sampler);
			checkError();
			return r;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetTextureSamplerHandleNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetTextureSamplerHandleNV.xhtml">registry.khronos.org</a>
	 */
	public long glGetTextureSamplerHandleNV(int texture, int sampler) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetTextureSamplerHandleNV(\{texture}, \{sampler})");
		try {
		    var r = (long)functions.glGetTextureSamplerHandleNV().invokeExact(texture, sampler);
			checkError();
			return r;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetTextureSubImage">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetTextureSubImage.xhtml">registry.khronos.org</a>
	 */
	public void glGetTextureSubImage(int texture, int level, int xoffset, int yoffset, int zoffset, int width, int height, int depth, GLPixelFormat format, GLPixelType type, int bufSize, MutPointer<Void> pixels) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetTextureSubImage(\{texture}, \{level}, \{xoffset}, \{yoffset}, \{zoffset}, \{width}, \{height}, \{depth}, \{format}, \{type}, \{bufSize}, \{pixels})");
		try {
		    functions.glGetTextureSubImage().invokeExact(texture, level, xoffset, yoffset, zoffset, width, height, depth, format, type, bufSize, pixels);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were format=\{format}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetTrackMatrixivNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetTrackMatrixivNV.xhtml">registry.khronos.org</a>
	 */
	public void glGetTrackMatrixivNV(GLVertexAttribEnumNV target, int address, GLVertexAttribEnumNV pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetTrackMatrixivNV(\{target}, \{address}, \{pname}, \{params})");
		try {
		    functions.glGetTrackMatrixivNV().invokeExact(target, address, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetTransformFeedbackVarying">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetTransformFeedbackVarying.xhtml">registry.khronos.org</a>
	 */
	public void glGetTransformFeedbackVarying(int program, int index, int bufSize, MutPointer<Integer> length, MutPointer<Integer> size, MutPointer<GLAttributeType> type, MutPointer<Byte> name) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetTransformFeedbackVarying(\{program}, \{index}, \{bufSize}, \{length}, \{size}, \{type}, \{name})");
		try {
		    functions.glGetTransformFeedbackVarying().invokeExact(program, index, bufSize, length, size, type, name);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetTransformFeedbackVaryingEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetTransformFeedbackVaryingEXT.xhtml">registry.khronos.org</a>
	 */
	public void glGetTransformFeedbackVaryingEXT(int program, int index, int bufSize, MutPointer<Integer> length, MutPointer<Integer> size, MutPointer<GLAttributeType> type, MutPointer<Byte> name) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetTransformFeedbackVaryingEXT(\{program}, \{index}, \{bufSize}, \{length}, \{size}, \{type}, \{name})");
		try {
		    functions.glGetTransformFeedbackVaryingEXT().invokeExact(program, index, bufSize, length, size, type, name);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetTransformFeedbackVaryingNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetTransformFeedbackVaryingNV.xhtml">registry.khronos.org</a>
	 */
	public void glGetTransformFeedbackVaryingNV(int program, int index, MutPointer<Integer> location) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetTransformFeedbackVaryingNV(\{program}, \{index}, \{location})");
		try {
		    functions.glGetTransformFeedbackVaryingNV().invokeExact(program, index, location);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetTransformFeedbacki64_v">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetTransformFeedbacki64_v.xhtml">registry.khronos.org</a>
	 */
	public void glGetTransformFeedbacki64_v(int xfb, GLTransformFeedbackPName pname, int index, MutPointer<Long> param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetTransformFeedbacki64_v(\{xfb}, \{pname}, \{index}, \{param})");
		try {
		    functions.glGetTransformFeedbacki64_v().invokeExact(xfb, pname, index, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetTransformFeedbacki_v">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetTransformFeedbacki_v.xhtml">registry.khronos.org</a>
	 */
	public void glGetTransformFeedbacki_v(int xfb, GLTransformFeedbackPName pname, int index, MutPointer<Integer> param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetTransformFeedbacki_v(\{xfb}, \{pname}, \{index}, \{param})");
		try {
		    functions.glGetTransformFeedbacki_v().invokeExact(xfb, pname, index, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetTransformFeedbackiv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetTransformFeedbackiv.xhtml">registry.khronos.org</a>
	 */
	public void glGetTransformFeedbackiv(int xfb, GLTransformFeedbackPName pname, MutPointer<Integer> param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetTransformFeedbackiv(\{xfb}, \{pname}, \{param})");
		try {
		    functions.glGetTransformFeedbackiv().invokeExact(xfb, pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetTranslatedShaderSourceANGLE">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetTranslatedShaderSourceANGLE.xhtml">registry.khronos.org</a>
	 */
	public void glGetTranslatedShaderSourceANGLE(int shader, int bufSize, MutPointer<Integer> length, MutPointer<Byte> source) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetTranslatedShaderSourceANGLE(\{shader}, \{bufSize}, \{length}, \{source})");
		try {
		    functions.glGetTranslatedShaderSourceANGLE().invokeExact(shader, bufSize, length, source);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetUniformBlockIndex">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetUniformBlockIndex.xhtml">registry.khronos.org</a>
	 */
	public int glGetUniformBlockIndex(int program, ConstPointer<Byte> uniformBlockName) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetUniformBlockIndex(\{program}, \{uniformBlockName})");
		try {
		    var r = (int)functions.glGetUniformBlockIndex().invokeExact(program, uniformBlockName);
			checkError();
			return r;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetUniformBufferSizeEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetUniformBufferSizeEXT.xhtml">registry.khronos.org</a>
	 */
	public int glGetUniformBufferSizeEXT(int program, int location) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetUniformBufferSizeEXT(\{program}, \{location})");
		try {
		    var r = (int)functions.glGetUniformBufferSizeEXT().invokeExact(program, location);
			checkError();
			return r;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetUniformIndices">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetUniformIndices.xhtml">registry.khronos.org</a>
	 */
	public void glGetUniformIndices(int program, int uniformCount, ConstPointer<ConstPointer<Byte>> uniformNames, MutPointer<Integer> uniformIndices) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetUniformIndices(\{program}, \{uniformCount}, \{uniformNames}, \{uniformIndices})");
		try {
		    functions.glGetUniformIndices().invokeExact(program, uniformCount, uniformNames, uniformIndices);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetUniformLocation">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetUniformLocation.xhtml">registry.khronos.org</a>
	 */
	public int glGetUniformLocation(int program, ConstPointer<Byte> name) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetUniformLocation(\{program}, \{name})");
		try {
		    var r = (int)functions.glGetUniformLocation().invokeExact(program, name);
			checkError();
			return r;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetUniformLocationARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetUniformLocationARB.xhtml">registry.khronos.org</a>
	 */
	public int glGetUniformLocationARB(int programObj, ConstPointer<Byte> name) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetUniformLocationARB(\{programObj}, \{name})");
		try {
		    var r = (int)functions.glGetUniformLocationARB().invokeExact(programObj, name);
			checkError();
			return r;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetUniformOffsetEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetUniformOffsetEXT.xhtml">registry.khronos.org</a>
	 */
	public long glGetUniformOffsetEXT(int program, int location) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetUniformOffsetEXT(\{program}, \{location})");
		try {
		    var r = (long)functions.glGetUniformOffsetEXT().invokeExact(program, location);
			checkError();
			return r;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetUniformSubroutineuiv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetUniformSubroutineuiv.xhtml">registry.khronos.org</a>
	 */
	public void glGetUniformSubroutineuiv(GLShaderType shadertype, int location, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetUniformSubroutineuiv(\{shadertype}, \{location}, \{params})");
		try {
		    functions.glGetUniformSubroutineuiv().invokeExact(shadertype, location, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were shadertype=\{shadertype}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetUniformdv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetUniformdv.xhtml">registry.khronos.org</a>
	 */
	public void glGetUniformdv(int program, int location, MutPointer<Double> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetUniformdv(\{program}, \{location}, \{params})");
		try {
		    functions.glGetUniformdv().invokeExact(program, location, params);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetUniformfv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetUniformfv.xhtml">registry.khronos.org</a>
	 */
	public void glGetUniformfv(int program, int location, MutPointer<Float> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetUniformfv(\{program}, \{location}, \{params})");
		try {
		    functions.glGetUniformfv().invokeExact(program, location, params);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetUniformfvARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetUniformfvARB.xhtml">registry.khronos.org</a>
	 */
	public void glGetUniformfvARB(int programObj, int location, MutPointer<Float> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetUniformfvARB(\{programObj}, \{location}, \{params})");
		try {
		    functions.glGetUniformfvARB().invokeExact(programObj, location, params);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetUniformi64vARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetUniformi64vARB.xhtml">registry.khronos.org</a>
	 */
	public void glGetUniformi64vARB(int program, int location, MutPointer<Long> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetUniformi64vARB(\{program}, \{location}, \{params})");
		try {
		    functions.glGetUniformi64vARB().invokeExact(program, location, params);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetUniformi64vNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetUniformi64vNV.xhtml">registry.khronos.org</a>
	 */
	public void glGetUniformi64vNV(int program, int location, MutPointer<Long> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetUniformi64vNV(\{program}, \{location}, \{params})");
		try {
		    functions.glGetUniformi64vNV().invokeExact(program, location, params);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetUniformiv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetUniformiv.xhtml">registry.khronos.org</a>
	 */
	public void glGetUniformiv(int program, int location, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetUniformiv(\{program}, \{location}, \{params})");
		try {
		    functions.glGetUniformiv().invokeExact(program, location, params);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetUniformivARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetUniformivARB.xhtml">registry.khronos.org</a>
	 */
	public void glGetUniformivARB(int programObj, int location, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetUniformivARB(\{programObj}, \{location}, \{params})");
		try {
		    functions.glGetUniformivARB().invokeExact(programObj, location, params);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetUniformui64vARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetUniformui64vARB.xhtml">registry.khronos.org</a>
	 */
	public void glGetUniformui64vARB(int program, int location, MutPointer<Long> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetUniformui64vARB(\{program}, \{location}, \{params})");
		try {
		    functions.glGetUniformui64vARB().invokeExact(program, location, params);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetUniformui64vNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetUniformui64vNV.xhtml">registry.khronos.org</a>
	 */
	public void glGetUniformui64vNV(int program, int location, MutPointer<Long> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetUniformui64vNV(\{program}, \{location}, \{params})");
		try {
		    functions.glGetUniformui64vNV().invokeExact(program, location, params);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetUniformuiv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetUniformuiv.xhtml">registry.khronos.org</a>
	 */
	public void glGetUniformuiv(int program, int location, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetUniformuiv(\{program}, \{location}, \{params})");
		try {
		    functions.glGetUniformuiv().invokeExact(program, location, params);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetUniformuivEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetUniformuivEXT.xhtml">registry.khronos.org</a>
	 */
	public void glGetUniformuivEXT(int program, int location, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetUniformuivEXT(\{program}, \{location}, \{params})");
		try {
		    functions.glGetUniformuivEXT().invokeExact(program, location, params);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetUnsignedBytevEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetUnsignedBytevEXT.xhtml">registry.khronos.org</a>
	 */
	public void glGetUnsignedBytevEXT(GLGetPName pname, MutPointer<Byte> data) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetUnsignedBytevEXT(\{pname}, \{data})");
		try {
		    functions.glGetUnsignedBytevEXT().invokeExact(pname, data);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetUnsignedBytei_vEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetUnsignedBytei_vEXT.xhtml">registry.khronos.org</a>
	 */
	public void glGetUnsignedBytei_vEXT(GLEnum target, int index, MutPointer<Byte> data) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetUnsignedBytei_vEXT(\{target}, \{index}, \{data})");
		try {
		    functions.glGetUnsignedBytei_vEXT().invokeExact(target, index, data);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetVariantArrayObjectfvATI">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetVariantArrayObjectfvATI.xhtml">registry.khronos.org</a>
	 */
	public void glGetVariantArrayObjectfvATI(int id, GLArrayObjectPNameATI pname, MutPointer<Float> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetVariantArrayObjectfvATI(\{id}, \{pname}, \{params})");
		try {
		    functions.glGetVariantArrayObjectfvATI().invokeExact(id, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetVariantArrayObjectivATI">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetVariantArrayObjectivATI.xhtml">registry.khronos.org</a>
	 */
	public void glGetVariantArrayObjectivATI(int id, GLArrayObjectPNameATI pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetVariantArrayObjectivATI(\{id}, \{pname}, \{params})");
		try {
		    functions.glGetVariantArrayObjectivATI().invokeExact(id, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetVariantBooleanvEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetVariantBooleanvEXT.xhtml">registry.khronos.org</a>
	 */
	public void glGetVariantBooleanvEXT(int id, GLGetVariantValueEXT value, MutPointer<Boolean> data) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetVariantBooleanvEXT(\{id}, \{value}, \{data})");
		try {
		    functions.glGetVariantBooleanvEXT().invokeExact(id, value, data);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were value=\{value}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetVariantFloatvEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetVariantFloatvEXT.xhtml">registry.khronos.org</a>
	 */
	public void glGetVariantFloatvEXT(int id, GLGetVariantValueEXT value, MutPointer<Float> data) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetVariantFloatvEXT(\{id}, \{value}, \{data})");
		try {
		    functions.glGetVariantFloatvEXT().invokeExact(id, value, data);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were value=\{value}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetVariantIntegervEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetVariantIntegervEXT.xhtml">registry.khronos.org</a>
	 */
	public void glGetVariantIntegervEXT(int id, GLGetVariantValueEXT value, MutPointer<Integer> data) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetVariantIntegervEXT(\{id}, \{value}, \{data})");
		try {
		    functions.glGetVariantIntegervEXT().invokeExact(id, value, data);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were value=\{value}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetVariantPointervEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetVariantPointervEXT.xhtml">registry.khronos.org</a>
	 */
	public void glGetVariantPointervEXT(int id, GLGetVariantValueEXT value, MutPointer<MutPointer<Void>> data) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetVariantPointervEXT(\{id}, \{value}, \{data})");
		try {
		    functions.glGetVariantPointervEXT().invokeExact(id, value, data);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were value=\{value}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetVaryingLocationNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetVaryingLocationNV.xhtml">registry.khronos.org</a>
	 */
	public int glGetVaryingLocationNV(int program, ConstPointer<Byte> name) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetVaryingLocationNV(\{program}, \{name})");
		try {
		    var r = (int)functions.glGetVaryingLocationNV().invokeExact(program, name);
			checkError();
			return r;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetVertexArrayIndexed64iv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetVertexArrayIndexed64iv.xhtml">registry.khronos.org</a>
	 */
	public void glGetVertexArrayIndexed64iv(int vaobj, int index, GLVertexArrayPName pname, MutPointer<Long> param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetVertexArrayIndexed64iv(\{vaobj}, \{index}, \{pname}, \{param})");
		try {
		    functions.glGetVertexArrayIndexed64iv().invokeExact(vaobj, index, pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetVertexArrayIndexediv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetVertexArrayIndexediv.xhtml">registry.khronos.org</a>
	 */
	public void glGetVertexArrayIndexediv(int vaobj, int index, GLVertexArrayPName pname, MutPointer<Integer> param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetVertexArrayIndexediv(\{vaobj}, \{index}, \{pname}, \{param})");
		try {
		    functions.glGetVertexArrayIndexediv().invokeExact(vaobj, index, pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetVertexArrayIntegeri_vEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetVertexArrayIntegeri_vEXT.xhtml">registry.khronos.org</a>
	 */
	public void glGetVertexArrayIntegeri_vEXT(int vaobj, int index, GLVertexArrayPName pname, MutPointer<Integer> param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetVertexArrayIntegeri_vEXT(\{vaobj}, \{index}, \{pname}, \{param})");
		try {
		    functions.glGetVertexArrayIntegeri_vEXT().invokeExact(vaobj, index, pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetVertexArrayIntegervEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetVertexArrayIntegervEXT.xhtml">registry.khronos.org</a>
	 */
	public void glGetVertexArrayIntegervEXT(int vaobj, GLVertexArrayPName pname, MutPointer<Integer> param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetVertexArrayIntegervEXT(\{vaobj}, \{pname}, \{param})");
		try {
		    functions.glGetVertexArrayIntegervEXT().invokeExact(vaobj, pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetVertexArrayPointeri_vEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetVertexArrayPointeri_vEXT.xhtml">registry.khronos.org</a>
	 */
	public void glGetVertexArrayPointeri_vEXT(int vaobj, int index, GLVertexArrayPName pname, MutPointer<MutPointer<Void>> param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetVertexArrayPointeri_vEXT(\{vaobj}, \{index}, \{pname}, \{param})");
		try {
		    functions.glGetVertexArrayPointeri_vEXT().invokeExact(vaobj, index, pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetVertexArrayPointervEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetVertexArrayPointervEXT.xhtml">registry.khronos.org</a>
	 */
	public void glGetVertexArrayPointervEXT(int vaobj, GLVertexArrayPName pname, MutPointer<MutPointer<Void>> param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetVertexArrayPointervEXT(\{vaobj}, \{pname}, \{param})");
		try {
		    functions.glGetVertexArrayPointervEXT().invokeExact(vaobj, pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetVertexArrayiv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetVertexArrayiv.xhtml">registry.khronos.org</a>
	 */
	public void glGetVertexArrayiv(int vaobj, GLVertexArrayPName pname, MutPointer<Integer> param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetVertexArrayiv(\{vaobj}, \{pname}, \{param})");
		try {
		    functions.glGetVertexArrayiv().invokeExact(vaobj, pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetVertexAttribArrayObjectfvATI">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetVertexAttribArrayObjectfvATI.xhtml">registry.khronos.org</a>
	 */
	public void glGetVertexAttribArrayObjectfvATI(int index, GLArrayObjectPNameATI pname, MutPointer<Float> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetVertexAttribArrayObjectfvATI(\{index}, \{pname}, \{params})");
		try {
		    functions.glGetVertexAttribArrayObjectfvATI().invokeExact(index, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetVertexAttribArrayObjectivATI">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetVertexAttribArrayObjectivATI.xhtml">registry.khronos.org</a>
	 */
	public void glGetVertexAttribArrayObjectivATI(int index, GLArrayObjectPNameATI pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetVertexAttribArrayObjectivATI(\{index}, \{pname}, \{params})");
		try {
		    functions.glGetVertexAttribArrayObjectivATI().invokeExact(index, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetVertexAttribIiv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetVertexAttribIiv.xhtml">registry.khronos.org</a>
	 */
	public void glGetVertexAttribIiv(int index, GLVertexAttribEnum pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetVertexAttribIiv(\{index}, \{pname}, \{params})");
		try {
		    functions.glGetVertexAttribIiv().invokeExact(index, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetVertexAttribIivEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetVertexAttribIivEXT.xhtml">registry.khronos.org</a>
	 */
	public void glGetVertexAttribIivEXT(int index, GLVertexAttribEnum pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetVertexAttribIivEXT(\{index}, \{pname}, \{params})");
		try {
		    functions.glGetVertexAttribIivEXT().invokeExact(index, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetVertexAttribIuiv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetVertexAttribIuiv.xhtml">registry.khronos.org</a>
	 */
	public void glGetVertexAttribIuiv(int index, GLVertexAttribEnum pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetVertexAttribIuiv(\{index}, \{pname}, \{params})");
		try {
		    functions.glGetVertexAttribIuiv().invokeExact(index, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetVertexAttribIuivEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetVertexAttribIuivEXT.xhtml">registry.khronos.org</a>
	 */
	public void glGetVertexAttribIuivEXT(int index, GLVertexAttribEnum pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetVertexAttribIuivEXT(\{index}, \{pname}, \{params})");
		try {
		    functions.glGetVertexAttribIuivEXT().invokeExact(index, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetVertexAttribLdv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetVertexAttribLdv.xhtml">registry.khronos.org</a>
	 */
	public void glGetVertexAttribLdv(int index, GLVertexAttribEnum pname, MutPointer<Double> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetVertexAttribLdv(\{index}, \{pname}, \{params})");
		try {
		    functions.glGetVertexAttribLdv().invokeExact(index, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetVertexAttribLdvEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetVertexAttribLdvEXT.xhtml">registry.khronos.org</a>
	 */
	public void glGetVertexAttribLdvEXT(int index, GLVertexAttribEnum pname, MutPointer<Double> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetVertexAttribLdvEXT(\{index}, \{pname}, \{params})");
		try {
		    functions.glGetVertexAttribLdvEXT().invokeExact(index, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetVertexAttribLi64vNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetVertexAttribLi64vNV.xhtml">registry.khronos.org</a>
	 */
	public void glGetVertexAttribLi64vNV(int index, GLVertexAttribEnum pname, MutPointer<Long> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetVertexAttribLi64vNV(\{index}, \{pname}, \{params})");
		try {
		    functions.glGetVertexAttribLi64vNV().invokeExact(index, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetVertexAttribLui64vARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetVertexAttribLui64vARB.xhtml">registry.khronos.org</a>
	 */
	public void glGetVertexAttribLui64vARB(int index, GLVertexAttribEnum pname, MutPointer<Long> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetVertexAttribLui64vARB(\{index}, \{pname}, \{params})");
		try {
		    functions.glGetVertexAttribLui64vARB().invokeExact(index, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetVertexAttribLui64vNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetVertexAttribLui64vNV.xhtml">registry.khronos.org</a>
	 */
	public void glGetVertexAttribLui64vNV(int index, GLVertexAttribEnum pname, MutPointer<Long> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetVertexAttribLui64vNV(\{index}, \{pname}, \{params})");
		try {
		    functions.glGetVertexAttribLui64vNV().invokeExact(index, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetVertexAttribPointerv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetVertexAttribPointerv.xhtml">registry.khronos.org</a>
	 */
	public void glGetVertexAttribPointerv(int index, GLVertexAttribPointerPropertyARB pname, MutPointer<MutPointer<Void>> pointer) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetVertexAttribPointerv(\{index}, \{pname}, \{pointer})");
		try {
		    functions.glGetVertexAttribPointerv().invokeExact(index, pname, pointer);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetVertexAttribPointervARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetVertexAttribPointervARB.xhtml">registry.khronos.org</a>
	 */
	public void glGetVertexAttribPointervARB(int index, GLVertexAttribPointerPropertyARB pname, MutPointer<MutPointer<Void>> pointer) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetVertexAttribPointervARB(\{index}, \{pname}, \{pointer})");
		try {
		    functions.glGetVertexAttribPointervARB().invokeExact(index, pname, pointer);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetVertexAttribPointervNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetVertexAttribPointervNV.xhtml">registry.khronos.org</a>
	 */
	public void glGetVertexAttribPointervNV(int index, GLVertexAttribEnumNV pname, MutPointer<MutPointer<Void>> pointer) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetVertexAttribPointervNV(\{index}, \{pname}, \{pointer})");
		try {
		    functions.glGetVertexAttribPointervNV().invokeExact(index, pname, pointer);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetVertexAttribdv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetVertexAttribdv.xhtml">registry.khronos.org</a>
	 */
	public void glGetVertexAttribdv(int index, GLVertexAttribPropertyARB pname, MutPointer<Double> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetVertexAttribdv(\{index}, \{pname}, \{params})");
		try {
		    functions.glGetVertexAttribdv().invokeExact(index, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetVertexAttribdvARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetVertexAttribdvARB.xhtml">registry.khronos.org</a>
	 */
	public void glGetVertexAttribdvARB(int index, GLVertexAttribPropertyARB pname, MutPointer<Double> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetVertexAttribdvARB(\{index}, \{pname}, \{params})");
		try {
		    functions.glGetVertexAttribdvARB().invokeExact(index, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetVertexAttribdvNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetVertexAttribdvNV.xhtml">registry.khronos.org</a>
	 */
	public void glGetVertexAttribdvNV(int index, GLVertexAttribEnumNV pname, MutPointer<Double> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetVertexAttribdvNV(\{index}, \{pname}, \{params})");
		try {
		    functions.glGetVertexAttribdvNV().invokeExact(index, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetVertexAttribfv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetVertexAttribfv.xhtml">registry.khronos.org</a>
	 */
	public void glGetVertexAttribfv(int index, GLVertexAttribPropertyARB pname, MutPointer<Float> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetVertexAttribfv(\{index}, \{pname}, \{params})");
		try {
		    functions.glGetVertexAttribfv().invokeExact(index, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetVertexAttribfvARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetVertexAttribfvARB.xhtml">registry.khronos.org</a>
	 */
	public void glGetVertexAttribfvARB(int index, GLVertexAttribPropertyARB pname, MutPointer<Float> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetVertexAttribfvARB(\{index}, \{pname}, \{params})");
		try {
		    functions.glGetVertexAttribfvARB().invokeExact(index, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetVertexAttribfvNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetVertexAttribfvNV.xhtml">registry.khronos.org</a>
	 */
	public void glGetVertexAttribfvNV(int index, GLVertexAttribEnumNV pname, MutPointer<Float> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetVertexAttribfvNV(\{index}, \{pname}, \{params})");
		try {
		    functions.glGetVertexAttribfvNV().invokeExact(index, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetVertexAttribiv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetVertexAttribiv.xhtml">registry.khronos.org</a>
	 */
	public void glGetVertexAttribiv(int index, GLVertexAttribPropertyARB pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetVertexAttribiv(\{index}, \{pname}, \{params})");
		try {
		    functions.glGetVertexAttribiv().invokeExact(index, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetVertexAttribivARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetVertexAttribivARB.xhtml">registry.khronos.org</a>
	 */
	public void glGetVertexAttribivARB(int index, GLVertexAttribPropertyARB pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetVertexAttribivARB(\{index}, \{pname}, \{params})");
		try {
		    functions.glGetVertexAttribivARB().invokeExact(index, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetVertexAttribivNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetVertexAttribivNV.xhtml">registry.khronos.org</a>
	 */
	public void glGetVertexAttribivNV(int index, GLVertexAttribEnumNV pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetVertexAttribivNV(\{index}, \{pname}, \{params})");
		try {
		    functions.glGetVertexAttribivNV().invokeExact(index, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetVideoCaptureStreamdvNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetVideoCaptureStreamdvNV.xhtml">registry.khronos.org</a>
	 */
	public void glGetVideoCaptureStreamdvNV(int video_capture_slot, int stream, GLEnum pname, MutPointer<Double> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetVideoCaptureStreamdvNV(\{video_capture_slot}, \{stream}, \{pname}, \{params})");
		try {
		    functions.glGetVideoCaptureStreamdvNV().invokeExact(video_capture_slot, stream, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetVideoCaptureStreamfvNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetVideoCaptureStreamfvNV.xhtml">registry.khronos.org</a>
	 */
	public void glGetVideoCaptureStreamfvNV(int video_capture_slot, int stream, GLEnum pname, MutPointer<Float> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetVideoCaptureStreamfvNV(\{video_capture_slot}, \{stream}, \{pname}, \{params})");
		try {
		    functions.glGetVideoCaptureStreamfvNV().invokeExact(video_capture_slot, stream, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetVideoCaptureStreamivNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetVideoCaptureStreamivNV.xhtml">registry.khronos.org</a>
	 */
	public void glGetVideoCaptureStreamivNV(int video_capture_slot, int stream, GLEnum pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetVideoCaptureStreamivNV(\{video_capture_slot}, \{stream}, \{pname}, \{params})");
		try {
		    functions.glGetVideoCaptureStreamivNV().invokeExact(video_capture_slot, stream, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetVideoCaptureivNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetVideoCaptureivNV.xhtml">registry.khronos.org</a>
	 */
	public void glGetVideoCaptureivNV(int video_capture_slot, GLEnum pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetVideoCaptureivNV(\{video_capture_slot}, \{pname}, \{params})");
		try {
		    functions.glGetVideoCaptureivNV().invokeExact(video_capture_slot, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetVideoi64vNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetVideoi64vNV.xhtml">registry.khronos.org</a>
	 */
	public void glGetVideoi64vNV(int video_slot, GLEnum pname, MutPointer<Long> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetVideoi64vNV(\{video_slot}, \{pname}, \{params})");
		try {
		    functions.glGetVideoi64vNV().invokeExact(video_slot, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetVideoivNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetVideoivNV.xhtml">registry.khronos.org</a>
	 */
	public void glGetVideoivNV(int video_slot, GLEnum pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetVideoivNV(\{video_slot}, \{pname}, \{params})");
		try {
		    functions.glGetVideoivNV().invokeExact(video_slot, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetVideoui64vNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetVideoui64vNV.xhtml">registry.khronos.org</a>
	 */
	public void glGetVideoui64vNV(int video_slot, GLEnum pname, MutPointer<Long> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetVideoui64vNV(\{video_slot}, \{pname}, \{params})");
		try {
		    functions.glGetVideoui64vNV().invokeExact(video_slot, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetVideouivNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetVideouivNV.xhtml">registry.khronos.org</a>
	 */
	public void glGetVideouivNV(int video_slot, GLEnum pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetVideouivNV(\{video_slot}, \{pname}, \{params})");
		try {
		    functions.glGetVideouivNV().invokeExact(video_slot, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetnColorTable">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetnColorTable.xhtml">registry.khronos.org</a>
	 */
	public void glGetnColorTable(GLColorTableTarget target, GLPixelFormat format, GLPixelType type, int bufSize, MutPointer<Void> table) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetnColorTable(\{target}, \{format}, \{type}, \{bufSize}, \{table})");
		try {
		    functions.glGetnColorTable().invokeExact(target, format, type, bufSize, table);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, format=\{format}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetnColorTableARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetnColorTableARB.xhtml">registry.khronos.org</a>
	 */
	public void glGetnColorTableARB(GLColorTableTarget target, GLPixelFormat format, GLPixelType type, int bufSize, MutPointer<Void> table) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetnColorTableARB(\{target}, \{format}, \{type}, \{bufSize}, \{table})");
		try {
		    functions.glGetnColorTableARB().invokeExact(target, format, type, bufSize, table);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, format=\{format}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetnCompressedTexImage">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetnCompressedTexImage.xhtml">registry.khronos.org</a>
	 */
	public void glGetnCompressedTexImage(GLTextureTarget target, int lod, int bufSize, MutPointer<Void> pixels) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetnCompressedTexImage(\{target}, \{lod}, \{bufSize}, \{pixels})");
		try {
		    functions.glGetnCompressedTexImage().invokeExact(target, lod, bufSize, pixels);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetnCompressedTexImageARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetnCompressedTexImageARB.xhtml">registry.khronos.org</a>
	 */
	public void glGetnCompressedTexImageARB(GLTextureTarget target, int lod, int bufSize, MutPointer<Void> img) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetnCompressedTexImageARB(\{target}, \{lod}, \{bufSize}, \{img})");
		try {
		    functions.glGetnCompressedTexImageARB().invokeExact(target, lod, bufSize, img);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetnConvolutionFilter">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetnConvolutionFilter.xhtml">registry.khronos.org</a>
	 */
	public void glGetnConvolutionFilter(GLConvolutionTarget target, GLPixelFormat format, GLPixelType type, int bufSize, MutPointer<Void> image) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetnConvolutionFilter(\{target}, \{format}, \{type}, \{bufSize}, \{image})");
		try {
		    functions.glGetnConvolutionFilter().invokeExact(target, format, type, bufSize, image);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, format=\{format}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetnConvolutionFilterARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetnConvolutionFilterARB.xhtml">registry.khronos.org</a>
	 */
	public void glGetnConvolutionFilterARB(GLConvolutionTarget target, GLPixelFormat format, GLPixelType type, int bufSize, MutPointer<Void> image) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetnConvolutionFilterARB(\{target}, \{format}, \{type}, \{bufSize}, \{image})");
		try {
		    functions.glGetnConvolutionFilterARB().invokeExact(target, format, type, bufSize, image);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, format=\{format}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetnHistogram">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetnHistogram.xhtml">registry.khronos.org</a>
	 */
	public void glGetnHistogram(GLHistogramTarget target, boolean reset, GLPixelFormat format, GLPixelType type, int bufSize, MutPointer<Void> values) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetnHistogram(\{target}, \{reset}, \{format}, \{type}, \{bufSize}, \{values})");
		try {
		    functions.glGetnHistogram().invokeExact(target, reset, format, type, bufSize, values);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, format=\{format}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetnHistogramARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetnHistogramARB.xhtml">registry.khronos.org</a>
	 */
	public void glGetnHistogramARB(GLHistogramTargetEXT target, boolean reset, GLPixelFormat format, GLPixelType type, int bufSize, MutPointer<Void> values) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetnHistogramARB(\{target}, \{reset}, \{format}, \{type}, \{bufSize}, \{values})");
		try {
		    functions.glGetnHistogramARB().invokeExact(target, reset, format, type, bufSize, values);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, format=\{format}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetnMapdv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetnMapdv.xhtml">registry.khronos.org</a>
	 */
	public void glGetnMapdv(GLMapTarget target, GLMapQuery query, int bufSize, MutPointer<Double> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetnMapdv(\{target}, \{query}, \{bufSize}, \{v})");
		try {
		    functions.glGetnMapdv().invokeExact(target, query, bufSize, v);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, query=\{query}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetnMapdvARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetnMapdvARB.xhtml">registry.khronos.org</a>
	 */
	public void glGetnMapdvARB(GLMapTarget target, GLMapQuery query, int bufSize, MutPointer<Double> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetnMapdvARB(\{target}, \{query}, \{bufSize}, \{v})");
		try {
		    functions.glGetnMapdvARB().invokeExact(target, query, bufSize, v);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, query=\{query}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetnMapfv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetnMapfv.xhtml">registry.khronos.org</a>
	 */
	public void glGetnMapfv(GLMapTarget target, GLMapQuery query, int bufSize, MutPointer<Float> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetnMapfv(\{target}, \{query}, \{bufSize}, \{v})");
		try {
		    functions.glGetnMapfv().invokeExact(target, query, bufSize, v);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, query=\{query}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetnMapfvARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetnMapfvARB.xhtml">registry.khronos.org</a>
	 */
	public void glGetnMapfvARB(GLMapTarget target, GLMapQuery query, int bufSize, MutPointer<Float> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetnMapfvARB(\{target}, \{query}, \{bufSize}, \{v})");
		try {
		    functions.glGetnMapfvARB().invokeExact(target, query, bufSize, v);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, query=\{query}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetnMapiv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetnMapiv.xhtml">registry.khronos.org</a>
	 */
	public void glGetnMapiv(GLMapTarget target, GLMapQuery query, int bufSize, MutPointer<Integer> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetnMapiv(\{target}, \{query}, \{bufSize}, \{v})");
		try {
		    functions.glGetnMapiv().invokeExact(target, query, bufSize, v);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, query=\{query}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetnMapivARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetnMapivARB.xhtml">registry.khronos.org</a>
	 */
	public void glGetnMapivARB(GLMapTarget target, GLMapQuery query, int bufSize, MutPointer<Integer> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetnMapivARB(\{target}, \{query}, \{bufSize}, \{v})");
		try {
		    functions.glGetnMapivARB().invokeExact(target, query, bufSize, v);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, query=\{query}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetnMinmax">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetnMinmax.xhtml">registry.khronos.org</a>
	 */
	public void glGetnMinmax(GLMinmaxTarget target, boolean reset, GLPixelFormat format, GLPixelType type, int bufSize, MutPointer<Void> values) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetnMinmax(\{target}, \{reset}, \{format}, \{type}, \{bufSize}, \{values})");
		try {
		    functions.glGetnMinmax().invokeExact(target, reset, format, type, bufSize, values);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, format=\{format}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetnMinmaxARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetnMinmaxARB.xhtml">registry.khronos.org</a>
	 */
	public void glGetnMinmaxARB(GLMinmaxTargetEXT target, boolean reset, GLPixelFormat format, GLPixelType type, int bufSize, MutPointer<Void> values) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetnMinmaxARB(\{target}, \{reset}, \{format}, \{type}, \{bufSize}, \{values})");
		try {
		    functions.glGetnMinmaxARB().invokeExact(target, reset, format, type, bufSize, values);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, format=\{format}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetnPixelMapfv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetnPixelMapfv.xhtml">registry.khronos.org</a>
	 */
	public void glGetnPixelMapfv(GLPixelMap map, int bufSize, MutPointer<Float> values) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetnPixelMapfv(\{map}, \{bufSize}, \{values})");
		try {
		    functions.glGetnPixelMapfv().invokeExact(map, bufSize, values);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were map=\{map}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetnPixelMapfvARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetnPixelMapfvARB.xhtml">registry.khronos.org</a>
	 */
	public void glGetnPixelMapfvARB(GLPixelMap map, int bufSize, MutPointer<Float> values) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetnPixelMapfvARB(\{map}, \{bufSize}, \{values})");
		try {
		    functions.glGetnPixelMapfvARB().invokeExact(map, bufSize, values);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were map=\{map}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetnPixelMapuiv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetnPixelMapuiv.xhtml">registry.khronos.org</a>
	 */
	public void glGetnPixelMapuiv(GLPixelMap map, int bufSize, MutPointer<Integer> values) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetnPixelMapuiv(\{map}, \{bufSize}, \{values})");
		try {
		    functions.glGetnPixelMapuiv().invokeExact(map, bufSize, values);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were map=\{map}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetnPixelMapuivARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetnPixelMapuivARB.xhtml">registry.khronos.org</a>
	 */
	public void glGetnPixelMapuivARB(GLPixelMap map, int bufSize, MutPointer<Integer> values) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetnPixelMapuivARB(\{map}, \{bufSize}, \{values})");
		try {
		    functions.glGetnPixelMapuivARB().invokeExact(map, bufSize, values);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were map=\{map}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetnPixelMapusv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetnPixelMapusv.xhtml">registry.khronos.org</a>
	 */
	public void glGetnPixelMapusv(GLPixelMap map, int bufSize, MutPointer<Character> values) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetnPixelMapusv(\{map}, \{bufSize}, \{values})");
		try {
		    functions.glGetnPixelMapusv().invokeExact(map, bufSize, values);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were map=\{map}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetnPixelMapusvARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetnPixelMapusvARB.xhtml">registry.khronos.org</a>
	 */
	public void glGetnPixelMapusvARB(GLPixelMap map, int bufSize, MutPointer<Character> values) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetnPixelMapusvARB(\{map}, \{bufSize}, \{values})");
		try {
		    functions.glGetnPixelMapusvARB().invokeExact(map, bufSize, values);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were map=\{map}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetnPolygonStipple">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetnPolygonStipple.xhtml">registry.khronos.org</a>
	 */
	public void glGetnPolygonStipple(int bufSize, MutPointer<Byte> pattern) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetnPolygonStipple(\{bufSize}, \{pattern})");
		try {
		    functions.glGetnPolygonStipple().invokeExact(bufSize, pattern);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetnPolygonStippleARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetnPolygonStippleARB.xhtml">registry.khronos.org</a>
	 */
	public void glGetnPolygonStippleARB(int bufSize, MutPointer<Byte> pattern) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetnPolygonStippleARB(\{bufSize}, \{pattern})");
		try {
		    functions.glGetnPolygonStippleARB().invokeExact(bufSize, pattern);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetnSeparableFilter">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetnSeparableFilter.xhtml">registry.khronos.org</a>
	 */
	public void glGetnSeparableFilter(GLSeparableTarget target, GLPixelFormat format, GLPixelType type, int rowBufSize, MutPointer<Void> row, int columnBufSize, MutPointer<Void> column, MutPointer<Void> span) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetnSeparableFilter(\{target}, \{format}, \{type}, \{rowBufSize}, \{row}, \{columnBufSize}, \{column}, \{span})");
		try {
		    functions.glGetnSeparableFilter().invokeExact(target, format, type, rowBufSize, row, columnBufSize, column, span);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, format=\{format}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetnSeparableFilterARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetnSeparableFilterARB.xhtml">registry.khronos.org</a>
	 */
	public void glGetnSeparableFilterARB(GLSeparableTargetEXT target, GLPixelFormat format, GLPixelType type, int rowBufSize, MutPointer<Void> row, int columnBufSize, MutPointer<Void> column, MutPointer<Void> span) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetnSeparableFilterARB(\{target}, \{format}, \{type}, \{rowBufSize}, \{row}, \{columnBufSize}, \{column}, \{span})");
		try {
		    functions.glGetnSeparableFilterARB().invokeExact(target, format, type, rowBufSize, row, columnBufSize, column, span);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, format=\{format}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetnTexImage">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetnTexImage.xhtml">registry.khronos.org</a>
	 */
	public void glGetnTexImage(GLTextureTarget target, int level, GLPixelFormat format, GLPixelType type, int bufSize, MutPointer<Void> pixels) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetnTexImage(\{target}, \{level}, \{format}, \{type}, \{bufSize}, \{pixels})");
		try {
		    functions.glGetnTexImage().invokeExact(target, level, format, type, bufSize, pixels);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, format=\{format}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetnTexImageARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetnTexImageARB.xhtml">registry.khronos.org</a>
	 */
	public void glGetnTexImageARB(GLTextureTarget target, int level, GLPixelFormat format, GLPixelType type, int bufSize, MutPointer<Void> img) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetnTexImageARB(\{target}, \{level}, \{format}, \{type}, \{bufSize}, \{img})");
		try {
		    functions.glGetnTexImageARB().invokeExact(target, level, format, type, bufSize, img);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, format=\{format}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetnUniformdv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetnUniformdv.xhtml">registry.khronos.org</a>
	 */
	public void glGetnUniformdv(int program, int location, int bufSize, MutPointer<Double> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetnUniformdv(\{program}, \{location}, \{bufSize}, \{params})");
		try {
		    functions.glGetnUniformdv().invokeExact(program, location, bufSize, params);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetnUniformdvARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetnUniformdvARB.xhtml">registry.khronos.org</a>
	 */
	public void glGetnUniformdvARB(int program, int location, int bufSize, MutPointer<Double> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetnUniformdvARB(\{program}, \{location}, \{bufSize}, \{params})");
		try {
		    functions.glGetnUniformdvARB().invokeExact(program, location, bufSize, params);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetnUniformfv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetnUniformfv.xhtml">registry.khronos.org</a>
	 */
	public void glGetnUniformfv(int program, int location, int bufSize, MutPointer<Float> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetnUniformfv(\{program}, \{location}, \{bufSize}, \{params})");
		try {
		    functions.glGetnUniformfv().invokeExact(program, location, bufSize, params);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetnUniformfvARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetnUniformfvARB.xhtml">registry.khronos.org</a>
	 */
	public void glGetnUniformfvARB(int program, int location, int bufSize, MutPointer<Float> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetnUniformfvARB(\{program}, \{location}, \{bufSize}, \{params})");
		try {
		    functions.glGetnUniformfvARB().invokeExact(program, location, bufSize, params);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetnUniformfvEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetnUniformfvEXT.xhtml">registry.khronos.org</a>
	 */
	public void glGetnUniformfvEXT(int program, int location, int bufSize, MutPointer<Float> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetnUniformfvEXT(\{program}, \{location}, \{bufSize}, \{params})");
		try {
		    functions.glGetnUniformfvEXT().invokeExact(program, location, bufSize, params);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetnUniformfvKHR">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetnUniformfvKHR.xhtml">registry.khronos.org</a>
	 */
	public void glGetnUniformfvKHR(int program, int location, int bufSize, MutPointer<Float> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetnUniformfvKHR(\{program}, \{location}, \{bufSize}, \{params})");
		try {
		    functions.glGetnUniformfvKHR().invokeExact(program, location, bufSize, params);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetnUniformi64vARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetnUniformi64vARB.xhtml">registry.khronos.org</a>
	 */
	public void glGetnUniformi64vARB(int program, int location, int bufSize, MutPointer<Long> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetnUniformi64vARB(\{program}, \{location}, \{bufSize}, \{params})");
		try {
		    functions.glGetnUniformi64vARB().invokeExact(program, location, bufSize, params);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetnUniformiv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetnUniformiv.xhtml">registry.khronos.org</a>
	 */
	public void glGetnUniformiv(int program, int location, int bufSize, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetnUniformiv(\{program}, \{location}, \{bufSize}, \{params})");
		try {
		    functions.glGetnUniformiv().invokeExact(program, location, bufSize, params);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetnUniformivARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetnUniformivARB.xhtml">registry.khronos.org</a>
	 */
	public void glGetnUniformivARB(int program, int location, int bufSize, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetnUniformivARB(\{program}, \{location}, \{bufSize}, \{params})");
		try {
		    functions.glGetnUniformivARB().invokeExact(program, location, bufSize, params);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetnUniformivEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetnUniformivEXT.xhtml">registry.khronos.org</a>
	 */
	public void glGetnUniformivEXT(int program, int location, int bufSize, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetnUniformivEXT(\{program}, \{location}, \{bufSize}, \{params})");
		try {
		    functions.glGetnUniformivEXT().invokeExact(program, location, bufSize, params);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetnUniformivKHR">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetnUniformivKHR.xhtml">registry.khronos.org</a>
	 */
	public void glGetnUniformivKHR(int program, int location, int bufSize, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetnUniformivKHR(\{program}, \{location}, \{bufSize}, \{params})");
		try {
		    functions.glGetnUniformivKHR().invokeExact(program, location, bufSize, params);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetnUniformui64vARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetnUniformui64vARB.xhtml">registry.khronos.org</a>
	 */
	public void glGetnUniformui64vARB(int program, int location, int bufSize, MutPointer<Long> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetnUniformui64vARB(\{program}, \{location}, \{bufSize}, \{params})");
		try {
		    functions.glGetnUniformui64vARB().invokeExact(program, location, bufSize, params);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetnUniformuiv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetnUniformuiv.xhtml">registry.khronos.org</a>
	 */
	public void glGetnUniformuiv(int program, int location, int bufSize, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetnUniformuiv(\{program}, \{location}, \{bufSize}, \{params})");
		try {
		    functions.glGetnUniformuiv().invokeExact(program, location, bufSize, params);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetnUniformuivARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetnUniformuivARB.xhtml">registry.khronos.org</a>
	 */
	public void glGetnUniformuivARB(int program, int location, int bufSize, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetnUniformuivARB(\{program}, \{location}, \{bufSize}, \{params})");
		try {
		    functions.glGetnUniformuivARB().invokeExact(program, location, bufSize, params);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetnUniformuivKHR">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetnUniformuivKHR.xhtml">registry.khronos.org</a>
	 */
	public void glGetnUniformuivKHR(int program, int location, int bufSize, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetnUniformuivKHR(\{program}, \{location}, \{bufSize}, \{params})");
		try {
		    functions.glGetnUniformuivKHR().invokeExact(program, location, bufSize, params);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGlobalAlphaFactorbSUN">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGlobalAlphaFactorbSUN.xhtml">registry.khronos.org</a>
	 */
	public void glGlobalAlphaFactorbSUN(byte factor) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGlobalAlphaFactorbSUN(\{factor})");
		try {
		    functions.glGlobalAlphaFactorbSUN().invokeExact(factor);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGlobalAlphaFactordSUN">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGlobalAlphaFactordSUN.xhtml">registry.khronos.org</a>
	 */
	public void glGlobalAlphaFactordSUN(double factor) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGlobalAlphaFactordSUN(\{factor})");
		try {
		    functions.glGlobalAlphaFactordSUN().invokeExact(factor);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGlobalAlphaFactorfSUN">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGlobalAlphaFactorfSUN.xhtml">registry.khronos.org</a>
	 */
	public void glGlobalAlphaFactorfSUN(float factor) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGlobalAlphaFactorfSUN(\{factor})");
		try {
		    functions.glGlobalAlphaFactorfSUN().invokeExact(factor);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGlobalAlphaFactoriSUN">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGlobalAlphaFactoriSUN.xhtml">registry.khronos.org</a>
	 */
	public void glGlobalAlphaFactoriSUN(int factor) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGlobalAlphaFactoriSUN(\{factor})");
		try {
		    functions.glGlobalAlphaFactoriSUN().invokeExact(factor);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGlobalAlphaFactorsSUN">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGlobalAlphaFactorsSUN.xhtml">registry.khronos.org</a>
	 */
	public void glGlobalAlphaFactorsSUN(short factor) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGlobalAlphaFactorsSUN(\{factor})");
		try {
		    functions.glGlobalAlphaFactorsSUN().invokeExact(factor);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGlobalAlphaFactorubSUN">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGlobalAlphaFactorubSUN.xhtml">registry.khronos.org</a>
	 */
	public void glGlobalAlphaFactorubSUN(byte factor) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGlobalAlphaFactorubSUN(\{factor})");
		try {
		    functions.glGlobalAlphaFactorubSUN().invokeExact(factor);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGlobalAlphaFactoruiSUN">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGlobalAlphaFactoruiSUN.xhtml">registry.khronos.org</a>
	 */
	public void glGlobalAlphaFactoruiSUN(int factor) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGlobalAlphaFactoruiSUN(\{factor})");
		try {
		    functions.glGlobalAlphaFactoruiSUN().invokeExact(factor);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGlobalAlphaFactorusSUN">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGlobalAlphaFactorusSUN.xhtml">registry.khronos.org</a>
	 */
	public void glGlobalAlphaFactorusSUN(char factor) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGlobalAlphaFactorusSUN(\{factor})");
		try {
		    functions.glGlobalAlphaFactorusSUN().invokeExact(factor);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glHint">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glHint.xhtml">registry.khronos.org</a>
	 */
	public void glHint(GLHintTarget target, GLHintMode mode) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glHint(\{target}, \{mode})");
		try {
		    functions.glHint().invokeExact(target, mode);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, mode=\{mode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glHintPGI">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glHintPGI.xhtml">registry.khronos.org</a>
	 */
	public void glHintPGI(GLHintTargetPGI target, GLVertexHintsMaskPGI mode) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glHintPGI(\{target}, \{mode})");
		try {
		    functions.glHintPGI().invokeExact(target, mode);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, mode=\{mode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glHistogram">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glHistogram.xhtml">registry.khronos.org</a>
	 */
	public void glHistogram(GLHistogramTargetEXT target, int width, GLInternalFormat internalformat, boolean sink) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glHistogram(\{target}, \{width}, \{internalformat}, \{sink})");
		try {
		    functions.glHistogram().invokeExact(target, width, internalformat, sink);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, internalformat=\{internalformat}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glHistogramEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glHistogramEXT.xhtml">registry.khronos.org</a>
	 */
	public void glHistogramEXT(GLHistogramTargetEXT target, int width, GLInternalFormat internalformat, boolean sink) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glHistogramEXT(\{target}, \{width}, \{internalformat}, \{sink})");
		try {
		    functions.glHistogramEXT().invokeExact(target, width, internalformat, sink);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, internalformat=\{internalformat}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glIglooInterfaceSGIX">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glIglooInterfaceSGIX.xhtml">registry.khronos.org</a>
	 */
	public void glIglooInterfaceSGIX(GLEnum pname, ConstPointer<Void> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glIglooInterfaceSGIX(\{pname}, \{params})");
		try {
		    functions.glIglooInterfaceSGIX().invokeExact(pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glImageTransformParameterfHP">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glImageTransformParameterfHP.xhtml">registry.khronos.org</a>
	 */
	public void glImageTransformParameterfHP(GLImageTransformTargetHP target, GLImageTransformPNameHP pname, float param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glImageTransformParameterfHP(\{target}, \{pname}, \{param})");
		try {
		    functions.glImageTransformParameterfHP().invokeExact(target, pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glImageTransformParameterfvHP">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glImageTransformParameterfvHP.xhtml">registry.khronos.org</a>
	 */
	public void glImageTransformParameterfvHP(GLImageTransformTargetHP target, GLImageTransformPNameHP pname, ConstPointer<Float> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glImageTransformParameterfvHP(\{target}, \{pname}, \{params})");
		try {
		    functions.glImageTransformParameterfvHP().invokeExact(target, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glImageTransformParameteriHP">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glImageTransformParameteriHP.xhtml">registry.khronos.org</a>
	 */
	public void glImageTransformParameteriHP(GLImageTransformTargetHP target, GLImageTransformPNameHP pname, int param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glImageTransformParameteriHP(\{target}, \{pname}, \{param})");
		try {
		    functions.glImageTransformParameteriHP().invokeExact(target, pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glImageTransformParameterivHP">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glImageTransformParameterivHP.xhtml">registry.khronos.org</a>
	 */
	public void glImageTransformParameterivHP(GLImageTransformTargetHP target, GLImageTransformPNameHP pname, ConstPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glImageTransformParameterivHP(\{target}, \{pname}, \{params})");
		try {
		    functions.glImageTransformParameterivHP().invokeExact(target, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glImportMemoryFdEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glImportMemoryFdEXT.xhtml">registry.khronos.org</a>
	 */
	public void glImportMemoryFdEXT(int memory, long size, GLExternalHandleType handleType, int fd) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glImportMemoryFdEXT(\{memory}, \{size}, \{handleType}, \{fd})");
		try {
		    functions.glImportMemoryFdEXT().invokeExact(memory, size, handleType, fd);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were handleType=\{handleType}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glImportMemoryWin32HandleEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glImportMemoryWin32HandleEXT.xhtml">registry.khronos.org</a>
	 */
	public void glImportMemoryWin32HandleEXT(int memory, long size, GLExternalHandleType handleType, MutPointer<Void> handle) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glImportMemoryWin32HandleEXT(\{memory}, \{size}, \{handleType}, \{handle})");
		try {
		    functions.glImportMemoryWin32HandleEXT().invokeExact(memory, size, handleType, handle);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were handleType=\{handleType}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glImportMemoryWin32NameEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glImportMemoryWin32NameEXT.xhtml">registry.khronos.org</a>
	 */
	public void glImportMemoryWin32NameEXT(int memory, long size, GLExternalHandleType handleType, ConstPointer<Void> name) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glImportMemoryWin32NameEXT(\{memory}, \{size}, \{handleType}, \{name})");
		try {
		    functions.glImportMemoryWin32NameEXT().invokeExact(memory, size, handleType, name);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were handleType=\{handleType}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glImportSemaphoreFdEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glImportSemaphoreFdEXT.xhtml">registry.khronos.org</a>
	 */
	public void glImportSemaphoreFdEXT(int semaphore, GLExternalHandleType handleType, int fd) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glImportSemaphoreFdEXT(\{semaphore}, \{handleType}, \{fd})");
		try {
		    functions.glImportSemaphoreFdEXT().invokeExact(semaphore, handleType, fd);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were handleType=\{handleType}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glImportSemaphoreWin32HandleEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glImportSemaphoreWin32HandleEXT.xhtml">registry.khronos.org</a>
	 */
	public void glImportSemaphoreWin32HandleEXT(int semaphore, GLExternalHandleType handleType, MutPointer<Void> handle) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glImportSemaphoreWin32HandleEXT(\{semaphore}, \{handleType}, \{handle})");
		try {
		    functions.glImportSemaphoreWin32HandleEXT().invokeExact(semaphore, handleType, handle);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were handleType=\{handleType}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glImportSemaphoreWin32NameEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glImportSemaphoreWin32NameEXT.xhtml">registry.khronos.org</a>
	 */
	public void glImportSemaphoreWin32NameEXT(int semaphore, GLExternalHandleType handleType, ConstPointer<Void> name) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glImportSemaphoreWin32NameEXT(\{semaphore}, \{handleType}, \{name})");
		try {
		    functions.glImportSemaphoreWin32NameEXT().invokeExact(semaphore, handleType, name);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were handleType=\{handleType}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glImportSyncEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glImportSyncEXT.xhtml">registry.khronos.org</a>
	 */
	public ConstPointer<Void> glImportSyncEXT(GLEnum external_sync_type, long external_sync, int flags) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glImportSyncEXT(\{external_sync_type}, \{external_sync}, \{flags})");
		try {
		    var r = (ConstPointer<Void>)functions.glImportSyncEXT().invokeExact(external_sync_type, external_sync, flags);
			checkError();
			return r;
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were external_sync_type=\{external_sync_type}.");
			}
			throw e;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glIndexFormatNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glIndexFormatNV.xhtml">registry.khronos.org</a>
	 */
	public void glIndexFormatNV(GLEnum type, int stride) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glIndexFormatNV(\{type}, \{stride})");
		try {
		    functions.glIndexFormatNV().invokeExact(type, stride);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glIndexFuncEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glIndexFuncEXT.xhtml">registry.khronos.org</a>
	 */
	public void glIndexFuncEXT(GLIndexFunctionEXT func, float ref) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glIndexFuncEXT(\{func}, \{ref})");
		try {
		    functions.glIndexFuncEXT().invokeExact(func, ref);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were func=\{func}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glIndexMask">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glIndexMask.xhtml">registry.khronos.org</a>
	 */
	public void glIndexMask(int mask) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glIndexMask(\{mask})");
		try {
		    functions.glIndexMask().invokeExact(mask);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glIndexMaterialEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glIndexMaterialEXT.xhtml">registry.khronos.org</a>
	 */
	public void glIndexMaterialEXT(GLTriangleFace face, GLIndexMaterialParameterEXT mode) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glIndexMaterialEXT(\{face}, \{mode})");
		try {
		    functions.glIndexMaterialEXT().invokeExact(face, mode);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were face=\{face}, mode=\{mode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glIndexPointer">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glIndexPointer.xhtml">registry.khronos.org</a>
	 */
	public void glIndexPointer(GLIndexPointerType type, int stride, ConstPointer<Void> pointer) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glIndexPointer(\{type}, \{stride}, \{pointer})");
		try {
		    functions.glIndexPointer().invokeExact(type, stride, pointer);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glIndexPointerEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glIndexPointerEXT.xhtml">registry.khronos.org</a>
	 */
	public void glIndexPointerEXT(GLIndexPointerType type, int stride, int count, ConstPointer<Void> pointer) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glIndexPointerEXT(\{type}, \{stride}, \{count}, \{pointer})");
		try {
		    functions.glIndexPointerEXT().invokeExact(type, stride, count, pointer);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glIndexPointerListIBM">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glIndexPointerListIBM.xhtml">registry.khronos.org</a>
	 */
	public void glIndexPointerListIBM(GLIndexPointerType type, int stride, MutPointer<ConstPointer<Void>> pointer, int ptrstride) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glIndexPointerListIBM(\{type}, \{stride}, \{pointer}, \{ptrstride})");
		try {
		    functions.glIndexPointerListIBM().invokeExact(type, stride, pointer, ptrstride);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glIndexd">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glIndexd.xhtml">registry.khronos.org</a>
	 */
	public void glIndexd(double c) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glIndexd(\{c})");
		try {
		    functions.glIndexd().invokeExact(c);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glIndexdv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glIndexdv.xhtml">registry.khronos.org</a>
	 */
	public void glIndexdv(ConstPointer<Double> c) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glIndexdv(\{c})");
		try {
		    functions.glIndexdv().invokeExact(c);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glIndexf">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glIndexf.xhtml">registry.khronos.org</a>
	 */
	public void glIndexf(float c) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glIndexf(\{c})");
		try {
		    functions.glIndexf().invokeExact(c);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glIndexfv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glIndexfv.xhtml">registry.khronos.org</a>
	 */
	public void glIndexfv(ConstPointer<Float> c) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glIndexfv(\{c})");
		try {
		    functions.glIndexfv().invokeExact(c);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glIndexi">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glIndexi.xhtml">registry.khronos.org</a>
	 */
	public void glIndexi(int c) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glIndexi(\{c})");
		try {
		    functions.glIndexi().invokeExact(c);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glIndexiv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glIndexiv.xhtml">registry.khronos.org</a>
	 */
	public void glIndexiv(ConstPointer<Integer> c) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glIndexiv(\{c})");
		try {
		    functions.glIndexiv().invokeExact(c);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glIndexs">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glIndexs.xhtml">registry.khronos.org</a>
	 */
	public void glIndexs(short c) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glIndexs(\{c})");
		try {
		    functions.glIndexs().invokeExact(c);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glIndexsv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glIndexsv.xhtml">registry.khronos.org</a>
	 */
	public void glIndexsv(ConstPointer<Short> c) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glIndexsv(\{c})");
		try {
		    functions.glIndexsv().invokeExact(c);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glIndexub">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glIndexub.xhtml">registry.khronos.org</a>
	 */
	public void glIndexub(byte c) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glIndexub(\{c})");
		try {
		    functions.glIndexub().invokeExact(c);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glIndexubv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glIndexubv.xhtml">registry.khronos.org</a>
	 */
	public void glIndexubv(ConstPointer<Byte> c) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glIndexubv(\{c})");
		try {
		    functions.glIndexubv().invokeExact(c);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glIndexxOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glIndexxOES.xhtml">registry.khronos.org</a>
	 */
	public void glIndexxOES(int component) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glIndexxOES(\{component})");
		try {
		    functions.glIndexxOES().invokeExact(component);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glIndexxvOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glIndexxvOES.xhtml">registry.khronos.org</a>
	 */
	public void glIndexxvOES(ConstPointer<Integer> component) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glIndexxvOES(\{component})");
		try {
		    functions.glIndexxvOES().invokeExact(component);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glInitNames">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glInitNames.xhtml">registry.khronos.org</a>
	 */
	public void glInitNames() {
	    if (LOG_COMMANDS) logCmd(() -> STR."glInitNames()");
		try {
		    functions.glInitNames().invokeExact();
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glInsertComponentEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glInsertComponentEXT.xhtml">registry.khronos.org</a>
	 */
	public void glInsertComponentEXT(int res, int src, int num) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glInsertComponentEXT(\{res}, \{src}, \{num})");
		try {
		    functions.glInsertComponentEXT().invokeExact(res, src, num);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glInsertEventMarkerEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glInsertEventMarkerEXT.xhtml">registry.khronos.org</a>
	 */
	public void glInsertEventMarkerEXT(int length, ConstPointer<Byte> marker) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glInsertEventMarkerEXT(\{length}, \{marker})");
		try {
		    functions.glInsertEventMarkerEXT().invokeExact(length, marker);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glInstrumentsBufferSGIX">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glInstrumentsBufferSGIX.xhtml">registry.khronos.org</a>
	 */
	public void glInstrumentsBufferSGIX(int size, MutPointer<Integer> buffer) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glInstrumentsBufferSGIX(\{size}, \{buffer})");
		try {
		    functions.glInstrumentsBufferSGIX().invokeExact(size, buffer);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glInterleavedArrays">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glInterleavedArrays.xhtml">registry.khronos.org</a>
	 */
	public void glInterleavedArrays(GLInterleavedArrayFormat format, int stride, ConstPointer<Void> pointer) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glInterleavedArrays(\{format}, \{stride}, \{pointer})");
		try {
		    functions.glInterleavedArrays().invokeExact(format, stride, pointer);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were format=\{format}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glInterpolatePathsNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glInterpolatePathsNV.xhtml">registry.khronos.org</a>
	 */
	public void glInterpolatePathsNV(int resultPath, int pathA, int pathB, float weight) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glInterpolatePathsNV(\{resultPath}, \{pathA}, \{pathB}, \{weight})");
		try {
		    functions.glInterpolatePathsNV().invokeExact(resultPath, pathA, pathB, weight);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glInvalidateBufferData">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glInvalidateBufferData.xhtml">registry.khronos.org</a>
	 */
	public void glInvalidateBufferData(int buffer) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glInvalidateBufferData(\{buffer})");
		try {
		    functions.glInvalidateBufferData().invokeExact(buffer);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glInvalidateBufferSubData">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glInvalidateBufferSubData.xhtml">registry.khronos.org</a>
	 */
	public void glInvalidateBufferSubData(int buffer, long offset, long length) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glInvalidateBufferSubData(\{buffer}, \{offset}, \{length})");
		try {
		    functions.glInvalidateBufferSubData().invokeExact(buffer, offset, length);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glInvalidateFramebuffer">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glInvalidateFramebuffer.xhtml">registry.khronos.org</a>
	 */
	public void glInvalidateFramebuffer(GLFramebufferTarget target, int numAttachments, ConstPointer<GLInvalidateFramebufferAttachment> attachments) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glInvalidateFramebuffer(\{target}, \{numAttachments}, \{attachments})");
		try {
		    functions.glInvalidateFramebuffer().invokeExact(target, numAttachments, attachments);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glInvalidateNamedFramebufferData">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glInvalidateNamedFramebufferData.xhtml">registry.khronos.org</a>
	 */
	public void glInvalidateNamedFramebufferData(int framebuffer, int numAttachments, ConstPointer<GLFramebufferAttachment> attachments) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glInvalidateNamedFramebufferData(\{framebuffer}, \{numAttachments}, \{attachments})");
		try {
		    functions.glInvalidateNamedFramebufferData().invokeExact(framebuffer, numAttachments, attachments);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glInvalidateNamedFramebufferSubData">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glInvalidateNamedFramebufferSubData.xhtml">registry.khronos.org</a>
	 */
	public void glInvalidateNamedFramebufferSubData(int framebuffer, int numAttachments, ConstPointer<GLFramebufferAttachment> attachments, int x, int y, int width, int height) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glInvalidateNamedFramebufferSubData(\{framebuffer}, \{numAttachments}, \{attachments}, \{x}, \{y}, \{width}, \{height})");
		try {
		    functions.glInvalidateNamedFramebufferSubData().invokeExact(framebuffer, numAttachments, attachments, x, y, width, height);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glInvalidateSubFramebuffer">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glInvalidateSubFramebuffer.xhtml">registry.khronos.org</a>
	 */
	public void glInvalidateSubFramebuffer(GLFramebufferTarget target, int numAttachments, ConstPointer<GLInvalidateFramebufferAttachment> attachments, int x, int y, int width, int height) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glInvalidateSubFramebuffer(\{target}, \{numAttachments}, \{attachments}, \{x}, \{y}, \{width}, \{height})");
		try {
		    functions.glInvalidateSubFramebuffer().invokeExact(target, numAttachments, attachments, x, y, width, height);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glInvalidateTexImage">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glInvalidateTexImage.xhtml">registry.khronos.org</a>
	 */
	public void glInvalidateTexImage(int texture, int level) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glInvalidateTexImage(\{texture}, \{level})");
		try {
		    functions.glInvalidateTexImage().invokeExact(texture, level);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glInvalidateTexSubImage">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glInvalidateTexSubImage.xhtml">registry.khronos.org</a>
	 */
	public void glInvalidateTexSubImage(int texture, int level, int xoffset, int yoffset, int zoffset, int width, int height, int depth) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glInvalidateTexSubImage(\{texture}, \{level}, \{xoffset}, \{yoffset}, \{zoffset}, \{width}, \{height}, \{depth})");
		try {
		    functions.glInvalidateTexSubImage().invokeExact(texture, level, xoffset, yoffset, zoffset, width, height, depth);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glIsAsyncMarkerSGIX">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glIsAsyncMarkerSGIX.xhtml">registry.khronos.org</a>
	 */
	public boolean glIsAsyncMarkerSGIX(int marker) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glIsAsyncMarkerSGIX(\{marker})");
		try {
		    var r = (boolean)functions.glIsAsyncMarkerSGIX().invokeExact(marker);
			checkError();
			return r;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glIsBuffer">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glIsBuffer.xhtml">registry.khronos.org</a>
	 */
	public boolean glIsBuffer(int buffer) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glIsBuffer(\{buffer})");
		try {
		    var r = (boolean)functions.glIsBuffer().invokeExact(buffer);
			checkError();
			return r;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glIsBufferARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glIsBufferARB.xhtml">registry.khronos.org</a>
	 */
	public boolean glIsBufferARB(int buffer) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glIsBufferARB(\{buffer})");
		try {
		    var r = (boolean)functions.glIsBufferARB().invokeExact(buffer);
			checkError();
			return r;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glIsBufferResidentNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glIsBufferResidentNV.xhtml">registry.khronos.org</a>
	 */
	public boolean glIsBufferResidentNV(GLEnum target) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glIsBufferResidentNV(\{target})");
		try {
		    var r = (boolean)functions.glIsBufferResidentNV().invokeExact(target);
			checkError();
			return r;
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glIsCommandListNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glIsCommandListNV.xhtml">registry.khronos.org</a>
	 */
	public boolean glIsCommandListNV(int list) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glIsCommandListNV(\{list})");
		try {
		    var r = (boolean)functions.glIsCommandListNV().invokeExact(list);
			checkError();
			return r;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glIsEnabled">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glIsEnabled.xhtml">registry.khronos.org</a>
	 */
	public boolean glIsEnabled(GLEnableCap cap) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glIsEnabled(\{cap})");
		try {
		    var r = (boolean)functions.glIsEnabled().invokeExact(cap);
			checkError();
			return r;
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were cap=\{cap}.");
			}
			throw e;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glIsEnabledIndexedEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glIsEnabledIndexedEXT.xhtml">registry.khronos.org</a>
	 */
	public boolean glIsEnabledIndexedEXT(GLEnableCap target, int index) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glIsEnabledIndexedEXT(\{target}, \{index})");
		try {
		    var r = (boolean)functions.glIsEnabledIndexedEXT().invokeExact(target, index);
			checkError();
			return r;
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glIsEnabledi">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glIsEnabledi.xhtml">registry.khronos.org</a>
	 */
	public boolean glIsEnabledi(GLEnableCap target, int index) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glIsEnabledi(\{target}, \{index})");
		try {
		    var r = (boolean)functions.glIsEnabledi().invokeExact(target, index);
			checkError();
			return r;
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glIsEnablediEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glIsEnablediEXT.xhtml">registry.khronos.org</a>
	 */
	public boolean glIsEnablediEXT(GLEnableCap target, int index) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glIsEnablediEXT(\{target}, \{index})");
		try {
		    var r = (boolean)functions.glIsEnablediEXT().invokeExact(target, index);
			checkError();
			return r;
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glIsEnablediNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glIsEnablediNV.xhtml">registry.khronos.org</a>
	 */
	public boolean glIsEnablediNV(GLEnableCap target, int index) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glIsEnablediNV(\{target}, \{index})");
		try {
		    var r = (boolean)functions.glIsEnablediNV().invokeExact(target, index);
			checkError();
			return r;
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glIsEnablediOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glIsEnablediOES.xhtml">registry.khronos.org</a>
	 */
	public boolean glIsEnablediOES(GLEnableCap target, int index) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glIsEnablediOES(\{target}, \{index})");
		try {
		    var r = (boolean)functions.glIsEnablediOES().invokeExact(target, index);
			checkError();
			return r;
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glIsFenceAPPLE">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glIsFenceAPPLE.xhtml">registry.khronos.org</a>
	 */
	public boolean glIsFenceAPPLE(int fence) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glIsFenceAPPLE(\{fence})");
		try {
		    var r = (boolean)functions.glIsFenceAPPLE().invokeExact(fence);
			checkError();
			return r;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glIsFenceNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glIsFenceNV.xhtml">registry.khronos.org</a>
	 */
	public boolean glIsFenceNV(int fence) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glIsFenceNV(\{fence})");
		try {
		    var r = (boolean)functions.glIsFenceNV().invokeExact(fence);
			checkError();
			return r;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glIsFramebuffer">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glIsFramebuffer.xhtml">registry.khronos.org</a>
	 */
	public boolean glIsFramebuffer(int framebuffer) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glIsFramebuffer(\{framebuffer})");
		try {
		    var r = (boolean)functions.glIsFramebuffer().invokeExact(framebuffer);
			checkError();
			return r;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glIsFramebufferEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glIsFramebufferEXT.xhtml">registry.khronos.org</a>
	 */
	public boolean glIsFramebufferEXT(int framebuffer) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glIsFramebufferEXT(\{framebuffer})");
		try {
		    var r = (boolean)functions.glIsFramebufferEXT().invokeExact(framebuffer);
			checkError();
			return r;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glIsFramebufferOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glIsFramebufferOES.xhtml">registry.khronos.org</a>
	 */
	public boolean glIsFramebufferOES(int framebuffer) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glIsFramebufferOES(\{framebuffer})");
		try {
		    var r = (boolean)functions.glIsFramebufferOES().invokeExact(framebuffer);
			checkError();
			return r;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glIsImageHandleResidentARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glIsImageHandleResidentARB.xhtml">registry.khronos.org</a>
	 */
	public boolean glIsImageHandleResidentARB(long handle) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glIsImageHandleResidentARB(\{handle})");
		try {
		    var r = (boolean)functions.glIsImageHandleResidentARB().invokeExact(handle);
			checkError();
			return r;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glIsImageHandleResidentNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glIsImageHandleResidentNV.xhtml">registry.khronos.org</a>
	 */
	public boolean glIsImageHandleResidentNV(long handle) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glIsImageHandleResidentNV(\{handle})");
		try {
		    var r = (boolean)functions.glIsImageHandleResidentNV().invokeExact(handle);
			checkError();
			return r;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glIsList">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glIsList.xhtml">registry.khronos.org</a>
	 */
	public boolean glIsList(int list) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glIsList(\{list})");
		try {
		    var r = (boolean)functions.glIsList().invokeExact(list);
			checkError();
			return r;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glIsMemoryObjectEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glIsMemoryObjectEXT.xhtml">registry.khronos.org</a>
	 */
	public boolean glIsMemoryObjectEXT(int memoryObject) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glIsMemoryObjectEXT(\{memoryObject})");
		try {
		    var r = (boolean)functions.glIsMemoryObjectEXT().invokeExact(memoryObject);
			checkError();
			return r;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glIsNameAMD">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glIsNameAMD.xhtml">registry.khronos.org</a>
	 */
	public boolean glIsNameAMD(GLEnum identifier, int name) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glIsNameAMD(\{identifier}, \{name})");
		try {
		    var r = (boolean)functions.glIsNameAMD().invokeExact(identifier, name);
			checkError();
			return r;
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were identifier=\{identifier}.");
			}
			throw e;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glIsNamedBufferResidentNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glIsNamedBufferResidentNV.xhtml">registry.khronos.org</a>
	 */
	public boolean glIsNamedBufferResidentNV(int buffer) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glIsNamedBufferResidentNV(\{buffer})");
		try {
		    var r = (boolean)functions.glIsNamedBufferResidentNV().invokeExact(buffer);
			checkError();
			return r;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glIsNamedStringARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glIsNamedStringARB.xhtml">registry.khronos.org</a>
	 */
	public boolean glIsNamedStringARB(int namelen, ConstPointer<Byte> name) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glIsNamedStringARB(\{namelen}, \{name})");
		try {
		    var r = (boolean)functions.glIsNamedStringARB().invokeExact(namelen, name);
			checkError();
			return r;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glIsObjectBufferATI">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glIsObjectBufferATI.xhtml">registry.khronos.org</a>
	 */
	public boolean glIsObjectBufferATI(int buffer) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glIsObjectBufferATI(\{buffer})");
		try {
		    var r = (boolean)functions.glIsObjectBufferATI().invokeExact(buffer);
			checkError();
			return r;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glIsOcclusionQueryNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glIsOcclusionQueryNV.xhtml">registry.khronos.org</a>
	 */
	public boolean glIsOcclusionQueryNV(int id) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glIsOcclusionQueryNV(\{id})");
		try {
		    var r = (boolean)functions.glIsOcclusionQueryNV().invokeExact(id);
			checkError();
			return r;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glIsPathNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glIsPathNV.xhtml">registry.khronos.org</a>
	 */
	public boolean glIsPathNV(int path) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glIsPathNV(\{path})");
		try {
		    var r = (boolean)functions.glIsPathNV().invokeExact(path);
			checkError();
			return r;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glIsPointInFillPathNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glIsPointInFillPathNV.xhtml">registry.khronos.org</a>
	 */
	public boolean glIsPointInFillPathNV(int path, int mask, float x, float y) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glIsPointInFillPathNV(\{path}, \{mask}, \{x}, \{y})");
		try {
		    var r = (boolean)functions.glIsPointInFillPathNV().invokeExact(path, mask, x, y);
			checkError();
			return r;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glIsPointInStrokePathNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glIsPointInStrokePathNV.xhtml">registry.khronos.org</a>
	 */
	public boolean glIsPointInStrokePathNV(int path, float x, float y) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glIsPointInStrokePathNV(\{path}, \{x}, \{y})");
		try {
		    var r = (boolean)functions.glIsPointInStrokePathNV().invokeExact(path, x, y);
			checkError();
			return r;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glIsProgram">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glIsProgram.xhtml">registry.khronos.org</a>
	 */
	public boolean glIsProgram(int program) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glIsProgram(\{program})");
		try {
		    var r = (boolean)functions.glIsProgram().invokeExact(program);
			checkError();
			return r;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glIsProgramARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glIsProgramARB.xhtml">registry.khronos.org</a>
	 */
	public boolean glIsProgramARB(int program) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glIsProgramARB(\{program})");
		try {
		    var r = (boolean)functions.glIsProgramARB().invokeExact(program);
			checkError();
			return r;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glIsProgramNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glIsProgramNV.xhtml">registry.khronos.org</a>
	 */
	public boolean glIsProgramNV(int id) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glIsProgramNV(\{id})");
		try {
		    var r = (boolean)functions.glIsProgramNV().invokeExact(id);
			checkError();
			return r;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glIsProgramPipeline">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glIsProgramPipeline.xhtml">registry.khronos.org</a>
	 */
	public boolean glIsProgramPipeline(int pipeline) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glIsProgramPipeline(\{pipeline})");
		try {
		    var r = (boolean)functions.glIsProgramPipeline().invokeExact(pipeline);
			checkError();
			return r;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glIsProgramPipelineEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glIsProgramPipelineEXT.xhtml">registry.khronos.org</a>
	 */
	public boolean glIsProgramPipelineEXT(int pipeline) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glIsProgramPipelineEXT(\{pipeline})");
		try {
		    var r = (boolean)functions.glIsProgramPipelineEXT().invokeExact(pipeline);
			checkError();
			return r;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glIsQuery">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glIsQuery.xhtml">registry.khronos.org</a>
	 */
	public boolean glIsQuery(int id) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glIsQuery(\{id})");
		try {
		    var r = (boolean)functions.glIsQuery().invokeExact(id);
			checkError();
			return r;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glIsQueryARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glIsQueryARB.xhtml">registry.khronos.org</a>
	 */
	public boolean glIsQueryARB(int id) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glIsQueryARB(\{id})");
		try {
		    var r = (boolean)functions.glIsQueryARB().invokeExact(id);
			checkError();
			return r;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glIsQueryEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glIsQueryEXT.xhtml">registry.khronos.org</a>
	 */
	public boolean glIsQueryEXT(int id) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glIsQueryEXT(\{id})");
		try {
		    var r = (boolean)functions.glIsQueryEXT().invokeExact(id);
			checkError();
			return r;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glIsRenderbuffer">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glIsRenderbuffer.xhtml">registry.khronos.org</a>
	 */
	public boolean glIsRenderbuffer(int renderbuffer) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glIsRenderbuffer(\{renderbuffer})");
		try {
		    var r = (boolean)functions.glIsRenderbuffer().invokeExact(renderbuffer);
			checkError();
			return r;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glIsRenderbufferEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glIsRenderbufferEXT.xhtml">registry.khronos.org</a>
	 */
	public boolean glIsRenderbufferEXT(int renderbuffer) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glIsRenderbufferEXT(\{renderbuffer})");
		try {
		    var r = (boolean)functions.glIsRenderbufferEXT().invokeExact(renderbuffer);
			checkError();
			return r;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glIsRenderbufferOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glIsRenderbufferOES.xhtml">registry.khronos.org</a>
	 */
	public boolean glIsRenderbufferOES(int renderbuffer) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glIsRenderbufferOES(\{renderbuffer})");
		try {
		    var r = (boolean)functions.glIsRenderbufferOES().invokeExact(renderbuffer);
			checkError();
			return r;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glIsSemaphoreEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glIsSemaphoreEXT.xhtml">registry.khronos.org</a>
	 */
	public boolean glIsSemaphoreEXT(int semaphore) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glIsSemaphoreEXT(\{semaphore})");
		try {
		    var r = (boolean)functions.glIsSemaphoreEXT().invokeExact(semaphore);
			checkError();
			return r;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glIsSampler">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glIsSampler.xhtml">registry.khronos.org</a>
	 */
	public boolean glIsSampler(int sampler) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glIsSampler(\{sampler})");
		try {
		    var r = (boolean)functions.glIsSampler().invokeExact(sampler);
			checkError();
			return r;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glIsShader">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glIsShader.xhtml">registry.khronos.org</a>
	 */
	public boolean glIsShader(int shader) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glIsShader(\{shader})");
		try {
		    var r = (boolean)functions.glIsShader().invokeExact(shader);
			checkError();
			return r;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glIsStateNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glIsStateNV.xhtml">registry.khronos.org</a>
	 */
	public boolean glIsStateNV(int state) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glIsStateNV(\{state})");
		try {
		    var r = (boolean)functions.glIsStateNV().invokeExact(state);
			checkError();
			return r;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glIsSync">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glIsSync.xhtml">registry.khronos.org</a>
	 */
	public boolean glIsSync(ConstPointer<Void> sync) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glIsSync(\{sync})");
		try {
		    var r = (boolean)functions.glIsSync().invokeExact(sync);
			checkError();
			return r;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glIsSyncAPPLE">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glIsSyncAPPLE.xhtml">registry.khronos.org</a>
	 */
	public boolean glIsSyncAPPLE(ConstPointer<Void> sync) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glIsSyncAPPLE(\{sync})");
		try {
		    var r = (boolean)functions.glIsSyncAPPLE().invokeExact(sync);
			checkError();
			return r;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glIsTexture">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glIsTexture.xhtml">registry.khronos.org</a>
	 */
	public boolean glIsTexture(int texture) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glIsTexture(\{texture})");
		try {
		    var r = (boolean)functions.glIsTexture().invokeExact(texture);
			checkError();
			return r;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glIsTextureEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glIsTextureEXT.xhtml">registry.khronos.org</a>
	 */
	public boolean glIsTextureEXT(int texture) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glIsTextureEXT(\{texture})");
		try {
		    var r = (boolean)functions.glIsTextureEXT().invokeExact(texture);
			checkError();
			return r;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glIsTextureHandleResidentARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glIsTextureHandleResidentARB.xhtml">registry.khronos.org</a>
	 */
	public boolean glIsTextureHandleResidentARB(long handle) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glIsTextureHandleResidentARB(\{handle})");
		try {
		    var r = (boolean)functions.glIsTextureHandleResidentARB().invokeExact(handle);
			checkError();
			return r;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glIsTextureHandleResidentNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glIsTextureHandleResidentNV.xhtml">registry.khronos.org</a>
	 */
	public boolean glIsTextureHandleResidentNV(long handle) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glIsTextureHandleResidentNV(\{handle})");
		try {
		    var r = (boolean)functions.glIsTextureHandleResidentNV().invokeExact(handle);
			checkError();
			return r;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glIsTransformFeedback">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glIsTransformFeedback.xhtml">registry.khronos.org</a>
	 */
	public boolean glIsTransformFeedback(int id) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glIsTransformFeedback(\{id})");
		try {
		    var r = (boolean)functions.glIsTransformFeedback().invokeExact(id);
			checkError();
			return r;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glIsTransformFeedbackNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glIsTransformFeedbackNV.xhtml">registry.khronos.org</a>
	 */
	public boolean glIsTransformFeedbackNV(int id) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glIsTransformFeedbackNV(\{id})");
		try {
		    var r = (boolean)functions.glIsTransformFeedbackNV().invokeExact(id);
			checkError();
			return r;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glIsVariantEnabledEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glIsVariantEnabledEXT.xhtml">registry.khronos.org</a>
	 */
	public boolean glIsVariantEnabledEXT(int id, GLVariantCapEXT cap) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glIsVariantEnabledEXT(\{id}, \{cap})");
		try {
		    var r = (boolean)functions.glIsVariantEnabledEXT().invokeExact(id, cap);
			checkError();
			return r;
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were cap=\{cap}.");
			}
			throw e;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glIsVertexArray">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glIsVertexArray.xhtml">registry.khronos.org</a>
	 */
	public boolean glIsVertexArray(int array) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glIsVertexArray(\{array})");
		try {
		    var r = (boolean)functions.glIsVertexArray().invokeExact(array);
			checkError();
			return r;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glIsVertexArrayAPPLE">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glIsVertexArrayAPPLE.xhtml">registry.khronos.org</a>
	 */
	public boolean glIsVertexArrayAPPLE(int array) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glIsVertexArrayAPPLE(\{array})");
		try {
		    var r = (boolean)functions.glIsVertexArrayAPPLE().invokeExact(array);
			checkError();
			return r;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glIsVertexArrayOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glIsVertexArrayOES.xhtml">registry.khronos.org</a>
	 */
	public boolean glIsVertexArrayOES(int array) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glIsVertexArrayOES(\{array})");
		try {
		    var r = (boolean)functions.glIsVertexArrayOES().invokeExact(array);
			checkError();
			return r;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glIsVertexAttribEnabledAPPLE">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glIsVertexAttribEnabledAPPLE.xhtml">registry.khronos.org</a>
	 */
	public boolean glIsVertexAttribEnabledAPPLE(int index, GLEnum pname) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glIsVertexAttribEnabledAPPLE(\{index}, \{pname})");
		try {
		    var r = (boolean)functions.glIsVertexAttribEnabledAPPLE().invokeExact(index, pname);
			checkError();
			return r;
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glLGPUCopyImageSubDataNVX">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glLGPUCopyImageSubDataNVX.xhtml">registry.khronos.org</a>
	 */
	public void glLGPUCopyImageSubDataNVX(int sourceGpu, int destinationGpuMask, int srcName, GLEnum srcTarget, int srcLevel, int srcX, int srxY, int srcZ, int dstName, GLEnum dstTarget, int dstLevel, int dstX, int dstY, int dstZ, int width, int height, int depth) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glLGPUCopyImageSubDataNVX(\{sourceGpu}, \{destinationGpuMask}, \{srcName}, \{srcTarget}, \{srcLevel}, \{srcX}, \{srxY}, \{srcZ}, \{dstName}, \{dstTarget}, \{dstLevel}, \{dstX}, \{dstY}, \{dstZ}, \{width}, \{height}, \{depth})");
		try {
		    functions.glLGPUCopyImageSubDataNVX().invokeExact(sourceGpu, destinationGpuMask, srcName, srcTarget, srcLevel, srcX, srxY, srcZ, dstName, dstTarget, dstLevel, dstX, dstY, dstZ, width, height, depth);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were srcTarget=\{srcTarget}, dstTarget=\{dstTarget}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glLGPUInterlockNVX">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glLGPUInterlockNVX.xhtml">registry.khronos.org</a>
	 */
	public void glLGPUInterlockNVX() {
	    if (LOG_COMMANDS) logCmd(() -> STR."glLGPUInterlockNVX()");
		try {
		    functions.glLGPUInterlockNVX().invokeExact();
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glLGPUNamedBufferSubDataNVX">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glLGPUNamedBufferSubDataNVX.xhtml">registry.khronos.org</a>
	 */
	public void glLGPUNamedBufferSubDataNVX(int gpuMask, int buffer, long offset, long size, ConstPointer<Void> data) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glLGPUNamedBufferSubDataNVX(\{gpuMask}, \{buffer}, \{offset}, \{size}, \{data})");
		try {
		    functions.glLGPUNamedBufferSubDataNVX().invokeExact(gpuMask, buffer, offset, size, data);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glLabelObjectEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glLabelObjectEXT.xhtml">registry.khronos.org</a>
	 */
	public void glLabelObjectEXT(GLEnum type, int object, int length, ConstPointer<Byte> label) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glLabelObjectEXT(\{type}, \{object}, \{length}, \{label})");
		try {
		    functions.glLabelObjectEXT().invokeExact(type, object, length, label);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glLightEnviSGIX">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glLightEnviSGIX.xhtml">registry.khronos.org</a>
	 */
	public void glLightEnviSGIX(GLLightEnvParameterSGIX pname, GLLightEnvModeSGIX param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glLightEnviSGIX(\{pname}, \{param})");
		try {
		    functions.glLightEnviSGIX().invokeExact(pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}, param=\{param}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glLightModelf">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glLightModelf.xhtml">registry.khronos.org</a>
	 */
	public void glLightModelf(GLLightModelParameter pname, float param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glLightModelf(\{pname}, \{param})");
		try {
		    functions.glLightModelf().invokeExact(pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glLightModelfv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glLightModelfv.xhtml">registry.khronos.org</a>
	 */
	public void glLightModelfv(GLLightModelParameter pname, ConstPointer<Float> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glLightModelfv(\{pname}, \{params})");
		try {
		    functions.glLightModelfv().invokeExact(pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glLightModeli">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glLightModeli.xhtml">registry.khronos.org</a>
	 */
	public void glLightModeli(GLLightModelParameter pname, int param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glLightModeli(\{pname}, \{param})");
		try {
		    functions.glLightModeli().invokeExact(pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glLightModeliv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glLightModeliv.xhtml">registry.khronos.org</a>
	 */
	public void glLightModeliv(GLLightModelParameter pname, ConstPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glLightModeliv(\{pname}, \{params})");
		try {
		    functions.glLightModeliv().invokeExact(pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glLightModelx">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glLightModelx.xhtml">registry.khronos.org</a>
	 */
	public void glLightModelx(GLLightModelParameter pname, int param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glLightModelx(\{pname}, \{param})");
		try {
		    functions.glLightModelx().invokeExact(pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glLightModelxOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glLightModelxOES.xhtml">registry.khronos.org</a>
	 */
	public void glLightModelxOES(GLLightModelParameter pname, int param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glLightModelxOES(\{pname}, \{param})");
		try {
		    functions.glLightModelxOES().invokeExact(pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glLightModelxv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glLightModelxv.xhtml">registry.khronos.org</a>
	 */
	public void glLightModelxv(GLLightModelParameter pname, ConstPointer<Integer> param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glLightModelxv(\{pname}, \{param})");
		try {
		    functions.glLightModelxv().invokeExact(pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glLightModelxvOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glLightModelxvOES.xhtml">registry.khronos.org</a>
	 */
	public void glLightModelxvOES(GLLightModelParameter pname, ConstPointer<Integer> param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glLightModelxvOES(\{pname}, \{param})");
		try {
		    functions.glLightModelxvOES().invokeExact(pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glLightf">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glLightf.xhtml">registry.khronos.org</a>
	 */
	public void glLightf(GLLightName light, GLLightParameter pname, float param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glLightf(\{light}, \{pname}, \{param})");
		try {
		    functions.glLightf().invokeExact(light, pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were light=\{light}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glLightfv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glLightfv.xhtml">registry.khronos.org</a>
	 */
	public void glLightfv(GLLightName light, GLLightParameter pname, ConstPointer<Float> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glLightfv(\{light}, \{pname}, \{params})");
		try {
		    functions.glLightfv().invokeExact(light, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were light=\{light}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glLighti">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glLighti.xhtml">registry.khronos.org</a>
	 */
	public void glLighti(GLLightName light, GLLightParameter pname, int param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glLighti(\{light}, \{pname}, \{param})");
		try {
		    functions.glLighti().invokeExact(light, pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were light=\{light}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glLightiv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glLightiv.xhtml">registry.khronos.org</a>
	 */
	public void glLightiv(GLLightName light, GLLightParameter pname, ConstPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glLightiv(\{light}, \{pname}, \{params})");
		try {
		    functions.glLightiv().invokeExact(light, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were light=\{light}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glLightx">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glLightx.xhtml">registry.khronos.org</a>
	 */
	public void glLightx(GLLightName light, GLLightParameter pname, int param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glLightx(\{light}, \{pname}, \{param})");
		try {
		    functions.glLightx().invokeExact(light, pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were light=\{light}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glLightxOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glLightxOES.xhtml">registry.khronos.org</a>
	 */
	public void glLightxOES(GLLightName light, GLLightParameter pname, int param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glLightxOES(\{light}, \{pname}, \{param})");
		try {
		    functions.glLightxOES().invokeExact(light, pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were light=\{light}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glLightxv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glLightxv.xhtml">registry.khronos.org</a>
	 */
	public void glLightxv(GLLightName light, GLLightParameter pname, ConstPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glLightxv(\{light}, \{pname}, \{params})");
		try {
		    functions.glLightxv().invokeExact(light, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were light=\{light}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glLightxvOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glLightxvOES.xhtml">registry.khronos.org</a>
	 */
	public void glLightxvOES(GLLightName light, GLLightParameter pname, ConstPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glLightxvOES(\{light}, \{pname}, \{params})");
		try {
		    functions.glLightxvOES().invokeExact(light, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were light=\{light}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glLineStipple">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glLineStipple.xhtml">registry.khronos.org</a>
	 */
	public void glLineStipple(int factor, char pattern) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glLineStipple(\{factor}, \{pattern})");
		try {
		    functions.glLineStipple().invokeExact(factor, pattern);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glLineWidth">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glLineWidth.xhtml">registry.khronos.org</a>
	 */
	public void glLineWidth(float width) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glLineWidth(\{width})");
		try {
		    functions.glLineWidth().invokeExact(width);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glLineWidthx">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glLineWidthx.xhtml">registry.khronos.org</a>
	 */
	public void glLineWidthx(int width) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glLineWidthx(\{width})");
		try {
		    functions.glLineWidthx().invokeExact(width);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glLineWidthxOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glLineWidthxOES.xhtml">registry.khronos.org</a>
	 */
	public void glLineWidthxOES(int width) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glLineWidthxOES(\{width})");
		try {
		    functions.glLineWidthxOES().invokeExact(width);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glLinkProgram">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glLinkProgram.xhtml">registry.khronos.org</a>
	 */
	public void glLinkProgram(int program) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glLinkProgram(\{program})");
		try {
		    functions.glLinkProgram().invokeExact(program);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glLinkProgramARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glLinkProgramARB.xhtml">registry.khronos.org</a>
	 */
	public void glLinkProgramARB(int programObj) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glLinkProgramARB(\{programObj})");
		try {
		    functions.glLinkProgramARB().invokeExact(programObj);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glListBase">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glListBase.xhtml">registry.khronos.org</a>
	 */
	public void glListBase(int base) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glListBase(\{base})");
		try {
		    functions.glListBase().invokeExact(base);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glListDrawCommandsStatesClientNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glListDrawCommandsStatesClientNV.xhtml">registry.khronos.org</a>
	 */
	public void glListDrawCommandsStatesClientNV(int list, int segment, MutPointer<ConstPointer<Void>> indirects, ConstPointer<Integer> sizes, ConstPointer<Integer> states, ConstPointer<Integer> fbos, int count) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glListDrawCommandsStatesClientNV(\{list}, \{segment}, \{indirects}, \{sizes}, \{states}, \{fbos}, \{count})");
		try {
		    functions.glListDrawCommandsStatesClientNV().invokeExact(list, segment, indirects, sizes, states, fbos, count);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glListParameterfSGIX">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glListParameterfSGIX.xhtml">registry.khronos.org</a>
	 */
	public void glListParameterfSGIX(int list, GLListParameterName pname, float param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glListParameterfSGIX(\{list}, \{pname}, \{param})");
		try {
		    functions.glListParameterfSGIX().invokeExact(list, pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glListParameterfvSGIX">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glListParameterfvSGIX.xhtml">registry.khronos.org</a>
	 */
	public void glListParameterfvSGIX(int list, GLListParameterName pname, ConstPointer<Float> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glListParameterfvSGIX(\{list}, \{pname}, \{params})");
		try {
		    functions.glListParameterfvSGIX().invokeExact(list, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glListParameteriSGIX">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glListParameteriSGIX.xhtml">registry.khronos.org</a>
	 */
	public void glListParameteriSGIX(int list, GLListParameterName pname, int param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glListParameteriSGIX(\{list}, \{pname}, \{param})");
		try {
		    functions.glListParameteriSGIX().invokeExact(list, pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glListParameterivSGIX">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glListParameterivSGIX.xhtml">registry.khronos.org</a>
	 */
	public void glListParameterivSGIX(int list, GLListParameterName pname, ConstPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glListParameterivSGIX(\{list}, \{pname}, \{params})");
		try {
		    functions.glListParameterivSGIX().invokeExact(list, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glLoadIdentity">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glLoadIdentity.xhtml">registry.khronos.org</a>
	 */
	public void glLoadIdentity() {
	    if (LOG_COMMANDS) logCmd(() -> STR."glLoadIdentity()");
		try {
		    functions.glLoadIdentity().invokeExact();
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glLoadIdentityDeformationMapSGIX">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glLoadIdentityDeformationMapSGIX.xhtml">registry.khronos.org</a>
	 */
	public void glLoadIdentityDeformationMapSGIX(int mask) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glLoadIdentityDeformationMapSGIX(\{mask})");
		try {
		    functions.glLoadIdentityDeformationMapSGIX().invokeExact(mask);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glLoadMatrixd">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glLoadMatrixd.xhtml">registry.khronos.org</a>
	 */
	public void glLoadMatrixd(ConstPointer<Double> m) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glLoadMatrixd(\{m})");
		try {
		    functions.glLoadMatrixd().invokeExact(m);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glLoadMatrixf">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glLoadMatrixf.xhtml">registry.khronos.org</a>
	 */
	public void glLoadMatrixf(ConstPointer<Float> m) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glLoadMatrixf(\{m})");
		try {
		    functions.glLoadMatrixf().invokeExact(m);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glLoadMatrixx">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glLoadMatrixx.xhtml">registry.khronos.org</a>
	 */
	public void glLoadMatrixx(ConstPointer<Integer> m) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glLoadMatrixx(\{m})");
		try {
		    functions.glLoadMatrixx().invokeExact(m);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glLoadMatrixxOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glLoadMatrixxOES.xhtml">registry.khronos.org</a>
	 */
	public void glLoadMatrixxOES(ConstPointer<Integer> m) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glLoadMatrixxOES(\{m})");
		try {
		    functions.glLoadMatrixxOES().invokeExact(m);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glLoadName">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glLoadName.xhtml">registry.khronos.org</a>
	 */
	public void glLoadName(int name) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glLoadName(\{name})");
		try {
		    functions.glLoadName().invokeExact(name);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glLoadPaletteFromModelViewMatrixOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glLoadPaletteFromModelViewMatrixOES.xhtml">registry.khronos.org</a>
	 */
	public void glLoadPaletteFromModelViewMatrixOES() {
	    if (LOG_COMMANDS) logCmd(() -> STR."glLoadPaletteFromModelViewMatrixOES()");
		try {
		    functions.glLoadPaletteFromModelViewMatrixOES().invokeExact();
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glLoadProgramNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glLoadProgramNV.xhtml">registry.khronos.org</a>
	 */
	public void glLoadProgramNV(GLVertexAttribEnumNV target, int id, int len, ConstPointer<Byte> program) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glLoadProgramNV(\{target}, \{id}, \{len}, \{program})");
		try {
		    functions.glLoadProgramNV().invokeExact(target, id, len, program);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glLoadTransposeMatrixd">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glLoadTransposeMatrixd.xhtml">registry.khronos.org</a>
	 */
	public void glLoadTransposeMatrixd(ConstPointer<Double> m) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glLoadTransposeMatrixd(\{m})");
		try {
		    functions.glLoadTransposeMatrixd().invokeExact(m);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glLoadTransposeMatrixdARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glLoadTransposeMatrixdARB.xhtml">registry.khronos.org</a>
	 */
	public void glLoadTransposeMatrixdARB(ConstPointer<Double> m) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glLoadTransposeMatrixdARB(\{m})");
		try {
		    functions.glLoadTransposeMatrixdARB().invokeExact(m);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glLoadTransposeMatrixf">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glLoadTransposeMatrixf.xhtml">registry.khronos.org</a>
	 */
	public void glLoadTransposeMatrixf(ConstPointer<Float> m) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glLoadTransposeMatrixf(\{m})");
		try {
		    functions.glLoadTransposeMatrixf().invokeExact(m);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glLoadTransposeMatrixfARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glLoadTransposeMatrixfARB.xhtml">registry.khronos.org</a>
	 */
	public void glLoadTransposeMatrixfARB(ConstPointer<Float> m) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glLoadTransposeMatrixfARB(\{m})");
		try {
		    functions.glLoadTransposeMatrixfARB().invokeExact(m);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glLoadTransposeMatrixxOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glLoadTransposeMatrixxOES.xhtml">registry.khronos.org</a>
	 */
	public void glLoadTransposeMatrixxOES(ConstPointer<Integer> m) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glLoadTransposeMatrixxOES(\{m})");
		try {
		    functions.glLoadTransposeMatrixxOES().invokeExact(m);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glLockArraysEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glLockArraysEXT.xhtml">registry.khronos.org</a>
	 */
	public void glLockArraysEXT(int first, int count) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glLockArraysEXT(\{first}, \{count})");
		try {
		    functions.glLockArraysEXT().invokeExact(first, count);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glLogicOp">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glLogicOp.xhtml">registry.khronos.org</a>
	 */
	public void glLogicOp(GLLogicOp opcode) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glLogicOp(\{opcode})");
		try {
		    functions.glLogicOp().invokeExact(opcode);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were opcode=\{opcode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMakeBufferNonResidentNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMakeBufferNonResidentNV.xhtml">registry.khronos.org</a>
	 */
	public void glMakeBufferNonResidentNV(GLEnum target) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMakeBufferNonResidentNV(\{target})");
		try {
		    functions.glMakeBufferNonResidentNV().invokeExact(target);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMakeBufferResidentNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMakeBufferResidentNV.xhtml">registry.khronos.org</a>
	 */
	public void glMakeBufferResidentNV(GLEnum target, GLEnum access) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMakeBufferResidentNV(\{target}, \{access})");
		try {
		    functions.glMakeBufferResidentNV().invokeExact(target, access);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, access=\{access}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMakeImageHandleNonResidentARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMakeImageHandleNonResidentARB.xhtml">registry.khronos.org</a>
	 */
	public void glMakeImageHandleNonResidentARB(long handle) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMakeImageHandleNonResidentARB(\{handle})");
		try {
		    functions.glMakeImageHandleNonResidentARB().invokeExact(handle);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMakeImageHandleNonResidentNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMakeImageHandleNonResidentNV.xhtml">registry.khronos.org</a>
	 */
	public void glMakeImageHandleNonResidentNV(long handle) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMakeImageHandleNonResidentNV(\{handle})");
		try {
		    functions.glMakeImageHandleNonResidentNV().invokeExact(handle);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMakeImageHandleResidentARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMakeImageHandleResidentARB.xhtml">registry.khronos.org</a>
	 */
	public void glMakeImageHandleResidentARB(long handle, GLEnum access) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMakeImageHandleResidentARB(\{handle}, \{access})");
		try {
		    functions.glMakeImageHandleResidentARB().invokeExact(handle, access);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were access=\{access}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMakeImageHandleResidentNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMakeImageHandleResidentNV.xhtml">registry.khronos.org</a>
	 */
	public void glMakeImageHandleResidentNV(long handle, GLEnum access) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMakeImageHandleResidentNV(\{handle}, \{access})");
		try {
		    functions.glMakeImageHandleResidentNV().invokeExact(handle, access);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were access=\{access}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMakeNamedBufferNonResidentNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMakeNamedBufferNonResidentNV.xhtml">registry.khronos.org</a>
	 */
	public void glMakeNamedBufferNonResidentNV(int buffer) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMakeNamedBufferNonResidentNV(\{buffer})");
		try {
		    functions.glMakeNamedBufferNonResidentNV().invokeExact(buffer);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMakeNamedBufferResidentNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMakeNamedBufferResidentNV.xhtml">registry.khronos.org</a>
	 */
	public void glMakeNamedBufferResidentNV(int buffer, GLEnum access) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMakeNamedBufferResidentNV(\{buffer}, \{access})");
		try {
		    functions.glMakeNamedBufferResidentNV().invokeExact(buffer, access);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were access=\{access}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMakeTextureHandleNonResidentARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMakeTextureHandleNonResidentARB.xhtml">registry.khronos.org</a>
	 */
	public void glMakeTextureHandleNonResidentARB(long handle) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMakeTextureHandleNonResidentARB(\{handle})");
		try {
		    functions.glMakeTextureHandleNonResidentARB().invokeExact(handle);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMakeTextureHandleNonResidentNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMakeTextureHandleNonResidentNV.xhtml">registry.khronos.org</a>
	 */
	public void glMakeTextureHandleNonResidentNV(long handle) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMakeTextureHandleNonResidentNV(\{handle})");
		try {
		    functions.glMakeTextureHandleNonResidentNV().invokeExact(handle);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMakeTextureHandleResidentARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMakeTextureHandleResidentARB.xhtml">registry.khronos.org</a>
	 */
	public void glMakeTextureHandleResidentARB(long handle) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMakeTextureHandleResidentARB(\{handle})");
		try {
		    functions.glMakeTextureHandleResidentARB().invokeExact(handle);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMakeTextureHandleResidentNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMakeTextureHandleResidentNV.xhtml">registry.khronos.org</a>
	 */
	public void glMakeTextureHandleResidentNV(long handle) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMakeTextureHandleResidentNV(\{handle})");
		try {
		    functions.glMakeTextureHandleResidentNV().invokeExact(handle);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMap1d">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMap1d.xhtml">registry.khronos.org</a>
	 */
	public void glMap1d(GLMapTarget target, double u1, double u2, int stride, int order, ConstPointer<Double> points) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMap1d(\{target}, \{u1}, \{u2}, \{stride}, \{order}, \{points})");
		try {
		    functions.glMap1d().invokeExact(target, u1, u2, stride, order, points);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMap1f">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMap1f.xhtml">registry.khronos.org</a>
	 */
	public void glMap1f(GLMapTarget target, float u1, float u2, int stride, int order, ConstPointer<Float> points) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMap1f(\{target}, \{u1}, \{u2}, \{stride}, \{order}, \{points})");
		try {
		    functions.glMap1f().invokeExact(target, u1, u2, stride, order, points);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMap1xOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMap1xOES.xhtml">registry.khronos.org</a>
	 */
	public void glMap1xOES(GLMapTarget target, int u1, int u2, int stride, int order, int points) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMap1xOES(\{target}, \{u1}, \{u2}, \{stride}, \{order}, \{points})");
		try {
		    functions.glMap1xOES().invokeExact(target, u1, u2, stride, order, points);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMap2d">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMap2d.xhtml">registry.khronos.org</a>
	 */
	public void glMap2d(GLMapTarget target, double u1, double u2, int ustride, int uorder, double v1, double v2, int vstride, int vorder, ConstPointer<Double> points) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMap2d(\{target}, \{u1}, \{u2}, \{ustride}, \{uorder}, \{v1}, \{v2}, \{vstride}, \{vorder}, \{points})");
		try {
		    functions.glMap2d().invokeExact(target, u1, u2, ustride, uorder, v1, v2, vstride, vorder, points);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMap2f">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMap2f.xhtml">registry.khronos.org</a>
	 */
	public void glMap2f(GLMapTarget target, float u1, float u2, int ustride, int uorder, float v1, float v2, int vstride, int vorder, ConstPointer<Float> points) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMap2f(\{target}, \{u1}, \{u2}, \{ustride}, \{uorder}, \{v1}, \{v2}, \{vstride}, \{vorder}, \{points})");
		try {
		    functions.glMap2f().invokeExact(target, u1, u2, ustride, uorder, v1, v2, vstride, vorder, points);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMap2xOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMap2xOES.xhtml">registry.khronos.org</a>
	 */
	public void glMap2xOES(GLMapTarget target, int u1, int u2, int ustride, int uorder, int v1, int v2, int vstride, int vorder, int points) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMap2xOES(\{target}, \{u1}, \{u2}, \{ustride}, \{uorder}, \{v1}, \{v2}, \{vstride}, \{vorder}, \{points})");
		try {
		    functions.glMap2xOES().invokeExact(target, u1, u2, ustride, uorder, v1, v2, vstride, vorder, points);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMapBuffer">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMapBuffer.xhtml">registry.khronos.org</a>
	 */
	public MutPointer<Void> glMapBuffer(GLBufferTarget target, GLBufferAccess access) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMapBuffer(\{target}, \{access})");
		try {
		    var r = (MutPointer<Void>)functions.glMapBuffer().invokeExact(target, access);
			checkError();
			return r;
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, access=\{access}.");
			}
			throw e;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMapBufferARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMapBufferARB.xhtml">registry.khronos.org</a>
	 */
	public MutPointer<Void> glMapBufferARB(GLBufferTarget target, GLBufferAccess access) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMapBufferARB(\{target}, \{access})");
		try {
		    var r = (MutPointer<Void>)functions.glMapBufferARB().invokeExact(target, access);
			checkError();
			return r;
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, access=\{access}.");
			}
			throw e;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMapBufferOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMapBufferOES.xhtml">registry.khronos.org</a>
	 */
	public MutPointer<Void> glMapBufferOES(GLBufferTarget target, GLBufferAccess access) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMapBufferOES(\{target}, \{access})");
		try {
		    var r = (MutPointer<Void>)functions.glMapBufferOES().invokeExact(target, access);
			checkError();
			return r;
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, access=\{access}.");
			}
			throw e;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMapBufferRange">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMapBufferRange.xhtml">registry.khronos.org</a>
	 */
	public MutPointer<Void> glMapBufferRange(GLBufferTarget target, long offset, long length, int access) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMapBufferRange(\{target}, \{offset}, \{length}, \{access})");
		try {
		    var r = (MutPointer<Void>)functions.glMapBufferRange().invokeExact(target, offset, length, access);
			checkError();
			return r;
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMapBufferRangeEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMapBufferRangeEXT.xhtml">registry.khronos.org</a>
	 */
	public MutPointer<Void> glMapBufferRangeEXT(GLBufferTarget target, long offset, long length, int access) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMapBufferRangeEXT(\{target}, \{offset}, \{length}, \{access})");
		try {
		    var r = (MutPointer<Void>)functions.glMapBufferRangeEXT().invokeExact(target, offset, length, access);
			checkError();
			return r;
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMapControlPointsNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMapControlPointsNV.xhtml">registry.khronos.org</a>
	 */
	public void glMapControlPointsNV(GLEvalTargetNV target, int index, GLMapTypeNV type, int ustride, int vstride, int uorder, int vorder, boolean packed, ConstPointer<Void> points) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMapControlPointsNV(\{target}, \{index}, \{type}, \{ustride}, \{vstride}, \{uorder}, \{vorder}, \{packed}, \{points})");
		try {
		    functions.glMapControlPointsNV().invokeExact(target, index, type, ustride, vstride, uorder, vorder, packed, points);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMapGrid1d">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMapGrid1d.xhtml">registry.khronos.org</a>
	 */
	public void glMapGrid1d(int un, double u1, double u2) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMapGrid1d(\{un}, \{u1}, \{u2})");
		try {
		    functions.glMapGrid1d().invokeExact(un, u1, u2);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMapGrid1f">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMapGrid1f.xhtml">registry.khronos.org</a>
	 */
	public void glMapGrid1f(int un, float u1, float u2) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMapGrid1f(\{un}, \{u1}, \{u2})");
		try {
		    functions.glMapGrid1f().invokeExact(un, u1, u2);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMapGrid1xOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMapGrid1xOES.xhtml">registry.khronos.org</a>
	 */
	public void glMapGrid1xOES(int n, int u1, int u2) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMapGrid1xOES(\{n}, \{u1}, \{u2})");
		try {
		    functions.glMapGrid1xOES().invokeExact(n, u1, u2);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMapGrid2d">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMapGrid2d.xhtml">registry.khronos.org</a>
	 */
	public void glMapGrid2d(int un, double u1, double u2, int vn, double v1, double v2) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMapGrid2d(\{un}, \{u1}, \{u2}, \{vn}, \{v1}, \{v2})");
		try {
		    functions.glMapGrid2d().invokeExact(un, u1, u2, vn, v1, v2);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMapGrid2f">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMapGrid2f.xhtml">registry.khronos.org</a>
	 */
	public void glMapGrid2f(int un, float u1, float u2, int vn, float v1, float v2) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMapGrid2f(\{un}, \{u1}, \{u2}, \{vn}, \{v1}, \{v2})");
		try {
		    functions.glMapGrid2f().invokeExact(un, u1, u2, vn, v1, v2);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMapGrid2xOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMapGrid2xOES.xhtml">registry.khronos.org</a>
	 */
	public void glMapGrid2xOES(int n, int u1, int u2, int v1, int v2) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMapGrid2xOES(\{n}, \{u1}, \{u2}, \{v1}, \{v2})");
		try {
		    functions.glMapGrid2xOES().invokeExact(n, u1, u2, v1, v2);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMapNamedBuffer">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMapNamedBuffer.xhtml">registry.khronos.org</a>
	 */
	public MutPointer<Void> glMapNamedBuffer(int buffer, GLBufferAccess access) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMapNamedBuffer(\{buffer}, \{access})");
		try {
		    var r = (MutPointer<Void>)functions.glMapNamedBuffer().invokeExact(buffer, access);
			checkError();
			return r;
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were access=\{access}.");
			}
			throw e;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMapNamedBufferEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMapNamedBufferEXT.xhtml">registry.khronos.org</a>
	 */
	public MutPointer<Void> glMapNamedBufferEXT(int buffer, GLBufferAccess access) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMapNamedBufferEXT(\{buffer}, \{access})");
		try {
		    var r = (MutPointer<Void>)functions.glMapNamedBufferEXT().invokeExact(buffer, access);
			checkError();
			return r;
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were access=\{access}.");
			}
			throw e;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMapNamedBufferRange">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMapNamedBufferRange.xhtml">registry.khronos.org</a>
	 */
	public MutPointer<Void> glMapNamedBufferRange(int buffer, long offset, long length, int access) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMapNamedBufferRange(\{buffer}, \{offset}, \{length}, \{access})");
		try {
		    var r = (MutPointer<Void>)functions.glMapNamedBufferRange().invokeExact(buffer, offset, length, access);
			checkError();
			return r;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMapNamedBufferRangeEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMapNamedBufferRangeEXT.xhtml">registry.khronos.org</a>
	 */
	public MutPointer<Void> glMapNamedBufferRangeEXT(int buffer, long offset, long length, int access) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMapNamedBufferRangeEXT(\{buffer}, \{offset}, \{length}, \{access})");
		try {
		    var r = (MutPointer<Void>)functions.glMapNamedBufferRangeEXT().invokeExact(buffer, offset, length, access);
			checkError();
			return r;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMapObjectBufferATI">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMapObjectBufferATI.xhtml">registry.khronos.org</a>
	 */
	public MutPointer<Void> glMapObjectBufferATI(int buffer) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMapObjectBufferATI(\{buffer})");
		try {
		    var r = (MutPointer<Void>)functions.glMapObjectBufferATI().invokeExact(buffer);
			checkError();
			return r;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMapParameterfvNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMapParameterfvNV.xhtml">registry.khronos.org</a>
	 */
	public void glMapParameterfvNV(GLEvalTargetNV target, GLMapParameterNV pname, ConstPointer<Float> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMapParameterfvNV(\{target}, \{pname}, \{params})");
		try {
		    functions.glMapParameterfvNV().invokeExact(target, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMapParameterivNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMapParameterivNV.xhtml">registry.khronos.org</a>
	 */
	public void glMapParameterivNV(GLEvalTargetNV target, GLMapParameterNV pname, ConstPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMapParameterivNV(\{target}, \{pname}, \{params})");
		try {
		    functions.glMapParameterivNV().invokeExact(target, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMapTexture2DINTEL">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMapTexture2DINTEL.xhtml">registry.khronos.org</a>
	 */
	public MutPointer<Void> glMapTexture2DINTEL(int texture, int level, int access, MutPointer<Integer> stride, MutPointer<GLEnum> layout) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMapTexture2DINTEL(\{texture}, \{level}, \{access}, \{stride}, \{layout})");
		try {
		    var r = (MutPointer<Void>)functions.glMapTexture2DINTEL().invokeExact(texture, level, access, stride, layout);
			checkError();
			return r;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMapVertexAttrib1dAPPLE">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMapVertexAttrib1dAPPLE.xhtml">registry.khronos.org</a>
	 */
	public void glMapVertexAttrib1dAPPLE(int index, int size, double u1, double u2, int stride, int order, ConstPointer<Double> points) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMapVertexAttrib1dAPPLE(\{index}, \{size}, \{u1}, \{u2}, \{stride}, \{order}, \{points})");
		try {
		    functions.glMapVertexAttrib1dAPPLE().invokeExact(index, size, u1, u2, stride, order, points);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMapVertexAttrib1fAPPLE">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMapVertexAttrib1fAPPLE.xhtml">registry.khronos.org</a>
	 */
	public void glMapVertexAttrib1fAPPLE(int index, int size, float u1, float u2, int stride, int order, ConstPointer<Float> points) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMapVertexAttrib1fAPPLE(\{index}, \{size}, \{u1}, \{u2}, \{stride}, \{order}, \{points})");
		try {
		    functions.glMapVertexAttrib1fAPPLE().invokeExact(index, size, u1, u2, stride, order, points);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMapVertexAttrib2dAPPLE">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMapVertexAttrib2dAPPLE.xhtml">registry.khronos.org</a>
	 */
	public void glMapVertexAttrib2dAPPLE(int index, int size, double u1, double u2, int ustride, int uorder, double v1, double v2, int vstride, int vorder, ConstPointer<Double> points) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMapVertexAttrib2dAPPLE(\{index}, \{size}, \{u1}, \{u2}, \{ustride}, \{uorder}, \{v1}, \{v2}, \{vstride}, \{vorder}, \{points})");
		try {
		    functions.glMapVertexAttrib2dAPPLE().invokeExact(index, size, u1, u2, ustride, uorder, v1, v2, vstride, vorder, points);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMapVertexAttrib2fAPPLE">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMapVertexAttrib2fAPPLE.xhtml">registry.khronos.org</a>
	 */
	public void glMapVertexAttrib2fAPPLE(int index, int size, float u1, float u2, int ustride, int uorder, float v1, float v2, int vstride, int vorder, ConstPointer<Float> points) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMapVertexAttrib2fAPPLE(\{index}, \{size}, \{u1}, \{u2}, \{ustride}, \{uorder}, \{v1}, \{v2}, \{vstride}, \{vorder}, \{points})");
		try {
		    functions.glMapVertexAttrib2fAPPLE().invokeExact(index, size, u1, u2, ustride, uorder, v1, v2, vstride, vorder, points);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMaterialf">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMaterialf.xhtml">registry.khronos.org</a>
	 */
	public void glMaterialf(GLTriangleFace face, GLMaterialParameter pname, float param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMaterialf(\{face}, \{pname}, \{param})");
		try {
		    functions.glMaterialf().invokeExact(face, pname, param);
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were face=\{face}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMaterialfv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMaterialfv.xhtml">registry.khronos.org</a>
	 */
	public void glMaterialfv(GLTriangleFace face, GLMaterialParameter pname, ConstPointer<Float> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMaterialfv(\{face}, \{pname}, \{params})");
		try {
		    functions.glMaterialfv().invokeExact(face, pname, params);
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were face=\{face}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMateriali">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMateriali.xhtml">registry.khronos.org</a>
	 */
	public void glMateriali(GLTriangleFace face, GLMaterialParameter pname, int param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMateriali(\{face}, \{pname}, \{param})");
		try {
		    functions.glMateriali().invokeExact(face, pname, param);
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were face=\{face}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMaterialiv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMaterialiv.xhtml">registry.khronos.org</a>
	 */
	public void glMaterialiv(GLTriangleFace face, GLMaterialParameter pname, ConstPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMaterialiv(\{face}, \{pname}, \{params})");
		try {
		    functions.glMaterialiv().invokeExact(face, pname, params);
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were face=\{face}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMaterialx">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMaterialx.xhtml">registry.khronos.org</a>
	 */
	public void glMaterialx(GLTriangleFace face, GLMaterialParameter pname, int param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMaterialx(\{face}, \{pname}, \{param})");
		try {
		    functions.glMaterialx().invokeExact(face, pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were face=\{face}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMaterialxOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMaterialxOES.xhtml">registry.khronos.org</a>
	 */
	public void glMaterialxOES(GLTriangleFace face, GLMaterialParameter pname, int param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMaterialxOES(\{face}, \{pname}, \{param})");
		try {
		    functions.glMaterialxOES().invokeExact(face, pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were face=\{face}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMaterialxv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMaterialxv.xhtml">registry.khronos.org</a>
	 */
	public void glMaterialxv(GLTriangleFace face, GLMaterialParameter pname, ConstPointer<Integer> param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMaterialxv(\{face}, \{pname}, \{param})");
		try {
		    functions.glMaterialxv().invokeExact(face, pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were face=\{face}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMaterialxvOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMaterialxvOES.xhtml">registry.khronos.org</a>
	 */
	public void glMaterialxvOES(GLTriangleFace face, GLMaterialParameter pname, ConstPointer<Integer> param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMaterialxvOES(\{face}, \{pname}, \{param})");
		try {
		    functions.glMaterialxvOES().invokeExact(face, pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were face=\{face}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMatrixFrustumEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMatrixFrustumEXT.xhtml">registry.khronos.org</a>
	 */
	public void glMatrixFrustumEXT(GLMatrixMode mode, double left, double right, double bottom, double top, double zNear, double zFar) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMatrixFrustumEXT(\{mode}, \{left}, \{right}, \{bottom}, \{top}, \{zNear}, \{zFar})");
		try {
		    functions.glMatrixFrustumEXT().invokeExact(mode, left, right, bottom, top, zNear, zFar);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMatrixIndexPointerARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMatrixIndexPointerARB.xhtml">registry.khronos.org</a>
	 */
	public void glMatrixIndexPointerARB(int size, GLMatrixIndexPointerTypeARB type, int stride, ConstPointer<Void> pointer) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMatrixIndexPointerARB(\{size}, \{type}, \{stride}, \{pointer})");
		try {
		    functions.glMatrixIndexPointerARB().invokeExact(size, type, stride, pointer);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMatrixIndexPointerOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMatrixIndexPointerOES.xhtml">registry.khronos.org</a>
	 */
	public void glMatrixIndexPointerOES(int size, GLMatrixIndexPointerTypeARB type, int stride, ConstPointer<Void> pointer) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMatrixIndexPointerOES(\{size}, \{type}, \{stride}, \{pointer})");
		try {
		    functions.glMatrixIndexPointerOES().invokeExact(size, type, stride, pointer);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMatrixIndexubvARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMatrixIndexubvARB.xhtml">registry.khronos.org</a>
	 */
	public void glMatrixIndexubvARB(int size, ConstPointer<Byte> indices) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMatrixIndexubvARB(\{size}, \{indices})");
		try {
		    functions.glMatrixIndexubvARB().invokeExact(size, indices);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMatrixIndexuivARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMatrixIndexuivARB.xhtml">registry.khronos.org</a>
	 */
	public void glMatrixIndexuivARB(int size, ConstPointer<Integer> indices) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMatrixIndexuivARB(\{size}, \{indices})");
		try {
		    functions.glMatrixIndexuivARB().invokeExact(size, indices);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMatrixIndexusvARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMatrixIndexusvARB.xhtml">registry.khronos.org</a>
	 */
	public void glMatrixIndexusvARB(int size, ConstPointer<Character> indices) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMatrixIndexusvARB(\{size}, \{indices})");
		try {
		    functions.glMatrixIndexusvARB().invokeExact(size, indices);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMatrixLoad3x2fNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMatrixLoad3x2fNV.xhtml">registry.khronos.org</a>
	 */
	public void glMatrixLoad3x2fNV(GLEnum matrixMode, ConstPointer<Float> m) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMatrixLoad3x2fNV(\{matrixMode}, \{m})");
		try {
		    functions.glMatrixLoad3x2fNV().invokeExact(matrixMode, m);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were matrixMode=\{matrixMode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMatrixLoad3x3fNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMatrixLoad3x3fNV.xhtml">registry.khronos.org</a>
	 */
	public void glMatrixLoad3x3fNV(GLEnum matrixMode, ConstPointer<Float> m) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMatrixLoad3x3fNV(\{matrixMode}, \{m})");
		try {
		    functions.glMatrixLoad3x3fNV().invokeExact(matrixMode, m);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were matrixMode=\{matrixMode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMatrixLoadIdentityEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMatrixLoadIdentityEXT.xhtml">registry.khronos.org</a>
	 */
	public void glMatrixLoadIdentityEXT(GLMatrixMode mode) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMatrixLoadIdentityEXT(\{mode})");
		try {
		    functions.glMatrixLoadIdentityEXT().invokeExact(mode);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMatrixLoadTranspose3x3fNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMatrixLoadTranspose3x3fNV.xhtml">registry.khronos.org</a>
	 */
	public void glMatrixLoadTranspose3x3fNV(GLEnum matrixMode, ConstPointer<Float> m) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMatrixLoadTranspose3x3fNV(\{matrixMode}, \{m})");
		try {
		    functions.glMatrixLoadTranspose3x3fNV().invokeExact(matrixMode, m);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were matrixMode=\{matrixMode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMatrixLoadTransposedEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMatrixLoadTransposedEXT.xhtml">registry.khronos.org</a>
	 */
	public void glMatrixLoadTransposedEXT(GLMatrixMode mode, ConstPointer<Double> m) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMatrixLoadTransposedEXT(\{mode}, \{m})");
		try {
		    functions.glMatrixLoadTransposedEXT().invokeExact(mode, m);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMatrixLoadTransposefEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMatrixLoadTransposefEXT.xhtml">registry.khronos.org</a>
	 */
	public void glMatrixLoadTransposefEXT(GLMatrixMode mode, ConstPointer<Float> m) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMatrixLoadTransposefEXT(\{mode}, \{m})");
		try {
		    functions.glMatrixLoadTransposefEXT().invokeExact(mode, m);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMatrixLoaddEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMatrixLoaddEXT.xhtml">registry.khronos.org</a>
	 */
	public void glMatrixLoaddEXT(GLMatrixMode mode, ConstPointer<Double> m) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMatrixLoaddEXT(\{mode}, \{m})");
		try {
		    functions.glMatrixLoaddEXT().invokeExact(mode, m);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMatrixLoadfEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMatrixLoadfEXT.xhtml">registry.khronos.org</a>
	 */
	public void glMatrixLoadfEXT(GLMatrixMode mode, ConstPointer<Float> m) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMatrixLoadfEXT(\{mode}, \{m})");
		try {
		    functions.glMatrixLoadfEXT().invokeExact(mode, m);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMatrixMode">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMatrixMode.xhtml">registry.khronos.org</a>
	 */
	public void glMatrixMode(GLMatrixMode mode) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMatrixMode(\{mode})");
		try {
		    functions.glMatrixMode().invokeExact(mode);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMatrixMult3x2fNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMatrixMult3x2fNV.xhtml">registry.khronos.org</a>
	 */
	public void glMatrixMult3x2fNV(GLEnum matrixMode, ConstPointer<Float> m) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMatrixMult3x2fNV(\{matrixMode}, \{m})");
		try {
		    functions.glMatrixMult3x2fNV().invokeExact(matrixMode, m);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were matrixMode=\{matrixMode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMatrixMult3x3fNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMatrixMult3x3fNV.xhtml">registry.khronos.org</a>
	 */
	public void glMatrixMult3x3fNV(GLEnum matrixMode, ConstPointer<Float> m) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMatrixMult3x3fNV(\{matrixMode}, \{m})");
		try {
		    functions.glMatrixMult3x3fNV().invokeExact(matrixMode, m);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were matrixMode=\{matrixMode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMatrixMultTranspose3x3fNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMatrixMultTranspose3x3fNV.xhtml">registry.khronos.org</a>
	 */
	public void glMatrixMultTranspose3x3fNV(GLEnum matrixMode, ConstPointer<Float> m) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMatrixMultTranspose3x3fNV(\{matrixMode}, \{m})");
		try {
		    functions.glMatrixMultTranspose3x3fNV().invokeExact(matrixMode, m);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were matrixMode=\{matrixMode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMatrixMultTransposedEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMatrixMultTransposedEXT.xhtml">registry.khronos.org</a>
	 */
	public void glMatrixMultTransposedEXT(GLMatrixMode mode, ConstPointer<Double> m) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMatrixMultTransposedEXT(\{mode}, \{m})");
		try {
		    functions.glMatrixMultTransposedEXT().invokeExact(mode, m);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMatrixMultTransposefEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMatrixMultTransposefEXT.xhtml">registry.khronos.org</a>
	 */
	public void glMatrixMultTransposefEXT(GLMatrixMode mode, ConstPointer<Float> m) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMatrixMultTransposefEXT(\{mode}, \{m})");
		try {
		    functions.glMatrixMultTransposefEXT().invokeExact(mode, m);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMatrixMultdEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMatrixMultdEXT.xhtml">registry.khronos.org</a>
	 */
	public void glMatrixMultdEXT(GLMatrixMode mode, ConstPointer<Double> m) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMatrixMultdEXT(\{mode}, \{m})");
		try {
		    functions.glMatrixMultdEXT().invokeExact(mode, m);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMatrixMultfEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMatrixMultfEXT.xhtml">registry.khronos.org</a>
	 */
	public void glMatrixMultfEXT(GLMatrixMode mode, ConstPointer<Float> m) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMatrixMultfEXT(\{mode}, \{m})");
		try {
		    functions.glMatrixMultfEXT().invokeExact(mode, m);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMatrixOrthoEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMatrixOrthoEXT.xhtml">registry.khronos.org</a>
	 */
	public void glMatrixOrthoEXT(GLMatrixMode mode, double left, double right, double bottom, double top, double zNear, double zFar) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMatrixOrthoEXT(\{mode}, \{left}, \{right}, \{bottom}, \{top}, \{zNear}, \{zFar})");
		try {
		    functions.glMatrixOrthoEXT().invokeExact(mode, left, right, bottom, top, zNear, zFar);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMatrixPopEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMatrixPopEXT.xhtml">registry.khronos.org</a>
	 */
	public void glMatrixPopEXT(GLMatrixMode mode) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMatrixPopEXT(\{mode})");
		try {
		    functions.glMatrixPopEXT().invokeExact(mode);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMatrixPushEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMatrixPushEXT.xhtml">registry.khronos.org</a>
	 */
	public void glMatrixPushEXT(GLMatrixMode mode) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMatrixPushEXT(\{mode})");
		try {
		    functions.glMatrixPushEXT().invokeExact(mode);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMatrixRotatedEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMatrixRotatedEXT.xhtml">registry.khronos.org</a>
	 */
	public void glMatrixRotatedEXT(GLMatrixMode mode, double angle, double x, double y, double z) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMatrixRotatedEXT(\{mode}, \{angle}, \{x}, \{y}, \{z})");
		try {
		    functions.glMatrixRotatedEXT().invokeExact(mode, angle, x, y, z);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMatrixRotatefEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMatrixRotatefEXT.xhtml">registry.khronos.org</a>
	 */
	public void glMatrixRotatefEXT(GLMatrixMode mode, float angle, float x, float y, float z) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMatrixRotatefEXT(\{mode}, \{angle}, \{x}, \{y}, \{z})");
		try {
		    functions.glMatrixRotatefEXT().invokeExact(mode, angle, x, y, z);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMatrixScaledEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMatrixScaledEXT.xhtml">registry.khronos.org</a>
	 */
	public void glMatrixScaledEXT(GLMatrixMode mode, double x, double y, double z) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMatrixScaledEXT(\{mode}, \{x}, \{y}, \{z})");
		try {
		    functions.glMatrixScaledEXT().invokeExact(mode, x, y, z);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMatrixScalefEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMatrixScalefEXT.xhtml">registry.khronos.org</a>
	 */
	public void glMatrixScalefEXT(GLMatrixMode mode, float x, float y, float z) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMatrixScalefEXT(\{mode}, \{x}, \{y}, \{z})");
		try {
		    functions.glMatrixScalefEXT().invokeExact(mode, x, y, z);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMatrixTranslatedEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMatrixTranslatedEXT.xhtml">registry.khronos.org</a>
	 */
	public void glMatrixTranslatedEXT(GLMatrixMode mode, double x, double y, double z) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMatrixTranslatedEXT(\{mode}, \{x}, \{y}, \{z})");
		try {
		    functions.glMatrixTranslatedEXT().invokeExact(mode, x, y, z);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMatrixTranslatefEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMatrixTranslatefEXT.xhtml">registry.khronos.org</a>
	 */
	public void glMatrixTranslatefEXT(GLMatrixMode mode, float x, float y, float z) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMatrixTranslatefEXT(\{mode}, \{x}, \{y}, \{z})");
		try {
		    functions.glMatrixTranslatefEXT().invokeExact(mode, x, y, z);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMaxActiveShaderCoresARM">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMaxActiveShaderCoresARM.xhtml">registry.khronos.org</a>
	 */
	public void glMaxActiveShaderCoresARM(int count) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMaxActiveShaderCoresARM(\{count})");
		try {
		    functions.glMaxActiveShaderCoresARM().invokeExact(count);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMaxShaderCompilerThreadsKHR">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMaxShaderCompilerThreadsKHR.xhtml">registry.khronos.org</a>
	 */
	public void glMaxShaderCompilerThreadsKHR(int count) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMaxShaderCompilerThreadsKHR(\{count})");
		try {
		    functions.glMaxShaderCompilerThreadsKHR().invokeExact(count);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMaxShaderCompilerThreadsARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMaxShaderCompilerThreadsARB.xhtml">registry.khronos.org</a>
	 */
	public void glMaxShaderCompilerThreadsARB(int count) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMaxShaderCompilerThreadsARB(\{count})");
		try {
		    functions.glMaxShaderCompilerThreadsARB().invokeExact(count);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMemoryBarrier">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMemoryBarrier.xhtml">registry.khronos.org</a>
	 */
	public void glMemoryBarrier(int barriers) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMemoryBarrier(\{barriers})");
		try {
		    functions.glMemoryBarrier().invokeExact(barriers);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMemoryBarrierByRegion">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMemoryBarrierByRegion.xhtml">registry.khronos.org</a>
	 */
	public void glMemoryBarrierByRegion(int barriers) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMemoryBarrierByRegion(\{barriers})");
		try {
		    functions.glMemoryBarrierByRegion().invokeExact(barriers);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMemoryBarrierEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMemoryBarrierEXT.xhtml">registry.khronos.org</a>
	 */
	public void glMemoryBarrierEXT(int barriers) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMemoryBarrierEXT(\{barriers})");
		try {
		    functions.glMemoryBarrierEXT().invokeExact(barriers);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMemoryObjectParameterivEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMemoryObjectParameterivEXT.xhtml">registry.khronos.org</a>
	 */
	public void glMemoryObjectParameterivEXT(int memoryObject, GLMemoryObjectParameterName pname, ConstPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMemoryObjectParameterivEXT(\{memoryObject}, \{pname}, \{params})");
		try {
		    functions.glMemoryObjectParameterivEXT().invokeExact(memoryObject, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMinSampleShading">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMinSampleShading.xhtml">registry.khronos.org</a>
	 */
	public void glMinSampleShading(float value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMinSampleShading(\{value})");
		try {
		    functions.glMinSampleShading().invokeExact(value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMinSampleShadingARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMinSampleShadingARB.xhtml">registry.khronos.org</a>
	 */
	public void glMinSampleShadingARB(float value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMinSampleShadingARB(\{value})");
		try {
		    functions.glMinSampleShadingARB().invokeExact(value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMinSampleShadingOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMinSampleShadingOES.xhtml">registry.khronos.org</a>
	 */
	public void glMinSampleShadingOES(float value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMinSampleShadingOES(\{value})");
		try {
		    functions.glMinSampleShadingOES().invokeExact(value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMinmax">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMinmax.xhtml">registry.khronos.org</a>
	 */
	public void glMinmax(GLMinmaxTargetEXT target, GLInternalFormat internalformat, boolean sink) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMinmax(\{target}, \{internalformat}, \{sink})");
		try {
		    functions.glMinmax().invokeExact(target, internalformat, sink);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, internalformat=\{internalformat}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMinmaxEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMinmaxEXT.xhtml">registry.khronos.org</a>
	 */
	public void glMinmaxEXT(GLMinmaxTargetEXT target, GLInternalFormat internalformat, boolean sink) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMinmaxEXT(\{target}, \{internalformat}, \{sink})");
		try {
		    functions.glMinmaxEXT().invokeExact(target, internalformat, sink);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, internalformat=\{internalformat}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultMatrixd">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultMatrixd.xhtml">registry.khronos.org</a>
	 */
	public void glMultMatrixd(ConstPointer<Double> m) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultMatrixd(\{m})");
		try {
		    functions.glMultMatrixd().invokeExact(m);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultMatrixf">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultMatrixf.xhtml">registry.khronos.org</a>
	 */
	public void glMultMatrixf(ConstPointer<Float> m) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultMatrixf(\{m})");
		try {
		    functions.glMultMatrixf().invokeExact(m);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultMatrixx">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultMatrixx.xhtml">registry.khronos.org</a>
	 */
	public void glMultMatrixx(ConstPointer<Integer> m) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultMatrixx(\{m})");
		try {
		    functions.glMultMatrixx().invokeExact(m);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultMatrixxOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultMatrixxOES.xhtml">registry.khronos.org</a>
	 */
	public void glMultMatrixxOES(ConstPointer<Integer> m) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultMatrixxOES(\{m})");
		try {
		    functions.glMultMatrixxOES().invokeExact(m);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultTransposeMatrixd">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultTransposeMatrixd.xhtml">registry.khronos.org</a>
	 */
	public void glMultTransposeMatrixd(ConstPointer<Double> m) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultTransposeMatrixd(\{m})");
		try {
		    functions.glMultTransposeMatrixd().invokeExact(m);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultTransposeMatrixdARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultTransposeMatrixdARB.xhtml">registry.khronos.org</a>
	 */
	public void glMultTransposeMatrixdARB(ConstPointer<Double> m) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultTransposeMatrixdARB(\{m})");
		try {
		    functions.glMultTransposeMatrixdARB().invokeExact(m);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultTransposeMatrixf">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultTransposeMatrixf.xhtml">registry.khronos.org</a>
	 */
	public void glMultTransposeMatrixf(ConstPointer<Float> m) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultTransposeMatrixf(\{m})");
		try {
		    functions.glMultTransposeMatrixf().invokeExact(m);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultTransposeMatrixfARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultTransposeMatrixfARB.xhtml">registry.khronos.org</a>
	 */
	public void glMultTransposeMatrixfARB(ConstPointer<Float> m) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultTransposeMatrixfARB(\{m})");
		try {
		    functions.glMultTransposeMatrixfARB().invokeExact(m);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultTransposeMatrixxOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultTransposeMatrixxOES.xhtml">registry.khronos.org</a>
	 */
	public void glMultTransposeMatrixxOES(ConstPointer<Integer> m) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultTransposeMatrixxOES(\{m})");
		try {
		    functions.glMultTransposeMatrixxOES().invokeExact(m);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiDrawArrays">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiDrawArrays.xhtml">registry.khronos.org</a>
	 */
	public void glMultiDrawArrays(GLPrimitiveType mode, ConstPointer<Integer> first, ConstPointer<Integer> count, int drawcount) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiDrawArrays(\{mode}, \{first}, \{count}, \{drawcount})");
		try {
		    functions.glMultiDrawArrays().invokeExact(mode, first, count, drawcount);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiDrawArraysEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiDrawArraysEXT.xhtml">registry.khronos.org</a>
	 */
	public void glMultiDrawArraysEXT(GLPrimitiveType mode, ConstPointer<Integer> first, ConstPointer<Integer> count, int primcount) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiDrawArraysEXT(\{mode}, \{first}, \{count}, \{primcount})");
		try {
		    functions.glMultiDrawArraysEXT().invokeExact(mode, first, count, primcount);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiDrawArraysIndirect">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiDrawArraysIndirect.xhtml">registry.khronos.org</a>
	 */
	public void glMultiDrawArraysIndirect(GLPrimitiveType mode, ConstPointer<Void> indirect, int drawcount, int stride) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiDrawArraysIndirect(\{mode}, \{indirect}, \{drawcount}, \{stride})");
		try {
		    functions.glMultiDrawArraysIndirect().invokeExact(mode, indirect, drawcount, stride);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiDrawArraysIndirectAMD">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiDrawArraysIndirectAMD.xhtml">registry.khronos.org</a>
	 */
	public void glMultiDrawArraysIndirectAMD(GLPrimitiveType mode, ConstPointer<Void> indirect, int primcount, int stride) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiDrawArraysIndirectAMD(\{mode}, \{indirect}, \{primcount}, \{stride})");
		try {
		    functions.glMultiDrawArraysIndirectAMD().invokeExact(mode, indirect, primcount, stride);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiDrawArraysIndirectBindlessCountNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiDrawArraysIndirectBindlessCountNV.xhtml">registry.khronos.org</a>
	 */
	public void glMultiDrawArraysIndirectBindlessCountNV(GLPrimitiveType mode, ConstPointer<Void> indirect, int drawCount, int maxDrawCount, int stride, int vertexBufferCount) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiDrawArraysIndirectBindlessCountNV(\{mode}, \{indirect}, \{drawCount}, \{maxDrawCount}, \{stride}, \{vertexBufferCount})");
		try {
		    functions.glMultiDrawArraysIndirectBindlessCountNV().invokeExact(mode, indirect, drawCount, maxDrawCount, stride, vertexBufferCount);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiDrawArraysIndirectBindlessNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiDrawArraysIndirectBindlessNV.xhtml">registry.khronos.org</a>
	 */
	public void glMultiDrawArraysIndirectBindlessNV(GLPrimitiveType mode, ConstPointer<Void> indirect, int drawCount, int stride, int vertexBufferCount) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiDrawArraysIndirectBindlessNV(\{mode}, \{indirect}, \{drawCount}, \{stride}, \{vertexBufferCount})");
		try {
		    functions.glMultiDrawArraysIndirectBindlessNV().invokeExact(mode, indirect, drawCount, stride, vertexBufferCount);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiDrawArraysIndirectCount">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiDrawArraysIndirectCount.xhtml">registry.khronos.org</a>
	 */
	public void glMultiDrawArraysIndirectCount(GLPrimitiveType mode, ConstPointer<Void> indirect, long drawcount, int maxdrawcount, int stride) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiDrawArraysIndirectCount(\{mode}, \{indirect}, \{drawcount}, \{maxdrawcount}, \{stride})");
		try {
		    functions.glMultiDrawArraysIndirectCount().invokeExact(mode, indirect, drawcount, maxdrawcount, stride);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiDrawArraysIndirectCountARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiDrawArraysIndirectCountARB.xhtml">registry.khronos.org</a>
	 */
	public void glMultiDrawArraysIndirectCountARB(GLPrimitiveType mode, ConstPointer<Void> indirect, long drawcount, int maxdrawcount, int stride) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiDrawArraysIndirectCountARB(\{mode}, \{indirect}, \{drawcount}, \{maxdrawcount}, \{stride})");
		try {
		    functions.glMultiDrawArraysIndirectCountARB().invokeExact(mode, indirect, drawcount, maxdrawcount, stride);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiDrawArraysIndirectEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiDrawArraysIndirectEXT.xhtml">registry.khronos.org</a>
	 */
	public void glMultiDrawArraysIndirectEXT(GLPrimitiveType mode, ConstPointer<Void> indirect, int drawcount, int stride) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiDrawArraysIndirectEXT(\{mode}, \{indirect}, \{drawcount}, \{stride})");
		try {
		    functions.glMultiDrawArraysIndirectEXT().invokeExact(mode, indirect, drawcount, stride);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiDrawElementArrayAPPLE">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiDrawElementArrayAPPLE.xhtml">registry.khronos.org</a>
	 */
	public void glMultiDrawElementArrayAPPLE(GLPrimitiveType mode, ConstPointer<Integer> first, ConstPointer<Integer> count, int primcount) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiDrawElementArrayAPPLE(\{mode}, \{first}, \{count}, \{primcount})");
		try {
		    functions.glMultiDrawElementArrayAPPLE().invokeExact(mode, first, count, primcount);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiDrawElements">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiDrawElements.xhtml">registry.khronos.org</a>
	 */
	public void glMultiDrawElements(GLPrimitiveType mode, ConstPointer<Integer> count, GLDrawElementsType type, ConstPointer<ConstPointer<Void>> indices, int drawcount) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiDrawElements(\{mode}, \{count}, \{type}, \{indices}, \{drawcount})");
		try {
		    functions.glMultiDrawElements().invokeExact(mode, count, type, indices, drawcount);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiDrawElementsBaseVertex">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiDrawElementsBaseVertex.xhtml">registry.khronos.org</a>
	 */
	public void glMultiDrawElementsBaseVertex(GLPrimitiveType mode, ConstPointer<Integer> count, GLDrawElementsType type, ConstPointer<ConstPointer<Void>> indices, int drawcount, ConstPointer<Integer> basevertex) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiDrawElementsBaseVertex(\{mode}, \{count}, \{type}, \{indices}, \{drawcount}, \{basevertex})");
		try {
		    functions.glMultiDrawElementsBaseVertex().invokeExact(mode, count, type, indices, drawcount, basevertex);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiDrawElementsBaseVertexEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiDrawElementsBaseVertexEXT.xhtml">registry.khronos.org</a>
	 */
	public void glMultiDrawElementsBaseVertexEXT(GLPrimitiveType mode, ConstPointer<Integer> count, GLDrawElementsType type, ConstPointer<ConstPointer<Void>> indices, int drawcount, ConstPointer<Integer> basevertex) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiDrawElementsBaseVertexEXT(\{mode}, \{count}, \{type}, \{indices}, \{drawcount}, \{basevertex})");
		try {
		    functions.glMultiDrawElementsBaseVertexEXT().invokeExact(mode, count, type, indices, drawcount, basevertex);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiDrawElementsEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiDrawElementsEXT.xhtml">registry.khronos.org</a>
	 */
	public void glMultiDrawElementsEXT(GLPrimitiveType mode, ConstPointer<Integer> count, GLDrawElementsType type, ConstPointer<ConstPointer<Void>> indices, int primcount) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiDrawElementsEXT(\{mode}, \{count}, \{type}, \{indices}, \{primcount})");
		try {
		    functions.glMultiDrawElementsEXT().invokeExact(mode, count, type, indices, primcount);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiDrawElementsIndirect">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiDrawElementsIndirect.xhtml">registry.khronos.org</a>
	 */
	public void glMultiDrawElementsIndirect(GLPrimitiveType mode, GLDrawElementsType type, ConstPointer<Void> indirect, int drawcount, int stride) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiDrawElementsIndirect(\{mode}, \{type}, \{indirect}, \{drawcount}, \{stride})");
		try {
		    functions.glMultiDrawElementsIndirect().invokeExact(mode, type, indirect, drawcount, stride);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiDrawElementsIndirectAMD">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiDrawElementsIndirectAMD.xhtml">registry.khronos.org</a>
	 */
	public void glMultiDrawElementsIndirectAMD(GLPrimitiveType mode, GLDrawElementsType type, ConstPointer<Void> indirect, int primcount, int stride) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiDrawElementsIndirectAMD(\{mode}, \{type}, \{indirect}, \{primcount}, \{stride})");
		try {
		    functions.glMultiDrawElementsIndirectAMD().invokeExact(mode, type, indirect, primcount, stride);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiDrawElementsIndirectBindlessCountNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiDrawElementsIndirectBindlessCountNV.xhtml">registry.khronos.org</a>
	 */
	public void glMultiDrawElementsIndirectBindlessCountNV(GLPrimitiveType mode, GLDrawElementsType type, ConstPointer<Void> indirect, int drawCount, int maxDrawCount, int stride, int vertexBufferCount) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiDrawElementsIndirectBindlessCountNV(\{mode}, \{type}, \{indirect}, \{drawCount}, \{maxDrawCount}, \{stride}, \{vertexBufferCount})");
		try {
		    functions.glMultiDrawElementsIndirectBindlessCountNV().invokeExact(mode, type, indirect, drawCount, maxDrawCount, stride, vertexBufferCount);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiDrawElementsIndirectBindlessNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiDrawElementsIndirectBindlessNV.xhtml">registry.khronos.org</a>
	 */
	public void glMultiDrawElementsIndirectBindlessNV(GLPrimitiveType mode, GLDrawElementsType type, ConstPointer<Void> indirect, int drawCount, int stride, int vertexBufferCount) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiDrawElementsIndirectBindlessNV(\{mode}, \{type}, \{indirect}, \{drawCount}, \{stride}, \{vertexBufferCount})");
		try {
		    functions.glMultiDrawElementsIndirectBindlessNV().invokeExact(mode, type, indirect, drawCount, stride, vertexBufferCount);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiDrawElementsIndirectCount">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiDrawElementsIndirectCount.xhtml">registry.khronos.org</a>
	 */
	public void glMultiDrawElementsIndirectCount(GLPrimitiveType mode, GLDrawElementsType type, ConstPointer<Void> indirect, long drawcount, int maxdrawcount, int stride) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiDrawElementsIndirectCount(\{mode}, \{type}, \{indirect}, \{drawcount}, \{maxdrawcount}, \{stride})");
		try {
		    functions.glMultiDrawElementsIndirectCount().invokeExact(mode, type, indirect, drawcount, maxdrawcount, stride);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiDrawElementsIndirectCountARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiDrawElementsIndirectCountARB.xhtml">registry.khronos.org</a>
	 */
	public void glMultiDrawElementsIndirectCountARB(GLPrimitiveType mode, GLDrawElementsType type, ConstPointer<Void> indirect, long drawcount, int maxdrawcount, int stride) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiDrawElementsIndirectCountARB(\{mode}, \{type}, \{indirect}, \{drawcount}, \{maxdrawcount}, \{stride})");
		try {
		    functions.glMultiDrawElementsIndirectCountARB().invokeExact(mode, type, indirect, drawcount, maxdrawcount, stride);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiDrawElementsIndirectEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiDrawElementsIndirectEXT.xhtml">registry.khronos.org</a>
	 */
	public void glMultiDrawElementsIndirectEXT(GLPrimitiveType mode, GLDrawElementsType type, ConstPointer<Void> indirect, int drawcount, int stride) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiDrawElementsIndirectEXT(\{mode}, \{type}, \{indirect}, \{drawcount}, \{stride})");
		try {
		    functions.glMultiDrawElementsIndirectEXT().invokeExact(mode, type, indirect, drawcount, stride);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiDrawMeshTasksIndirectNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiDrawMeshTasksIndirectNV.xhtml">registry.khronos.org</a>
	 */
	public void glMultiDrawMeshTasksIndirectNV(long indirect, int drawcount, int stride) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiDrawMeshTasksIndirectNV(\{indirect}, \{drawcount}, \{stride})");
		try {
		    functions.glMultiDrawMeshTasksIndirectNV().invokeExact(indirect, drawcount, stride);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiDrawMeshTasksIndirectCountNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiDrawMeshTasksIndirectCountNV.xhtml">registry.khronos.org</a>
	 */
	public void glMultiDrawMeshTasksIndirectCountNV(long indirect, long drawcount, int maxdrawcount, int stride) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiDrawMeshTasksIndirectCountNV(\{indirect}, \{drawcount}, \{maxdrawcount}, \{stride})");
		try {
		    functions.glMultiDrawMeshTasksIndirectCountNV().invokeExact(indirect, drawcount, maxdrawcount, stride);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiDrawRangeElementArrayAPPLE">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiDrawRangeElementArrayAPPLE.xhtml">registry.khronos.org</a>
	 */
	public void glMultiDrawRangeElementArrayAPPLE(GLPrimitiveType mode, int start, int end, ConstPointer<Integer> first, ConstPointer<Integer> count, int primcount) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiDrawRangeElementArrayAPPLE(\{mode}, \{start}, \{end}, \{first}, \{count}, \{primcount})");
		try {
		    functions.glMultiDrawRangeElementArrayAPPLE().invokeExact(mode, start, end, first, count, primcount);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiModeDrawArraysIBM">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiModeDrawArraysIBM.xhtml">registry.khronos.org</a>
	 */
	public void glMultiModeDrawArraysIBM(ConstPointer<GLPrimitiveType> mode, ConstPointer<Integer> first, ConstPointer<Integer> count, int primcount, int modestride) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiModeDrawArraysIBM(\{mode}, \{first}, \{count}, \{primcount}, \{modestride})");
		try {
		    functions.glMultiModeDrawArraysIBM().invokeExact(mode, first, count, primcount, modestride);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiModeDrawElementsIBM">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiModeDrawElementsIBM.xhtml">registry.khronos.org</a>
	 */
	public void glMultiModeDrawElementsIBM(ConstPointer<GLPrimitiveType> mode, ConstPointer<Integer> count, GLDrawElementsType type, ConstPointer<ConstPointer<Void>> indices, int primcount, int modestride) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiModeDrawElementsIBM(\{mode}, \{count}, \{type}, \{indices}, \{primcount}, \{modestride})");
		try {
		    functions.glMultiModeDrawElementsIBM().invokeExact(mode, count, type, indices, primcount, modestride);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexBufferEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexBufferEXT.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexBufferEXT(GLTextureUnit texunit, GLTextureTarget target, GLInternalFormat internalformat, int buffer) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexBufferEXT(\{texunit}, \{target}, \{internalformat}, \{buffer})");
		try {
		    functions.glMultiTexBufferEXT().invokeExact(texunit, target, internalformat, buffer);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were texunit=\{texunit}, target=\{target}, internalformat=\{internalformat}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexCoord1bOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexCoord1bOES.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexCoord1bOES(GLTextureUnit texture, byte s) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexCoord1bOES(\{texture}, \{s})");
		try {
		    functions.glMultiTexCoord1bOES().invokeExact(texture, s);
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were texture=\{texture}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexCoord1bvOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexCoord1bvOES.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexCoord1bvOES(GLTextureUnit texture, ConstPointer<Byte> coords) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexCoord1bvOES(\{texture}, \{coords})");
		try {
		    functions.glMultiTexCoord1bvOES().invokeExact(texture, coords);
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were texture=\{texture}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexCoord1d">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexCoord1d.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexCoord1d(GLTextureUnit target, double s) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexCoord1d(\{target}, \{s})");
		try {
		    functions.glMultiTexCoord1d().invokeExact(target, s);
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexCoord1dARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexCoord1dARB.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexCoord1dARB(GLTextureUnit target, double s) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexCoord1dARB(\{target}, \{s})");
		try {
		    functions.glMultiTexCoord1dARB().invokeExact(target, s);
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexCoord1dv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexCoord1dv.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexCoord1dv(GLTextureUnit target, ConstPointer<Double> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexCoord1dv(\{target}, \{v})");
		try {
		    functions.glMultiTexCoord1dv().invokeExact(target, v);
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexCoord1dvARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexCoord1dvARB.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexCoord1dvARB(GLTextureUnit target, ConstPointer<Double> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexCoord1dvARB(\{target}, \{v})");
		try {
		    functions.glMultiTexCoord1dvARB().invokeExact(target, v);
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexCoord1f">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexCoord1f.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexCoord1f(GLTextureUnit target, float s) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexCoord1f(\{target}, \{s})");
		try {
		    functions.glMultiTexCoord1f().invokeExact(target, s);
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexCoord1fARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexCoord1fARB.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexCoord1fARB(GLTextureUnit target, float s) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexCoord1fARB(\{target}, \{s})");
		try {
		    functions.glMultiTexCoord1fARB().invokeExact(target, s);
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexCoord1fv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexCoord1fv.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexCoord1fv(GLTextureUnit target, ConstPointer<Float> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexCoord1fv(\{target}, \{v})");
		try {
		    functions.glMultiTexCoord1fv().invokeExact(target, v);
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexCoord1fvARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexCoord1fvARB.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexCoord1fvARB(GLTextureUnit target, ConstPointer<Float> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexCoord1fvARB(\{target}, \{v})");
		try {
		    functions.glMultiTexCoord1fvARB().invokeExact(target, v);
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexCoord1hNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexCoord1hNV.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexCoord1hNV(GLTextureUnit target, char s) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexCoord1hNV(\{target}, \{s})");
		try {
		    functions.glMultiTexCoord1hNV().invokeExact(target, s);
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexCoord1hvNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexCoord1hvNV.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexCoord1hvNV(GLTextureUnit target, ConstPointer<Character> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexCoord1hvNV(\{target}, \{v})");
		try {
		    functions.glMultiTexCoord1hvNV().invokeExact(target, v);
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexCoord1i">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexCoord1i.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexCoord1i(GLTextureUnit target, int s) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexCoord1i(\{target}, \{s})");
		try {
		    functions.glMultiTexCoord1i().invokeExact(target, s);
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexCoord1iARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexCoord1iARB.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexCoord1iARB(GLTextureUnit target, int s) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexCoord1iARB(\{target}, \{s})");
		try {
		    functions.glMultiTexCoord1iARB().invokeExact(target, s);
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexCoord1iv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexCoord1iv.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexCoord1iv(GLTextureUnit target, ConstPointer<Integer> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexCoord1iv(\{target}, \{v})");
		try {
		    functions.glMultiTexCoord1iv().invokeExact(target, v);
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexCoord1ivARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexCoord1ivARB.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexCoord1ivARB(GLTextureUnit target, ConstPointer<Integer> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexCoord1ivARB(\{target}, \{v})");
		try {
		    functions.glMultiTexCoord1ivARB().invokeExact(target, v);
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexCoord1s">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexCoord1s.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexCoord1s(GLTextureUnit target, short s) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexCoord1s(\{target}, \{s})");
		try {
		    functions.glMultiTexCoord1s().invokeExact(target, s);
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexCoord1sARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexCoord1sARB.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexCoord1sARB(GLTextureUnit target, short s) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexCoord1sARB(\{target}, \{s})");
		try {
		    functions.glMultiTexCoord1sARB().invokeExact(target, s);
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexCoord1sv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexCoord1sv.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexCoord1sv(GLTextureUnit target, ConstPointer<Short> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexCoord1sv(\{target}, \{v})");
		try {
		    functions.glMultiTexCoord1sv().invokeExact(target, v);
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexCoord1svARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexCoord1svARB.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexCoord1svARB(GLTextureUnit target, ConstPointer<Short> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexCoord1svARB(\{target}, \{v})");
		try {
		    functions.glMultiTexCoord1svARB().invokeExact(target, v);
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexCoord1xOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexCoord1xOES.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexCoord1xOES(GLTextureUnit texture, int s) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexCoord1xOES(\{texture}, \{s})");
		try {
		    functions.glMultiTexCoord1xOES().invokeExact(texture, s);
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were texture=\{texture}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexCoord1xvOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexCoord1xvOES.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexCoord1xvOES(GLTextureUnit texture, ConstPointer<Integer> coords) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexCoord1xvOES(\{texture}, \{coords})");
		try {
		    functions.glMultiTexCoord1xvOES().invokeExact(texture, coords);
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were texture=\{texture}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexCoord2bOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexCoord2bOES.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexCoord2bOES(GLTextureUnit texture, byte s, byte t) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexCoord2bOES(\{texture}, \{s}, \{t})");
		try {
		    functions.glMultiTexCoord2bOES().invokeExact(texture, s, t);
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were texture=\{texture}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexCoord2bvOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexCoord2bvOES.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexCoord2bvOES(GLTextureUnit texture, ConstPointer<Byte> coords) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexCoord2bvOES(\{texture}, \{coords})");
		try {
		    functions.glMultiTexCoord2bvOES().invokeExact(texture, coords);
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were texture=\{texture}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexCoord2d">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexCoord2d.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexCoord2d(GLTextureUnit target, double s, double t) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexCoord2d(\{target}, \{s}, \{t})");
		try {
		    functions.glMultiTexCoord2d().invokeExact(target, s, t);
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexCoord2dARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexCoord2dARB.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexCoord2dARB(GLTextureUnit target, double s, double t) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexCoord2dARB(\{target}, \{s}, \{t})");
		try {
		    functions.glMultiTexCoord2dARB().invokeExact(target, s, t);
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexCoord2dv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexCoord2dv.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexCoord2dv(GLTextureUnit target, ConstPointer<Double> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexCoord2dv(\{target}, \{v})");
		try {
		    functions.glMultiTexCoord2dv().invokeExact(target, v);
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexCoord2dvARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexCoord2dvARB.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexCoord2dvARB(GLTextureUnit target, ConstPointer<Double> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexCoord2dvARB(\{target}, \{v})");
		try {
		    functions.glMultiTexCoord2dvARB().invokeExact(target, v);
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexCoord2f">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexCoord2f.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexCoord2f(GLTextureUnit target, float s, float t) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexCoord2f(\{target}, \{s}, \{t})");
		try {
		    functions.glMultiTexCoord2f().invokeExact(target, s, t);
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexCoord2fARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexCoord2fARB.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexCoord2fARB(GLTextureUnit target, float s, float t) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexCoord2fARB(\{target}, \{s}, \{t})");
		try {
		    functions.glMultiTexCoord2fARB().invokeExact(target, s, t);
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexCoord2fv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexCoord2fv.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexCoord2fv(GLTextureUnit target, ConstPointer<Float> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexCoord2fv(\{target}, \{v})");
		try {
		    functions.glMultiTexCoord2fv().invokeExact(target, v);
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexCoord2fvARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexCoord2fvARB.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexCoord2fvARB(GLTextureUnit target, ConstPointer<Float> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexCoord2fvARB(\{target}, \{v})");
		try {
		    functions.glMultiTexCoord2fvARB().invokeExact(target, v);
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexCoord2hNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexCoord2hNV.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexCoord2hNV(GLTextureUnit target, char s, char t) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexCoord2hNV(\{target}, \{s}, \{t})");
		try {
		    functions.glMultiTexCoord2hNV().invokeExact(target, s, t);
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexCoord2hvNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexCoord2hvNV.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexCoord2hvNV(GLTextureUnit target, ConstPointer<Character> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexCoord2hvNV(\{target}, \{v})");
		try {
		    functions.glMultiTexCoord2hvNV().invokeExact(target, v);
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexCoord2i">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexCoord2i.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexCoord2i(GLTextureUnit target, int s, int t) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexCoord2i(\{target}, \{s}, \{t})");
		try {
		    functions.glMultiTexCoord2i().invokeExact(target, s, t);
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexCoord2iARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexCoord2iARB.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexCoord2iARB(GLTextureUnit target, int s, int t) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexCoord2iARB(\{target}, \{s}, \{t})");
		try {
		    functions.glMultiTexCoord2iARB().invokeExact(target, s, t);
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexCoord2iv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexCoord2iv.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexCoord2iv(GLTextureUnit target, ConstPointer<Integer> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexCoord2iv(\{target}, \{v})");
		try {
		    functions.glMultiTexCoord2iv().invokeExact(target, v);
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexCoord2ivARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexCoord2ivARB.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexCoord2ivARB(GLTextureUnit target, ConstPointer<Integer> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexCoord2ivARB(\{target}, \{v})");
		try {
		    functions.glMultiTexCoord2ivARB().invokeExact(target, v);
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexCoord2s">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexCoord2s.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexCoord2s(GLTextureUnit target, short s, short t) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexCoord2s(\{target}, \{s}, \{t})");
		try {
		    functions.glMultiTexCoord2s().invokeExact(target, s, t);
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexCoord2sARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexCoord2sARB.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexCoord2sARB(GLTextureUnit target, short s, short t) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexCoord2sARB(\{target}, \{s}, \{t})");
		try {
		    functions.glMultiTexCoord2sARB().invokeExact(target, s, t);
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexCoord2sv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexCoord2sv.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexCoord2sv(GLTextureUnit target, ConstPointer<Short> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexCoord2sv(\{target}, \{v})");
		try {
		    functions.glMultiTexCoord2sv().invokeExact(target, v);
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexCoord2svARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexCoord2svARB.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexCoord2svARB(GLTextureUnit target, ConstPointer<Short> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexCoord2svARB(\{target}, \{v})");
		try {
		    functions.glMultiTexCoord2svARB().invokeExact(target, v);
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexCoord2xOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexCoord2xOES.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexCoord2xOES(GLTextureUnit texture, int s, int t) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexCoord2xOES(\{texture}, \{s}, \{t})");
		try {
		    functions.glMultiTexCoord2xOES().invokeExact(texture, s, t);
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were texture=\{texture}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexCoord2xvOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexCoord2xvOES.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexCoord2xvOES(GLTextureUnit texture, ConstPointer<Integer> coords) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexCoord2xvOES(\{texture}, \{coords})");
		try {
		    functions.glMultiTexCoord2xvOES().invokeExact(texture, coords);
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were texture=\{texture}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexCoord3bOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexCoord3bOES.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexCoord3bOES(GLTextureUnit texture, byte s, byte t, byte r) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexCoord3bOES(\{texture}, \{s}, \{t}, \{r})");
		try {
		    functions.glMultiTexCoord3bOES().invokeExact(texture, s, t, r);
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were texture=\{texture}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexCoord3bvOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexCoord3bvOES.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexCoord3bvOES(GLTextureUnit texture, ConstPointer<Byte> coords) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexCoord3bvOES(\{texture}, \{coords})");
		try {
		    functions.glMultiTexCoord3bvOES().invokeExact(texture, coords);
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were texture=\{texture}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexCoord3d">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexCoord3d.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexCoord3d(GLTextureUnit target, double s, double t, double r) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexCoord3d(\{target}, \{s}, \{t}, \{r})");
		try {
		    functions.glMultiTexCoord3d().invokeExact(target, s, t, r);
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexCoord3dARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexCoord3dARB.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexCoord3dARB(GLTextureUnit target, double s, double t, double r) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexCoord3dARB(\{target}, \{s}, \{t}, \{r})");
		try {
		    functions.glMultiTexCoord3dARB().invokeExact(target, s, t, r);
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexCoord3dv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexCoord3dv.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexCoord3dv(GLTextureUnit target, ConstPointer<Double> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexCoord3dv(\{target}, \{v})");
		try {
		    functions.glMultiTexCoord3dv().invokeExact(target, v);
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexCoord3dvARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexCoord3dvARB.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexCoord3dvARB(GLTextureUnit target, ConstPointer<Double> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexCoord3dvARB(\{target}, \{v})");
		try {
		    functions.glMultiTexCoord3dvARB().invokeExact(target, v);
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexCoord3f">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexCoord3f.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexCoord3f(GLTextureUnit target, float s, float t, float r) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexCoord3f(\{target}, \{s}, \{t}, \{r})");
		try {
		    functions.glMultiTexCoord3f().invokeExact(target, s, t, r);
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexCoord3fARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexCoord3fARB.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexCoord3fARB(GLTextureUnit target, float s, float t, float r) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexCoord3fARB(\{target}, \{s}, \{t}, \{r})");
		try {
		    functions.glMultiTexCoord3fARB().invokeExact(target, s, t, r);
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexCoord3fv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexCoord3fv.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexCoord3fv(GLTextureUnit target, ConstPointer<Float> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexCoord3fv(\{target}, \{v})");
		try {
		    functions.glMultiTexCoord3fv().invokeExact(target, v);
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexCoord3fvARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexCoord3fvARB.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexCoord3fvARB(GLTextureUnit target, ConstPointer<Float> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexCoord3fvARB(\{target}, \{v})");
		try {
		    functions.glMultiTexCoord3fvARB().invokeExact(target, v);
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexCoord3hNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexCoord3hNV.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexCoord3hNV(GLTextureUnit target, char s, char t, char r) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexCoord3hNV(\{target}, \{s}, \{t}, \{r})");
		try {
		    functions.glMultiTexCoord3hNV().invokeExact(target, s, t, r);
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexCoord3hvNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexCoord3hvNV.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexCoord3hvNV(GLTextureUnit target, ConstPointer<Character> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexCoord3hvNV(\{target}, \{v})");
		try {
		    functions.glMultiTexCoord3hvNV().invokeExact(target, v);
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexCoord3i">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexCoord3i.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexCoord3i(GLTextureUnit target, int s, int t, int r) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexCoord3i(\{target}, \{s}, \{t}, \{r})");
		try {
		    functions.glMultiTexCoord3i().invokeExact(target, s, t, r);
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexCoord3iARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexCoord3iARB.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexCoord3iARB(GLTextureUnit target, int s, int t, int r) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexCoord3iARB(\{target}, \{s}, \{t}, \{r})");
		try {
		    functions.glMultiTexCoord3iARB().invokeExact(target, s, t, r);
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexCoord3iv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexCoord3iv.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexCoord3iv(GLTextureUnit target, ConstPointer<Integer> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexCoord3iv(\{target}, \{v})");
		try {
		    functions.glMultiTexCoord3iv().invokeExact(target, v);
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexCoord3ivARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexCoord3ivARB.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexCoord3ivARB(GLTextureUnit target, ConstPointer<Integer> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexCoord3ivARB(\{target}, \{v})");
		try {
		    functions.glMultiTexCoord3ivARB().invokeExact(target, v);
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexCoord3s">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexCoord3s.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexCoord3s(GLTextureUnit target, short s, short t, short r) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexCoord3s(\{target}, \{s}, \{t}, \{r})");
		try {
		    functions.glMultiTexCoord3s().invokeExact(target, s, t, r);
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexCoord3sARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexCoord3sARB.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexCoord3sARB(GLTextureUnit target, short s, short t, short r) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexCoord3sARB(\{target}, \{s}, \{t}, \{r})");
		try {
		    functions.glMultiTexCoord3sARB().invokeExact(target, s, t, r);
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexCoord3sv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexCoord3sv.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexCoord3sv(GLTextureUnit target, ConstPointer<Short> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexCoord3sv(\{target}, \{v})");
		try {
		    functions.glMultiTexCoord3sv().invokeExact(target, v);
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexCoord3svARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexCoord3svARB.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexCoord3svARB(GLTextureUnit target, ConstPointer<Short> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexCoord3svARB(\{target}, \{v})");
		try {
		    functions.glMultiTexCoord3svARB().invokeExact(target, v);
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexCoord3xOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexCoord3xOES.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexCoord3xOES(GLTextureUnit texture, int s, int t, int r) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexCoord3xOES(\{texture}, \{s}, \{t}, \{r})");
		try {
		    functions.glMultiTexCoord3xOES().invokeExact(texture, s, t, r);
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were texture=\{texture}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexCoord3xvOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexCoord3xvOES.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexCoord3xvOES(GLTextureUnit texture, ConstPointer<Integer> coords) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexCoord3xvOES(\{texture}, \{coords})");
		try {
		    functions.glMultiTexCoord3xvOES().invokeExact(texture, coords);
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were texture=\{texture}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexCoord4bOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexCoord4bOES.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexCoord4bOES(GLTextureUnit texture, byte s, byte t, byte r, byte q) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexCoord4bOES(\{texture}, \{s}, \{t}, \{r}, \{q})");
		try {
		    functions.glMultiTexCoord4bOES().invokeExact(texture, s, t, r, q);
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were texture=\{texture}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexCoord4bvOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexCoord4bvOES.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexCoord4bvOES(GLTextureUnit texture, ConstPointer<Byte> coords) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexCoord4bvOES(\{texture}, \{coords})");
		try {
		    functions.glMultiTexCoord4bvOES().invokeExact(texture, coords);
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were texture=\{texture}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexCoord4d">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexCoord4d.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexCoord4d(GLTextureUnit target, double s, double t, double r, double q) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexCoord4d(\{target}, \{s}, \{t}, \{r}, \{q})");
		try {
		    functions.glMultiTexCoord4d().invokeExact(target, s, t, r, q);
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexCoord4dARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexCoord4dARB.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexCoord4dARB(GLTextureUnit target, double s, double t, double r, double q) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexCoord4dARB(\{target}, \{s}, \{t}, \{r}, \{q})");
		try {
		    functions.glMultiTexCoord4dARB().invokeExact(target, s, t, r, q);
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexCoord4dv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexCoord4dv.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexCoord4dv(GLTextureUnit target, ConstPointer<Double> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexCoord4dv(\{target}, \{v})");
		try {
		    functions.glMultiTexCoord4dv().invokeExact(target, v);
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexCoord4dvARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexCoord4dvARB.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexCoord4dvARB(GLTextureUnit target, ConstPointer<Double> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexCoord4dvARB(\{target}, \{v})");
		try {
		    functions.glMultiTexCoord4dvARB().invokeExact(target, v);
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexCoord4f">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexCoord4f.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexCoord4f(GLTextureUnit target, float s, float t, float r, float q) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexCoord4f(\{target}, \{s}, \{t}, \{r}, \{q})");
		try {
		    functions.glMultiTexCoord4f().invokeExact(target, s, t, r, q);
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexCoord4fARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexCoord4fARB.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexCoord4fARB(GLTextureUnit target, float s, float t, float r, float q) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexCoord4fARB(\{target}, \{s}, \{t}, \{r}, \{q})");
		try {
		    functions.glMultiTexCoord4fARB().invokeExact(target, s, t, r, q);
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexCoord4fv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexCoord4fv.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexCoord4fv(GLTextureUnit target, ConstPointer<Float> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexCoord4fv(\{target}, \{v})");
		try {
		    functions.glMultiTexCoord4fv().invokeExact(target, v);
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexCoord4fvARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexCoord4fvARB.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexCoord4fvARB(GLTextureUnit target, ConstPointer<Float> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexCoord4fvARB(\{target}, \{v})");
		try {
		    functions.glMultiTexCoord4fvARB().invokeExact(target, v);
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexCoord4hNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexCoord4hNV.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexCoord4hNV(GLTextureUnit target, char s, char t, char r, char q) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexCoord4hNV(\{target}, \{s}, \{t}, \{r}, \{q})");
		try {
		    functions.glMultiTexCoord4hNV().invokeExact(target, s, t, r, q);
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexCoord4hvNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexCoord4hvNV.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexCoord4hvNV(GLTextureUnit target, ConstPointer<Character> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexCoord4hvNV(\{target}, \{v})");
		try {
		    functions.glMultiTexCoord4hvNV().invokeExact(target, v);
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexCoord4i">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexCoord4i.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexCoord4i(GLTextureUnit target, int s, int t, int r, int q) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexCoord4i(\{target}, \{s}, \{t}, \{r}, \{q})");
		try {
		    functions.glMultiTexCoord4i().invokeExact(target, s, t, r, q);
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexCoord4iARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexCoord4iARB.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexCoord4iARB(GLTextureUnit target, int s, int t, int r, int q) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexCoord4iARB(\{target}, \{s}, \{t}, \{r}, \{q})");
		try {
		    functions.glMultiTexCoord4iARB().invokeExact(target, s, t, r, q);
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexCoord4iv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexCoord4iv.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexCoord4iv(GLTextureUnit target, ConstPointer<Integer> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexCoord4iv(\{target}, \{v})");
		try {
		    functions.glMultiTexCoord4iv().invokeExact(target, v);
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexCoord4ivARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexCoord4ivARB.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexCoord4ivARB(GLTextureUnit target, ConstPointer<Integer> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexCoord4ivARB(\{target}, \{v})");
		try {
		    functions.glMultiTexCoord4ivARB().invokeExact(target, v);
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexCoord4s">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexCoord4s.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexCoord4s(GLTextureUnit target, short s, short t, short r, short q) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexCoord4s(\{target}, \{s}, \{t}, \{r}, \{q})");
		try {
		    functions.glMultiTexCoord4s().invokeExact(target, s, t, r, q);
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexCoord4sARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexCoord4sARB.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexCoord4sARB(GLTextureUnit target, short s, short t, short r, short q) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexCoord4sARB(\{target}, \{s}, \{t}, \{r}, \{q})");
		try {
		    functions.glMultiTexCoord4sARB().invokeExact(target, s, t, r, q);
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexCoord4sv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexCoord4sv.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexCoord4sv(GLTextureUnit target, ConstPointer<Short> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexCoord4sv(\{target}, \{v})");
		try {
		    functions.glMultiTexCoord4sv().invokeExact(target, v);
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexCoord4svARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexCoord4svARB.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexCoord4svARB(GLTextureUnit target, ConstPointer<Short> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexCoord4svARB(\{target}, \{v})");
		try {
		    functions.glMultiTexCoord4svARB().invokeExact(target, v);
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexCoord4x">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexCoord4x.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexCoord4x(GLTextureUnit texture, int s, int t, int r, int q) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexCoord4x(\{texture}, \{s}, \{t}, \{r}, \{q})");
		try {
		    functions.glMultiTexCoord4x().invokeExact(texture, s, t, r, q);
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were texture=\{texture}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexCoord4xOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexCoord4xOES.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexCoord4xOES(GLTextureUnit texture, int s, int t, int r, int q) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexCoord4xOES(\{texture}, \{s}, \{t}, \{r}, \{q})");
		try {
		    functions.glMultiTexCoord4xOES().invokeExact(texture, s, t, r, q);
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were texture=\{texture}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexCoord4xvOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexCoord4xvOES.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexCoord4xvOES(GLTextureUnit texture, ConstPointer<Integer> coords) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexCoord4xvOES(\{texture}, \{coords})");
		try {
		    functions.glMultiTexCoord4xvOES().invokeExact(texture, coords);
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were texture=\{texture}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexCoordP1ui">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexCoordP1ui.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexCoordP1ui(GLTextureUnit texture, GLTexCoordPointerType type, int coords) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexCoordP1ui(\{texture}, \{type}, \{coords})");
		try {
		    functions.glMultiTexCoordP1ui().invokeExact(texture, type, coords);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were texture=\{texture}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexCoordP1uiv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexCoordP1uiv.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexCoordP1uiv(GLTextureUnit texture, GLTexCoordPointerType type, ConstPointer<Integer> coords) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexCoordP1uiv(\{texture}, \{type}, \{coords})");
		try {
		    functions.glMultiTexCoordP1uiv().invokeExact(texture, type, coords);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were texture=\{texture}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexCoordP2ui">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexCoordP2ui.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexCoordP2ui(GLTextureUnit texture, GLTexCoordPointerType type, int coords) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexCoordP2ui(\{texture}, \{type}, \{coords})");
		try {
		    functions.glMultiTexCoordP2ui().invokeExact(texture, type, coords);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were texture=\{texture}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexCoordP2uiv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexCoordP2uiv.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexCoordP2uiv(GLTextureUnit texture, GLTexCoordPointerType type, ConstPointer<Integer> coords) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexCoordP2uiv(\{texture}, \{type}, \{coords})");
		try {
		    functions.glMultiTexCoordP2uiv().invokeExact(texture, type, coords);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were texture=\{texture}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexCoordP3ui">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexCoordP3ui.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexCoordP3ui(GLTextureUnit texture, GLTexCoordPointerType type, int coords) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexCoordP3ui(\{texture}, \{type}, \{coords})");
		try {
		    functions.glMultiTexCoordP3ui().invokeExact(texture, type, coords);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were texture=\{texture}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexCoordP3uiv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexCoordP3uiv.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexCoordP3uiv(GLTextureUnit texture, GLTexCoordPointerType type, ConstPointer<Integer> coords) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexCoordP3uiv(\{texture}, \{type}, \{coords})");
		try {
		    functions.glMultiTexCoordP3uiv().invokeExact(texture, type, coords);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were texture=\{texture}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexCoordP4ui">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexCoordP4ui.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexCoordP4ui(GLTextureUnit texture, GLTexCoordPointerType type, int coords) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexCoordP4ui(\{texture}, \{type}, \{coords})");
		try {
		    functions.glMultiTexCoordP4ui().invokeExact(texture, type, coords);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were texture=\{texture}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexCoordP4uiv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexCoordP4uiv.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexCoordP4uiv(GLTextureUnit texture, GLTexCoordPointerType type, ConstPointer<Integer> coords) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexCoordP4uiv(\{texture}, \{type}, \{coords})");
		try {
		    functions.glMultiTexCoordP4uiv().invokeExact(texture, type, coords);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were texture=\{texture}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexCoordPointerEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexCoordPointerEXT.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexCoordPointerEXT(GLTextureUnit texunit, int size, GLTexCoordPointerType type, int stride, ConstPointer<Void> pointer) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexCoordPointerEXT(\{texunit}, \{size}, \{type}, \{stride}, \{pointer})");
		try {
		    functions.glMultiTexCoordPointerEXT().invokeExact(texunit, size, type, stride, pointer);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were texunit=\{texunit}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexEnvfEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexEnvfEXT.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexEnvfEXT(GLTextureUnit texunit, GLTextureEnvTarget target, GLTextureEnvParameter pname, float param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexEnvfEXT(\{texunit}, \{target}, \{pname}, \{param})");
		try {
		    functions.glMultiTexEnvfEXT().invokeExact(texunit, target, pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were texunit=\{texunit}, target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexEnvfvEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexEnvfvEXT.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexEnvfvEXT(GLTextureUnit texunit, GLTextureEnvTarget target, GLTextureEnvParameter pname, ConstPointer<Float> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexEnvfvEXT(\{texunit}, \{target}, \{pname}, \{params})");
		try {
		    functions.glMultiTexEnvfvEXT().invokeExact(texunit, target, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were texunit=\{texunit}, target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexEnviEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexEnviEXT.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexEnviEXT(GLTextureUnit texunit, GLTextureEnvTarget target, GLTextureEnvParameter pname, int param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexEnviEXT(\{texunit}, \{target}, \{pname}, \{param})");
		try {
		    functions.glMultiTexEnviEXT().invokeExact(texunit, target, pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were texunit=\{texunit}, target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexEnvivEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexEnvivEXT.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexEnvivEXT(GLTextureUnit texunit, GLTextureEnvTarget target, GLTextureEnvParameter pname, ConstPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexEnvivEXT(\{texunit}, \{target}, \{pname}, \{params})");
		try {
		    functions.glMultiTexEnvivEXT().invokeExact(texunit, target, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were texunit=\{texunit}, target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexGendEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexGendEXT.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexGendEXT(GLTextureUnit texunit, GLTextureCoordName coord, GLTextureGenParameter pname, double param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexGendEXT(\{texunit}, \{coord}, \{pname}, \{param})");
		try {
		    functions.glMultiTexGendEXT().invokeExact(texunit, coord, pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were texunit=\{texunit}, coord=\{coord}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexGendvEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexGendvEXT.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexGendvEXT(GLTextureUnit texunit, GLTextureCoordName coord, GLTextureGenParameter pname, ConstPointer<Double> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexGendvEXT(\{texunit}, \{coord}, \{pname}, \{params})");
		try {
		    functions.glMultiTexGendvEXT().invokeExact(texunit, coord, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were texunit=\{texunit}, coord=\{coord}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexGenfEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexGenfEXT.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexGenfEXT(GLTextureUnit texunit, GLTextureCoordName coord, GLTextureGenParameter pname, float param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexGenfEXT(\{texunit}, \{coord}, \{pname}, \{param})");
		try {
		    functions.glMultiTexGenfEXT().invokeExact(texunit, coord, pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were texunit=\{texunit}, coord=\{coord}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexGenfvEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexGenfvEXT.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexGenfvEXT(GLTextureUnit texunit, GLTextureCoordName coord, GLTextureGenParameter pname, ConstPointer<Float> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexGenfvEXT(\{texunit}, \{coord}, \{pname}, \{params})");
		try {
		    functions.glMultiTexGenfvEXT().invokeExact(texunit, coord, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were texunit=\{texunit}, coord=\{coord}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexGeniEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexGeniEXT.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexGeniEXT(GLTextureUnit texunit, GLTextureCoordName coord, GLTextureGenParameter pname, int param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexGeniEXT(\{texunit}, \{coord}, \{pname}, \{param})");
		try {
		    functions.glMultiTexGeniEXT().invokeExact(texunit, coord, pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were texunit=\{texunit}, coord=\{coord}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexGenivEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexGenivEXT.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexGenivEXT(GLTextureUnit texunit, GLTextureCoordName coord, GLTextureGenParameter pname, ConstPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexGenivEXT(\{texunit}, \{coord}, \{pname}, \{params})");
		try {
		    functions.glMultiTexGenivEXT().invokeExact(texunit, coord, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were texunit=\{texunit}, coord=\{coord}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexImage1DEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexImage1DEXT.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexImage1DEXT(GLTextureUnit texunit, GLTextureTarget target, int level, GLInternalFormat internalformat, int width, int border, GLPixelFormat format, GLPixelType type, ConstPointer<Void> pixels) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexImage1DEXT(\{texunit}, \{target}, \{level}, \{internalformat}, \{width}, \{border}, \{format}, \{type}, \{pixels})");
		try {
		    functions.glMultiTexImage1DEXT().invokeExact(texunit, target, level, internalformat, width, border, format, type, pixels);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were texunit=\{texunit}, target=\{target}, internalformat=\{internalformat}, format=\{format}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexImage2DEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexImage2DEXT.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexImage2DEXT(GLTextureUnit texunit, GLTextureTarget target, int level, GLInternalFormat internalformat, int width, int height, int border, GLPixelFormat format, GLPixelType type, ConstPointer<Void> pixels) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexImage2DEXT(\{texunit}, \{target}, \{level}, \{internalformat}, \{width}, \{height}, \{border}, \{format}, \{type}, \{pixels})");
		try {
		    functions.glMultiTexImage2DEXT().invokeExact(texunit, target, level, internalformat, width, height, border, format, type, pixels);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were texunit=\{texunit}, target=\{target}, internalformat=\{internalformat}, format=\{format}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexImage3DEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexImage3DEXT.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexImage3DEXT(GLTextureUnit texunit, GLTextureTarget target, int level, GLInternalFormat internalformat, int width, int height, int depth, int border, GLPixelFormat format, GLPixelType type, ConstPointer<Void> pixels) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexImage3DEXT(\{texunit}, \{target}, \{level}, \{internalformat}, \{width}, \{height}, \{depth}, \{border}, \{format}, \{type}, \{pixels})");
		try {
		    functions.glMultiTexImage3DEXT().invokeExact(texunit, target, level, internalformat, width, height, depth, border, format, type, pixels);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were texunit=\{texunit}, target=\{target}, internalformat=\{internalformat}, format=\{format}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexParameterIivEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexParameterIivEXT.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexParameterIivEXT(GLTextureUnit texunit, GLTextureTarget target, GLTextureParameterName pname, ConstPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexParameterIivEXT(\{texunit}, \{target}, \{pname}, \{params})");
		try {
		    functions.glMultiTexParameterIivEXT().invokeExact(texunit, target, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were texunit=\{texunit}, target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexParameterIuivEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexParameterIuivEXT.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexParameterIuivEXT(GLTextureUnit texunit, GLTextureTarget target, GLTextureParameterName pname, ConstPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexParameterIuivEXT(\{texunit}, \{target}, \{pname}, \{params})");
		try {
		    functions.glMultiTexParameterIuivEXT().invokeExact(texunit, target, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were texunit=\{texunit}, target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexParameterfEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexParameterfEXT.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexParameterfEXT(GLTextureUnit texunit, GLTextureTarget target, GLTextureParameterName pname, float param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexParameterfEXT(\{texunit}, \{target}, \{pname}, \{param})");
		try {
		    functions.glMultiTexParameterfEXT().invokeExact(texunit, target, pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were texunit=\{texunit}, target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexParameterfvEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexParameterfvEXT.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexParameterfvEXT(GLTextureUnit texunit, GLTextureTarget target, GLTextureParameterName pname, ConstPointer<Float> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexParameterfvEXT(\{texunit}, \{target}, \{pname}, \{params})");
		try {
		    functions.glMultiTexParameterfvEXT().invokeExact(texunit, target, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were texunit=\{texunit}, target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexParameteriEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexParameteriEXT.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexParameteriEXT(GLTextureUnit texunit, GLTextureTarget target, GLTextureParameterName pname, int param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexParameteriEXT(\{texunit}, \{target}, \{pname}, \{param})");
		try {
		    functions.glMultiTexParameteriEXT().invokeExact(texunit, target, pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were texunit=\{texunit}, target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexParameterivEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexParameterivEXT.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexParameterivEXT(GLTextureUnit texunit, GLTextureTarget target, GLTextureParameterName pname, ConstPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexParameterivEXT(\{texunit}, \{target}, \{pname}, \{params})");
		try {
		    functions.glMultiTexParameterivEXT().invokeExact(texunit, target, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were texunit=\{texunit}, target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexRenderbufferEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexRenderbufferEXT.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexRenderbufferEXT(GLTextureUnit texunit, GLTextureTarget target, int renderbuffer) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexRenderbufferEXT(\{texunit}, \{target}, \{renderbuffer})");
		try {
		    functions.glMultiTexRenderbufferEXT().invokeExact(texunit, target, renderbuffer);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were texunit=\{texunit}, target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexSubImage1DEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexSubImage1DEXT.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexSubImage1DEXT(GLTextureUnit texunit, GLTextureTarget target, int level, int xoffset, int width, GLPixelFormat format, GLPixelType type, ConstPointer<Void> pixels) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexSubImage1DEXT(\{texunit}, \{target}, \{level}, \{xoffset}, \{width}, \{format}, \{type}, \{pixels})");
		try {
		    functions.glMultiTexSubImage1DEXT().invokeExact(texunit, target, level, xoffset, width, format, type, pixels);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were texunit=\{texunit}, target=\{target}, format=\{format}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexSubImage2DEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexSubImage2DEXT.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexSubImage2DEXT(GLTextureUnit texunit, GLTextureTarget target, int level, int xoffset, int yoffset, int width, int height, GLPixelFormat format, GLPixelType type, ConstPointer<Void> pixels) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexSubImage2DEXT(\{texunit}, \{target}, \{level}, \{xoffset}, \{yoffset}, \{width}, \{height}, \{format}, \{type}, \{pixels})");
		try {
		    functions.glMultiTexSubImage2DEXT().invokeExact(texunit, target, level, xoffset, yoffset, width, height, format, type, pixels);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were texunit=\{texunit}, target=\{target}, format=\{format}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMultiTexSubImage3DEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMultiTexSubImage3DEXT.xhtml">registry.khronos.org</a>
	 */
	public void glMultiTexSubImage3DEXT(GLTextureUnit texunit, GLTextureTarget target, int level, int xoffset, int yoffset, int zoffset, int width, int height, int depth, GLPixelFormat format, GLPixelType type, ConstPointer<Void> pixels) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMultiTexSubImage3DEXT(\{texunit}, \{target}, \{level}, \{xoffset}, \{yoffset}, \{zoffset}, \{width}, \{height}, \{depth}, \{format}, \{type}, \{pixels})");
		try {
		    functions.glMultiTexSubImage3DEXT().invokeExact(texunit, target, level, xoffset, yoffset, zoffset, width, height, depth, format, type, pixels);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were texunit=\{texunit}, target=\{target}, format=\{format}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMulticastBarrierNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMulticastBarrierNV.xhtml">registry.khronos.org</a>
	 */
	public void glMulticastBarrierNV() {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMulticastBarrierNV()");
		try {
		    functions.glMulticastBarrierNV().invokeExact();
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMulticastBlitFramebufferNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMulticastBlitFramebufferNV.xhtml">registry.khronos.org</a>
	 */
	public void glMulticastBlitFramebufferNV(int srcGpu, int dstGpu, int srcX0, int srcY0, int srcX1, int srcY1, int dstX0, int dstY0, int dstX1, int dstY1, int mask, GLEnum filter) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMulticastBlitFramebufferNV(\{srcGpu}, \{dstGpu}, \{srcX0}, \{srcY0}, \{srcX1}, \{srcY1}, \{dstX0}, \{dstY0}, \{dstX1}, \{dstY1}, \{mask}, \{filter})");
		try {
		    functions.glMulticastBlitFramebufferNV().invokeExact(srcGpu, dstGpu, srcX0, srcY0, srcX1, srcY1, dstX0, dstY0, dstX1, dstY1, mask, filter);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were filter=\{filter}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMulticastBufferSubDataNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMulticastBufferSubDataNV.xhtml">registry.khronos.org</a>
	 */
	public void glMulticastBufferSubDataNV(int gpuMask, int buffer, long offset, long size, ConstPointer<Void> data) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMulticastBufferSubDataNV(\{gpuMask}, \{buffer}, \{offset}, \{size}, \{data})");
		try {
		    functions.glMulticastBufferSubDataNV().invokeExact(gpuMask, buffer, offset, size, data);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMulticastCopyBufferSubDataNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMulticastCopyBufferSubDataNV.xhtml">registry.khronos.org</a>
	 */
	public void glMulticastCopyBufferSubDataNV(int readGpu, int writeGpuMask, int readBuffer, int writeBuffer, long readOffset, long writeOffset, long size) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMulticastCopyBufferSubDataNV(\{readGpu}, \{writeGpuMask}, \{readBuffer}, \{writeBuffer}, \{readOffset}, \{writeOffset}, \{size})");
		try {
		    functions.glMulticastCopyBufferSubDataNV().invokeExact(readGpu, writeGpuMask, readBuffer, writeBuffer, readOffset, writeOffset, size);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMulticastCopyImageSubDataNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMulticastCopyImageSubDataNV.xhtml">registry.khronos.org</a>
	 */
	public void glMulticastCopyImageSubDataNV(int srcGpu, int dstGpuMask, int srcName, GLEnum srcTarget, int srcLevel, int srcX, int srcY, int srcZ, int dstName, GLEnum dstTarget, int dstLevel, int dstX, int dstY, int dstZ, int srcWidth, int srcHeight, int srcDepth) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMulticastCopyImageSubDataNV(\{srcGpu}, \{dstGpuMask}, \{srcName}, \{srcTarget}, \{srcLevel}, \{srcX}, \{srcY}, \{srcZ}, \{dstName}, \{dstTarget}, \{dstLevel}, \{dstX}, \{dstY}, \{dstZ}, \{srcWidth}, \{srcHeight}, \{srcDepth})");
		try {
		    functions.glMulticastCopyImageSubDataNV().invokeExact(srcGpu, dstGpuMask, srcName, srcTarget, srcLevel, srcX, srcY, srcZ, dstName, dstTarget, dstLevel, dstX, dstY, dstZ, srcWidth, srcHeight, srcDepth);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were srcTarget=\{srcTarget}, dstTarget=\{dstTarget}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMulticastFramebufferSampleLocationsfvNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMulticastFramebufferSampleLocationsfvNV.xhtml">registry.khronos.org</a>
	 */
	public void glMulticastFramebufferSampleLocationsfvNV(int gpu, int framebuffer, int start, int count, ConstPointer<Float> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMulticastFramebufferSampleLocationsfvNV(\{gpu}, \{framebuffer}, \{start}, \{count}, \{v})");
		try {
		    functions.glMulticastFramebufferSampleLocationsfvNV().invokeExact(gpu, framebuffer, start, count, v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMulticastGetQueryObjecti64vNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMulticastGetQueryObjecti64vNV.xhtml">registry.khronos.org</a>
	 */
	public void glMulticastGetQueryObjecti64vNV(int gpu, int id, GLEnum pname, MutPointer<Long> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMulticastGetQueryObjecti64vNV(\{gpu}, \{id}, \{pname}, \{params})");
		try {
		    functions.glMulticastGetQueryObjecti64vNV().invokeExact(gpu, id, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMulticastGetQueryObjectivNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMulticastGetQueryObjectivNV.xhtml">registry.khronos.org</a>
	 */
	public void glMulticastGetQueryObjectivNV(int gpu, int id, GLEnum pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMulticastGetQueryObjectivNV(\{gpu}, \{id}, \{pname}, \{params})");
		try {
		    functions.glMulticastGetQueryObjectivNV().invokeExact(gpu, id, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMulticastGetQueryObjectui64vNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMulticastGetQueryObjectui64vNV.xhtml">registry.khronos.org</a>
	 */
	public void glMulticastGetQueryObjectui64vNV(int gpu, int id, GLEnum pname, MutPointer<Long> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMulticastGetQueryObjectui64vNV(\{gpu}, \{id}, \{pname}, \{params})");
		try {
		    functions.glMulticastGetQueryObjectui64vNV().invokeExact(gpu, id, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMulticastGetQueryObjectuivNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMulticastGetQueryObjectuivNV.xhtml">registry.khronos.org</a>
	 */
	public void glMulticastGetQueryObjectuivNV(int gpu, int id, GLEnum pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMulticastGetQueryObjectuivNV(\{gpu}, \{id}, \{pname}, \{params})");
		try {
		    functions.glMulticastGetQueryObjectuivNV().invokeExact(gpu, id, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMulticastScissorArrayvNVX">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMulticastScissorArrayvNVX.xhtml">registry.khronos.org</a>
	 */
	public void glMulticastScissorArrayvNVX(int gpu, int first, int count, ConstPointer<Integer> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMulticastScissorArrayvNVX(\{gpu}, \{first}, \{count}, \{v})");
		try {
		    functions.glMulticastScissorArrayvNVX().invokeExact(gpu, first, count, v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMulticastViewportArrayvNVX">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMulticastViewportArrayvNVX.xhtml">registry.khronos.org</a>
	 */
	public void glMulticastViewportArrayvNVX(int gpu, int first, int count, ConstPointer<Float> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMulticastViewportArrayvNVX(\{gpu}, \{first}, \{count}, \{v})");
		try {
		    functions.glMulticastViewportArrayvNVX().invokeExact(gpu, first, count, v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMulticastViewportPositionWScaleNVX">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMulticastViewportPositionWScaleNVX.xhtml">registry.khronos.org</a>
	 */
	public void glMulticastViewportPositionWScaleNVX(int gpu, int index, float xcoeff, float ycoeff) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMulticastViewportPositionWScaleNVX(\{gpu}, \{index}, \{xcoeff}, \{ycoeff})");
		try {
		    functions.glMulticastViewportPositionWScaleNVX().invokeExact(gpu, index, xcoeff, ycoeff);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glMulticastWaitSyncNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glMulticastWaitSyncNV.xhtml">registry.khronos.org</a>
	 */
	public void glMulticastWaitSyncNV(int signalGpu, int waitGpuMask) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glMulticastWaitSyncNV(\{signalGpu}, \{waitGpuMask})");
		try {
		    functions.glMulticastWaitSyncNV().invokeExact(signalGpu, waitGpuMask);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glNamedBufferAttachMemoryNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glNamedBufferAttachMemoryNV.xhtml">registry.khronos.org</a>
	 */
	public void glNamedBufferAttachMemoryNV(int buffer, int memory, long offset) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glNamedBufferAttachMemoryNV(\{buffer}, \{memory}, \{offset})");
		try {
		    functions.glNamedBufferAttachMemoryNV().invokeExact(buffer, memory, offset);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glNamedBufferData">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glNamedBufferData.xhtml">registry.khronos.org</a>
	 */
	public void glNamedBufferData(int buffer, long size, ConstPointer<Void> data, GLVertexBufferObjectUsage usage) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glNamedBufferData(\{buffer}, \{size}, \{data}, \{usage})");
		try {
		    functions.glNamedBufferData().invokeExact(buffer, size, data, usage);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were usage=\{usage}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glNamedBufferDataEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glNamedBufferDataEXT.xhtml">registry.khronos.org</a>
	 */
	public void glNamedBufferDataEXT(int buffer, long size, ConstPointer<Void> data, GLVertexBufferObjectUsage usage) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glNamedBufferDataEXT(\{buffer}, \{size}, \{data}, \{usage})");
		try {
		    functions.glNamedBufferDataEXT().invokeExact(buffer, size, data, usage);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were usage=\{usage}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glNamedBufferPageCommitmentARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glNamedBufferPageCommitmentARB.xhtml">registry.khronos.org</a>
	 */
	public void glNamedBufferPageCommitmentARB(int buffer, long offset, long size, boolean commit) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glNamedBufferPageCommitmentARB(\{buffer}, \{offset}, \{size}, \{commit})");
		try {
		    functions.glNamedBufferPageCommitmentARB().invokeExact(buffer, offset, size, commit);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glNamedBufferPageCommitmentEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glNamedBufferPageCommitmentEXT.xhtml">registry.khronos.org</a>
	 */
	public void glNamedBufferPageCommitmentEXT(int buffer, long offset, long size, boolean commit) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glNamedBufferPageCommitmentEXT(\{buffer}, \{offset}, \{size}, \{commit})");
		try {
		    functions.glNamedBufferPageCommitmentEXT().invokeExact(buffer, offset, size, commit);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glNamedBufferPageCommitmentMemNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glNamedBufferPageCommitmentMemNV.xhtml">registry.khronos.org</a>
	 */
	public void glNamedBufferPageCommitmentMemNV(int buffer, long offset, long size, int memory, long memOffset, boolean commit) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glNamedBufferPageCommitmentMemNV(\{buffer}, \{offset}, \{size}, \{memory}, \{memOffset}, \{commit})");
		try {
		    functions.glNamedBufferPageCommitmentMemNV().invokeExact(buffer, offset, size, memory, memOffset, commit);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glNamedBufferStorage">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glNamedBufferStorage.xhtml">registry.khronos.org</a>
	 */
	public void glNamedBufferStorage(int buffer, long size, ConstPointer<Void> data, int flags) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glNamedBufferStorage(\{buffer}, \{size}, \{data}, \{flags})");
		try {
		    functions.glNamedBufferStorage().invokeExact(buffer, size, data, flags);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glNamedBufferStorageExternalEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glNamedBufferStorageExternalEXT.xhtml">registry.khronos.org</a>
	 */
	public void glNamedBufferStorageExternalEXT(int buffer, long offset, long size, int clientBuffer, int flags) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glNamedBufferStorageExternalEXT(\{buffer}, \{offset}, \{size}, \{clientBuffer}, \{flags})");
		try {
		    functions.glNamedBufferStorageExternalEXT().invokeExact(buffer, offset, size, clientBuffer, flags);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glNamedBufferStorageEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glNamedBufferStorageEXT.xhtml">registry.khronos.org</a>
	 */
	public void glNamedBufferStorageEXT(int buffer, long size, ConstPointer<Void> data, int flags) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glNamedBufferStorageEXT(\{buffer}, \{size}, \{data}, \{flags})");
		try {
		    functions.glNamedBufferStorageEXT().invokeExact(buffer, size, data, flags);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glNamedBufferStorageMemEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glNamedBufferStorageMemEXT.xhtml">registry.khronos.org</a>
	 */
	public void glNamedBufferStorageMemEXT(int buffer, long size, int memory, long offset) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glNamedBufferStorageMemEXT(\{buffer}, \{size}, \{memory}, \{offset})");
		try {
		    functions.glNamedBufferStorageMemEXT().invokeExact(buffer, size, memory, offset);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glNamedBufferSubData">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glNamedBufferSubData.xhtml">registry.khronos.org</a>
	 */
	public void glNamedBufferSubData(int buffer, long offset, long size, ConstPointer<Void> data) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glNamedBufferSubData(\{buffer}, \{offset}, \{size}, \{data})");
		try {
		    functions.glNamedBufferSubData().invokeExact(buffer, offset, size, data);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glNamedBufferSubDataEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glNamedBufferSubDataEXT.xhtml">registry.khronos.org</a>
	 */
	public void glNamedBufferSubDataEXT(int buffer, long offset, long size, ConstPointer<Void> data) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glNamedBufferSubDataEXT(\{buffer}, \{offset}, \{size}, \{data})");
		try {
		    functions.glNamedBufferSubDataEXT().invokeExact(buffer, offset, size, data);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glNamedCopyBufferSubDataEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glNamedCopyBufferSubDataEXT.xhtml">registry.khronos.org</a>
	 */
	public void glNamedCopyBufferSubDataEXT(int readBuffer, int writeBuffer, long readOffset, long writeOffset, long size) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glNamedCopyBufferSubDataEXT(\{readBuffer}, \{writeBuffer}, \{readOffset}, \{writeOffset}, \{size})");
		try {
		    functions.glNamedCopyBufferSubDataEXT().invokeExact(readBuffer, writeBuffer, readOffset, writeOffset, size);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glNamedFramebufferDrawBuffer">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glNamedFramebufferDrawBuffer.xhtml">registry.khronos.org</a>
	 */
	public void glNamedFramebufferDrawBuffer(int framebuffer, GLColorBuffer buf) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glNamedFramebufferDrawBuffer(\{framebuffer}, \{buf})");
		try {
		    functions.glNamedFramebufferDrawBuffer().invokeExact(framebuffer, buf);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were buf=\{buf}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glNamedFramebufferDrawBuffers">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glNamedFramebufferDrawBuffers.xhtml">registry.khronos.org</a>
	 */
	public void glNamedFramebufferDrawBuffers(int framebuffer, int n, ConstPointer<GLColorBuffer> bufs) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glNamedFramebufferDrawBuffers(\{framebuffer}, \{n}, \{bufs})");
		try {
		    functions.glNamedFramebufferDrawBuffers().invokeExact(framebuffer, n, bufs);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glNamedFramebufferParameteri">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glNamedFramebufferParameteri.xhtml">registry.khronos.org</a>
	 */
	public void glNamedFramebufferParameteri(int framebuffer, GLFramebufferParameterName pname, int param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glNamedFramebufferParameteri(\{framebuffer}, \{pname}, \{param})");
		try {
		    functions.glNamedFramebufferParameteri().invokeExact(framebuffer, pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glNamedFramebufferParameteriEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glNamedFramebufferParameteriEXT.xhtml">registry.khronos.org</a>
	 */
	public void glNamedFramebufferParameteriEXT(int framebuffer, GLFramebufferParameterName pname, int param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glNamedFramebufferParameteriEXT(\{framebuffer}, \{pname}, \{param})");
		try {
		    functions.glNamedFramebufferParameteriEXT().invokeExact(framebuffer, pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glNamedFramebufferReadBuffer">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glNamedFramebufferReadBuffer.xhtml">registry.khronos.org</a>
	 */
	public void glNamedFramebufferReadBuffer(int framebuffer, GLColorBuffer src) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glNamedFramebufferReadBuffer(\{framebuffer}, \{src})");
		try {
		    functions.glNamedFramebufferReadBuffer().invokeExact(framebuffer, src);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were src=\{src}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glNamedFramebufferRenderbuffer">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glNamedFramebufferRenderbuffer.xhtml">registry.khronos.org</a>
	 */
	public void glNamedFramebufferRenderbuffer(int framebuffer, GLFramebufferAttachment attachment, GLRenderbufferTarget renderbuffertarget, int renderbuffer) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glNamedFramebufferRenderbuffer(\{framebuffer}, \{attachment}, \{renderbuffertarget}, \{renderbuffer})");
		try {
		    functions.glNamedFramebufferRenderbuffer().invokeExact(framebuffer, attachment, renderbuffertarget, renderbuffer);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were attachment=\{attachment}, renderbuffertarget=\{renderbuffertarget}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glNamedFramebufferRenderbufferEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glNamedFramebufferRenderbufferEXT.xhtml">registry.khronos.org</a>
	 */
	public void glNamedFramebufferRenderbufferEXT(int framebuffer, GLFramebufferAttachment attachment, GLRenderbufferTarget renderbuffertarget, int renderbuffer) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glNamedFramebufferRenderbufferEXT(\{framebuffer}, \{attachment}, \{renderbuffertarget}, \{renderbuffer})");
		try {
		    functions.glNamedFramebufferRenderbufferEXT().invokeExact(framebuffer, attachment, renderbuffertarget, renderbuffer);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were attachment=\{attachment}, renderbuffertarget=\{renderbuffertarget}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glNamedFramebufferSampleLocationsfvARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glNamedFramebufferSampleLocationsfvARB.xhtml">registry.khronos.org</a>
	 */
	public void glNamedFramebufferSampleLocationsfvARB(int framebuffer, int start, int count, ConstPointer<Float> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glNamedFramebufferSampleLocationsfvARB(\{framebuffer}, \{start}, \{count}, \{v})");
		try {
		    functions.glNamedFramebufferSampleLocationsfvARB().invokeExact(framebuffer, start, count, v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glNamedFramebufferSampleLocationsfvNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glNamedFramebufferSampleLocationsfvNV.xhtml">registry.khronos.org</a>
	 */
	public void glNamedFramebufferSampleLocationsfvNV(int framebuffer, int start, int count, ConstPointer<Float> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glNamedFramebufferSampleLocationsfvNV(\{framebuffer}, \{start}, \{count}, \{v})");
		try {
		    functions.glNamedFramebufferSampleLocationsfvNV().invokeExact(framebuffer, start, count, v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glNamedFramebufferTexture">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glNamedFramebufferTexture.xhtml">registry.khronos.org</a>
	 */
	public void glNamedFramebufferTexture(int framebuffer, GLFramebufferAttachment attachment, int texture, int level) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glNamedFramebufferTexture(\{framebuffer}, \{attachment}, \{texture}, \{level})");
		try {
		    functions.glNamedFramebufferTexture().invokeExact(framebuffer, attachment, texture, level);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were attachment=\{attachment}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glNamedFramebufferSamplePositionsfvAMD">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glNamedFramebufferSamplePositionsfvAMD.xhtml">registry.khronos.org</a>
	 */
	public void glNamedFramebufferSamplePositionsfvAMD(int framebuffer, int numsamples, int pixelindex, ConstPointer<Float> values) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glNamedFramebufferSamplePositionsfvAMD(\{framebuffer}, \{numsamples}, \{pixelindex}, \{values})");
		try {
		    functions.glNamedFramebufferSamplePositionsfvAMD().invokeExact(framebuffer, numsamples, pixelindex, values);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glNamedFramebufferTexture1DEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glNamedFramebufferTexture1DEXT.xhtml">registry.khronos.org</a>
	 */
	public void glNamedFramebufferTexture1DEXT(int framebuffer, GLFramebufferAttachment attachment, GLTextureTarget textarget, int texture, int level) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glNamedFramebufferTexture1DEXT(\{framebuffer}, \{attachment}, \{textarget}, \{texture}, \{level})");
		try {
		    functions.glNamedFramebufferTexture1DEXT().invokeExact(framebuffer, attachment, textarget, texture, level);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were attachment=\{attachment}, textarget=\{textarget}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glNamedFramebufferTexture2DEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glNamedFramebufferTexture2DEXT.xhtml">registry.khronos.org</a>
	 */
	public void glNamedFramebufferTexture2DEXT(int framebuffer, GLFramebufferAttachment attachment, GLTextureTarget textarget, int texture, int level) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glNamedFramebufferTexture2DEXT(\{framebuffer}, \{attachment}, \{textarget}, \{texture}, \{level})");
		try {
		    functions.glNamedFramebufferTexture2DEXT().invokeExact(framebuffer, attachment, textarget, texture, level);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were attachment=\{attachment}, textarget=\{textarget}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glNamedFramebufferTexture3DEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glNamedFramebufferTexture3DEXT.xhtml">registry.khronos.org</a>
	 */
	public void glNamedFramebufferTexture3DEXT(int framebuffer, GLFramebufferAttachment attachment, GLTextureTarget textarget, int texture, int level, int zoffset) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glNamedFramebufferTexture3DEXT(\{framebuffer}, \{attachment}, \{textarget}, \{texture}, \{level}, \{zoffset})");
		try {
		    functions.glNamedFramebufferTexture3DEXT().invokeExact(framebuffer, attachment, textarget, texture, level, zoffset);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were attachment=\{attachment}, textarget=\{textarget}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glNamedFramebufferTextureEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glNamedFramebufferTextureEXT.xhtml">registry.khronos.org</a>
	 */
	public void glNamedFramebufferTextureEXT(int framebuffer, GLFramebufferAttachment attachment, int texture, int level) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glNamedFramebufferTextureEXT(\{framebuffer}, \{attachment}, \{texture}, \{level})");
		try {
		    functions.glNamedFramebufferTextureEXT().invokeExact(framebuffer, attachment, texture, level);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were attachment=\{attachment}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glNamedFramebufferTextureFaceEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glNamedFramebufferTextureFaceEXT.xhtml">registry.khronos.org</a>
	 */
	public void glNamedFramebufferTextureFaceEXT(int framebuffer, GLFramebufferAttachment attachment, int texture, int level, GLTextureTarget face) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glNamedFramebufferTextureFaceEXT(\{framebuffer}, \{attachment}, \{texture}, \{level}, \{face})");
		try {
		    functions.glNamedFramebufferTextureFaceEXT().invokeExact(framebuffer, attachment, texture, level, face);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were attachment=\{attachment}, face=\{face}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glNamedFramebufferTextureLayer">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glNamedFramebufferTextureLayer.xhtml">registry.khronos.org</a>
	 */
	public void glNamedFramebufferTextureLayer(int framebuffer, GLFramebufferAttachment attachment, int texture, int level, int layer) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glNamedFramebufferTextureLayer(\{framebuffer}, \{attachment}, \{texture}, \{level}, \{layer})");
		try {
		    functions.glNamedFramebufferTextureLayer().invokeExact(framebuffer, attachment, texture, level, layer);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were attachment=\{attachment}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glNamedFramebufferTextureLayerEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glNamedFramebufferTextureLayerEXT.xhtml">registry.khronos.org</a>
	 */
	public void glNamedFramebufferTextureLayerEXT(int framebuffer, GLFramebufferAttachment attachment, int texture, int level, int layer) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glNamedFramebufferTextureLayerEXT(\{framebuffer}, \{attachment}, \{texture}, \{level}, \{layer})");
		try {
		    functions.glNamedFramebufferTextureLayerEXT().invokeExact(framebuffer, attachment, texture, level, layer);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were attachment=\{attachment}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glNamedFramebufferTextureMultiviewOVR">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glNamedFramebufferTextureMultiviewOVR.xhtml">registry.khronos.org</a>
	 */
	public void glNamedFramebufferTextureMultiviewOVR(int framebuffer, GLFramebufferAttachment attachment, int texture, int level, int baseViewIndex, int numViews) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glNamedFramebufferTextureMultiviewOVR(\{framebuffer}, \{attachment}, \{texture}, \{level}, \{baseViewIndex}, \{numViews})");
		try {
		    functions.glNamedFramebufferTextureMultiviewOVR().invokeExact(framebuffer, attachment, texture, level, baseViewIndex, numViews);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were attachment=\{attachment}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glNamedProgramLocalParameter4dEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glNamedProgramLocalParameter4dEXT.xhtml">registry.khronos.org</a>
	 */
	public void glNamedProgramLocalParameter4dEXT(int program, GLProgramTarget target, int index, double x, double y, double z, double w) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glNamedProgramLocalParameter4dEXT(\{program}, \{target}, \{index}, \{x}, \{y}, \{z}, \{w})");
		try {
		    functions.glNamedProgramLocalParameter4dEXT().invokeExact(program, target, index, x, y, z, w);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glNamedProgramLocalParameter4dvEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glNamedProgramLocalParameter4dvEXT.xhtml">registry.khronos.org</a>
	 */
	public void glNamedProgramLocalParameter4dvEXT(int program, GLProgramTarget target, int index, ConstPointer<Double> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glNamedProgramLocalParameter4dvEXT(\{program}, \{target}, \{index}, \{params})");
		try {
		    functions.glNamedProgramLocalParameter4dvEXT().invokeExact(program, target, index, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glNamedProgramLocalParameter4fEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glNamedProgramLocalParameter4fEXT.xhtml">registry.khronos.org</a>
	 */
	public void glNamedProgramLocalParameter4fEXT(int program, GLProgramTarget target, int index, float x, float y, float z, float w) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glNamedProgramLocalParameter4fEXT(\{program}, \{target}, \{index}, \{x}, \{y}, \{z}, \{w})");
		try {
		    functions.glNamedProgramLocalParameter4fEXT().invokeExact(program, target, index, x, y, z, w);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glNamedProgramLocalParameter4fvEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glNamedProgramLocalParameter4fvEXT.xhtml">registry.khronos.org</a>
	 */
	public void glNamedProgramLocalParameter4fvEXT(int program, GLProgramTarget target, int index, ConstPointer<Float> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glNamedProgramLocalParameter4fvEXT(\{program}, \{target}, \{index}, \{params})");
		try {
		    functions.glNamedProgramLocalParameter4fvEXT().invokeExact(program, target, index, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glNamedProgramLocalParameterI4iEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glNamedProgramLocalParameterI4iEXT.xhtml">registry.khronos.org</a>
	 */
	public void glNamedProgramLocalParameterI4iEXT(int program, GLProgramTarget target, int index, int x, int y, int z, int w) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glNamedProgramLocalParameterI4iEXT(\{program}, \{target}, \{index}, \{x}, \{y}, \{z}, \{w})");
		try {
		    functions.glNamedProgramLocalParameterI4iEXT().invokeExact(program, target, index, x, y, z, w);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glNamedProgramLocalParameterI4ivEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glNamedProgramLocalParameterI4ivEXT.xhtml">registry.khronos.org</a>
	 */
	public void glNamedProgramLocalParameterI4ivEXT(int program, GLProgramTarget target, int index, ConstPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glNamedProgramLocalParameterI4ivEXT(\{program}, \{target}, \{index}, \{params})");
		try {
		    functions.glNamedProgramLocalParameterI4ivEXT().invokeExact(program, target, index, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glNamedProgramLocalParameterI4uiEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glNamedProgramLocalParameterI4uiEXT.xhtml">registry.khronos.org</a>
	 */
	public void glNamedProgramLocalParameterI4uiEXT(int program, GLProgramTarget target, int index, int x, int y, int z, int w) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glNamedProgramLocalParameterI4uiEXT(\{program}, \{target}, \{index}, \{x}, \{y}, \{z}, \{w})");
		try {
		    functions.glNamedProgramLocalParameterI4uiEXT().invokeExact(program, target, index, x, y, z, w);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glNamedProgramLocalParameterI4uivEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glNamedProgramLocalParameterI4uivEXT.xhtml">registry.khronos.org</a>
	 */
	public void glNamedProgramLocalParameterI4uivEXT(int program, GLProgramTarget target, int index, ConstPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glNamedProgramLocalParameterI4uivEXT(\{program}, \{target}, \{index}, \{params})");
		try {
		    functions.glNamedProgramLocalParameterI4uivEXT().invokeExact(program, target, index, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glNamedProgramLocalParameters4fvEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glNamedProgramLocalParameters4fvEXT.xhtml">registry.khronos.org</a>
	 */
	public void glNamedProgramLocalParameters4fvEXT(int program, GLProgramTarget target, int index, int count, ConstPointer<Float> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glNamedProgramLocalParameters4fvEXT(\{program}, \{target}, \{index}, \{count}, \{params})");
		try {
		    functions.glNamedProgramLocalParameters4fvEXT().invokeExact(program, target, index, count, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glNamedProgramLocalParametersI4ivEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glNamedProgramLocalParametersI4ivEXT.xhtml">registry.khronos.org</a>
	 */
	public void glNamedProgramLocalParametersI4ivEXT(int program, GLProgramTarget target, int index, int count, ConstPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glNamedProgramLocalParametersI4ivEXT(\{program}, \{target}, \{index}, \{count}, \{params})");
		try {
		    functions.glNamedProgramLocalParametersI4ivEXT().invokeExact(program, target, index, count, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glNamedProgramLocalParametersI4uivEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glNamedProgramLocalParametersI4uivEXT.xhtml">registry.khronos.org</a>
	 */
	public void glNamedProgramLocalParametersI4uivEXT(int program, GLProgramTarget target, int index, int count, ConstPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glNamedProgramLocalParametersI4uivEXT(\{program}, \{target}, \{index}, \{count}, \{params})");
		try {
		    functions.glNamedProgramLocalParametersI4uivEXT().invokeExact(program, target, index, count, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glNamedProgramStringEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glNamedProgramStringEXT.xhtml">registry.khronos.org</a>
	 */
	public void glNamedProgramStringEXT(int program, GLProgramTarget target, GLProgramFormat format, int len, ConstPointer<Void> string) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glNamedProgramStringEXT(\{program}, \{target}, \{format}, \{len}, \{string})");
		try {
		    functions.glNamedProgramStringEXT().invokeExact(program, target, format, len, string);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, format=\{format}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glNamedRenderbufferStorage">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glNamedRenderbufferStorage.xhtml">registry.khronos.org</a>
	 */
	public void glNamedRenderbufferStorage(int renderbuffer, GLInternalFormat internalformat, int width, int height) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glNamedRenderbufferStorage(\{renderbuffer}, \{internalformat}, \{width}, \{height})");
		try {
		    functions.glNamedRenderbufferStorage().invokeExact(renderbuffer, internalformat, width, height);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were internalformat=\{internalformat}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glNamedRenderbufferStorageEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glNamedRenderbufferStorageEXT.xhtml">registry.khronos.org</a>
	 */
	public void glNamedRenderbufferStorageEXT(int renderbuffer, GLInternalFormat internalformat, int width, int height) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glNamedRenderbufferStorageEXT(\{renderbuffer}, \{internalformat}, \{width}, \{height})");
		try {
		    functions.glNamedRenderbufferStorageEXT().invokeExact(renderbuffer, internalformat, width, height);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were internalformat=\{internalformat}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glNamedRenderbufferStorageMultisample">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glNamedRenderbufferStorageMultisample.xhtml">registry.khronos.org</a>
	 */
	public void glNamedRenderbufferStorageMultisample(int renderbuffer, int samples, GLInternalFormat internalformat, int width, int height) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glNamedRenderbufferStorageMultisample(\{renderbuffer}, \{samples}, \{internalformat}, \{width}, \{height})");
		try {
		    functions.glNamedRenderbufferStorageMultisample().invokeExact(renderbuffer, samples, internalformat, width, height);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were internalformat=\{internalformat}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glNamedRenderbufferStorageMultisampleAdvancedAMD">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glNamedRenderbufferStorageMultisampleAdvancedAMD.xhtml">registry.khronos.org</a>
	 */
	public void glNamedRenderbufferStorageMultisampleAdvancedAMD(int renderbuffer, int samples, int storageSamples, GLInternalFormat internalformat, int width, int height) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glNamedRenderbufferStorageMultisampleAdvancedAMD(\{renderbuffer}, \{samples}, \{storageSamples}, \{internalformat}, \{width}, \{height})");
		try {
		    functions.glNamedRenderbufferStorageMultisampleAdvancedAMD().invokeExact(renderbuffer, samples, storageSamples, internalformat, width, height);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were internalformat=\{internalformat}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glNamedRenderbufferStorageMultisampleCoverageEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glNamedRenderbufferStorageMultisampleCoverageEXT.xhtml">registry.khronos.org</a>
	 */
	public void glNamedRenderbufferStorageMultisampleCoverageEXT(int renderbuffer, int coverageSamples, int colorSamples, GLInternalFormat internalformat, int width, int height) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glNamedRenderbufferStorageMultisampleCoverageEXT(\{renderbuffer}, \{coverageSamples}, \{colorSamples}, \{internalformat}, \{width}, \{height})");
		try {
		    functions.glNamedRenderbufferStorageMultisampleCoverageEXT().invokeExact(renderbuffer, coverageSamples, colorSamples, internalformat, width, height);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were internalformat=\{internalformat}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glNamedRenderbufferStorageMultisampleEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glNamedRenderbufferStorageMultisampleEXT.xhtml">registry.khronos.org</a>
	 */
	public void glNamedRenderbufferStorageMultisampleEXT(int renderbuffer, int samples, GLInternalFormat internalformat, int width, int height) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glNamedRenderbufferStorageMultisampleEXT(\{renderbuffer}, \{samples}, \{internalformat}, \{width}, \{height})");
		try {
		    functions.glNamedRenderbufferStorageMultisampleEXT().invokeExact(renderbuffer, samples, internalformat, width, height);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were internalformat=\{internalformat}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glNamedStringARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glNamedStringARB.xhtml">registry.khronos.org</a>
	 */
	public void glNamedStringARB(GLEnum type, int namelen, ConstPointer<Byte> name, int stringlen, ConstPointer<Byte> string) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glNamedStringARB(\{type}, \{namelen}, \{name}, \{stringlen}, \{string})");
		try {
		    functions.glNamedStringARB().invokeExact(type, namelen, name, stringlen, string);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glNewList">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glNewList.xhtml">registry.khronos.org</a>
	 */
	public void glNewList(int list, GLListMode mode) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glNewList(\{list}, \{mode})");
		try {
		    functions.glNewList().invokeExact(list, mode);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glNewObjectBufferATI">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glNewObjectBufferATI.xhtml">registry.khronos.org</a>
	 */
	public int glNewObjectBufferATI(int size, ConstPointer<Void> pointer, GLArrayObjectUsageATI usage) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glNewObjectBufferATI(\{size}, \{pointer}, \{usage})");
		try {
		    var r = (int)functions.glNewObjectBufferATI().invokeExact(size, pointer, usage);
			checkError();
			return r;
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were usage=\{usage}.");
			}
			throw e;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glNormal3b">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glNormal3b.xhtml">registry.khronos.org</a>
	 */
	public void glNormal3b(byte nx, byte ny, byte nz) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glNormal3b(\{nx}, \{ny}, \{nz})");
		try {
		    functions.glNormal3b().invokeExact(nx, ny, nz);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glNormal3bv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glNormal3bv.xhtml">registry.khronos.org</a>
	 */
	public void glNormal3bv(ConstPointer<Byte> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glNormal3bv(\{v})");
		try {
		    functions.glNormal3bv().invokeExact(v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glNormal3d">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glNormal3d.xhtml">registry.khronos.org</a>
	 */
	public void glNormal3d(double nx, double ny, double nz) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glNormal3d(\{nx}, \{ny}, \{nz})");
		try {
		    functions.glNormal3d().invokeExact(nx, ny, nz);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glNormal3dv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glNormal3dv.xhtml">registry.khronos.org</a>
	 */
	public void glNormal3dv(ConstPointer<Double> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glNormal3dv(\{v})");
		try {
		    functions.glNormal3dv().invokeExact(v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glNormal3f">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glNormal3f.xhtml">registry.khronos.org</a>
	 */
	public void glNormal3f(float nx, float ny, float nz) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glNormal3f(\{nx}, \{ny}, \{nz})");
		try {
		    functions.glNormal3f().invokeExact(nx, ny, nz);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glNormal3fVertex3fSUN">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glNormal3fVertex3fSUN.xhtml">registry.khronos.org</a>
	 */
	public void glNormal3fVertex3fSUN(float nx, float ny, float nz, float x, float y, float z) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glNormal3fVertex3fSUN(\{nx}, \{ny}, \{nz}, \{x}, \{y}, \{z})");
		try {
		    functions.glNormal3fVertex3fSUN().invokeExact(nx, ny, nz, x, y, z);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glNormal3fVertex3fvSUN">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glNormal3fVertex3fvSUN.xhtml">registry.khronos.org</a>
	 */
	public void glNormal3fVertex3fvSUN(ConstPointer<Float> n, ConstPointer<Float> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glNormal3fVertex3fvSUN(\{n}, \{v})");
		try {
		    functions.glNormal3fVertex3fvSUN().invokeExact(n, v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glNormal3fv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glNormal3fv.xhtml">registry.khronos.org</a>
	 */
	public void glNormal3fv(ConstPointer<Float> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glNormal3fv(\{v})");
		try {
		    functions.glNormal3fv().invokeExact(v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glNormal3hNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glNormal3hNV.xhtml">registry.khronos.org</a>
	 */
	public void glNormal3hNV(char nx, char ny, char nz) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glNormal3hNV(\{nx}, \{ny}, \{nz})");
		try {
		    functions.glNormal3hNV().invokeExact(nx, ny, nz);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glNormal3hvNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glNormal3hvNV.xhtml">registry.khronos.org</a>
	 */
	public void glNormal3hvNV(ConstPointer<Character> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glNormal3hvNV(\{v})");
		try {
		    functions.glNormal3hvNV().invokeExact(v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glNormal3i">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glNormal3i.xhtml">registry.khronos.org</a>
	 */
	public void glNormal3i(int nx, int ny, int nz) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glNormal3i(\{nx}, \{ny}, \{nz})");
		try {
		    functions.glNormal3i().invokeExact(nx, ny, nz);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glNormal3iv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glNormal3iv.xhtml">registry.khronos.org</a>
	 */
	public void glNormal3iv(ConstPointer<Integer> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glNormal3iv(\{v})");
		try {
		    functions.glNormal3iv().invokeExact(v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glNormal3s">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glNormal3s.xhtml">registry.khronos.org</a>
	 */
	public void glNormal3s(short nx, short ny, short nz) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glNormal3s(\{nx}, \{ny}, \{nz})");
		try {
		    functions.glNormal3s().invokeExact(nx, ny, nz);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glNormal3sv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glNormal3sv.xhtml">registry.khronos.org</a>
	 */
	public void glNormal3sv(ConstPointer<Short> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glNormal3sv(\{v})");
		try {
		    functions.glNormal3sv().invokeExact(v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glNormal3x">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glNormal3x.xhtml">registry.khronos.org</a>
	 */
	public void glNormal3x(int nx, int ny, int nz) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glNormal3x(\{nx}, \{ny}, \{nz})");
		try {
		    functions.glNormal3x().invokeExact(nx, ny, nz);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glNormal3xOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glNormal3xOES.xhtml">registry.khronos.org</a>
	 */
	public void glNormal3xOES(int nx, int ny, int nz) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glNormal3xOES(\{nx}, \{ny}, \{nz})");
		try {
		    functions.glNormal3xOES().invokeExact(nx, ny, nz);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glNormal3xvOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glNormal3xvOES.xhtml">registry.khronos.org</a>
	 */
	public void glNormal3xvOES(ConstPointer<Integer> coords) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glNormal3xvOES(\{coords})");
		try {
		    functions.glNormal3xvOES().invokeExact(coords);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glNormalFormatNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glNormalFormatNV.xhtml">registry.khronos.org</a>
	 */
	public void glNormalFormatNV(GLEnum type, int stride) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glNormalFormatNV(\{type}, \{stride})");
		try {
		    functions.glNormalFormatNV().invokeExact(type, stride);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glNormalP3ui">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glNormalP3ui.xhtml">registry.khronos.org</a>
	 */
	public void glNormalP3ui(GLNormalPointerType type, int coords) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glNormalP3ui(\{type}, \{coords})");
		try {
		    functions.glNormalP3ui().invokeExact(type, coords);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glNormalP3uiv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glNormalP3uiv.xhtml">registry.khronos.org</a>
	 */
	public void glNormalP3uiv(GLNormalPointerType type, ConstPointer<Integer> coords) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glNormalP3uiv(\{type}, \{coords})");
		try {
		    functions.glNormalP3uiv().invokeExact(type, coords);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glNormalPointer">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glNormalPointer.xhtml">registry.khronos.org</a>
	 */
	public void glNormalPointer(GLNormalPointerType type, int stride, ConstPointer<Void> pointer) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glNormalPointer(\{type}, \{stride}, \{pointer})");
		try {
		    functions.glNormalPointer().invokeExact(type, stride, pointer);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glNormalPointerEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glNormalPointerEXT.xhtml">registry.khronos.org</a>
	 */
	public void glNormalPointerEXT(GLNormalPointerType type, int stride, int count, ConstPointer<Void> pointer) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glNormalPointerEXT(\{type}, \{stride}, \{count}, \{pointer})");
		try {
		    functions.glNormalPointerEXT().invokeExact(type, stride, count, pointer);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glNormalPointerListIBM">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glNormalPointerListIBM.xhtml">registry.khronos.org</a>
	 */
	public void glNormalPointerListIBM(GLNormalPointerType type, int stride, MutPointer<ConstPointer<Void>> pointer, int ptrstride) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glNormalPointerListIBM(\{type}, \{stride}, \{pointer}, \{ptrstride})");
		try {
		    functions.glNormalPointerListIBM().invokeExact(type, stride, pointer, ptrstride);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glNormalPointervINTEL">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glNormalPointervINTEL.xhtml">registry.khronos.org</a>
	 */
	public void glNormalPointervINTEL(GLNormalPointerType type, MutPointer<ConstPointer<Void>> pointer) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glNormalPointervINTEL(\{type}, \{pointer})");
		try {
		    functions.glNormalPointervINTEL().invokeExact(type, pointer);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glNormalStream3bATI">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glNormalStream3bATI.xhtml">registry.khronos.org</a>
	 */
	public void glNormalStream3bATI(GLVertexStreamATI stream, byte nx, byte ny, byte nz) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glNormalStream3bATI(\{stream}, \{nx}, \{ny}, \{nz})");
		try {
		    functions.glNormalStream3bATI().invokeExact(stream, nx, ny, nz);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were stream=\{stream}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glNormalStream3bvATI">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glNormalStream3bvATI.xhtml">registry.khronos.org</a>
	 */
	public void glNormalStream3bvATI(GLVertexStreamATI stream, ConstPointer<Byte> coords) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glNormalStream3bvATI(\{stream}, \{coords})");
		try {
		    functions.glNormalStream3bvATI().invokeExact(stream, coords);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were stream=\{stream}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glNormalStream3dATI">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glNormalStream3dATI.xhtml">registry.khronos.org</a>
	 */
	public void glNormalStream3dATI(GLVertexStreamATI stream, double nx, double ny, double nz) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glNormalStream3dATI(\{stream}, \{nx}, \{ny}, \{nz})");
		try {
		    functions.glNormalStream3dATI().invokeExact(stream, nx, ny, nz);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were stream=\{stream}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glNormalStream3dvATI">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glNormalStream3dvATI.xhtml">registry.khronos.org</a>
	 */
	public void glNormalStream3dvATI(GLVertexStreamATI stream, ConstPointer<Double> coords) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glNormalStream3dvATI(\{stream}, \{coords})");
		try {
		    functions.glNormalStream3dvATI().invokeExact(stream, coords);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were stream=\{stream}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glNormalStream3fATI">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glNormalStream3fATI.xhtml">registry.khronos.org</a>
	 */
	public void glNormalStream3fATI(GLVertexStreamATI stream, float nx, float ny, float nz) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glNormalStream3fATI(\{stream}, \{nx}, \{ny}, \{nz})");
		try {
		    functions.glNormalStream3fATI().invokeExact(stream, nx, ny, nz);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were stream=\{stream}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glNormalStream3fvATI">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glNormalStream3fvATI.xhtml">registry.khronos.org</a>
	 */
	public void glNormalStream3fvATI(GLVertexStreamATI stream, ConstPointer<Float> coords) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glNormalStream3fvATI(\{stream}, \{coords})");
		try {
		    functions.glNormalStream3fvATI().invokeExact(stream, coords);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were stream=\{stream}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glNormalStream3iATI">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glNormalStream3iATI.xhtml">registry.khronos.org</a>
	 */
	public void glNormalStream3iATI(GLVertexStreamATI stream, int nx, int ny, int nz) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glNormalStream3iATI(\{stream}, \{nx}, \{ny}, \{nz})");
		try {
		    functions.glNormalStream3iATI().invokeExact(stream, nx, ny, nz);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were stream=\{stream}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glNormalStream3ivATI">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glNormalStream3ivATI.xhtml">registry.khronos.org</a>
	 */
	public void glNormalStream3ivATI(GLVertexStreamATI stream, ConstPointer<Integer> coords) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glNormalStream3ivATI(\{stream}, \{coords})");
		try {
		    functions.glNormalStream3ivATI().invokeExact(stream, coords);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were stream=\{stream}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glNormalStream3sATI">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glNormalStream3sATI.xhtml">registry.khronos.org</a>
	 */
	public void glNormalStream3sATI(GLVertexStreamATI stream, short nx, short ny, short nz) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glNormalStream3sATI(\{stream}, \{nx}, \{ny}, \{nz})");
		try {
		    functions.glNormalStream3sATI().invokeExact(stream, nx, ny, nz);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were stream=\{stream}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glNormalStream3svATI">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glNormalStream3svATI.xhtml">registry.khronos.org</a>
	 */
	public void glNormalStream3svATI(GLVertexStreamATI stream, ConstPointer<Short> coords) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glNormalStream3svATI(\{stream}, \{coords})");
		try {
		    functions.glNormalStream3svATI().invokeExact(stream, coords);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were stream=\{stream}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glObjectLabel">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glObjectLabel.xhtml">registry.khronos.org</a>
	 */
	public void glObjectLabel(GLObjectIdentifier identifier, int name, int length, ConstPointer<Byte> label) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glObjectLabel(\{identifier}, \{name}, \{length}, \{label})");
		try {
		    functions.glObjectLabel().invokeExact(identifier, name, length, label);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were identifier=\{identifier}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glObjectLabelKHR">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glObjectLabelKHR.xhtml">registry.khronos.org</a>
	 */
	public void glObjectLabelKHR(GLObjectIdentifier identifier, int name, int length, ConstPointer<Byte> label) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glObjectLabelKHR(\{identifier}, \{name}, \{length}, \{label})");
		try {
		    functions.glObjectLabelKHR().invokeExact(identifier, name, length, label);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were identifier=\{identifier}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glObjectPtrLabel">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glObjectPtrLabel.xhtml">registry.khronos.org</a>
	 */
	public void glObjectPtrLabel(ConstPointer<Void> ptr, int length, ConstPointer<Byte> label) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glObjectPtrLabel(\{ptr}, \{length}, \{label})");
		try {
		    functions.glObjectPtrLabel().invokeExact(ptr, length, label);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glObjectPtrLabelKHR">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glObjectPtrLabelKHR.xhtml">registry.khronos.org</a>
	 */
	public void glObjectPtrLabelKHR(ConstPointer<Void> ptr, int length, ConstPointer<Byte> label) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glObjectPtrLabelKHR(\{ptr}, \{length}, \{label})");
		try {
		    functions.glObjectPtrLabelKHR().invokeExact(ptr, length, label);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glObjectPurgeableAPPLE">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glObjectPurgeableAPPLE.xhtml">registry.khronos.org</a>
	 */
	public GLEnum glObjectPurgeableAPPLE(GLEnum objectType, int name, GLEnum option) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glObjectPurgeableAPPLE(\{objectType}, \{name}, \{option})");
		try {
		    var r = (GLEnum)functions.glObjectPurgeableAPPLE().invokeExact(objectType, name, option);
			checkError();
			return r;
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were objectType=\{objectType}, option=\{option}.");
			}
			throw e;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glObjectUnpurgeableAPPLE">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glObjectUnpurgeableAPPLE.xhtml">registry.khronos.org</a>
	 */
	public GLEnum glObjectUnpurgeableAPPLE(GLEnum objectType, int name, GLEnum option) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glObjectUnpurgeableAPPLE(\{objectType}, \{name}, \{option})");
		try {
		    var r = (GLEnum)functions.glObjectUnpurgeableAPPLE().invokeExact(objectType, name, option);
			checkError();
			return r;
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were objectType=\{objectType}, option=\{option}.");
			}
			throw e;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glOrtho">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glOrtho.xhtml">registry.khronos.org</a>
	 */
	public void glOrtho(double left, double right, double bottom, double top, double zNear, double zFar) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glOrtho(\{left}, \{right}, \{bottom}, \{top}, \{zNear}, \{zFar})");
		try {
		    functions.glOrtho().invokeExact(left, right, bottom, top, zNear, zFar);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glOrthof">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glOrthof.xhtml">registry.khronos.org</a>
	 */
	public void glOrthof(float l, float r, float b, float t, float n, float f) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glOrthof(\{l}, \{r}, \{b}, \{t}, \{n}, \{f})");
		try {
		    functions.glOrthof().invokeExact(l, r, b, t, n, f);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glOrthofOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glOrthofOES.xhtml">registry.khronos.org</a>
	 */
	public void glOrthofOES(float l, float r, float b, float t, float n, float f) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glOrthofOES(\{l}, \{r}, \{b}, \{t}, \{n}, \{f})");
		try {
		    functions.glOrthofOES().invokeExact(l, r, b, t, n, f);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glOrthox">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glOrthox.xhtml">registry.khronos.org</a>
	 */
	public void glOrthox(int l, int r, int b, int t, int n, int f) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glOrthox(\{l}, \{r}, \{b}, \{t}, \{n}, \{f})");
		try {
		    functions.glOrthox().invokeExact(l, r, b, t, n, f);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glOrthoxOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glOrthoxOES.xhtml">registry.khronos.org</a>
	 */
	public void glOrthoxOES(int l, int r, int b, int t, int n, int f) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glOrthoxOES(\{l}, \{r}, \{b}, \{t}, \{n}, \{f})");
		try {
		    functions.glOrthoxOES().invokeExact(l, r, b, t, n, f);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glPNTrianglesfATI">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glPNTrianglesfATI.xhtml">registry.khronos.org</a>
	 */
	public void glPNTrianglesfATI(GLPNTrianglesPNameATI pname, float param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glPNTrianglesfATI(\{pname}, \{param})");
		try {
		    functions.glPNTrianglesfATI().invokeExact(pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glPNTrianglesiATI">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glPNTrianglesiATI.xhtml">registry.khronos.org</a>
	 */
	public void glPNTrianglesiATI(GLPNTrianglesPNameATI pname, int param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glPNTrianglesiATI(\{pname}, \{param})");
		try {
		    functions.glPNTrianglesiATI().invokeExact(pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glPassTexCoordATI">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glPassTexCoordATI.xhtml">registry.khronos.org</a>
	 */
	public void glPassTexCoordATI(int dst, int coord, GLSwizzleOpATI swizzle) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glPassTexCoordATI(\{dst}, \{coord}, \{swizzle})");
		try {
		    functions.glPassTexCoordATI().invokeExact(dst, coord, swizzle);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were swizzle=\{swizzle}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glPassThrough">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glPassThrough.xhtml">registry.khronos.org</a>
	 */
	public void glPassThrough(float token) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glPassThrough(\{token})");
		try {
		    functions.glPassThrough().invokeExact(token);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glPassThroughxOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glPassThroughxOES.xhtml">registry.khronos.org</a>
	 */
	public void glPassThroughxOES(int token) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glPassThroughxOES(\{token})");
		try {
		    functions.glPassThroughxOES().invokeExact(token);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glPatchParameterfv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glPatchParameterfv.xhtml">registry.khronos.org</a>
	 */
	public void glPatchParameterfv(GLPatchParameterName pname, ConstPointer<Float> values) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glPatchParameterfv(\{pname}, \{values})");
		try {
		    functions.glPatchParameterfv().invokeExact(pname, values);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glPatchParameteri">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glPatchParameteri.xhtml">registry.khronos.org</a>
	 */
	public void glPatchParameteri(GLPatchParameterName pname, int value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glPatchParameteri(\{pname}, \{value})");
		try {
		    functions.glPatchParameteri().invokeExact(pname, value);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glPatchParameteriEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glPatchParameteriEXT.xhtml">registry.khronos.org</a>
	 */
	public void glPatchParameteriEXT(GLPatchParameterName pname, int value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glPatchParameteriEXT(\{pname}, \{value})");
		try {
		    functions.glPatchParameteriEXT().invokeExact(pname, value);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glPatchParameteriOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glPatchParameteriOES.xhtml">registry.khronos.org</a>
	 */
	public void glPatchParameteriOES(GLPatchParameterName pname, int value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glPatchParameteriOES(\{pname}, \{value})");
		try {
		    functions.glPatchParameteriOES().invokeExact(pname, value);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glPathColorGenNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glPathColorGenNV.xhtml">registry.khronos.org</a>
	 */
	public void glPathColorGenNV(GLPathColor color, GLPathGenMode genMode, GLPathColorFormat colorFormat, ConstPointer<Float> coeffs) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glPathColorGenNV(\{color}, \{genMode}, \{colorFormat}, \{coeffs})");
		try {
		    functions.glPathColorGenNV().invokeExact(color, genMode, colorFormat, coeffs);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were color=\{color}, genMode=\{genMode}, colorFormat=\{colorFormat}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glPathCommandsNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glPathCommandsNV.xhtml">registry.khronos.org</a>
	 */
	public void glPathCommandsNV(int path, int numCommands, ConstPointer<Byte> commands, int numCoords, GLPathCoordType coordType, ConstPointer<Void> coords) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glPathCommandsNV(\{path}, \{numCommands}, \{commands}, \{numCoords}, \{coordType}, \{coords})");
		try {
		    functions.glPathCommandsNV().invokeExact(path, numCommands, commands, numCoords, coordType, coords);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were coordType=\{coordType}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glPathCoordsNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glPathCoordsNV.xhtml">registry.khronos.org</a>
	 */
	public void glPathCoordsNV(int path, int numCoords, GLPathCoordType coordType, ConstPointer<Void> coords) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glPathCoordsNV(\{path}, \{numCoords}, \{coordType}, \{coords})");
		try {
		    functions.glPathCoordsNV().invokeExact(path, numCoords, coordType, coords);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were coordType=\{coordType}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glPathCoverDepthFuncNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glPathCoverDepthFuncNV.xhtml">registry.khronos.org</a>
	 */
	public void glPathCoverDepthFuncNV(GLDepthFunction func) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glPathCoverDepthFuncNV(\{func})");
		try {
		    functions.glPathCoverDepthFuncNV().invokeExact(func);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were func=\{func}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glPathDashArrayNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glPathDashArrayNV.xhtml">registry.khronos.org</a>
	 */
	public void glPathDashArrayNV(int path, int dashCount, ConstPointer<Float> dashArray) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glPathDashArrayNV(\{path}, \{dashCount}, \{dashArray})");
		try {
		    functions.glPathDashArrayNV().invokeExact(path, dashCount, dashArray);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glPathFogGenNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glPathFogGenNV.xhtml">registry.khronos.org</a>
	 */
	public void glPathFogGenNV(GLPathGenMode genMode) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glPathFogGenNV(\{genMode})");
		try {
		    functions.glPathFogGenNV().invokeExact(genMode);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were genMode=\{genMode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glPathGlyphIndexArrayNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glPathGlyphIndexArrayNV.xhtml">registry.khronos.org</a>
	 */
	public GLEnum glPathGlyphIndexArrayNV(int firstPathName, GLEnum fontTarget, ConstPointer<Void> fontName, int fontStyle, int firstGlyphIndex, int numGlyphs, int pathParameterTemplate, float emScale) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glPathGlyphIndexArrayNV(\{firstPathName}, \{fontTarget}, \{fontName}, \{fontStyle}, \{firstGlyphIndex}, \{numGlyphs}, \{pathParameterTemplate}, \{emScale})");
		try {
		    var r = (GLEnum)functions.glPathGlyphIndexArrayNV().invokeExact(firstPathName, fontTarget, fontName, fontStyle, firstGlyphIndex, numGlyphs, pathParameterTemplate, emScale);
			checkError();
			return r;
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were fontTarget=\{fontTarget}.");
			}
			throw e;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glPathGlyphIndexRangeNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glPathGlyphIndexRangeNV.xhtml">registry.khronos.org</a>
	 */
	public GLEnum glPathGlyphIndexRangeNV(GLEnum fontTarget, ConstPointer<Void> fontName, int fontStyle, int pathParameterTemplate, float emScale, MutPointer<Integer> baseAndCount) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glPathGlyphIndexRangeNV(\{fontTarget}, \{fontName}, \{fontStyle}, \{pathParameterTemplate}, \{emScale}, \{baseAndCount})");
		try {
		    var r = (GLEnum)functions.glPathGlyphIndexRangeNV().invokeExact(fontTarget, fontName, fontStyle, pathParameterTemplate, emScale, baseAndCount);
			checkError();
			return r;
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were fontTarget=\{fontTarget}.");
			}
			throw e;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glPathGlyphRangeNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glPathGlyphRangeNV.xhtml">registry.khronos.org</a>
	 */
	public void glPathGlyphRangeNV(int firstPathName, GLPathFontTarget fontTarget, ConstPointer<Void> fontName, int fontStyle, int firstGlyph, int numGlyphs, GLPathHandleMissingGlyphs handleMissingGlyphs, int pathParameterTemplate, float emScale) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glPathGlyphRangeNV(\{firstPathName}, \{fontTarget}, \{fontName}, \{fontStyle}, \{firstGlyph}, \{numGlyphs}, \{handleMissingGlyphs}, \{pathParameterTemplate}, \{emScale})");
		try {
		    functions.glPathGlyphRangeNV().invokeExact(firstPathName, fontTarget, fontName, fontStyle, firstGlyph, numGlyphs, handleMissingGlyphs, pathParameterTemplate, emScale);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were fontTarget=\{fontTarget}, handleMissingGlyphs=\{handleMissingGlyphs}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glPathGlyphsNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glPathGlyphsNV.xhtml">registry.khronos.org</a>
	 */
	public void glPathGlyphsNV(int firstPathName, GLPathFontTarget fontTarget, ConstPointer<Void> fontName, int fontStyle, int numGlyphs, GLPathElementType type, ConstPointer<Void> charcodes, GLPathHandleMissingGlyphs handleMissingGlyphs, int pathParameterTemplate, float emScale) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glPathGlyphsNV(\{firstPathName}, \{fontTarget}, \{fontName}, \{fontStyle}, \{numGlyphs}, \{type}, \{charcodes}, \{handleMissingGlyphs}, \{pathParameterTemplate}, \{emScale})");
		try {
		    functions.glPathGlyphsNV().invokeExact(firstPathName, fontTarget, fontName, fontStyle, numGlyphs, type, charcodes, handleMissingGlyphs, pathParameterTemplate, emScale);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were fontTarget=\{fontTarget}, type=\{type}, handleMissingGlyphs=\{handleMissingGlyphs}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glPathMemoryGlyphIndexArrayNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glPathMemoryGlyphIndexArrayNV.xhtml">registry.khronos.org</a>
	 */
	public GLEnum glPathMemoryGlyphIndexArrayNV(int firstPathName, GLEnum fontTarget, long fontSize, ConstPointer<Void> fontData, int faceIndex, int firstGlyphIndex, int numGlyphs, int pathParameterTemplate, float emScale) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glPathMemoryGlyphIndexArrayNV(\{firstPathName}, \{fontTarget}, \{fontSize}, \{fontData}, \{faceIndex}, \{firstGlyphIndex}, \{numGlyphs}, \{pathParameterTemplate}, \{emScale})");
		try {
		    var r = (GLEnum)functions.glPathMemoryGlyphIndexArrayNV().invokeExact(firstPathName, fontTarget, fontSize, fontData, faceIndex, firstGlyphIndex, numGlyphs, pathParameterTemplate, emScale);
			checkError();
			return r;
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were fontTarget=\{fontTarget}.");
			}
			throw e;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glPathParameterfNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glPathParameterfNV.xhtml">registry.khronos.org</a>
	 */
	public void glPathParameterfNV(int path, GLPathParameter pname, float value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glPathParameterfNV(\{path}, \{pname}, \{value})");
		try {
		    functions.glPathParameterfNV().invokeExact(path, pname, value);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glPathParameterfvNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glPathParameterfvNV.xhtml">registry.khronos.org</a>
	 */
	public void glPathParameterfvNV(int path, GLPathParameter pname, ConstPointer<Float> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glPathParameterfvNV(\{path}, \{pname}, \{value})");
		try {
		    functions.glPathParameterfvNV().invokeExact(path, pname, value);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glPathParameteriNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glPathParameteriNV.xhtml">registry.khronos.org</a>
	 */
	public void glPathParameteriNV(int path, GLPathParameter pname, int value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glPathParameteriNV(\{path}, \{pname}, \{value})");
		try {
		    functions.glPathParameteriNV().invokeExact(path, pname, value);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glPathParameterivNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glPathParameterivNV.xhtml">registry.khronos.org</a>
	 */
	public void glPathParameterivNV(int path, GLPathParameter pname, ConstPointer<Integer> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glPathParameterivNV(\{path}, \{pname}, \{value})");
		try {
		    functions.glPathParameterivNV().invokeExact(path, pname, value);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glPathStencilDepthOffsetNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glPathStencilDepthOffsetNV.xhtml">registry.khronos.org</a>
	 */
	public void glPathStencilDepthOffsetNV(float factor, float units) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glPathStencilDepthOffsetNV(\{factor}, \{units})");
		try {
		    functions.glPathStencilDepthOffsetNV().invokeExact(factor, units);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glPathStencilFuncNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glPathStencilFuncNV.xhtml">registry.khronos.org</a>
	 */
	public void glPathStencilFuncNV(GLStencilFunction func, int ref, int mask) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glPathStencilFuncNV(\{func}, \{ref}, \{mask})");
		try {
		    functions.glPathStencilFuncNV().invokeExact(func, ref, mask);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were func=\{func}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glPathStringNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glPathStringNV.xhtml">registry.khronos.org</a>
	 */
	public void glPathStringNV(int path, GLPathStringFormat format, int length, ConstPointer<Void> pathString) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glPathStringNV(\{path}, \{format}, \{length}, \{pathString})");
		try {
		    functions.glPathStringNV().invokeExact(path, format, length, pathString);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were format=\{format}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glPathSubCommandsNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glPathSubCommandsNV.xhtml">registry.khronos.org</a>
	 */
	public void glPathSubCommandsNV(int path, int commandStart, int commandsToDelete, int numCommands, ConstPointer<Byte> commands, int numCoords, GLPathCoordType coordType, ConstPointer<Void> coords) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glPathSubCommandsNV(\{path}, \{commandStart}, \{commandsToDelete}, \{numCommands}, \{commands}, \{numCoords}, \{coordType}, \{coords})");
		try {
		    functions.glPathSubCommandsNV().invokeExact(path, commandStart, commandsToDelete, numCommands, commands, numCoords, coordType, coords);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were coordType=\{coordType}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glPathSubCoordsNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glPathSubCoordsNV.xhtml">registry.khronos.org</a>
	 */
	public void glPathSubCoordsNV(int path, int coordStart, int numCoords, GLPathCoordType coordType, ConstPointer<Void> coords) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glPathSubCoordsNV(\{path}, \{coordStart}, \{numCoords}, \{coordType}, \{coords})");
		try {
		    functions.glPathSubCoordsNV().invokeExact(path, coordStart, numCoords, coordType, coords);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were coordType=\{coordType}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glPathTexGenNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glPathTexGenNV.xhtml">registry.khronos.org</a>
	 */
	public void glPathTexGenNV(GLPathColor texCoordSet, GLPathGenMode genMode, int components, ConstPointer<Float> coeffs) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glPathTexGenNV(\{texCoordSet}, \{genMode}, \{components}, \{coeffs})");
		try {
		    functions.glPathTexGenNV().invokeExact(texCoordSet, genMode, components, coeffs);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were texCoordSet=\{texCoordSet}, genMode=\{genMode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glPauseTransformFeedback">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glPauseTransformFeedback.xhtml">registry.khronos.org</a>
	 */
	public void glPauseTransformFeedback() {
	    if (LOG_COMMANDS) logCmd(() -> STR."glPauseTransformFeedback()");
		try {
		    functions.glPauseTransformFeedback().invokeExact();
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glPauseTransformFeedbackNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glPauseTransformFeedbackNV.xhtml">registry.khronos.org</a>
	 */
	public void glPauseTransformFeedbackNV() {
	    if (LOG_COMMANDS) logCmd(() -> STR."glPauseTransformFeedbackNV()");
		try {
		    functions.glPauseTransformFeedbackNV().invokeExact();
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glPixelDataRangeNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glPixelDataRangeNV.xhtml">registry.khronos.org</a>
	 */
	public void glPixelDataRangeNV(GLPixelDataRangeTargetNV target, int length, ConstPointer<Void> pointer) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glPixelDataRangeNV(\{target}, \{length}, \{pointer})");
		try {
		    functions.glPixelDataRangeNV().invokeExact(target, length, pointer);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glPixelMapfv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glPixelMapfv.xhtml">registry.khronos.org</a>
	 */
	public void glPixelMapfv(GLPixelMap map, int mapsize, ConstPointer<Float> values) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glPixelMapfv(\{map}, \{mapsize}, \{values})");
		try {
		    functions.glPixelMapfv().invokeExact(map, mapsize, values);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were map=\{map}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glPixelMapuiv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glPixelMapuiv.xhtml">registry.khronos.org</a>
	 */
	public void glPixelMapuiv(GLPixelMap map, int mapsize, ConstPointer<Integer> values) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glPixelMapuiv(\{map}, \{mapsize}, \{values})");
		try {
		    functions.glPixelMapuiv().invokeExact(map, mapsize, values);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were map=\{map}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glPixelMapusv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glPixelMapusv.xhtml">registry.khronos.org</a>
	 */
	public void glPixelMapusv(GLPixelMap map, int mapsize, ConstPointer<Character> values) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glPixelMapusv(\{map}, \{mapsize}, \{values})");
		try {
		    functions.glPixelMapusv().invokeExact(map, mapsize, values);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were map=\{map}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glPixelMapx">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glPixelMapx.xhtml">registry.khronos.org</a>
	 */
	public void glPixelMapx(GLPixelMap map, int size, ConstPointer<Integer> values) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glPixelMapx(\{map}, \{size}, \{values})");
		try {
		    functions.glPixelMapx().invokeExact(map, size, values);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were map=\{map}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glPixelStoref">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glPixelStoref.xhtml">registry.khronos.org</a>
	 */
	public void glPixelStoref(GLPixelStoreParameter pname, float param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glPixelStoref(\{pname}, \{param})");
		try {
		    functions.glPixelStoref().invokeExact(pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glPixelStorei">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glPixelStorei.xhtml">registry.khronos.org</a>
	 */
	public void glPixelStorei(GLPixelStoreParameter pname, int param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glPixelStorei(\{pname}, \{param})");
		try {
		    functions.glPixelStorei().invokeExact(pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glPixelStorex">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glPixelStorex.xhtml">registry.khronos.org</a>
	 */
	public void glPixelStorex(GLPixelStoreParameter pname, int param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glPixelStorex(\{pname}, \{param})");
		try {
		    functions.glPixelStorex().invokeExact(pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glPixelTexGenParameterfSGIS">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glPixelTexGenParameterfSGIS.xhtml">registry.khronos.org</a>
	 */
	public void glPixelTexGenParameterfSGIS(GLPixelTexGenParameterNameSGIS pname, float param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glPixelTexGenParameterfSGIS(\{pname}, \{param})");
		try {
		    functions.glPixelTexGenParameterfSGIS().invokeExact(pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glPixelTexGenParameterfvSGIS">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glPixelTexGenParameterfvSGIS.xhtml">registry.khronos.org</a>
	 */
	public void glPixelTexGenParameterfvSGIS(GLPixelTexGenParameterNameSGIS pname, ConstPointer<Float> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glPixelTexGenParameterfvSGIS(\{pname}, \{params})");
		try {
		    functions.glPixelTexGenParameterfvSGIS().invokeExact(pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glPixelTexGenParameteriSGIS">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glPixelTexGenParameteriSGIS.xhtml">registry.khronos.org</a>
	 */
	public void glPixelTexGenParameteriSGIS(GLPixelTexGenParameterNameSGIS pname, int param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glPixelTexGenParameteriSGIS(\{pname}, \{param})");
		try {
		    functions.glPixelTexGenParameteriSGIS().invokeExact(pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glPixelTexGenParameterivSGIS">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glPixelTexGenParameterivSGIS.xhtml">registry.khronos.org</a>
	 */
	public void glPixelTexGenParameterivSGIS(GLPixelTexGenParameterNameSGIS pname, ConstPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glPixelTexGenParameterivSGIS(\{pname}, \{params})");
		try {
		    functions.glPixelTexGenParameterivSGIS().invokeExact(pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glPixelTexGenSGIX">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glPixelTexGenSGIX.xhtml">registry.khronos.org</a>
	 */
	public void glPixelTexGenSGIX(GLPixelTexGenModeSGIX mode) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glPixelTexGenSGIX(\{mode})");
		try {
		    functions.glPixelTexGenSGIX().invokeExact(mode);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glPixelTransferf">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glPixelTransferf.xhtml">registry.khronos.org</a>
	 */
	public void glPixelTransferf(GLPixelTransferParameter pname, float param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glPixelTransferf(\{pname}, \{param})");
		try {
		    functions.glPixelTransferf().invokeExact(pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glPixelTransferi">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glPixelTransferi.xhtml">registry.khronos.org</a>
	 */
	public void glPixelTransferi(GLPixelTransferParameter pname, int param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glPixelTransferi(\{pname}, \{param})");
		try {
		    functions.glPixelTransferi().invokeExact(pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glPixelTransferxOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glPixelTransferxOES.xhtml">registry.khronos.org</a>
	 */
	public void glPixelTransferxOES(GLPixelTransferParameter pname, int param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glPixelTransferxOES(\{pname}, \{param})");
		try {
		    functions.glPixelTransferxOES().invokeExact(pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glPixelTransformParameterfEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glPixelTransformParameterfEXT.xhtml">registry.khronos.org</a>
	 */
	public void glPixelTransformParameterfEXT(GLPixelTransformTargetEXT target, GLPixelTransformPNameEXT pname, float param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glPixelTransformParameterfEXT(\{target}, \{pname}, \{param})");
		try {
		    functions.glPixelTransformParameterfEXT().invokeExact(target, pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glPixelTransformParameterfvEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glPixelTransformParameterfvEXT.xhtml">registry.khronos.org</a>
	 */
	public void glPixelTransformParameterfvEXT(GLPixelTransformTargetEXT target, GLPixelTransformPNameEXT pname, ConstPointer<Float> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glPixelTransformParameterfvEXT(\{target}, \{pname}, \{params})");
		try {
		    functions.glPixelTransformParameterfvEXT().invokeExact(target, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glPixelTransformParameteriEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glPixelTransformParameteriEXT.xhtml">registry.khronos.org</a>
	 */
	public void glPixelTransformParameteriEXT(GLPixelTransformTargetEXT target, GLPixelTransformPNameEXT pname, int param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glPixelTransformParameteriEXT(\{target}, \{pname}, \{param})");
		try {
		    functions.glPixelTransformParameteriEXT().invokeExact(target, pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glPixelTransformParameterivEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glPixelTransformParameterivEXT.xhtml">registry.khronos.org</a>
	 */
	public void glPixelTransformParameterivEXT(GLPixelTransformTargetEXT target, GLPixelTransformPNameEXT pname, ConstPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glPixelTransformParameterivEXT(\{target}, \{pname}, \{params})");
		try {
		    functions.glPixelTransformParameterivEXT().invokeExact(target, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glPixelZoom">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glPixelZoom.xhtml">registry.khronos.org</a>
	 */
	public void glPixelZoom(float xfactor, float yfactor) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glPixelZoom(\{xfactor}, \{yfactor})");
		try {
		    functions.glPixelZoom().invokeExact(xfactor, yfactor);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glPixelZoomxOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glPixelZoomxOES.xhtml">registry.khronos.org</a>
	 */
	public void glPixelZoomxOES(int xfactor, int yfactor) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glPixelZoomxOES(\{xfactor}, \{yfactor})");
		try {
		    functions.glPixelZoomxOES().invokeExact(xfactor, yfactor);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glPointAlongPathNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glPointAlongPathNV.xhtml">registry.khronos.org</a>
	 */
	public boolean glPointAlongPathNV(int path, int startSegment, int numSegments, float distance, MutPointer<Float> x, MutPointer<Float> y, MutPointer<Float> tangentX, MutPointer<Float> tangentY) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glPointAlongPathNV(\{path}, \{startSegment}, \{numSegments}, \{distance}, \{x}, \{y}, \{tangentX}, \{tangentY})");
		try {
		    var r = (boolean)functions.glPointAlongPathNV().invokeExact(path, startSegment, numSegments, distance, x, y, tangentX, tangentY);
			checkError();
			return r;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glPointParameterf">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glPointParameterf.xhtml">registry.khronos.org</a>
	 */
	public void glPointParameterf(GLPointParameterNameARB pname, float param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glPointParameterf(\{pname}, \{param})");
		try {
		    functions.glPointParameterf().invokeExact(pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glPointParameterfARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glPointParameterfARB.xhtml">registry.khronos.org</a>
	 */
	public void glPointParameterfARB(GLPointParameterNameARB pname, float param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glPointParameterfARB(\{pname}, \{param})");
		try {
		    functions.glPointParameterfARB().invokeExact(pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glPointParameterfEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glPointParameterfEXT.xhtml">registry.khronos.org</a>
	 */
	public void glPointParameterfEXT(GLPointParameterNameARB pname, float param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glPointParameterfEXT(\{pname}, \{param})");
		try {
		    functions.glPointParameterfEXT().invokeExact(pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glPointParameterfSGIS">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glPointParameterfSGIS.xhtml">registry.khronos.org</a>
	 */
	public void glPointParameterfSGIS(GLPointParameterNameARB pname, float param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glPointParameterfSGIS(\{pname}, \{param})");
		try {
		    functions.glPointParameterfSGIS().invokeExact(pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glPointParameterfv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glPointParameterfv.xhtml">registry.khronos.org</a>
	 */
	public void glPointParameterfv(GLPointParameterNameARB pname, ConstPointer<Float> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glPointParameterfv(\{pname}, \{params})");
		try {
		    functions.glPointParameterfv().invokeExact(pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glPointParameterfvARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glPointParameterfvARB.xhtml">registry.khronos.org</a>
	 */
	public void glPointParameterfvARB(GLPointParameterNameARB pname, ConstPointer<Float> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glPointParameterfvARB(\{pname}, \{params})");
		try {
		    functions.glPointParameterfvARB().invokeExact(pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glPointParameterfvEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glPointParameterfvEXT.xhtml">registry.khronos.org</a>
	 */
	public void glPointParameterfvEXT(GLPointParameterNameARB pname, ConstPointer<Float> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glPointParameterfvEXT(\{pname}, \{params})");
		try {
		    functions.glPointParameterfvEXT().invokeExact(pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glPointParameterfvSGIS">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glPointParameterfvSGIS.xhtml">registry.khronos.org</a>
	 */
	public void glPointParameterfvSGIS(GLPointParameterNameARB pname, ConstPointer<Float> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glPointParameterfvSGIS(\{pname}, \{params})");
		try {
		    functions.glPointParameterfvSGIS().invokeExact(pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glPointParameteri">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glPointParameteri.xhtml">registry.khronos.org</a>
	 */
	public void glPointParameteri(GLPointParameterNameARB pname, int param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glPointParameteri(\{pname}, \{param})");
		try {
		    functions.glPointParameteri().invokeExact(pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glPointParameteriNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glPointParameteriNV.xhtml">registry.khronos.org</a>
	 */
	public void glPointParameteriNV(GLPointParameterNameARB pname, int param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glPointParameteriNV(\{pname}, \{param})");
		try {
		    functions.glPointParameteriNV().invokeExact(pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glPointParameteriv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glPointParameteriv.xhtml">registry.khronos.org</a>
	 */
	public void glPointParameteriv(GLPointParameterNameARB pname, ConstPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glPointParameteriv(\{pname}, \{params})");
		try {
		    functions.glPointParameteriv().invokeExact(pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glPointParameterivNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glPointParameterivNV.xhtml">registry.khronos.org</a>
	 */
	public void glPointParameterivNV(GLPointParameterNameARB pname, ConstPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glPointParameterivNV(\{pname}, \{params})");
		try {
		    functions.glPointParameterivNV().invokeExact(pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glPointParameterx">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glPointParameterx.xhtml">registry.khronos.org</a>
	 */
	public void glPointParameterx(GLPointParameterNameARB pname, int param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glPointParameterx(\{pname}, \{param})");
		try {
		    functions.glPointParameterx().invokeExact(pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glPointParameterxOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glPointParameterxOES.xhtml">registry.khronos.org</a>
	 */
	public void glPointParameterxOES(GLPointParameterNameARB pname, int param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glPointParameterxOES(\{pname}, \{param})");
		try {
		    functions.glPointParameterxOES().invokeExact(pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glPointParameterxv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glPointParameterxv.xhtml">registry.khronos.org</a>
	 */
	public void glPointParameterxv(GLPointParameterNameARB pname, ConstPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glPointParameterxv(\{pname}, \{params})");
		try {
		    functions.glPointParameterxv().invokeExact(pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glPointParameterxvOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glPointParameterxvOES.xhtml">registry.khronos.org</a>
	 */
	public void glPointParameterxvOES(GLPointParameterNameARB pname, ConstPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glPointParameterxvOES(\{pname}, \{params})");
		try {
		    functions.glPointParameterxvOES().invokeExact(pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glPointSize">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glPointSize.xhtml">registry.khronos.org</a>
	 */
	public void glPointSize(float size) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glPointSize(\{size})");
		try {
		    functions.glPointSize().invokeExact(size);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glPointSizePointerOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glPointSizePointerOES.xhtml">registry.khronos.org</a>
	 */
	public void glPointSizePointerOES(GLEnum type, int stride, ConstPointer<Void> pointer) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glPointSizePointerOES(\{type}, \{stride}, \{pointer})");
		try {
		    functions.glPointSizePointerOES().invokeExact(type, stride, pointer);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glPointSizex">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glPointSizex.xhtml">registry.khronos.org</a>
	 */
	public void glPointSizex(int size) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glPointSizex(\{size})");
		try {
		    functions.glPointSizex().invokeExact(size);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glPointSizexOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glPointSizexOES.xhtml">registry.khronos.org</a>
	 */
	public void glPointSizexOES(int size) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glPointSizexOES(\{size})");
		try {
		    functions.glPointSizexOES().invokeExact(size);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glPollAsyncSGIX">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glPollAsyncSGIX.xhtml">registry.khronos.org</a>
	 */
	public int glPollAsyncSGIX(MutPointer<Integer> markerp) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glPollAsyncSGIX(\{markerp})");
		try {
		    var r = (int)functions.glPollAsyncSGIX().invokeExact(markerp);
			checkError();
			return r;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glPollInstrumentsSGIX">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glPollInstrumentsSGIX.xhtml">registry.khronos.org</a>
	 */
	public int glPollInstrumentsSGIX(MutPointer<Integer> marker_p) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glPollInstrumentsSGIX(\{marker_p})");
		try {
		    var r = (int)functions.glPollInstrumentsSGIX().invokeExact(marker_p);
			checkError();
			return r;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glPolygonMode">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glPolygonMode.xhtml">registry.khronos.org</a>
	 */
	public void glPolygonMode(GLTriangleFace face, GLPolygonMode mode) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glPolygonMode(\{face}, \{mode})");
		try {
		    functions.glPolygonMode().invokeExact(face, mode);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were face=\{face}, mode=\{mode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glPolygonModeNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glPolygonModeNV.xhtml">registry.khronos.org</a>
	 */
	public void glPolygonModeNV(GLTriangleFace face, GLPolygonMode mode) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glPolygonModeNV(\{face}, \{mode})");
		try {
		    functions.glPolygonModeNV().invokeExact(face, mode);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were face=\{face}, mode=\{mode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glPolygonOffset">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glPolygonOffset.xhtml">registry.khronos.org</a>
	 */
	public void glPolygonOffset(float factor, float units) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glPolygonOffset(\{factor}, \{units})");
		try {
		    functions.glPolygonOffset().invokeExact(factor, units);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glPolygonOffsetClamp">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glPolygonOffsetClamp.xhtml">registry.khronos.org</a>
	 */
	public void glPolygonOffsetClamp(float factor, float units, float clamp) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glPolygonOffsetClamp(\{factor}, \{units}, \{clamp})");
		try {
		    functions.glPolygonOffsetClamp().invokeExact(factor, units, clamp);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glPolygonOffsetClampEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glPolygonOffsetClampEXT.xhtml">registry.khronos.org</a>
	 */
	public void glPolygonOffsetClampEXT(float factor, float units, float clamp) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glPolygonOffsetClampEXT(\{factor}, \{units}, \{clamp})");
		try {
		    functions.glPolygonOffsetClampEXT().invokeExact(factor, units, clamp);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glPolygonOffsetEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glPolygonOffsetEXT.xhtml">registry.khronos.org</a>
	 */
	public void glPolygonOffsetEXT(float factor, float bias) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glPolygonOffsetEXT(\{factor}, \{bias})");
		try {
		    functions.glPolygonOffsetEXT().invokeExact(factor, bias);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glPolygonOffsetx">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glPolygonOffsetx.xhtml">registry.khronos.org</a>
	 */
	public void glPolygonOffsetx(int factor, int units) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glPolygonOffsetx(\{factor}, \{units})");
		try {
		    functions.glPolygonOffsetx().invokeExact(factor, units);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glPolygonOffsetxOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glPolygonOffsetxOES.xhtml">registry.khronos.org</a>
	 */
	public void glPolygonOffsetxOES(int factor, int units) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glPolygonOffsetxOES(\{factor}, \{units})");
		try {
		    functions.glPolygonOffsetxOES().invokeExact(factor, units);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glPolygonStipple">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glPolygonStipple.xhtml">registry.khronos.org</a>
	 */
	public void glPolygonStipple(ConstPointer<Byte> mask) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glPolygonStipple(\{mask})");
		try {
		    functions.glPolygonStipple().invokeExact(mask);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glPopAttrib">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glPopAttrib.xhtml">registry.khronos.org</a>
	 */
	public void glPopAttrib() {
	    if (LOG_COMMANDS) logCmd(() -> STR."glPopAttrib()");
		try {
		    functions.glPopAttrib().invokeExact();
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glPopClientAttrib">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glPopClientAttrib.xhtml">registry.khronos.org</a>
	 */
	public void glPopClientAttrib() {
	    if (LOG_COMMANDS) logCmd(() -> STR."glPopClientAttrib()");
		try {
		    functions.glPopClientAttrib().invokeExact();
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glPopDebugGroup">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glPopDebugGroup.xhtml">registry.khronos.org</a>
	 */
	public void glPopDebugGroup() {
	    if (LOG_COMMANDS) logCmd(() -> STR."glPopDebugGroup()");
		try {
		    functions.glPopDebugGroup().invokeExact();
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glPopDebugGroupKHR">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glPopDebugGroupKHR.xhtml">registry.khronos.org</a>
	 */
	public void glPopDebugGroupKHR() {
	    if (LOG_COMMANDS) logCmd(() -> STR."glPopDebugGroupKHR()");
		try {
		    functions.glPopDebugGroupKHR().invokeExact();
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glPopGroupMarkerEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glPopGroupMarkerEXT.xhtml">registry.khronos.org</a>
	 */
	public void glPopGroupMarkerEXT() {
	    if (LOG_COMMANDS) logCmd(() -> STR."glPopGroupMarkerEXT()");
		try {
		    functions.glPopGroupMarkerEXT().invokeExact();
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glPopMatrix">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glPopMatrix.xhtml">registry.khronos.org</a>
	 */
	public void glPopMatrix() {
	    if (LOG_COMMANDS) logCmd(() -> STR."glPopMatrix()");
		try {
		    functions.glPopMatrix().invokeExact();
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glPopName">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glPopName.xhtml">registry.khronos.org</a>
	 */
	public void glPopName() {
	    if (LOG_COMMANDS) logCmd(() -> STR."glPopName()");
		try {
		    functions.glPopName().invokeExact();
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glPresentFrameDualFillNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glPresentFrameDualFillNV.xhtml">registry.khronos.org</a>
	 */
	public void glPresentFrameDualFillNV(int video_slot, long minPresentTime, int beginPresentTimeId, int presentDurationId, GLEnum type, GLEnum target0, int fill0, GLEnum target1, int fill1, GLEnum target2, int fill2, GLEnum target3, int fill3) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glPresentFrameDualFillNV(\{video_slot}, \{minPresentTime}, \{beginPresentTimeId}, \{presentDurationId}, \{type}, \{target0}, \{fill0}, \{target1}, \{fill1}, \{target2}, \{fill2}, \{target3}, \{fill3})");
		try {
		    functions.glPresentFrameDualFillNV().invokeExact(video_slot, minPresentTime, beginPresentTimeId, presentDurationId, type, target0, fill0, target1, fill1, target2, fill2, target3, fill3);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}, target0=\{target0}, target1=\{target1}, target2=\{target2}, target3=\{target3}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glPresentFrameKeyedNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glPresentFrameKeyedNV.xhtml">registry.khronos.org</a>
	 */
	public void glPresentFrameKeyedNV(int video_slot, long minPresentTime, int beginPresentTimeId, int presentDurationId, GLEnum type, GLEnum target0, int fill0, int key0, GLEnum target1, int fill1, int key1) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glPresentFrameKeyedNV(\{video_slot}, \{minPresentTime}, \{beginPresentTimeId}, \{presentDurationId}, \{type}, \{target0}, \{fill0}, \{key0}, \{target1}, \{fill1}, \{key1})");
		try {
		    functions.glPresentFrameKeyedNV().invokeExact(video_slot, minPresentTime, beginPresentTimeId, presentDurationId, type, target0, fill0, key0, target1, fill1, key1);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}, target0=\{target0}, target1=\{target1}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glPrimitiveBoundingBox">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glPrimitiveBoundingBox.xhtml">registry.khronos.org</a>
	 */
	public void glPrimitiveBoundingBox(float minX, float minY, float minZ, float minW, float maxX, float maxY, float maxZ, float maxW) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glPrimitiveBoundingBox(\{minX}, \{minY}, \{minZ}, \{minW}, \{maxX}, \{maxY}, \{maxZ}, \{maxW})");
		try {
		    functions.glPrimitiveBoundingBox().invokeExact(minX, minY, minZ, minW, maxX, maxY, maxZ, maxW);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glPrimitiveBoundingBoxARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glPrimitiveBoundingBoxARB.xhtml">registry.khronos.org</a>
	 */
	public void glPrimitiveBoundingBoxARB(float minX, float minY, float minZ, float minW, float maxX, float maxY, float maxZ, float maxW) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glPrimitiveBoundingBoxARB(\{minX}, \{minY}, \{minZ}, \{minW}, \{maxX}, \{maxY}, \{maxZ}, \{maxW})");
		try {
		    functions.glPrimitiveBoundingBoxARB().invokeExact(minX, minY, minZ, minW, maxX, maxY, maxZ, maxW);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glPrimitiveBoundingBoxEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glPrimitiveBoundingBoxEXT.xhtml">registry.khronos.org</a>
	 */
	public void glPrimitiveBoundingBoxEXT(float minX, float minY, float minZ, float minW, float maxX, float maxY, float maxZ, float maxW) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glPrimitiveBoundingBoxEXT(\{minX}, \{minY}, \{minZ}, \{minW}, \{maxX}, \{maxY}, \{maxZ}, \{maxW})");
		try {
		    functions.glPrimitiveBoundingBoxEXT().invokeExact(minX, minY, minZ, minW, maxX, maxY, maxZ, maxW);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glPrimitiveBoundingBoxOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glPrimitiveBoundingBoxOES.xhtml">registry.khronos.org</a>
	 */
	public void glPrimitiveBoundingBoxOES(float minX, float minY, float minZ, float minW, float maxX, float maxY, float maxZ, float maxW) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glPrimitiveBoundingBoxOES(\{minX}, \{minY}, \{minZ}, \{minW}, \{maxX}, \{maxY}, \{maxZ}, \{maxW})");
		try {
		    functions.glPrimitiveBoundingBoxOES().invokeExact(minX, minY, minZ, minW, maxX, maxY, maxZ, maxW);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glPrimitiveRestartIndex">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glPrimitiveRestartIndex.xhtml">registry.khronos.org</a>
	 */
	public void glPrimitiveRestartIndex(int index) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glPrimitiveRestartIndex(\{index})");
		try {
		    functions.glPrimitiveRestartIndex().invokeExact(index);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glPrimitiveRestartIndexNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glPrimitiveRestartIndexNV.xhtml">registry.khronos.org</a>
	 */
	public void glPrimitiveRestartIndexNV(int index) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glPrimitiveRestartIndexNV(\{index})");
		try {
		    functions.glPrimitiveRestartIndexNV().invokeExact(index);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glPrimitiveRestartNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glPrimitiveRestartNV.xhtml">registry.khronos.org</a>
	 */
	public void glPrimitiveRestartNV() {
	    if (LOG_COMMANDS) logCmd(() -> STR."glPrimitiveRestartNV()");
		try {
		    functions.glPrimitiveRestartNV().invokeExact();
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glPrioritizeTextures">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glPrioritizeTextures.xhtml">registry.khronos.org</a>
	 */
	public void glPrioritizeTextures(int n, ConstPointer<Integer> textures, ConstPointer<Float> priorities) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glPrioritizeTextures(\{n}, \{textures}, \{priorities})");
		try {
		    functions.glPrioritizeTextures().invokeExact(n, textures, priorities);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glPrioritizeTexturesEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glPrioritizeTexturesEXT.xhtml">registry.khronos.org</a>
	 */
	public void glPrioritizeTexturesEXT(int n, ConstPointer<Integer> textures, ConstPointer<Float> priorities) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glPrioritizeTexturesEXT(\{n}, \{textures}, \{priorities})");
		try {
		    functions.glPrioritizeTexturesEXT().invokeExact(n, textures, priorities);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glPrioritizeTexturesxOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glPrioritizeTexturesxOES.xhtml">registry.khronos.org</a>
	 */
	public void glPrioritizeTexturesxOES(int n, ConstPointer<Integer> textures, ConstPointer<Integer> priorities) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glPrioritizeTexturesxOES(\{n}, \{textures}, \{priorities})");
		try {
		    functions.glPrioritizeTexturesxOES().invokeExact(n, textures, priorities);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramBinary">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramBinary.xhtml">registry.khronos.org</a>
	 */
	public void glProgramBinary(int program, GLEnum binaryFormat, ConstPointer<Void> binary, int length) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramBinary(\{program}, \{binaryFormat}, \{binary}, \{length})");
		try {
		    functions.glProgramBinary().invokeExact(program, binaryFormat, binary, length);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were binaryFormat=\{binaryFormat}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramBinaryOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramBinaryOES.xhtml">registry.khronos.org</a>
	 */
	public void glProgramBinaryOES(int program, GLEnum binaryFormat, ConstPointer<Void> binary, int length) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramBinaryOES(\{program}, \{binaryFormat}, \{binary}, \{length})");
		try {
		    functions.glProgramBinaryOES().invokeExact(program, binaryFormat, binary, length);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were binaryFormat=\{binaryFormat}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramBufferParametersIivNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramBufferParametersIivNV.xhtml">registry.khronos.org</a>
	 */
	public void glProgramBufferParametersIivNV(GLProgramTarget target, int bindingIndex, int wordIndex, int count, ConstPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramBufferParametersIivNV(\{target}, \{bindingIndex}, \{wordIndex}, \{count}, \{params})");
		try {
		    functions.glProgramBufferParametersIivNV().invokeExact(target, bindingIndex, wordIndex, count, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramBufferParametersIuivNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramBufferParametersIuivNV.xhtml">registry.khronos.org</a>
	 */
	public void glProgramBufferParametersIuivNV(GLProgramTarget target, int bindingIndex, int wordIndex, int count, ConstPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramBufferParametersIuivNV(\{target}, \{bindingIndex}, \{wordIndex}, \{count}, \{params})");
		try {
		    functions.glProgramBufferParametersIuivNV().invokeExact(target, bindingIndex, wordIndex, count, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramBufferParametersfvNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramBufferParametersfvNV.xhtml">registry.khronos.org</a>
	 */
	public void glProgramBufferParametersfvNV(GLProgramTarget target, int bindingIndex, int wordIndex, int count, ConstPointer<Float> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramBufferParametersfvNV(\{target}, \{bindingIndex}, \{wordIndex}, \{count}, \{params})");
		try {
		    functions.glProgramBufferParametersfvNV().invokeExact(target, bindingIndex, wordIndex, count, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramEnvParameter4dARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramEnvParameter4dARB.xhtml">registry.khronos.org</a>
	 */
	public void glProgramEnvParameter4dARB(GLProgramTarget target, int index, double x, double y, double z, double w) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramEnvParameter4dARB(\{target}, \{index}, \{x}, \{y}, \{z}, \{w})");
		try {
		    functions.glProgramEnvParameter4dARB().invokeExact(target, index, x, y, z, w);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramEnvParameter4dvARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramEnvParameter4dvARB.xhtml">registry.khronos.org</a>
	 */
	public void glProgramEnvParameter4dvARB(GLProgramTarget target, int index, ConstPointer<Double> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramEnvParameter4dvARB(\{target}, \{index}, \{params})");
		try {
		    functions.glProgramEnvParameter4dvARB().invokeExact(target, index, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramEnvParameter4fARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramEnvParameter4fARB.xhtml">registry.khronos.org</a>
	 */
	public void glProgramEnvParameter4fARB(GLProgramTarget target, int index, float x, float y, float z, float w) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramEnvParameter4fARB(\{target}, \{index}, \{x}, \{y}, \{z}, \{w})");
		try {
		    functions.glProgramEnvParameter4fARB().invokeExact(target, index, x, y, z, w);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramEnvParameter4fvARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramEnvParameter4fvARB.xhtml">registry.khronos.org</a>
	 */
	public void glProgramEnvParameter4fvARB(GLProgramTarget target, int index, ConstPointer<Float> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramEnvParameter4fvARB(\{target}, \{index}, \{params})");
		try {
		    functions.glProgramEnvParameter4fvARB().invokeExact(target, index, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramEnvParameterI4iNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramEnvParameterI4iNV.xhtml">registry.khronos.org</a>
	 */
	public void glProgramEnvParameterI4iNV(GLProgramTarget target, int index, int x, int y, int z, int w) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramEnvParameterI4iNV(\{target}, \{index}, \{x}, \{y}, \{z}, \{w})");
		try {
		    functions.glProgramEnvParameterI4iNV().invokeExact(target, index, x, y, z, w);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramEnvParameterI4ivNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramEnvParameterI4ivNV.xhtml">registry.khronos.org</a>
	 */
	public void glProgramEnvParameterI4ivNV(GLProgramTarget target, int index, ConstPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramEnvParameterI4ivNV(\{target}, \{index}, \{params})");
		try {
		    functions.glProgramEnvParameterI4ivNV().invokeExact(target, index, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramEnvParameterI4uiNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramEnvParameterI4uiNV.xhtml">registry.khronos.org</a>
	 */
	public void glProgramEnvParameterI4uiNV(GLProgramTarget target, int index, int x, int y, int z, int w) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramEnvParameterI4uiNV(\{target}, \{index}, \{x}, \{y}, \{z}, \{w})");
		try {
		    functions.glProgramEnvParameterI4uiNV().invokeExact(target, index, x, y, z, w);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramEnvParameterI4uivNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramEnvParameterI4uivNV.xhtml">registry.khronos.org</a>
	 */
	public void glProgramEnvParameterI4uivNV(GLProgramTarget target, int index, ConstPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramEnvParameterI4uivNV(\{target}, \{index}, \{params})");
		try {
		    functions.glProgramEnvParameterI4uivNV().invokeExact(target, index, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramEnvParameters4fvEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramEnvParameters4fvEXT.xhtml">registry.khronos.org</a>
	 */
	public void glProgramEnvParameters4fvEXT(GLProgramTarget target, int index, int count, ConstPointer<Float> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramEnvParameters4fvEXT(\{target}, \{index}, \{count}, \{params})");
		try {
		    functions.glProgramEnvParameters4fvEXT().invokeExact(target, index, count, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramEnvParametersI4ivNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramEnvParametersI4ivNV.xhtml">registry.khronos.org</a>
	 */
	public void glProgramEnvParametersI4ivNV(GLProgramTarget target, int index, int count, ConstPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramEnvParametersI4ivNV(\{target}, \{index}, \{count}, \{params})");
		try {
		    functions.glProgramEnvParametersI4ivNV().invokeExact(target, index, count, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramEnvParametersI4uivNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramEnvParametersI4uivNV.xhtml">registry.khronos.org</a>
	 */
	public void glProgramEnvParametersI4uivNV(GLProgramTarget target, int index, int count, ConstPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramEnvParametersI4uivNV(\{target}, \{index}, \{count}, \{params})");
		try {
		    functions.glProgramEnvParametersI4uivNV().invokeExact(target, index, count, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramLocalParameter4dARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramLocalParameter4dARB.xhtml">registry.khronos.org</a>
	 */
	public void glProgramLocalParameter4dARB(GLProgramTarget target, int index, double x, double y, double z, double w) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramLocalParameter4dARB(\{target}, \{index}, \{x}, \{y}, \{z}, \{w})");
		try {
		    functions.glProgramLocalParameter4dARB().invokeExact(target, index, x, y, z, w);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramLocalParameter4dvARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramLocalParameter4dvARB.xhtml">registry.khronos.org</a>
	 */
	public void glProgramLocalParameter4dvARB(GLProgramTarget target, int index, ConstPointer<Double> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramLocalParameter4dvARB(\{target}, \{index}, \{params})");
		try {
		    functions.glProgramLocalParameter4dvARB().invokeExact(target, index, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramLocalParameter4fARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramLocalParameter4fARB.xhtml">registry.khronos.org</a>
	 */
	public void glProgramLocalParameter4fARB(GLProgramTarget target, int index, float x, float y, float z, float w) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramLocalParameter4fARB(\{target}, \{index}, \{x}, \{y}, \{z}, \{w})");
		try {
		    functions.glProgramLocalParameter4fARB().invokeExact(target, index, x, y, z, w);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramLocalParameter4fvARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramLocalParameter4fvARB.xhtml">registry.khronos.org</a>
	 */
	public void glProgramLocalParameter4fvARB(GLProgramTarget target, int index, ConstPointer<Float> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramLocalParameter4fvARB(\{target}, \{index}, \{params})");
		try {
		    functions.glProgramLocalParameter4fvARB().invokeExact(target, index, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramLocalParameterI4iNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramLocalParameterI4iNV.xhtml">registry.khronos.org</a>
	 */
	public void glProgramLocalParameterI4iNV(GLProgramTarget target, int index, int x, int y, int z, int w) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramLocalParameterI4iNV(\{target}, \{index}, \{x}, \{y}, \{z}, \{w})");
		try {
		    functions.glProgramLocalParameterI4iNV().invokeExact(target, index, x, y, z, w);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramLocalParameterI4ivNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramLocalParameterI4ivNV.xhtml">registry.khronos.org</a>
	 */
	public void glProgramLocalParameterI4ivNV(GLProgramTarget target, int index, ConstPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramLocalParameterI4ivNV(\{target}, \{index}, \{params})");
		try {
		    functions.glProgramLocalParameterI4ivNV().invokeExact(target, index, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramLocalParameterI4uiNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramLocalParameterI4uiNV.xhtml">registry.khronos.org</a>
	 */
	public void glProgramLocalParameterI4uiNV(GLProgramTarget target, int index, int x, int y, int z, int w) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramLocalParameterI4uiNV(\{target}, \{index}, \{x}, \{y}, \{z}, \{w})");
		try {
		    functions.glProgramLocalParameterI4uiNV().invokeExact(target, index, x, y, z, w);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramLocalParameterI4uivNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramLocalParameterI4uivNV.xhtml">registry.khronos.org</a>
	 */
	public void glProgramLocalParameterI4uivNV(GLProgramTarget target, int index, ConstPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramLocalParameterI4uivNV(\{target}, \{index}, \{params})");
		try {
		    functions.glProgramLocalParameterI4uivNV().invokeExact(target, index, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramLocalParameters4fvEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramLocalParameters4fvEXT.xhtml">registry.khronos.org</a>
	 */
	public void glProgramLocalParameters4fvEXT(GLProgramTarget target, int index, int count, ConstPointer<Float> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramLocalParameters4fvEXT(\{target}, \{index}, \{count}, \{params})");
		try {
		    functions.glProgramLocalParameters4fvEXT().invokeExact(target, index, count, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramLocalParametersI4ivNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramLocalParametersI4ivNV.xhtml">registry.khronos.org</a>
	 */
	public void glProgramLocalParametersI4ivNV(GLProgramTarget target, int index, int count, ConstPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramLocalParametersI4ivNV(\{target}, \{index}, \{count}, \{params})");
		try {
		    functions.glProgramLocalParametersI4ivNV().invokeExact(target, index, count, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramLocalParametersI4uivNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramLocalParametersI4uivNV.xhtml">registry.khronos.org</a>
	 */
	public void glProgramLocalParametersI4uivNV(GLProgramTarget target, int index, int count, ConstPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramLocalParametersI4uivNV(\{target}, \{index}, \{count}, \{params})");
		try {
		    functions.glProgramLocalParametersI4uivNV().invokeExact(target, index, count, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramNamedParameter4dNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramNamedParameter4dNV.xhtml">registry.khronos.org</a>
	 */
	public void glProgramNamedParameter4dNV(int id, int len, ConstPointer<Byte> name, double x, double y, double z, double w) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramNamedParameter4dNV(\{id}, \{len}, \{name}, \{x}, \{y}, \{z}, \{w})");
		try {
		    functions.glProgramNamedParameter4dNV().invokeExact(id, len, name, x, y, z, w);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramNamedParameter4dvNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramNamedParameter4dvNV.xhtml">registry.khronos.org</a>
	 */
	public void glProgramNamedParameter4dvNV(int id, int len, ConstPointer<Byte> name, ConstPointer<Double> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramNamedParameter4dvNV(\{id}, \{len}, \{name}, \{v})");
		try {
		    functions.glProgramNamedParameter4dvNV().invokeExact(id, len, name, v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramNamedParameter4fNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramNamedParameter4fNV.xhtml">registry.khronos.org</a>
	 */
	public void glProgramNamedParameter4fNV(int id, int len, ConstPointer<Byte> name, float x, float y, float z, float w) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramNamedParameter4fNV(\{id}, \{len}, \{name}, \{x}, \{y}, \{z}, \{w})");
		try {
		    functions.glProgramNamedParameter4fNV().invokeExact(id, len, name, x, y, z, w);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramNamedParameter4fvNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramNamedParameter4fvNV.xhtml">registry.khronos.org</a>
	 */
	public void glProgramNamedParameter4fvNV(int id, int len, ConstPointer<Byte> name, ConstPointer<Float> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramNamedParameter4fvNV(\{id}, \{len}, \{name}, \{v})");
		try {
		    functions.glProgramNamedParameter4fvNV().invokeExact(id, len, name, v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramParameter4dNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramParameter4dNV.xhtml">registry.khronos.org</a>
	 */
	public void glProgramParameter4dNV(GLVertexAttribEnumNV target, int index, double x, double y, double z, double w) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramParameter4dNV(\{target}, \{index}, \{x}, \{y}, \{z}, \{w})");
		try {
		    functions.glProgramParameter4dNV().invokeExact(target, index, x, y, z, w);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramParameter4dvNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramParameter4dvNV.xhtml">registry.khronos.org</a>
	 */
	public void glProgramParameter4dvNV(GLVertexAttribEnumNV target, int index, ConstPointer<Double> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramParameter4dvNV(\{target}, \{index}, \{v})");
		try {
		    functions.glProgramParameter4dvNV().invokeExact(target, index, v);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramParameter4fNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramParameter4fNV.xhtml">registry.khronos.org</a>
	 */
	public void glProgramParameter4fNV(GLVertexAttribEnumNV target, int index, float x, float y, float z, float w) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramParameter4fNV(\{target}, \{index}, \{x}, \{y}, \{z}, \{w})");
		try {
		    functions.glProgramParameter4fNV().invokeExact(target, index, x, y, z, w);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramParameter4fvNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramParameter4fvNV.xhtml">registry.khronos.org</a>
	 */
	public void glProgramParameter4fvNV(GLVertexAttribEnumNV target, int index, ConstPointer<Float> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramParameter4fvNV(\{target}, \{index}, \{v})");
		try {
		    functions.glProgramParameter4fvNV().invokeExact(target, index, v);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramParameteri">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramParameteri.xhtml">registry.khronos.org</a>
	 */
	public void glProgramParameteri(int program, GLProgramParameterPName pname, int value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramParameteri(\{program}, \{pname}, \{value})");
		try {
		    functions.glProgramParameteri().invokeExact(program, pname, value);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramParameteriARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramParameteriARB.xhtml">registry.khronos.org</a>
	 */
	public void glProgramParameteriARB(int program, GLProgramParameterPName pname, int value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramParameteriARB(\{program}, \{pname}, \{value})");
		try {
		    functions.glProgramParameteriARB().invokeExact(program, pname, value);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramParameteriEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramParameteriEXT.xhtml">registry.khronos.org</a>
	 */
	public void glProgramParameteriEXT(int program, GLProgramParameterPName pname, int value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramParameteriEXT(\{program}, \{pname}, \{value})");
		try {
		    functions.glProgramParameteriEXT().invokeExact(program, pname, value);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramParameters4dvNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramParameters4dvNV.xhtml">registry.khronos.org</a>
	 */
	public void glProgramParameters4dvNV(GLVertexAttribEnumNV target, int index, int count, ConstPointer<Double> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramParameters4dvNV(\{target}, \{index}, \{count}, \{v})");
		try {
		    functions.glProgramParameters4dvNV().invokeExact(target, index, count, v);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramParameters4fvNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramParameters4fvNV.xhtml">registry.khronos.org</a>
	 */
	public void glProgramParameters4fvNV(GLVertexAttribEnumNV target, int index, int count, ConstPointer<Float> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramParameters4fvNV(\{target}, \{index}, \{count}, \{v})");
		try {
		    functions.glProgramParameters4fvNV().invokeExact(target, index, count, v);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramPathFragmentInputGenNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramPathFragmentInputGenNV.xhtml">registry.khronos.org</a>
	 */
	public void glProgramPathFragmentInputGenNV(int program, int location, GLEnum genMode, int components, ConstPointer<Float> coeffs) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramPathFragmentInputGenNV(\{program}, \{location}, \{genMode}, \{components}, \{coeffs})");
		try {
		    functions.glProgramPathFragmentInputGenNV().invokeExact(program, location, genMode, components, coeffs);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were genMode=\{genMode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramStringARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramStringARB.xhtml">registry.khronos.org</a>
	 */
	public void glProgramStringARB(GLProgramTarget target, GLProgramFormat format, int len, ConstPointer<Void> string) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramStringARB(\{target}, \{format}, \{len}, \{string})");
		try {
		    functions.glProgramStringARB().invokeExact(target, format, len, string);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, format=\{format}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramSubroutineParametersuivNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramSubroutineParametersuivNV.xhtml">registry.khronos.org</a>
	 */
	public void glProgramSubroutineParametersuivNV(GLEnum target, int count, ConstPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramSubroutineParametersuivNV(\{target}, \{count}, \{params})");
		try {
		    functions.glProgramSubroutineParametersuivNV().invokeExact(target, count, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniform1d">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniform1d.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniform1d(int program, int location, double v0) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniform1d(\{program}, \{location}, \{v0})");
		try {
		    functions.glProgramUniform1d().invokeExact(program, location, v0);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniform1dEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniform1dEXT.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniform1dEXT(int program, int location, double x) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniform1dEXT(\{program}, \{location}, \{x})");
		try {
		    functions.glProgramUniform1dEXT().invokeExact(program, location, x);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniform1dv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniform1dv.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniform1dv(int program, int location, int count, ConstPointer<Double> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniform1dv(\{program}, \{location}, \{count}, \{value})");
		try {
		    functions.glProgramUniform1dv().invokeExact(program, location, count, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniform1dvEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniform1dvEXT.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniform1dvEXT(int program, int location, int count, ConstPointer<Double> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniform1dvEXT(\{program}, \{location}, \{count}, \{value})");
		try {
		    functions.glProgramUniform1dvEXT().invokeExact(program, location, count, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniform1f">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniform1f.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniform1f(int program, int location, float v0) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniform1f(\{program}, \{location}, \{v0})");
		try {
		    functions.glProgramUniform1f().invokeExact(program, location, v0);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniform1fEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniform1fEXT.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniform1fEXT(int program, int location, float v0) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniform1fEXT(\{program}, \{location}, \{v0})");
		try {
		    functions.glProgramUniform1fEXT().invokeExact(program, location, v0);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniform1fv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniform1fv.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniform1fv(int program, int location, int count, ConstPointer<Float> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniform1fv(\{program}, \{location}, \{count}, \{value})");
		try {
		    functions.glProgramUniform1fv().invokeExact(program, location, count, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniform1fvEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniform1fvEXT.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniform1fvEXT(int program, int location, int count, ConstPointer<Float> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniform1fvEXT(\{program}, \{location}, \{count}, \{value})");
		try {
		    functions.glProgramUniform1fvEXT().invokeExact(program, location, count, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniform1i">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniform1i.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniform1i(int program, int location, int v0) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniform1i(\{program}, \{location}, \{v0})");
		try {
		    functions.glProgramUniform1i().invokeExact(program, location, v0);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniform1i64ARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniform1i64ARB.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniform1i64ARB(int program, int location, long x) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniform1i64ARB(\{program}, \{location}, \{x})");
		try {
		    functions.glProgramUniform1i64ARB().invokeExact(program, location, x);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniform1i64NV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniform1i64NV.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniform1i64NV(int program, int location, long x) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniform1i64NV(\{program}, \{location}, \{x})");
		try {
		    functions.glProgramUniform1i64NV().invokeExact(program, location, x);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniform1i64vARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniform1i64vARB.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniform1i64vARB(int program, int location, int count, ConstPointer<Long> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniform1i64vARB(\{program}, \{location}, \{count}, \{value})");
		try {
		    functions.glProgramUniform1i64vARB().invokeExact(program, location, count, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniform1i64vNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniform1i64vNV.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniform1i64vNV(int program, int location, int count, ConstPointer<Long> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniform1i64vNV(\{program}, \{location}, \{count}, \{value})");
		try {
		    functions.glProgramUniform1i64vNV().invokeExact(program, location, count, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniform1iEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniform1iEXT.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniform1iEXT(int program, int location, int v0) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniform1iEXT(\{program}, \{location}, \{v0})");
		try {
		    functions.glProgramUniform1iEXT().invokeExact(program, location, v0);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniform1iv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniform1iv.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniform1iv(int program, int location, int count, ConstPointer<Integer> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniform1iv(\{program}, \{location}, \{count}, \{value})");
		try {
		    functions.glProgramUniform1iv().invokeExact(program, location, count, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniform1ivEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniform1ivEXT.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniform1ivEXT(int program, int location, int count, ConstPointer<Integer> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniform1ivEXT(\{program}, \{location}, \{count}, \{value})");
		try {
		    functions.glProgramUniform1ivEXT().invokeExact(program, location, count, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniform1ui">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniform1ui.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniform1ui(int program, int location, int v0) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniform1ui(\{program}, \{location}, \{v0})");
		try {
		    functions.glProgramUniform1ui().invokeExact(program, location, v0);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniform1ui64ARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniform1ui64ARB.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniform1ui64ARB(int program, int location, long x) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniform1ui64ARB(\{program}, \{location}, \{x})");
		try {
		    functions.glProgramUniform1ui64ARB().invokeExact(program, location, x);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniform1ui64NV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniform1ui64NV.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniform1ui64NV(int program, int location, long x) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniform1ui64NV(\{program}, \{location}, \{x})");
		try {
		    functions.glProgramUniform1ui64NV().invokeExact(program, location, x);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniform1ui64vARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniform1ui64vARB.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniform1ui64vARB(int program, int location, int count, ConstPointer<Long> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniform1ui64vARB(\{program}, \{location}, \{count}, \{value})");
		try {
		    functions.glProgramUniform1ui64vARB().invokeExact(program, location, count, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniform1ui64vNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniform1ui64vNV.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniform1ui64vNV(int program, int location, int count, ConstPointer<Long> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniform1ui64vNV(\{program}, \{location}, \{count}, \{value})");
		try {
		    functions.glProgramUniform1ui64vNV().invokeExact(program, location, count, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniform1uiEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniform1uiEXT.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniform1uiEXT(int program, int location, int v0) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniform1uiEXT(\{program}, \{location}, \{v0})");
		try {
		    functions.glProgramUniform1uiEXT().invokeExact(program, location, v0);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniform1uiv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniform1uiv.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniform1uiv(int program, int location, int count, ConstPointer<Integer> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniform1uiv(\{program}, \{location}, \{count}, \{value})");
		try {
		    functions.glProgramUniform1uiv().invokeExact(program, location, count, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniform1uivEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniform1uivEXT.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniform1uivEXT(int program, int location, int count, ConstPointer<Integer> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniform1uivEXT(\{program}, \{location}, \{count}, \{value})");
		try {
		    functions.glProgramUniform1uivEXT().invokeExact(program, location, count, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniform2d">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniform2d.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniform2d(int program, int location, double v0, double v1) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniform2d(\{program}, \{location}, \{v0}, \{v1})");
		try {
		    functions.glProgramUniform2d().invokeExact(program, location, v0, v1);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniform2dEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniform2dEXT.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniform2dEXT(int program, int location, double x, double y) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniform2dEXT(\{program}, \{location}, \{x}, \{y})");
		try {
		    functions.glProgramUniform2dEXT().invokeExact(program, location, x, y);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniform2dv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniform2dv.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniform2dv(int program, int location, int count, ConstPointer<Double> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniform2dv(\{program}, \{location}, \{count}, \{value})");
		try {
		    functions.glProgramUniform2dv().invokeExact(program, location, count, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniform2dvEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniform2dvEXT.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniform2dvEXT(int program, int location, int count, ConstPointer<Double> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniform2dvEXT(\{program}, \{location}, \{count}, \{value})");
		try {
		    functions.glProgramUniform2dvEXT().invokeExact(program, location, count, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniform2f">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniform2f.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniform2f(int program, int location, float v0, float v1) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniform2f(\{program}, \{location}, \{v0}, \{v1})");
		try {
		    functions.glProgramUniform2f().invokeExact(program, location, v0, v1);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniform2fEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniform2fEXT.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniform2fEXT(int program, int location, float v0, float v1) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniform2fEXT(\{program}, \{location}, \{v0}, \{v1})");
		try {
		    functions.glProgramUniform2fEXT().invokeExact(program, location, v0, v1);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniform2fv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniform2fv.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniform2fv(int program, int location, int count, ConstPointer<Float> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniform2fv(\{program}, \{location}, \{count}, \{value})");
		try {
		    functions.glProgramUniform2fv().invokeExact(program, location, count, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniform2fvEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniform2fvEXT.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniform2fvEXT(int program, int location, int count, ConstPointer<Float> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniform2fvEXT(\{program}, \{location}, \{count}, \{value})");
		try {
		    functions.glProgramUniform2fvEXT().invokeExact(program, location, count, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniform2i">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniform2i.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniform2i(int program, int location, int v0, int v1) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniform2i(\{program}, \{location}, \{v0}, \{v1})");
		try {
		    functions.glProgramUniform2i().invokeExact(program, location, v0, v1);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniform2i64ARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniform2i64ARB.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniform2i64ARB(int program, int location, long x, long y) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniform2i64ARB(\{program}, \{location}, \{x}, \{y})");
		try {
		    functions.glProgramUniform2i64ARB().invokeExact(program, location, x, y);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniform2i64NV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniform2i64NV.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniform2i64NV(int program, int location, long x, long y) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniform2i64NV(\{program}, \{location}, \{x}, \{y})");
		try {
		    functions.glProgramUniform2i64NV().invokeExact(program, location, x, y);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniform2i64vARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniform2i64vARB.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniform2i64vARB(int program, int location, int count, ConstPointer<Long> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniform2i64vARB(\{program}, \{location}, \{count}, \{value})");
		try {
		    functions.glProgramUniform2i64vARB().invokeExact(program, location, count, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniform2i64vNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniform2i64vNV.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniform2i64vNV(int program, int location, int count, ConstPointer<Long> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniform2i64vNV(\{program}, \{location}, \{count}, \{value})");
		try {
		    functions.glProgramUniform2i64vNV().invokeExact(program, location, count, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniform2iEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniform2iEXT.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniform2iEXT(int program, int location, int v0, int v1) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniform2iEXT(\{program}, \{location}, \{v0}, \{v1})");
		try {
		    functions.glProgramUniform2iEXT().invokeExact(program, location, v0, v1);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniform2iv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniform2iv.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniform2iv(int program, int location, int count, ConstPointer<Integer> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniform2iv(\{program}, \{location}, \{count}, \{value})");
		try {
		    functions.glProgramUniform2iv().invokeExact(program, location, count, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniform2ivEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniform2ivEXT.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniform2ivEXT(int program, int location, int count, ConstPointer<Integer> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniform2ivEXT(\{program}, \{location}, \{count}, \{value})");
		try {
		    functions.glProgramUniform2ivEXT().invokeExact(program, location, count, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniform2ui">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniform2ui.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniform2ui(int program, int location, int v0, int v1) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniform2ui(\{program}, \{location}, \{v0}, \{v1})");
		try {
		    functions.glProgramUniform2ui().invokeExact(program, location, v0, v1);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniform2ui64ARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniform2ui64ARB.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniform2ui64ARB(int program, int location, long x, long y) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniform2ui64ARB(\{program}, \{location}, \{x}, \{y})");
		try {
		    functions.glProgramUniform2ui64ARB().invokeExact(program, location, x, y);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniform2ui64NV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniform2ui64NV.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniform2ui64NV(int program, int location, long x, long y) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniform2ui64NV(\{program}, \{location}, \{x}, \{y})");
		try {
		    functions.glProgramUniform2ui64NV().invokeExact(program, location, x, y);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniform2ui64vARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniform2ui64vARB.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniform2ui64vARB(int program, int location, int count, ConstPointer<Long> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniform2ui64vARB(\{program}, \{location}, \{count}, \{value})");
		try {
		    functions.glProgramUniform2ui64vARB().invokeExact(program, location, count, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniform2ui64vNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniform2ui64vNV.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniform2ui64vNV(int program, int location, int count, ConstPointer<Long> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniform2ui64vNV(\{program}, \{location}, \{count}, \{value})");
		try {
		    functions.glProgramUniform2ui64vNV().invokeExact(program, location, count, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniform2uiEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniform2uiEXT.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniform2uiEXT(int program, int location, int v0, int v1) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniform2uiEXT(\{program}, \{location}, \{v0}, \{v1})");
		try {
		    functions.glProgramUniform2uiEXT().invokeExact(program, location, v0, v1);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniform2uiv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniform2uiv.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniform2uiv(int program, int location, int count, ConstPointer<Integer> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniform2uiv(\{program}, \{location}, \{count}, \{value})");
		try {
		    functions.glProgramUniform2uiv().invokeExact(program, location, count, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniform2uivEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniform2uivEXT.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniform2uivEXT(int program, int location, int count, ConstPointer<Integer> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniform2uivEXT(\{program}, \{location}, \{count}, \{value})");
		try {
		    functions.glProgramUniform2uivEXT().invokeExact(program, location, count, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniform3d">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniform3d.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniform3d(int program, int location, double v0, double v1, double v2) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniform3d(\{program}, \{location}, \{v0}, \{v1}, \{v2})");
		try {
		    functions.glProgramUniform3d().invokeExact(program, location, v0, v1, v2);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniform3dEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniform3dEXT.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniform3dEXT(int program, int location, double x, double y, double z) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniform3dEXT(\{program}, \{location}, \{x}, \{y}, \{z})");
		try {
		    functions.glProgramUniform3dEXT().invokeExact(program, location, x, y, z);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniform3dv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniform3dv.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniform3dv(int program, int location, int count, ConstPointer<Double> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniform3dv(\{program}, \{location}, \{count}, \{value})");
		try {
		    functions.glProgramUniform3dv().invokeExact(program, location, count, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniform3dvEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniform3dvEXT.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniform3dvEXT(int program, int location, int count, ConstPointer<Double> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniform3dvEXT(\{program}, \{location}, \{count}, \{value})");
		try {
		    functions.glProgramUniform3dvEXT().invokeExact(program, location, count, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniform3f">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniform3f.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniform3f(int program, int location, float v0, float v1, float v2) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniform3f(\{program}, \{location}, \{v0}, \{v1}, \{v2})");
		try {
		    functions.glProgramUniform3f().invokeExact(program, location, v0, v1, v2);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniform3fEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniform3fEXT.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniform3fEXT(int program, int location, float v0, float v1, float v2) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniform3fEXT(\{program}, \{location}, \{v0}, \{v1}, \{v2})");
		try {
		    functions.glProgramUniform3fEXT().invokeExact(program, location, v0, v1, v2);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniform3fv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniform3fv.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniform3fv(int program, int location, int count, ConstPointer<Float> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniform3fv(\{program}, \{location}, \{count}, \{value})");
		try {
		    functions.glProgramUniform3fv().invokeExact(program, location, count, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniform3fvEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniform3fvEXT.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniform3fvEXT(int program, int location, int count, ConstPointer<Float> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniform3fvEXT(\{program}, \{location}, \{count}, \{value})");
		try {
		    functions.glProgramUniform3fvEXT().invokeExact(program, location, count, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniform3i">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniform3i.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniform3i(int program, int location, int v0, int v1, int v2) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniform3i(\{program}, \{location}, \{v0}, \{v1}, \{v2})");
		try {
		    functions.glProgramUniform3i().invokeExact(program, location, v0, v1, v2);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniform3i64ARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniform3i64ARB.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniform3i64ARB(int program, int location, long x, long y, long z) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniform3i64ARB(\{program}, \{location}, \{x}, \{y}, \{z})");
		try {
		    functions.glProgramUniform3i64ARB().invokeExact(program, location, x, y, z);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniform3i64NV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniform3i64NV.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniform3i64NV(int program, int location, long x, long y, long z) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniform3i64NV(\{program}, \{location}, \{x}, \{y}, \{z})");
		try {
		    functions.glProgramUniform3i64NV().invokeExact(program, location, x, y, z);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniform3i64vARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniform3i64vARB.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniform3i64vARB(int program, int location, int count, ConstPointer<Long> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniform3i64vARB(\{program}, \{location}, \{count}, \{value})");
		try {
		    functions.glProgramUniform3i64vARB().invokeExact(program, location, count, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniform3i64vNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniform3i64vNV.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniform3i64vNV(int program, int location, int count, ConstPointer<Long> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniform3i64vNV(\{program}, \{location}, \{count}, \{value})");
		try {
		    functions.glProgramUniform3i64vNV().invokeExact(program, location, count, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniform3iEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniform3iEXT.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniform3iEXT(int program, int location, int v0, int v1, int v2) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniform3iEXT(\{program}, \{location}, \{v0}, \{v1}, \{v2})");
		try {
		    functions.glProgramUniform3iEXT().invokeExact(program, location, v0, v1, v2);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniform3iv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniform3iv.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniform3iv(int program, int location, int count, ConstPointer<Integer> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniform3iv(\{program}, \{location}, \{count}, \{value})");
		try {
		    functions.glProgramUniform3iv().invokeExact(program, location, count, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniform3ivEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniform3ivEXT.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniform3ivEXT(int program, int location, int count, ConstPointer<Integer> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniform3ivEXT(\{program}, \{location}, \{count}, \{value})");
		try {
		    functions.glProgramUniform3ivEXT().invokeExact(program, location, count, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniform3ui">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniform3ui.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniform3ui(int program, int location, int v0, int v1, int v2) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniform3ui(\{program}, \{location}, \{v0}, \{v1}, \{v2})");
		try {
		    functions.glProgramUniform3ui().invokeExact(program, location, v0, v1, v2);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniform3ui64ARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniform3ui64ARB.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniform3ui64ARB(int program, int location, long x, long y, long z) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniform3ui64ARB(\{program}, \{location}, \{x}, \{y}, \{z})");
		try {
		    functions.glProgramUniform3ui64ARB().invokeExact(program, location, x, y, z);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniform3ui64NV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniform3ui64NV.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniform3ui64NV(int program, int location, long x, long y, long z) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniform3ui64NV(\{program}, \{location}, \{x}, \{y}, \{z})");
		try {
		    functions.glProgramUniform3ui64NV().invokeExact(program, location, x, y, z);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniform3ui64vARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniform3ui64vARB.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniform3ui64vARB(int program, int location, int count, ConstPointer<Long> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniform3ui64vARB(\{program}, \{location}, \{count}, \{value})");
		try {
		    functions.glProgramUniform3ui64vARB().invokeExact(program, location, count, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniform3ui64vNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniform3ui64vNV.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniform3ui64vNV(int program, int location, int count, ConstPointer<Long> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniform3ui64vNV(\{program}, \{location}, \{count}, \{value})");
		try {
		    functions.glProgramUniform3ui64vNV().invokeExact(program, location, count, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniform3uiEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniform3uiEXT.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniform3uiEXT(int program, int location, int v0, int v1, int v2) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniform3uiEXT(\{program}, \{location}, \{v0}, \{v1}, \{v2})");
		try {
		    functions.glProgramUniform3uiEXT().invokeExact(program, location, v0, v1, v2);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniform3uiv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniform3uiv.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniform3uiv(int program, int location, int count, ConstPointer<Integer> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniform3uiv(\{program}, \{location}, \{count}, \{value})");
		try {
		    functions.glProgramUniform3uiv().invokeExact(program, location, count, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniform3uivEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniform3uivEXT.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniform3uivEXT(int program, int location, int count, ConstPointer<Integer> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniform3uivEXT(\{program}, \{location}, \{count}, \{value})");
		try {
		    functions.glProgramUniform3uivEXT().invokeExact(program, location, count, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniform4d">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniform4d.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniform4d(int program, int location, double v0, double v1, double v2, double v3) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniform4d(\{program}, \{location}, \{v0}, \{v1}, \{v2}, \{v3})");
		try {
		    functions.glProgramUniform4d().invokeExact(program, location, v0, v1, v2, v3);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniform4dEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniform4dEXT.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniform4dEXT(int program, int location, double x, double y, double z, double w) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniform4dEXT(\{program}, \{location}, \{x}, \{y}, \{z}, \{w})");
		try {
		    functions.glProgramUniform4dEXT().invokeExact(program, location, x, y, z, w);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniform4dv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniform4dv.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniform4dv(int program, int location, int count, ConstPointer<Double> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniform4dv(\{program}, \{location}, \{count}, \{value})");
		try {
		    functions.glProgramUniform4dv().invokeExact(program, location, count, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniform4dvEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniform4dvEXT.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniform4dvEXT(int program, int location, int count, ConstPointer<Double> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniform4dvEXT(\{program}, \{location}, \{count}, \{value})");
		try {
		    functions.glProgramUniform4dvEXT().invokeExact(program, location, count, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniform4f">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniform4f.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniform4f(int program, int location, float v0, float v1, float v2, float v3) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniform4f(\{program}, \{location}, \{v0}, \{v1}, \{v2}, \{v3})");
		try {
		    functions.glProgramUniform4f().invokeExact(program, location, v0, v1, v2, v3);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniform4fEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniform4fEXT.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniform4fEXT(int program, int location, float v0, float v1, float v2, float v3) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniform4fEXT(\{program}, \{location}, \{v0}, \{v1}, \{v2}, \{v3})");
		try {
		    functions.glProgramUniform4fEXT().invokeExact(program, location, v0, v1, v2, v3);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniform4fv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniform4fv.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniform4fv(int program, int location, int count, ConstPointer<Float> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniform4fv(\{program}, \{location}, \{count}, \{value})");
		try {
		    functions.glProgramUniform4fv().invokeExact(program, location, count, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniform4fvEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniform4fvEXT.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniform4fvEXT(int program, int location, int count, ConstPointer<Float> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniform4fvEXT(\{program}, \{location}, \{count}, \{value})");
		try {
		    functions.glProgramUniform4fvEXT().invokeExact(program, location, count, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniform4i">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniform4i.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniform4i(int program, int location, int v0, int v1, int v2, int v3) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniform4i(\{program}, \{location}, \{v0}, \{v1}, \{v2}, \{v3})");
		try {
		    functions.glProgramUniform4i().invokeExact(program, location, v0, v1, v2, v3);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniform4i64ARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniform4i64ARB.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniform4i64ARB(int program, int location, long x, long y, long z, long w) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniform4i64ARB(\{program}, \{location}, \{x}, \{y}, \{z}, \{w})");
		try {
		    functions.glProgramUniform4i64ARB().invokeExact(program, location, x, y, z, w);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniform4i64NV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniform4i64NV.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniform4i64NV(int program, int location, long x, long y, long z, long w) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniform4i64NV(\{program}, \{location}, \{x}, \{y}, \{z}, \{w})");
		try {
		    functions.glProgramUniform4i64NV().invokeExact(program, location, x, y, z, w);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniform4i64vARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniform4i64vARB.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniform4i64vARB(int program, int location, int count, ConstPointer<Long> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniform4i64vARB(\{program}, \{location}, \{count}, \{value})");
		try {
		    functions.glProgramUniform4i64vARB().invokeExact(program, location, count, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniform4i64vNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniform4i64vNV.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniform4i64vNV(int program, int location, int count, ConstPointer<Long> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniform4i64vNV(\{program}, \{location}, \{count}, \{value})");
		try {
		    functions.glProgramUniform4i64vNV().invokeExact(program, location, count, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniform4iEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniform4iEXT.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniform4iEXT(int program, int location, int v0, int v1, int v2, int v3) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniform4iEXT(\{program}, \{location}, \{v0}, \{v1}, \{v2}, \{v3})");
		try {
		    functions.glProgramUniform4iEXT().invokeExact(program, location, v0, v1, v2, v3);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniform4iv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniform4iv.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniform4iv(int program, int location, int count, ConstPointer<Integer> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniform4iv(\{program}, \{location}, \{count}, \{value})");
		try {
		    functions.glProgramUniform4iv().invokeExact(program, location, count, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniform4ivEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniform4ivEXT.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniform4ivEXT(int program, int location, int count, ConstPointer<Integer> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniform4ivEXT(\{program}, \{location}, \{count}, \{value})");
		try {
		    functions.glProgramUniform4ivEXT().invokeExact(program, location, count, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniform4ui">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniform4ui.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniform4ui(int program, int location, int v0, int v1, int v2, int v3) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniform4ui(\{program}, \{location}, \{v0}, \{v1}, \{v2}, \{v3})");
		try {
		    functions.glProgramUniform4ui().invokeExact(program, location, v0, v1, v2, v3);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniform4ui64ARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniform4ui64ARB.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniform4ui64ARB(int program, int location, long x, long y, long z, long w) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniform4ui64ARB(\{program}, \{location}, \{x}, \{y}, \{z}, \{w})");
		try {
		    functions.glProgramUniform4ui64ARB().invokeExact(program, location, x, y, z, w);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniform4ui64NV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniform4ui64NV.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniform4ui64NV(int program, int location, long x, long y, long z, long w) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniform4ui64NV(\{program}, \{location}, \{x}, \{y}, \{z}, \{w})");
		try {
		    functions.glProgramUniform4ui64NV().invokeExact(program, location, x, y, z, w);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniform4ui64vARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniform4ui64vARB.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniform4ui64vARB(int program, int location, int count, ConstPointer<Long> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniform4ui64vARB(\{program}, \{location}, \{count}, \{value})");
		try {
		    functions.glProgramUniform4ui64vARB().invokeExact(program, location, count, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniform4ui64vNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniform4ui64vNV.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniform4ui64vNV(int program, int location, int count, ConstPointer<Long> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniform4ui64vNV(\{program}, \{location}, \{count}, \{value})");
		try {
		    functions.glProgramUniform4ui64vNV().invokeExact(program, location, count, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniform4uiEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniform4uiEXT.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniform4uiEXT(int program, int location, int v0, int v1, int v2, int v3) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniform4uiEXT(\{program}, \{location}, \{v0}, \{v1}, \{v2}, \{v3})");
		try {
		    functions.glProgramUniform4uiEXT().invokeExact(program, location, v0, v1, v2, v3);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniform4uiv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniform4uiv.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniform4uiv(int program, int location, int count, ConstPointer<Integer> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniform4uiv(\{program}, \{location}, \{count}, \{value})");
		try {
		    functions.glProgramUniform4uiv().invokeExact(program, location, count, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniform4uivEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniform4uivEXT.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniform4uivEXT(int program, int location, int count, ConstPointer<Integer> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniform4uivEXT(\{program}, \{location}, \{count}, \{value})");
		try {
		    functions.glProgramUniform4uivEXT().invokeExact(program, location, count, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniformHandleui64ARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniformHandleui64ARB.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniformHandleui64ARB(int program, int location, long value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniformHandleui64ARB(\{program}, \{location}, \{value})");
		try {
		    functions.glProgramUniformHandleui64ARB().invokeExact(program, location, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniformHandleui64IMG">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniformHandleui64IMG.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniformHandleui64IMG(int program, int location, long value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniformHandleui64IMG(\{program}, \{location}, \{value})");
		try {
		    functions.glProgramUniformHandleui64IMG().invokeExact(program, location, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniformHandleui64NV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniformHandleui64NV.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniformHandleui64NV(int program, int location, long value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniformHandleui64NV(\{program}, \{location}, \{value})");
		try {
		    functions.glProgramUniformHandleui64NV().invokeExact(program, location, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniformHandleui64vARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniformHandleui64vARB.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniformHandleui64vARB(int program, int location, int count, ConstPointer<Long> values) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniformHandleui64vARB(\{program}, \{location}, \{count}, \{values})");
		try {
		    functions.glProgramUniformHandleui64vARB().invokeExact(program, location, count, values);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniformHandleui64vIMG">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniformHandleui64vIMG.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniformHandleui64vIMG(int program, int location, int count, ConstPointer<Long> values) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniformHandleui64vIMG(\{program}, \{location}, \{count}, \{values})");
		try {
		    functions.glProgramUniformHandleui64vIMG().invokeExact(program, location, count, values);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniformHandleui64vNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniformHandleui64vNV.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniformHandleui64vNV(int program, int location, int count, ConstPointer<Long> values) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniformHandleui64vNV(\{program}, \{location}, \{count}, \{values})");
		try {
		    functions.glProgramUniformHandleui64vNV().invokeExact(program, location, count, values);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniformMatrix2dv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniformMatrix2dv.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniformMatrix2dv(int program, int location, int count, boolean transpose, ConstPointer<Double> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniformMatrix2dv(\{program}, \{location}, \{count}, \{transpose}, \{value})");
		try {
		    functions.glProgramUniformMatrix2dv().invokeExact(program, location, count, transpose, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniformMatrix2dvEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniformMatrix2dvEXT.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniformMatrix2dvEXT(int program, int location, int count, boolean transpose, ConstPointer<Double> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniformMatrix2dvEXT(\{program}, \{location}, \{count}, \{transpose}, \{value})");
		try {
		    functions.glProgramUniformMatrix2dvEXT().invokeExact(program, location, count, transpose, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniformMatrix2fv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniformMatrix2fv.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniformMatrix2fv(int program, int location, int count, boolean transpose, ConstPointer<Float> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniformMatrix2fv(\{program}, \{location}, \{count}, \{transpose}, \{value})");
		try {
		    functions.glProgramUniformMatrix2fv().invokeExact(program, location, count, transpose, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniformMatrix2fvEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniformMatrix2fvEXT.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniformMatrix2fvEXT(int program, int location, int count, boolean transpose, ConstPointer<Float> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniformMatrix2fvEXT(\{program}, \{location}, \{count}, \{transpose}, \{value})");
		try {
		    functions.glProgramUniformMatrix2fvEXT().invokeExact(program, location, count, transpose, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniformMatrix2x3dv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniformMatrix2x3dv.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniformMatrix2x3dv(int program, int location, int count, boolean transpose, ConstPointer<Double> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniformMatrix2x3dv(\{program}, \{location}, \{count}, \{transpose}, \{value})");
		try {
		    functions.glProgramUniformMatrix2x3dv().invokeExact(program, location, count, transpose, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniformMatrix2x3dvEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniformMatrix2x3dvEXT.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniformMatrix2x3dvEXT(int program, int location, int count, boolean transpose, ConstPointer<Double> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniformMatrix2x3dvEXT(\{program}, \{location}, \{count}, \{transpose}, \{value})");
		try {
		    functions.glProgramUniformMatrix2x3dvEXT().invokeExact(program, location, count, transpose, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniformMatrix2x3fv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniformMatrix2x3fv.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniformMatrix2x3fv(int program, int location, int count, boolean transpose, ConstPointer<Float> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniformMatrix2x3fv(\{program}, \{location}, \{count}, \{transpose}, \{value})");
		try {
		    functions.glProgramUniformMatrix2x3fv().invokeExact(program, location, count, transpose, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniformMatrix2x3fvEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniformMatrix2x3fvEXT.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniformMatrix2x3fvEXT(int program, int location, int count, boolean transpose, ConstPointer<Float> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniformMatrix2x3fvEXT(\{program}, \{location}, \{count}, \{transpose}, \{value})");
		try {
		    functions.glProgramUniformMatrix2x3fvEXT().invokeExact(program, location, count, transpose, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniformMatrix2x4dv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniformMatrix2x4dv.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniformMatrix2x4dv(int program, int location, int count, boolean transpose, ConstPointer<Double> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniformMatrix2x4dv(\{program}, \{location}, \{count}, \{transpose}, \{value})");
		try {
		    functions.glProgramUniformMatrix2x4dv().invokeExact(program, location, count, transpose, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniformMatrix2x4dvEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniformMatrix2x4dvEXT.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniformMatrix2x4dvEXT(int program, int location, int count, boolean transpose, ConstPointer<Double> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniformMatrix2x4dvEXT(\{program}, \{location}, \{count}, \{transpose}, \{value})");
		try {
		    functions.glProgramUniformMatrix2x4dvEXT().invokeExact(program, location, count, transpose, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniformMatrix2x4fv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniformMatrix2x4fv.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniformMatrix2x4fv(int program, int location, int count, boolean transpose, ConstPointer<Float> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniformMatrix2x4fv(\{program}, \{location}, \{count}, \{transpose}, \{value})");
		try {
		    functions.glProgramUniformMatrix2x4fv().invokeExact(program, location, count, transpose, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniformMatrix2x4fvEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniformMatrix2x4fvEXT.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniformMatrix2x4fvEXT(int program, int location, int count, boolean transpose, ConstPointer<Float> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniformMatrix2x4fvEXT(\{program}, \{location}, \{count}, \{transpose}, \{value})");
		try {
		    functions.glProgramUniformMatrix2x4fvEXT().invokeExact(program, location, count, transpose, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniformMatrix3dv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniformMatrix3dv.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniformMatrix3dv(int program, int location, int count, boolean transpose, ConstPointer<Double> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniformMatrix3dv(\{program}, \{location}, \{count}, \{transpose}, \{value})");
		try {
		    functions.glProgramUniformMatrix3dv().invokeExact(program, location, count, transpose, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniformMatrix3dvEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniformMatrix3dvEXT.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniformMatrix3dvEXT(int program, int location, int count, boolean transpose, ConstPointer<Double> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniformMatrix3dvEXT(\{program}, \{location}, \{count}, \{transpose}, \{value})");
		try {
		    functions.glProgramUniformMatrix3dvEXT().invokeExact(program, location, count, transpose, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniformMatrix3fv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniformMatrix3fv.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniformMatrix3fv(int program, int location, int count, boolean transpose, ConstPointer<Float> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniformMatrix3fv(\{program}, \{location}, \{count}, \{transpose}, \{value})");
		try {
		    functions.glProgramUniformMatrix3fv().invokeExact(program, location, count, transpose, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniformMatrix3fvEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniformMatrix3fvEXT.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniformMatrix3fvEXT(int program, int location, int count, boolean transpose, ConstPointer<Float> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniformMatrix3fvEXT(\{program}, \{location}, \{count}, \{transpose}, \{value})");
		try {
		    functions.glProgramUniformMatrix3fvEXT().invokeExact(program, location, count, transpose, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniformMatrix3x2dv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniformMatrix3x2dv.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniformMatrix3x2dv(int program, int location, int count, boolean transpose, ConstPointer<Double> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniformMatrix3x2dv(\{program}, \{location}, \{count}, \{transpose}, \{value})");
		try {
		    functions.glProgramUniformMatrix3x2dv().invokeExact(program, location, count, transpose, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniformMatrix3x2dvEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniformMatrix3x2dvEXT.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniformMatrix3x2dvEXT(int program, int location, int count, boolean transpose, ConstPointer<Double> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniformMatrix3x2dvEXT(\{program}, \{location}, \{count}, \{transpose}, \{value})");
		try {
		    functions.glProgramUniformMatrix3x2dvEXT().invokeExact(program, location, count, transpose, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniformMatrix3x2fv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniformMatrix3x2fv.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniformMatrix3x2fv(int program, int location, int count, boolean transpose, ConstPointer<Float> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniformMatrix3x2fv(\{program}, \{location}, \{count}, \{transpose}, \{value})");
		try {
		    functions.glProgramUniformMatrix3x2fv().invokeExact(program, location, count, transpose, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniformMatrix3x2fvEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniformMatrix3x2fvEXT.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniformMatrix3x2fvEXT(int program, int location, int count, boolean transpose, ConstPointer<Float> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniformMatrix3x2fvEXT(\{program}, \{location}, \{count}, \{transpose}, \{value})");
		try {
		    functions.glProgramUniformMatrix3x2fvEXT().invokeExact(program, location, count, transpose, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniformMatrix3x4dv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniformMatrix3x4dv.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniformMatrix3x4dv(int program, int location, int count, boolean transpose, ConstPointer<Double> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniformMatrix3x4dv(\{program}, \{location}, \{count}, \{transpose}, \{value})");
		try {
		    functions.glProgramUniformMatrix3x4dv().invokeExact(program, location, count, transpose, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniformMatrix3x4dvEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniformMatrix3x4dvEXT.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniformMatrix3x4dvEXT(int program, int location, int count, boolean transpose, ConstPointer<Double> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniformMatrix3x4dvEXT(\{program}, \{location}, \{count}, \{transpose}, \{value})");
		try {
		    functions.glProgramUniformMatrix3x4dvEXT().invokeExact(program, location, count, transpose, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniformMatrix3x4fv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniformMatrix3x4fv.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniformMatrix3x4fv(int program, int location, int count, boolean transpose, ConstPointer<Float> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniformMatrix3x4fv(\{program}, \{location}, \{count}, \{transpose}, \{value})");
		try {
		    functions.glProgramUniformMatrix3x4fv().invokeExact(program, location, count, transpose, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniformMatrix3x4fvEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniformMatrix3x4fvEXT.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniformMatrix3x4fvEXT(int program, int location, int count, boolean transpose, ConstPointer<Float> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniformMatrix3x4fvEXT(\{program}, \{location}, \{count}, \{transpose}, \{value})");
		try {
		    functions.glProgramUniformMatrix3x4fvEXT().invokeExact(program, location, count, transpose, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniformMatrix4dv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniformMatrix4dv.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniformMatrix4dv(int program, int location, int count, boolean transpose, ConstPointer<Double> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniformMatrix4dv(\{program}, \{location}, \{count}, \{transpose}, \{value})");
		try {
		    functions.glProgramUniformMatrix4dv().invokeExact(program, location, count, transpose, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniformMatrix4dvEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniformMatrix4dvEXT.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniformMatrix4dvEXT(int program, int location, int count, boolean transpose, ConstPointer<Double> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniformMatrix4dvEXT(\{program}, \{location}, \{count}, \{transpose}, \{value})");
		try {
		    functions.glProgramUniformMatrix4dvEXT().invokeExact(program, location, count, transpose, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniformMatrix4fv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniformMatrix4fv.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniformMatrix4fv(int program, int location, int count, boolean transpose, ConstPointer<Float> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniformMatrix4fv(\{program}, \{location}, \{count}, \{transpose}, \{value})");
		try {
		    functions.glProgramUniformMatrix4fv().invokeExact(program, location, count, transpose, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniformMatrix4fvEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniformMatrix4fvEXT.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniformMatrix4fvEXT(int program, int location, int count, boolean transpose, ConstPointer<Float> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniformMatrix4fvEXT(\{program}, \{location}, \{count}, \{transpose}, \{value})");
		try {
		    functions.glProgramUniformMatrix4fvEXT().invokeExact(program, location, count, transpose, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniformMatrix4x2dv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniformMatrix4x2dv.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniformMatrix4x2dv(int program, int location, int count, boolean transpose, ConstPointer<Double> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniformMatrix4x2dv(\{program}, \{location}, \{count}, \{transpose}, \{value})");
		try {
		    functions.glProgramUniformMatrix4x2dv().invokeExact(program, location, count, transpose, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniformMatrix4x2dvEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniformMatrix4x2dvEXT.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniformMatrix4x2dvEXT(int program, int location, int count, boolean transpose, ConstPointer<Double> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniformMatrix4x2dvEXT(\{program}, \{location}, \{count}, \{transpose}, \{value})");
		try {
		    functions.glProgramUniformMatrix4x2dvEXT().invokeExact(program, location, count, transpose, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniformMatrix4x2fv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniformMatrix4x2fv.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniformMatrix4x2fv(int program, int location, int count, boolean transpose, ConstPointer<Float> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniformMatrix4x2fv(\{program}, \{location}, \{count}, \{transpose}, \{value})");
		try {
		    functions.glProgramUniformMatrix4x2fv().invokeExact(program, location, count, transpose, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniformMatrix4x2fvEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniformMatrix4x2fvEXT.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniformMatrix4x2fvEXT(int program, int location, int count, boolean transpose, ConstPointer<Float> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniformMatrix4x2fvEXT(\{program}, \{location}, \{count}, \{transpose}, \{value})");
		try {
		    functions.glProgramUniformMatrix4x2fvEXT().invokeExact(program, location, count, transpose, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniformMatrix4x3dv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniformMatrix4x3dv.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniformMatrix4x3dv(int program, int location, int count, boolean transpose, ConstPointer<Double> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniformMatrix4x3dv(\{program}, \{location}, \{count}, \{transpose}, \{value})");
		try {
		    functions.glProgramUniformMatrix4x3dv().invokeExact(program, location, count, transpose, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniformMatrix4x3dvEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniformMatrix4x3dvEXT.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniformMatrix4x3dvEXT(int program, int location, int count, boolean transpose, ConstPointer<Double> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniformMatrix4x3dvEXT(\{program}, \{location}, \{count}, \{transpose}, \{value})");
		try {
		    functions.glProgramUniformMatrix4x3dvEXT().invokeExact(program, location, count, transpose, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniformMatrix4x3fv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniformMatrix4x3fv.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniformMatrix4x3fv(int program, int location, int count, boolean transpose, ConstPointer<Float> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniformMatrix4x3fv(\{program}, \{location}, \{count}, \{transpose}, \{value})");
		try {
		    functions.glProgramUniformMatrix4x3fv().invokeExact(program, location, count, transpose, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniformMatrix4x3fvEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniformMatrix4x3fvEXT.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniformMatrix4x3fvEXT(int program, int location, int count, boolean transpose, ConstPointer<Float> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniformMatrix4x3fvEXT(\{program}, \{location}, \{count}, \{transpose}, \{value})");
		try {
		    functions.glProgramUniformMatrix4x3fvEXT().invokeExact(program, location, count, transpose, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniformui64NV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniformui64NV.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniformui64NV(int program, int location, long value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniformui64NV(\{program}, \{location}, \{value})");
		try {
		    functions.glProgramUniformui64NV().invokeExact(program, location, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramUniformui64vNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramUniformui64vNV.xhtml">registry.khronos.org</a>
	 */
	public void glProgramUniformui64vNV(int program, int location, int count, ConstPointer<Long> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramUniformui64vNV(\{program}, \{location}, \{count}, \{value})");
		try {
		    functions.glProgramUniformui64vNV().invokeExact(program, location, count, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProgramVertexLimitNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProgramVertexLimitNV.xhtml">registry.khronos.org</a>
	 */
	public void glProgramVertexLimitNV(GLProgramTarget target, int limit) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProgramVertexLimitNV(\{target}, \{limit})");
		try {
		    functions.glProgramVertexLimitNV().invokeExact(target, limit);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProvokingVertex">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProvokingVertex.xhtml">registry.khronos.org</a>
	 */
	public void glProvokingVertex(GLVertexProvokingMode mode) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProvokingVertex(\{mode})");
		try {
		    functions.glProvokingVertex().invokeExact(mode);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glProvokingVertexEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glProvokingVertexEXT.xhtml">registry.khronos.org</a>
	 */
	public void glProvokingVertexEXT(GLVertexProvokingMode mode) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glProvokingVertexEXT(\{mode})");
		try {
		    functions.glProvokingVertexEXT().invokeExact(mode);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glPushAttrib">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glPushAttrib.xhtml">registry.khronos.org</a>
	 */
	public void glPushAttrib(int mask) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glPushAttrib(\{mask})");
		try {
		    functions.glPushAttrib().invokeExact(mask);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glPushClientAttrib">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glPushClientAttrib.xhtml">registry.khronos.org</a>
	 */
	public void glPushClientAttrib(int mask) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glPushClientAttrib(\{mask})");
		try {
		    functions.glPushClientAttrib().invokeExact(mask);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glPushClientAttribDefaultEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glPushClientAttribDefaultEXT.xhtml">registry.khronos.org</a>
	 */
	public void glPushClientAttribDefaultEXT(int mask) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glPushClientAttribDefaultEXT(\{mask})");
		try {
		    functions.glPushClientAttribDefaultEXT().invokeExact(mask);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glPushDebugGroup">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glPushDebugGroup.xhtml">registry.khronos.org</a>
	 */
	public void glPushDebugGroup(GLDebugSource source, int id, int length, ConstPointer<Byte> message) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glPushDebugGroup(\{source}, \{id}, \{length}, \{message})");
		try {
		    functions.glPushDebugGroup().invokeExact(source, id, length, message);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were source=\{source}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glPushDebugGroupKHR">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glPushDebugGroupKHR.xhtml">registry.khronos.org</a>
	 */
	public void glPushDebugGroupKHR(GLDebugSource source, int id, int length, ConstPointer<Byte> message) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glPushDebugGroupKHR(\{source}, \{id}, \{length}, \{message})");
		try {
		    functions.glPushDebugGroupKHR().invokeExact(source, id, length, message);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were source=\{source}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glPushGroupMarkerEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glPushGroupMarkerEXT.xhtml">registry.khronos.org</a>
	 */
	public void glPushGroupMarkerEXT(int length, ConstPointer<Byte> marker) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glPushGroupMarkerEXT(\{length}, \{marker})");
		try {
		    functions.glPushGroupMarkerEXT().invokeExact(length, marker);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glPushMatrix">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glPushMatrix.xhtml">registry.khronos.org</a>
	 */
	public void glPushMatrix() {
	    if (LOG_COMMANDS) logCmd(() -> STR."glPushMatrix()");
		try {
		    functions.glPushMatrix().invokeExact();
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glPushName">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glPushName.xhtml">registry.khronos.org</a>
	 */
	public void glPushName(int name) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glPushName(\{name})");
		try {
		    functions.glPushName().invokeExact(name);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glQueryCounter">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glQueryCounter.xhtml">registry.khronos.org</a>
	 */
	public void glQueryCounter(int id, GLQueryCounterTarget target) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glQueryCounter(\{id}, \{target})");
		try {
		    functions.glQueryCounter().invokeExact(id, target);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glQueryCounterEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glQueryCounterEXT.xhtml">registry.khronos.org</a>
	 */
	public void glQueryCounterEXT(int id, GLQueryCounterTarget target) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glQueryCounterEXT(\{id}, \{target})");
		try {
		    functions.glQueryCounterEXT().invokeExact(id, target);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glQueryMatrixxOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glQueryMatrixxOES.xhtml">registry.khronos.org</a>
	 */
	public int glQueryMatrixxOES(MutPointer<Integer> mantissa, MutPointer<Integer> exponent) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glQueryMatrixxOES(\{mantissa}, \{exponent})");
		try {
		    var r = (int)functions.glQueryMatrixxOES().invokeExact(mantissa, exponent);
			checkError();
			return r;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glQueryObjectParameteruiAMD">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glQueryObjectParameteruiAMD.xhtml">registry.khronos.org</a>
	 */
	public void glQueryObjectParameteruiAMD(GLQueryTarget target, int id, GLEnum pname, int param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glQueryObjectParameteruiAMD(\{target}, \{id}, \{pname}, \{param})");
		try {
		    functions.glQueryObjectParameteruiAMD().invokeExact(target, id, pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glQueryResourceNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glQueryResourceNV.xhtml">registry.khronos.org</a>
	 */
	public int glQueryResourceNV(GLEnum queryType, int tagId, int count, MutPointer<Integer> buffer) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glQueryResourceNV(\{queryType}, \{tagId}, \{count}, \{buffer})");
		try {
		    var r = (int)functions.glQueryResourceNV().invokeExact(queryType, tagId, count, buffer);
			checkError();
			return r;
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were queryType=\{queryType}.");
			}
			throw e;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glQueryResourceTagNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glQueryResourceTagNV.xhtml">registry.khronos.org</a>
	 */
	public void glQueryResourceTagNV(int tagId, ConstPointer<Byte> tagString) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glQueryResourceTagNV(\{tagId}, \{tagString})");
		try {
		    functions.glQueryResourceTagNV().invokeExact(tagId, tagString);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glRasterPos2d">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glRasterPos2d.xhtml">registry.khronos.org</a>
	 */
	public void glRasterPos2d(double x, double y) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glRasterPos2d(\{x}, \{y})");
		try {
		    functions.glRasterPos2d().invokeExact(x, y);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glRasterPos2dv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glRasterPos2dv.xhtml">registry.khronos.org</a>
	 */
	public void glRasterPos2dv(ConstPointer<Double> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glRasterPos2dv(\{v})");
		try {
		    functions.glRasterPos2dv().invokeExact(v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glRasterPos2f">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glRasterPos2f.xhtml">registry.khronos.org</a>
	 */
	public void glRasterPos2f(float x, float y) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glRasterPos2f(\{x}, \{y})");
		try {
		    functions.glRasterPos2f().invokeExact(x, y);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glRasterPos2fv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glRasterPos2fv.xhtml">registry.khronos.org</a>
	 */
	public void glRasterPos2fv(ConstPointer<Float> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glRasterPos2fv(\{v})");
		try {
		    functions.glRasterPos2fv().invokeExact(v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glRasterPos2i">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glRasterPos2i.xhtml">registry.khronos.org</a>
	 */
	public void glRasterPos2i(int x, int y) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glRasterPos2i(\{x}, \{y})");
		try {
		    functions.glRasterPos2i().invokeExact(x, y);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glRasterPos2iv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glRasterPos2iv.xhtml">registry.khronos.org</a>
	 */
	public void glRasterPos2iv(ConstPointer<Integer> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glRasterPos2iv(\{v})");
		try {
		    functions.glRasterPos2iv().invokeExact(v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glRasterPos2s">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glRasterPos2s.xhtml">registry.khronos.org</a>
	 */
	public void glRasterPos2s(short x, short y) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glRasterPos2s(\{x}, \{y})");
		try {
		    functions.glRasterPos2s().invokeExact(x, y);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glRasterPos2sv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glRasterPos2sv.xhtml">registry.khronos.org</a>
	 */
	public void glRasterPos2sv(ConstPointer<Short> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glRasterPos2sv(\{v})");
		try {
		    functions.glRasterPos2sv().invokeExact(v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glRasterPos2xOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glRasterPos2xOES.xhtml">registry.khronos.org</a>
	 */
	public void glRasterPos2xOES(int x, int y) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glRasterPos2xOES(\{x}, \{y})");
		try {
		    functions.glRasterPos2xOES().invokeExact(x, y);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glRasterPos2xvOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glRasterPos2xvOES.xhtml">registry.khronos.org</a>
	 */
	public void glRasterPos2xvOES(ConstPointer<Integer> coords) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glRasterPos2xvOES(\{coords})");
		try {
		    functions.glRasterPos2xvOES().invokeExact(coords);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glRasterPos3d">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glRasterPos3d.xhtml">registry.khronos.org</a>
	 */
	public void glRasterPos3d(double x, double y, double z) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glRasterPos3d(\{x}, \{y}, \{z})");
		try {
		    functions.glRasterPos3d().invokeExact(x, y, z);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glRasterPos3dv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glRasterPos3dv.xhtml">registry.khronos.org</a>
	 */
	public void glRasterPos3dv(ConstPointer<Double> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glRasterPos3dv(\{v})");
		try {
		    functions.glRasterPos3dv().invokeExact(v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glRasterPos3f">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glRasterPos3f.xhtml">registry.khronos.org</a>
	 */
	public void glRasterPos3f(float x, float y, float z) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glRasterPos3f(\{x}, \{y}, \{z})");
		try {
		    functions.glRasterPos3f().invokeExact(x, y, z);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glRasterPos3fv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glRasterPos3fv.xhtml">registry.khronos.org</a>
	 */
	public void glRasterPos3fv(ConstPointer<Float> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glRasterPos3fv(\{v})");
		try {
		    functions.glRasterPos3fv().invokeExact(v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glRasterPos3i">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glRasterPos3i.xhtml">registry.khronos.org</a>
	 */
	public void glRasterPos3i(int x, int y, int z) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glRasterPos3i(\{x}, \{y}, \{z})");
		try {
		    functions.glRasterPos3i().invokeExact(x, y, z);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glRasterPos3iv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glRasterPos3iv.xhtml">registry.khronos.org</a>
	 */
	public void glRasterPos3iv(ConstPointer<Integer> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glRasterPos3iv(\{v})");
		try {
		    functions.glRasterPos3iv().invokeExact(v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glRasterPos3s">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glRasterPos3s.xhtml">registry.khronos.org</a>
	 */
	public void glRasterPos3s(short x, short y, short z) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glRasterPos3s(\{x}, \{y}, \{z})");
		try {
		    functions.glRasterPos3s().invokeExact(x, y, z);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glRasterPos3sv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glRasterPos3sv.xhtml">registry.khronos.org</a>
	 */
	public void glRasterPos3sv(ConstPointer<Short> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glRasterPos3sv(\{v})");
		try {
		    functions.glRasterPos3sv().invokeExact(v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glRasterPos3xOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glRasterPos3xOES.xhtml">registry.khronos.org</a>
	 */
	public void glRasterPos3xOES(int x, int y, int z) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glRasterPos3xOES(\{x}, \{y}, \{z})");
		try {
		    functions.glRasterPos3xOES().invokeExact(x, y, z);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glRasterPos3xvOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glRasterPos3xvOES.xhtml">registry.khronos.org</a>
	 */
	public void glRasterPos3xvOES(ConstPointer<Integer> coords) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glRasterPos3xvOES(\{coords})");
		try {
		    functions.glRasterPos3xvOES().invokeExact(coords);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glRasterPos4d">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glRasterPos4d.xhtml">registry.khronos.org</a>
	 */
	public void glRasterPos4d(double x, double y, double z, double w) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glRasterPos4d(\{x}, \{y}, \{z}, \{w})");
		try {
		    functions.glRasterPos4d().invokeExact(x, y, z, w);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glRasterPos4dv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glRasterPos4dv.xhtml">registry.khronos.org</a>
	 */
	public void glRasterPos4dv(ConstPointer<Double> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glRasterPos4dv(\{v})");
		try {
		    functions.glRasterPos4dv().invokeExact(v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glRasterPos4f">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glRasterPos4f.xhtml">registry.khronos.org</a>
	 */
	public void glRasterPos4f(float x, float y, float z, float w) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glRasterPos4f(\{x}, \{y}, \{z}, \{w})");
		try {
		    functions.glRasterPos4f().invokeExact(x, y, z, w);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glRasterPos4fv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glRasterPos4fv.xhtml">registry.khronos.org</a>
	 */
	public void glRasterPos4fv(ConstPointer<Float> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glRasterPos4fv(\{v})");
		try {
		    functions.glRasterPos4fv().invokeExact(v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glRasterPos4i">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glRasterPos4i.xhtml">registry.khronos.org</a>
	 */
	public void glRasterPos4i(int x, int y, int z, int w) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glRasterPos4i(\{x}, \{y}, \{z}, \{w})");
		try {
		    functions.glRasterPos4i().invokeExact(x, y, z, w);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glRasterPos4iv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glRasterPos4iv.xhtml">registry.khronos.org</a>
	 */
	public void glRasterPos4iv(ConstPointer<Integer> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glRasterPos4iv(\{v})");
		try {
		    functions.glRasterPos4iv().invokeExact(v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glRasterPos4s">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glRasterPos4s.xhtml">registry.khronos.org</a>
	 */
	public void glRasterPos4s(short x, short y, short z, short w) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glRasterPos4s(\{x}, \{y}, \{z}, \{w})");
		try {
		    functions.glRasterPos4s().invokeExact(x, y, z, w);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glRasterPos4sv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glRasterPos4sv.xhtml">registry.khronos.org</a>
	 */
	public void glRasterPos4sv(ConstPointer<Short> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glRasterPos4sv(\{v})");
		try {
		    functions.glRasterPos4sv().invokeExact(v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glRasterPos4xOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glRasterPos4xOES.xhtml">registry.khronos.org</a>
	 */
	public void glRasterPos4xOES(int x, int y, int z, int w) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glRasterPos4xOES(\{x}, \{y}, \{z}, \{w})");
		try {
		    functions.glRasterPos4xOES().invokeExact(x, y, z, w);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glRasterPos4xvOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glRasterPos4xvOES.xhtml">registry.khronos.org</a>
	 */
	public void glRasterPos4xvOES(ConstPointer<Integer> coords) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glRasterPos4xvOES(\{coords})");
		try {
		    functions.glRasterPos4xvOES().invokeExact(coords);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glRasterSamplesEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glRasterSamplesEXT.xhtml">registry.khronos.org</a>
	 */
	public void glRasterSamplesEXT(int samples, boolean fixedsamplelocations) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glRasterSamplesEXT(\{samples}, \{fixedsamplelocations})");
		try {
		    functions.glRasterSamplesEXT().invokeExact(samples, fixedsamplelocations);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glReadBuffer">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glReadBuffer.xhtml">registry.khronos.org</a>
	 */
	public void glReadBuffer(GLReadBufferMode src) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glReadBuffer(\{src})");
		try {
		    functions.glReadBuffer().invokeExact(src);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were src=\{src}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glReadBufferIndexedEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glReadBufferIndexedEXT.xhtml">registry.khronos.org</a>
	 */
	public void glReadBufferIndexedEXT(GLReadBufferMode src, int index) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glReadBufferIndexedEXT(\{src}, \{index})");
		try {
		    functions.glReadBufferIndexedEXT().invokeExact(src, index);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were src=\{src}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glReadBufferNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glReadBufferNV.xhtml">registry.khronos.org</a>
	 */
	public void glReadBufferNV(GLEnum mode) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glReadBufferNV(\{mode})");
		try {
		    functions.glReadBufferNV().invokeExact(mode);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glReadInstrumentsSGIX">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glReadInstrumentsSGIX.xhtml">registry.khronos.org</a>
	 */
	public void glReadInstrumentsSGIX(int marker) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glReadInstrumentsSGIX(\{marker})");
		try {
		    functions.glReadInstrumentsSGIX().invokeExact(marker);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glReadPixels">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glReadPixels.xhtml">registry.khronos.org</a>
	 */
	public void glReadPixels(int x, int y, int width, int height, GLPixelFormat format, GLPixelType type, MutPointer<Void> pixels) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glReadPixels(\{x}, \{y}, \{width}, \{height}, \{format}, \{type}, \{pixels})");
		try {
		    functions.glReadPixels().invokeExact(x, y, width, height, format, type, pixels);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were format=\{format}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glReadnPixels">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glReadnPixels.xhtml">registry.khronos.org</a>
	 */
	public void glReadnPixels(int x, int y, int width, int height, GLPixelFormat format, GLPixelType type, int bufSize, MutPointer<Void> data) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glReadnPixels(\{x}, \{y}, \{width}, \{height}, \{format}, \{type}, \{bufSize}, \{data})");
		try {
		    functions.glReadnPixels().invokeExact(x, y, width, height, format, type, bufSize, data);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were format=\{format}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glReadnPixelsARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glReadnPixelsARB.xhtml">registry.khronos.org</a>
	 */
	public void glReadnPixelsARB(int x, int y, int width, int height, GLPixelFormat format, GLPixelType type, int bufSize, MutPointer<Void> data) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glReadnPixelsARB(\{x}, \{y}, \{width}, \{height}, \{format}, \{type}, \{bufSize}, \{data})");
		try {
		    functions.glReadnPixelsARB().invokeExact(x, y, width, height, format, type, bufSize, data);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were format=\{format}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glReadnPixelsEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glReadnPixelsEXT.xhtml">registry.khronos.org</a>
	 */
	public void glReadnPixelsEXT(int x, int y, int width, int height, GLPixelFormat format, GLPixelType type, int bufSize, MutPointer<Void> data) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glReadnPixelsEXT(\{x}, \{y}, \{width}, \{height}, \{format}, \{type}, \{bufSize}, \{data})");
		try {
		    functions.glReadnPixelsEXT().invokeExact(x, y, width, height, format, type, bufSize, data);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were format=\{format}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glReadnPixelsKHR">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glReadnPixelsKHR.xhtml">registry.khronos.org</a>
	 */
	public void glReadnPixelsKHR(int x, int y, int width, int height, GLPixelFormat format, GLPixelType type, int bufSize, MutPointer<Void> data) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glReadnPixelsKHR(\{x}, \{y}, \{width}, \{height}, \{format}, \{type}, \{bufSize}, \{data})");
		try {
		    functions.glReadnPixelsKHR().invokeExact(x, y, width, height, format, type, bufSize, data);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were format=\{format}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glReleaseKeyedMutexWin32EXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glReleaseKeyedMutexWin32EXT.xhtml">registry.khronos.org</a>
	 */
	public boolean glReleaseKeyedMutexWin32EXT(int memory, long key) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glReleaseKeyedMutexWin32EXT(\{memory}, \{key})");
		try {
		    var r = (boolean)functions.glReleaseKeyedMutexWin32EXT().invokeExact(memory, key);
			checkError();
			return r;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glRectd">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glRectd.xhtml">registry.khronos.org</a>
	 */
	public void glRectd(double x1, double y1, double x2, double y2) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glRectd(\{x1}, \{y1}, \{x2}, \{y2})");
		try {
		    functions.glRectd().invokeExact(x1, y1, x2, y2);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glRectdv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glRectdv.xhtml">registry.khronos.org</a>
	 */
	public void glRectdv(ConstPointer<Double> v1, ConstPointer<Double> v2) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glRectdv(\{v1}, \{v2})");
		try {
		    functions.glRectdv().invokeExact(v1, v2);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glRectf">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glRectf.xhtml">registry.khronos.org</a>
	 */
	public void glRectf(float x1, float y1, float x2, float y2) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glRectf(\{x1}, \{y1}, \{x2}, \{y2})");
		try {
		    functions.glRectf().invokeExact(x1, y1, x2, y2);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glRectfv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glRectfv.xhtml">registry.khronos.org</a>
	 */
	public void glRectfv(ConstPointer<Float> v1, ConstPointer<Float> v2) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glRectfv(\{v1}, \{v2})");
		try {
		    functions.glRectfv().invokeExact(v1, v2);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glRecti">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glRecti.xhtml">registry.khronos.org</a>
	 */
	public void glRecti(int x1, int y1, int x2, int y2) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glRecti(\{x1}, \{y1}, \{x2}, \{y2})");
		try {
		    functions.glRecti().invokeExact(x1, y1, x2, y2);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glRectiv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glRectiv.xhtml">registry.khronos.org</a>
	 */
	public void glRectiv(ConstPointer<Integer> v1, ConstPointer<Integer> v2) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glRectiv(\{v1}, \{v2})");
		try {
		    functions.glRectiv().invokeExact(v1, v2);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glRects">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glRects.xhtml">registry.khronos.org</a>
	 */
	public void glRects(short x1, short y1, short x2, short y2) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glRects(\{x1}, \{y1}, \{x2}, \{y2})");
		try {
		    functions.glRects().invokeExact(x1, y1, x2, y2);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glRectsv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glRectsv.xhtml">registry.khronos.org</a>
	 */
	public void glRectsv(ConstPointer<Short> v1, ConstPointer<Short> v2) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glRectsv(\{v1}, \{v2})");
		try {
		    functions.glRectsv().invokeExact(v1, v2);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glRectxOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glRectxOES.xhtml">registry.khronos.org</a>
	 */
	public void glRectxOES(int x1, int y1, int x2, int y2) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glRectxOES(\{x1}, \{y1}, \{x2}, \{y2})");
		try {
		    functions.glRectxOES().invokeExact(x1, y1, x2, y2);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glRectxvOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glRectxvOES.xhtml">registry.khronos.org</a>
	 */
	public void glRectxvOES(ConstPointer<Integer> v1, ConstPointer<Integer> v2) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glRectxvOES(\{v1}, \{v2})");
		try {
		    functions.glRectxvOES().invokeExact(v1, v2);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glReferencePlaneSGIX">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glReferencePlaneSGIX.xhtml">registry.khronos.org</a>
	 */
	public void glReferencePlaneSGIX(ConstPointer<Double> equation) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glReferencePlaneSGIX(\{equation})");
		try {
		    functions.glReferencePlaneSGIX().invokeExact(equation);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glReleaseShaderCompiler">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glReleaseShaderCompiler.xhtml">registry.khronos.org</a>
	 */
	public void glReleaseShaderCompiler() {
	    if (LOG_COMMANDS) logCmd(() -> STR."glReleaseShaderCompiler()");
		try {
		    functions.glReleaseShaderCompiler().invokeExact();
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glRenderGpuMaskNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glRenderGpuMaskNV.xhtml">registry.khronos.org</a>
	 */
	public void glRenderGpuMaskNV(int mask) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glRenderGpuMaskNV(\{mask})");
		try {
		    functions.glRenderGpuMaskNV().invokeExact(mask);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glRenderMode">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glRenderMode.xhtml">registry.khronos.org</a>
	 */
	public int glRenderMode(GLRenderingMode mode) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glRenderMode(\{mode})");
		try {
		    var r = (int)functions.glRenderMode().invokeExact(mode);
			checkError();
			return r;
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}.");
			}
			throw e;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glRenderbufferStorage">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glRenderbufferStorage.xhtml">registry.khronos.org</a>
	 */
	public void glRenderbufferStorage(GLRenderbufferTarget target, GLInternalFormat internalformat, int width, int height) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glRenderbufferStorage(\{target}, \{internalformat}, \{width}, \{height})");
		try {
		    functions.glRenderbufferStorage().invokeExact(target, internalformat, width, height);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, internalformat=\{internalformat}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glRenderbufferStorageEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glRenderbufferStorageEXT.xhtml">registry.khronos.org</a>
	 */
	public void glRenderbufferStorageEXT(GLRenderbufferTarget target, GLInternalFormat internalformat, int width, int height) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glRenderbufferStorageEXT(\{target}, \{internalformat}, \{width}, \{height})");
		try {
		    functions.glRenderbufferStorageEXT().invokeExact(target, internalformat, width, height);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, internalformat=\{internalformat}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glRenderbufferStorageMultisample">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glRenderbufferStorageMultisample.xhtml">registry.khronos.org</a>
	 */
	public void glRenderbufferStorageMultisample(GLRenderbufferTarget target, int samples, GLInternalFormat internalformat, int width, int height) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glRenderbufferStorageMultisample(\{target}, \{samples}, \{internalformat}, \{width}, \{height})");
		try {
		    functions.glRenderbufferStorageMultisample().invokeExact(target, samples, internalformat, width, height);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, internalformat=\{internalformat}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glRenderbufferStorageMultisampleANGLE">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glRenderbufferStorageMultisampleANGLE.xhtml">registry.khronos.org</a>
	 */
	public void glRenderbufferStorageMultisampleANGLE(GLRenderbufferTarget target, int samples, GLInternalFormat internalformat, int width, int height) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glRenderbufferStorageMultisampleANGLE(\{target}, \{samples}, \{internalformat}, \{width}, \{height})");
		try {
		    functions.glRenderbufferStorageMultisampleANGLE().invokeExact(target, samples, internalformat, width, height);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, internalformat=\{internalformat}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glRenderbufferStorageMultisampleAPPLE">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glRenderbufferStorageMultisampleAPPLE.xhtml">registry.khronos.org</a>
	 */
	public void glRenderbufferStorageMultisampleAPPLE(GLRenderbufferTarget target, int samples, GLInternalFormat internalformat, int width, int height) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glRenderbufferStorageMultisampleAPPLE(\{target}, \{samples}, \{internalformat}, \{width}, \{height})");
		try {
		    functions.glRenderbufferStorageMultisampleAPPLE().invokeExact(target, samples, internalformat, width, height);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, internalformat=\{internalformat}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glRenderbufferStorageMultisampleAdvancedAMD">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glRenderbufferStorageMultisampleAdvancedAMD.xhtml">registry.khronos.org</a>
	 */
	public void glRenderbufferStorageMultisampleAdvancedAMD(GLRenderbufferTarget target, int samples, int storageSamples, GLInternalFormat internalformat, int width, int height) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glRenderbufferStorageMultisampleAdvancedAMD(\{target}, \{samples}, \{storageSamples}, \{internalformat}, \{width}, \{height})");
		try {
		    functions.glRenderbufferStorageMultisampleAdvancedAMD().invokeExact(target, samples, storageSamples, internalformat, width, height);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, internalformat=\{internalformat}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glRenderbufferStorageMultisampleCoverageNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glRenderbufferStorageMultisampleCoverageNV.xhtml">registry.khronos.org</a>
	 */
	public void glRenderbufferStorageMultisampleCoverageNV(GLRenderbufferTarget target, int coverageSamples, int colorSamples, GLInternalFormat internalformat, int width, int height) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glRenderbufferStorageMultisampleCoverageNV(\{target}, \{coverageSamples}, \{colorSamples}, \{internalformat}, \{width}, \{height})");
		try {
		    functions.glRenderbufferStorageMultisampleCoverageNV().invokeExact(target, coverageSamples, colorSamples, internalformat, width, height);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, internalformat=\{internalformat}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glRenderbufferStorageMultisampleEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glRenderbufferStorageMultisampleEXT.xhtml">registry.khronos.org</a>
	 */
	public void glRenderbufferStorageMultisampleEXT(GLRenderbufferTarget target, int samples, GLInternalFormat internalformat, int width, int height) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glRenderbufferStorageMultisampleEXT(\{target}, \{samples}, \{internalformat}, \{width}, \{height})");
		try {
		    functions.glRenderbufferStorageMultisampleEXT().invokeExact(target, samples, internalformat, width, height);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, internalformat=\{internalformat}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glRenderbufferStorageMultisampleIMG">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glRenderbufferStorageMultisampleIMG.xhtml">registry.khronos.org</a>
	 */
	public void glRenderbufferStorageMultisampleIMG(GLRenderbufferTarget target, int samples, GLInternalFormat internalformat, int width, int height) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glRenderbufferStorageMultisampleIMG(\{target}, \{samples}, \{internalformat}, \{width}, \{height})");
		try {
		    functions.glRenderbufferStorageMultisampleIMG().invokeExact(target, samples, internalformat, width, height);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, internalformat=\{internalformat}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glRenderbufferStorageMultisampleNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glRenderbufferStorageMultisampleNV.xhtml">registry.khronos.org</a>
	 */
	public void glRenderbufferStorageMultisampleNV(GLRenderbufferTarget target, int samples, GLInternalFormat internalformat, int width, int height) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glRenderbufferStorageMultisampleNV(\{target}, \{samples}, \{internalformat}, \{width}, \{height})");
		try {
		    functions.glRenderbufferStorageMultisampleNV().invokeExact(target, samples, internalformat, width, height);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, internalformat=\{internalformat}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glRenderbufferStorageOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glRenderbufferStorageOES.xhtml">registry.khronos.org</a>
	 */
	public void glRenderbufferStorageOES(GLRenderbufferTarget target, GLInternalFormat internalformat, int width, int height) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glRenderbufferStorageOES(\{target}, \{internalformat}, \{width}, \{height})");
		try {
		    functions.glRenderbufferStorageOES().invokeExact(target, internalformat, width, height);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, internalformat=\{internalformat}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glReplacementCodePointerSUN">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glReplacementCodePointerSUN.xhtml">registry.khronos.org</a>
	 */
	public void glReplacementCodePointerSUN(GLReplacementCodeTypeSUN type, int stride, MutPointer<ConstPointer<Void>> pointer) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glReplacementCodePointerSUN(\{type}, \{stride}, \{pointer})");
		try {
		    functions.glReplacementCodePointerSUN().invokeExact(type, stride, pointer);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glReplacementCodeubSUN">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glReplacementCodeubSUN.xhtml">registry.khronos.org</a>
	 */
	public void glReplacementCodeubSUN(byte code) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glReplacementCodeubSUN(\{code})");
		try {
		    functions.glReplacementCodeubSUN().invokeExact(code);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glReplacementCodeubvSUN">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glReplacementCodeubvSUN.xhtml">registry.khronos.org</a>
	 */
	public void glReplacementCodeubvSUN(ConstPointer<Byte> code) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glReplacementCodeubvSUN(\{code})");
		try {
		    functions.glReplacementCodeubvSUN().invokeExact(code);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glReplacementCodeuiColor3fVertex3fSUN">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glReplacementCodeuiColor3fVertex3fSUN.xhtml">registry.khronos.org</a>
	 */
	public void glReplacementCodeuiColor3fVertex3fSUN(int rc, float r, float g, float b, float x, float y, float z) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glReplacementCodeuiColor3fVertex3fSUN(\{rc}, \{r}, \{g}, \{b}, \{x}, \{y}, \{z})");
		try {
		    functions.glReplacementCodeuiColor3fVertex3fSUN().invokeExact(rc, r, g, b, x, y, z);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glReplacementCodeuiColor3fVertex3fvSUN">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glReplacementCodeuiColor3fVertex3fvSUN.xhtml">registry.khronos.org</a>
	 */
	public void glReplacementCodeuiColor3fVertex3fvSUN(ConstPointer<Integer> rc, ConstPointer<Float> c, ConstPointer<Float> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glReplacementCodeuiColor3fVertex3fvSUN(\{rc}, \{c}, \{v})");
		try {
		    functions.glReplacementCodeuiColor3fVertex3fvSUN().invokeExact(rc, c, v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glReplacementCodeuiColor4fNormal3fVertex3fSUN">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glReplacementCodeuiColor4fNormal3fVertex3fSUN.xhtml">registry.khronos.org</a>
	 */
	public void glReplacementCodeuiColor4fNormal3fVertex3fSUN(int rc, float r, float g, float b, float a, float nx, float ny, float nz, float x, float y, float z) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glReplacementCodeuiColor4fNormal3fVertex3fSUN(\{rc}, \{r}, \{g}, \{b}, \{a}, \{nx}, \{ny}, \{nz}, \{x}, \{y}, \{z})");
		try {
		    functions.glReplacementCodeuiColor4fNormal3fVertex3fSUN().invokeExact(rc, r, g, b, a, nx, ny, nz, x, y, z);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glReplacementCodeuiColor4fNormal3fVertex3fvSUN">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glReplacementCodeuiColor4fNormal3fVertex3fvSUN.xhtml">registry.khronos.org</a>
	 */
	public void glReplacementCodeuiColor4fNormal3fVertex3fvSUN(ConstPointer<Integer> rc, ConstPointer<Float> c, ConstPointer<Float> n, ConstPointer<Float> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glReplacementCodeuiColor4fNormal3fVertex3fvSUN(\{rc}, \{c}, \{n}, \{v})");
		try {
		    functions.glReplacementCodeuiColor4fNormal3fVertex3fvSUN().invokeExact(rc, c, n, v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glReplacementCodeuiColor4ubVertex3fSUN">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glReplacementCodeuiColor4ubVertex3fSUN.xhtml">registry.khronos.org</a>
	 */
	public void glReplacementCodeuiColor4ubVertex3fSUN(int rc, byte r, byte g, byte b, byte a, float x, float y, float z) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glReplacementCodeuiColor4ubVertex3fSUN(\{rc}, \{r}, \{g}, \{b}, \{a}, \{x}, \{y}, \{z})");
		try {
		    functions.glReplacementCodeuiColor4ubVertex3fSUN().invokeExact(rc, r, g, b, a, x, y, z);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glReplacementCodeuiColor4ubVertex3fvSUN">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glReplacementCodeuiColor4ubVertex3fvSUN.xhtml">registry.khronos.org</a>
	 */
	public void glReplacementCodeuiColor4ubVertex3fvSUN(ConstPointer<Integer> rc, ConstPointer<Byte> c, ConstPointer<Float> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glReplacementCodeuiColor4ubVertex3fvSUN(\{rc}, \{c}, \{v})");
		try {
		    functions.glReplacementCodeuiColor4ubVertex3fvSUN().invokeExact(rc, c, v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glReplacementCodeuiNormal3fVertex3fSUN">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glReplacementCodeuiNormal3fVertex3fSUN.xhtml">registry.khronos.org</a>
	 */
	public void glReplacementCodeuiNormal3fVertex3fSUN(int rc, float nx, float ny, float nz, float x, float y, float z) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glReplacementCodeuiNormal3fVertex3fSUN(\{rc}, \{nx}, \{ny}, \{nz}, \{x}, \{y}, \{z})");
		try {
		    functions.glReplacementCodeuiNormal3fVertex3fSUN().invokeExact(rc, nx, ny, nz, x, y, z);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glReplacementCodeuiNormal3fVertex3fvSUN">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glReplacementCodeuiNormal3fVertex3fvSUN.xhtml">registry.khronos.org</a>
	 */
	public void glReplacementCodeuiNormal3fVertex3fvSUN(ConstPointer<Integer> rc, ConstPointer<Float> n, ConstPointer<Float> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glReplacementCodeuiNormal3fVertex3fvSUN(\{rc}, \{n}, \{v})");
		try {
		    functions.glReplacementCodeuiNormal3fVertex3fvSUN().invokeExact(rc, n, v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glReplacementCodeuiSUN">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glReplacementCodeuiSUN.xhtml">registry.khronos.org</a>
	 */
	public void glReplacementCodeuiSUN(int code) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glReplacementCodeuiSUN(\{code})");
		try {
		    functions.glReplacementCodeuiSUN().invokeExact(code);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glReplacementCodeuiTexCoord2fColor4fNormal3fVertex3fSUN">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glReplacementCodeuiTexCoord2fColor4fNormal3fVertex3fSUN.xhtml">registry.khronos.org</a>
	 */
	public void glReplacementCodeuiTexCoord2fColor4fNormal3fVertex3fSUN(int rc, float s, float t, float r, float g, float b, float a, float nx, float ny, float nz, float x, float y, float z) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glReplacementCodeuiTexCoord2fColor4fNormal3fVertex3fSUN(\{rc}, \{s}, \{t}, \{r}, \{g}, \{b}, \{a}, \{nx}, \{ny}, \{nz}, \{x}, \{y}, \{z})");
		try {
		    functions.glReplacementCodeuiTexCoord2fColor4fNormal3fVertex3fSUN().invokeExact(rc, s, t, r, g, b, a, nx, ny, nz, x, y, z);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glReplacementCodeuiTexCoord2fColor4fNormal3fVertex3fvSUN">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glReplacementCodeuiTexCoord2fColor4fNormal3fVertex3fvSUN.xhtml">registry.khronos.org</a>
	 */
	public void glReplacementCodeuiTexCoord2fColor4fNormal3fVertex3fvSUN(ConstPointer<Integer> rc, ConstPointer<Float> tc, ConstPointer<Float> c, ConstPointer<Float> n, ConstPointer<Float> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glReplacementCodeuiTexCoord2fColor4fNormal3fVertex3fvSUN(\{rc}, \{tc}, \{c}, \{n}, \{v})");
		try {
		    functions.glReplacementCodeuiTexCoord2fColor4fNormal3fVertex3fvSUN().invokeExact(rc, tc, c, n, v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glReplacementCodeuiTexCoord2fNormal3fVertex3fSUN">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glReplacementCodeuiTexCoord2fNormal3fVertex3fSUN.xhtml">registry.khronos.org</a>
	 */
	public void glReplacementCodeuiTexCoord2fNormal3fVertex3fSUN(int rc, float s, float t, float nx, float ny, float nz, float x, float y, float z) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glReplacementCodeuiTexCoord2fNormal3fVertex3fSUN(\{rc}, \{s}, \{t}, \{nx}, \{ny}, \{nz}, \{x}, \{y}, \{z})");
		try {
		    functions.glReplacementCodeuiTexCoord2fNormal3fVertex3fSUN().invokeExact(rc, s, t, nx, ny, nz, x, y, z);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glReplacementCodeuiTexCoord2fNormal3fVertex3fvSUN">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glReplacementCodeuiTexCoord2fNormal3fVertex3fvSUN.xhtml">registry.khronos.org</a>
	 */
	public void glReplacementCodeuiTexCoord2fNormal3fVertex3fvSUN(ConstPointer<Integer> rc, ConstPointer<Float> tc, ConstPointer<Float> n, ConstPointer<Float> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glReplacementCodeuiTexCoord2fNormal3fVertex3fvSUN(\{rc}, \{tc}, \{n}, \{v})");
		try {
		    functions.glReplacementCodeuiTexCoord2fNormal3fVertex3fvSUN().invokeExact(rc, tc, n, v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glReplacementCodeuiTexCoord2fVertex3fSUN">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glReplacementCodeuiTexCoord2fVertex3fSUN.xhtml">registry.khronos.org</a>
	 */
	public void glReplacementCodeuiTexCoord2fVertex3fSUN(int rc, float s, float t, float x, float y, float z) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glReplacementCodeuiTexCoord2fVertex3fSUN(\{rc}, \{s}, \{t}, \{x}, \{y}, \{z})");
		try {
		    functions.glReplacementCodeuiTexCoord2fVertex3fSUN().invokeExact(rc, s, t, x, y, z);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glReplacementCodeuiTexCoord2fVertex3fvSUN">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glReplacementCodeuiTexCoord2fVertex3fvSUN.xhtml">registry.khronos.org</a>
	 */
	public void glReplacementCodeuiTexCoord2fVertex3fvSUN(ConstPointer<Integer> rc, ConstPointer<Float> tc, ConstPointer<Float> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glReplacementCodeuiTexCoord2fVertex3fvSUN(\{rc}, \{tc}, \{v})");
		try {
		    functions.glReplacementCodeuiTexCoord2fVertex3fvSUN().invokeExact(rc, tc, v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glReplacementCodeuiVertex3fSUN">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glReplacementCodeuiVertex3fSUN.xhtml">registry.khronos.org</a>
	 */
	public void glReplacementCodeuiVertex3fSUN(int rc, float x, float y, float z) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glReplacementCodeuiVertex3fSUN(\{rc}, \{x}, \{y}, \{z})");
		try {
		    functions.glReplacementCodeuiVertex3fSUN().invokeExact(rc, x, y, z);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glReplacementCodeuiVertex3fvSUN">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glReplacementCodeuiVertex3fvSUN.xhtml">registry.khronos.org</a>
	 */
	public void glReplacementCodeuiVertex3fvSUN(ConstPointer<Integer> rc, ConstPointer<Float> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glReplacementCodeuiVertex3fvSUN(\{rc}, \{v})");
		try {
		    functions.glReplacementCodeuiVertex3fvSUN().invokeExact(rc, v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glReplacementCodeuivSUN">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glReplacementCodeuivSUN.xhtml">registry.khronos.org</a>
	 */
	public void glReplacementCodeuivSUN(ConstPointer<Integer> code) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glReplacementCodeuivSUN(\{code})");
		try {
		    functions.glReplacementCodeuivSUN().invokeExact(code);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glReplacementCodeusSUN">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glReplacementCodeusSUN.xhtml">registry.khronos.org</a>
	 */
	public void glReplacementCodeusSUN(char code) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glReplacementCodeusSUN(\{code})");
		try {
		    functions.glReplacementCodeusSUN().invokeExact(code);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glReplacementCodeusvSUN">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glReplacementCodeusvSUN.xhtml">registry.khronos.org</a>
	 */
	public void glReplacementCodeusvSUN(ConstPointer<Character> code) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glReplacementCodeusvSUN(\{code})");
		try {
		    functions.glReplacementCodeusvSUN().invokeExact(code);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glRequestResidentProgramsNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glRequestResidentProgramsNV.xhtml">registry.khronos.org</a>
	 */
	public void glRequestResidentProgramsNV(int n, ConstPointer<Integer> programs) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glRequestResidentProgramsNV(\{n}, \{programs})");
		try {
		    functions.glRequestResidentProgramsNV().invokeExact(n, programs);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glResetHistogram">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glResetHistogram.xhtml">registry.khronos.org</a>
	 */
	public void glResetHistogram(GLHistogramTargetEXT target) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glResetHistogram(\{target})");
		try {
		    functions.glResetHistogram().invokeExact(target);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glResetHistogramEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glResetHistogramEXT.xhtml">registry.khronos.org</a>
	 */
	public void glResetHistogramEXT(GLHistogramTargetEXT target) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glResetHistogramEXT(\{target})");
		try {
		    functions.glResetHistogramEXT().invokeExact(target);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glResetMemoryObjectParameterNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glResetMemoryObjectParameterNV.xhtml">registry.khronos.org</a>
	 */
	public void glResetMemoryObjectParameterNV(int memory, GLEnum pname) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glResetMemoryObjectParameterNV(\{memory}, \{pname})");
		try {
		    functions.glResetMemoryObjectParameterNV().invokeExact(memory, pname);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glResetMinmax">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glResetMinmax.xhtml">registry.khronos.org</a>
	 */
	public void glResetMinmax(GLMinmaxTargetEXT target) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glResetMinmax(\{target})");
		try {
		    functions.glResetMinmax().invokeExact(target);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glResetMinmaxEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glResetMinmaxEXT.xhtml">registry.khronos.org</a>
	 */
	public void glResetMinmaxEXT(GLMinmaxTargetEXT target) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glResetMinmaxEXT(\{target})");
		try {
		    functions.glResetMinmaxEXT().invokeExact(target);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glResizeBuffersMESA">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glResizeBuffersMESA.xhtml">registry.khronos.org</a>
	 */
	public void glResizeBuffersMESA() {
	    if (LOG_COMMANDS) logCmd(() -> STR."glResizeBuffersMESA()");
		try {
		    functions.glResizeBuffersMESA().invokeExact();
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glResolveDepthValuesNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glResolveDepthValuesNV.xhtml">registry.khronos.org</a>
	 */
	public void glResolveDepthValuesNV() {
	    if (LOG_COMMANDS) logCmd(() -> STR."glResolveDepthValuesNV()");
		try {
		    functions.glResolveDepthValuesNV().invokeExact();
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glResolveMultisampleFramebufferAPPLE">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glResolveMultisampleFramebufferAPPLE.xhtml">registry.khronos.org</a>
	 */
	public void glResolveMultisampleFramebufferAPPLE() {
	    if (LOG_COMMANDS) logCmd(() -> STR."glResolveMultisampleFramebufferAPPLE()");
		try {
		    functions.glResolveMultisampleFramebufferAPPLE().invokeExact();
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glResumeTransformFeedback">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glResumeTransformFeedback.xhtml">registry.khronos.org</a>
	 */
	public void glResumeTransformFeedback() {
	    if (LOG_COMMANDS) logCmd(() -> STR."glResumeTransformFeedback()");
		try {
		    functions.glResumeTransformFeedback().invokeExact();
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glResumeTransformFeedbackNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glResumeTransformFeedbackNV.xhtml">registry.khronos.org</a>
	 */
	public void glResumeTransformFeedbackNV() {
	    if (LOG_COMMANDS) logCmd(() -> STR."glResumeTransformFeedbackNV()");
		try {
		    functions.glResumeTransformFeedbackNV().invokeExact();
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glRotated">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glRotated.xhtml">registry.khronos.org</a>
	 */
	public void glRotated(double angle, double x, double y, double z) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glRotated(\{angle}, \{x}, \{y}, \{z})");
		try {
		    functions.glRotated().invokeExact(angle, x, y, z);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glRotatef">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glRotatef.xhtml">registry.khronos.org</a>
	 */
	public void glRotatef(float angle, float x, float y, float z) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glRotatef(\{angle}, \{x}, \{y}, \{z})");
		try {
		    functions.glRotatef().invokeExact(angle, x, y, z);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glRotatex">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glRotatex.xhtml">registry.khronos.org</a>
	 */
	public void glRotatex(int angle, int x, int y, int z) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glRotatex(\{angle}, \{x}, \{y}, \{z})");
		try {
		    functions.glRotatex().invokeExact(angle, x, y, z);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glRotatexOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glRotatexOES.xhtml">registry.khronos.org</a>
	 */
	public void glRotatexOES(int angle, int x, int y, int z) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glRotatexOES(\{angle}, \{x}, \{y}, \{z})");
		try {
		    functions.glRotatexOES().invokeExact(angle, x, y, z);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glSampleCoverage">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glSampleCoverage.xhtml">registry.khronos.org</a>
	 */
	public void glSampleCoverage(float value, boolean invert) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glSampleCoverage(\{value}, \{invert})");
		try {
		    functions.glSampleCoverage().invokeExact(value, invert);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glSampleCoverageARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glSampleCoverageARB.xhtml">registry.khronos.org</a>
	 */
	public void glSampleCoverageARB(float value, boolean invert) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glSampleCoverageARB(\{value}, \{invert})");
		try {
		    functions.glSampleCoverageARB().invokeExact(value, invert);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glSampleCoveragex">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glSampleCoveragex.xhtml">registry.khronos.org</a>
	 */
	public void glSampleCoveragex(int value, boolean invert) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glSampleCoveragex(\{value}, \{invert})");
		try {
		    functions.glSampleCoveragex().invokeExact(value, invert);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glSampleCoveragexOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glSampleCoveragexOES.xhtml">registry.khronos.org</a>
	 */
	public void glSampleCoveragexOES(int value, boolean invert) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glSampleCoveragexOES(\{value}, \{invert})");
		try {
		    functions.glSampleCoveragexOES().invokeExact(value, invert);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glSampleMapATI">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glSampleMapATI.xhtml">registry.khronos.org</a>
	 */
	public void glSampleMapATI(int dst, int interp, GLSwizzleOpATI swizzle) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glSampleMapATI(\{dst}, \{interp}, \{swizzle})");
		try {
		    functions.glSampleMapATI().invokeExact(dst, interp, swizzle);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were swizzle=\{swizzle}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glSampleMaskEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glSampleMaskEXT.xhtml">registry.khronos.org</a>
	 */
	public void glSampleMaskEXT(float value, boolean invert) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glSampleMaskEXT(\{value}, \{invert})");
		try {
		    functions.glSampleMaskEXT().invokeExact(value, invert);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glSampleMaskIndexedNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glSampleMaskIndexedNV.xhtml">registry.khronos.org</a>
	 */
	public void glSampleMaskIndexedNV(int index, int mask) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glSampleMaskIndexedNV(\{index}, \{mask})");
		try {
		    functions.glSampleMaskIndexedNV().invokeExact(index, mask);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glSampleMaskSGIS">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glSampleMaskSGIS.xhtml">registry.khronos.org</a>
	 */
	public void glSampleMaskSGIS(float value, boolean invert) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glSampleMaskSGIS(\{value}, \{invert})");
		try {
		    functions.glSampleMaskSGIS().invokeExact(value, invert);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glSampleMaski">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glSampleMaski.xhtml">registry.khronos.org</a>
	 */
	public void glSampleMaski(int maskNumber, int mask) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glSampleMaski(\{maskNumber}, \{mask})");
		try {
		    functions.glSampleMaski().invokeExact(maskNumber, mask);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glSamplePatternEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glSamplePatternEXT.xhtml">registry.khronos.org</a>
	 */
	public void glSamplePatternEXT(GLSamplePatternEXT pattern) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glSamplePatternEXT(\{pattern})");
		try {
		    functions.glSamplePatternEXT().invokeExact(pattern);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pattern=\{pattern}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glSamplePatternSGIS">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glSamplePatternSGIS.xhtml">registry.khronos.org</a>
	 */
	public void glSamplePatternSGIS(GLSamplePatternSGIS pattern) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glSamplePatternSGIS(\{pattern})");
		try {
		    functions.glSamplePatternSGIS().invokeExact(pattern);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pattern=\{pattern}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glSamplerParameterIiv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glSamplerParameterIiv.xhtml">registry.khronos.org</a>
	 */
	public void glSamplerParameterIiv(int sampler, GLSamplerParameterI pname, ConstPointer<Integer> param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glSamplerParameterIiv(\{sampler}, \{pname}, \{param})");
		try {
		    functions.glSamplerParameterIiv().invokeExact(sampler, pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glSamplerParameterIivEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glSamplerParameterIivEXT.xhtml">registry.khronos.org</a>
	 */
	public void glSamplerParameterIivEXT(int sampler, GLSamplerParameterI pname, ConstPointer<Integer> param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glSamplerParameterIivEXT(\{sampler}, \{pname}, \{param})");
		try {
		    functions.glSamplerParameterIivEXT().invokeExact(sampler, pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glSamplerParameterIivOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glSamplerParameterIivOES.xhtml">registry.khronos.org</a>
	 */
	public void glSamplerParameterIivOES(int sampler, GLSamplerParameterI pname, ConstPointer<Integer> param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glSamplerParameterIivOES(\{sampler}, \{pname}, \{param})");
		try {
		    functions.glSamplerParameterIivOES().invokeExact(sampler, pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glSamplerParameterIuiv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glSamplerParameterIuiv.xhtml">registry.khronos.org</a>
	 */
	public void glSamplerParameterIuiv(int sampler, GLSamplerParameterI pname, ConstPointer<Integer> param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glSamplerParameterIuiv(\{sampler}, \{pname}, \{param})");
		try {
		    functions.glSamplerParameterIuiv().invokeExact(sampler, pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glSamplerParameterIuivEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glSamplerParameterIuivEXT.xhtml">registry.khronos.org</a>
	 */
	public void glSamplerParameterIuivEXT(int sampler, GLSamplerParameterI pname, ConstPointer<Integer> param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glSamplerParameterIuivEXT(\{sampler}, \{pname}, \{param})");
		try {
		    functions.glSamplerParameterIuivEXT().invokeExact(sampler, pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glSamplerParameterIuivOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glSamplerParameterIuivOES.xhtml">registry.khronos.org</a>
	 */
	public void glSamplerParameterIuivOES(int sampler, GLSamplerParameterI pname, ConstPointer<Integer> param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glSamplerParameterIuivOES(\{sampler}, \{pname}, \{param})");
		try {
		    functions.glSamplerParameterIuivOES().invokeExact(sampler, pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glSamplerParameterf">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glSamplerParameterf.xhtml">registry.khronos.org</a>
	 */
	public void glSamplerParameterf(int sampler, GLSamplerParameterF pname, float param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glSamplerParameterf(\{sampler}, \{pname}, \{param})");
		try {
		    functions.glSamplerParameterf().invokeExact(sampler, pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glSamplerParameterfv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glSamplerParameterfv.xhtml">registry.khronos.org</a>
	 */
	public void glSamplerParameterfv(int sampler, GLSamplerParameterF pname, ConstPointer<Float> param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glSamplerParameterfv(\{sampler}, \{pname}, \{param})");
		try {
		    functions.glSamplerParameterfv().invokeExact(sampler, pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glSamplerParameteri">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glSamplerParameteri.xhtml">registry.khronos.org</a>
	 */
	public void glSamplerParameteri(int sampler, GLSamplerParameterI pname, int param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glSamplerParameteri(\{sampler}, \{pname}, \{param})");
		try {
		    functions.glSamplerParameteri().invokeExact(sampler, pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glSamplerParameteriv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glSamplerParameteriv.xhtml">registry.khronos.org</a>
	 */
	public void glSamplerParameteriv(int sampler, GLSamplerParameterI pname, ConstPointer<Integer> param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glSamplerParameteriv(\{sampler}, \{pname}, \{param})");
		try {
		    functions.glSamplerParameteriv().invokeExact(sampler, pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glScaled">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glScaled.xhtml">registry.khronos.org</a>
	 */
	public void glScaled(double x, double y, double z) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glScaled(\{x}, \{y}, \{z})");
		try {
		    functions.glScaled().invokeExact(x, y, z);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glScalef">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glScalef.xhtml">registry.khronos.org</a>
	 */
	public void glScalef(float x, float y, float z) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glScalef(\{x}, \{y}, \{z})");
		try {
		    functions.glScalef().invokeExact(x, y, z);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glScalex">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glScalex.xhtml">registry.khronos.org</a>
	 */
	public void glScalex(int x, int y, int z) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glScalex(\{x}, \{y}, \{z})");
		try {
		    functions.glScalex().invokeExact(x, y, z);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glScalexOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glScalexOES.xhtml">registry.khronos.org</a>
	 */
	public void glScalexOES(int x, int y, int z) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glScalexOES(\{x}, \{y}, \{z})");
		try {
		    functions.glScalexOES().invokeExact(x, y, z);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glScissor">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glScissor.xhtml">registry.khronos.org</a>
	 */
	public void glScissor(int x, int y, int width, int height) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glScissor(\{x}, \{y}, \{width}, \{height})");
		try {
		    functions.glScissor().invokeExact(x, y, width, height);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glScissorArrayv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glScissorArrayv.xhtml">registry.khronos.org</a>
	 */
	public void glScissorArrayv(int first, int count, ConstPointer<Integer> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glScissorArrayv(\{first}, \{count}, \{v})");
		try {
		    functions.glScissorArrayv().invokeExact(first, count, v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glScissorArrayvNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glScissorArrayvNV.xhtml">registry.khronos.org</a>
	 */
	public void glScissorArrayvNV(int first, int count, ConstPointer<Integer> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glScissorArrayvNV(\{first}, \{count}, \{v})");
		try {
		    functions.glScissorArrayvNV().invokeExact(first, count, v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glScissorArrayvOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glScissorArrayvOES.xhtml">registry.khronos.org</a>
	 */
	public void glScissorArrayvOES(int first, int count, ConstPointer<Integer> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glScissorArrayvOES(\{first}, \{count}, \{v})");
		try {
		    functions.glScissorArrayvOES().invokeExact(first, count, v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glScissorExclusiveArrayvNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glScissorExclusiveArrayvNV.xhtml">registry.khronos.org</a>
	 */
	public void glScissorExclusiveArrayvNV(int first, int count, ConstPointer<Integer> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glScissorExclusiveArrayvNV(\{first}, \{count}, \{v})");
		try {
		    functions.glScissorExclusiveArrayvNV().invokeExact(first, count, v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glScissorExclusiveNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glScissorExclusiveNV.xhtml">registry.khronos.org</a>
	 */
	public void glScissorExclusiveNV(int x, int y, int width, int height) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glScissorExclusiveNV(\{x}, \{y}, \{width}, \{height})");
		try {
		    functions.glScissorExclusiveNV().invokeExact(x, y, width, height);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glScissorIndexed">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glScissorIndexed.xhtml">registry.khronos.org</a>
	 */
	public void glScissorIndexed(int index, int left, int bottom, int width, int height) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glScissorIndexed(\{index}, \{left}, \{bottom}, \{width}, \{height})");
		try {
		    functions.glScissorIndexed().invokeExact(index, left, bottom, width, height);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glScissorIndexedNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glScissorIndexedNV.xhtml">registry.khronos.org</a>
	 */
	public void glScissorIndexedNV(int index, int left, int bottom, int width, int height) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glScissorIndexedNV(\{index}, \{left}, \{bottom}, \{width}, \{height})");
		try {
		    functions.glScissorIndexedNV().invokeExact(index, left, bottom, width, height);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glScissorIndexedOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glScissorIndexedOES.xhtml">registry.khronos.org</a>
	 */
	public void glScissorIndexedOES(int index, int left, int bottom, int width, int height) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glScissorIndexedOES(\{index}, \{left}, \{bottom}, \{width}, \{height})");
		try {
		    functions.glScissorIndexedOES().invokeExact(index, left, bottom, width, height);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glScissorIndexedv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glScissorIndexedv.xhtml">registry.khronos.org</a>
	 */
	public void glScissorIndexedv(int index, ConstPointer<Integer> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glScissorIndexedv(\{index}, \{v})");
		try {
		    functions.glScissorIndexedv().invokeExact(index, v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glScissorIndexedvNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glScissorIndexedvNV.xhtml">registry.khronos.org</a>
	 */
	public void glScissorIndexedvNV(int index, ConstPointer<Integer> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glScissorIndexedvNV(\{index}, \{v})");
		try {
		    functions.glScissorIndexedvNV().invokeExact(index, v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glScissorIndexedvOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glScissorIndexedvOES.xhtml">registry.khronos.org</a>
	 */
	public void glScissorIndexedvOES(int index, ConstPointer<Integer> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glScissorIndexedvOES(\{index}, \{v})");
		try {
		    functions.glScissorIndexedvOES().invokeExact(index, v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glSecondaryColor3b">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glSecondaryColor3b.xhtml">registry.khronos.org</a>
	 */
	public void glSecondaryColor3b(byte red, byte green, byte blue) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glSecondaryColor3b(\{red}, \{green}, \{blue})");
		try {
		    functions.glSecondaryColor3b().invokeExact(red, green, blue);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glSecondaryColor3bEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glSecondaryColor3bEXT.xhtml">registry.khronos.org</a>
	 */
	public void glSecondaryColor3bEXT(byte red, byte green, byte blue) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glSecondaryColor3bEXT(\{red}, \{green}, \{blue})");
		try {
		    functions.glSecondaryColor3bEXT().invokeExact(red, green, blue);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glSecondaryColor3bv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glSecondaryColor3bv.xhtml">registry.khronos.org</a>
	 */
	public void glSecondaryColor3bv(ConstPointer<Byte> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glSecondaryColor3bv(\{v})");
		try {
		    functions.glSecondaryColor3bv().invokeExact(v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glSecondaryColor3bvEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glSecondaryColor3bvEXT.xhtml">registry.khronos.org</a>
	 */
	public void glSecondaryColor3bvEXT(ConstPointer<Byte> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glSecondaryColor3bvEXT(\{v})");
		try {
		    functions.glSecondaryColor3bvEXT().invokeExact(v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glSecondaryColor3d">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glSecondaryColor3d.xhtml">registry.khronos.org</a>
	 */
	public void glSecondaryColor3d(double red, double green, double blue) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glSecondaryColor3d(\{red}, \{green}, \{blue})");
		try {
		    functions.glSecondaryColor3d().invokeExact(red, green, blue);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glSecondaryColor3dEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glSecondaryColor3dEXT.xhtml">registry.khronos.org</a>
	 */
	public void glSecondaryColor3dEXT(double red, double green, double blue) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glSecondaryColor3dEXT(\{red}, \{green}, \{blue})");
		try {
		    functions.glSecondaryColor3dEXT().invokeExact(red, green, blue);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glSecondaryColor3dv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glSecondaryColor3dv.xhtml">registry.khronos.org</a>
	 */
	public void glSecondaryColor3dv(ConstPointer<Double> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glSecondaryColor3dv(\{v})");
		try {
		    functions.glSecondaryColor3dv().invokeExact(v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glSecondaryColor3dvEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glSecondaryColor3dvEXT.xhtml">registry.khronos.org</a>
	 */
	public void glSecondaryColor3dvEXT(ConstPointer<Double> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glSecondaryColor3dvEXT(\{v})");
		try {
		    functions.glSecondaryColor3dvEXT().invokeExact(v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glSecondaryColor3f">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glSecondaryColor3f.xhtml">registry.khronos.org</a>
	 */
	public void glSecondaryColor3f(float red, float green, float blue) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glSecondaryColor3f(\{red}, \{green}, \{blue})");
		try {
		    functions.glSecondaryColor3f().invokeExact(red, green, blue);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glSecondaryColor3fEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glSecondaryColor3fEXT.xhtml">registry.khronos.org</a>
	 */
	public void glSecondaryColor3fEXT(float red, float green, float blue) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glSecondaryColor3fEXT(\{red}, \{green}, \{blue})");
		try {
		    functions.glSecondaryColor3fEXT().invokeExact(red, green, blue);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glSecondaryColor3fv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glSecondaryColor3fv.xhtml">registry.khronos.org</a>
	 */
	public void glSecondaryColor3fv(ConstPointer<Float> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glSecondaryColor3fv(\{v})");
		try {
		    functions.glSecondaryColor3fv().invokeExact(v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glSecondaryColor3fvEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glSecondaryColor3fvEXT.xhtml">registry.khronos.org</a>
	 */
	public void glSecondaryColor3fvEXT(ConstPointer<Float> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glSecondaryColor3fvEXT(\{v})");
		try {
		    functions.glSecondaryColor3fvEXT().invokeExact(v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glSecondaryColor3hNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glSecondaryColor3hNV.xhtml">registry.khronos.org</a>
	 */
	public void glSecondaryColor3hNV(char red, char green, char blue) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glSecondaryColor3hNV(\{red}, \{green}, \{blue})");
		try {
		    functions.glSecondaryColor3hNV().invokeExact(red, green, blue);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glSecondaryColor3hvNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glSecondaryColor3hvNV.xhtml">registry.khronos.org</a>
	 */
	public void glSecondaryColor3hvNV(ConstPointer<Character> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glSecondaryColor3hvNV(\{v})");
		try {
		    functions.glSecondaryColor3hvNV().invokeExact(v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glSecondaryColor3i">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glSecondaryColor3i.xhtml">registry.khronos.org</a>
	 */
	public void glSecondaryColor3i(int red, int green, int blue) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glSecondaryColor3i(\{red}, \{green}, \{blue})");
		try {
		    functions.glSecondaryColor3i().invokeExact(red, green, blue);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glSecondaryColor3iEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glSecondaryColor3iEXT.xhtml">registry.khronos.org</a>
	 */
	public void glSecondaryColor3iEXT(int red, int green, int blue) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glSecondaryColor3iEXT(\{red}, \{green}, \{blue})");
		try {
		    functions.glSecondaryColor3iEXT().invokeExact(red, green, blue);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glSecondaryColor3iv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glSecondaryColor3iv.xhtml">registry.khronos.org</a>
	 */
	public void glSecondaryColor3iv(ConstPointer<Integer> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glSecondaryColor3iv(\{v})");
		try {
		    functions.glSecondaryColor3iv().invokeExact(v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glSecondaryColor3ivEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glSecondaryColor3ivEXT.xhtml">registry.khronos.org</a>
	 */
	public void glSecondaryColor3ivEXT(ConstPointer<Integer> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glSecondaryColor3ivEXT(\{v})");
		try {
		    functions.glSecondaryColor3ivEXT().invokeExact(v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glSecondaryColor3s">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glSecondaryColor3s.xhtml">registry.khronos.org</a>
	 */
	public void glSecondaryColor3s(short red, short green, short blue) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glSecondaryColor3s(\{red}, \{green}, \{blue})");
		try {
		    functions.glSecondaryColor3s().invokeExact(red, green, blue);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glSecondaryColor3sEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glSecondaryColor3sEXT.xhtml">registry.khronos.org</a>
	 */
	public void glSecondaryColor3sEXT(short red, short green, short blue) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glSecondaryColor3sEXT(\{red}, \{green}, \{blue})");
		try {
		    functions.glSecondaryColor3sEXT().invokeExact(red, green, blue);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glSecondaryColor3sv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glSecondaryColor3sv.xhtml">registry.khronos.org</a>
	 */
	public void glSecondaryColor3sv(ConstPointer<Short> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glSecondaryColor3sv(\{v})");
		try {
		    functions.glSecondaryColor3sv().invokeExact(v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glSecondaryColor3svEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glSecondaryColor3svEXT.xhtml">registry.khronos.org</a>
	 */
	public void glSecondaryColor3svEXT(ConstPointer<Short> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glSecondaryColor3svEXT(\{v})");
		try {
		    functions.glSecondaryColor3svEXT().invokeExact(v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glSecondaryColor3ub">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glSecondaryColor3ub.xhtml">registry.khronos.org</a>
	 */
	public void glSecondaryColor3ub(byte red, byte green, byte blue) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glSecondaryColor3ub(\{red}, \{green}, \{blue})");
		try {
		    functions.glSecondaryColor3ub().invokeExact(red, green, blue);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glSecondaryColor3ubEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glSecondaryColor3ubEXT.xhtml">registry.khronos.org</a>
	 */
	public void glSecondaryColor3ubEXT(byte red, byte green, byte blue) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glSecondaryColor3ubEXT(\{red}, \{green}, \{blue})");
		try {
		    functions.glSecondaryColor3ubEXT().invokeExact(red, green, blue);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glSecondaryColor3ubv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glSecondaryColor3ubv.xhtml">registry.khronos.org</a>
	 */
	public void glSecondaryColor3ubv(ConstPointer<Byte> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glSecondaryColor3ubv(\{v})");
		try {
		    functions.glSecondaryColor3ubv().invokeExact(v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glSecondaryColor3ubvEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glSecondaryColor3ubvEXT.xhtml">registry.khronos.org</a>
	 */
	public void glSecondaryColor3ubvEXT(ConstPointer<Byte> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glSecondaryColor3ubvEXT(\{v})");
		try {
		    functions.glSecondaryColor3ubvEXT().invokeExact(v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glSecondaryColor3ui">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glSecondaryColor3ui.xhtml">registry.khronos.org</a>
	 */
	public void glSecondaryColor3ui(int red, int green, int blue) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glSecondaryColor3ui(\{red}, \{green}, \{blue})");
		try {
		    functions.glSecondaryColor3ui().invokeExact(red, green, blue);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glSecondaryColor3uiEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glSecondaryColor3uiEXT.xhtml">registry.khronos.org</a>
	 */
	public void glSecondaryColor3uiEXT(int red, int green, int blue) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glSecondaryColor3uiEXT(\{red}, \{green}, \{blue})");
		try {
		    functions.glSecondaryColor3uiEXT().invokeExact(red, green, blue);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glSecondaryColor3uiv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glSecondaryColor3uiv.xhtml">registry.khronos.org</a>
	 */
	public void glSecondaryColor3uiv(ConstPointer<Integer> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glSecondaryColor3uiv(\{v})");
		try {
		    functions.glSecondaryColor3uiv().invokeExact(v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glSecondaryColor3uivEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glSecondaryColor3uivEXT.xhtml">registry.khronos.org</a>
	 */
	public void glSecondaryColor3uivEXT(ConstPointer<Integer> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glSecondaryColor3uivEXT(\{v})");
		try {
		    functions.glSecondaryColor3uivEXT().invokeExact(v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glSecondaryColor3us">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glSecondaryColor3us.xhtml">registry.khronos.org</a>
	 */
	public void glSecondaryColor3us(char red, char green, char blue) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glSecondaryColor3us(\{red}, \{green}, \{blue})");
		try {
		    functions.glSecondaryColor3us().invokeExact(red, green, blue);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glSecondaryColor3usEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glSecondaryColor3usEXT.xhtml">registry.khronos.org</a>
	 */
	public void glSecondaryColor3usEXT(char red, char green, char blue) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glSecondaryColor3usEXT(\{red}, \{green}, \{blue})");
		try {
		    functions.glSecondaryColor3usEXT().invokeExact(red, green, blue);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glSecondaryColor3usv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glSecondaryColor3usv.xhtml">registry.khronos.org</a>
	 */
	public void glSecondaryColor3usv(ConstPointer<Character> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glSecondaryColor3usv(\{v})");
		try {
		    functions.glSecondaryColor3usv().invokeExact(v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glSecondaryColor3usvEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glSecondaryColor3usvEXT.xhtml">registry.khronos.org</a>
	 */
	public void glSecondaryColor3usvEXT(ConstPointer<Character> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glSecondaryColor3usvEXT(\{v})");
		try {
		    functions.glSecondaryColor3usvEXT().invokeExact(v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glSecondaryColorFormatNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glSecondaryColorFormatNV.xhtml">registry.khronos.org</a>
	 */
	public void glSecondaryColorFormatNV(int size, GLColorPointerType type, int stride) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glSecondaryColorFormatNV(\{size}, \{type}, \{stride})");
		try {
		    functions.glSecondaryColorFormatNV().invokeExact(size, type, stride);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glSecondaryColorP3ui">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glSecondaryColorP3ui.xhtml">registry.khronos.org</a>
	 */
	public void glSecondaryColorP3ui(GLColorPointerType type, int color) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glSecondaryColorP3ui(\{type}, \{color})");
		try {
		    functions.glSecondaryColorP3ui().invokeExact(type, color);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glSecondaryColorP3uiv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glSecondaryColorP3uiv.xhtml">registry.khronos.org</a>
	 */
	public void glSecondaryColorP3uiv(GLColorPointerType type, ConstPointer<Integer> color) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glSecondaryColorP3uiv(\{type}, \{color})");
		try {
		    functions.glSecondaryColorP3uiv().invokeExact(type, color);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glSecondaryColorPointer">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glSecondaryColorPointer.xhtml">registry.khronos.org</a>
	 */
	public void glSecondaryColorPointer(int size, GLColorPointerType type, int stride, ConstPointer<Void> pointer) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glSecondaryColorPointer(\{size}, \{type}, \{stride}, \{pointer})");
		try {
		    functions.glSecondaryColorPointer().invokeExact(size, type, stride, pointer);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glSecondaryColorPointerEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glSecondaryColorPointerEXT.xhtml">registry.khronos.org</a>
	 */
	public void glSecondaryColorPointerEXT(int size, GLColorPointerType type, int stride, ConstPointer<Void> pointer) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glSecondaryColorPointerEXT(\{size}, \{type}, \{stride}, \{pointer})");
		try {
		    functions.glSecondaryColorPointerEXT().invokeExact(size, type, stride, pointer);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glSecondaryColorPointerListIBM">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glSecondaryColorPointerListIBM.xhtml">registry.khronos.org</a>
	 */
	public void glSecondaryColorPointerListIBM(int size, GLSecondaryColorPointerTypeIBM type, int stride, MutPointer<ConstPointer<Void>> pointer, int ptrstride) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glSecondaryColorPointerListIBM(\{size}, \{type}, \{stride}, \{pointer}, \{ptrstride})");
		try {
		    functions.glSecondaryColorPointerListIBM().invokeExact(size, type, stride, pointer, ptrstride);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glSelectBuffer">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glSelectBuffer.xhtml">registry.khronos.org</a>
	 */
	public void glSelectBuffer(int size, MutPointer<Integer> buffer) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glSelectBuffer(\{size}, \{buffer})");
		try {
		    functions.glSelectBuffer().invokeExact(size, buffer);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glSelectPerfMonitorCountersAMD">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glSelectPerfMonitorCountersAMD.xhtml">registry.khronos.org</a>
	 */
	public void glSelectPerfMonitorCountersAMD(int monitor, boolean enable, int group, int numCounters, MutPointer<Integer> counterList) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glSelectPerfMonitorCountersAMD(\{monitor}, \{enable}, \{group}, \{numCounters}, \{counterList})");
		try {
		    functions.glSelectPerfMonitorCountersAMD().invokeExact(monitor, enable, group, numCounters, counterList);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glSemaphoreParameterivNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glSemaphoreParameterivNV.xhtml">registry.khronos.org</a>
	 */
	public void glSemaphoreParameterivNV(int semaphore, GLSemaphoreParameterName pname, ConstPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glSemaphoreParameterivNV(\{semaphore}, \{pname}, \{params})");
		try {
		    functions.glSemaphoreParameterivNV().invokeExact(semaphore, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glSemaphoreParameterui64vEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glSemaphoreParameterui64vEXT.xhtml">registry.khronos.org</a>
	 */
	public void glSemaphoreParameterui64vEXT(int semaphore, GLSemaphoreParameterName pname, ConstPointer<Long> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glSemaphoreParameterui64vEXT(\{semaphore}, \{pname}, \{params})");
		try {
		    functions.glSemaphoreParameterui64vEXT().invokeExact(semaphore, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glSeparableFilter2D">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glSeparableFilter2D.xhtml">registry.khronos.org</a>
	 */
	public void glSeparableFilter2D(GLSeparableTargetEXT target, GLInternalFormat internalformat, int width, int height, GLPixelFormat format, GLPixelType type, ConstPointer<Void> row, ConstPointer<Void> column) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glSeparableFilter2D(\{target}, \{internalformat}, \{width}, \{height}, \{format}, \{type}, \{row}, \{column})");
		try {
		    functions.glSeparableFilter2D().invokeExact(target, internalformat, width, height, format, type, row, column);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, internalformat=\{internalformat}, format=\{format}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glSeparableFilter2DEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glSeparableFilter2DEXT.xhtml">registry.khronos.org</a>
	 */
	public void glSeparableFilter2DEXT(GLSeparableTargetEXT target, GLInternalFormat internalformat, int width, int height, GLPixelFormat format, GLPixelType type, ConstPointer<Void> row, ConstPointer<Void> column) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glSeparableFilter2DEXT(\{target}, \{internalformat}, \{width}, \{height}, \{format}, \{type}, \{row}, \{column})");
		try {
		    functions.glSeparableFilter2DEXT().invokeExact(target, internalformat, width, height, format, type, row, column);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, internalformat=\{internalformat}, format=\{format}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glSetFenceAPPLE">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glSetFenceAPPLE.xhtml">registry.khronos.org</a>
	 */
	public void glSetFenceAPPLE(int fence) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glSetFenceAPPLE(\{fence})");
		try {
		    functions.glSetFenceAPPLE().invokeExact(fence);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glSetFenceNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glSetFenceNV.xhtml">registry.khronos.org</a>
	 */
	public void glSetFenceNV(int fence, GLFenceConditionNV condition) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glSetFenceNV(\{fence}, \{condition})");
		try {
		    functions.glSetFenceNV().invokeExact(fence, condition);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were condition=\{condition}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glSetFragmentShaderConstantATI">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glSetFragmentShaderConstantATI.xhtml">registry.khronos.org</a>
	 */
	public void glSetFragmentShaderConstantATI(int dst, ConstPointer<Float> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glSetFragmentShaderConstantATI(\{dst}, \{value})");
		try {
		    functions.glSetFragmentShaderConstantATI().invokeExact(dst, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glSetInvariantEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glSetInvariantEXT.xhtml">registry.khronos.org</a>
	 */
	public void glSetInvariantEXT(int id, GLScalarType type, ConstPointer<Void> addr) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glSetInvariantEXT(\{id}, \{type}, \{addr})");
		try {
		    functions.glSetInvariantEXT().invokeExact(id, type, addr);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glSetLocalConstantEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glSetLocalConstantEXT.xhtml">registry.khronos.org</a>
	 */
	public void glSetLocalConstantEXT(int id, GLScalarType type, ConstPointer<Void> addr) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glSetLocalConstantEXT(\{id}, \{type}, \{addr})");
		try {
		    functions.glSetLocalConstantEXT().invokeExact(id, type, addr);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glSetMultisamplefvAMD">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glSetMultisamplefvAMD.xhtml">registry.khronos.org</a>
	 */
	public void glSetMultisamplefvAMD(GLEnum pname, int index, ConstPointer<Float> val) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glSetMultisamplefvAMD(\{pname}, \{index}, \{val})");
		try {
		    functions.glSetMultisamplefvAMD().invokeExact(pname, index, val);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glShadeModel">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glShadeModel.xhtml">registry.khronos.org</a>
	 */
	public void glShadeModel(GLShadingModel mode) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glShadeModel(\{mode})");
		try {
		    functions.glShadeModel().invokeExact(mode);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glShaderBinary">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glShaderBinary.xhtml">registry.khronos.org</a>
	 */
	public void glShaderBinary(int count, ConstPointer<Integer> shaders, GLShaderBinaryFormat binaryFormat, ConstPointer<Void> binary, int length) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glShaderBinary(\{count}, \{shaders}, \{binaryFormat}, \{binary}, \{length})");
		try {
		    functions.glShaderBinary().invokeExact(count, shaders, binaryFormat, binary, length);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were binaryFormat=\{binaryFormat}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glShaderOp1EXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glShaderOp1EXT.xhtml">registry.khronos.org</a>
	 */
	public void glShaderOp1EXT(GLVertexShaderOpEXT op, int res, int arg1) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glShaderOp1EXT(\{op}, \{res}, \{arg1})");
		try {
		    functions.glShaderOp1EXT().invokeExact(op, res, arg1);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were op=\{op}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glShaderOp2EXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glShaderOp2EXT.xhtml">registry.khronos.org</a>
	 */
	public void glShaderOp2EXT(GLVertexShaderOpEXT op, int res, int arg1, int arg2) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glShaderOp2EXT(\{op}, \{res}, \{arg1}, \{arg2})");
		try {
		    functions.glShaderOp2EXT().invokeExact(op, res, arg1, arg2);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were op=\{op}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glShaderOp3EXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glShaderOp3EXT.xhtml">registry.khronos.org</a>
	 */
	public void glShaderOp3EXT(GLVertexShaderOpEXT op, int res, int arg1, int arg2, int arg3) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glShaderOp3EXT(\{op}, \{res}, \{arg1}, \{arg2}, \{arg3})");
		try {
		    functions.glShaderOp3EXT().invokeExact(op, res, arg1, arg2, arg3);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were op=\{op}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glShaderSource">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glShaderSource.xhtml">registry.khronos.org</a>
	 */
	public void glShaderSource(int shader, int count, ConstPointer<ConstPointer<Byte>> string, ConstPointer<Integer> length) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glShaderSource(\{shader}, \{count}, \{string}, \{length})");
		try {
		    functions.glShaderSource().invokeExact(shader, count, string, length);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glShaderSourceARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glShaderSourceARB.xhtml">registry.khronos.org</a>
	 */
	public void glShaderSourceARB(int shaderObj, int count, MutPointer<ConstPointer<Byte>> string, ConstPointer<Integer> length) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glShaderSourceARB(\{shaderObj}, \{count}, \{string}, \{length})");
		try {
		    functions.glShaderSourceARB().invokeExact(shaderObj, count, string, length);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glShaderStorageBlockBinding">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glShaderStorageBlockBinding.xhtml">registry.khronos.org</a>
	 */
	public void glShaderStorageBlockBinding(int program, int storageBlockIndex, int storageBlockBinding) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glShaderStorageBlockBinding(\{program}, \{storageBlockIndex}, \{storageBlockBinding})");
		try {
		    functions.glShaderStorageBlockBinding().invokeExact(program, storageBlockIndex, storageBlockBinding);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glShadingRateEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glShadingRateEXT.xhtml">registry.khronos.org</a>
	 */
	public void glShadingRateEXT(GLShadingRate rate) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glShadingRateEXT(\{rate})");
		try {
		    functions.glShadingRateEXT().invokeExact(rate);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were rate=\{rate}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glShadingRateCombinerOpsEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glShadingRateCombinerOpsEXT.xhtml">registry.khronos.org</a>
	 */
	public void glShadingRateCombinerOpsEXT(GLShadingRateCombinerOp combinerOp0, GLShadingRateCombinerOp combinerOp1) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glShadingRateCombinerOpsEXT(\{combinerOp0}, \{combinerOp1})");
		try {
		    functions.glShadingRateCombinerOpsEXT().invokeExact(combinerOp0, combinerOp1);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were combinerOp0=\{combinerOp0}, combinerOp1=\{combinerOp1}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glShadingRateImageBarrierNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glShadingRateImageBarrierNV.xhtml">registry.khronos.org</a>
	 */
	public void glShadingRateImageBarrierNV(boolean synchronize) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glShadingRateImageBarrierNV(\{synchronize})");
		try {
		    functions.glShadingRateImageBarrierNV().invokeExact(synchronize);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glShadingRateQCOM">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glShadingRateQCOM.xhtml">registry.khronos.org</a>
	 */
	public void glShadingRateQCOM(GLShadingRateQCOM rate) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glShadingRateQCOM(\{rate})");
		try {
		    functions.glShadingRateQCOM().invokeExact(rate);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were rate=\{rate}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glShadingRateImagePaletteNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glShadingRateImagePaletteNV.xhtml">registry.khronos.org</a>
	 */
	public void glShadingRateImagePaletteNV(int viewport, int first, int count, ConstPointer<GLEnum> rates) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glShadingRateImagePaletteNV(\{viewport}, \{first}, \{count}, \{rates})");
		try {
		    functions.glShadingRateImagePaletteNV().invokeExact(viewport, first, count, rates);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glShadingRateSampleOrderNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glShadingRateSampleOrderNV.xhtml">registry.khronos.org</a>
	 */
	public void glShadingRateSampleOrderNV(GLEnum order) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glShadingRateSampleOrderNV(\{order})");
		try {
		    functions.glShadingRateSampleOrderNV().invokeExact(order);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were order=\{order}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glShadingRateSampleOrderCustomNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glShadingRateSampleOrderCustomNV.xhtml">registry.khronos.org</a>
	 */
	public void glShadingRateSampleOrderCustomNV(GLEnum rate, int samples, ConstPointer<Integer> locations) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glShadingRateSampleOrderCustomNV(\{rate}, \{samples}, \{locations})");
		try {
		    functions.glShadingRateSampleOrderCustomNV().invokeExact(rate, samples, locations);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were rate=\{rate}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glSharpenTexFuncSGIS">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glSharpenTexFuncSGIS.xhtml">registry.khronos.org</a>
	 */
	public void glSharpenTexFuncSGIS(GLTextureTarget target, int n, ConstPointer<Float> points) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glSharpenTexFuncSGIS(\{target}, \{n}, \{points})");
		try {
		    functions.glSharpenTexFuncSGIS().invokeExact(target, n, points);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glSignalSemaphoreEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glSignalSemaphoreEXT.xhtml">registry.khronos.org</a>
	 */
	public void glSignalSemaphoreEXT(int semaphore, int numBufferBarriers, ConstPointer<Integer> buffers, int numTextureBarriers, ConstPointer<Integer> textures, ConstPointer<GLTextureLayout> dstLayouts) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glSignalSemaphoreEXT(\{semaphore}, \{numBufferBarriers}, \{buffers}, \{numTextureBarriers}, \{textures}, \{dstLayouts})");
		try {
		    functions.glSignalSemaphoreEXT().invokeExact(semaphore, numBufferBarriers, buffers, numTextureBarriers, textures, dstLayouts);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glSignalSemaphoreui64NVX">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glSignalSemaphoreui64NVX.xhtml">registry.khronos.org</a>
	 */
	public void glSignalSemaphoreui64NVX(int signalGpu, int fenceObjectCount, ConstPointer<Integer> semaphoreArray, ConstPointer<Long> fenceValueArray) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glSignalSemaphoreui64NVX(\{signalGpu}, \{fenceObjectCount}, \{semaphoreArray}, \{fenceValueArray})");
		try {
		    functions.glSignalSemaphoreui64NVX().invokeExact(signalGpu, fenceObjectCount, semaphoreArray, fenceValueArray);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glSpecializeShader">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glSpecializeShader.xhtml">registry.khronos.org</a>
	 */
	public void glSpecializeShader(int shader, ConstPointer<Byte> pEntryPoint, int numSpecializationConstants, ConstPointer<Integer> pConstantIndex, ConstPointer<Integer> pConstantValue) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glSpecializeShader(\{shader}, \{pEntryPoint}, \{numSpecializationConstants}, \{pConstantIndex}, \{pConstantValue})");
		try {
		    functions.glSpecializeShader().invokeExact(shader, pEntryPoint, numSpecializationConstants, pConstantIndex, pConstantValue);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glSpecializeShaderARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glSpecializeShaderARB.xhtml">registry.khronos.org</a>
	 */
	public void glSpecializeShaderARB(int shader, ConstPointer<Byte> pEntryPoint, int numSpecializationConstants, ConstPointer<Integer> pConstantIndex, ConstPointer<Integer> pConstantValue) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glSpecializeShaderARB(\{shader}, \{pEntryPoint}, \{numSpecializationConstants}, \{pConstantIndex}, \{pConstantValue})");
		try {
		    functions.glSpecializeShaderARB().invokeExact(shader, pEntryPoint, numSpecializationConstants, pConstantIndex, pConstantValue);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glSpriteParameterfSGIX">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glSpriteParameterfSGIX.xhtml">registry.khronos.org</a>
	 */
	public void glSpriteParameterfSGIX(GLSpriteParameterNameSGIX pname, float param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glSpriteParameterfSGIX(\{pname}, \{param})");
		try {
		    functions.glSpriteParameterfSGIX().invokeExact(pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glSpriteParameterfvSGIX">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glSpriteParameterfvSGIX.xhtml">registry.khronos.org</a>
	 */
	public void glSpriteParameterfvSGIX(GLSpriteParameterNameSGIX pname, ConstPointer<Float> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glSpriteParameterfvSGIX(\{pname}, \{params})");
		try {
		    functions.glSpriteParameterfvSGIX().invokeExact(pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glSpriteParameteriSGIX">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glSpriteParameteriSGIX.xhtml">registry.khronos.org</a>
	 */
	public void glSpriteParameteriSGIX(GLSpriteParameterNameSGIX pname, GLSpriteModeSGIX param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glSpriteParameteriSGIX(\{pname}, \{param})");
		try {
		    functions.glSpriteParameteriSGIX().invokeExact(pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}, param=\{param}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glSpriteParameterivSGIX">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glSpriteParameterivSGIX.xhtml">registry.khronos.org</a>
	 */
	public void glSpriteParameterivSGIX(GLSpriteParameterNameSGIX pname, ConstPointer<GLSpriteModeSGIX> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glSpriteParameterivSGIX(\{pname}, \{params})");
		try {
		    functions.glSpriteParameterivSGIX().invokeExact(pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glStartInstrumentsSGIX">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glStartInstrumentsSGIX.xhtml">registry.khronos.org</a>
	 */
	public void glStartInstrumentsSGIX() {
	    if (LOG_COMMANDS) logCmd(() -> STR."glStartInstrumentsSGIX()");
		try {
		    functions.glStartInstrumentsSGIX().invokeExact();
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glStartTilingQCOM">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glStartTilingQCOM.xhtml">registry.khronos.org</a>
	 */
	public void glStartTilingQCOM(int x, int y, int width, int height, int preserveMask) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glStartTilingQCOM(\{x}, \{y}, \{width}, \{height}, \{preserveMask})");
		try {
		    functions.glStartTilingQCOM().invokeExact(x, y, width, height, preserveMask);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glStateCaptureNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glStateCaptureNV.xhtml">registry.khronos.org</a>
	 */
	public void glStateCaptureNV(int state, GLEnum mode) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glStateCaptureNV(\{state}, \{mode})");
		try {
		    functions.glStateCaptureNV().invokeExact(state, mode);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glStencilClearTagEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glStencilClearTagEXT.xhtml">registry.khronos.org</a>
	 */
	public void glStencilClearTagEXT(int stencilTagBits, int stencilClearTag) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glStencilClearTagEXT(\{stencilTagBits}, \{stencilClearTag})");
		try {
		    functions.glStencilClearTagEXT().invokeExact(stencilTagBits, stencilClearTag);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glStencilFillPathInstancedNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glStencilFillPathInstancedNV.xhtml">registry.khronos.org</a>
	 */
	public void glStencilFillPathInstancedNV(int numPaths, GLPathElementType pathNameType, ConstPointer<Void> paths, int pathBase, GLPathFillMode fillMode, int mask, GLPathTransformType transformType, ConstPointer<Float> transformValues) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glStencilFillPathInstancedNV(\{numPaths}, \{pathNameType}, \{paths}, \{pathBase}, \{fillMode}, \{mask}, \{transformType}, \{transformValues})");
		try {
		    functions.glStencilFillPathInstancedNV().invokeExact(numPaths, pathNameType, paths, pathBase, fillMode, mask, transformType, transformValues);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pathNameType=\{pathNameType}, fillMode=\{fillMode}, transformType=\{transformType}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glStencilFillPathNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glStencilFillPathNV.xhtml">registry.khronos.org</a>
	 */
	public void glStencilFillPathNV(int path, GLPathFillMode fillMode, int mask) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glStencilFillPathNV(\{path}, \{fillMode}, \{mask})");
		try {
		    functions.glStencilFillPathNV().invokeExact(path, fillMode, mask);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were fillMode=\{fillMode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glStencilFunc">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glStencilFunc.xhtml">registry.khronos.org</a>
	 */
	public void glStencilFunc(GLStencilFunction func, int ref, int mask) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glStencilFunc(\{func}, \{ref}, \{mask})");
		try {
		    functions.glStencilFunc().invokeExact(func, ref, mask);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were func=\{func}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glStencilFuncSeparate">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glStencilFuncSeparate.xhtml">registry.khronos.org</a>
	 */
	public void glStencilFuncSeparate(GLTriangleFace face, GLStencilFunction func, int ref, int mask) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glStencilFuncSeparate(\{face}, \{func}, \{ref}, \{mask})");
		try {
		    functions.glStencilFuncSeparate().invokeExact(face, func, ref, mask);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were face=\{face}, func=\{func}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glStencilFuncSeparateATI">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glStencilFuncSeparateATI.xhtml">registry.khronos.org</a>
	 */
	public void glStencilFuncSeparateATI(GLStencilFunction frontfunc, GLStencilFunction backfunc, int ref, int mask) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glStencilFuncSeparateATI(\{frontfunc}, \{backfunc}, \{ref}, \{mask})");
		try {
		    functions.glStencilFuncSeparateATI().invokeExact(frontfunc, backfunc, ref, mask);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were frontfunc=\{frontfunc}, backfunc=\{backfunc}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glStencilMask">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glStencilMask.xhtml">registry.khronos.org</a>
	 */
	public void glStencilMask(int mask) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glStencilMask(\{mask})");
		try {
		    functions.glStencilMask().invokeExact(mask);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glStencilMaskSeparate">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glStencilMaskSeparate.xhtml">registry.khronos.org</a>
	 */
	public void glStencilMaskSeparate(GLTriangleFace face, int mask) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glStencilMaskSeparate(\{face}, \{mask})");
		try {
		    functions.glStencilMaskSeparate().invokeExact(face, mask);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were face=\{face}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glStencilOp">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glStencilOp.xhtml">registry.khronos.org</a>
	 */
	public void glStencilOp(GLStencilOp fail, GLStencilOp zfail, GLStencilOp zpass) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glStencilOp(\{fail}, \{zfail}, \{zpass})");
		try {
		    functions.glStencilOp().invokeExact(fail, zfail, zpass);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were fail=\{fail}, zfail=\{zfail}, zpass=\{zpass}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glStencilOpSeparate">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glStencilOpSeparate.xhtml">registry.khronos.org</a>
	 */
	public void glStencilOpSeparate(GLTriangleFace face, GLStencilOp sfail, GLStencilOp dpfail, GLStencilOp dppass) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glStencilOpSeparate(\{face}, \{sfail}, \{dpfail}, \{dppass})");
		try {
		    functions.glStencilOpSeparate().invokeExact(face, sfail, dpfail, dppass);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were face=\{face}, sfail=\{sfail}, dpfail=\{dpfail}, dppass=\{dppass}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glStencilOpSeparateATI">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glStencilOpSeparateATI.xhtml">registry.khronos.org</a>
	 */
	public void glStencilOpSeparateATI(GLTriangleFace face, GLStencilOp sfail, GLStencilOp dpfail, GLStencilOp dppass) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glStencilOpSeparateATI(\{face}, \{sfail}, \{dpfail}, \{dppass})");
		try {
		    functions.glStencilOpSeparateATI().invokeExact(face, sfail, dpfail, dppass);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were face=\{face}, sfail=\{sfail}, dpfail=\{dpfail}, dppass=\{dppass}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glStencilOpValueAMD">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glStencilOpValueAMD.xhtml">registry.khronos.org</a>
	 */
	public void glStencilOpValueAMD(GLTriangleFace face, int value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glStencilOpValueAMD(\{face}, \{value})");
		try {
		    functions.glStencilOpValueAMD().invokeExact(face, value);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were face=\{face}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glStencilStrokePathInstancedNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glStencilStrokePathInstancedNV.xhtml">registry.khronos.org</a>
	 */
	public void glStencilStrokePathInstancedNV(int numPaths, GLPathElementType pathNameType, ConstPointer<Void> paths, int pathBase, int reference, int mask, GLPathTransformType transformType, ConstPointer<Float> transformValues) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glStencilStrokePathInstancedNV(\{numPaths}, \{pathNameType}, \{paths}, \{pathBase}, \{reference}, \{mask}, \{transformType}, \{transformValues})");
		try {
		    functions.glStencilStrokePathInstancedNV().invokeExact(numPaths, pathNameType, paths, pathBase, reference, mask, transformType, transformValues);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pathNameType=\{pathNameType}, transformType=\{transformType}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glStencilStrokePathNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glStencilStrokePathNV.xhtml">registry.khronos.org</a>
	 */
	public void glStencilStrokePathNV(int path, int reference, int mask) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glStencilStrokePathNV(\{path}, \{reference}, \{mask})");
		try {
		    functions.glStencilStrokePathNV().invokeExact(path, reference, mask);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glStencilThenCoverFillPathInstancedNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glStencilThenCoverFillPathInstancedNV.xhtml">registry.khronos.org</a>
	 */
	public void glStencilThenCoverFillPathInstancedNV(int numPaths, GLPathElementType pathNameType, ConstPointer<Void> paths, int pathBase, GLPathFillMode fillMode, int mask, GLInstancedPathCoverMode coverMode, GLPathTransformType transformType, ConstPointer<Float> transformValues) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glStencilThenCoverFillPathInstancedNV(\{numPaths}, \{pathNameType}, \{paths}, \{pathBase}, \{fillMode}, \{mask}, \{coverMode}, \{transformType}, \{transformValues})");
		try {
		    functions.glStencilThenCoverFillPathInstancedNV().invokeExact(numPaths, pathNameType, paths, pathBase, fillMode, mask, coverMode, transformType, transformValues);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pathNameType=\{pathNameType}, fillMode=\{fillMode}, coverMode=\{coverMode}, transformType=\{transformType}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glStencilThenCoverFillPathNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glStencilThenCoverFillPathNV.xhtml">registry.khronos.org</a>
	 */
	public void glStencilThenCoverFillPathNV(int path, GLPathFillMode fillMode, int mask, GLPathCoverMode coverMode) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glStencilThenCoverFillPathNV(\{path}, \{fillMode}, \{mask}, \{coverMode})");
		try {
		    functions.glStencilThenCoverFillPathNV().invokeExact(path, fillMode, mask, coverMode);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were fillMode=\{fillMode}, coverMode=\{coverMode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glStencilThenCoverStrokePathInstancedNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glStencilThenCoverStrokePathInstancedNV.xhtml">registry.khronos.org</a>
	 */
	public void glStencilThenCoverStrokePathInstancedNV(int numPaths, GLPathElementType pathNameType, ConstPointer<Void> paths, int pathBase, int reference, int mask, GLInstancedPathCoverMode coverMode, GLPathTransformType transformType, ConstPointer<Float> transformValues) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glStencilThenCoverStrokePathInstancedNV(\{numPaths}, \{pathNameType}, \{paths}, \{pathBase}, \{reference}, \{mask}, \{coverMode}, \{transformType}, \{transformValues})");
		try {
		    functions.glStencilThenCoverStrokePathInstancedNV().invokeExact(numPaths, pathNameType, paths, pathBase, reference, mask, coverMode, transformType, transformValues);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pathNameType=\{pathNameType}, coverMode=\{coverMode}, transformType=\{transformType}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glStencilThenCoverStrokePathNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glStencilThenCoverStrokePathNV.xhtml">registry.khronos.org</a>
	 */
	public void glStencilThenCoverStrokePathNV(int path, int reference, int mask, GLPathCoverMode coverMode) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glStencilThenCoverStrokePathNV(\{path}, \{reference}, \{mask}, \{coverMode})");
		try {
		    functions.glStencilThenCoverStrokePathNV().invokeExact(path, reference, mask, coverMode);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were coverMode=\{coverMode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glStopInstrumentsSGIX">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glStopInstrumentsSGIX.xhtml">registry.khronos.org</a>
	 */
	public void glStopInstrumentsSGIX(int marker) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glStopInstrumentsSGIX(\{marker})");
		try {
		    functions.glStopInstrumentsSGIX().invokeExact(marker);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glStringMarkerGREMEDY">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glStringMarkerGREMEDY.xhtml">registry.khronos.org</a>
	 */
	public void glStringMarkerGREMEDY(int len, ConstPointer<Void> string) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glStringMarkerGREMEDY(\{len}, \{string})");
		try {
		    functions.glStringMarkerGREMEDY().invokeExact(len, string);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glSubpixelPrecisionBiasNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glSubpixelPrecisionBiasNV.xhtml">registry.khronos.org</a>
	 */
	public void glSubpixelPrecisionBiasNV(int xbits, int ybits) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glSubpixelPrecisionBiasNV(\{xbits}, \{ybits})");
		try {
		    functions.glSubpixelPrecisionBiasNV().invokeExact(xbits, ybits);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glSwizzleEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glSwizzleEXT.xhtml">registry.khronos.org</a>
	 */
	public void glSwizzleEXT(int res, int in, GLVertexShaderCoordOutEXT outX, GLVertexShaderCoordOutEXT outY, GLVertexShaderCoordOutEXT outZ, GLVertexShaderCoordOutEXT outW) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glSwizzleEXT(\{res}, \{in}, \{outX}, \{outY}, \{outZ}, \{outW})");
		try {
		    functions.glSwizzleEXT().invokeExact(res, in, outX, outY, outZ, outW);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were outX=\{outX}, outY=\{outY}, outZ=\{outZ}, outW=\{outW}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glSyncTextureINTEL">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glSyncTextureINTEL.xhtml">registry.khronos.org</a>
	 */
	public void glSyncTextureINTEL(int texture) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glSyncTextureINTEL(\{texture})");
		try {
		    functions.glSyncTextureINTEL().invokeExact(texture);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTagSampleBufferSGIX">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTagSampleBufferSGIX.xhtml">registry.khronos.org</a>
	 */
	public void glTagSampleBufferSGIX() {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTagSampleBufferSGIX()");
		try {
		    functions.glTagSampleBufferSGIX().invokeExact();
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTangent3bEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTangent3bEXT.xhtml">registry.khronos.org</a>
	 */
	public void glTangent3bEXT(byte tx, byte ty, byte tz) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTangent3bEXT(\{tx}, \{ty}, \{tz})");
		try {
		    functions.glTangent3bEXT().invokeExact(tx, ty, tz);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTangent3bvEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTangent3bvEXT.xhtml">registry.khronos.org</a>
	 */
	public void glTangent3bvEXT(ConstPointer<Byte> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTangent3bvEXT(\{v})");
		try {
		    functions.glTangent3bvEXT().invokeExact(v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTangent3dEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTangent3dEXT.xhtml">registry.khronos.org</a>
	 */
	public void glTangent3dEXT(double tx, double ty, double tz) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTangent3dEXT(\{tx}, \{ty}, \{tz})");
		try {
		    functions.glTangent3dEXT().invokeExact(tx, ty, tz);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTangent3dvEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTangent3dvEXT.xhtml">registry.khronos.org</a>
	 */
	public void glTangent3dvEXT(ConstPointer<Double> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTangent3dvEXT(\{v})");
		try {
		    functions.glTangent3dvEXT().invokeExact(v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTangent3fEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTangent3fEXT.xhtml">registry.khronos.org</a>
	 */
	public void glTangent3fEXT(float tx, float ty, float tz) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTangent3fEXT(\{tx}, \{ty}, \{tz})");
		try {
		    functions.glTangent3fEXT().invokeExact(tx, ty, tz);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTangent3fvEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTangent3fvEXT.xhtml">registry.khronos.org</a>
	 */
	public void glTangent3fvEXT(ConstPointer<Float> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTangent3fvEXT(\{v})");
		try {
		    functions.glTangent3fvEXT().invokeExact(v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTangent3iEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTangent3iEXT.xhtml">registry.khronos.org</a>
	 */
	public void glTangent3iEXT(int tx, int ty, int tz) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTangent3iEXT(\{tx}, \{ty}, \{tz})");
		try {
		    functions.glTangent3iEXT().invokeExact(tx, ty, tz);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTangent3ivEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTangent3ivEXT.xhtml">registry.khronos.org</a>
	 */
	public void glTangent3ivEXT(ConstPointer<Integer> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTangent3ivEXT(\{v})");
		try {
		    functions.glTangent3ivEXT().invokeExact(v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTangent3sEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTangent3sEXT.xhtml">registry.khronos.org</a>
	 */
	public void glTangent3sEXT(short tx, short ty, short tz) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTangent3sEXT(\{tx}, \{ty}, \{tz})");
		try {
		    functions.glTangent3sEXT().invokeExact(tx, ty, tz);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTangent3svEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTangent3svEXT.xhtml">registry.khronos.org</a>
	 */
	public void glTangent3svEXT(ConstPointer<Short> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTangent3svEXT(\{v})");
		try {
		    functions.glTangent3svEXT().invokeExact(v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTangentPointerEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTangentPointerEXT.xhtml">registry.khronos.org</a>
	 */
	public void glTangentPointerEXT(GLTangentPointerTypeEXT type, int stride, ConstPointer<Void> pointer) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTangentPointerEXT(\{type}, \{stride}, \{pointer})");
		try {
		    functions.glTangentPointerEXT().invokeExact(type, stride, pointer);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTbufferMask3DFX">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTbufferMask3DFX.xhtml">registry.khronos.org</a>
	 */
	public void glTbufferMask3DFX(int mask) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTbufferMask3DFX(\{mask})");
		try {
		    functions.glTbufferMask3DFX().invokeExact(mask);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTessellationFactorAMD">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTessellationFactorAMD.xhtml">registry.khronos.org</a>
	 */
	public void glTessellationFactorAMD(float factor) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTessellationFactorAMD(\{factor})");
		try {
		    functions.glTessellationFactorAMD().invokeExact(factor);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTessellationModeAMD">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTessellationModeAMD.xhtml">registry.khronos.org</a>
	 */
	public void glTessellationModeAMD(GLEnum mode) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTessellationModeAMD(\{mode})");
		try {
		    functions.glTessellationModeAMD().invokeExact(mode);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTestFenceAPPLE">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTestFenceAPPLE.xhtml">registry.khronos.org</a>
	 */
	public boolean glTestFenceAPPLE(int fence) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTestFenceAPPLE(\{fence})");
		try {
		    var r = (boolean)functions.glTestFenceAPPLE().invokeExact(fence);
			checkError();
			return r;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTestFenceNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTestFenceNV.xhtml">registry.khronos.org</a>
	 */
	public boolean glTestFenceNV(int fence) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTestFenceNV(\{fence})");
		try {
		    var r = (boolean)functions.glTestFenceNV().invokeExact(fence);
			checkError();
			return r;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTestObjectAPPLE">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTestObjectAPPLE.xhtml">registry.khronos.org</a>
	 */
	public boolean glTestObjectAPPLE(GLObjectTypeAPPLE object, int name) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTestObjectAPPLE(\{object}, \{name})");
		try {
		    var r = (boolean)functions.glTestObjectAPPLE().invokeExact(object, name);
			checkError();
			return r;
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were object=\{object}.");
			}
			throw e;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexAttachMemoryNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexAttachMemoryNV.xhtml">registry.khronos.org</a>
	 */
	public void glTexAttachMemoryNV(GLTextureTarget target, int memory, long offset) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexAttachMemoryNV(\{target}, \{memory}, \{offset})");
		try {
		    functions.glTexAttachMemoryNV().invokeExact(target, memory, offset);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexBuffer">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexBuffer.xhtml">registry.khronos.org</a>
	 */
	public void glTexBuffer(GLTextureTarget target, GLSizedInternalFormat internalformat, int buffer) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexBuffer(\{target}, \{internalformat}, \{buffer})");
		try {
		    functions.glTexBuffer().invokeExact(target, internalformat, buffer);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, internalformat=\{internalformat}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexBufferARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexBufferARB.xhtml">registry.khronos.org</a>
	 */
	public void glTexBufferARB(GLTextureTarget target, GLSizedInternalFormat internalformat, int buffer) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexBufferARB(\{target}, \{internalformat}, \{buffer})");
		try {
		    functions.glTexBufferARB().invokeExact(target, internalformat, buffer);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, internalformat=\{internalformat}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexBufferEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexBufferEXT.xhtml">registry.khronos.org</a>
	 */
	public void glTexBufferEXT(GLTextureTarget target, GLSizedInternalFormat internalformat, int buffer) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexBufferEXT(\{target}, \{internalformat}, \{buffer})");
		try {
		    functions.glTexBufferEXT().invokeExact(target, internalformat, buffer);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, internalformat=\{internalformat}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexBufferOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexBufferOES.xhtml">registry.khronos.org</a>
	 */
	public void glTexBufferOES(GLTextureTarget target, GLSizedInternalFormat internalformat, int buffer) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexBufferOES(\{target}, \{internalformat}, \{buffer})");
		try {
		    functions.glTexBufferOES().invokeExact(target, internalformat, buffer);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, internalformat=\{internalformat}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexBufferRange">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexBufferRange.xhtml">registry.khronos.org</a>
	 */
	public void glTexBufferRange(GLTextureTarget target, GLSizedInternalFormat internalformat, int buffer, long offset, long size) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexBufferRange(\{target}, \{internalformat}, \{buffer}, \{offset}, \{size})");
		try {
		    functions.glTexBufferRange().invokeExact(target, internalformat, buffer, offset, size);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, internalformat=\{internalformat}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexBufferRangeEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexBufferRangeEXT.xhtml">registry.khronos.org</a>
	 */
	public void glTexBufferRangeEXT(GLTextureTarget target, GLSizedInternalFormat internalformat, int buffer, long offset, long size) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexBufferRangeEXT(\{target}, \{internalformat}, \{buffer}, \{offset}, \{size})");
		try {
		    functions.glTexBufferRangeEXT().invokeExact(target, internalformat, buffer, offset, size);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, internalformat=\{internalformat}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexBufferRangeOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexBufferRangeOES.xhtml">registry.khronos.org</a>
	 */
	public void glTexBufferRangeOES(GLTextureTarget target, GLSizedInternalFormat internalformat, int buffer, long offset, long size) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexBufferRangeOES(\{target}, \{internalformat}, \{buffer}, \{offset}, \{size})");
		try {
		    functions.glTexBufferRangeOES().invokeExact(target, internalformat, buffer, offset, size);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, internalformat=\{internalformat}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexBumpParameterfvATI">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexBumpParameterfvATI.xhtml">registry.khronos.org</a>
	 */
	public void glTexBumpParameterfvATI(GLTexBumpParameterATI pname, ConstPointer<Float> param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexBumpParameterfvATI(\{pname}, \{param})");
		try {
		    functions.glTexBumpParameterfvATI().invokeExact(pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexBumpParameterivATI">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexBumpParameterivATI.xhtml">registry.khronos.org</a>
	 */
	public void glTexBumpParameterivATI(GLTexBumpParameterATI pname, ConstPointer<Integer> param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexBumpParameterivATI(\{pname}, \{param})");
		try {
		    functions.glTexBumpParameterivATI().invokeExact(pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexCoord1bOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexCoord1bOES.xhtml">registry.khronos.org</a>
	 */
	public void glTexCoord1bOES(byte s) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexCoord1bOES(\{s})");
		try {
		    functions.glTexCoord1bOES().invokeExact(s);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexCoord1bvOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexCoord1bvOES.xhtml">registry.khronos.org</a>
	 */
	public void glTexCoord1bvOES(ConstPointer<Byte> coords) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexCoord1bvOES(\{coords})");
		try {
		    functions.glTexCoord1bvOES().invokeExact(coords);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexCoord1d">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexCoord1d.xhtml">registry.khronos.org</a>
	 */
	public void glTexCoord1d(double s) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexCoord1d(\{s})");
		try {
		    functions.glTexCoord1d().invokeExact(s);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexCoord1dv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexCoord1dv.xhtml">registry.khronos.org</a>
	 */
	public void glTexCoord1dv(ConstPointer<Double> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexCoord1dv(\{v})");
		try {
		    functions.glTexCoord1dv().invokeExact(v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexCoord1f">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexCoord1f.xhtml">registry.khronos.org</a>
	 */
	public void glTexCoord1f(float s) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexCoord1f(\{s})");
		try {
		    functions.glTexCoord1f().invokeExact(s);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexCoord1fv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexCoord1fv.xhtml">registry.khronos.org</a>
	 */
	public void glTexCoord1fv(ConstPointer<Float> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexCoord1fv(\{v})");
		try {
		    functions.glTexCoord1fv().invokeExact(v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexCoord1hNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexCoord1hNV.xhtml">registry.khronos.org</a>
	 */
	public void glTexCoord1hNV(char s) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexCoord1hNV(\{s})");
		try {
		    functions.glTexCoord1hNV().invokeExact(s);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexCoord1hvNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexCoord1hvNV.xhtml">registry.khronos.org</a>
	 */
	public void glTexCoord1hvNV(ConstPointer<Character> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexCoord1hvNV(\{v})");
		try {
		    functions.glTexCoord1hvNV().invokeExact(v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexCoord1i">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexCoord1i.xhtml">registry.khronos.org</a>
	 */
	public void glTexCoord1i(int s) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexCoord1i(\{s})");
		try {
		    functions.glTexCoord1i().invokeExact(s);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexCoord1iv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexCoord1iv.xhtml">registry.khronos.org</a>
	 */
	public void glTexCoord1iv(ConstPointer<Integer> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexCoord1iv(\{v})");
		try {
		    functions.glTexCoord1iv().invokeExact(v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexCoord1s">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexCoord1s.xhtml">registry.khronos.org</a>
	 */
	public void glTexCoord1s(short s) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexCoord1s(\{s})");
		try {
		    functions.glTexCoord1s().invokeExact(s);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexCoord1sv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexCoord1sv.xhtml">registry.khronos.org</a>
	 */
	public void glTexCoord1sv(ConstPointer<Short> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexCoord1sv(\{v})");
		try {
		    functions.glTexCoord1sv().invokeExact(v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexCoord1xOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexCoord1xOES.xhtml">registry.khronos.org</a>
	 */
	public void glTexCoord1xOES(int s) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexCoord1xOES(\{s})");
		try {
		    functions.glTexCoord1xOES().invokeExact(s);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexCoord1xvOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexCoord1xvOES.xhtml">registry.khronos.org</a>
	 */
	public void glTexCoord1xvOES(ConstPointer<Integer> coords) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexCoord1xvOES(\{coords})");
		try {
		    functions.glTexCoord1xvOES().invokeExact(coords);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexCoord2bOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexCoord2bOES.xhtml">registry.khronos.org</a>
	 */
	public void glTexCoord2bOES(byte s, byte t) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexCoord2bOES(\{s}, \{t})");
		try {
		    functions.glTexCoord2bOES().invokeExact(s, t);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexCoord2bvOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexCoord2bvOES.xhtml">registry.khronos.org</a>
	 */
	public void glTexCoord2bvOES(ConstPointer<Byte> coords) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexCoord2bvOES(\{coords})");
		try {
		    functions.glTexCoord2bvOES().invokeExact(coords);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexCoord2d">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexCoord2d.xhtml">registry.khronos.org</a>
	 */
	public void glTexCoord2d(double s, double t) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexCoord2d(\{s}, \{t})");
		try {
		    functions.glTexCoord2d().invokeExact(s, t);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexCoord2dv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexCoord2dv.xhtml">registry.khronos.org</a>
	 */
	public void glTexCoord2dv(ConstPointer<Double> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexCoord2dv(\{v})");
		try {
		    functions.glTexCoord2dv().invokeExact(v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexCoord2f">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexCoord2f.xhtml">registry.khronos.org</a>
	 */
	public void glTexCoord2f(float s, float t) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexCoord2f(\{s}, \{t})");
		try {
		    functions.glTexCoord2f().invokeExact(s, t);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexCoord2fColor3fVertex3fSUN">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexCoord2fColor3fVertex3fSUN.xhtml">registry.khronos.org</a>
	 */
	public void glTexCoord2fColor3fVertex3fSUN(float s, float t, float r, float g, float b, float x, float y, float z) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexCoord2fColor3fVertex3fSUN(\{s}, \{t}, \{r}, \{g}, \{b}, \{x}, \{y}, \{z})");
		try {
		    functions.glTexCoord2fColor3fVertex3fSUN().invokeExact(s, t, r, g, b, x, y, z);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexCoord2fColor3fVertex3fvSUN">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexCoord2fColor3fVertex3fvSUN.xhtml">registry.khronos.org</a>
	 */
	public void glTexCoord2fColor3fVertex3fvSUN(ConstPointer<Float> tc, ConstPointer<Float> c, ConstPointer<Float> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexCoord2fColor3fVertex3fvSUN(\{tc}, \{c}, \{v})");
		try {
		    functions.glTexCoord2fColor3fVertex3fvSUN().invokeExact(tc, c, v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexCoord2fColor4fNormal3fVertex3fSUN">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexCoord2fColor4fNormal3fVertex3fSUN.xhtml">registry.khronos.org</a>
	 */
	public void glTexCoord2fColor4fNormal3fVertex3fSUN(float s, float t, float r, float g, float b, float a, float nx, float ny, float nz, float x, float y, float z) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexCoord2fColor4fNormal3fVertex3fSUN(\{s}, \{t}, \{r}, \{g}, \{b}, \{a}, \{nx}, \{ny}, \{nz}, \{x}, \{y}, \{z})");
		try {
		    functions.glTexCoord2fColor4fNormal3fVertex3fSUN().invokeExact(s, t, r, g, b, a, nx, ny, nz, x, y, z);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexCoord2fColor4fNormal3fVertex3fvSUN">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexCoord2fColor4fNormal3fVertex3fvSUN.xhtml">registry.khronos.org</a>
	 */
	public void glTexCoord2fColor4fNormal3fVertex3fvSUN(ConstPointer<Float> tc, ConstPointer<Float> c, ConstPointer<Float> n, ConstPointer<Float> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexCoord2fColor4fNormal3fVertex3fvSUN(\{tc}, \{c}, \{n}, \{v})");
		try {
		    functions.glTexCoord2fColor4fNormal3fVertex3fvSUN().invokeExact(tc, c, n, v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexCoord2fColor4ubVertex3fSUN">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexCoord2fColor4ubVertex3fSUN.xhtml">registry.khronos.org</a>
	 */
	public void glTexCoord2fColor4ubVertex3fSUN(float s, float t, byte r, byte g, byte b, byte a, float x, float y, float z) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexCoord2fColor4ubVertex3fSUN(\{s}, \{t}, \{r}, \{g}, \{b}, \{a}, \{x}, \{y}, \{z})");
		try {
		    functions.glTexCoord2fColor4ubVertex3fSUN().invokeExact(s, t, r, g, b, a, x, y, z);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexCoord2fColor4ubVertex3fvSUN">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexCoord2fColor4ubVertex3fvSUN.xhtml">registry.khronos.org</a>
	 */
	public void glTexCoord2fColor4ubVertex3fvSUN(ConstPointer<Float> tc, ConstPointer<Byte> c, ConstPointer<Float> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexCoord2fColor4ubVertex3fvSUN(\{tc}, \{c}, \{v})");
		try {
		    functions.glTexCoord2fColor4ubVertex3fvSUN().invokeExact(tc, c, v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexCoord2fNormal3fVertex3fSUN">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexCoord2fNormal3fVertex3fSUN.xhtml">registry.khronos.org</a>
	 */
	public void glTexCoord2fNormal3fVertex3fSUN(float s, float t, float nx, float ny, float nz, float x, float y, float z) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexCoord2fNormal3fVertex3fSUN(\{s}, \{t}, \{nx}, \{ny}, \{nz}, \{x}, \{y}, \{z})");
		try {
		    functions.glTexCoord2fNormal3fVertex3fSUN().invokeExact(s, t, nx, ny, nz, x, y, z);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexCoord2fNormal3fVertex3fvSUN">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexCoord2fNormal3fVertex3fvSUN.xhtml">registry.khronos.org</a>
	 */
	public void glTexCoord2fNormal3fVertex3fvSUN(ConstPointer<Float> tc, ConstPointer<Float> n, ConstPointer<Float> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexCoord2fNormal3fVertex3fvSUN(\{tc}, \{n}, \{v})");
		try {
		    functions.glTexCoord2fNormal3fVertex3fvSUN().invokeExact(tc, n, v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexCoord2fVertex3fSUN">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexCoord2fVertex3fSUN.xhtml">registry.khronos.org</a>
	 */
	public void glTexCoord2fVertex3fSUN(float s, float t, float x, float y, float z) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexCoord2fVertex3fSUN(\{s}, \{t}, \{x}, \{y}, \{z})");
		try {
		    functions.glTexCoord2fVertex3fSUN().invokeExact(s, t, x, y, z);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexCoord2fVertex3fvSUN">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexCoord2fVertex3fvSUN.xhtml">registry.khronos.org</a>
	 */
	public void glTexCoord2fVertex3fvSUN(ConstPointer<Float> tc, ConstPointer<Float> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexCoord2fVertex3fvSUN(\{tc}, \{v})");
		try {
		    functions.glTexCoord2fVertex3fvSUN().invokeExact(tc, v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexCoord2fv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexCoord2fv.xhtml">registry.khronos.org</a>
	 */
	public void glTexCoord2fv(ConstPointer<Float> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexCoord2fv(\{v})");
		try {
		    functions.glTexCoord2fv().invokeExact(v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexCoord2hNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexCoord2hNV.xhtml">registry.khronos.org</a>
	 */
	public void glTexCoord2hNV(char s, char t) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexCoord2hNV(\{s}, \{t})");
		try {
		    functions.glTexCoord2hNV().invokeExact(s, t);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexCoord2hvNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexCoord2hvNV.xhtml">registry.khronos.org</a>
	 */
	public void glTexCoord2hvNV(ConstPointer<Character> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexCoord2hvNV(\{v})");
		try {
		    functions.glTexCoord2hvNV().invokeExact(v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexCoord2i">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexCoord2i.xhtml">registry.khronos.org</a>
	 */
	public void glTexCoord2i(int s, int t) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexCoord2i(\{s}, \{t})");
		try {
		    functions.glTexCoord2i().invokeExact(s, t);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexCoord2iv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexCoord2iv.xhtml">registry.khronos.org</a>
	 */
	public void glTexCoord2iv(ConstPointer<Integer> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexCoord2iv(\{v})");
		try {
		    functions.glTexCoord2iv().invokeExact(v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexCoord2s">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexCoord2s.xhtml">registry.khronos.org</a>
	 */
	public void glTexCoord2s(short s, short t) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexCoord2s(\{s}, \{t})");
		try {
		    functions.glTexCoord2s().invokeExact(s, t);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexCoord2sv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexCoord2sv.xhtml">registry.khronos.org</a>
	 */
	public void glTexCoord2sv(ConstPointer<Short> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexCoord2sv(\{v})");
		try {
		    functions.glTexCoord2sv().invokeExact(v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexCoord2xOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexCoord2xOES.xhtml">registry.khronos.org</a>
	 */
	public void glTexCoord2xOES(int s, int t) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexCoord2xOES(\{s}, \{t})");
		try {
		    functions.glTexCoord2xOES().invokeExact(s, t);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexCoord2xvOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexCoord2xvOES.xhtml">registry.khronos.org</a>
	 */
	public void glTexCoord2xvOES(ConstPointer<Integer> coords) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexCoord2xvOES(\{coords})");
		try {
		    functions.glTexCoord2xvOES().invokeExact(coords);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexCoord3bOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexCoord3bOES.xhtml">registry.khronos.org</a>
	 */
	public void glTexCoord3bOES(byte s, byte t, byte r) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexCoord3bOES(\{s}, \{t}, \{r})");
		try {
		    functions.glTexCoord3bOES().invokeExact(s, t, r);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexCoord3bvOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexCoord3bvOES.xhtml">registry.khronos.org</a>
	 */
	public void glTexCoord3bvOES(ConstPointer<Byte> coords) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexCoord3bvOES(\{coords})");
		try {
		    functions.glTexCoord3bvOES().invokeExact(coords);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexCoord3d">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexCoord3d.xhtml">registry.khronos.org</a>
	 */
	public void glTexCoord3d(double s, double t, double r) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexCoord3d(\{s}, \{t}, \{r})");
		try {
		    functions.glTexCoord3d().invokeExact(s, t, r);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexCoord3dv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexCoord3dv.xhtml">registry.khronos.org</a>
	 */
	public void glTexCoord3dv(ConstPointer<Double> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexCoord3dv(\{v})");
		try {
		    functions.glTexCoord3dv().invokeExact(v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexCoord3f">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexCoord3f.xhtml">registry.khronos.org</a>
	 */
	public void glTexCoord3f(float s, float t, float r) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexCoord3f(\{s}, \{t}, \{r})");
		try {
		    functions.glTexCoord3f().invokeExact(s, t, r);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexCoord3fv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexCoord3fv.xhtml">registry.khronos.org</a>
	 */
	public void glTexCoord3fv(ConstPointer<Float> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexCoord3fv(\{v})");
		try {
		    functions.glTexCoord3fv().invokeExact(v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexCoord3hNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexCoord3hNV.xhtml">registry.khronos.org</a>
	 */
	public void glTexCoord3hNV(char s, char t, char r) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexCoord3hNV(\{s}, \{t}, \{r})");
		try {
		    functions.glTexCoord3hNV().invokeExact(s, t, r);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexCoord3hvNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexCoord3hvNV.xhtml">registry.khronos.org</a>
	 */
	public void glTexCoord3hvNV(ConstPointer<Character> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexCoord3hvNV(\{v})");
		try {
		    functions.glTexCoord3hvNV().invokeExact(v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexCoord3i">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexCoord3i.xhtml">registry.khronos.org</a>
	 */
	public void glTexCoord3i(int s, int t, int r) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexCoord3i(\{s}, \{t}, \{r})");
		try {
		    functions.glTexCoord3i().invokeExact(s, t, r);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexCoord3iv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexCoord3iv.xhtml">registry.khronos.org</a>
	 */
	public void glTexCoord3iv(ConstPointer<Integer> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexCoord3iv(\{v})");
		try {
		    functions.glTexCoord3iv().invokeExact(v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexCoord3s">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexCoord3s.xhtml">registry.khronos.org</a>
	 */
	public void glTexCoord3s(short s, short t, short r) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexCoord3s(\{s}, \{t}, \{r})");
		try {
		    functions.glTexCoord3s().invokeExact(s, t, r);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexCoord3sv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexCoord3sv.xhtml">registry.khronos.org</a>
	 */
	public void glTexCoord3sv(ConstPointer<Short> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexCoord3sv(\{v})");
		try {
		    functions.glTexCoord3sv().invokeExact(v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexCoord3xOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexCoord3xOES.xhtml">registry.khronos.org</a>
	 */
	public void glTexCoord3xOES(int s, int t, int r) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexCoord3xOES(\{s}, \{t}, \{r})");
		try {
		    functions.glTexCoord3xOES().invokeExact(s, t, r);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexCoord3xvOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexCoord3xvOES.xhtml">registry.khronos.org</a>
	 */
	public void glTexCoord3xvOES(ConstPointer<Integer> coords) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexCoord3xvOES(\{coords})");
		try {
		    functions.glTexCoord3xvOES().invokeExact(coords);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexCoord4bOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexCoord4bOES.xhtml">registry.khronos.org</a>
	 */
	public void glTexCoord4bOES(byte s, byte t, byte r, byte q) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexCoord4bOES(\{s}, \{t}, \{r}, \{q})");
		try {
		    functions.glTexCoord4bOES().invokeExact(s, t, r, q);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexCoord4bvOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexCoord4bvOES.xhtml">registry.khronos.org</a>
	 */
	public void glTexCoord4bvOES(ConstPointer<Byte> coords) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexCoord4bvOES(\{coords})");
		try {
		    functions.glTexCoord4bvOES().invokeExact(coords);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexCoord4d">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexCoord4d.xhtml">registry.khronos.org</a>
	 */
	public void glTexCoord4d(double s, double t, double r, double q) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexCoord4d(\{s}, \{t}, \{r}, \{q})");
		try {
		    functions.glTexCoord4d().invokeExact(s, t, r, q);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexCoord4dv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexCoord4dv.xhtml">registry.khronos.org</a>
	 */
	public void glTexCoord4dv(ConstPointer<Double> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexCoord4dv(\{v})");
		try {
		    functions.glTexCoord4dv().invokeExact(v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexCoord4f">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexCoord4f.xhtml">registry.khronos.org</a>
	 */
	public void glTexCoord4f(float s, float t, float r, float q) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexCoord4f(\{s}, \{t}, \{r}, \{q})");
		try {
		    functions.glTexCoord4f().invokeExact(s, t, r, q);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexCoord4fColor4fNormal3fVertex4fSUN">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexCoord4fColor4fNormal3fVertex4fSUN.xhtml">registry.khronos.org</a>
	 */
	public void glTexCoord4fColor4fNormal3fVertex4fSUN(float s, float t, float p, float q, float r, float g, float b, float a, float nx, float ny, float nz, float x, float y, float z, float w) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexCoord4fColor4fNormal3fVertex4fSUN(\{s}, \{t}, \{p}, \{q}, \{r}, \{g}, \{b}, \{a}, \{nx}, \{ny}, \{nz}, \{x}, \{y}, \{z}, \{w})");
		try {
		    functions.glTexCoord4fColor4fNormal3fVertex4fSUN().invokeExact(s, t, p, q, r, g, b, a, nx, ny, nz, x, y, z, w);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexCoord4fColor4fNormal3fVertex4fvSUN">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexCoord4fColor4fNormal3fVertex4fvSUN.xhtml">registry.khronos.org</a>
	 */
	public void glTexCoord4fColor4fNormal3fVertex4fvSUN(ConstPointer<Float> tc, ConstPointer<Float> c, ConstPointer<Float> n, ConstPointer<Float> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexCoord4fColor4fNormal3fVertex4fvSUN(\{tc}, \{c}, \{n}, \{v})");
		try {
		    functions.glTexCoord4fColor4fNormal3fVertex4fvSUN().invokeExact(tc, c, n, v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexCoord4fVertex4fSUN">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexCoord4fVertex4fSUN.xhtml">registry.khronos.org</a>
	 */
	public void glTexCoord4fVertex4fSUN(float s, float t, float p, float q, float x, float y, float z, float w) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexCoord4fVertex4fSUN(\{s}, \{t}, \{p}, \{q}, \{x}, \{y}, \{z}, \{w})");
		try {
		    functions.glTexCoord4fVertex4fSUN().invokeExact(s, t, p, q, x, y, z, w);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexCoord4fVertex4fvSUN">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexCoord4fVertex4fvSUN.xhtml">registry.khronos.org</a>
	 */
	public void glTexCoord4fVertex4fvSUN(ConstPointer<Float> tc, ConstPointer<Float> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexCoord4fVertex4fvSUN(\{tc}, \{v})");
		try {
		    functions.glTexCoord4fVertex4fvSUN().invokeExact(tc, v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexCoord4fv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexCoord4fv.xhtml">registry.khronos.org</a>
	 */
	public void glTexCoord4fv(ConstPointer<Float> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexCoord4fv(\{v})");
		try {
		    functions.glTexCoord4fv().invokeExact(v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexCoord4hNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexCoord4hNV.xhtml">registry.khronos.org</a>
	 */
	public void glTexCoord4hNV(char s, char t, char r, char q) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexCoord4hNV(\{s}, \{t}, \{r}, \{q})");
		try {
		    functions.glTexCoord4hNV().invokeExact(s, t, r, q);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexCoord4hvNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexCoord4hvNV.xhtml">registry.khronos.org</a>
	 */
	public void glTexCoord4hvNV(ConstPointer<Character> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexCoord4hvNV(\{v})");
		try {
		    functions.glTexCoord4hvNV().invokeExact(v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexCoord4i">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexCoord4i.xhtml">registry.khronos.org</a>
	 */
	public void glTexCoord4i(int s, int t, int r, int q) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexCoord4i(\{s}, \{t}, \{r}, \{q})");
		try {
		    functions.glTexCoord4i().invokeExact(s, t, r, q);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexCoord4iv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexCoord4iv.xhtml">registry.khronos.org</a>
	 */
	public void glTexCoord4iv(ConstPointer<Integer> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexCoord4iv(\{v})");
		try {
		    functions.glTexCoord4iv().invokeExact(v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexCoord4s">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexCoord4s.xhtml">registry.khronos.org</a>
	 */
	public void glTexCoord4s(short s, short t, short r, short q) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexCoord4s(\{s}, \{t}, \{r}, \{q})");
		try {
		    functions.glTexCoord4s().invokeExact(s, t, r, q);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexCoord4sv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexCoord4sv.xhtml">registry.khronos.org</a>
	 */
	public void glTexCoord4sv(ConstPointer<Short> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexCoord4sv(\{v})");
		try {
		    functions.glTexCoord4sv().invokeExact(v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexCoord4xOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexCoord4xOES.xhtml">registry.khronos.org</a>
	 */
	public void glTexCoord4xOES(int s, int t, int r, int q) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexCoord4xOES(\{s}, \{t}, \{r}, \{q})");
		try {
		    functions.glTexCoord4xOES().invokeExact(s, t, r, q);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexCoord4xvOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexCoord4xvOES.xhtml">registry.khronos.org</a>
	 */
	public void glTexCoord4xvOES(ConstPointer<Integer> coords) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexCoord4xvOES(\{coords})");
		try {
		    functions.glTexCoord4xvOES().invokeExact(coords);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexCoordFormatNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexCoordFormatNV.xhtml">registry.khronos.org</a>
	 */
	public void glTexCoordFormatNV(int size, GLEnum type, int stride) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexCoordFormatNV(\{size}, \{type}, \{stride})");
		try {
		    functions.glTexCoordFormatNV().invokeExact(size, type, stride);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexCoordP1ui">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexCoordP1ui.xhtml">registry.khronos.org</a>
	 */
	public void glTexCoordP1ui(GLTexCoordPointerType type, int coords) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexCoordP1ui(\{type}, \{coords})");
		try {
		    functions.glTexCoordP1ui().invokeExact(type, coords);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexCoordP1uiv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexCoordP1uiv.xhtml">registry.khronos.org</a>
	 */
	public void glTexCoordP1uiv(GLTexCoordPointerType type, ConstPointer<Integer> coords) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexCoordP1uiv(\{type}, \{coords})");
		try {
		    functions.glTexCoordP1uiv().invokeExact(type, coords);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexCoordP2ui">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexCoordP2ui.xhtml">registry.khronos.org</a>
	 */
	public void glTexCoordP2ui(GLTexCoordPointerType type, int coords) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexCoordP2ui(\{type}, \{coords})");
		try {
		    functions.glTexCoordP2ui().invokeExact(type, coords);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexCoordP2uiv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexCoordP2uiv.xhtml">registry.khronos.org</a>
	 */
	public void glTexCoordP2uiv(GLTexCoordPointerType type, ConstPointer<Integer> coords) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexCoordP2uiv(\{type}, \{coords})");
		try {
		    functions.glTexCoordP2uiv().invokeExact(type, coords);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexCoordP3ui">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexCoordP3ui.xhtml">registry.khronos.org</a>
	 */
	public void glTexCoordP3ui(GLTexCoordPointerType type, int coords) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexCoordP3ui(\{type}, \{coords})");
		try {
		    functions.glTexCoordP3ui().invokeExact(type, coords);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexCoordP3uiv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexCoordP3uiv.xhtml">registry.khronos.org</a>
	 */
	public void glTexCoordP3uiv(GLTexCoordPointerType type, ConstPointer<Integer> coords) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexCoordP3uiv(\{type}, \{coords})");
		try {
		    functions.glTexCoordP3uiv().invokeExact(type, coords);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexCoordP4ui">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexCoordP4ui.xhtml">registry.khronos.org</a>
	 */
	public void glTexCoordP4ui(GLTexCoordPointerType type, int coords) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexCoordP4ui(\{type}, \{coords})");
		try {
		    functions.glTexCoordP4ui().invokeExact(type, coords);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexCoordP4uiv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexCoordP4uiv.xhtml">registry.khronos.org</a>
	 */
	public void glTexCoordP4uiv(GLTexCoordPointerType type, ConstPointer<Integer> coords) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexCoordP4uiv(\{type}, \{coords})");
		try {
		    functions.glTexCoordP4uiv().invokeExact(type, coords);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexCoordPointer">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexCoordPointer.xhtml">registry.khronos.org</a>
	 */
	public void glTexCoordPointer(int size, GLTexCoordPointerType type, int stride, ConstPointer<Void> pointer) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexCoordPointer(\{size}, \{type}, \{stride}, \{pointer})");
		try {
		    functions.glTexCoordPointer().invokeExact(size, type, stride, pointer);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexCoordPointerEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexCoordPointerEXT.xhtml">registry.khronos.org</a>
	 */
	public void glTexCoordPointerEXT(int size, GLTexCoordPointerType type, int stride, int count, ConstPointer<Void> pointer) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexCoordPointerEXT(\{size}, \{type}, \{stride}, \{count}, \{pointer})");
		try {
		    functions.glTexCoordPointerEXT().invokeExact(size, type, stride, count, pointer);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexCoordPointerListIBM">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexCoordPointerListIBM.xhtml">registry.khronos.org</a>
	 */
	public void glTexCoordPointerListIBM(int size, GLTexCoordPointerType type, int stride, MutPointer<ConstPointer<Void>> pointer, int ptrstride) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexCoordPointerListIBM(\{size}, \{type}, \{stride}, \{pointer}, \{ptrstride})");
		try {
		    functions.glTexCoordPointerListIBM().invokeExact(size, type, stride, pointer, ptrstride);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexCoordPointervINTEL">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexCoordPointervINTEL.xhtml">registry.khronos.org</a>
	 */
	public void glTexCoordPointervINTEL(int size, GLVertexPointerType type, MutPointer<ConstPointer<Void>> pointer) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexCoordPointervINTEL(\{size}, \{type}, \{pointer})");
		try {
		    functions.glTexCoordPointervINTEL().invokeExact(size, type, pointer);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexEnvf">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexEnvf.xhtml">registry.khronos.org</a>
	 */
	public void glTexEnvf(GLTextureEnvTarget target, GLTextureEnvParameter pname, float param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexEnvf(\{target}, \{pname}, \{param})");
		try {
		    functions.glTexEnvf().invokeExact(target, pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexEnvfv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexEnvfv.xhtml">registry.khronos.org</a>
	 */
	public void glTexEnvfv(GLTextureEnvTarget target, GLTextureEnvParameter pname, ConstPointer<Float> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexEnvfv(\{target}, \{pname}, \{params})");
		try {
		    functions.glTexEnvfv().invokeExact(target, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexEnvi">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexEnvi.xhtml">registry.khronos.org</a>
	 */
	public void glTexEnvi(GLTextureEnvTarget target, GLTextureEnvParameter pname, int param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexEnvi(\{target}, \{pname}, \{param})");
		try {
		    functions.glTexEnvi().invokeExact(target, pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexEnviv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexEnviv.xhtml">registry.khronos.org</a>
	 */
	public void glTexEnviv(GLTextureEnvTarget target, GLTextureEnvParameter pname, ConstPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexEnviv(\{target}, \{pname}, \{params})");
		try {
		    functions.glTexEnviv().invokeExact(target, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexEnvx">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexEnvx.xhtml">registry.khronos.org</a>
	 */
	public void glTexEnvx(GLTextureEnvTarget target, GLTextureEnvParameter pname, int param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexEnvx(\{target}, \{pname}, \{param})");
		try {
		    functions.glTexEnvx().invokeExact(target, pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexEnvxOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexEnvxOES.xhtml">registry.khronos.org</a>
	 */
	public void glTexEnvxOES(GLTextureEnvTarget target, GLTextureEnvParameter pname, int param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexEnvxOES(\{target}, \{pname}, \{param})");
		try {
		    functions.glTexEnvxOES().invokeExact(target, pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexEnvxv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexEnvxv.xhtml">registry.khronos.org</a>
	 */
	public void glTexEnvxv(GLTextureEnvTarget target, GLTextureEnvParameter pname, ConstPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexEnvxv(\{target}, \{pname}, \{params})");
		try {
		    functions.glTexEnvxv().invokeExact(target, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexEnvxvOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexEnvxvOES.xhtml">registry.khronos.org</a>
	 */
	public void glTexEnvxvOES(GLTextureEnvTarget target, GLTextureEnvParameter pname, ConstPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexEnvxvOES(\{target}, \{pname}, \{params})");
		try {
		    functions.glTexEnvxvOES().invokeExact(target, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexEstimateMotionQCOM">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexEstimateMotionQCOM.xhtml">registry.khronos.org</a>
	 */
	public void glTexEstimateMotionQCOM(int ref, int target, int output) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexEstimateMotionQCOM(\{ref}, \{target}, \{output})");
		try {
		    functions.glTexEstimateMotionQCOM().invokeExact(ref, target, output);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexEstimateMotionRegionsQCOM">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexEstimateMotionRegionsQCOM.xhtml">registry.khronos.org</a>
	 */
	public void glTexEstimateMotionRegionsQCOM(int ref, int target, int output, int mask) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexEstimateMotionRegionsQCOM(\{ref}, \{target}, \{output}, \{mask})");
		try {
		    functions.glTexEstimateMotionRegionsQCOM().invokeExact(ref, target, output, mask);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glExtrapolateTex2DQCOM">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glExtrapolateTex2DQCOM.xhtml">registry.khronos.org</a>
	 */
	public void glExtrapolateTex2DQCOM(int src1, int src2, int output, float scaleFactor) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glExtrapolateTex2DQCOM(\{src1}, \{src2}, \{output}, \{scaleFactor})");
		try {
		    functions.glExtrapolateTex2DQCOM().invokeExact(src1, src2, output, scaleFactor);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexFilterFuncSGIS">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexFilterFuncSGIS.xhtml">registry.khronos.org</a>
	 */
	public void glTexFilterFuncSGIS(GLTextureTarget target, GLTextureFilterSGIS filter, int n, ConstPointer<Float> weights) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexFilterFuncSGIS(\{target}, \{filter}, \{n}, \{weights})");
		try {
		    functions.glTexFilterFuncSGIS().invokeExact(target, filter, n, weights);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, filter=\{filter}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexGend">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexGend.xhtml">registry.khronos.org</a>
	 */
	public void glTexGend(GLTextureCoordName coord, GLTextureGenParameter pname, double param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexGend(\{coord}, \{pname}, \{param})");
		try {
		    functions.glTexGend().invokeExact(coord, pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were coord=\{coord}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexGendv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexGendv.xhtml">registry.khronos.org</a>
	 */
	public void glTexGendv(GLTextureCoordName coord, GLTextureGenParameter pname, ConstPointer<Double> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexGendv(\{coord}, \{pname}, \{params})");
		try {
		    functions.glTexGendv().invokeExact(coord, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were coord=\{coord}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexGenf">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexGenf.xhtml">registry.khronos.org</a>
	 */
	public void glTexGenf(GLTextureCoordName coord, GLTextureGenParameter pname, float param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexGenf(\{coord}, \{pname}, \{param})");
		try {
		    functions.glTexGenf().invokeExact(coord, pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were coord=\{coord}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexGenfOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexGenfOES.xhtml">registry.khronos.org</a>
	 */
	public void glTexGenfOES(GLTextureCoordName coord, GLTextureGenParameter pname, float param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexGenfOES(\{coord}, \{pname}, \{param})");
		try {
		    functions.glTexGenfOES().invokeExact(coord, pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were coord=\{coord}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexGenfv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexGenfv.xhtml">registry.khronos.org</a>
	 */
	public void glTexGenfv(GLTextureCoordName coord, GLTextureGenParameter pname, ConstPointer<Float> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexGenfv(\{coord}, \{pname}, \{params})");
		try {
		    functions.glTexGenfv().invokeExact(coord, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were coord=\{coord}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexGenfvOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexGenfvOES.xhtml">registry.khronos.org</a>
	 */
	public void glTexGenfvOES(GLTextureCoordName coord, GLTextureGenParameter pname, ConstPointer<Float> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexGenfvOES(\{coord}, \{pname}, \{params})");
		try {
		    functions.glTexGenfvOES().invokeExact(coord, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were coord=\{coord}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexGeni">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexGeni.xhtml">registry.khronos.org</a>
	 */
	public void glTexGeni(GLTextureCoordName coord, GLTextureGenParameter pname, int param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexGeni(\{coord}, \{pname}, \{param})");
		try {
		    functions.glTexGeni().invokeExact(coord, pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were coord=\{coord}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexGeniOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexGeniOES.xhtml">registry.khronos.org</a>
	 */
	public void glTexGeniOES(GLTextureCoordName coord, GLTextureGenParameter pname, int param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexGeniOES(\{coord}, \{pname}, \{param})");
		try {
		    functions.glTexGeniOES().invokeExact(coord, pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were coord=\{coord}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexGeniv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexGeniv.xhtml">registry.khronos.org</a>
	 */
	public void glTexGeniv(GLTextureCoordName coord, GLTextureGenParameter pname, ConstPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexGeniv(\{coord}, \{pname}, \{params})");
		try {
		    functions.glTexGeniv().invokeExact(coord, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were coord=\{coord}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexGenivOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexGenivOES.xhtml">registry.khronos.org</a>
	 */
	public void glTexGenivOES(GLTextureCoordName coord, GLTextureGenParameter pname, ConstPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexGenivOES(\{coord}, \{pname}, \{params})");
		try {
		    functions.glTexGenivOES().invokeExact(coord, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were coord=\{coord}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexGenxOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexGenxOES.xhtml">registry.khronos.org</a>
	 */
	public void glTexGenxOES(GLTextureCoordName coord, GLTextureGenParameter pname, int param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexGenxOES(\{coord}, \{pname}, \{param})");
		try {
		    functions.glTexGenxOES().invokeExact(coord, pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were coord=\{coord}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexGenxvOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexGenxvOES.xhtml">registry.khronos.org</a>
	 */
	public void glTexGenxvOES(GLTextureCoordName coord, GLTextureGenParameter pname, ConstPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexGenxvOES(\{coord}, \{pname}, \{params})");
		try {
		    functions.glTexGenxvOES().invokeExact(coord, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were coord=\{coord}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexImage1D">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexImage1D.xhtml">registry.khronos.org</a>
	 */
	public void glTexImage1D(GLTextureTarget target, int level, GLInternalFormat internalformat, int width, int border, GLPixelFormat format, GLPixelType type, ConstPointer<Void> pixels) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexImage1D(\{target}, \{level}, \{internalformat}, \{width}, \{border}, \{format}, \{type}, \{pixels})");
		try {
		    functions.glTexImage1D().invokeExact(target, level, internalformat, width, border, format, type, pixels);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, internalformat=\{internalformat}, format=\{format}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexImage2D">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexImage2D.xhtml">registry.khronos.org</a>
	 */
	public void glTexImage2D(GLTextureTarget target, int level, GLInternalFormat internalformat, int width, int height, int border, GLPixelFormat format, GLPixelType type, ConstPointer<Void> pixels) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexImage2D(\{target}, \{level}, \{internalformat}, \{width}, \{height}, \{border}, \{format}, \{type}, \{pixels})");
		try {
		    functions.glTexImage2D().invokeExact(target, level, internalformat, width, height, border, format, type, pixels);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, internalformat=\{internalformat}, format=\{format}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexImage2DMultisample">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexImage2DMultisample.xhtml">registry.khronos.org</a>
	 */
	public void glTexImage2DMultisample(GLTextureTarget target, int samples, GLInternalFormat internalformat, int width, int height, boolean fixedsamplelocations) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexImage2DMultisample(\{target}, \{samples}, \{internalformat}, \{width}, \{height}, \{fixedsamplelocations})");
		try {
		    functions.glTexImage2DMultisample().invokeExact(target, samples, internalformat, width, height, fixedsamplelocations);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, internalformat=\{internalformat}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexImage2DMultisampleCoverageNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexImage2DMultisampleCoverageNV.xhtml">registry.khronos.org</a>
	 */
	public void glTexImage2DMultisampleCoverageNV(GLTextureTarget target, int coverageSamples, int colorSamples, GLInternalFormat internalFormat, int width, int height, boolean fixedSampleLocations) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexImage2DMultisampleCoverageNV(\{target}, \{coverageSamples}, \{colorSamples}, \{internalFormat}, \{width}, \{height}, \{fixedSampleLocations})");
		try {
		    functions.glTexImage2DMultisampleCoverageNV().invokeExact(target, coverageSamples, colorSamples, internalFormat, width, height, fixedSampleLocations);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, internalFormat=\{internalFormat}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexImage3D">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexImage3D.xhtml">registry.khronos.org</a>
	 */
	public void glTexImage3D(GLTextureTarget target, int level, GLInternalFormat internalformat, int width, int height, int depth, int border, GLPixelFormat format, GLPixelType type, ConstPointer<Void> pixels) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexImage3D(\{target}, \{level}, \{internalformat}, \{width}, \{height}, \{depth}, \{border}, \{format}, \{type}, \{pixels})");
		try {
		    functions.glTexImage3D().invokeExact(target, level, internalformat, width, height, depth, border, format, type, pixels);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, internalformat=\{internalformat}, format=\{format}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexImage3DEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexImage3DEXT.xhtml">registry.khronos.org</a>
	 */
	public void glTexImage3DEXT(GLTextureTarget target, int level, GLInternalFormat internalformat, int width, int height, int depth, int border, GLPixelFormat format, GLPixelType type, ConstPointer<Void> pixels) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexImage3DEXT(\{target}, \{level}, \{internalformat}, \{width}, \{height}, \{depth}, \{border}, \{format}, \{type}, \{pixels})");
		try {
		    functions.glTexImage3DEXT().invokeExact(target, level, internalformat, width, height, depth, border, format, type, pixels);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, internalformat=\{internalformat}, format=\{format}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexImage3DMultisample">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexImage3DMultisample.xhtml">registry.khronos.org</a>
	 */
	public void glTexImage3DMultisample(GLTextureTarget target, int samples, GLInternalFormat internalformat, int width, int height, int depth, boolean fixedsamplelocations) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexImage3DMultisample(\{target}, \{samples}, \{internalformat}, \{width}, \{height}, \{depth}, \{fixedsamplelocations})");
		try {
		    functions.glTexImage3DMultisample().invokeExact(target, samples, internalformat, width, height, depth, fixedsamplelocations);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, internalformat=\{internalformat}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexImage3DMultisampleCoverageNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexImage3DMultisampleCoverageNV.xhtml">registry.khronos.org</a>
	 */
	public void glTexImage3DMultisampleCoverageNV(GLTextureTarget target, int coverageSamples, int colorSamples, GLInternalFormat internalFormat, int width, int height, int depth, boolean fixedSampleLocations) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexImage3DMultisampleCoverageNV(\{target}, \{coverageSamples}, \{colorSamples}, \{internalFormat}, \{width}, \{height}, \{depth}, \{fixedSampleLocations})");
		try {
		    functions.glTexImage3DMultisampleCoverageNV().invokeExact(target, coverageSamples, colorSamples, internalFormat, width, height, depth, fixedSampleLocations);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, internalFormat=\{internalFormat}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexImage3DOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexImage3DOES.xhtml">registry.khronos.org</a>
	 */
	public void glTexImage3DOES(GLTextureTarget target, int level, GLInternalFormat internalformat, int width, int height, int depth, int border, GLPixelFormat format, GLPixelType type, ConstPointer<Void> pixels) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexImage3DOES(\{target}, \{level}, \{internalformat}, \{width}, \{height}, \{depth}, \{border}, \{format}, \{type}, \{pixels})");
		try {
		    functions.glTexImage3DOES().invokeExact(target, level, internalformat, width, height, depth, border, format, type, pixels);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, internalformat=\{internalformat}, format=\{format}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexImage4DSGIS">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexImage4DSGIS.xhtml">registry.khronos.org</a>
	 */
	public void glTexImage4DSGIS(GLTextureTarget target, int level, GLInternalFormat internalformat, int width, int height, int depth, int size4d, int border, GLPixelFormat format, GLPixelType type, ConstPointer<Void> pixels) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexImage4DSGIS(\{target}, \{level}, \{internalformat}, \{width}, \{height}, \{depth}, \{size4d}, \{border}, \{format}, \{type}, \{pixels})");
		try {
		    functions.glTexImage4DSGIS().invokeExact(target, level, internalformat, width, height, depth, size4d, border, format, type, pixels);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, internalformat=\{internalformat}, format=\{format}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexPageCommitmentARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexPageCommitmentARB.xhtml">registry.khronos.org</a>
	 */
	public void glTexPageCommitmentARB(GLEnum target, int level, int xoffset, int yoffset, int zoffset, int width, int height, int depth, boolean commit) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexPageCommitmentARB(\{target}, \{level}, \{xoffset}, \{yoffset}, \{zoffset}, \{width}, \{height}, \{depth}, \{commit})");
		try {
		    functions.glTexPageCommitmentARB().invokeExact(target, level, xoffset, yoffset, zoffset, width, height, depth, commit);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexPageCommitmentEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexPageCommitmentEXT.xhtml">registry.khronos.org</a>
	 */
	public void glTexPageCommitmentEXT(GLEnum target, int level, int xoffset, int yoffset, int zoffset, int width, int height, int depth, boolean commit) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexPageCommitmentEXT(\{target}, \{level}, \{xoffset}, \{yoffset}, \{zoffset}, \{width}, \{height}, \{depth}, \{commit})");
		try {
		    functions.glTexPageCommitmentEXT().invokeExact(target, level, xoffset, yoffset, zoffset, width, height, depth, commit);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexPageCommitmentMemNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexPageCommitmentMemNV.xhtml">registry.khronos.org</a>
	 */
	public void glTexPageCommitmentMemNV(GLTextureTarget target, int layer, int level, int xoffset, int yoffset, int zoffset, int width, int height, int depth, int memory, long offset, boolean commit) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexPageCommitmentMemNV(\{target}, \{layer}, \{level}, \{xoffset}, \{yoffset}, \{zoffset}, \{width}, \{height}, \{depth}, \{memory}, \{offset}, \{commit})");
		try {
		    functions.glTexPageCommitmentMemNV().invokeExact(target, layer, level, xoffset, yoffset, zoffset, width, height, depth, memory, offset, commit);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexParameterIiv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexParameterIiv.xhtml">registry.khronos.org</a>
	 */
	public void glTexParameterIiv(GLTextureTarget target, GLTextureParameterName pname, ConstPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexParameterIiv(\{target}, \{pname}, \{params})");
		try {
		    functions.glTexParameterIiv().invokeExact(target, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexParameterIivEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexParameterIivEXT.xhtml">registry.khronos.org</a>
	 */
	public void glTexParameterIivEXT(GLTextureTarget target, GLTextureParameterName pname, ConstPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexParameterIivEXT(\{target}, \{pname}, \{params})");
		try {
		    functions.glTexParameterIivEXT().invokeExact(target, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexParameterIivOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexParameterIivOES.xhtml">registry.khronos.org</a>
	 */
	public void glTexParameterIivOES(GLTextureTarget target, GLTextureParameterName pname, ConstPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexParameterIivOES(\{target}, \{pname}, \{params})");
		try {
		    functions.glTexParameterIivOES().invokeExact(target, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexParameterIuiv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexParameterIuiv.xhtml">registry.khronos.org</a>
	 */
	public void glTexParameterIuiv(GLTextureTarget target, GLTextureParameterName pname, ConstPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexParameterIuiv(\{target}, \{pname}, \{params})");
		try {
		    functions.glTexParameterIuiv().invokeExact(target, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexParameterIuivEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexParameterIuivEXT.xhtml">registry.khronos.org</a>
	 */
	public void glTexParameterIuivEXT(GLTextureTarget target, GLTextureParameterName pname, ConstPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexParameterIuivEXT(\{target}, \{pname}, \{params})");
		try {
		    functions.glTexParameterIuivEXT().invokeExact(target, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexParameterIuivOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexParameterIuivOES.xhtml">registry.khronos.org</a>
	 */
	public void glTexParameterIuivOES(GLTextureTarget target, GLTextureParameterName pname, ConstPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexParameterIuivOES(\{target}, \{pname}, \{params})");
		try {
		    functions.glTexParameterIuivOES().invokeExact(target, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexParameterf">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexParameterf.xhtml">registry.khronos.org</a>
	 */
	public void glTexParameterf(GLTextureTarget target, GLTextureParameterName pname, float param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexParameterf(\{target}, \{pname}, \{param})");
		try {
		    functions.glTexParameterf().invokeExact(target, pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexParameterfv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexParameterfv.xhtml">registry.khronos.org</a>
	 */
	public void glTexParameterfv(GLTextureTarget target, GLTextureParameterName pname, ConstPointer<Float> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexParameterfv(\{target}, \{pname}, \{params})");
		try {
		    functions.glTexParameterfv().invokeExact(target, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexParameteri">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexParameteri.xhtml">registry.khronos.org</a>
	 */
	public void glTexParameteri(GLTextureTarget target, GLTextureParameterName pname, int param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexParameteri(\{target}, \{pname}, \{param})");
		try {
		    functions.glTexParameteri().invokeExact(target, pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexParameteriv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexParameteriv.xhtml">registry.khronos.org</a>
	 */
	public void glTexParameteriv(GLTextureTarget target, GLTextureParameterName pname, ConstPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexParameteriv(\{target}, \{pname}, \{params})");
		try {
		    functions.glTexParameteriv().invokeExact(target, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexParameterx">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexParameterx.xhtml">registry.khronos.org</a>
	 */
	public void glTexParameterx(GLTextureTarget target, GLGetTextureParameter pname, int param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexParameterx(\{target}, \{pname}, \{param})");
		try {
		    functions.glTexParameterx().invokeExact(target, pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexParameterxOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexParameterxOES.xhtml">registry.khronos.org</a>
	 */
	public void glTexParameterxOES(GLTextureTarget target, GLGetTextureParameter pname, int param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexParameterxOES(\{target}, \{pname}, \{param})");
		try {
		    functions.glTexParameterxOES().invokeExact(target, pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexParameterxv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexParameterxv.xhtml">registry.khronos.org</a>
	 */
	public void glTexParameterxv(GLTextureTarget target, GLGetTextureParameter pname, ConstPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexParameterxv(\{target}, \{pname}, \{params})");
		try {
		    functions.glTexParameterxv().invokeExact(target, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexParameterxvOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexParameterxvOES.xhtml">registry.khronos.org</a>
	 */
	public void glTexParameterxvOES(GLTextureTarget target, GLGetTextureParameter pname, ConstPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexParameterxvOES(\{target}, \{pname}, \{params})");
		try {
		    functions.glTexParameterxvOES().invokeExact(target, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexRenderbufferNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexRenderbufferNV.xhtml">registry.khronos.org</a>
	 */
	public void glTexRenderbufferNV(GLTextureTarget target, int renderbuffer) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexRenderbufferNV(\{target}, \{renderbuffer})");
		try {
		    functions.glTexRenderbufferNV().invokeExact(target, renderbuffer);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexStorage1D">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexStorage1D.xhtml">registry.khronos.org</a>
	 */
	public void glTexStorage1D(GLTextureTarget target, int levels, GLSizedInternalFormat internalformat, int width) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexStorage1D(\{target}, \{levels}, \{internalformat}, \{width})");
		try {
		    functions.glTexStorage1D().invokeExact(target, levels, internalformat, width);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, internalformat=\{internalformat}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexStorage1DEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexStorage1DEXT.xhtml">registry.khronos.org</a>
	 */
	public void glTexStorage1DEXT(GLTextureTarget target, int levels, GLSizedInternalFormat internalformat, int width) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexStorage1DEXT(\{target}, \{levels}, \{internalformat}, \{width})");
		try {
		    functions.glTexStorage1DEXT().invokeExact(target, levels, internalformat, width);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, internalformat=\{internalformat}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexStorage2D">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexStorage2D.xhtml">registry.khronos.org</a>
	 */
	public void glTexStorage2D(GLTextureTarget target, int levels, GLSizedInternalFormat internalformat, int width, int height) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexStorage2D(\{target}, \{levels}, \{internalformat}, \{width}, \{height})");
		try {
		    functions.glTexStorage2D().invokeExact(target, levels, internalformat, width, height);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, internalformat=\{internalformat}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexStorage2DEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexStorage2DEXT.xhtml">registry.khronos.org</a>
	 */
	public void glTexStorage2DEXT(GLTextureTarget target, int levels, GLSizedInternalFormat internalformat, int width, int height) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexStorage2DEXT(\{target}, \{levels}, \{internalformat}, \{width}, \{height})");
		try {
		    functions.glTexStorage2DEXT().invokeExact(target, levels, internalformat, width, height);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, internalformat=\{internalformat}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexStorage2DMultisample">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexStorage2DMultisample.xhtml">registry.khronos.org</a>
	 */
	public void glTexStorage2DMultisample(GLTextureTarget target, int samples, GLSizedInternalFormat internalformat, int width, int height, boolean fixedsamplelocations) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexStorage2DMultisample(\{target}, \{samples}, \{internalformat}, \{width}, \{height}, \{fixedsamplelocations})");
		try {
		    functions.glTexStorage2DMultisample().invokeExact(target, samples, internalformat, width, height, fixedsamplelocations);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, internalformat=\{internalformat}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexStorage3D">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexStorage3D.xhtml">registry.khronos.org</a>
	 */
	public void glTexStorage3D(GLTextureTarget target, int levels, GLSizedInternalFormat internalformat, int width, int height, int depth) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexStorage3D(\{target}, \{levels}, \{internalformat}, \{width}, \{height}, \{depth})");
		try {
		    functions.glTexStorage3D().invokeExact(target, levels, internalformat, width, height, depth);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, internalformat=\{internalformat}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexStorage3DEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexStorage3DEXT.xhtml">registry.khronos.org</a>
	 */
	public void glTexStorage3DEXT(GLTextureTarget target, int levels, GLSizedInternalFormat internalformat, int width, int height, int depth) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexStorage3DEXT(\{target}, \{levels}, \{internalformat}, \{width}, \{height}, \{depth})");
		try {
		    functions.glTexStorage3DEXT().invokeExact(target, levels, internalformat, width, height, depth);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, internalformat=\{internalformat}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexStorage3DMultisample">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexStorage3DMultisample.xhtml">registry.khronos.org</a>
	 */
	public void glTexStorage3DMultisample(GLTextureTarget target, int samples, GLSizedInternalFormat internalformat, int width, int height, int depth, boolean fixedsamplelocations) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexStorage3DMultisample(\{target}, \{samples}, \{internalformat}, \{width}, \{height}, \{depth}, \{fixedsamplelocations})");
		try {
		    functions.glTexStorage3DMultisample().invokeExact(target, samples, internalformat, width, height, depth, fixedsamplelocations);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, internalformat=\{internalformat}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexStorage3DMultisampleOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexStorage3DMultisampleOES.xhtml">registry.khronos.org</a>
	 */
	public void glTexStorage3DMultisampleOES(GLTextureTarget target, int samples, GLSizedInternalFormat internalformat, int width, int height, int depth, boolean fixedsamplelocations) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexStorage3DMultisampleOES(\{target}, \{samples}, \{internalformat}, \{width}, \{height}, \{depth}, \{fixedsamplelocations})");
		try {
		    functions.glTexStorage3DMultisampleOES().invokeExact(target, samples, internalformat, width, height, depth, fixedsamplelocations);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, internalformat=\{internalformat}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexStorageAttribs2DEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexStorageAttribs2DEXT.xhtml">registry.khronos.org</a>
	 */
	public void glTexStorageAttribs2DEXT(GLTextureTarget target, int levels, GLSizedInternalFormat internalformat, int width, int height, ConstPointer<GLTexStorageAttribs> attrib_list) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexStorageAttribs2DEXT(\{target}, \{levels}, \{internalformat}, \{width}, \{height}, \{attrib_list})");
		try {
		    functions.glTexStorageAttribs2DEXT().invokeExact(target, levels, internalformat, width, height, attrib_list);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, internalformat=\{internalformat}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexStorageAttribs3DEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexStorageAttribs3DEXT.xhtml">registry.khronos.org</a>
	 */
	public void glTexStorageAttribs3DEXT(GLTextureTarget target, int levels, GLSizedInternalFormat internalformat, int width, int height, int depth, ConstPointer<GLTexStorageAttribs> attrib_list) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexStorageAttribs3DEXT(\{target}, \{levels}, \{internalformat}, \{width}, \{height}, \{depth}, \{attrib_list})");
		try {
		    functions.glTexStorageAttribs3DEXT().invokeExact(target, levels, internalformat, width, height, depth, attrib_list);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, internalformat=\{internalformat}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexStorageMem1DEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexStorageMem1DEXT.xhtml">registry.khronos.org</a>
	 */
	public void glTexStorageMem1DEXT(GLTextureTarget target, int levels, GLSizedInternalFormat internalFormat, int width, int memory, long offset) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexStorageMem1DEXT(\{target}, \{levels}, \{internalFormat}, \{width}, \{memory}, \{offset})");
		try {
		    functions.glTexStorageMem1DEXT().invokeExact(target, levels, internalFormat, width, memory, offset);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, internalFormat=\{internalFormat}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexStorageMem2DEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexStorageMem2DEXT.xhtml">registry.khronos.org</a>
	 */
	public void glTexStorageMem2DEXT(GLTextureTarget target, int levels, GLSizedInternalFormat internalFormat, int width, int height, int memory, long offset) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexStorageMem2DEXT(\{target}, \{levels}, \{internalFormat}, \{width}, \{height}, \{memory}, \{offset})");
		try {
		    functions.glTexStorageMem2DEXT().invokeExact(target, levels, internalFormat, width, height, memory, offset);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, internalFormat=\{internalFormat}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexStorageMem2DMultisampleEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexStorageMem2DMultisampleEXT.xhtml">registry.khronos.org</a>
	 */
	public void glTexStorageMem2DMultisampleEXT(GLTextureTarget target, int samples, GLSizedInternalFormat internalFormat, int width, int height, boolean fixedSampleLocations, int memory, long offset) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexStorageMem2DMultisampleEXT(\{target}, \{samples}, \{internalFormat}, \{width}, \{height}, \{fixedSampleLocations}, \{memory}, \{offset})");
		try {
		    functions.glTexStorageMem2DMultisampleEXT().invokeExact(target, samples, internalFormat, width, height, fixedSampleLocations, memory, offset);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, internalFormat=\{internalFormat}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexStorageMem3DEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexStorageMem3DEXT.xhtml">registry.khronos.org</a>
	 */
	public void glTexStorageMem3DEXT(GLTextureTarget target, int levels, GLSizedInternalFormat internalFormat, int width, int height, int depth, int memory, long offset) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexStorageMem3DEXT(\{target}, \{levels}, \{internalFormat}, \{width}, \{height}, \{depth}, \{memory}, \{offset})");
		try {
		    functions.glTexStorageMem3DEXT().invokeExact(target, levels, internalFormat, width, height, depth, memory, offset);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, internalFormat=\{internalFormat}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexStorageMem3DMultisampleEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexStorageMem3DMultisampleEXT.xhtml">registry.khronos.org</a>
	 */
	public void glTexStorageMem3DMultisampleEXT(GLTextureTarget target, int samples, GLSizedInternalFormat internalFormat, int width, int height, int depth, boolean fixedSampleLocations, int memory, long offset) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexStorageMem3DMultisampleEXT(\{target}, \{samples}, \{internalFormat}, \{width}, \{height}, \{depth}, \{fixedSampleLocations}, \{memory}, \{offset})");
		try {
		    functions.glTexStorageMem3DMultisampleEXT().invokeExact(target, samples, internalFormat, width, height, depth, fixedSampleLocations, memory, offset);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, internalFormat=\{internalFormat}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexStorageSparseAMD">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexStorageSparseAMD.xhtml">registry.khronos.org</a>
	 */
	public void glTexStorageSparseAMD(GLTextureTarget target, GLSizedInternalFormat internalFormat, int width, int height, int depth, int layers, int flags) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexStorageSparseAMD(\{target}, \{internalFormat}, \{width}, \{height}, \{depth}, \{layers}, \{flags})");
		try {
		    functions.glTexStorageSparseAMD().invokeExact(target, internalFormat, width, height, depth, layers, flags);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, internalFormat=\{internalFormat}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexSubImage1D">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexSubImage1D.xhtml">registry.khronos.org</a>
	 */
	public void glTexSubImage1D(GLTextureTarget target, int level, int xoffset, int width, GLPixelFormat format, GLPixelType type, ConstPointer<Void> pixels) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexSubImage1D(\{target}, \{level}, \{xoffset}, \{width}, \{format}, \{type}, \{pixels})");
		try {
		    functions.glTexSubImage1D().invokeExact(target, level, xoffset, width, format, type, pixels);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, format=\{format}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexSubImage1DEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexSubImage1DEXT.xhtml">registry.khronos.org</a>
	 */
	public void glTexSubImage1DEXT(GLTextureTarget target, int level, int xoffset, int width, GLPixelFormat format, GLPixelType type, ConstPointer<Void> pixels) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexSubImage1DEXT(\{target}, \{level}, \{xoffset}, \{width}, \{format}, \{type}, \{pixels})");
		try {
		    functions.glTexSubImage1DEXT().invokeExact(target, level, xoffset, width, format, type, pixels);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, format=\{format}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexSubImage2D">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexSubImage2D.xhtml">registry.khronos.org</a>
	 */
	public void glTexSubImage2D(GLTextureTarget target, int level, int xoffset, int yoffset, int width, int height, GLPixelFormat format, GLPixelType type, ConstPointer<Void> pixels) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexSubImage2D(\{target}, \{level}, \{xoffset}, \{yoffset}, \{width}, \{height}, \{format}, \{type}, \{pixels})");
		try {
		    functions.glTexSubImage2D().invokeExact(target, level, xoffset, yoffset, width, height, format, type, pixels);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, format=\{format}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexSubImage2DEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexSubImage2DEXT.xhtml">registry.khronos.org</a>
	 */
	public void glTexSubImage2DEXT(GLTextureTarget target, int level, int xoffset, int yoffset, int width, int height, GLPixelFormat format, GLPixelType type, ConstPointer<Void> pixels) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexSubImage2DEXT(\{target}, \{level}, \{xoffset}, \{yoffset}, \{width}, \{height}, \{format}, \{type}, \{pixels})");
		try {
		    functions.glTexSubImage2DEXT().invokeExact(target, level, xoffset, yoffset, width, height, format, type, pixels);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, format=\{format}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexSubImage3D">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexSubImage3D.xhtml">registry.khronos.org</a>
	 */
	public void glTexSubImage3D(GLTextureTarget target, int level, int xoffset, int yoffset, int zoffset, int width, int height, int depth, GLPixelFormat format, GLPixelType type, ConstPointer<Void> pixels) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexSubImage3D(\{target}, \{level}, \{xoffset}, \{yoffset}, \{zoffset}, \{width}, \{height}, \{depth}, \{format}, \{type}, \{pixels})");
		try {
		    functions.glTexSubImage3D().invokeExact(target, level, xoffset, yoffset, zoffset, width, height, depth, format, type, pixels);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, format=\{format}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexSubImage3DEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexSubImage3DEXT.xhtml">registry.khronos.org</a>
	 */
	public void glTexSubImage3DEXT(GLTextureTarget target, int level, int xoffset, int yoffset, int zoffset, int width, int height, int depth, GLPixelFormat format, GLPixelType type, ConstPointer<Void> pixels) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexSubImage3DEXT(\{target}, \{level}, \{xoffset}, \{yoffset}, \{zoffset}, \{width}, \{height}, \{depth}, \{format}, \{type}, \{pixels})");
		try {
		    functions.glTexSubImage3DEXT().invokeExact(target, level, xoffset, yoffset, zoffset, width, height, depth, format, type, pixels);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, format=\{format}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexSubImage3DOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexSubImage3DOES.xhtml">registry.khronos.org</a>
	 */
	public void glTexSubImage3DOES(GLTextureTarget target, int level, int xoffset, int yoffset, int zoffset, int width, int height, int depth, GLPixelFormat format, GLPixelType type, ConstPointer<Void> pixels) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexSubImage3DOES(\{target}, \{level}, \{xoffset}, \{yoffset}, \{zoffset}, \{width}, \{height}, \{depth}, \{format}, \{type}, \{pixels})");
		try {
		    functions.glTexSubImage3DOES().invokeExact(target, level, xoffset, yoffset, zoffset, width, height, depth, format, type, pixels);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, format=\{format}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexSubImage4DSGIS">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexSubImage4DSGIS.xhtml">registry.khronos.org</a>
	 */
	public void glTexSubImage4DSGIS(GLTextureTarget target, int level, int xoffset, int yoffset, int zoffset, int woffset, int width, int height, int depth, int size4d, GLPixelFormat format, GLPixelType type, ConstPointer<Void> pixels) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexSubImage4DSGIS(\{target}, \{level}, \{xoffset}, \{yoffset}, \{zoffset}, \{woffset}, \{width}, \{height}, \{depth}, \{size4d}, \{format}, \{type}, \{pixels})");
		try {
		    functions.glTexSubImage4DSGIS().invokeExact(target, level, xoffset, yoffset, zoffset, woffset, width, height, depth, size4d, format, type, pixels);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, format=\{format}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTextureAttachMemoryNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTextureAttachMemoryNV.xhtml">registry.khronos.org</a>
	 */
	public void glTextureAttachMemoryNV(int texture, int memory, long offset) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTextureAttachMemoryNV(\{texture}, \{memory}, \{offset})");
		try {
		    functions.glTextureAttachMemoryNV().invokeExact(texture, memory, offset);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTextureBarrier">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTextureBarrier.xhtml">registry.khronos.org</a>
	 */
	public void glTextureBarrier() {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTextureBarrier()");
		try {
		    functions.glTextureBarrier().invokeExact();
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTextureBarrierNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTextureBarrierNV.xhtml">registry.khronos.org</a>
	 */
	public void glTextureBarrierNV() {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTextureBarrierNV()");
		try {
		    functions.glTextureBarrierNV().invokeExact();
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTextureBuffer">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTextureBuffer.xhtml">registry.khronos.org</a>
	 */
	public void glTextureBuffer(int texture, GLSizedInternalFormat internalformat, int buffer) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTextureBuffer(\{texture}, \{internalformat}, \{buffer})");
		try {
		    functions.glTextureBuffer().invokeExact(texture, internalformat, buffer);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were internalformat=\{internalformat}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTextureBufferEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTextureBufferEXT.xhtml">registry.khronos.org</a>
	 */
	public void glTextureBufferEXT(int texture, GLTextureTarget target, GLSizedInternalFormat internalformat, int buffer) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTextureBufferEXT(\{texture}, \{target}, \{internalformat}, \{buffer})");
		try {
		    functions.glTextureBufferEXT().invokeExact(texture, target, internalformat, buffer);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, internalformat=\{internalformat}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTextureBufferRange">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTextureBufferRange.xhtml">registry.khronos.org</a>
	 */
	public void glTextureBufferRange(int texture, GLSizedInternalFormat internalformat, int buffer, long offset, long size) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTextureBufferRange(\{texture}, \{internalformat}, \{buffer}, \{offset}, \{size})");
		try {
		    functions.glTextureBufferRange().invokeExact(texture, internalformat, buffer, offset, size);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were internalformat=\{internalformat}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTextureBufferRangeEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTextureBufferRangeEXT.xhtml">registry.khronos.org</a>
	 */
	public void glTextureBufferRangeEXT(int texture, GLTextureTarget target, GLSizedInternalFormat internalformat, int buffer, long offset, long size) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTextureBufferRangeEXT(\{texture}, \{target}, \{internalformat}, \{buffer}, \{offset}, \{size})");
		try {
		    functions.glTextureBufferRangeEXT().invokeExact(texture, target, internalformat, buffer, offset, size);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, internalformat=\{internalformat}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTextureColorMaskSGIS">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTextureColorMaskSGIS.xhtml">registry.khronos.org</a>
	 */
	public void glTextureColorMaskSGIS(boolean red, boolean green, boolean blue, boolean alpha) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTextureColorMaskSGIS(\{red}, \{green}, \{blue}, \{alpha})");
		try {
		    functions.glTextureColorMaskSGIS().invokeExact(red, green, blue, alpha);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTextureFoveationParametersQCOM">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTextureFoveationParametersQCOM.xhtml">registry.khronos.org</a>
	 */
	public void glTextureFoveationParametersQCOM(int texture, int layer, int focalPoint, float focalX, float focalY, float gainX, float gainY, float foveaArea) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTextureFoveationParametersQCOM(\{texture}, \{layer}, \{focalPoint}, \{focalX}, \{focalY}, \{gainX}, \{gainY}, \{foveaArea})");
		try {
		    functions.glTextureFoveationParametersQCOM().invokeExact(texture, layer, focalPoint, focalX, focalY, gainX, gainY, foveaArea);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTextureImage1DEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTextureImage1DEXT.xhtml">registry.khronos.org</a>
	 */
	public void glTextureImage1DEXT(int texture, GLTextureTarget target, int level, GLInternalFormat internalformat, int width, int border, GLPixelFormat format, GLPixelType type, ConstPointer<Void> pixels) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTextureImage1DEXT(\{texture}, \{target}, \{level}, \{internalformat}, \{width}, \{border}, \{format}, \{type}, \{pixels})");
		try {
		    functions.glTextureImage1DEXT().invokeExact(texture, target, level, internalformat, width, border, format, type, pixels);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, internalformat=\{internalformat}, format=\{format}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTextureImage2DEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTextureImage2DEXT.xhtml">registry.khronos.org</a>
	 */
	public void glTextureImage2DEXT(int texture, GLTextureTarget target, int level, GLInternalFormat internalformat, int width, int height, int border, GLPixelFormat format, GLPixelType type, ConstPointer<Void> pixels) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTextureImage2DEXT(\{texture}, \{target}, \{level}, \{internalformat}, \{width}, \{height}, \{border}, \{format}, \{type}, \{pixels})");
		try {
		    functions.glTextureImage2DEXT().invokeExact(texture, target, level, internalformat, width, height, border, format, type, pixels);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, internalformat=\{internalformat}, format=\{format}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTextureImage2DMultisampleCoverageNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTextureImage2DMultisampleCoverageNV.xhtml">registry.khronos.org</a>
	 */
	public void glTextureImage2DMultisampleCoverageNV(int texture, GLTextureTarget target, int coverageSamples, int colorSamples, GLInternalFormat internalFormat, int width, int height, boolean fixedSampleLocations) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTextureImage2DMultisampleCoverageNV(\{texture}, \{target}, \{coverageSamples}, \{colorSamples}, \{internalFormat}, \{width}, \{height}, \{fixedSampleLocations})");
		try {
		    functions.glTextureImage2DMultisampleCoverageNV().invokeExact(texture, target, coverageSamples, colorSamples, internalFormat, width, height, fixedSampleLocations);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, internalFormat=\{internalFormat}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTextureImage2DMultisampleNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTextureImage2DMultisampleNV.xhtml">registry.khronos.org</a>
	 */
	public void glTextureImage2DMultisampleNV(int texture, GLTextureTarget target, int samples, GLInternalFormat internalFormat, int width, int height, boolean fixedSampleLocations) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTextureImage2DMultisampleNV(\{texture}, \{target}, \{samples}, \{internalFormat}, \{width}, \{height}, \{fixedSampleLocations})");
		try {
		    functions.glTextureImage2DMultisampleNV().invokeExact(texture, target, samples, internalFormat, width, height, fixedSampleLocations);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, internalFormat=\{internalFormat}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTextureImage3DEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTextureImage3DEXT.xhtml">registry.khronos.org</a>
	 */
	public void glTextureImage3DEXT(int texture, GLTextureTarget target, int level, GLInternalFormat internalformat, int width, int height, int depth, int border, GLPixelFormat format, GLPixelType type, ConstPointer<Void> pixels) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTextureImage3DEXT(\{texture}, \{target}, \{level}, \{internalformat}, \{width}, \{height}, \{depth}, \{border}, \{format}, \{type}, \{pixels})");
		try {
		    functions.glTextureImage3DEXT().invokeExact(texture, target, level, internalformat, width, height, depth, border, format, type, pixels);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, internalformat=\{internalformat}, format=\{format}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTextureImage3DMultisampleCoverageNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTextureImage3DMultisampleCoverageNV.xhtml">registry.khronos.org</a>
	 */
	public void glTextureImage3DMultisampleCoverageNV(int texture, GLTextureTarget target, int coverageSamples, int colorSamples, GLInternalFormat internalFormat, int width, int height, int depth, boolean fixedSampleLocations) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTextureImage3DMultisampleCoverageNV(\{texture}, \{target}, \{coverageSamples}, \{colorSamples}, \{internalFormat}, \{width}, \{height}, \{depth}, \{fixedSampleLocations})");
		try {
		    functions.glTextureImage3DMultisampleCoverageNV().invokeExact(texture, target, coverageSamples, colorSamples, internalFormat, width, height, depth, fixedSampleLocations);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, internalFormat=\{internalFormat}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTextureImage3DMultisampleNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTextureImage3DMultisampleNV.xhtml">registry.khronos.org</a>
	 */
	public void glTextureImage3DMultisampleNV(int texture, GLTextureTarget target, int samples, GLInternalFormat internalFormat, int width, int height, int depth, boolean fixedSampleLocations) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTextureImage3DMultisampleNV(\{texture}, \{target}, \{samples}, \{internalFormat}, \{width}, \{height}, \{depth}, \{fixedSampleLocations})");
		try {
		    functions.glTextureImage3DMultisampleNV().invokeExact(texture, target, samples, internalFormat, width, height, depth, fixedSampleLocations);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, internalFormat=\{internalFormat}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTextureLightEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTextureLightEXT.xhtml">registry.khronos.org</a>
	 */
	public void glTextureLightEXT(GLLightTexturePNameEXT pname) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTextureLightEXT(\{pname})");
		try {
		    functions.glTextureLightEXT().invokeExact(pname);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTextureMaterialEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTextureMaterialEXT.xhtml">registry.khronos.org</a>
	 */
	public void glTextureMaterialEXT(GLTriangleFace face, GLMaterialParameter mode) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTextureMaterialEXT(\{face}, \{mode})");
		try {
		    functions.glTextureMaterialEXT().invokeExact(face, mode);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were face=\{face}, mode=\{mode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTextureNormalEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTextureNormalEXT.xhtml">registry.khronos.org</a>
	 */
	public void glTextureNormalEXT(GLTextureNormalModeEXT mode) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTextureNormalEXT(\{mode})");
		try {
		    functions.glTextureNormalEXT().invokeExact(mode);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexturePageCommitmentEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexturePageCommitmentEXT.xhtml">registry.khronos.org</a>
	 */
	public void glTexturePageCommitmentEXT(int texture, int level, int xoffset, int yoffset, int zoffset, int width, int height, int depth, boolean commit) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexturePageCommitmentEXT(\{texture}, \{level}, \{xoffset}, \{yoffset}, \{zoffset}, \{width}, \{height}, \{depth}, \{commit})");
		try {
		    functions.glTexturePageCommitmentEXT().invokeExact(texture, level, xoffset, yoffset, zoffset, width, height, depth, commit);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTexturePageCommitmentMemNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTexturePageCommitmentMemNV.xhtml">registry.khronos.org</a>
	 */
	public void glTexturePageCommitmentMemNV(int texture, int layer, int level, int xoffset, int yoffset, int zoffset, int width, int height, int depth, int memory, long offset, boolean commit) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTexturePageCommitmentMemNV(\{texture}, \{layer}, \{level}, \{xoffset}, \{yoffset}, \{zoffset}, \{width}, \{height}, \{depth}, \{memory}, \{offset}, \{commit})");
		try {
		    functions.glTexturePageCommitmentMemNV().invokeExact(texture, layer, level, xoffset, yoffset, zoffset, width, height, depth, memory, offset, commit);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTextureParameterIiv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTextureParameterIiv.xhtml">registry.khronos.org</a>
	 */
	public void glTextureParameterIiv(int texture, GLTextureParameterName pname, ConstPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTextureParameterIiv(\{texture}, \{pname}, \{params})");
		try {
		    functions.glTextureParameterIiv().invokeExact(texture, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTextureParameterIivEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTextureParameterIivEXT.xhtml">registry.khronos.org</a>
	 */
	public void glTextureParameterIivEXT(int texture, GLTextureTarget target, GLTextureParameterName pname, ConstPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTextureParameterIivEXT(\{texture}, \{target}, \{pname}, \{params})");
		try {
		    functions.glTextureParameterIivEXT().invokeExact(texture, target, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTextureParameterIuiv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTextureParameterIuiv.xhtml">registry.khronos.org</a>
	 */
	public void glTextureParameterIuiv(int texture, GLTextureParameterName pname, ConstPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTextureParameterIuiv(\{texture}, \{pname}, \{params})");
		try {
		    functions.glTextureParameterIuiv().invokeExact(texture, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTextureParameterIuivEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTextureParameterIuivEXT.xhtml">registry.khronos.org</a>
	 */
	public void glTextureParameterIuivEXT(int texture, GLTextureTarget target, GLTextureParameterName pname, ConstPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTextureParameterIuivEXT(\{texture}, \{target}, \{pname}, \{params})");
		try {
		    functions.glTextureParameterIuivEXT().invokeExact(texture, target, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTextureParameterf">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTextureParameterf.xhtml">registry.khronos.org</a>
	 */
	public void glTextureParameterf(int texture, GLTextureParameterName pname, float param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTextureParameterf(\{texture}, \{pname}, \{param})");
		try {
		    functions.glTextureParameterf().invokeExact(texture, pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTextureParameterfEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTextureParameterfEXT.xhtml">registry.khronos.org</a>
	 */
	public void glTextureParameterfEXT(int texture, GLTextureTarget target, GLTextureParameterName pname, float param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTextureParameterfEXT(\{texture}, \{target}, \{pname}, \{param})");
		try {
		    functions.glTextureParameterfEXT().invokeExact(texture, target, pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTextureParameterfv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTextureParameterfv.xhtml">registry.khronos.org</a>
	 */
	public void glTextureParameterfv(int texture, GLTextureParameterName pname, ConstPointer<Float> param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTextureParameterfv(\{texture}, \{pname}, \{param})");
		try {
		    functions.glTextureParameterfv().invokeExact(texture, pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTextureParameterfvEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTextureParameterfvEXT.xhtml">registry.khronos.org</a>
	 */
	public void glTextureParameterfvEXT(int texture, GLTextureTarget target, GLTextureParameterName pname, ConstPointer<Float> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTextureParameterfvEXT(\{texture}, \{target}, \{pname}, \{params})");
		try {
		    functions.glTextureParameterfvEXT().invokeExact(texture, target, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTextureParameteri">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTextureParameteri.xhtml">registry.khronos.org</a>
	 */
	public void glTextureParameteri(int texture, GLTextureParameterName pname, int param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTextureParameteri(\{texture}, \{pname}, \{param})");
		try {
		    functions.glTextureParameteri().invokeExact(texture, pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTextureParameteriEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTextureParameteriEXT.xhtml">registry.khronos.org</a>
	 */
	public void glTextureParameteriEXT(int texture, GLTextureTarget target, GLTextureParameterName pname, int param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTextureParameteriEXT(\{texture}, \{target}, \{pname}, \{param})");
		try {
		    functions.glTextureParameteriEXT().invokeExact(texture, target, pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTextureParameteriv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTextureParameteriv.xhtml">registry.khronos.org</a>
	 */
	public void glTextureParameteriv(int texture, GLTextureParameterName pname, ConstPointer<Integer> param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTextureParameteriv(\{texture}, \{pname}, \{param})");
		try {
		    functions.glTextureParameteriv().invokeExact(texture, pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTextureParameterivEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTextureParameterivEXT.xhtml">registry.khronos.org</a>
	 */
	public void glTextureParameterivEXT(int texture, GLTextureTarget target, GLTextureParameterName pname, ConstPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTextureParameterivEXT(\{texture}, \{target}, \{pname}, \{params})");
		try {
		    functions.glTextureParameterivEXT().invokeExact(texture, target, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTextureRangeAPPLE">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTextureRangeAPPLE.xhtml">registry.khronos.org</a>
	 */
	public void glTextureRangeAPPLE(GLEnum target, int length, ConstPointer<Void> pointer) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTextureRangeAPPLE(\{target}, \{length}, \{pointer})");
		try {
		    functions.glTextureRangeAPPLE().invokeExact(target, length, pointer);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTextureRenderbufferEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTextureRenderbufferEXT.xhtml">registry.khronos.org</a>
	 */
	public void glTextureRenderbufferEXT(int texture, GLTextureTarget target, int renderbuffer) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTextureRenderbufferEXT(\{texture}, \{target}, \{renderbuffer})");
		try {
		    functions.glTextureRenderbufferEXT().invokeExact(texture, target, renderbuffer);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTextureStorage1D">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTextureStorage1D.xhtml">registry.khronos.org</a>
	 */
	public void glTextureStorage1D(int texture, int levels, GLSizedInternalFormat internalformat, int width) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTextureStorage1D(\{texture}, \{levels}, \{internalformat}, \{width})");
		try {
		    functions.glTextureStorage1D().invokeExact(texture, levels, internalformat, width);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were internalformat=\{internalformat}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTextureStorage1DEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTextureStorage1DEXT.xhtml">registry.khronos.org</a>
	 */
	public void glTextureStorage1DEXT(int texture, GLEnum target, int levels, GLSizedInternalFormat internalformat, int width) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTextureStorage1DEXT(\{texture}, \{target}, \{levels}, \{internalformat}, \{width})");
		try {
		    functions.glTextureStorage1DEXT().invokeExact(texture, target, levels, internalformat, width);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, internalformat=\{internalformat}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTextureStorage2D">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTextureStorage2D.xhtml">registry.khronos.org</a>
	 */
	public void glTextureStorage2D(int texture, int levels, GLSizedInternalFormat internalformat, int width, int height) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTextureStorage2D(\{texture}, \{levels}, \{internalformat}, \{width}, \{height})");
		try {
		    functions.glTextureStorage2D().invokeExact(texture, levels, internalformat, width, height);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were internalformat=\{internalformat}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTextureStorage2DEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTextureStorage2DEXT.xhtml">registry.khronos.org</a>
	 */
	public void glTextureStorage2DEXT(int texture, GLEnum target, int levels, GLSizedInternalFormat internalformat, int width, int height) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTextureStorage2DEXT(\{texture}, \{target}, \{levels}, \{internalformat}, \{width}, \{height})");
		try {
		    functions.glTextureStorage2DEXT().invokeExact(texture, target, levels, internalformat, width, height);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, internalformat=\{internalformat}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTextureStorage2DMultisample">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTextureStorage2DMultisample.xhtml">registry.khronos.org</a>
	 */
	public void glTextureStorage2DMultisample(int texture, int samples, GLSizedInternalFormat internalformat, int width, int height, boolean fixedsamplelocations) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTextureStorage2DMultisample(\{texture}, \{samples}, \{internalformat}, \{width}, \{height}, \{fixedsamplelocations})");
		try {
		    functions.glTextureStorage2DMultisample().invokeExact(texture, samples, internalformat, width, height, fixedsamplelocations);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were internalformat=\{internalformat}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTextureStorage2DMultisampleEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTextureStorage2DMultisampleEXT.xhtml">registry.khronos.org</a>
	 */
	public void glTextureStorage2DMultisampleEXT(int texture, GLTextureTarget target, int samples, GLSizedInternalFormat internalformat, int width, int height, boolean fixedsamplelocations) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTextureStorage2DMultisampleEXT(\{texture}, \{target}, \{samples}, \{internalformat}, \{width}, \{height}, \{fixedsamplelocations})");
		try {
		    functions.glTextureStorage2DMultisampleEXT().invokeExact(texture, target, samples, internalformat, width, height, fixedsamplelocations);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, internalformat=\{internalformat}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTextureStorage3D">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTextureStorage3D.xhtml">registry.khronos.org</a>
	 */
	public void glTextureStorage3D(int texture, int levels, GLSizedInternalFormat internalformat, int width, int height, int depth) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTextureStorage3D(\{texture}, \{levels}, \{internalformat}, \{width}, \{height}, \{depth})");
		try {
		    functions.glTextureStorage3D().invokeExact(texture, levels, internalformat, width, height, depth);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were internalformat=\{internalformat}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTextureStorage3DEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTextureStorage3DEXT.xhtml">registry.khronos.org</a>
	 */
	public void glTextureStorage3DEXT(int texture, GLEnum target, int levels, GLSizedInternalFormat internalformat, int width, int height, int depth) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTextureStorage3DEXT(\{texture}, \{target}, \{levels}, \{internalformat}, \{width}, \{height}, \{depth})");
		try {
		    functions.glTextureStorage3DEXT().invokeExact(texture, target, levels, internalformat, width, height, depth);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, internalformat=\{internalformat}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTextureStorage3DMultisample">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTextureStorage3DMultisample.xhtml">registry.khronos.org</a>
	 */
	public void glTextureStorage3DMultisample(int texture, int samples, GLSizedInternalFormat internalformat, int width, int height, int depth, boolean fixedsamplelocations) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTextureStorage3DMultisample(\{texture}, \{samples}, \{internalformat}, \{width}, \{height}, \{depth}, \{fixedsamplelocations})");
		try {
		    functions.glTextureStorage3DMultisample().invokeExact(texture, samples, internalformat, width, height, depth, fixedsamplelocations);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were internalformat=\{internalformat}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTextureStorage3DMultisampleEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTextureStorage3DMultisampleEXT.xhtml">registry.khronos.org</a>
	 */
	public void glTextureStorage3DMultisampleEXT(int texture, GLEnum target, int samples, GLSizedInternalFormat internalformat, int width, int height, int depth, boolean fixedsamplelocations) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTextureStorage3DMultisampleEXT(\{texture}, \{target}, \{samples}, \{internalformat}, \{width}, \{height}, \{depth}, \{fixedsamplelocations})");
		try {
		    functions.glTextureStorage3DMultisampleEXT().invokeExact(texture, target, samples, internalformat, width, height, depth, fixedsamplelocations);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, internalformat=\{internalformat}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTextureStorageMem1DEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTextureStorageMem1DEXT.xhtml">registry.khronos.org</a>
	 */
	public void glTextureStorageMem1DEXT(int texture, int levels, GLSizedInternalFormat internalFormat, int width, int memory, long offset) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTextureStorageMem1DEXT(\{texture}, \{levels}, \{internalFormat}, \{width}, \{memory}, \{offset})");
		try {
		    functions.glTextureStorageMem1DEXT().invokeExact(texture, levels, internalFormat, width, memory, offset);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were internalFormat=\{internalFormat}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTextureStorageMem2DEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTextureStorageMem2DEXT.xhtml">registry.khronos.org</a>
	 */
	public void glTextureStorageMem2DEXT(int texture, int levels, GLSizedInternalFormat internalFormat, int width, int height, int memory, long offset) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTextureStorageMem2DEXT(\{texture}, \{levels}, \{internalFormat}, \{width}, \{height}, \{memory}, \{offset})");
		try {
		    functions.glTextureStorageMem2DEXT().invokeExact(texture, levels, internalFormat, width, height, memory, offset);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were internalFormat=\{internalFormat}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTextureStorageMem2DMultisampleEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTextureStorageMem2DMultisampleEXT.xhtml">registry.khronos.org</a>
	 */
	public void glTextureStorageMem2DMultisampleEXT(int texture, int samples, GLSizedInternalFormat internalFormat, int width, int height, boolean fixedSampleLocations, int memory, long offset) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTextureStorageMem2DMultisampleEXT(\{texture}, \{samples}, \{internalFormat}, \{width}, \{height}, \{fixedSampleLocations}, \{memory}, \{offset})");
		try {
		    functions.glTextureStorageMem2DMultisampleEXT().invokeExact(texture, samples, internalFormat, width, height, fixedSampleLocations, memory, offset);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were internalFormat=\{internalFormat}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTextureStorageMem3DEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTextureStorageMem3DEXT.xhtml">registry.khronos.org</a>
	 */
	public void glTextureStorageMem3DEXT(int texture, int levels, GLSizedInternalFormat internalFormat, int width, int height, int depth, int memory, long offset) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTextureStorageMem3DEXT(\{texture}, \{levels}, \{internalFormat}, \{width}, \{height}, \{depth}, \{memory}, \{offset})");
		try {
		    functions.glTextureStorageMem3DEXT().invokeExact(texture, levels, internalFormat, width, height, depth, memory, offset);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were internalFormat=\{internalFormat}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTextureStorageMem3DMultisampleEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTextureStorageMem3DMultisampleEXT.xhtml">registry.khronos.org</a>
	 */
	public void glTextureStorageMem3DMultisampleEXT(int texture, int samples, GLSizedInternalFormat internalFormat, int width, int height, int depth, boolean fixedSampleLocations, int memory, long offset) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTextureStorageMem3DMultisampleEXT(\{texture}, \{samples}, \{internalFormat}, \{width}, \{height}, \{depth}, \{fixedSampleLocations}, \{memory}, \{offset})");
		try {
		    functions.glTextureStorageMem3DMultisampleEXT().invokeExact(texture, samples, internalFormat, width, height, depth, fixedSampleLocations, memory, offset);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were internalFormat=\{internalFormat}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTextureStorageSparseAMD">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTextureStorageSparseAMD.xhtml">registry.khronos.org</a>
	 */
	public void glTextureStorageSparseAMD(int texture, GLEnum target, GLSizedInternalFormat internalFormat, int width, int height, int depth, int layers, int flags) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTextureStorageSparseAMD(\{texture}, \{target}, \{internalFormat}, \{width}, \{height}, \{depth}, \{layers}, \{flags})");
		try {
		    functions.glTextureStorageSparseAMD().invokeExact(texture, target, internalFormat, width, height, depth, layers, flags);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, internalFormat=\{internalFormat}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTextureSubImage1D">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTextureSubImage1D.xhtml">registry.khronos.org</a>
	 */
	public void glTextureSubImage1D(int texture, int level, int xoffset, int width, GLPixelFormat format, GLPixelType type, ConstPointer<Void> pixels) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTextureSubImage1D(\{texture}, \{level}, \{xoffset}, \{width}, \{format}, \{type}, \{pixels})");
		try {
		    functions.glTextureSubImage1D().invokeExact(texture, level, xoffset, width, format, type, pixels);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were format=\{format}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTextureSubImage1DEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTextureSubImage1DEXT.xhtml">registry.khronos.org</a>
	 */
	public void glTextureSubImage1DEXT(int texture, GLTextureTarget target, int level, int xoffset, int width, GLPixelFormat format, GLPixelType type, ConstPointer<Void> pixels) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTextureSubImage1DEXT(\{texture}, \{target}, \{level}, \{xoffset}, \{width}, \{format}, \{type}, \{pixels})");
		try {
		    functions.glTextureSubImage1DEXT().invokeExact(texture, target, level, xoffset, width, format, type, pixels);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, format=\{format}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTextureSubImage2D">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTextureSubImage2D.xhtml">registry.khronos.org</a>
	 */
	public void glTextureSubImage2D(int texture, int level, int xoffset, int yoffset, int width, int height, GLPixelFormat format, GLPixelType type, ConstPointer<Void> pixels) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTextureSubImage2D(\{texture}, \{level}, \{xoffset}, \{yoffset}, \{width}, \{height}, \{format}, \{type}, \{pixels})");
		try {
		    functions.glTextureSubImage2D().invokeExact(texture, level, xoffset, yoffset, width, height, format, type, pixels);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were format=\{format}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTextureSubImage2DEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTextureSubImage2DEXT.xhtml">registry.khronos.org</a>
	 */
	public void glTextureSubImage2DEXT(int texture, GLTextureTarget target, int level, int xoffset, int yoffset, int width, int height, GLPixelFormat format, GLPixelType type, ConstPointer<Void> pixels) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTextureSubImage2DEXT(\{texture}, \{target}, \{level}, \{xoffset}, \{yoffset}, \{width}, \{height}, \{format}, \{type}, \{pixels})");
		try {
		    functions.glTextureSubImage2DEXT().invokeExact(texture, target, level, xoffset, yoffset, width, height, format, type, pixels);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, format=\{format}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTextureSubImage3D">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTextureSubImage3D.xhtml">registry.khronos.org</a>
	 */
	public void glTextureSubImage3D(int texture, int level, int xoffset, int yoffset, int zoffset, int width, int height, int depth, GLPixelFormat format, GLPixelType type, ConstPointer<Void> pixels) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTextureSubImage3D(\{texture}, \{level}, \{xoffset}, \{yoffset}, \{zoffset}, \{width}, \{height}, \{depth}, \{format}, \{type}, \{pixels})");
		try {
		    functions.glTextureSubImage3D().invokeExact(texture, level, xoffset, yoffset, zoffset, width, height, depth, format, type, pixels);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were format=\{format}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTextureSubImage3DEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTextureSubImage3DEXT.xhtml">registry.khronos.org</a>
	 */
	public void glTextureSubImage3DEXT(int texture, GLTextureTarget target, int level, int xoffset, int yoffset, int zoffset, int width, int height, int depth, GLPixelFormat format, GLPixelType type, ConstPointer<Void> pixels) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTextureSubImage3DEXT(\{texture}, \{target}, \{level}, \{xoffset}, \{yoffset}, \{zoffset}, \{width}, \{height}, \{depth}, \{format}, \{type}, \{pixels})");
		try {
		    functions.glTextureSubImage3DEXT().invokeExact(texture, target, level, xoffset, yoffset, zoffset, width, height, depth, format, type, pixels);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, format=\{format}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTextureView">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTextureView.xhtml">registry.khronos.org</a>
	 */
	public void glTextureView(int texture, GLTextureTarget target, int origtexture, GLSizedInternalFormat internalformat, int minlevel, int numlevels, int minlayer, int numlayers) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTextureView(\{texture}, \{target}, \{origtexture}, \{internalformat}, \{minlevel}, \{numlevels}, \{minlayer}, \{numlayers})");
		try {
		    functions.glTextureView().invokeExact(texture, target, origtexture, internalformat, minlevel, numlevels, minlayer, numlayers);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, internalformat=\{internalformat}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTextureViewEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTextureViewEXT.xhtml">registry.khronos.org</a>
	 */
	public void glTextureViewEXT(int texture, GLTextureTarget target, int origtexture, GLSizedInternalFormat internalformat, int minlevel, int numlevels, int minlayer, int numlayers) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTextureViewEXT(\{texture}, \{target}, \{origtexture}, \{internalformat}, \{minlevel}, \{numlevels}, \{minlayer}, \{numlayers})");
		try {
		    functions.glTextureViewEXT().invokeExact(texture, target, origtexture, internalformat, minlevel, numlevels, minlayer, numlayers);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, internalformat=\{internalformat}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTextureViewOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTextureViewOES.xhtml">registry.khronos.org</a>
	 */
	public void glTextureViewOES(int texture, GLTextureTarget target, int origtexture, GLSizedInternalFormat internalformat, int minlevel, int numlevels, int minlayer, int numlayers) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTextureViewOES(\{texture}, \{target}, \{origtexture}, \{internalformat}, \{minlevel}, \{numlevels}, \{minlayer}, \{numlayers})");
		try {
		    functions.glTextureViewOES().invokeExact(texture, target, origtexture, internalformat, minlevel, numlevels, minlayer, numlayers);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, internalformat=\{internalformat}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTrackMatrixNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTrackMatrixNV.xhtml">registry.khronos.org</a>
	 */
	public void glTrackMatrixNV(GLVertexAttribEnumNV target, int address, GLVertexAttribEnumNV matrix, GLVertexAttribEnumNV transform) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTrackMatrixNV(\{target}, \{address}, \{matrix}, \{transform})");
		try {
		    functions.glTrackMatrixNV().invokeExact(target, address, matrix, transform);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, matrix=\{matrix}, transform=\{transform}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTransformFeedbackAttribsNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTransformFeedbackAttribsNV.xhtml">registry.khronos.org</a>
	 */
	public void glTransformFeedbackAttribsNV(int count, ConstPointer<Integer> attribs, GLEnum bufferMode) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTransformFeedbackAttribsNV(\{count}, \{attribs}, \{bufferMode})");
		try {
		    functions.glTransformFeedbackAttribsNV().invokeExact(count, attribs, bufferMode);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were bufferMode=\{bufferMode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTransformFeedbackBufferBase">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTransformFeedbackBufferBase.xhtml">registry.khronos.org</a>
	 */
	public void glTransformFeedbackBufferBase(int xfb, int index, int buffer) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTransformFeedbackBufferBase(\{xfb}, \{index}, \{buffer})");
		try {
		    functions.glTransformFeedbackBufferBase().invokeExact(xfb, index, buffer);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTransformFeedbackBufferRange">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTransformFeedbackBufferRange.xhtml">registry.khronos.org</a>
	 */
	public void glTransformFeedbackBufferRange(int xfb, int index, int buffer, long offset, long size) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTransformFeedbackBufferRange(\{xfb}, \{index}, \{buffer}, \{offset}, \{size})");
		try {
		    functions.glTransformFeedbackBufferRange().invokeExact(xfb, index, buffer, offset, size);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTransformFeedbackStreamAttribsNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTransformFeedbackStreamAttribsNV.xhtml">registry.khronos.org</a>
	 */
	public void glTransformFeedbackStreamAttribsNV(int count, ConstPointer<Integer> attribs, int nbuffers, ConstPointer<Integer> bufstreams, GLEnum bufferMode) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTransformFeedbackStreamAttribsNV(\{count}, \{attribs}, \{nbuffers}, \{bufstreams}, \{bufferMode})");
		try {
		    functions.glTransformFeedbackStreamAttribsNV().invokeExact(count, attribs, nbuffers, bufstreams, bufferMode);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were bufferMode=\{bufferMode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTransformFeedbackVaryings">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTransformFeedbackVaryings.xhtml">registry.khronos.org</a>
	 */
	public void glTransformFeedbackVaryings(int program, int count, ConstPointer<ConstPointer<Byte>> varyings, GLTransformFeedbackBufferMode bufferMode) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTransformFeedbackVaryings(\{program}, \{count}, \{varyings}, \{bufferMode})");
		try {
		    functions.glTransformFeedbackVaryings().invokeExact(program, count, varyings, bufferMode);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were bufferMode=\{bufferMode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTransformFeedbackVaryingsEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTransformFeedbackVaryingsEXT.xhtml">registry.khronos.org</a>
	 */
	public void glTransformFeedbackVaryingsEXT(int program, int count, ConstPointer<ConstPointer<Byte>> varyings, GLTransformFeedbackBufferMode bufferMode) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTransformFeedbackVaryingsEXT(\{program}, \{count}, \{varyings}, \{bufferMode})");
		try {
		    functions.glTransformFeedbackVaryingsEXT().invokeExact(program, count, varyings, bufferMode);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were bufferMode=\{bufferMode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTransformFeedbackVaryingsNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTransformFeedbackVaryingsNV.xhtml">registry.khronos.org</a>
	 */
	public void glTransformFeedbackVaryingsNV(int program, int count, ConstPointer<GLTransformFeedbackTokenNV> locations, GLTransformFeedbackBufferMode bufferMode) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTransformFeedbackVaryingsNV(\{program}, \{count}, \{locations}, \{bufferMode})");
		try {
		    functions.glTransformFeedbackVaryingsNV().invokeExact(program, count, locations, bufferMode);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were bufferMode=\{bufferMode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTransformPathNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTransformPathNV.xhtml">registry.khronos.org</a>
	 */
	public void glTransformPathNV(int resultPath, int srcPath, GLPathTransformType transformType, ConstPointer<Float> transformValues) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTransformPathNV(\{resultPath}, \{srcPath}, \{transformType}, \{transformValues})");
		try {
		    functions.glTransformPathNV().invokeExact(resultPath, srcPath, transformType, transformValues);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were transformType=\{transformType}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTranslated">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTranslated.xhtml">registry.khronos.org</a>
	 */
	public void glTranslated(double x, double y, double z) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTranslated(\{x}, \{y}, \{z})");
		try {
		    functions.glTranslated().invokeExact(x, y, z);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTranslatef">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTranslatef.xhtml">registry.khronos.org</a>
	 */
	public void glTranslatef(float x, float y, float z) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTranslatef(\{x}, \{y}, \{z})");
		try {
		    functions.glTranslatef().invokeExact(x, y, z);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTranslatex">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTranslatex.xhtml">registry.khronos.org</a>
	 */
	public void glTranslatex(int x, int y, int z) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTranslatex(\{x}, \{y}, \{z})");
		try {
		    functions.glTranslatex().invokeExact(x, y, z);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glTranslatexOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glTranslatexOES.xhtml">registry.khronos.org</a>
	 */
	public void glTranslatexOES(int x, int y, int z) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glTranslatexOES(\{x}, \{y}, \{z})");
		try {
		    functions.glTranslatexOES().invokeExact(x, y, z);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniform1d">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniform1d.xhtml">registry.khronos.org</a>
	 */
	public void glUniform1d(int location, double x) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniform1d(\{location}, \{x})");
		try {
		    functions.glUniform1d().invokeExact(location, x);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniform1dv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniform1dv.xhtml">registry.khronos.org</a>
	 */
	public void glUniform1dv(int location, int count, ConstPointer<Double> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniform1dv(\{location}, \{count}, \{value})");
		try {
		    functions.glUniform1dv().invokeExact(location, count, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniform1f">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniform1f.xhtml">registry.khronos.org</a>
	 */
	public void glUniform1f(int location, float v0) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniform1f(\{location}, \{v0})");
		try {
		    functions.glUniform1f().invokeExact(location, v0);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniform1fARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniform1fARB.xhtml">registry.khronos.org</a>
	 */
	public void glUniform1fARB(int location, float v0) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniform1fARB(\{location}, \{v0})");
		try {
		    functions.glUniform1fARB().invokeExact(location, v0);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniform1fv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniform1fv.xhtml">registry.khronos.org</a>
	 */
	public void glUniform1fv(int location, int count, ConstPointer<Float> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniform1fv(\{location}, \{count}, \{value})");
		try {
		    functions.glUniform1fv().invokeExact(location, count, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniform1fvARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniform1fvARB.xhtml">registry.khronos.org</a>
	 */
	public void glUniform1fvARB(int location, int count, ConstPointer<Float> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniform1fvARB(\{location}, \{count}, \{value})");
		try {
		    functions.glUniform1fvARB().invokeExact(location, count, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniform1i">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniform1i.xhtml">registry.khronos.org</a>
	 */
	public void glUniform1i(int location, int v0) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniform1i(\{location}, \{v0})");
		try {
		    functions.glUniform1i().invokeExact(location, v0);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniform1i64ARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniform1i64ARB.xhtml">registry.khronos.org</a>
	 */
	public void glUniform1i64ARB(int location, long x) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniform1i64ARB(\{location}, \{x})");
		try {
		    functions.glUniform1i64ARB().invokeExact(location, x);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniform1i64NV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniform1i64NV.xhtml">registry.khronos.org</a>
	 */
	public void glUniform1i64NV(int location, long x) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniform1i64NV(\{location}, \{x})");
		try {
		    functions.glUniform1i64NV().invokeExact(location, x);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniform1i64vARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniform1i64vARB.xhtml">registry.khronos.org</a>
	 */
	public void glUniform1i64vARB(int location, int count, ConstPointer<Long> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniform1i64vARB(\{location}, \{count}, \{value})");
		try {
		    functions.glUniform1i64vARB().invokeExact(location, count, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniform1i64vNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniform1i64vNV.xhtml">registry.khronos.org</a>
	 */
	public void glUniform1i64vNV(int location, int count, ConstPointer<Long> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniform1i64vNV(\{location}, \{count}, \{value})");
		try {
		    functions.glUniform1i64vNV().invokeExact(location, count, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniform1iARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniform1iARB.xhtml">registry.khronos.org</a>
	 */
	public void glUniform1iARB(int location, int v0) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniform1iARB(\{location}, \{v0})");
		try {
		    functions.glUniform1iARB().invokeExact(location, v0);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniform1iv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniform1iv.xhtml">registry.khronos.org</a>
	 */
	public void glUniform1iv(int location, int count, ConstPointer<Integer> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniform1iv(\{location}, \{count}, \{value})");
		try {
		    functions.glUniform1iv().invokeExact(location, count, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniform1ivARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniform1ivARB.xhtml">registry.khronos.org</a>
	 */
	public void glUniform1ivARB(int location, int count, ConstPointer<Integer> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniform1ivARB(\{location}, \{count}, \{value})");
		try {
		    functions.glUniform1ivARB().invokeExact(location, count, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniform1ui">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniform1ui.xhtml">registry.khronos.org</a>
	 */
	public void glUniform1ui(int location, int v0) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniform1ui(\{location}, \{v0})");
		try {
		    functions.glUniform1ui().invokeExact(location, v0);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniform1ui64ARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniform1ui64ARB.xhtml">registry.khronos.org</a>
	 */
	public void glUniform1ui64ARB(int location, long x) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniform1ui64ARB(\{location}, \{x})");
		try {
		    functions.glUniform1ui64ARB().invokeExact(location, x);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniform1ui64NV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniform1ui64NV.xhtml">registry.khronos.org</a>
	 */
	public void glUniform1ui64NV(int location, long x) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniform1ui64NV(\{location}, \{x})");
		try {
		    functions.glUniform1ui64NV().invokeExact(location, x);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniform1ui64vARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniform1ui64vARB.xhtml">registry.khronos.org</a>
	 */
	public void glUniform1ui64vARB(int location, int count, ConstPointer<Long> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniform1ui64vARB(\{location}, \{count}, \{value})");
		try {
		    functions.glUniform1ui64vARB().invokeExact(location, count, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniform1ui64vNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniform1ui64vNV.xhtml">registry.khronos.org</a>
	 */
	public void glUniform1ui64vNV(int location, int count, ConstPointer<Long> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniform1ui64vNV(\{location}, \{count}, \{value})");
		try {
		    functions.glUniform1ui64vNV().invokeExact(location, count, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniform1uiEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniform1uiEXT.xhtml">registry.khronos.org</a>
	 */
	public void glUniform1uiEXT(int location, int v0) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniform1uiEXT(\{location}, \{v0})");
		try {
		    functions.glUniform1uiEXT().invokeExact(location, v0);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniform1uiv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniform1uiv.xhtml">registry.khronos.org</a>
	 */
	public void glUniform1uiv(int location, int count, ConstPointer<Integer> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniform1uiv(\{location}, \{count}, \{value})");
		try {
		    functions.glUniform1uiv().invokeExact(location, count, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniform1uivEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniform1uivEXT.xhtml">registry.khronos.org</a>
	 */
	public void glUniform1uivEXT(int location, int count, ConstPointer<Integer> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniform1uivEXT(\{location}, \{count}, \{value})");
		try {
		    functions.glUniform1uivEXT().invokeExact(location, count, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniform2d">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniform2d.xhtml">registry.khronos.org</a>
	 */
	public void glUniform2d(int location, double x, double y) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniform2d(\{location}, \{x}, \{y})");
		try {
		    functions.glUniform2d().invokeExact(location, x, y);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniform2dv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniform2dv.xhtml">registry.khronos.org</a>
	 */
	public void glUniform2dv(int location, int count, ConstPointer<Double> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniform2dv(\{location}, \{count}, \{value})");
		try {
		    functions.glUniform2dv().invokeExact(location, count, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniform2f">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniform2f.xhtml">registry.khronos.org</a>
	 */
	public void glUniform2f(int location, float v0, float v1) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniform2f(\{location}, \{v0}, \{v1})");
		try {
		    functions.glUniform2f().invokeExact(location, v0, v1);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniform2fARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniform2fARB.xhtml">registry.khronos.org</a>
	 */
	public void glUniform2fARB(int location, float v0, float v1) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniform2fARB(\{location}, \{v0}, \{v1})");
		try {
		    functions.glUniform2fARB().invokeExact(location, v0, v1);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniform2fv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniform2fv.xhtml">registry.khronos.org</a>
	 */
	public void glUniform2fv(int location, int count, ConstPointer<Float> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniform2fv(\{location}, \{count}, \{value})");
		try {
		    functions.glUniform2fv().invokeExact(location, count, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniform2fvARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniform2fvARB.xhtml">registry.khronos.org</a>
	 */
	public void glUniform2fvARB(int location, int count, ConstPointer<Float> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniform2fvARB(\{location}, \{count}, \{value})");
		try {
		    functions.glUniform2fvARB().invokeExact(location, count, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniform2i">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniform2i.xhtml">registry.khronos.org</a>
	 */
	public void glUniform2i(int location, int v0, int v1) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniform2i(\{location}, \{v0}, \{v1})");
		try {
		    functions.glUniform2i().invokeExact(location, v0, v1);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniform2i64ARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniform2i64ARB.xhtml">registry.khronos.org</a>
	 */
	public void glUniform2i64ARB(int location, long x, long y) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniform2i64ARB(\{location}, \{x}, \{y})");
		try {
		    functions.glUniform2i64ARB().invokeExact(location, x, y);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniform2i64NV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniform2i64NV.xhtml">registry.khronos.org</a>
	 */
	public void glUniform2i64NV(int location, long x, long y) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniform2i64NV(\{location}, \{x}, \{y})");
		try {
		    functions.glUniform2i64NV().invokeExact(location, x, y);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniform2i64vARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniform2i64vARB.xhtml">registry.khronos.org</a>
	 */
	public void glUniform2i64vARB(int location, int count, ConstPointer<Long> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniform2i64vARB(\{location}, \{count}, \{value})");
		try {
		    functions.glUniform2i64vARB().invokeExact(location, count, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniform2i64vNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniform2i64vNV.xhtml">registry.khronos.org</a>
	 */
	public void glUniform2i64vNV(int location, int count, ConstPointer<Long> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniform2i64vNV(\{location}, \{count}, \{value})");
		try {
		    functions.glUniform2i64vNV().invokeExact(location, count, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniform2iARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniform2iARB.xhtml">registry.khronos.org</a>
	 */
	public void glUniform2iARB(int location, int v0, int v1) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniform2iARB(\{location}, \{v0}, \{v1})");
		try {
		    functions.glUniform2iARB().invokeExact(location, v0, v1);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniform2iv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniform2iv.xhtml">registry.khronos.org</a>
	 */
	public void glUniform2iv(int location, int count, ConstPointer<Integer> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniform2iv(\{location}, \{count}, \{value})");
		try {
		    functions.glUniform2iv().invokeExact(location, count, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniform2ivARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniform2ivARB.xhtml">registry.khronos.org</a>
	 */
	public void glUniform2ivARB(int location, int count, ConstPointer<Integer> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniform2ivARB(\{location}, \{count}, \{value})");
		try {
		    functions.glUniform2ivARB().invokeExact(location, count, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniform2ui">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniform2ui.xhtml">registry.khronos.org</a>
	 */
	public void glUniform2ui(int location, int v0, int v1) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniform2ui(\{location}, \{v0}, \{v1})");
		try {
		    functions.glUniform2ui().invokeExact(location, v0, v1);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniform2ui64ARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniform2ui64ARB.xhtml">registry.khronos.org</a>
	 */
	public void glUniform2ui64ARB(int location, long x, long y) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniform2ui64ARB(\{location}, \{x}, \{y})");
		try {
		    functions.glUniform2ui64ARB().invokeExact(location, x, y);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniform2ui64NV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniform2ui64NV.xhtml">registry.khronos.org</a>
	 */
	public void glUniform2ui64NV(int location, long x, long y) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniform2ui64NV(\{location}, \{x}, \{y})");
		try {
		    functions.glUniform2ui64NV().invokeExact(location, x, y);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniform2ui64vARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniform2ui64vARB.xhtml">registry.khronos.org</a>
	 */
	public void glUniform2ui64vARB(int location, int count, ConstPointer<Long> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniform2ui64vARB(\{location}, \{count}, \{value})");
		try {
		    functions.glUniform2ui64vARB().invokeExact(location, count, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniform2ui64vNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniform2ui64vNV.xhtml">registry.khronos.org</a>
	 */
	public void glUniform2ui64vNV(int location, int count, ConstPointer<Long> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniform2ui64vNV(\{location}, \{count}, \{value})");
		try {
		    functions.glUniform2ui64vNV().invokeExact(location, count, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniform2uiEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniform2uiEXT.xhtml">registry.khronos.org</a>
	 */
	public void glUniform2uiEXT(int location, int v0, int v1) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniform2uiEXT(\{location}, \{v0}, \{v1})");
		try {
		    functions.glUniform2uiEXT().invokeExact(location, v0, v1);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniform2uiv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniform2uiv.xhtml">registry.khronos.org</a>
	 */
	public void glUniform2uiv(int location, int count, ConstPointer<Integer> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniform2uiv(\{location}, \{count}, \{value})");
		try {
		    functions.glUniform2uiv().invokeExact(location, count, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniform2uivEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniform2uivEXT.xhtml">registry.khronos.org</a>
	 */
	public void glUniform2uivEXT(int location, int count, ConstPointer<Integer> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniform2uivEXT(\{location}, \{count}, \{value})");
		try {
		    functions.glUniform2uivEXT().invokeExact(location, count, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniform3d">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniform3d.xhtml">registry.khronos.org</a>
	 */
	public void glUniform3d(int location, double x, double y, double z) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniform3d(\{location}, \{x}, \{y}, \{z})");
		try {
		    functions.glUniform3d().invokeExact(location, x, y, z);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniform3dv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniform3dv.xhtml">registry.khronos.org</a>
	 */
	public void glUniform3dv(int location, int count, ConstPointer<Double> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniform3dv(\{location}, \{count}, \{value})");
		try {
		    functions.glUniform3dv().invokeExact(location, count, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniform3f">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniform3f.xhtml">registry.khronos.org</a>
	 */
	public void glUniform3f(int location, float v0, float v1, float v2) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniform3f(\{location}, \{v0}, \{v1}, \{v2})");
		try {
		    functions.glUniform3f().invokeExact(location, v0, v1, v2);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniform3fARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniform3fARB.xhtml">registry.khronos.org</a>
	 */
	public void glUniform3fARB(int location, float v0, float v1, float v2) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniform3fARB(\{location}, \{v0}, \{v1}, \{v2})");
		try {
		    functions.glUniform3fARB().invokeExact(location, v0, v1, v2);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniform3fv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniform3fv.xhtml">registry.khronos.org</a>
	 */
	public void glUniform3fv(int location, int count, ConstPointer<Float> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniform3fv(\{location}, \{count}, \{value})");
		try {
		    functions.glUniform3fv().invokeExact(location, count, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniform3fvARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniform3fvARB.xhtml">registry.khronos.org</a>
	 */
	public void glUniform3fvARB(int location, int count, ConstPointer<Float> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniform3fvARB(\{location}, \{count}, \{value})");
		try {
		    functions.glUniform3fvARB().invokeExact(location, count, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniform3i">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniform3i.xhtml">registry.khronos.org</a>
	 */
	public void glUniform3i(int location, int v0, int v1, int v2) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniform3i(\{location}, \{v0}, \{v1}, \{v2})");
		try {
		    functions.glUniform3i().invokeExact(location, v0, v1, v2);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniform3i64ARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniform3i64ARB.xhtml">registry.khronos.org</a>
	 */
	public void glUniform3i64ARB(int location, long x, long y, long z) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniform3i64ARB(\{location}, \{x}, \{y}, \{z})");
		try {
		    functions.glUniform3i64ARB().invokeExact(location, x, y, z);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniform3i64NV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniform3i64NV.xhtml">registry.khronos.org</a>
	 */
	public void glUniform3i64NV(int location, long x, long y, long z) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniform3i64NV(\{location}, \{x}, \{y}, \{z})");
		try {
		    functions.glUniform3i64NV().invokeExact(location, x, y, z);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniform3i64vARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniform3i64vARB.xhtml">registry.khronos.org</a>
	 */
	public void glUniform3i64vARB(int location, int count, ConstPointer<Long> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniform3i64vARB(\{location}, \{count}, \{value})");
		try {
		    functions.glUniform3i64vARB().invokeExact(location, count, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniform3i64vNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniform3i64vNV.xhtml">registry.khronos.org</a>
	 */
	public void glUniform3i64vNV(int location, int count, ConstPointer<Long> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniform3i64vNV(\{location}, \{count}, \{value})");
		try {
		    functions.glUniform3i64vNV().invokeExact(location, count, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniform3iARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniform3iARB.xhtml">registry.khronos.org</a>
	 */
	public void glUniform3iARB(int location, int v0, int v1, int v2) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniform3iARB(\{location}, \{v0}, \{v1}, \{v2})");
		try {
		    functions.glUniform3iARB().invokeExact(location, v0, v1, v2);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniform3iv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniform3iv.xhtml">registry.khronos.org</a>
	 */
	public void glUniform3iv(int location, int count, ConstPointer<Integer> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniform3iv(\{location}, \{count}, \{value})");
		try {
		    functions.glUniform3iv().invokeExact(location, count, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniform3ivARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniform3ivARB.xhtml">registry.khronos.org</a>
	 */
	public void glUniform3ivARB(int location, int count, ConstPointer<Integer> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniform3ivARB(\{location}, \{count}, \{value})");
		try {
		    functions.glUniform3ivARB().invokeExact(location, count, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniform3ui">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniform3ui.xhtml">registry.khronos.org</a>
	 */
	public void glUniform3ui(int location, int v0, int v1, int v2) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniform3ui(\{location}, \{v0}, \{v1}, \{v2})");
		try {
		    functions.glUniform3ui().invokeExact(location, v0, v1, v2);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniform3ui64ARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniform3ui64ARB.xhtml">registry.khronos.org</a>
	 */
	public void glUniform3ui64ARB(int location, long x, long y, long z) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniform3ui64ARB(\{location}, \{x}, \{y}, \{z})");
		try {
		    functions.glUniform3ui64ARB().invokeExact(location, x, y, z);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniform3ui64NV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniform3ui64NV.xhtml">registry.khronos.org</a>
	 */
	public void glUniform3ui64NV(int location, long x, long y, long z) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniform3ui64NV(\{location}, \{x}, \{y}, \{z})");
		try {
		    functions.glUniform3ui64NV().invokeExact(location, x, y, z);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniform3ui64vARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniform3ui64vARB.xhtml">registry.khronos.org</a>
	 */
	public void glUniform3ui64vARB(int location, int count, ConstPointer<Long> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniform3ui64vARB(\{location}, \{count}, \{value})");
		try {
		    functions.glUniform3ui64vARB().invokeExact(location, count, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniform3ui64vNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniform3ui64vNV.xhtml">registry.khronos.org</a>
	 */
	public void glUniform3ui64vNV(int location, int count, ConstPointer<Long> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniform3ui64vNV(\{location}, \{count}, \{value})");
		try {
		    functions.glUniform3ui64vNV().invokeExact(location, count, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniform3uiEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniform3uiEXT.xhtml">registry.khronos.org</a>
	 */
	public void glUniform3uiEXT(int location, int v0, int v1, int v2) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniform3uiEXT(\{location}, \{v0}, \{v1}, \{v2})");
		try {
		    functions.glUniform3uiEXT().invokeExact(location, v0, v1, v2);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniform3uiv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniform3uiv.xhtml">registry.khronos.org</a>
	 */
	public void glUniform3uiv(int location, int count, ConstPointer<Integer> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniform3uiv(\{location}, \{count}, \{value})");
		try {
		    functions.glUniform3uiv().invokeExact(location, count, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniform3uivEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniform3uivEXT.xhtml">registry.khronos.org</a>
	 */
	public void glUniform3uivEXT(int location, int count, ConstPointer<Integer> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniform3uivEXT(\{location}, \{count}, \{value})");
		try {
		    functions.glUniform3uivEXT().invokeExact(location, count, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniform4d">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniform4d.xhtml">registry.khronos.org</a>
	 */
	public void glUniform4d(int location, double x, double y, double z, double w) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniform4d(\{location}, \{x}, \{y}, \{z}, \{w})");
		try {
		    functions.glUniform4d().invokeExact(location, x, y, z, w);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniform4dv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniform4dv.xhtml">registry.khronos.org</a>
	 */
	public void glUniform4dv(int location, int count, ConstPointer<Double> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniform4dv(\{location}, \{count}, \{value})");
		try {
		    functions.glUniform4dv().invokeExact(location, count, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniform4f">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniform4f.xhtml">registry.khronos.org</a>
	 */
	public void glUniform4f(int location, float v0, float v1, float v2, float v3) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniform4f(\{location}, \{v0}, \{v1}, \{v2}, \{v3})");
		try {
		    functions.glUniform4f().invokeExact(location, v0, v1, v2, v3);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniform4fARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniform4fARB.xhtml">registry.khronos.org</a>
	 */
	public void glUniform4fARB(int location, float v0, float v1, float v2, float v3) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniform4fARB(\{location}, \{v0}, \{v1}, \{v2}, \{v3})");
		try {
		    functions.glUniform4fARB().invokeExact(location, v0, v1, v2, v3);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniform4fv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniform4fv.xhtml">registry.khronos.org</a>
	 */
	public void glUniform4fv(int location, int count, ConstPointer<Float> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniform4fv(\{location}, \{count}, \{value})");
		try {
		    functions.glUniform4fv().invokeExact(location, count, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniform4fvARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniform4fvARB.xhtml">registry.khronos.org</a>
	 */
	public void glUniform4fvARB(int location, int count, ConstPointer<Float> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniform4fvARB(\{location}, \{count}, \{value})");
		try {
		    functions.glUniform4fvARB().invokeExact(location, count, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniform4i">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniform4i.xhtml">registry.khronos.org</a>
	 */
	public void glUniform4i(int location, int v0, int v1, int v2, int v3) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniform4i(\{location}, \{v0}, \{v1}, \{v2}, \{v3})");
		try {
		    functions.glUniform4i().invokeExact(location, v0, v1, v2, v3);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniform4i64ARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniform4i64ARB.xhtml">registry.khronos.org</a>
	 */
	public void glUniform4i64ARB(int location, long x, long y, long z, long w) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniform4i64ARB(\{location}, \{x}, \{y}, \{z}, \{w})");
		try {
		    functions.glUniform4i64ARB().invokeExact(location, x, y, z, w);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniform4i64NV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniform4i64NV.xhtml">registry.khronos.org</a>
	 */
	public void glUniform4i64NV(int location, long x, long y, long z, long w) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniform4i64NV(\{location}, \{x}, \{y}, \{z}, \{w})");
		try {
		    functions.glUniform4i64NV().invokeExact(location, x, y, z, w);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniform4i64vARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniform4i64vARB.xhtml">registry.khronos.org</a>
	 */
	public void glUniform4i64vARB(int location, int count, ConstPointer<Long> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniform4i64vARB(\{location}, \{count}, \{value})");
		try {
		    functions.glUniform4i64vARB().invokeExact(location, count, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniform4i64vNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniform4i64vNV.xhtml">registry.khronos.org</a>
	 */
	public void glUniform4i64vNV(int location, int count, ConstPointer<Long> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniform4i64vNV(\{location}, \{count}, \{value})");
		try {
		    functions.glUniform4i64vNV().invokeExact(location, count, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniform4iARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniform4iARB.xhtml">registry.khronos.org</a>
	 */
	public void glUniform4iARB(int location, int v0, int v1, int v2, int v3) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniform4iARB(\{location}, \{v0}, \{v1}, \{v2}, \{v3})");
		try {
		    functions.glUniform4iARB().invokeExact(location, v0, v1, v2, v3);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniform4iv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniform4iv.xhtml">registry.khronos.org</a>
	 */
	public void glUniform4iv(int location, int count, ConstPointer<Integer> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniform4iv(\{location}, \{count}, \{value})");
		try {
		    functions.glUniform4iv().invokeExact(location, count, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniform4ivARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniform4ivARB.xhtml">registry.khronos.org</a>
	 */
	public void glUniform4ivARB(int location, int count, ConstPointer<Integer> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniform4ivARB(\{location}, \{count}, \{value})");
		try {
		    functions.glUniform4ivARB().invokeExact(location, count, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniform4ui">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniform4ui.xhtml">registry.khronos.org</a>
	 */
	public void glUniform4ui(int location, int v0, int v1, int v2, int v3) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniform4ui(\{location}, \{v0}, \{v1}, \{v2}, \{v3})");
		try {
		    functions.glUniform4ui().invokeExact(location, v0, v1, v2, v3);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniform4ui64ARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniform4ui64ARB.xhtml">registry.khronos.org</a>
	 */
	public void glUniform4ui64ARB(int location, long x, long y, long z, long w) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniform4ui64ARB(\{location}, \{x}, \{y}, \{z}, \{w})");
		try {
		    functions.glUniform4ui64ARB().invokeExact(location, x, y, z, w);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniform4ui64NV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniform4ui64NV.xhtml">registry.khronos.org</a>
	 */
	public void glUniform4ui64NV(int location, long x, long y, long z, long w) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniform4ui64NV(\{location}, \{x}, \{y}, \{z}, \{w})");
		try {
		    functions.glUniform4ui64NV().invokeExact(location, x, y, z, w);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniform4ui64vARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniform4ui64vARB.xhtml">registry.khronos.org</a>
	 */
	public void glUniform4ui64vARB(int location, int count, ConstPointer<Long> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniform4ui64vARB(\{location}, \{count}, \{value})");
		try {
		    functions.glUniform4ui64vARB().invokeExact(location, count, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniform4ui64vNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniform4ui64vNV.xhtml">registry.khronos.org</a>
	 */
	public void glUniform4ui64vNV(int location, int count, ConstPointer<Long> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniform4ui64vNV(\{location}, \{count}, \{value})");
		try {
		    functions.glUniform4ui64vNV().invokeExact(location, count, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniform4uiEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniform4uiEXT.xhtml">registry.khronos.org</a>
	 */
	public void glUniform4uiEXT(int location, int v0, int v1, int v2, int v3) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniform4uiEXT(\{location}, \{v0}, \{v1}, \{v2}, \{v3})");
		try {
		    functions.glUniform4uiEXT().invokeExact(location, v0, v1, v2, v3);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniform4uiv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniform4uiv.xhtml">registry.khronos.org</a>
	 */
	public void glUniform4uiv(int location, int count, ConstPointer<Integer> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniform4uiv(\{location}, \{count}, \{value})");
		try {
		    functions.glUniform4uiv().invokeExact(location, count, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniform4uivEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniform4uivEXT.xhtml">registry.khronos.org</a>
	 */
	public void glUniform4uivEXT(int location, int count, ConstPointer<Integer> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniform4uivEXT(\{location}, \{count}, \{value})");
		try {
		    functions.glUniform4uivEXT().invokeExact(location, count, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniformBlockBinding">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniformBlockBinding.xhtml">registry.khronos.org</a>
	 */
	public void glUniformBlockBinding(int program, int uniformBlockIndex, int uniformBlockBinding) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniformBlockBinding(\{program}, \{uniformBlockIndex}, \{uniformBlockBinding})");
		try {
		    functions.glUniformBlockBinding().invokeExact(program, uniformBlockIndex, uniformBlockBinding);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniformBufferEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniformBufferEXT.xhtml">registry.khronos.org</a>
	 */
	public void glUniformBufferEXT(int program, int location, int buffer) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniformBufferEXT(\{program}, \{location}, \{buffer})");
		try {
		    functions.glUniformBufferEXT().invokeExact(program, location, buffer);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniformHandleui64ARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniformHandleui64ARB.xhtml">registry.khronos.org</a>
	 */
	public void glUniformHandleui64ARB(int location, long value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniformHandleui64ARB(\{location}, \{value})");
		try {
		    functions.glUniformHandleui64ARB().invokeExact(location, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniformHandleui64IMG">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniformHandleui64IMG.xhtml">registry.khronos.org</a>
	 */
	public void glUniformHandleui64IMG(int location, long value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniformHandleui64IMG(\{location}, \{value})");
		try {
		    functions.glUniformHandleui64IMG().invokeExact(location, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniformHandleui64NV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniformHandleui64NV.xhtml">registry.khronos.org</a>
	 */
	public void glUniformHandleui64NV(int location, long value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniformHandleui64NV(\{location}, \{value})");
		try {
		    functions.glUniformHandleui64NV().invokeExact(location, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniformHandleui64vARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniformHandleui64vARB.xhtml">registry.khronos.org</a>
	 */
	public void glUniformHandleui64vARB(int location, int count, ConstPointer<Long> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniformHandleui64vARB(\{location}, \{count}, \{value})");
		try {
		    functions.glUniformHandleui64vARB().invokeExact(location, count, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniformHandleui64vIMG">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniformHandleui64vIMG.xhtml">registry.khronos.org</a>
	 */
	public void glUniformHandleui64vIMG(int location, int count, ConstPointer<Long> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniformHandleui64vIMG(\{location}, \{count}, \{value})");
		try {
		    functions.glUniformHandleui64vIMG().invokeExact(location, count, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniformHandleui64vNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniformHandleui64vNV.xhtml">registry.khronos.org</a>
	 */
	public void glUniformHandleui64vNV(int location, int count, ConstPointer<Long> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniformHandleui64vNV(\{location}, \{count}, \{value})");
		try {
		    functions.glUniformHandleui64vNV().invokeExact(location, count, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniformMatrix2dv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniformMatrix2dv.xhtml">registry.khronos.org</a>
	 */
	public void glUniformMatrix2dv(int location, int count, boolean transpose, ConstPointer<Double> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniformMatrix2dv(\{location}, \{count}, \{transpose}, \{value})");
		try {
		    functions.glUniformMatrix2dv().invokeExact(location, count, transpose, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniformMatrix2fv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniformMatrix2fv.xhtml">registry.khronos.org</a>
	 */
	public void glUniformMatrix2fv(int location, int count, boolean transpose, ConstPointer<Float> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniformMatrix2fv(\{location}, \{count}, \{transpose}, \{value})");
		try {
		    functions.glUniformMatrix2fv().invokeExact(location, count, transpose, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniformMatrix2fvARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniformMatrix2fvARB.xhtml">registry.khronos.org</a>
	 */
	public void glUniformMatrix2fvARB(int location, int count, boolean transpose, ConstPointer<Float> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniformMatrix2fvARB(\{location}, \{count}, \{transpose}, \{value})");
		try {
		    functions.glUniformMatrix2fvARB().invokeExact(location, count, transpose, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniformMatrix2x3dv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniformMatrix2x3dv.xhtml">registry.khronos.org</a>
	 */
	public void glUniformMatrix2x3dv(int location, int count, boolean transpose, ConstPointer<Double> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniformMatrix2x3dv(\{location}, \{count}, \{transpose}, \{value})");
		try {
		    functions.glUniformMatrix2x3dv().invokeExact(location, count, transpose, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniformMatrix2x3fv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniformMatrix2x3fv.xhtml">registry.khronos.org</a>
	 */
	public void glUniformMatrix2x3fv(int location, int count, boolean transpose, ConstPointer<Float> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniformMatrix2x3fv(\{location}, \{count}, \{transpose}, \{value})");
		try {
		    functions.glUniformMatrix2x3fv().invokeExact(location, count, transpose, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniformMatrix2x3fvNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniformMatrix2x3fvNV.xhtml">registry.khronos.org</a>
	 */
	public void glUniformMatrix2x3fvNV(int location, int count, boolean transpose, ConstPointer<Float> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniformMatrix2x3fvNV(\{location}, \{count}, \{transpose}, \{value})");
		try {
		    functions.glUniformMatrix2x3fvNV().invokeExact(location, count, transpose, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniformMatrix2x4dv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniformMatrix2x4dv.xhtml">registry.khronos.org</a>
	 */
	public void glUniformMatrix2x4dv(int location, int count, boolean transpose, ConstPointer<Double> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniformMatrix2x4dv(\{location}, \{count}, \{transpose}, \{value})");
		try {
		    functions.glUniformMatrix2x4dv().invokeExact(location, count, transpose, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniformMatrix2x4fv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniformMatrix2x4fv.xhtml">registry.khronos.org</a>
	 */
	public void glUniformMatrix2x4fv(int location, int count, boolean transpose, ConstPointer<Float> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniformMatrix2x4fv(\{location}, \{count}, \{transpose}, \{value})");
		try {
		    functions.glUniformMatrix2x4fv().invokeExact(location, count, transpose, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniformMatrix2x4fvNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniformMatrix2x4fvNV.xhtml">registry.khronos.org</a>
	 */
	public void glUniformMatrix2x4fvNV(int location, int count, boolean transpose, ConstPointer<Float> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniformMatrix2x4fvNV(\{location}, \{count}, \{transpose}, \{value})");
		try {
		    functions.glUniformMatrix2x4fvNV().invokeExact(location, count, transpose, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniformMatrix3dv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniformMatrix3dv.xhtml">registry.khronos.org</a>
	 */
	public void glUniformMatrix3dv(int location, int count, boolean transpose, ConstPointer<Double> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniformMatrix3dv(\{location}, \{count}, \{transpose}, \{value})");
		try {
		    functions.glUniformMatrix3dv().invokeExact(location, count, transpose, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniformMatrix3fv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniformMatrix3fv.xhtml">registry.khronos.org</a>
	 */
	public void glUniformMatrix3fv(int location, int count, boolean transpose, ConstPointer<Float> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniformMatrix3fv(\{location}, \{count}, \{transpose}, \{value})");
		try {
		    functions.glUniformMatrix3fv().invokeExact(location, count, transpose, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniformMatrix3fvARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniformMatrix3fvARB.xhtml">registry.khronos.org</a>
	 */
	public void glUniformMatrix3fvARB(int location, int count, boolean transpose, ConstPointer<Float> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniformMatrix3fvARB(\{location}, \{count}, \{transpose}, \{value})");
		try {
		    functions.glUniformMatrix3fvARB().invokeExact(location, count, transpose, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniformMatrix3x2dv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniformMatrix3x2dv.xhtml">registry.khronos.org</a>
	 */
	public void glUniformMatrix3x2dv(int location, int count, boolean transpose, ConstPointer<Double> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniformMatrix3x2dv(\{location}, \{count}, \{transpose}, \{value})");
		try {
		    functions.glUniformMatrix3x2dv().invokeExact(location, count, transpose, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniformMatrix3x2fv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniformMatrix3x2fv.xhtml">registry.khronos.org</a>
	 */
	public void glUniformMatrix3x2fv(int location, int count, boolean transpose, ConstPointer<Float> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniformMatrix3x2fv(\{location}, \{count}, \{transpose}, \{value})");
		try {
		    functions.glUniformMatrix3x2fv().invokeExact(location, count, transpose, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniformMatrix3x2fvNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniformMatrix3x2fvNV.xhtml">registry.khronos.org</a>
	 */
	public void glUniformMatrix3x2fvNV(int location, int count, boolean transpose, ConstPointer<Float> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniformMatrix3x2fvNV(\{location}, \{count}, \{transpose}, \{value})");
		try {
		    functions.glUniformMatrix3x2fvNV().invokeExact(location, count, transpose, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniformMatrix3x4dv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniformMatrix3x4dv.xhtml">registry.khronos.org</a>
	 */
	public void glUniformMatrix3x4dv(int location, int count, boolean transpose, ConstPointer<Double> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniformMatrix3x4dv(\{location}, \{count}, \{transpose}, \{value})");
		try {
		    functions.glUniformMatrix3x4dv().invokeExact(location, count, transpose, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniformMatrix3x4fv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniformMatrix3x4fv.xhtml">registry.khronos.org</a>
	 */
	public void glUniformMatrix3x4fv(int location, int count, boolean transpose, ConstPointer<Float> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniformMatrix3x4fv(\{location}, \{count}, \{transpose}, \{value})");
		try {
		    functions.glUniformMatrix3x4fv().invokeExact(location, count, transpose, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniformMatrix3x4fvNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniformMatrix3x4fvNV.xhtml">registry.khronos.org</a>
	 */
	public void glUniformMatrix3x4fvNV(int location, int count, boolean transpose, ConstPointer<Float> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniformMatrix3x4fvNV(\{location}, \{count}, \{transpose}, \{value})");
		try {
		    functions.glUniformMatrix3x4fvNV().invokeExact(location, count, transpose, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniformMatrix4dv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniformMatrix4dv.xhtml">registry.khronos.org</a>
	 */
	public void glUniformMatrix4dv(int location, int count, boolean transpose, ConstPointer<Double> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniformMatrix4dv(\{location}, \{count}, \{transpose}, \{value})");
		try {
		    functions.glUniformMatrix4dv().invokeExact(location, count, transpose, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniformMatrix4fv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniformMatrix4fv.xhtml">registry.khronos.org</a>
	 */
	public void glUniformMatrix4fv(int location, int count, boolean transpose, ConstPointer<Float> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniformMatrix4fv(\{location}, \{count}, \{transpose}, \{value})");
		try {
		    functions.glUniformMatrix4fv().invokeExact(location, count, transpose, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniformMatrix4fvARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniformMatrix4fvARB.xhtml">registry.khronos.org</a>
	 */
	public void glUniformMatrix4fvARB(int location, int count, boolean transpose, ConstPointer<Float> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniformMatrix4fvARB(\{location}, \{count}, \{transpose}, \{value})");
		try {
		    functions.glUniformMatrix4fvARB().invokeExact(location, count, transpose, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniformMatrix4x2dv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniformMatrix4x2dv.xhtml">registry.khronos.org</a>
	 */
	public void glUniformMatrix4x2dv(int location, int count, boolean transpose, ConstPointer<Double> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniformMatrix4x2dv(\{location}, \{count}, \{transpose}, \{value})");
		try {
		    functions.glUniformMatrix4x2dv().invokeExact(location, count, transpose, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniformMatrix4x2fv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniformMatrix4x2fv.xhtml">registry.khronos.org</a>
	 */
	public void glUniformMatrix4x2fv(int location, int count, boolean transpose, ConstPointer<Float> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniformMatrix4x2fv(\{location}, \{count}, \{transpose}, \{value})");
		try {
		    functions.glUniformMatrix4x2fv().invokeExact(location, count, transpose, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniformMatrix4x2fvNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniformMatrix4x2fvNV.xhtml">registry.khronos.org</a>
	 */
	public void glUniformMatrix4x2fvNV(int location, int count, boolean transpose, ConstPointer<Float> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniformMatrix4x2fvNV(\{location}, \{count}, \{transpose}, \{value})");
		try {
		    functions.glUniformMatrix4x2fvNV().invokeExact(location, count, transpose, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniformMatrix4x3dv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniformMatrix4x3dv.xhtml">registry.khronos.org</a>
	 */
	public void glUniformMatrix4x3dv(int location, int count, boolean transpose, ConstPointer<Double> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniformMatrix4x3dv(\{location}, \{count}, \{transpose}, \{value})");
		try {
		    functions.glUniformMatrix4x3dv().invokeExact(location, count, transpose, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniformMatrix4x3fv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniformMatrix4x3fv.xhtml">registry.khronos.org</a>
	 */
	public void glUniformMatrix4x3fv(int location, int count, boolean transpose, ConstPointer<Float> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniformMatrix4x3fv(\{location}, \{count}, \{transpose}, \{value})");
		try {
		    functions.glUniformMatrix4x3fv().invokeExact(location, count, transpose, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniformMatrix4x3fvNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniformMatrix4x3fvNV.xhtml">registry.khronos.org</a>
	 */
	public void glUniformMatrix4x3fvNV(int location, int count, boolean transpose, ConstPointer<Float> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniformMatrix4x3fvNV(\{location}, \{count}, \{transpose}, \{value})");
		try {
		    functions.glUniformMatrix4x3fvNV().invokeExact(location, count, transpose, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniformSubroutinesuiv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniformSubroutinesuiv.xhtml">registry.khronos.org</a>
	 */
	public void glUniformSubroutinesuiv(GLShaderType shadertype, int count, ConstPointer<Integer> indices) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniformSubroutinesuiv(\{shadertype}, \{count}, \{indices})");
		try {
		    functions.glUniformSubroutinesuiv().invokeExact(shadertype, count, indices);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were shadertype=\{shadertype}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniformui64NV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniformui64NV.xhtml">registry.khronos.org</a>
	 */
	public void glUniformui64NV(int location, long value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniformui64NV(\{location}, \{value})");
		try {
		    functions.glUniformui64NV().invokeExact(location, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUniformui64vNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUniformui64vNV.xhtml">registry.khronos.org</a>
	 */
	public void glUniformui64vNV(int location, int count, ConstPointer<Long> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUniformui64vNV(\{location}, \{count}, \{value})");
		try {
		    functions.glUniformui64vNV().invokeExact(location, count, value);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUnlockArraysEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUnlockArraysEXT.xhtml">registry.khronos.org</a>
	 */
	public void glUnlockArraysEXT() {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUnlockArraysEXT()");
		try {
		    functions.glUnlockArraysEXT().invokeExact();
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUnmapBuffer">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUnmapBuffer.xhtml">registry.khronos.org</a>
	 */
	public boolean glUnmapBuffer(GLBufferTarget target) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUnmapBuffer(\{target})");
		try {
		    var r = (boolean)functions.glUnmapBuffer().invokeExact(target);
			checkError();
			return r;
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUnmapBufferARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUnmapBufferARB.xhtml">registry.khronos.org</a>
	 */
	public boolean glUnmapBufferARB(GLBufferTarget target) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUnmapBufferARB(\{target})");
		try {
		    var r = (boolean)functions.glUnmapBufferARB().invokeExact(target);
			checkError();
			return r;
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUnmapBufferOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUnmapBufferOES.xhtml">registry.khronos.org</a>
	 */
	public boolean glUnmapBufferOES(GLEnum target) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUnmapBufferOES(\{target})");
		try {
		    var r = (boolean)functions.glUnmapBufferOES().invokeExact(target);
			checkError();
			return r;
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUnmapNamedBuffer">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUnmapNamedBuffer.xhtml">registry.khronos.org</a>
	 */
	public boolean glUnmapNamedBuffer(int buffer) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUnmapNamedBuffer(\{buffer})");
		try {
		    var r = (boolean)functions.glUnmapNamedBuffer().invokeExact(buffer);
			checkError();
			return r;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUnmapNamedBufferEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUnmapNamedBufferEXT.xhtml">registry.khronos.org</a>
	 */
	public boolean glUnmapNamedBufferEXT(int buffer) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUnmapNamedBufferEXT(\{buffer})");
		try {
		    var r = (boolean)functions.glUnmapNamedBufferEXT().invokeExact(buffer);
			checkError();
			return r;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUnmapObjectBufferATI">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUnmapObjectBufferATI.xhtml">registry.khronos.org</a>
	 */
	public void glUnmapObjectBufferATI(int buffer) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUnmapObjectBufferATI(\{buffer})");
		try {
		    functions.glUnmapObjectBufferATI().invokeExact(buffer);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUnmapTexture2DINTEL">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUnmapTexture2DINTEL.xhtml">registry.khronos.org</a>
	 */
	public void glUnmapTexture2DINTEL(int texture, int level) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUnmapTexture2DINTEL(\{texture}, \{level})");
		try {
		    functions.glUnmapTexture2DINTEL().invokeExact(texture, level);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUpdateObjectBufferATI">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUpdateObjectBufferATI.xhtml">registry.khronos.org</a>
	 */
	public void glUpdateObjectBufferATI(int buffer, int offset, int size, ConstPointer<Void> pointer, GLPreserveModeATI preserve) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUpdateObjectBufferATI(\{buffer}, \{offset}, \{size}, \{pointer}, \{preserve})");
		try {
		    functions.glUpdateObjectBufferATI().invokeExact(buffer, offset, size, pointer, preserve);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were preserve=\{preserve}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUploadGpuMaskNVX">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUploadGpuMaskNVX.xhtml">registry.khronos.org</a>
	 */
	public void glUploadGpuMaskNVX(int mask) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUploadGpuMaskNVX(\{mask})");
		try {
		    functions.glUploadGpuMaskNVX().invokeExact(mask);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUseProgram">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUseProgram.xhtml">registry.khronos.org</a>
	 */
	public void glUseProgram(int program) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUseProgram(\{program})");
		try {
		    functions.glUseProgram().invokeExact(program);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUseProgramObjectARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUseProgramObjectARB.xhtml">registry.khronos.org</a>
	 */
	public void glUseProgramObjectARB(int programObj) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUseProgramObjectARB(\{programObj})");
		try {
		    functions.glUseProgramObjectARB().invokeExact(programObj);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUseProgramStages">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUseProgramStages.xhtml">registry.khronos.org</a>
	 */
	public void glUseProgramStages(int pipeline, int stages, int program) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUseProgramStages(\{pipeline}, \{stages}, \{program})");
		try {
		    functions.glUseProgramStages().invokeExact(pipeline, stages, program);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUseProgramStagesEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUseProgramStagesEXT.xhtml">registry.khronos.org</a>
	 */
	public void glUseProgramStagesEXT(int pipeline, int stages, int program) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUseProgramStagesEXT(\{pipeline}, \{stages}, \{program})");
		try {
		    functions.glUseProgramStagesEXT().invokeExact(pipeline, stages, program);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glUseShaderProgramEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glUseShaderProgramEXT.xhtml">registry.khronos.org</a>
	 */
	public void glUseShaderProgramEXT(GLEnum type, int program) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glUseShaderProgramEXT(\{type}, \{program})");
		try {
		    functions.glUseShaderProgramEXT().invokeExact(type, program);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVDPAUFiniNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVDPAUFiniNV.xhtml">registry.khronos.org</a>
	 */
	public void glVDPAUFiniNV() {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVDPAUFiniNV()");
		try {
		    functions.glVDPAUFiniNV().invokeExact();
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVDPAUGetSurfaceivNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVDPAUGetSurfaceivNV.xhtml">registry.khronos.org</a>
	 */
	public void glVDPAUGetSurfaceivNV(int surface, GLEnum pname, int count, MutPointer<Integer> length, MutPointer<Integer> values) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVDPAUGetSurfaceivNV(\{surface}, \{pname}, \{count}, \{length}, \{values})");
		try {
		    functions.glVDPAUGetSurfaceivNV().invokeExact(surface, pname, count, length, values);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVDPAUInitNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVDPAUInitNV.xhtml">registry.khronos.org</a>
	 */
	public void glVDPAUInitNV(ConstPointer<Void> vdpDevice, ConstPointer<Void> getProcAddress) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVDPAUInitNV(\{vdpDevice}, \{getProcAddress})");
		try {
		    functions.glVDPAUInitNV().invokeExact(vdpDevice, getProcAddress);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVDPAUIsSurfaceNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVDPAUIsSurfaceNV.xhtml">registry.khronos.org</a>
	 */
	public boolean glVDPAUIsSurfaceNV(int surface) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVDPAUIsSurfaceNV(\{surface})");
		try {
		    var r = (boolean)functions.glVDPAUIsSurfaceNV().invokeExact(surface);
			checkError();
			return r;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVDPAUMapSurfacesNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVDPAUMapSurfacesNV.xhtml">registry.khronos.org</a>
	 */
	public void glVDPAUMapSurfacesNV(int numSurfaces, ConstPointer<Integer> surfaces) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVDPAUMapSurfacesNV(\{numSurfaces}, \{surfaces})");
		try {
		    functions.glVDPAUMapSurfacesNV().invokeExact(numSurfaces, surfaces);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVDPAURegisterOutputSurfaceNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVDPAURegisterOutputSurfaceNV.xhtml">registry.khronos.org</a>
	 */
	public int glVDPAURegisterOutputSurfaceNV(ConstPointer<Void> vdpSurface, GLEnum target, int numTextureNames, ConstPointer<Integer> textureNames) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVDPAURegisterOutputSurfaceNV(\{vdpSurface}, \{target}, \{numTextureNames}, \{textureNames})");
		try {
		    var r = (int)functions.glVDPAURegisterOutputSurfaceNV().invokeExact(vdpSurface, target, numTextureNames, textureNames);
			checkError();
			return r;
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVDPAURegisterVideoSurfaceNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVDPAURegisterVideoSurfaceNV.xhtml">registry.khronos.org</a>
	 */
	public int glVDPAURegisterVideoSurfaceNV(ConstPointer<Void> vdpSurface, GLEnum target, int numTextureNames, ConstPointer<Integer> textureNames) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVDPAURegisterVideoSurfaceNV(\{vdpSurface}, \{target}, \{numTextureNames}, \{textureNames})");
		try {
		    var r = (int)functions.glVDPAURegisterVideoSurfaceNV().invokeExact(vdpSurface, target, numTextureNames, textureNames);
			checkError();
			return r;
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVDPAURegisterVideoSurfaceWithPictureStructureNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVDPAURegisterVideoSurfaceWithPictureStructureNV.xhtml">registry.khronos.org</a>
	 */
	public int glVDPAURegisterVideoSurfaceWithPictureStructureNV(ConstPointer<Void> vdpSurface, GLEnum target, int numTextureNames, ConstPointer<Integer> textureNames, boolean isFrameStructure) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVDPAURegisterVideoSurfaceWithPictureStructureNV(\{vdpSurface}, \{target}, \{numTextureNames}, \{textureNames}, \{isFrameStructure})");
		try {
		    var r = (int)functions.glVDPAURegisterVideoSurfaceWithPictureStructureNV().invokeExact(vdpSurface, target, numTextureNames, textureNames, isFrameStructure);
			checkError();
			return r;
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}.");
			}
			throw e;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVDPAUSurfaceAccessNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVDPAUSurfaceAccessNV.xhtml">registry.khronos.org</a>
	 */
	public void glVDPAUSurfaceAccessNV(int surface, GLEnum access) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVDPAUSurfaceAccessNV(\{surface}, \{access})");
		try {
		    functions.glVDPAUSurfaceAccessNV().invokeExact(surface, access);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were access=\{access}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVDPAUUnmapSurfacesNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVDPAUUnmapSurfacesNV.xhtml">registry.khronos.org</a>
	 */
	public void glVDPAUUnmapSurfacesNV(int numSurface, ConstPointer<Integer> surfaces) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVDPAUUnmapSurfacesNV(\{numSurface}, \{surfaces})");
		try {
		    functions.glVDPAUUnmapSurfacesNV().invokeExact(numSurface, surfaces);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVDPAUUnregisterSurfaceNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVDPAUUnregisterSurfaceNV.xhtml">registry.khronos.org</a>
	 */
	public void glVDPAUUnregisterSurfaceNV(int surface) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVDPAUUnregisterSurfaceNV(\{surface})");
		try {
		    functions.glVDPAUUnregisterSurfaceNV().invokeExact(surface);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glValidateProgram">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glValidateProgram.xhtml">registry.khronos.org</a>
	 */
	public void glValidateProgram(int program) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glValidateProgram(\{program})");
		try {
		    functions.glValidateProgram().invokeExact(program);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glValidateProgramARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glValidateProgramARB.xhtml">registry.khronos.org</a>
	 */
	public void glValidateProgramARB(int programObj) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glValidateProgramARB(\{programObj})");
		try {
		    functions.glValidateProgramARB().invokeExact(programObj);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glValidateProgramPipeline">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glValidateProgramPipeline.xhtml">registry.khronos.org</a>
	 */
	public void glValidateProgramPipeline(int pipeline) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glValidateProgramPipeline(\{pipeline})");
		try {
		    functions.glValidateProgramPipeline().invokeExact(pipeline);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glValidateProgramPipelineEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glValidateProgramPipelineEXT.xhtml">registry.khronos.org</a>
	 */
	public void glValidateProgramPipelineEXT(int pipeline) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glValidateProgramPipelineEXT(\{pipeline})");
		try {
		    functions.glValidateProgramPipelineEXT().invokeExact(pipeline);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVariantArrayObjectATI">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVariantArrayObjectATI.xhtml">registry.khronos.org</a>
	 */
	public void glVariantArrayObjectATI(int id, GLScalarType type, int stride, int buffer, int offset) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVariantArrayObjectATI(\{id}, \{type}, \{stride}, \{buffer}, \{offset})");
		try {
		    functions.glVariantArrayObjectATI().invokeExact(id, type, stride, buffer, offset);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVariantPointerEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVariantPointerEXT.xhtml">registry.khronos.org</a>
	 */
	public void glVariantPointerEXT(int id, GLScalarType type, int stride, ConstPointer<Void> addr) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVariantPointerEXT(\{id}, \{type}, \{stride}, \{addr})");
		try {
		    functions.glVariantPointerEXT().invokeExact(id, type, stride, addr);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVariantbvEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVariantbvEXT.xhtml">registry.khronos.org</a>
	 */
	public void glVariantbvEXT(int id, ConstPointer<Byte> addr) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVariantbvEXT(\{id}, \{addr})");
		try {
		    functions.glVariantbvEXT().invokeExact(id, addr);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVariantdvEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVariantdvEXT.xhtml">registry.khronos.org</a>
	 */
	public void glVariantdvEXT(int id, ConstPointer<Double> addr) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVariantdvEXT(\{id}, \{addr})");
		try {
		    functions.glVariantdvEXT().invokeExact(id, addr);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVariantfvEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVariantfvEXT.xhtml">registry.khronos.org</a>
	 */
	public void glVariantfvEXT(int id, ConstPointer<Float> addr) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVariantfvEXT(\{id}, \{addr})");
		try {
		    functions.glVariantfvEXT().invokeExact(id, addr);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVariantivEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVariantivEXT.xhtml">registry.khronos.org</a>
	 */
	public void glVariantivEXT(int id, ConstPointer<Integer> addr) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVariantivEXT(\{id}, \{addr})");
		try {
		    functions.glVariantivEXT().invokeExact(id, addr);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVariantsvEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVariantsvEXT.xhtml">registry.khronos.org</a>
	 */
	public void glVariantsvEXT(int id, ConstPointer<Short> addr) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVariantsvEXT(\{id}, \{addr})");
		try {
		    functions.glVariantsvEXT().invokeExact(id, addr);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVariantubvEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVariantubvEXT.xhtml">registry.khronos.org</a>
	 */
	public void glVariantubvEXT(int id, ConstPointer<Byte> addr) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVariantubvEXT(\{id}, \{addr})");
		try {
		    functions.glVariantubvEXT().invokeExact(id, addr);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVariantuivEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVariantuivEXT.xhtml">registry.khronos.org</a>
	 */
	public void glVariantuivEXT(int id, ConstPointer<Integer> addr) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVariantuivEXT(\{id}, \{addr})");
		try {
		    functions.glVariantuivEXT().invokeExact(id, addr);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVariantusvEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVariantusvEXT.xhtml">registry.khronos.org</a>
	 */
	public void glVariantusvEXT(int id, ConstPointer<Character> addr) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVariantusvEXT(\{id}, \{addr})");
		try {
		    functions.glVariantusvEXT().invokeExact(id, addr);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertex2bOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertex2bOES.xhtml">registry.khronos.org</a>
	 */
	public void glVertex2bOES(byte x, byte y) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertex2bOES(\{x}, \{y})");
		try {
		    functions.glVertex2bOES().invokeExact(x, y);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertex2bvOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertex2bvOES.xhtml">registry.khronos.org</a>
	 */
	public void glVertex2bvOES(ConstPointer<Byte> coords) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertex2bvOES(\{coords})");
		try {
		    functions.glVertex2bvOES().invokeExact(coords);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertex2d">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertex2d.xhtml">registry.khronos.org</a>
	 */
	public void glVertex2d(double x, double y) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertex2d(\{x}, \{y})");
		try {
		    functions.glVertex2d().invokeExact(x, y);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertex2dv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertex2dv.xhtml">registry.khronos.org</a>
	 */
	public void glVertex2dv(ConstPointer<Double> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertex2dv(\{v})");
		try {
		    functions.glVertex2dv().invokeExact(v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertex2f">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertex2f.xhtml">registry.khronos.org</a>
	 */
	public void glVertex2f(float x, float y) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertex2f(\{x}, \{y})");
		try {
		    functions.glVertex2f().invokeExact(x, y);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertex2fv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertex2fv.xhtml">registry.khronos.org</a>
	 */
	public void glVertex2fv(ConstPointer<Float> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertex2fv(\{v})");
		try {
		    functions.glVertex2fv().invokeExact(v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertex2hNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertex2hNV.xhtml">registry.khronos.org</a>
	 */
	public void glVertex2hNV(char x, char y) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertex2hNV(\{x}, \{y})");
		try {
		    functions.glVertex2hNV().invokeExact(x, y);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertex2hvNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertex2hvNV.xhtml">registry.khronos.org</a>
	 */
	public void glVertex2hvNV(ConstPointer<Character> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertex2hvNV(\{v})");
		try {
		    functions.glVertex2hvNV().invokeExact(v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertex2i">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertex2i.xhtml">registry.khronos.org</a>
	 */
	public void glVertex2i(int x, int y) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertex2i(\{x}, \{y})");
		try {
		    functions.glVertex2i().invokeExact(x, y);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertex2iv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertex2iv.xhtml">registry.khronos.org</a>
	 */
	public void glVertex2iv(ConstPointer<Integer> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertex2iv(\{v})");
		try {
		    functions.glVertex2iv().invokeExact(v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertex2s">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertex2s.xhtml">registry.khronos.org</a>
	 */
	public void glVertex2s(short x, short y) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertex2s(\{x}, \{y})");
		try {
		    functions.glVertex2s().invokeExact(x, y);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertex2sv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertex2sv.xhtml">registry.khronos.org</a>
	 */
	public void glVertex2sv(ConstPointer<Short> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertex2sv(\{v})");
		try {
		    functions.glVertex2sv().invokeExact(v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertex2xOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertex2xOES.xhtml">registry.khronos.org</a>
	 */
	public void glVertex2xOES(int x) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertex2xOES(\{x})");
		try {
		    functions.glVertex2xOES().invokeExact(x);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertex2xvOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertex2xvOES.xhtml">registry.khronos.org</a>
	 */
	public void glVertex2xvOES(ConstPointer<Integer> coords) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertex2xvOES(\{coords})");
		try {
		    functions.glVertex2xvOES().invokeExact(coords);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertex3bOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertex3bOES.xhtml">registry.khronos.org</a>
	 */
	public void glVertex3bOES(byte x, byte y, byte z) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertex3bOES(\{x}, \{y}, \{z})");
		try {
		    functions.glVertex3bOES().invokeExact(x, y, z);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertex3bvOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertex3bvOES.xhtml">registry.khronos.org</a>
	 */
	public void glVertex3bvOES(ConstPointer<Byte> coords) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertex3bvOES(\{coords})");
		try {
		    functions.glVertex3bvOES().invokeExact(coords);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertex3d">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertex3d.xhtml">registry.khronos.org</a>
	 */
	public void glVertex3d(double x, double y, double z) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertex3d(\{x}, \{y}, \{z})");
		try {
		    functions.glVertex3d().invokeExact(x, y, z);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertex3dv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertex3dv.xhtml">registry.khronos.org</a>
	 */
	public void glVertex3dv(ConstPointer<Double> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertex3dv(\{v})");
		try {
		    functions.glVertex3dv().invokeExact(v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertex3f">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertex3f.xhtml">registry.khronos.org</a>
	 */
	public void glVertex3f(float x, float y, float z) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertex3f(\{x}, \{y}, \{z})");
		try {
		    functions.glVertex3f().invokeExact(x, y, z);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertex3fv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertex3fv.xhtml">registry.khronos.org</a>
	 */
	public void glVertex3fv(ConstPointer<Float> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertex3fv(\{v})");
		try {
		    functions.glVertex3fv().invokeExact(v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertex3hNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertex3hNV.xhtml">registry.khronos.org</a>
	 */
	public void glVertex3hNV(char x, char y, char z) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertex3hNV(\{x}, \{y}, \{z})");
		try {
		    functions.glVertex3hNV().invokeExact(x, y, z);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertex3hvNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertex3hvNV.xhtml">registry.khronos.org</a>
	 */
	public void glVertex3hvNV(ConstPointer<Character> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertex3hvNV(\{v})");
		try {
		    functions.glVertex3hvNV().invokeExact(v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertex3i">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertex3i.xhtml">registry.khronos.org</a>
	 */
	public void glVertex3i(int x, int y, int z) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertex3i(\{x}, \{y}, \{z})");
		try {
		    functions.glVertex3i().invokeExact(x, y, z);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertex3iv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertex3iv.xhtml">registry.khronos.org</a>
	 */
	public void glVertex3iv(ConstPointer<Integer> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertex3iv(\{v})");
		try {
		    functions.glVertex3iv().invokeExact(v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertex3s">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertex3s.xhtml">registry.khronos.org</a>
	 */
	public void glVertex3s(short x, short y, short z) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertex3s(\{x}, \{y}, \{z})");
		try {
		    functions.glVertex3s().invokeExact(x, y, z);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertex3sv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertex3sv.xhtml">registry.khronos.org</a>
	 */
	public void glVertex3sv(ConstPointer<Short> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertex3sv(\{v})");
		try {
		    functions.glVertex3sv().invokeExact(v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertex3xOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertex3xOES.xhtml">registry.khronos.org</a>
	 */
	public void glVertex3xOES(int x, int y) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertex3xOES(\{x}, \{y})");
		try {
		    functions.glVertex3xOES().invokeExact(x, y);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertex3xvOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertex3xvOES.xhtml">registry.khronos.org</a>
	 */
	public void glVertex3xvOES(ConstPointer<Integer> coords) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertex3xvOES(\{coords})");
		try {
		    functions.glVertex3xvOES().invokeExact(coords);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertex4bOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertex4bOES.xhtml">registry.khronos.org</a>
	 */
	public void glVertex4bOES(byte x, byte y, byte z, byte w) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertex4bOES(\{x}, \{y}, \{z}, \{w})");
		try {
		    functions.glVertex4bOES().invokeExact(x, y, z, w);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertex4bvOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertex4bvOES.xhtml">registry.khronos.org</a>
	 */
	public void glVertex4bvOES(ConstPointer<Byte> coords) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertex4bvOES(\{coords})");
		try {
		    functions.glVertex4bvOES().invokeExact(coords);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertex4d">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertex4d.xhtml">registry.khronos.org</a>
	 */
	public void glVertex4d(double x, double y, double z, double w) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertex4d(\{x}, \{y}, \{z}, \{w})");
		try {
		    functions.glVertex4d().invokeExact(x, y, z, w);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertex4dv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertex4dv.xhtml">registry.khronos.org</a>
	 */
	public void glVertex4dv(ConstPointer<Double> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertex4dv(\{v})");
		try {
		    functions.glVertex4dv().invokeExact(v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertex4f">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertex4f.xhtml">registry.khronos.org</a>
	 */
	public void glVertex4f(float x, float y, float z, float w) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertex4f(\{x}, \{y}, \{z}, \{w})");
		try {
		    functions.glVertex4f().invokeExact(x, y, z, w);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertex4fv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertex4fv.xhtml">registry.khronos.org</a>
	 */
	public void glVertex4fv(ConstPointer<Float> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertex4fv(\{v})");
		try {
		    functions.glVertex4fv().invokeExact(v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertex4hNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertex4hNV.xhtml">registry.khronos.org</a>
	 */
	public void glVertex4hNV(char x, char y, char z, char w) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertex4hNV(\{x}, \{y}, \{z}, \{w})");
		try {
		    functions.glVertex4hNV().invokeExact(x, y, z, w);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertex4hvNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertex4hvNV.xhtml">registry.khronos.org</a>
	 */
	public void glVertex4hvNV(ConstPointer<Character> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertex4hvNV(\{v})");
		try {
		    functions.glVertex4hvNV().invokeExact(v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertex4i">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertex4i.xhtml">registry.khronos.org</a>
	 */
	public void glVertex4i(int x, int y, int z, int w) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertex4i(\{x}, \{y}, \{z}, \{w})");
		try {
		    functions.glVertex4i().invokeExact(x, y, z, w);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertex4iv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertex4iv.xhtml">registry.khronos.org</a>
	 */
	public void glVertex4iv(ConstPointer<Integer> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertex4iv(\{v})");
		try {
		    functions.glVertex4iv().invokeExact(v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertex4s">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertex4s.xhtml">registry.khronos.org</a>
	 */
	public void glVertex4s(short x, short y, short z, short w) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertex4s(\{x}, \{y}, \{z}, \{w})");
		try {
		    functions.glVertex4s().invokeExact(x, y, z, w);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertex4sv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertex4sv.xhtml">registry.khronos.org</a>
	 */
	public void glVertex4sv(ConstPointer<Short> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertex4sv(\{v})");
		try {
		    functions.glVertex4sv().invokeExact(v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertex4xOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertex4xOES.xhtml">registry.khronos.org</a>
	 */
	public void glVertex4xOES(int x, int y, int z) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertex4xOES(\{x}, \{y}, \{z})");
		try {
		    functions.glVertex4xOES().invokeExact(x, y, z);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertex4xvOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertex4xvOES.xhtml">registry.khronos.org</a>
	 */
	public void glVertex4xvOES(ConstPointer<Integer> coords) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertex4xvOES(\{coords})");
		try {
		    functions.glVertex4xvOES().invokeExact(coords);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexArrayAttribBinding">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexArrayAttribBinding.xhtml">registry.khronos.org</a>
	 */
	public void glVertexArrayAttribBinding(int vaobj, int attribindex, int bindingindex) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexArrayAttribBinding(\{vaobj}, \{attribindex}, \{bindingindex})");
		try {
		    functions.glVertexArrayAttribBinding().invokeExact(vaobj, attribindex, bindingindex);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexArrayAttribFormat">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexArrayAttribFormat.xhtml">registry.khronos.org</a>
	 */
	public void glVertexArrayAttribFormat(int vaobj, int attribindex, int size, GLVertexAttribType type, boolean normalized, int relativeoffset) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexArrayAttribFormat(\{vaobj}, \{attribindex}, \{size}, \{type}, \{normalized}, \{relativeoffset})");
		try {
		    functions.glVertexArrayAttribFormat().invokeExact(vaobj, attribindex, size, type, normalized, relativeoffset);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexArrayAttribIFormat">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexArrayAttribIFormat.xhtml">registry.khronos.org</a>
	 */
	public void glVertexArrayAttribIFormat(int vaobj, int attribindex, int size, GLVertexAttribIType type, int relativeoffset) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexArrayAttribIFormat(\{vaobj}, \{attribindex}, \{size}, \{type}, \{relativeoffset})");
		try {
		    functions.glVertexArrayAttribIFormat().invokeExact(vaobj, attribindex, size, type, relativeoffset);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexArrayAttribLFormat">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexArrayAttribLFormat.xhtml">registry.khronos.org</a>
	 */
	public void glVertexArrayAttribLFormat(int vaobj, int attribindex, int size, GLVertexAttribLType type, int relativeoffset) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexArrayAttribLFormat(\{vaobj}, \{attribindex}, \{size}, \{type}, \{relativeoffset})");
		try {
		    functions.glVertexArrayAttribLFormat().invokeExact(vaobj, attribindex, size, type, relativeoffset);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexArrayBindVertexBufferEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexArrayBindVertexBufferEXT.xhtml">registry.khronos.org</a>
	 */
	public void glVertexArrayBindVertexBufferEXT(int vaobj, int bindingindex, int buffer, long offset, int stride) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexArrayBindVertexBufferEXT(\{vaobj}, \{bindingindex}, \{buffer}, \{offset}, \{stride})");
		try {
		    functions.glVertexArrayBindVertexBufferEXT().invokeExact(vaobj, bindingindex, buffer, offset, stride);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexArrayBindingDivisor">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexArrayBindingDivisor.xhtml">registry.khronos.org</a>
	 */
	public void glVertexArrayBindingDivisor(int vaobj, int bindingindex, int divisor) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexArrayBindingDivisor(\{vaobj}, \{bindingindex}, \{divisor})");
		try {
		    functions.glVertexArrayBindingDivisor().invokeExact(vaobj, bindingindex, divisor);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexArrayColorOffsetEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexArrayColorOffsetEXT.xhtml">registry.khronos.org</a>
	 */
	public void glVertexArrayColorOffsetEXT(int vaobj, int buffer, int size, GLColorPointerType type, int stride, long offset) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexArrayColorOffsetEXT(\{vaobj}, \{buffer}, \{size}, \{type}, \{stride}, \{offset})");
		try {
		    functions.glVertexArrayColorOffsetEXT().invokeExact(vaobj, buffer, size, type, stride, offset);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexArrayEdgeFlagOffsetEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexArrayEdgeFlagOffsetEXT.xhtml">registry.khronos.org</a>
	 */
	public void glVertexArrayEdgeFlagOffsetEXT(int vaobj, int buffer, int stride, long offset) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexArrayEdgeFlagOffsetEXT(\{vaobj}, \{buffer}, \{stride}, \{offset})");
		try {
		    functions.glVertexArrayEdgeFlagOffsetEXT().invokeExact(vaobj, buffer, stride, offset);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexArrayElementBuffer">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexArrayElementBuffer.xhtml">registry.khronos.org</a>
	 */
	public void glVertexArrayElementBuffer(int vaobj, int buffer) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexArrayElementBuffer(\{vaobj}, \{buffer})");
		try {
		    functions.glVertexArrayElementBuffer().invokeExact(vaobj, buffer);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexArrayFogCoordOffsetEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexArrayFogCoordOffsetEXT.xhtml">registry.khronos.org</a>
	 */
	public void glVertexArrayFogCoordOffsetEXT(int vaobj, int buffer, GLFogCoordinatePointerType type, int stride, long offset) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexArrayFogCoordOffsetEXT(\{vaobj}, \{buffer}, \{type}, \{stride}, \{offset})");
		try {
		    functions.glVertexArrayFogCoordOffsetEXT().invokeExact(vaobj, buffer, type, stride, offset);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexArrayIndexOffsetEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexArrayIndexOffsetEXT.xhtml">registry.khronos.org</a>
	 */
	public void glVertexArrayIndexOffsetEXT(int vaobj, int buffer, GLIndexPointerType type, int stride, long offset) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexArrayIndexOffsetEXT(\{vaobj}, \{buffer}, \{type}, \{stride}, \{offset})");
		try {
		    functions.glVertexArrayIndexOffsetEXT().invokeExact(vaobj, buffer, type, stride, offset);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexArrayMultiTexCoordOffsetEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexArrayMultiTexCoordOffsetEXT.xhtml">registry.khronos.org</a>
	 */
	public void glVertexArrayMultiTexCoordOffsetEXT(int vaobj, int buffer, GLEnum texunit, int size, GLTexCoordPointerType type, int stride, long offset) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexArrayMultiTexCoordOffsetEXT(\{vaobj}, \{buffer}, \{texunit}, \{size}, \{type}, \{stride}, \{offset})");
		try {
		    functions.glVertexArrayMultiTexCoordOffsetEXT().invokeExact(vaobj, buffer, texunit, size, type, stride, offset);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were texunit=\{texunit}, type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexArrayNormalOffsetEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexArrayNormalOffsetEXT.xhtml">registry.khronos.org</a>
	 */
	public void glVertexArrayNormalOffsetEXT(int vaobj, int buffer, GLNormalPointerType type, int stride, long offset) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexArrayNormalOffsetEXT(\{vaobj}, \{buffer}, \{type}, \{stride}, \{offset})");
		try {
		    functions.glVertexArrayNormalOffsetEXT().invokeExact(vaobj, buffer, type, stride, offset);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexArrayParameteriAPPLE">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexArrayParameteriAPPLE.xhtml">registry.khronos.org</a>
	 */
	public void glVertexArrayParameteriAPPLE(GLVertexArrayPNameAPPLE pname, int param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexArrayParameteriAPPLE(\{pname}, \{param})");
		try {
		    functions.glVertexArrayParameteriAPPLE().invokeExact(pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexArrayRangeAPPLE">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexArrayRangeAPPLE.xhtml">registry.khronos.org</a>
	 */
	public void glVertexArrayRangeAPPLE(int length, MutPointer<Void> pointer) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexArrayRangeAPPLE(\{length}, \{pointer})");
		try {
		    functions.glVertexArrayRangeAPPLE().invokeExact(length, pointer);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexArrayRangeNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexArrayRangeNV.xhtml">registry.khronos.org</a>
	 */
	public void glVertexArrayRangeNV(int length, ConstPointer<Void> pointer) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexArrayRangeNV(\{length}, \{pointer})");
		try {
		    functions.glVertexArrayRangeNV().invokeExact(length, pointer);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexArraySecondaryColorOffsetEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexArraySecondaryColorOffsetEXT.xhtml">registry.khronos.org</a>
	 */
	public void glVertexArraySecondaryColorOffsetEXT(int vaobj, int buffer, int size, GLColorPointerType type, int stride, long offset) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexArraySecondaryColorOffsetEXT(\{vaobj}, \{buffer}, \{size}, \{type}, \{stride}, \{offset})");
		try {
		    functions.glVertexArraySecondaryColorOffsetEXT().invokeExact(vaobj, buffer, size, type, stride, offset);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexArrayTexCoordOffsetEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexArrayTexCoordOffsetEXT.xhtml">registry.khronos.org</a>
	 */
	public void glVertexArrayTexCoordOffsetEXT(int vaobj, int buffer, int size, GLTexCoordPointerType type, int stride, long offset) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexArrayTexCoordOffsetEXT(\{vaobj}, \{buffer}, \{size}, \{type}, \{stride}, \{offset})");
		try {
		    functions.glVertexArrayTexCoordOffsetEXT().invokeExact(vaobj, buffer, size, type, stride, offset);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexArrayVertexAttribBindingEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexArrayVertexAttribBindingEXT.xhtml">registry.khronos.org</a>
	 */
	public void glVertexArrayVertexAttribBindingEXT(int vaobj, int attribindex, int bindingindex) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexArrayVertexAttribBindingEXT(\{vaobj}, \{attribindex}, \{bindingindex})");
		try {
		    functions.glVertexArrayVertexAttribBindingEXT().invokeExact(vaobj, attribindex, bindingindex);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexArrayVertexAttribDivisorEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexArrayVertexAttribDivisorEXT.xhtml">registry.khronos.org</a>
	 */
	public void glVertexArrayVertexAttribDivisorEXT(int vaobj, int index, int divisor) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexArrayVertexAttribDivisorEXT(\{vaobj}, \{index}, \{divisor})");
		try {
		    functions.glVertexArrayVertexAttribDivisorEXT().invokeExact(vaobj, index, divisor);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexArrayVertexAttribFormatEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexArrayVertexAttribFormatEXT.xhtml">registry.khronos.org</a>
	 */
	public void glVertexArrayVertexAttribFormatEXT(int vaobj, int attribindex, int size, GLVertexAttribType type, boolean normalized, int relativeoffset) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexArrayVertexAttribFormatEXT(\{vaobj}, \{attribindex}, \{size}, \{type}, \{normalized}, \{relativeoffset})");
		try {
		    functions.glVertexArrayVertexAttribFormatEXT().invokeExact(vaobj, attribindex, size, type, normalized, relativeoffset);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexArrayVertexAttribIFormatEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexArrayVertexAttribIFormatEXT.xhtml">registry.khronos.org</a>
	 */
	public void glVertexArrayVertexAttribIFormatEXT(int vaobj, int attribindex, int size, GLVertexAttribIType type, int relativeoffset) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexArrayVertexAttribIFormatEXT(\{vaobj}, \{attribindex}, \{size}, \{type}, \{relativeoffset})");
		try {
		    functions.glVertexArrayVertexAttribIFormatEXT().invokeExact(vaobj, attribindex, size, type, relativeoffset);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexArrayVertexAttribIOffsetEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexArrayVertexAttribIOffsetEXT.xhtml">registry.khronos.org</a>
	 */
	public void glVertexArrayVertexAttribIOffsetEXT(int vaobj, int buffer, int index, int size, GLVertexAttribType type, int stride, long offset) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexArrayVertexAttribIOffsetEXT(\{vaobj}, \{buffer}, \{index}, \{size}, \{type}, \{stride}, \{offset})");
		try {
		    functions.glVertexArrayVertexAttribIOffsetEXT().invokeExact(vaobj, buffer, index, size, type, stride, offset);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexArrayVertexAttribLFormatEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexArrayVertexAttribLFormatEXT.xhtml">registry.khronos.org</a>
	 */
	public void glVertexArrayVertexAttribLFormatEXT(int vaobj, int attribindex, int size, GLVertexAttribLType type, int relativeoffset) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexArrayVertexAttribLFormatEXT(\{vaobj}, \{attribindex}, \{size}, \{type}, \{relativeoffset})");
		try {
		    functions.glVertexArrayVertexAttribLFormatEXT().invokeExact(vaobj, attribindex, size, type, relativeoffset);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexArrayVertexAttribLOffsetEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexArrayVertexAttribLOffsetEXT.xhtml">registry.khronos.org</a>
	 */
	public void glVertexArrayVertexAttribLOffsetEXT(int vaobj, int buffer, int index, int size, GLVertexAttribLType type, int stride, long offset) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexArrayVertexAttribLOffsetEXT(\{vaobj}, \{buffer}, \{index}, \{size}, \{type}, \{stride}, \{offset})");
		try {
		    functions.glVertexArrayVertexAttribLOffsetEXT().invokeExact(vaobj, buffer, index, size, type, stride, offset);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexArrayVertexAttribOffsetEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexArrayVertexAttribOffsetEXT.xhtml">registry.khronos.org</a>
	 */
	public void glVertexArrayVertexAttribOffsetEXT(int vaobj, int buffer, int index, int size, GLVertexAttribPointerType type, boolean normalized, int stride, long offset) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexArrayVertexAttribOffsetEXT(\{vaobj}, \{buffer}, \{index}, \{size}, \{type}, \{normalized}, \{stride}, \{offset})");
		try {
		    functions.glVertexArrayVertexAttribOffsetEXT().invokeExact(vaobj, buffer, index, size, type, normalized, stride, offset);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexArrayVertexBindingDivisorEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexArrayVertexBindingDivisorEXT.xhtml">registry.khronos.org</a>
	 */
	public void glVertexArrayVertexBindingDivisorEXT(int vaobj, int bindingindex, int divisor) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexArrayVertexBindingDivisorEXT(\{vaobj}, \{bindingindex}, \{divisor})");
		try {
		    functions.glVertexArrayVertexBindingDivisorEXT().invokeExact(vaobj, bindingindex, divisor);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexArrayVertexBuffer">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexArrayVertexBuffer.xhtml">registry.khronos.org</a>
	 */
	public void glVertexArrayVertexBuffer(int vaobj, int bindingindex, int buffer, long offset, int stride) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexArrayVertexBuffer(\{vaobj}, \{bindingindex}, \{buffer}, \{offset}, \{stride})");
		try {
		    functions.glVertexArrayVertexBuffer().invokeExact(vaobj, bindingindex, buffer, offset, stride);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexArrayVertexBuffers">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexArrayVertexBuffers.xhtml">registry.khronos.org</a>
	 */
	public void glVertexArrayVertexBuffers(int vaobj, int first, int count, ConstPointer<Integer> buffers, ConstPointer<Long> offsets, ConstPointer<Integer> strides) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexArrayVertexBuffers(\{vaobj}, \{first}, \{count}, \{buffers}, \{offsets}, \{strides})");
		try {
		    functions.glVertexArrayVertexBuffers().invokeExact(vaobj, first, count, buffers, offsets, strides);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexArrayVertexOffsetEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexArrayVertexOffsetEXT.xhtml">registry.khronos.org</a>
	 */
	public void glVertexArrayVertexOffsetEXT(int vaobj, int buffer, int size, GLVertexPointerType type, int stride, long offset) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexArrayVertexOffsetEXT(\{vaobj}, \{buffer}, \{size}, \{type}, \{stride}, \{offset})");
		try {
		    functions.glVertexArrayVertexOffsetEXT().invokeExact(vaobj, buffer, size, type, stride, offset);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttrib1d">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttrib1d.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttrib1d(int index, double x) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttrib1d(\{index}, \{x})");
		try {
		    functions.glVertexAttrib1d().invokeExact(index, x);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttrib1dARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttrib1dARB.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttrib1dARB(int index, double x) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttrib1dARB(\{index}, \{x})");
		try {
		    functions.glVertexAttrib1dARB().invokeExact(index, x);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttrib1dNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttrib1dNV.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttrib1dNV(int index, double x) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttrib1dNV(\{index}, \{x})");
		try {
		    functions.glVertexAttrib1dNV().invokeExact(index, x);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttrib1dv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttrib1dv.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttrib1dv(int index, ConstPointer<Double> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttrib1dv(\{index}, \{v})");
		try {
		    functions.glVertexAttrib1dv().invokeExact(index, v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttrib1dvARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttrib1dvARB.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttrib1dvARB(int index, ConstPointer<Double> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttrib1dvARB(\{index}, \{v})");
		try {
		    functions.glVertexAttrib1dvARB().invokeExact(index, v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttrib1dvNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttrib1dvNV.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttrib1dvNV(int index, ConstPointer<Double> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttrib1dvNV(\{index}, \{v})");
		try {
		    functions.glVertexAttrib1dvNV().invokeExact(index, v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttrib1f">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttrib1f.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttrib1f(int index, float x) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttrib1f(\{index}, \{x})");
		try {
		    functions.glVertexAttrib1f().invokeExact(index, x);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttrib1fARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttrib1fARB.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttrib1fARB(int index, float x) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttrib1fARB(\{index}, \{x})");
		try {
		    functions.glVertexAttrib1fARB().invokeExact(index, x);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttrib1fNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttrib1fNV.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttrib1fNV(int index, float x) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttrib1fNV(\{index}, \{x})");
		try {
		    functions.glVertexAttrib1fNV().invokeExact(index, x);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttrib1fv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttrib1fv.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttrib1fv(int index, ConstPointer<Float> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttrib1fv(\{index}, \{v})");
		try {
		    functions.glVertexAttrib1fv().invokeExact(index, v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttrib1fvARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttrib1fvARB.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttrib1fvARB(int index, ConstPointer<Float> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttrib1fvARB(\{index}, \{v})");
		try {
		    functions.glVertexAttrib1fvARB().invokeExact(index, v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttrib1fvNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttrib1fvNV.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttrib1fvNV(int index, ConstPointer<Float> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttrib1fvNV(\{index}, \{v})");
		try {
		    functions.glVertexAttrib1fvNV().invokeExact(index, v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttrib1hNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttrib1hNV.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttrib1hNV(int index, char x) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttrib1hNV(\{index}, \{x})");
		try {
		    functions.glVertexAttrib1hNV().invokeExact(index, x);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttrib1hvNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttrib1hvNV.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttrib1hvNV(int index, ConstPointer<Character> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttrib1hvNV(\{index}, \{v})");
		try {
		    functions.glVertexAttrib1hvNV().invokeExact(index, v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttrib1s">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttrib1s.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttrib1s(int index, short x) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttrib1s(\{index}, \{x})");
		try {
		    functions.glVertexAttrib1s().invokeExact(index, x);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttrib1sARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttrib1sARB.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttrib1sARB(int index, short x) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttrib1sARB(\{index}, \{x})");
		try {
		    functions.glVertexAttrib1sARB().invokeExact(index, x);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttrib1sNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttrib1sNV.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttrib1sNV(int index, short x) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttrib1sNV(\{index}, \{x})");
		try {
		    functions.glVertexAttrib1sNV().invokeExact(index, x);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttrib1sv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttrib1sv.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttrib1sv(int index, ConstPointer<Short> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttrib1sv(\{index}, \{v})");
		try {
		    functions.glVertexAttrib1sv().invokeExact(index, v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttrib1svARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttrib1svARB.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttrib1svARB(int index, ConstPointer<Short> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttrib1svARB(\{index}, \{v})");
		try {
		    functions.glVertexAttrib1svARB().invokeExact(index, v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttrib1svNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttrib1svNV.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttrib1svNV(int index, ConstPointer<Short> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttrib1svNV(\{index}, \{v})");
		try {
		    functions.glVertexAttrib1svNV().invokeExact(index, v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttrib2d">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttrib2d.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttrib2d(int index, double x, double y) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttrib2d(\{index}, \{x}, \{y})");
		try {
		    functions.glVertexAttrib2d().invokeExact(index, x, y);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttrib2dARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttrib2dARB.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttrib2dARB(int index, double x, double y) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttrib2dARB(\{index}, \{x}, \{y})");
		try {
		    functions.glVertexAttrib2dARB().invokeExact(index, x, y);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttrib2dNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttrib2dNV.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttrib2dNV(int index, double x, double y) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttrib2dNV(\{index}, \{x}, \{y})");
		try {
		    functions.glVertexAttrib2dNV().invokeExact(index, x, y);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttrib2dv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttrib2dv.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttrib2dv(int index, ConstPointer<Double> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttrib2dv(\{index}, \{v})");
		try {
		    functions.glVertexAttrib2dv().invokeExact(index, v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttrib2dvARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttrib2dvARB.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttrib2dvARB(int index, ConstPointer<Double> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttrib2dvARB(\{index}, \{v})");
		try {
		    functions.glVertexAttrib2dvARB().invokeExact(index, v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttrib2dvNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttrib2dvNV.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttrib2dvNV(int index, ConstPointer<Double> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttrib2dvNV(\{index}, \{v})");
		try {
		    functions.glVertexAttrib2dvNV().invokeExact(index, v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttrib2f">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttrib2f.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttrib2f(int index, float x, float y) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttrib2f(\{index}, \{x}, \{y})");
		try {
		    functions.glVertexAttrib2f().invokeExact(index, x, y);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttrib2fARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttrib2fARB.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttrib2fARB(int index, float x, float y) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttrib2fARB(\{index}, \{x}, \{y})");
		try {
		    functions.glVertexAttrib2fARB().invokeExact(index, x, y);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttrib2fNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttrib2fNV.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttrib2fNV(int index, float x, float y) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttrib2fNV(\{index}, \{x}, \{y})");
		try {
		    functions.glVertexAttrib2fNV().invokeExact(index, x, y);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttrib2fv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttrib2fv.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttrib2fv(int index, ConstPointer<Float> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttrib2fv(\{index}, \{v})");
		try {
		    functions.glVertexAttrib2fv().invokeExact(index, v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttrib2fvARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttrib2fvARB.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttrib2fvARB(int index, ConstPointer<Float> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttrib2fvARB(\{index}, \{v})");
		try {
		    functions.glVertexAttrib2fvARB().invokeExact(index, v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttrib2fvNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttrib2fvNV.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttrib2fvNV(int index, ConstPointer<Float> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttrib2fvNV(\{index}, \{v})");
		try {
		    functions.glVertexAttrib2fvNV().invokeExact(index, v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttrib2hNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttrib2hNV.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttrib2hNV(int index, char x, char y) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttrib2hNV(\{index}, \{x}, \{y})");
		try {
		    functions.glVertexAttrib2hNV().invokeExact(index, x, y);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttrib2hvNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttrib2hvNV.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttrib2hvNV(int index, ConstPointer<Character> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttrib2hvNV(\{index}, \{v})");
		try {
		    functions.glVertexAttrib2hvNV().invokeExact(index, v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttrib2s">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttrib2s.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttrib2s(int index, short x, short y) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttrib2s(\{index}, \{x}, \{y})");
		try {
		    functions.glVertexAttrib2s().invokeExact(index, x, y);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttrib2sARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttrib2sARB.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttrib2sARB(int index, short x, short y) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttrib2sARB(\{index}, \{x}, \{y})");
		try {
		    functions.glVertexAttrib2sARB().invokeExact(index, x, y);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttrib2sNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttrib2sNV.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttrib2sNV(int index, short x, short y) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttrib2sNV(\{index}, \{x}, \{y})");
		try {
		    functions.glVertexAttrib2sNV().invokeExact(index, x, y);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttrib2sv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttrib2sv.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttrib2sv(int index, ConstPointer<Short> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttrib2sv(\{index}, \{v})");
		try {
		    functions.glVertexAttrib2sv().invokeExact(index, v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttrib2svARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttrib2svARB.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttrib2svARB(int index, ConstPointer<Short> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttrib2svARB(\{index}, \{v})");
		try {
		    functions.glVertexAttrib2svARB().invokeExact(index, v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttrib2svNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttrib2svNV.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttrib2svNV(int index, ConstPointer<Short> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttrib2svNV(\{index}, \{v})");
		try {
		    functions.glVertexAttrib2svNV().invokeExact(index, v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttrib3d">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttrib3d.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttrib3d(int index, double x, double y, double z) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttrib3d(\{index}, \{x}, \{y}, \{z})");
		try {
		    functions.glVertexAttrib3d().invokeExact(index, x, y, z);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttrib3dARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttrib3dARB.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttrib3dARB(int index, double x, double y, double z) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttrib3dARB(\{index}, \{x}, \{y}, \{z})");
		try {
		    functions.glVertexAttrib3dARB().invokeExact(index, x, y, z);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttrib3dNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttrib3dNV.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttrib3dNV(int index, double x, double y, double z) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttrib3dNV(\{index}, \{x}, \{y}, \{z})");
		try {
		    functions.glVertexAttrib3dNV().invokeExact(index, x, y, z);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttrib3dv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttrib3dv.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttrib3dv(int index, ConstPointer<Double> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttrib3dv(\{index}, \{v})");
		try {
		    functions.glVertexAttrib3dv().invokeExact(index, v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttrib3dvARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttrib3dvARB.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttrib3dvARB(int index, ConstPointer<Double> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttrib3dvARB(\{index}, \{v})");
		try {
		    functions.glVertexAttrib3dvARB().invokeExact(index, v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttrib3dvNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttrib3dvNV.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttrib3dvNV(int index, ConstPointer<Double> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttrib3dvNV(\{index}, \{v})");
		try {
		    functions.glVertexAttrib3dvNV().invokeExact(index, v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttrib3f">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttrib3f.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttrib3f(int index, float x, float y, float z) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttrib3f(\{index}, \{x}, \{y}, \{z})");
		try {
		    functions.glVertexAttrib3f().invokeExact(index, x, y, z);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttrib3fARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttrib3fARB.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttrib3fARB(int index, float x, float y, float z) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttrib3fARB(\{index}, \{x}, \{y}, \{z})");
		try {
		    functions.glVertexAttrib3fARB().invokeExact(index, x, y, z);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttrib3fNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttrib3fNV.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttrib3fNV(int index, float x, float y, float z) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttrib3fNV(\{index}, \{x}, \{y}, \{z})");
		try {
		    functions.glVertexAttrib3fNV().invokeExact(index, x, y, z);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttrib3fv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttrib3fv.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttrib3fv(int index, ConstPointer<Float> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttrib3fv(\{index}, \{v})");
		try {
		    functions.glVertexAttrib3fv().invokeExact(index, v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttrib3fvARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttrib3fvARB.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttrib3fvARB(int index, ConstPointer<Float> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttrib3fvARB(\{index}, \{v})");
		try {
		    functions.glVertexAttrib3fvARB().invokeExact(index, v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttrib3fvNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttrib3fvNV.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttrib3fvNV(int index, ConstPointer<Float> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttrib3fvNV(\{index}, \{v})");
		try {
		    functions.glVertexAttrib3fvNV().invokeExact(index, v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttrib3hNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttrib3hNV.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttrib3hNV(int index, char x, char y, char z) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttrib3hNV(\{index}, \{x}, \{y}, \{z})");
		try {
		    functions.glVertexAttrib3hNV().invokeExact(index, x, y, z);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttrib3hvNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttrib3hvNV.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttrib3hvNV(int index, ConstPointer<Character> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttrib3hvNV(\{index}, \{v})");
		try {
		    functions.glVertexAttrib3hvNV().invokeExact(index, v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttrib3s">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttrib3s.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttrib3s(int index, short x, short y, short z) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttrib3s(\{index}, \{x}, \{y}, \{z})");
		try {
		    functions.glVertexAttrib3s().invokeExact(index, x, y, z);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttrib3sARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttrib3sARB.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttrib3sARB(int index, short x, short y, short z) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttrib3sARB(\{index}, \{x}, \{y}, \{z})");
		try {
		    functions.glVertexAttrib3sARB().invokeExact(index, x, y, z);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttrib3sNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttrib3sNV.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttrib3sNV(int index, short x, short y, short z) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttrib3sNV(\{index}, \{x}, \{y}, \{z})");
		try {
		    functions.glVertexAttrib3sNV().invokeExact(index, x, y, z);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttrib3sv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttrib3sv.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttrib3sv(int index, ConstPointer<Short> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttrib3sv(\{index}, \{v})");
		try {
		    functions.glVertexAttrib3sv().invokeExact(index, v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttrib3svARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttrib3svARB.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttrib3svARB(int index, ConstPointer<Short> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttrib3svARB(\{index}, \{v})");
		try {
		    functions.glVertexAttrib3svARB().invokeExact(index, v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttrib3svNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttrib3svNV.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttrib3svNV(int index, ConstPointer<Short> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttrib3svNV(\{index}, \{v})");
		try {
		    functions.glVertexAttrib3svNV().invokeExact(index, v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttrib4Nbv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttrib4Nbv.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttrib4Nbv(int index, ConstPointer<Byte> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttrib4Nbv(\{index}, \{v})");
		try {
		    functions.glVertexAttrib4Nbv().invokeExact(index, v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttrib4NbvARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttrib4NbvARB.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttrib4NbvARB(int index, ConstPointer<Byte> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttrib4NbvARB(\{index}, \{v})");
		try {
		    functions.glVertexAttrib4NbvARB().invokeExact(index, v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttrib4Niv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttrib4Niv.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttrib4Niv(int index, ConstPointer<Integer> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttrib4Niv(\{index}, \{v})");
		try {
		    functions.glVertexAttrib4Niv().invokeExact(index, v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttrib4NivARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttrib4NivARB.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttrib4NivARB(int index, ConstPointer<Integer> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttrib4NivARB(\{index}, \{v})");
		try {
		    functions.glVertexAttrib4NivARB().invokeExact(index, v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttrib4Nsv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttrib4Nsv.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttrib4Nsv(int index, ConstPointer<Short> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttrib4Nsv(\{index}, \{v})");
		try {
		    functions.glVertexAttrib4Nsv().invokeExact(index, v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttrib4NsvARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttrib4NsvARB.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttrib4NsvARB(int index, ConstPointer<Short> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttrib4NsvARB(\{index}, \{v})");
		try {
		    functions.glVertexAttrib4NsvARB().invokeExact(index, v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttrib4Nub">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttrib4Nub.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttrib4Nub(int index, byte x, byte y, byte z, byte w) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttrib4Nub(\{index}, \{x}, \{y}, \{z}, \{w})");
		try {
		    functions.glVertexAttrib4Nub().invokeExact(index, x, y, z, w);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttrib4NubARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttrib4NubARB.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttrib4NubARB(int index, byte x, byte y, byte z, byte w) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttrib4NubARB(\{index}, \{x}, \{y}, \{z}, \{w})");
		try {
		    functions.glVertexAttrib4NubARB().invokeExact(index, x, y, z, w);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttrib4Nubv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttrib4Nubv.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttrib4Nubv(int index, ConstPointer<Byte> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttrib4Nubv(\{index}, \{v})");
		try {
		    functions.glVertexAttrib4Nubv().invokeExact(index, v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttrib4NubvARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttrib4NubvARB.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttrib4NubvARB(int index, ConstPointer<Byte> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttrib4NubvARB(\{index}, \{v})");
		try {
		    functions.glVertexAttrib4NubvARB().invokeExact(index, v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttrib4Nuiv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttrib4Nuiv.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttrib4Nuiv(int index, ConstPointer<Integer> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttrib4Nuiv(\{index}, \{v})");
		try {
		    functions.glVertexAttrib4Nuiv().invokeExact(index, v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttrib4NuivARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttrib4NuivARB.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttrib4NuivARB(int index, ConstPointer<Integer> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttrib4NuivARB(\{index}, \{v})");
		try {
		    functions.glVertexAttrib4NuivARB().invokeExact(index, v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttrib4Nusv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttrib4Nusv.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttrib4Nusv(int index, ConstPointer<Character> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttrib4Nusv(\{index}, \{v})");
		try {
		    functions.glVertexAttrib4Nusv().invokeExact(index, v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttrib4NusvARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttrib4NusvARB.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttrib4NusvARB(int index, ConstPointer<Character> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttrib4NusvARB(\{index}, \{v})");
		try {
		    functions.glVertexAttrib4NusvARB().invokeExact(index, v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttrib4bv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttrib4bv.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttrib4bv(int index, ConstPointer<Byte> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttrib4bv(\{index}, \{v})");
		try {
		    functions.glVertexAttrib4bv().invokeExact(index, v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttrib4bvARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttrib4bvARB.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttrib4bvARB(int index, ConstPointer<Byte> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttrib4bvARB(\{index}, \{v})");
		try {
		    functions.glVertexAttrib4bvARB().invokeExact(index, v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttrib4d">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttrib4d.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttrib4d(int index, double x, double y, double z, double w) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttrib4d(\{index}, \{x}, \{y}, \{z}, \{w})");
		try {
		    functions.glVertexAttrib4d().invokeExact(index, x, y, z, w);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttrib4dARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttrib4dARB.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttrib4dARB(int index, double x, double y, double z, double w) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttrib4dARB(\{index}, \{x}, \{y}, \{z}, \{w})");
		try {
		    functions.glVertexAttrib4dARB().invokeExact(index, x, y, z, w);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttrib4dNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttrib4dNV.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttrib4dNV(int index, double x, double y, double z, double w) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttrib4dNV(\{index}, \{x}, \{y}, \{z}, \{w})");
		try {
		    functions.glVertexAttrib4dNV().invokeExact(index, x, y, z, w);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttrib4dv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttrib4dv.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttrib4dv(int index, ConstPointer<Double> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttrib4dv(\{index}, \{v})");
		try {
		    functions.glVertexAttrib4dv().invokeExact(index, v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttrib4dvARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttrib4dvARB.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttrib4dvARB(int index, ConstPointer<Double> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttrib4dvARB(\{index}, \{v})");
		try {
		    functions.glVertexAttrib4dvARB().invokeExact(index, v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttrib4dvNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttrib4dvNV.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttrib4dvNV(int index, ConstPointer<Double> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttrib4dvNV(\{index}, \{v})");
		try {
		    functions.glVertexAttrib4dvNV().invokeExact(index, v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttrib4f">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttrib4f.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttrib4f(int index, float x, float y, float z, float w) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttrib4f(\{index}, \{x}, \{y}, \{z}, \{w})");
		try {
		    functions.glVertexAttrib4f().invokeExact(index, x, y, z, w);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttrib4fARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttrib4fARB.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttrib4fARB(int index, float x, float y, float z, float w) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttrib4fARB(\{index}, \{x}, \{y}, \{z}, \{w})");
		try {
		    functions.glVertexAttrib4fARB().invokeExact(index, x, y, z, w);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttrib4fNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttrib4fNV.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttrib4fNV(int index, float x, float y, float z, float w) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttrib4fNV(\{index}, \{x}, \{y}, \{z}, \{w})");
		try {
		    functions.glVertexAttrib4fNV().invokeExact(index, x, y, z, w);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttrib4fv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttrib4fv.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttrib4fv(int index, ConstPointer<Float> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttrib4fv(\{index}, \{v})");
		try {
		    functions.glVertexAttrib4fv().invokeExact(index, v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttrib4fvARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttrib4fvARB.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttrib4fvARB(int index, ConstPointer<Float> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttrib4fvARB(\{index}, \{v})");
		try {
		    functions.glVertexAttrib4fvARB().invokeExact(index, v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttrib4fvNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttrib4fvNV.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttrib4fvNV(int index, ConstPointer<Float> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttrib4fvNV(\{index}, \{v})");
		try {
		    functions.glVertexAttrib4fvNV().invokeExact(index, v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttrib4hNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttrib4hNV.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttrib4hNV(int index, char x, char y, char z, char w) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttrib4hNV(\{index}, \{x}, \{y}, \{z}, \{w})");
		try {
		    functions.glVertexAttrib4hNV().invokeExact(index, x, y, z, w);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttrib4hvNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttrib4hvNV.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttrib4hvNV(int index, ConstPointer<Character> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttrib4hvNV(\{index}, \{v})");
		try {
		    functions.glVertexAttrib4hvNV().invokeExact(index, v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttrib4iv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttrib4iv.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttrib4iv(int index, ConstPointer<Integer> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttrib4iv(\{index}, \{v})");
		try {
		    functions.glVertexAttrib4iv().invokeExact(index, v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttrib4ivARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttrib4ivARB.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttrib4ivARB(int index, ConstPointer<Integer> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttrib4ivARB(\{index}, \{v})");
		try {
		    functions.glVertexAttrib4ivARB().invokeExact(index, v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttrib4s">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttrib4s.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttrib4s(int index, short x, short y, short z, short w) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttrib4s(\{index}, \{x}, \{y}, \{z}, \{w})");
		try {
		    functions.glVertexAttrib4s().invokeExact(index, x, y, z, w);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttrib4sARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttrib4sARB.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttrib4sARB(int index, short x, short y, short z, short w) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttrib4sARB(\{index}, \{x}, \{y}, \{z}, \{w})");
		try {
		    functions.glVertexAttrib4sARB().invokeExact(index, x, y, z, w);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttrib4sNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttrib4sNV.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttrib4sNV(int index, short x, short y, short z, short w) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttrib4sNV(\{index}, \{x}, \{y}, \{z}, \{w})");
		try {
		    functions.glVertexAttrib4sNV().invokeExact(index, x, y, z, w);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttrib4sv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttrib4sv.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttrib4sv(int index, ConstPointer<Short> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttrib4sv(\{index}, \{v})");
		try {
		    functions.glVertexAttrib4sv().invokeExact(index, v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttrib4svARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttrib4svARB.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttrib4svARB(int index, ConstPointer<Short> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttrib4svARB(\{index}, \{v})");
		try {
		    functions.glVertexAttrib4svARB().invokeExact(index, v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttrib4svNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttrib4svNV.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttrib4svNV(int index, ConstPointer<Short> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttrib4svNV(\{index}, \{v})");
		try {
		    functions.glVertexAttrib4svNV().invokeExact(index, v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttrib4ubNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttrib4ubNV.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttrib4ubNV(int index, byte x, byte y, byte z, byte w) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttrib4ubNV(\{index}, \{x}, \{y}, \{z}, \{w})");
		try {
		    functions.glVertexAttrib4ubNV().invokeExact(index, x, y, z, w);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttrib4ubv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttrib4ubv.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttrib4ubv(int index, ConstPointer<Byte> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttrib4ubv(\{index}, \{v})");
		try {
		    functions.glVertexAttrib4ubv().invokeExact(index, v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttrib4ubvARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttrib4ubvARB.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttrib4ubvARB(int index, ConstPointer<Byte> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttrib4ubvARB(\{index}, \{v})");
		try {
		    functions.glVertexAttrib4ubvARB().invokeExact(index, v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttrib4ubvNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttrib4ubvNV.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttrib4ubvNV(int index, ConstPointer<Byte> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttrib4ubvNV(\{index}, \{v})");
		try {
		    functions.glVertexAttrib4ubvNV().invokeExact(index, v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttrib4uiv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttrib4uiv.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttrib4uiv(int index, ConstPointer<Integer> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttrib4uiv(\{index}, \{v})");
		try {
		    functions.glVertexAttrib4uiv().invokeExact(index, v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttrib4uivARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttrib4uivARB.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttrib4uivARB(int index, ConstPointer<Integer> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttrib4uivARB(\{index}, \{v})");
		try {
		    functions.glVertexAttrib4uivARB().invokeExact(index, v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttrib4usv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttrib4usv.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttrib4usv(int index, ConstPointer<Character> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttrib4usv(\{index}, \{v})");
		try {
		    functions.glVertexAttrib4usv().invokeExact(index, v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttrib4usvARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttrib4usvARB.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttrib4usvARB(int index, ConstPointer<Character> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttrib4usvARB(\{index}, \{v})");
		try {
		    functions.glVertexAttrib4usvARB().invokeExact(index, v);
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribArrayObjectATI">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribArrayObjectATI.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribArrayObjectATI(int index, int size, GLVertexAttribPointerType type, boolean normalized, int stride, int buffer, int offset) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribArrayObjectATI(\{index}, \{size}, \{type}, \{normalized}, \{stride}, \{buffer}, \{offset})");
		try {
		    functions.glVertexAttribArrayObjectATI().invokeExact(index, size, type, normalized, stride, buffer, offset);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribBinding">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribBinding.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribBinding(int attribindex, int bindingindex) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribBinding(\{attribindex}, \{bindingindex})");
		try {
		    functions.glVertexAttribBinding().invokeExact(attribindex, bindingindex);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribDivisor">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribDivisor.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribDivisor(int index, int divisor) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribDivisor(\{index}, \{divisor})");
		try {
		    functions.glVertexAttribDivisor().invokeExact(index, divisor);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribDivisorANGLE">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribDivisorANGLE.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribDivisorANGLE(int index, int divisor) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribDivisorANGLE(\{index}, \{divisor})");
		try {
		    functions.glVertexAttribDivisorANGLE().invokeExact(index, divisor);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribDivisorARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribDivisorARB.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribDivisorARB(int index, int divisor) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribDivisorARB(\{index}, \{divisor})");
		try {
		    functions.glVertexAttribDivisorARB().invokeExact(index, divisor);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribDivisorEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribDivisorEXT.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribDivisorEXT(int index, int divisor) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribDivisorEXT(\{index}, \{divisor})");
		try {
		    functions.glVertexAttribDivisorEXT().invokeExact(index, divisor);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribDivisorNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribDivisorNV.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribDivisorNV(int index, int divisor) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribDivisorNV(\{index}, \{divisor})");
		try {
		    functions.glVertexAttribDivisorNV().invokeExact(index, divisor);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribFormat">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribFormat.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribFormat(int attribindex, int size, GLVertexAttribType type, boolean normalized, int relativeoffset) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribFormat(\{attribindex}, \{size}, \{type}, \{normalized}, \{relativeoffset})");
		try {
		    functions.glVertexAttribFormat().invokeExact(attribindex, size, type, normalized, relativeoffset);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribFormatNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribFormatNV.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribFormatNV(int index, int size, GLVertexAttribType type, boolean normalized, int stride) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribFormatNV(\{index}, \{size}, \{type}, \{normalized}, \{stride})");
		try {
		    functions.glVertexAttribFormatNV().invokeExact(index, size, type, normalized, stride);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribI1i">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribI1i.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribI1i(int index, int x) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribI1i(\{index}, \{x})");
		try {
		    functions.glVertexAttribI1i().invokeExact(index, x);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribI1iEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribI1iEXT.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribI1iEXT(int index, int x) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribI1iEXT(\{index}, \{x})");
		try {
		    functions.glVertexAttribI1iEXT().invokeExact(index, x);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribI1iv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribI1iv.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribI1iv(int index, ConstPointer<Integer> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribI1iv(\{index}, \{v})");
		try {
		    functions.glVertexAttribI1iv().invokeExact(index, v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribI1ivEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribI1ivEXT.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribI1ivEXT(int index, ConstPointer<Integer> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribI1ivEXT(\{index}, \{v})");
		try {
		    functions.glVertexAttribI1ivEXT().invokeExact(index, v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribI1ui">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribI1ui.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribI1ui(int index, int x) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribI1ui(\{index}, \{x})");
		try {
		    functions.glVertexAttribI1ui().invokeExact(index, x);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribI1uiEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribI1uiEXT.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribI1uiEXT(int index, int x) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribI1uiEXT(\{index}, \{x})");
		try {
		    functions.glVertexAttribI1uiEXT().invokeExact(index, x);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribI1uiv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribI1uiv.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribI1uiv(int index, ConstPointer<Integer> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribI1uiv(\{index}, \{v})");
		try {
		    functions.glVertexAttribI1uiv().invokeExact(index, v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribI1uivEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribI1uivEXT.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribI1uivEXT(int index, ConstPointer<Integer> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribI1uivEXT(\{index}, \{v})");
		try {
		    functions.glVertexAttribI1uivEXT().invokeExact(index, v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribI2i">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribI2i.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribI2i(int index, int x, int y) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribI2i(\{index}, \{x}, \{y})");
		try {
		    functions.glVertexAttribI2i().invokeExact(index, x, y);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribI2iEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribI2iEXT.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribI2iEXT(int index, int x, int y) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribI2iEXT(\{index}, \{x}, \{y})");
		try {
		    functions.glVertexAttribI2iEXT().invokeExact(index, x, y);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribI2iv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribI2iv.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribI2iv(int index, ConstPointer<Integer> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribI2iv(\{index}, \{v})");
		try {
		    functions.glVertexAttribI2iv().invokeExact(index, v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribI2ivEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribI2ivEXT.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribI2ivEXT(int index, ConstPointer<Integer> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribI2ivEXT(\{index}, \{v})");
		try {
		    functions.glVertexAttribI2ivEXT().invokeExact(index, v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribI2ui">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribI2ui.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribI2ui(int index, int x, int y) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribI2ui(\{index}, \{x}, \{y})");
		try {
		    functions.glVertexAttribI2ui().invokeExact(index, x, y);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribI2uiEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribI2uiEXT.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribI2uiEXT(int index, int x, int y) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribI2uiEXT(\{index}, \{x}, \{y})");
		try {
		    functions.glVertexAttribI2uiEXT().invokeExact(index, x, y);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribI2uiv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribI2uiv.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribI2uiv(int index, ConstPointer<Integer> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribI2uiv(\{index}, \{v})");
		try {
		    functions.glVertexAttribI2uiv().invokeExact(index, v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribI2uivEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribI2uivEXT.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribI2uivEXT(int index, ConstPointer<Integer> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribI2uivEXT(\{index}, \{v})");
		try {
		    functions.glVertexAttribI2uivEXT().invokeExact(index, v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribI3i">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribI3i.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribI3i(int index, int x, int y, int z) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribI3i(\{index}, \{x}, \{y}, \{z})");
		try {
		    functions.glVertexAttribI3i().invokeExact(index, x, y, z);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribI3iEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribI3iEXT.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribI3iEXT(int index, int x, int y, int z) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribI3iEXT(\{index}, \{x}, \{y}, \{z})");
		try {
		    functions.glVertexAttribI3iEXT().invokeExact(index, x, y, z);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribI3iv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribI3iv.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribI3iv(int index, ConstPointer<Integer> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribI3iv(\{index}, \{v})");
		try {
		    functions.glVertexAttribI3iv().invokeExact(index, v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribI3ivEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribI3ivEXT.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribI3ivEXT(int index, ConstPointer<Integer> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribI3ivEXT(\{index}, \{v})");
		try {
		    functions.glVertexAttribI3ivEXT().invokeExact(index, v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribI3ui">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribI3ui.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribI3ui(int index, int x, int y, int z) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribI3ui(\{index}, \{x}, \{y}, \{z})");
		try {
		    functions.glVertexAttribI3ui().invokeExact(index, x, y, z);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribI3uiEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribI3uiEXT.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribI3uiEXT(int index, int x, int y, int z) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribI3uiEXT(\{index}, \{x}, \{y}, \{z})");
		try {
		    functions.glVertexAttribI3uiEXT().invokeExact(index, x, y, z);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribI3uiv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribI3uiv.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribI3uiv(int index, ConstPointer<Integer> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribI3uiv(\{index}, \{v})");
		try {
		    functions.glVertexAttribI3uiv().invokeExact(index, v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribI3uivEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribI3uivEXT.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribI3uivEXT(int index, ConstPointer<Integer> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribI3uivEXT(\{index}, \{v})");
		try {
		    functions.glVertexAttribI3uivEXT().invokeExact(index, v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribI4bv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribI4bv.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribI4bv(int index, ConstPointer<Byte> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribI4bv(\{index}, \{v})");
		try {
		    functions.glVertexAttribI4bv().invokeExact(index, v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribI4bvEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribI4bvEXT.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribI4bvEXT(int index, ConstPointer<Byte> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribI4bvEXT(\{index}, \{v})");
		try {
		    functions.glVertexAttribI4bvEXT().invokeExact(index, v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribI4i">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribI4i.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribI4i(int index, int x, int y, int z, int w) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribI4i(\{index}, \{x}, \{y}, \{z}, \{w})");
		try {
		    functions.glVertexAttribI4i().invokeExact(index, x, y, z, w);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribI4iEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribI4iEXT.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribI4iEXT(int index, int x, int y, int z, int w) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribI4iEXT(\{index}, \{x}, \{y}, \{z}, \{w})");
		try {
		    functions.glVertexAttribI4iEXT().invokeExact(index, x, y, z, w);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribI4iv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribI4iv.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribI4iv(int index, ConstPointer<Integer> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribI4iv(\{index}, \{v})");
		try {
		    functions.glVertexAttribI4iv().invokeExact(index, v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribI4ivEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribI4ivEXT.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribI4ivEXT(int index, ConstPointer<Integer> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribI4ivEXT(\{index}, \{v})");
		try {
		    functions.glVertexAttribI4ivEXT().invokeExact(index, v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribI4sv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribI4sv.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribI4sv(int index, ConstPointer<Short> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribI4sv(\{index}, \{v})");
		try {
		    functions.glVertexAttribI4sv().invokeExact(index, v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribI4svEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribI4svEXT.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribI4svEXT(int index, ConstPointer<Short> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribI4svEXT(\{index}, \{v})");
		try {
		    functions.glVertexAttribI4svEXT().invokeExact(index, v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribI4ubv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribI4ubv.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribI4ubv(int index, ConstPointer<Byte> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribI4ubv(\{index}, \{v})");
		try {
		    functions.glVertexAttribI4ubv().invokeExact(index, v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribI4ubvEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribI4ubvEXT.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribI4ubvEXT(int index, ConstPointer<Byte> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribI4ubvEXT(\{index}, \{v})");
		try {
		    functions.glVertexAttribI4ubvEXT().invokeExact(index, v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribI4ui">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribI4ui.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribI4ui(int index, int x, int y, int z, int w) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribI4ui(\{index}, \{x}, \{y}, \{z}, \{w})");
		try {
		    functions.glVertexAttribI4ui().invokeExact(index, x, y, z, w);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribI4uiEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribI4uiEXT.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribI4uiEXT(int index, int x, int y, int z, int w) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribI4uiEXT(\{index}, \{x}, \{y}, \{z}, \{w})");
		try {
		    functions.glVertexAttribI4uiEXT().invokeExact(index, x, y, z, w);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribI4uiv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribI4uiv.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribI4uiv(int index, ConstPointer<Integer> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribI4uiv(\{index}, \{v})");
		try {
		    functions.glVertexAttribI4uiv().invokeExact(index, v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribI4uivEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribI4uivEXT.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribI4uivEXT(int index, ConstPointer<Integer> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribI4uivEXT(\{index}, \{v})");
		try {
		    functions.glVertexAttribI4uivEXT().invokeExact(index, v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribI4usv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribI4usv.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribI4usv(int index, ConstPointer<Character> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribI4usv(\{index}, \{v})");
		try {
		    functions.glVertexAttribI4usv().invokeExact(index, v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribI4usvEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribI4usvEXT.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribI4usvEXT(int index, ConstPointer<Character> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribI4usvEXT(\{index}, \{v})");
		try {
		    functions.glVertexAttribI4usvEXT().invokeExact(index, v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribIFormat">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribIFormat.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribIFormat(int attribindex, int size, GLVertexAttribIType type, int relativeoffset) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribIFormat(\{attribindex}, \{size}, \{type}, \{relativeoffset})");
		try {
		    functions.glVertexAttribIFormat().invokeExact(attribindex, size, type, relativeoffset);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribIFormatNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribIFormatNV.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribIFormatNV(int index, int size, GLVertexAttribIType type, int stride) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribIFormatNV(\{index}, \{size}, \{type}, \{stride})");
		try {
		    functions.glVertexAttribIFormatNV().invokeExact(index, size, type, stride);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribIPointer">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribIPointer.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribIPointer(int index, int size, GLVertexAttribIType type, int stride, ConstPointer<Void> pointer) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribIPointer(\{index}, \{size}, \{type}, \{stride}, \{pointer})");
		try {
		    functions.glVertexAttribIPointer().invokeExact(index, size, type, stride, pointer);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribIPointerEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribIPointerEXT.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribIPointerEXT(int index, int size, GLVertexAttribIType type, int stride, ConstPointer<Void> pointer) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribIPointerEXT(\{index}, \{size}, \{type}, \{stride}, \{pointer})");
		try {
		    functions.glVertexAttribIPointerEXT().invokeExact(index, size, type, stride, pointer);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribL1d">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribL1d.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribL1d(int index, double x) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribL1d(\{index}, \{x})");
		try {
		    functions.glVertexAttribL1d().invokeExact(index, x);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribL1dEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribL1dEXT.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribL1dEXT(int index, double x) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribL1dEXT(\{index}, \{x})");
		try {
		    functions.glVertexAttribL1dEXT().invokeExact(index, x);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribL1dv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribL1dv.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribL1dv(int index, ConstPointer<Double> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribL1dv(\{index}, \{v})");
		try {
		    functions.glVertexAttribL1dv().invokeExact(index, v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribL1dvEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribL1dvEXT.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribL1dvEXT(int index, ConstPointer<Double> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribL1dvEXT(\{index}, \{v})");
		try {
		    functions.glVertexAttribL1dvEXT().invokeExact(index, v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribL1i64NV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribL1i64NV.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribL1i64NV(int index, long x) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribL1i64NV(\{index}, \{x})");
		try {
		    functions.glVertexAttribL1i64NV().invokeExact(index, x);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribL1i64vNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribL1i64vNV.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribL1i64vNV(int index, ConstPointer<Long> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribL1i64vNV(\{index}, \{v})");
		try {
		    functions.glVertexAttribL1i64vNV().invokeExact(index, v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribL1ui64ARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribL1ui64ARB.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribL1ui64ARB(int index, long x) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribL1ui64ARB(\{index}, \{x})");
		try {
		    functions.glVertexAttribL1ui64ARB().invokeExact(index, x);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribL1ui64NV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribL1ui64NV.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribL1ui64NV(int index, long x) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribL1ui64NV(\{index}, \{x})");
		try {
		    functions.glVertexAttribL1ui64NV().invokeExact(index, x);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribL1ui64vARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribL1ui64vARB.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribL1ui64vARB(int index, ConstPointer<Long> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribL1ui64vARB(\{index}, \{v})");
		try {
		    functions.glVertexAttribL1ui64vARB().invokeExact(index, v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribL1ui64vNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribL1ui64vNV.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribL1ui64vNV(int index, ConstPointer<Long> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribL1ui64vNV(\{index}, \{v})");
		try {
		    functions.glVertexAttribL1ui64vNV().invokeExact(index, v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribL2d">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribL2d.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribL2d(int index, double x, double y) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribL2d(\{index}, \{x}, \{y})");
		try {
		    functions.glVertexAttribL2d().invokeExact(index, x, y);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribL2dEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribL2dEXT.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribL2dEXT(int index, double x, double y) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribL2dEXT(\{index}, \{x}, \{y})");
		try {
		    functions.glVertexAttribL2dEXT().invokeExact(index, x, y);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribL2dv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribL2dv.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribL2dv(int index, ConstPointer<Double> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribL2dv(\{index}, \{v})");
		try {
		    functions.glVertexAttribL2dv().invokeExact(index, v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribL2dvEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribL2dvEXT.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribL2dvEXT(int index, ConstPointer<Double> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribL2dvEXT(\{index}, \{v})");
		try {
		    functions.glVertexAttribL2dvEXT().invokeExact(index, v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribL2i64NV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribL2i64NV.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribL2i64NV(int index, long x, long y) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribL2i64NV(\{index}, \{x}, \{y})");
		try {
		    functions.glVertexAttribL2i64NV().invokeExact(index, x, y);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribL2i64vNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribL2i64vNV.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribL2i64vNV(int index, ConstPointer<Long> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribL2i64vNV(\{index}, \{v})");
		try {
		    functions.glVertexAttribL2i64vNV().invokeExact(index, v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribL2ui64NV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribL2ui64NV.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribL2ui64NV(int index, long x, long y) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribL2ui64NV(\{index}, \{x}, \{y})");
		try {
		    functions.glVertexAttribL2ui64NV().invokeExact(index, x, y);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribL2ui64vNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribL2ui64vNV.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribL2ui64vNV(int index, ConstPointer<Long> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribL2ui64vNV(\{index}, \{v})");
		try {
		    functions.glVertexAttribL2ui64vNV().invokeExact(index, v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribL3d">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribL3d.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribL3d(int index, double x, double y, double z) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribL3d(\{index}, \{x}, \{y}, \{z})");
		try {
		    functions.glVertexAttribL3d().invokeExact(index, x, y, z);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribL3dEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribL3dEXT.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribL3dEXT(int index, double x, double y, double z) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribL3dEXT(\{index}, \{x}, \{y}, \{z})");
		try {
		    functions.glVertexAttribL3dEXT().invokeExact(index, x, y, z);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribL3dv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribL3dv.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribL3dv(int index, ConstPointer<Double> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribL3dv(\{index}, \{v})");
		try {
		    functions.glVertexAttribL3dv().invokeExact(index, v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribL3dvEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribL3dvEXT.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribL3dvEXT(int index, ConstPointer<Double> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribL3dvEXT(\{index}, \{v})");
		try {
		    functions.glVertexAttribL3dvEXT().invokeExact(index, v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribL3i64NV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribL3i64NV.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribL3i64NV(int index, long x, long y, long z) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribL3i64NV(\{index}, \{x}, \{y}, \{z})");
		try {
		    functions.glVertexAttribL3i64NV().invokeExact(index, x, y, z);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribL3i64vNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribL3i64vNV.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribL3i64vNV(int index, ConstPointer<Long> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribL3i64vNV(\{index}, \{v})");
		try {
		    functions.glVertexAttribL3i64vNV().invokeExact(index, v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribL3ui64NV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribL3ui64NV.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribL3ui64NV(int index, long x, long y, long z) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribL3ui64NV(\{index}, \{x}, \{y}, \{z})");
		try {
		    functions.glVertexAttribL3ui64NV().invokeExact(index, x, y, z);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribL3ui64vNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribL3ui64vNV.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribL3ui64vNV(int index, ConstPointer<Long> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribL3ui64vNV(\{index}, \{v})");
		try {
		    functions.glVertexAttribL3ui64vNV().invokeExact(index, v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribL4d">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribL4d.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribL4d(int index, double x, double y, double z, double w) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribL4d(\{index}, \{x}, \{y}, \{z}, \{w})");
		try {
		    functions.glVertexAttribL4d().invokeExact(index, x, y, z, w);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribL4dEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribL4dEXT.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribL4dEXT(int index, double x, double y, double z, double w) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribL4dEXT(\{index}, \{x}, \{y}, \{z}, \{w})");
		try {
		    functions.glVertexAttribL4dEXT().invokeExact(index, x, y, z, w);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribL4dv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribL4dv.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribL4dv(int index, ConstPointer<Double> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribL4dv(\{index}, \{v})");
		try {
		    functions.glVertexAttribL4dv().invokeExact(index, v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribL4dvEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribL4dvEXT.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribL4dvEXT(int index, ConstPointer<Double> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribL4dvEXT(\{index}, \{v})");
		try {
		    functions.glVertexAttribL4dvEXT().invokeExact(index, v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribL4i64NV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribL4i64NV.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribL4i64NV(int index, long x, long y, long z, long w) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribL4i64NV(\{index}, \{x}, \{y}, \{z}, \{w})");
		try {
		    functions.glVertexAttribL4i64NV().invokeExact(index, x, y, z, w);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribL4i64vNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribL4i64vNV.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribL4i64vNV(int index, ConstPointer<Long> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribL4i64vNV(\{index}, \{v})");
		try {
		    functions.glVertexAttribL4i64vNV().invokeExact(index, v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribL4ui64NV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribL4ui64NV.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribL4ui64NV(int index, long x, long y, long z, long w) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribL4ui64NV(\{index}, \{x}, \{y}, \{z}, \{w})");
		try {
		    functions.glVertexAttribL4ui64NV().invokeExact(index, x, y, z, w);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribL4ui64vNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribL4ui64vNV.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribL4ui64vNV(int index, ConstPointer<Long> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribL4ui64vNV(\{index}, \{v})");
		try {
		    functions.glVertexAttribL4ui64vNV().invokeExact(index, v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribLFormat">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribLFormat.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribLFormat(int attribindex, int size, GLVertexAttribLType type, int relativeoffset) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribLFormat(\{attribindex}, \{size}, \{type}, \{relativeoffset})");
		try {
		    functions.glVertexAttribLFormat().invokeExact(attribindex, size, type, relativeoffset);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribLFormatNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribLFormatNV.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribLFormatNV(int index, int size, GLVertexAttribLType type, int stride) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribLFormatNV(\{index}, \{size}, \{type}, \{stride})");
		try {
		    functions.glVertexAttribLFormatNV().invokeExact(index, size, type, stride);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribLPointer">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribLPointer.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribLPointer(int index, int size, GLVertexAttribLType type, int stride, ConstPointer<Void> pointer) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribLPointer(\{index}, \{size}, \{type}, \{stride}, \{pointer})");
		try {
		    functions.glVertexAttribLPointer().invokeExact(index, size, type, stride, pointer);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribLPointerEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribLPointerEXT.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribLPointerEXT(int index, int size, GLVertexAttribLType type, int stride, ConstPointer<Void> pointer) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribLPointerEXT(\{index}, \{size}, \{type}, \{stride}, \{pointer})");
		try {
		    functions.glVertexAttribLPointerEXT().invokeExact(index, size, type, stride, pointer);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribP1ui">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribP1ui.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribP1ui(int index, GLVertexAttribPointerType type, boolean normalized, int value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribP1ui(\{index}, \{type}, \{normalized}, \{value})");
		try {
		    functions.glVertexAttribP1ui().invokeExact(index, type, normalized, value);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribP1uiv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribP1uiv.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribP1uiv(int index, GLVertexAttribPointerType type, boolean normalized, ConstPointer<Integer> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribP1uiv(\{index}, \{type}, \{normalized}, \{value})");
		try {
		    functions.glVertexAttribP1uiv().invokeExact(index, type, normalized, value);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribP2ui">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribP2ui.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribP2ui(int index, GLVertexAttribPointerType type, boolean normalized, int value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribP2ui(\{index}, \{type}, \{normalized}, \{value})");
		try {
		    functions.glVertexAttribP2ui().invokeExact(index, type, normalized, value);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribP2uiv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribP2uiv.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribP2uiv(int index, GLVertexAttribPointerType type, boolean normalized, ConstPointer<Integer> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribP2uiv(\{index}, \{type}, \{normalized}, \{value})");
		try {
		    functions.glVertexAttribP2uiv().invokeExact(index, type, normalized, value);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribP3ui">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribP3ui.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribP3ui(int index, GLVertexAttribPointerType type, boolean normalized, int value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribP3ui(\{index}, \{type}, \{normalized}, \{value})");
		try {
		    functions.glVertexAttribP3ui().invokeExact(index, type, normalized, value);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribP3uiv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribP3uiv.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribP3uiv(int index, GLVertexAttribPointerType type, boolean normalized, ConstPointer<Integer> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribP3uiv(\{index}, \{type}, \{normalized}, \{value})");
		try {
		    functions.glVertexAttribP3uiv().invokeExact(index, type, normalized, value);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribP4ui">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribP4ui.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribP4ui(int index, GLVertexAttribPointerType type, boolean normalized, int value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribP4ui(\{index}, \{type}, \{normalized}, \{value})");
		try {
		    functions.glVertexAttribP4ui().invokeExact(index, type, normalized, value);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribP4uiv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribP4uiv.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribP4uiv(int index, GLVertexAttribPointerType type, boolean normalized, ConstPointer<Integer> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribP4uiv(\{index}, \{type}, \{normalized}, \{value})");
		try {
		    functions.glVertexAttribP4uiv().invokeExact(index, type, normalized, value);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribParameteriAMD">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribParameteriAMD.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribParameteriAMD(int index, GLEnum pname, int param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribParameteriAMD(\{index}, \{pname}, \{param})");
		try {
		    functions.glVertexAttribParameteriAMD().invokeExact(index, pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribPointer">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribPointer.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribPointer(int index, int size, GLVertexAttribPointerType type, boolean normalized, int stride, ConstPointer<Void> pointer) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribPointer(\{index}, \{size}, \{type}, \{normalized}, \{stride}, \{pointer})");
		try {
		    functions.glVertexAttribPointer().invokeExact(index, size, type, normalized, stride, pointer);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribPointerARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribPointerARB.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribPointerARB(int index, int size, GLVertexAttribPointerType type, boolean normalized, int stride, ConstPointer<Void> pointer) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribPointerARB(\{index}, \{size}, \{type}, \{normalized}, \{stride}, \{pointer})");
		try {
		    functions.glVertexAttribPointerARB().invokeExact(index, size, type, normalized, stride, pointer);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribPointerNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribPointerNV.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribPointerNV(int index, int fsize, GLVertexAttribEnumNV type, int stride, ConstPointer<Void> pointer) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribPointerNV(\{index}, \{fsize}, \{type}, \{stride}, \{pointer})");
		try {
		    functions.glVertexAttribPointerNV().invokeExact(index, fsize, type, stride, pointer);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribs1dvNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribs1dvNV.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribs1dvNV(int index, int count, ConstPointer<Double> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribs1dvNV(\{index}, \{count}, \{v})");
		try {
		    functions.glVertexAttribs1dvNV().invokeExact(index, count, v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribs1fvNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribs1fvNV.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribs1fvNV(int index, int count, ConstPointer<Float> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribs1fvNV(\{index}, \{count}, \{v})");
		try {
		    functions.glVertexAttribs1fvNV().invokeExact(index, count, v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribs1hvNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribs1hvNV.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribs1hvNV(int index, int n, ConstPointer<Character> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribs1hvNV(\{index}, \{n}, \{v})");
		try {
		    functions.glVertexAttribs1hvNV().invokeExact(index, n, v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribs1svNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribs1svNV.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribs1svNV(int index, int count, ConstPointer<Short> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribs1svNV(\{index}, \{count}, \{v})");
		try {
		    functions.glVertexAttribs1svNV().invokeExact(index, count, v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribs2dvNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribs2dvNV.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribs2dvNV(int index, int count, ConstPointer<Double> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribs2dvNV(\{index}, \{count}, \{v})");
		try {
		    functions.glVertexAttribs2dvNV().invokeExact(index, count, v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribs2fvNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribs2fvNV.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribs2fvNV(int index, int count, ConstPointer<Float> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribs2fvNV(\{index}, \{count}, \{v})");
		try {
		    functions.glVertexAttribs2fvNV().invokeExact(index, count, v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribs2hvNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribs2hvNV.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribs2hvNV(int index, int n, ConstPointer<Character> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribs2hvNV(\{index}, \{n}, \{v})");
		try {
		    functions.glVertexAttribs2hvNV().invokeExact(index, n, v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribs2svNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribs2svNV.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribs2svNV(int index, int count, ConstPointer<Short> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribs2svNV(\{index}, \{count}, \{v})");
		try {
		    functions.glVertexAttribs2svNV().invokeExact(index, count, v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribs3dvNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribs3dvNV.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribs3dvNV(int index, int count, ConstPointer<Double> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribs3dvNV(\{index}, \{count}, \{v})");
		try {
		    functions.glVertexAttribs3dvNV().invokeExact(index, count, v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribs3fvNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribs3fvNV.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribs3fvNV(int index, int count, ConstPointer<Float> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribs3fvNV(\{index}, \{count}, \{v})");
		try {
		    functions.glVertexAttribs3fvNV().invokeExact(index, count, v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribs3hvNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribs3hvNV.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribs3hvNV(int index, int n, ConstPointer<Character> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribs3hvNV(\{index}, \{n}, \{v})");
		try {
		    functions.glVertexAttribs3hvNV().invokeExact(index, n, v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribs3svNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribs3svNV.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribs3svNV(int index, int count, ConstPointer<Short> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribs3svNV(\{index}, \{count}, \{v})");
		try {
		    functions.glVertexAttribs3svNV().invokeExact(index, count, v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribs4dvNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribs4dvNV.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribs4dvNV(int index, int count, ConstPointer<Double> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribs4dvNV(\{index}, \{count}, \{v})");
		try {
		    functions.glVertexAttribs4dvNV().invokeExact(index, count, v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribs4fvNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribs4fvNV.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribs4fvNV(int index, int count, ConstPointer<Float> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribs4fvNV(\{index}, \{count}, \{v})");
		try {
		    functions.glVertexAttribs4fvNV().invokeExact(index, count, v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribs4hvNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribs4hvNV.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribs4hvNV(int index, int n, ConstPointer<Character> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribs4hvNV(\{index}, \{n}, \{v})");
		try {
		    functions.glVertexAttribs4hvNV().invokeExact(index, n, v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribs4svNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribs4svNV.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribs4svNV(int index, int count, ConstPointer<Short> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribs4svNV(\{index}, \{count}, \{v})");
		try {
		    functions.glVertexAttribs4svNV().invokeExact(index, count, v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexAttribs4ubvNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribs4ubvNV.xhtml">registry.khronos.org</a>
	 */
	public void glVertexAttribs4ubvNV(int index, int count, ConstPointer<Byte> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexAttribs4ubvNV(\{index}, \{count}, \{v})");
		try {
		    functions.glVertexAttribs4ubvNV().invokeExact(index, count, v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexBindingDivisor">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexBindingDivisor.xhtml">registry.khronos.org</a>
	 */
	public void glVertexBindingDivisor(int bindingindex, int divisor) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexBindingDivisor(\{bindingindex}, \{divisor})");
		try {
		    functions.glVertexBindingDivisor().invokeExact(bindingindex, divisor);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexBlendARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexBlendARB.xhtml">registry.khronos.org</a>
	 */
	public void glVertexBlendARB(int count) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexBlendARB(\{count})");
		try {
		    functions.glVertexBlendARB().invokeExact(count);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexBlendEnvfATI">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexBlendEnvfATI.xhtml">registry.khronos.org</a>
	 */
	public void glVertexBlendEnvfATI(GLVertexStreamATI pname, float param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexBlendEnvfATI(\{pname}, \{param})");
		try {
		    functions.glVertexBlendEnvfATI().invokeExact(pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexBlendEnviATI">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexBlendEnviATI.xhtml">registry.khronos.org</a>
	 */
	public void glVertexBlendEnviATI(GLVertexStreamATI pname, int param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexBlendEnviATI(\{pname}, \{param})");
		try {
		    functions.glVertexBlendEnviATI().invokeExact(pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexFormatNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexFormatNV.xhtml">registry.khronos.org</a>
	 */
	public void glVertexFormatNV(int size, GLVertexPointerType type, int stride) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexFormatNV(\{size}, \{type}, \{stride})");
		try {
		    functions.glVertexFormatNV().invokeExact(size, type, stride);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexP2ui">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexP2ui.xhtml">registry.khronos.org</a>
	 */
	public void glVertexP2ui(GLVertexPointerType type, int value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexP2ui(\{type}, \{value})");
		try {
		    functions.glVertexP2ui().invokeExact(type, value);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexP2uiv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexP2uiv.xhtml">registry.khronos.org</a>
	 */
	public void glVertexP2uiv(GLVertexPointerType type, ConstPointer<Integer> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexP2uiv(\{type}, \{value})");
		try {
		    functions.glVertexP2uiv().invokeExact(type, value);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexP3ui">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexP3ui.xhtml">registry.khronos.org</a>
	 */
	public void glVertexP3ui(GLVertexPointerType type, int value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexP3ui(\{type}, \{value})");
		try {
		    functions.glVertexP3ui().invokeExact(type, value);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexP3uiv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexP3uiv.xhtml">registry.khronos.org</a>
	 */
	public void glVertexP3uiv(GLVertexPointerType type, ConstPointer<Integer> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexP3uiv(\{type}, \{value})");
		try {
		    functions.glVertexP3uiv().invokeExact(type, value);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexP4ui">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexP4ui.xhtml">registry.khronos.org</a>
	 */
	public void glVertexP4ui(GLVertexPointerType type, int value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexP4ui(\{type}, \{value})");
		try {
		    functions.glVertexP4ui().invokeExact(type, value);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexP4uiv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexP4uiv.xhtml">registry.khronos.org</a>
	 */
	public void glVertexP4uiv(GLVertexPointerType type, ConstPointer<Integer> value) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexP4uiv(\{type}, \{value})");
		try {
		    functions.glVertexP4uiv().invokeExact(type, value);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexPointer">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexPointer.xhtml">registry.khronos.org</a>
	 */
	public void glVertexPointer(int size, GLVertexPointerType type, int stride, ConstPointer<Void> pointer) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexPointer(\{size}, \{type}, \{stride}, \{pointer})");
		try {
		    functions.glVertexPointer().invokeExact(size, type, stride, pointer);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexPointerEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexPointerEXT.xhtml">registry.khronos.org</a>
	 */
	public void glVertexPointerEXT(int size, GLVertexPointerType type, int stride, int count, ConstPointer<Void> pointer) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexPointerEXT(\{size}, \{type}, \{stride}, \{count}, \{pointer})");
		try {
		    functions.glVertexPointerEXT().invokeExact(size, type, stride, count, pointer);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexPointerListIBM">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexPointerListIBM.xhtml">registry.khronos.org</a>
	 */
	public void glVertexPointerListIBM(int size, GLVertexPointerType type, int stride, MutPointer<ConstPointer<Void>> pointer, int ptrstride) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexPointerListIBM(\{size}, \{type}, \{stride}, \{pointer}, \{ptrstride})");
		try {
		    functions.glVertexPointerListIBM().invokeExact(size, type, stride, pointer, ptrstride);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexPointervINTEL">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexPointervINTEL.xhtml">registry.khronos.org</a>
	 */
	public void glVertexPointervINTEL(int size, GLVertexPointerType type, MutPointer<ConstPointer<Void>> pointer) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexPointervINTEL(\{size}, \{type}, \{pointer})");
		try {
		    functions.glVertexPointervINTEL().invokeExact(size, type, pointer);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexStream1dATI">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexStream1dATI.xhtml">registry.khronos.org</a>
	 */
	public void glVertexStream1dATI(GLVertexStreamATI stream, double x) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexStream1dATI(\{stream}, \{x})");
		try {
		    functions.glVertexStream1dATI().invokeExact(stream, x);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were stream=\{stream}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexStream1dvATI">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexStream1dvATI.xhtml">registry.khronos.org</a>
	 */
	public void glVertexStream1dvATI(GLVertexStreamATI stream, ConstPointer<Double> coords) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexStream1dvATI(\{stream}, \{coords})");
		try {
		    functions.glVertexStream1dvATI().invokeExact(stream, coords);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were stream=\{stream}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexStream1fATI">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexStream1fATI.xhtml">registry.khronos.org</a>
	 */
	public void glVertexStream1fATI(GLVertexStreamATI stream, float x) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexStream1fATI(\{stream}, \{x})");
		try {
		    functions.glVertexStream1fATI().invokeExact(stream, x);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were stream=\{stream}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexStream1fvATI">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexStream1fvATI.xhtml">registry.khronos.org</a>
	 */
	public void glVertexStream1fvATI(GLVertexStreamATI stream, ConstPointer<Float> coords) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexStream1fvATI(\{stream}, \{coords})");
		try {
		    functions.glVertexStream1fvATI().invokeExact(stream, coords);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were stream=\{stream}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexStream1iATI">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexStream1iATI.xhtml">registry.khronos.org</a>
	 */
	public void glVertexStream1iATI(GLVertexStreamATI stream, int x) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexStream1iATI(\{stream}, \{x})");
		try {
		    functions.glVertexStream1iATI().invokeExact(stream, x);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were stream=\{stream}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexStream1ivATI">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexStream1ivATI.xhtml">registry.khronos.org</a>
	 */
	public void glVertexStream1ivATI(GLVertexStreamATI stream, ConstPointer<Integer> coords) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexStream1ivATI(\{stream}, \{coords})");
		try {
		    functions.glVertexStream1ivATI().invokeExact(stream, coords);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were stream=\{stream}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexStream1sATI">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexStream1sATI.xhtml">registry.khronos.org</a>
	 */
	public void glVertexStream1sATI(GLVertexStreamATI stream, short x) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexStream1sATI(\{stream}, \{x})");
		try {
		    functions.glVertexStream1sATI().invokeExact(stream, x);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were stream=\{stream}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexStream1svATI">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexStream1svATI.xhtml">registry.khronos.org</a>
	 */
	public void glVertexStream1svATI(GLVertexStreamATI stream, ConstPointer<Short> coords) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexStream1svATI(\{stream}, \{coords})");
		try {
		    functions.glVertexStream1svATI().invokeExact(stream, coords);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were stream=\{stream}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexStream2dATI">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexStream2dATI.xhtml">registry.khronos.org</a>
	 */
	public void glVertexStream2dATI(GLVertexStreamATI stream, double x, double y) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexStream2dATI(\{stream}, \{x}, \{y})");
		try {
		    functions.glVertexStream2dATI().invokeExact(stream, x, y);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were stream=\{stream}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexStream2dvATI">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexStream2dvATI.xhtml">registry.khronos.org</a>
	 */
	public void glVertexStream2dvATI(GLVertexStreamATI stream, ConstPointer<Double> coords) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexStream2dvATI(\{stream}, \{coords})");
		try {
		    functions.glVertexStream2dvATI().invokeExact(stream, coords);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were stream=\{stream}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexStream2fATI">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexStream2fATI.xhtml">registry.khronos.org</a>
	 */
	public void glVertexStream2fATI(GLVertexStreamATI stream, float x, float y) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexStream2fATI(\{stream}, \{x}, \{y})");
		try {
		    functions.glVertexStream2fATI().invokeExact(stream, x, y);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were stream=\{stream}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexStream2fvATI">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexStream2fvATI.xhtml">registry.khronos.org</a>
	 */
	public void glVertexStream2fvATI(GLVertexStreamATI stream, ConstPointer<Float> coords) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexStream2fvATI(\{stream}, \{coords})");
		try {
		    functions.glVertexStream2fvATI().invokeExact(stream, coords);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were stream=\{stream}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexStream2iATI">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexStream2iATI.xhtml">registry.khronos.org</a>
	 */
	public void glVertexStream2iATI(GLVertexStreamATI stream, int x, int y) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexStream2iATI(\{stream}, \{x}, \{y})");
		try {
		    functions.glVertexStream2iATI().invokeExact(stream, x, y);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were stream=\{stream}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexStream2ivATI">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexStream2ivATI.xhtml">registry.khronos.org</a>
	 */
	public void glVertexStream2ivATI(GLVertexStreamATI stream, ConstPointer<Integer> coords) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexStream2ivATI(\{stream}, \{coords})");
		try {
		    functions.glVertexStream2ivATI().invokeExact(stream, coords);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were stream=\{stream}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexStream2sATI">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexStream2sATI.xhtml">registry.khronos.org</a>
	 */
	public void glVertexStream2sATI(GLVertexStreamATI stream, short x, short y) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexStream2sATI(\{stream}, \{x}, \{y})");
		try {
		    functions.glVertexStream2sATI().invokeExact(stream, x, y);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were stream=\{stream}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexStream2svATI">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexStream2svATI.xhtml">registry.khronos.org</a>
	 */
	public void glVertexStream2svATI(GLVertexStreamATI stream, ConstPointer<Short> coords) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexStream2svATI(\{stream}, \{coords})");
		try {
		    functions.glVertexStream2svATI().invokeExact(stream, coords);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were stream=\{stream}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexStream3dATI">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexStream3dATI.xhtml">registry.khronos.org</a>
	 */
	public void glVertexStream3dATI(GLVertexStreamATI stream, double x, double y, double z) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexStream3dATI(\{stream}, \{x}, \{y}, \{z})");
		try {
		    functions.glVertexStream3dATI().invokeExact(stream, x, y, z);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were stream=\{stream}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexStream3dvATI">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexStream3dvATI.xhtml">registry.khronos.org</a>
	 */
	public void glVertexStream3dvATI(GLVertexStreamATI stream, ConstPointer<Double> coords) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexStream3dvATI(\{stream}, \{coords})");
		try {
		    functions.glVertexStream3dvATI().invokeExact(stream, coords);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were stream=\{stream}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexStream3fATI">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexStream3fATI.xhtml">registry.khronos.org</a>
	 */
	public void glVertexStream3fATI(GLVertexStreamATI stream, float x, float y, float z) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexStream3fATI(\{stream}, \{x}, \{y}, \{z})");
		try {
		    functions.glVertexStream3fATI().invokeExact(stream, x, y, z);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were stream=\{stream}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexStream3fvATI">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexStream3fvATI.xhtml">registry.khronos.org</a>
	 */
	public void glVertexStream3fvATI(GLVertexStreamATI stream, ConstPointer<Float> coords) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexStream3fvATI(\{stream}, \{coords})");
		try {
		    functions.glVertexStream3fvATI().invokeExact(stream, coords);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were stream=\{stream}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexStream3iATI">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexStream3iATI.xhtml">registry.khronos.org</a>
	 */
	public void glVertexStream3iATI(GLVertexStreamATI stream, int x, int y, int z) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexStream3iATI(\{stream}, \{x}, \{y}, \{z})");
		try {
		    functions.glVertexStream3iATI().invokeExact(stream, x, y, z);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were stream=\{stream}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexStream3ivATI">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexStream3ivATI.xhtml">registry.khronos.org</a>
	 */
	public void glVertexStream3ivATI(GLVertexStreamATI stream, ConstPointer<Integer> coords) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexStream3ivATI(\{stream}, \{coords})");
		try {
		    functions.glVertexStream3ivATI().invokeExact(stream, coords);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were stream=\{stream}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexStream3sATI">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexStream3sATI.xhtml">registry.khronos.org</a>
	 */
	public void glVertexStream3sATI(GLVertexStreamATI stream, short x, short y, short z) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexStream3sATI(\{stream}, \{x}, \{y}, \{z})");
		try {
		    functions.glVertexStream3sATI().invokeExact(stream, x, y, z);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were stream=\{stream}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexStream3svATI">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexStream3svATI.xhtml">registry.khronos.org</a>
	 */
	public void glVertexStream3svATI(GLVertexStreamATI stream, ConstPointer<Short> coords) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexStream3svATI(\{stream}, \{coords})");
		try {
		    functions.glVertexStream3svATI().invokeExact(stream, coords);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were stream=\{stream}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexStream4dATI">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexStream4dATI.xhtml">registry.khronos.org</a>
	 */
	public void glVertexStream4dATI(GLVertexStreamATI stream, double x, double y, double z, double w) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexStream4dATI(\{stream}, \{x}, \{y}, \{z}, \{w})");
		try {
		    functions.glVertexStream4dATI().invokeExact(stream, x, y, z, w);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were stream=\{stream}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexStream4dvATI">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexStream4dvATI.xhtml">registry.khronos.org</a>
	 */
	public void glVertexStream4dvATI(GLVertexStreamATI stream, ConstPointer<Double> coords) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexStream4dvATI(\{stream}, \{coords})");
		try {
		    functions.glVertexStream4dvATI().invokeExact(stream, coords);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were stream=\{stream}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexStream4fATI">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexStream4fATI.xhtml">registry.khronos.org</a>
	 */
	public void glVertexStream4fATI(GLVertexStreamATI stream, float x, float y, float z, float w) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexStream4fATI(\{stream}, \{x}, \{y}, \{z}, \{w})");
		try {
		    functions.glVertexStream4fATI().invokeExact(stream, x, y, z, w);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were stream=\{stream}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexStream4fvATI">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexStream4fvATI.xhtml">registry.khronos.org</a>
	 */
	public void glVertexStream4fvATI(GLVertexStreamATI stream, ConstPointer<Float> coords) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexStream4fvATI(\{stream}, \{coords})");
		try {
		    functions.glVertexStream4fvATI().invokeExact(stream, coords);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were stream=\{stream}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexStream4iATI">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexStream4iATI.xhtml">registry.khronos.org</a>
	 */
	public void glVertexStream4iATI(GLVertexStreamATI stream, int x, int y, int z, int w) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexStream4iATI(\{stream}, \{x}, \{y}, \{z}, \{w})");
		try {
		    functions.glVertexStream4iATI().invokeExact(stream, x, y, z, w);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were stream=\{stream}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexStream4ivATI">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexStream4ivATI.xhtml">registry.khronos.org</a>
	 */
	public void glVertexStream4ivATI(GLVertexStreamATI stream, ConstPointer<Integer> coords) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexStream4ivATI(\{stream}, \{coords})");
		try {
		    functions.glVertexStream4ivATI().invokeExact(stream, coords);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were stream=\{stream}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexStream4sATI">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexStream4sATI.xhtml">registry.khronos.org</a>
	 */
	public void glVertexStream4sATI(GLVertexStreamATI stream, short x, short y, short z, short w) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexStream4sATI(\{stream}, \{x}, \{y}, \{z}, \{w})");
		try {
		    functions.glVertexStream4sATI().invokeExact(stream, x, y, z, w);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were stream=\{stream}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexStream4svATI">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexStream4svATI.xhtml">registry.khronos.org</a>
	 */
	public void glVertexStream4svATI(GLVertexStreamATI stream, ConstPointer<Short> coords) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexStream4svATI(\{stream}, \{coords})");
		try {
		    functions.glVertexStream4svATI().invokeExact(stream, coords);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were stream=\{stream}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexWeightPointerEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexWeightPointerEXT.xhtml">registry.khronos.org</a>
	 */
	public void glVertexWeightPointerEXT(int size, GLVertexWeightPointerTypeEXT type, int stride, ConstPointer<Void> pointer) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexWeightPointerEXT(\{size}, \{type}, \{stride}, \{pointer})");
		try {
		    functions.glVertexWeightPointerEXT().invokeExact(size, type, stride, pointer);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexWeightfEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexWeightfEXT.xhtml">registry.khronos.org</a>
	 */
	public void glVertexWeightfEXT(float weight) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexWeightfEXT(\{weight})");
		try {
		    functions.glVertexWeightfEXT().invokeExact(weight);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexWeightfvEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexWeightfvEXT.xhtml">registry.khronos.org</a>
	 */
	public void glVertexWeightfvEXT(ConstPointer<Float> weight) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexWeightfvEXT(\{weight})");
		try {
		    functions.glVertexWeightfvEXT().invokeExact(weight);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexWeighthNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexWeighthNV.xhtml">registry.khronos.org</a>
	 */
	public void glVertexWeighthNV(char weight) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexWeighthNV(\{weight})");
		try {
		    functions.glVertexWeighthNV().invokeExact(weight);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVertexWeighthvNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexWeighthvNV.xhtml">registry.khronos.org</a>
	 */
	public void glVertexWeighthvNV(ConstPointer<Character> weight) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVertexWeighthvNV(\{weight})");
		try {
		    functions.glVertexWeighthvNV().invokeExact(weight);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVideoCaptureNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVideoCaptureNV.xhtml">registry.khronos.org</a>
	 */
	public GLEnum glVideoCaptureNV(int video_capture_slot, MutPointer<Integer> sequence_num, MutPointer<Long> capture_time) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVideoCaptureNV(\{video_capture_slot}, \{sequence_num}, \{capture_time})");
		try {
		    var r = (GLEnum)functions.glVideoCaptureNV().invokeExact(video_capture_slot, sequence_num, capture_time);
			checkError();
			return r;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVideoCaptureStreamParameterdvNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVideoCaptureStreamParameterdvNV.xhtml">registry.khronos.org</a>
	 */
	public void glVideoCaptureStreamParameterdvNV(int video_capture_slot, int stream, GLEnum pname, ConstPointer<Double> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVideoCaptureStreamParameterdvNV(\{video_capture_slot}, \{stream}, \{pname}, \{params})");
		try {
		    functions.glVideoCaptureStreamParameterdvNV().invokeExact(video_capture_slot, stream, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVideoCaptureStreamParameterfvNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVideoCaptureStreamParameterfvNV.xhtml">registry.khronos.org</a>
	 */
	public void glVideoCaptureStreamParameterfvNV(int video_capture_slot, int stream, GLEnum pname, ConstPointer<Float> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVideoCaptureStreamParameterfvNV(\{video_capture_slot}, \{stream}, \{pname}, \{params})");
		try {
		    functions.glVideoCaptureStreamParameterfvNV().invokeExact(video_capture_slot, stream, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glVideoCaptureStreamParameterivNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVideoCaptureStreamParameterivNV.xhtml">registry.khronos.org</a>
	 */
	public void glVideoCaptureStreamParameterivNV(int video_capture_slot, int stream, GLEnum pname, ConstPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glVideoCaptureStreamParameterivNV(\{video_capture_slot}, \{stream}, \{pname}, \{params})");
		try {
		    functions.glVideoCaptureStreamParameterivNV().invokeExact(video_capture_slot, stream, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glViewport">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glViewport.xhtml">registry.khronos.org</a>
	 */
	public void glViewport(int x, int y, int width, int height) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glViewport(\{x}, \{y}, \{width}, \{height})");
		try {
		    functions.glViewport().invokeExact(x, y, width, height);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glViewportArrayv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glViewportArrayv.xhtml">registry.khronos.org</a>
	 */
	public void glViewportArrayv(int first, int count, ConstPointer<Float> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glViewportArrayv(\{first}, \{count}, \{v})");
		try {
		    functions.glViewportArrayv().invokeExact(first, count, v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glViewportArrayvNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glViewportArrayvNV.xhtml">registry.khronos.org</a>
	 */
	public void glViewportArrayvNV(int first, int count, ConstPointer<Float> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glViewportArrayvNV(\{first}, \{count}, \{v})");
		try {
		    functions.glViewportArrayvNV().invokeExact(first, count, v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glViewportArrayvOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glViewportArrayvOES.xhtml">registry.khronos.org</a>
	 */
	public void glViewportArrayvOES(int first, int count, ConstPointer<Float> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glViewportArrayvOES(\{first}, \{count}, \{v})");
		try {
		    functions.glViewportArrayvOES().invokeExact(first, count, v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glViewportIndexedf">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glViewportIndexedf.xhtml">registry.khronos.org</a>
	 */
	public void glViewportIndexedf(int index, float x, float y, float w, float h) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glViewportIndexedf(\{index}, \{x}, \{y}, \{w}, \{h})");
		try {
		    functions.glViewportIndexedf().invokeExact(index, x, y, w, h);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glViewportIndexedfOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glViewportIndexedfOES.xhtml">registry.khronos.org</a>
	 */
	public void glViewportIndexedfOES(int index, float x, float y, float w, float h) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glViewportIndexedfOES(\{index}, \{x}, \{y}, \{w}, \{h})");
		try {
		    functions.glViewportIndexedfOES().invokeExact(index, x, y, w, h);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glViewportIndexedfNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glViewportIndexedfNV.xhtml">registry.khronos.org</a>
	 */
	public void glViewportIndexedfNV(int index, float x, float y, float w, float h) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glViewportIndexedfNV(\{index}, \{x}, \{y}, \{w}, \{h})");
		try {
		    functions.glViewportIndexedfNV().invokeExact(index, x, y, w, h);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glViewportIndexedfv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glViewportIndexedfv.xhtml">registry.khronos.org</a>
	 */
	public void glViewportIndexedfv(int index, ConstPointer<Float> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glViewportIndexedfv(\{index}, \{v})");
		try {
		    functions.glViewportIndexedfv().invokeExact(index, v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glViewportIndexedfvOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glViewportIndexedfvOES.xhtml">registry.khronos.org</a>
	 */
	public void glViewportIndexedfvOES(int index, ConstPointer<Float> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glViewportIndexedfvOES(\{index}, \{v})");
		try {
		    functions.glViewportIndexedfvOES().invokeExact(index, v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glViewportIndexedfvNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glViewportIndexedfvNV.xhtml">registry.khronos.org</a>
	 */
	public void glViewportIndexedfvNV(int index, ConstPointer<Float> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glViewportIndexedfvNV(\{index}, \{v})");
		try {
		    functions.glViewportIndexedfvNV().invokeExact(index, v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glViewportPositionWScaleNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glViewportPositionWScaleNV.xhtml">registry.khronos.org</a>
	 */
	public void glViewportPositionWScaleNV(int index, float xcoeff, float ycoeff) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glViewportPositionWScaleNV(\{index}, \{xcoeff}, \{ycoeff})");
		try {
		    functions.glViewportPositionWScaleNV().invokeExact(index, xcoeff, ycoeff);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glViewportSwizzleNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glViewportSwizzleNV.xhtml">registry.khronos.org</a>
	 */
	public void glViewportSwizzleNV(int index, GLEnum swizzlex, GLEnum swizzley, GLEnum swizzlez, GLEnum swizzlew) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glViewportSwizzleNV(\{index}, \{swizzlex}, \{swizzley}, \{swizzlez}, \{swizzlew})");
		try {
		    functions.glViewportSwizzleNV().invokeExact(index, swizzlex, swizzley, swizzlez, swizzlew);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were swizzlex=\{swizzlex}, swizzley=\{swizzley}, swizzlez=\{swizzlez}, swizzlew=\{swizzlew}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glWaitSemaphoreEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glWaitSemaphoreEXT.xhtml">registry.khronos.org</a>
	 */
	public void glWaitSemaphoreEXT(int semaphore, int numBufferBarriers, ConstPointer<Integer> buffers, int numTextureBarriers, ConstPointer<Integer> textures, ConstPointer<GLTextureLayout> srcLayouts) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glWaitSemaphoreEXT(\{semaphore}, \{numBufferBarriers}, \{buffers}, \{numTextureBarriers}, \{textures}, \{srcLayouts})");
		try {
		    functions.glWaitSemaphoreEXT().invokeExact(semaphore, numBufferBarriers, buffers, numTextureBarriers, textures, srcLayouts);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glWaitSemaphoreui64NVX">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glWaitSemaphoreui64NVX.xhtml">registry.khronos.org</a>
	 */
	public void glWaitSemaphoreui64NVX(int waitGpu, int fenceObjectCount, ConstPointer<Integer> semaphoreArray, ConstPointer<Long> fenceValueArray) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glWaitSemaphoreui64NVX(\{waitGpu}, \{fenceObjectCount}, \{semaphoreArray}, \{fenceValueArray})");
		try {
		    functions.glWaitSemaphoreui64NVX().invokeExact(waitGpu, fenceObjectCount, semaphoreArray, fenceValueArray);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glWaitSync">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glWaitSync.xhtml">registry.khronos.org</a>
	 */
	public void glWaitSync(ConstPointer<Void> sync, int flags, long timeout) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glWaitSync(\{sync}, \{flags}, \{timeout})");
		try {
		    functions.glWaitSync().invokeExact(sync, flags, timeout);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glWaitSyncAPPLE">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glWaitSyncAPPLE.xhtml">registry.khronos.org</a>
	 */
	public void glWaitSyncAPPLE(ConstPointer<Void> sync, int flags, long timeout) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glWaitSyncAPPLE(\{sync}, \{flags}, \{timeout})");
		try {
		    functions.glWaitSyncAPPLE().invokeExact(sync, flags, timeout);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glWeightPathsNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glWeightPathsNV.xhtml">registry.khronos.org</a>
	 */
	public void glWeightPathsNV(int resultPath, int numPaths, ConstPointer<Integer> paths, ConstPointer<Float> weights) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glWeightPathsNV(\{resultPath}, \{numPaths}, \{paths}, \{weights})");
		try {
		    functions.glWeightPathsNV().invokeExact(resultPath, numPaths, paths, weights);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glWeightPointerARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glWeightPointerARB.xhtml">registry.khronos.org</a>
	 */
	public void glWeightPointerARB(int size, GLWeightPointerTypeARB type, int stride, ConstPointer<Void> pointer) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glWeightPointerARB(\{size}, \{type}, \{stride}, \{pointer})");
		try {
		    functions.glWeightPointerARB().invokeExact(size, type, stride, pointer);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glWeightPointerOES">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glWeightPointerOES.xhtml">registry.khronos.org</a>
	 */
	public void glWeightPointerOES(int size, GLEnum type, int stride, ConstPointer<Void> pointer) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glWeightPointerOES(\{size}, \{type}, \{stride}, \{pointer})");
		try {
		    functions.glWeightPointerOES().invokeExact(size, type, stride, pointer);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were type=\{type}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glWeightbvARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glWeightbvARB.xhtml">registry.khronos.org</a>
	 */
	public void glWeightbvARB(int size, ConstPointer<Byte> weights) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glWeightbvARB(\{size}, \{weights})");
		try {
		    functions.glWeightbvARB().invokeExact(size, weights);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glWeightdvARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glWeightdvARB.xhtml">registry.khronos.org</a>
	 */
	public void glWeightdvARB(int size, ConstPointer<Double> weights) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glWeightdvARB(\{size}, \{weights})");
		try {
		    functions.glWeightdvARB().invokeExact(size, weights);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glWeightfvARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glWeightfvARB.xhtml">registry.khronos.org</a>
	 */
	public void glWeightfvARB(int size, ConstPointer<Float> weights) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glWeightfvARB(\{size}, \{weights})");
		try {
		    functions.glWeightfvARB().invokeExact(size, weights);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glWeightivARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glWeightivARB.xhtml">registry.khronos.org</a>
	 */
	public void glWeightivARB(int size, ConstPointer<Integer> weights) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glWeightivARB(\{size}, \{weights})");
		try {
		    functions.glWeightivARB().invokeExact(size, weights);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glWeightsvARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glWeightsvARB.xhtml">registry.khronos.org</a>
	 */
	public void glWeightsvARB(int size, ConstPointer<Short> weights) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glWeightsvARB(\{size}, \{weights})");
		try {
		    functions.glWeightsvARB().invokeExact(size, weights);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glWeightubvARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glWeightubvARB.xhtml">registry.khronos.org</a>
	 */
	public void glWeightubvARB(int size, ConstPointer<Byte> weights) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glWeightubvARB(\{size}, \{weights})");
		try {
		    functions.glWeightubvARB().invokeExact(size, weights);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glWeightuivARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glWeightuivARB.xhtml">registry.khronos.org</a>
	 */
	public void glWeightuivARB(int size, ConstPointer<Integer> weights) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glWeightuivARB(\{size}, \{weights})");
		try {
		    functions.glWeightuivARB().invokeExact(size, weights);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glWeightusvARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glWeightusvARB.xhtml">registry.khronos.org</a>
	 */
	public void glWeightusvARB(int size, ConstPointer<Character> weights) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glWeightusvARB(\{size}, \{weights})");
		try {
		    functions.glWeightusvARB().invokeExact(size, weights);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glWindowPos2d">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glWindowPos2d.xhtml">registry.khronos.org</a>
	 */
	public void glWindowPos2d(double x, double y) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glWindowPos2d(\{x}, \{y})");
		try {
		    functions.glWindowPos2d().invokeExact(x, y);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glWindowPos2dARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glWindowPos2dARB.xhtml">registry.khronos.org</a>
	 */
	public void glWindowPos2dARB(double x, double y) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glWindowPos2dARB(\{x}, \{y})");
		try {
		    functions.glWindowPos2dARB().invokeExact(x, y);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glWindowPos2dMESA">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glWindowPos2dMESA.xhtml">registry.khronos.org</a>
	 */
	public void glWindowPos2dMESA(double x, double y) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glWindowPos2dMESA(\{x}, \{y})");
		try {
		    functions.glWindowPos2dMESA().invokeExact(x, y);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glWindowPos2dv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glWindowPos2dv.xhtml">registry.khronos.org</a>
	 */
	public void glWindowPos2dv(ConstPointer<Double> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glWindowPos2dv(\{v})");
		try {
		    functions.glWindowPos2dv().invokeExact(v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glWindowPos2dvARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glWindowPos2dvARB.xhtml">registry.khronos.org</a>
	 */
	public void glWindowPos2dvARB(ConstPointer<Double> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glWindowPos2dvARB(\{v})");
		try {
		    functions.glWindowPos2dvARB().invokeExact(v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glWindowPos2dvMESA">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glWindowPos2dvMESA.xhtml">registry.khronos.org</a>
	 */
	public void glWindowPos2dvMESA(ConstPointer<Double> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glWindowPos2dvMESA(\{v})");
		try {
		    functions.glWindowPos2dvMESA().invokeExact(v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glWindowPos2f">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glWindowPos2f.xhtml">registry.khronos.org</a>
	 */
	public void glWindowPos2f(float x, float y) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glWindowPos2f(\{x}, \{y})");
		try {
		    functions.glWindowPos2f().invokeExact(x, y);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glWindowPos2fARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glWindowPos2fARB.xhtml">registry.khronos.org</a>
	 */
	public void glWindowPos2fARB(float x, float y) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glWindowPos2fARB(\{x}, \{y})");
		try {
		    functions.glWindowPos2fARB().invokeExact(x, y);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glWindowPos2fMESA">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glWindowPos2fMESA.xhtml">registry.khronos.org</a>
	 */
	public void glWindowPos2fMESA(float x, float y) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glWindowPos2fMESA(\{x}, \{y})");
		try {
		    functions.glWindowPos2fMESA().invokeExact(x, y);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glWindowPos2fv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glWindowPos2fv.xhtml">registry.khronos.org</a>
	 */
	public void glWindowPos2fv(ConstPointer<Float> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glWindowPos2fv(\{v})");
		try {
		    functions.glWindowPos2fv().invokeExact(v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glWindowPos2fvARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glWindowPos2fvARB.xhtml">registry.khronos.org</a>
	 */
	public void glWindowPos2fvARB(ConstPointer<Float> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glWindowPos2fvARB(\{v})");
		try {
		    functions.glWindowPos2fvARB().invokeExact(v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glWindowPos2fvMESA">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glWindowPos2fvMESA.xhtml">registry.khronos.org</a>
	 */
	public void glWindowPos2fvMESA(ConstPointer<Float> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glWindowPos2fvMESA(\{v})");
		try {
		    functions.glWindowPos2fvMESA().invokeExact(v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glWindowPos2i">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glWindowPos2i.xhtml">registry.khronos.org</a>
	 */
	public void glWindowPos2i(int x, int y) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glWindowPos2i(\{x}, \{y})");
		try {
		    functions.glWindowPos2i().invokeExact(x, y);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glWindowPos2iARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glWindowPos2iARB.xhtml">registry.khronos.org</a>
	 */
	public void glWindowPos2iARB(int x, int y) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glWindowPos2iARB(\{x}, \{y})");
		try {
		    functions.glWindowPos2iARB().invokeExact(x, y);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glWindowPos2iMESA">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glWindowPos2iMESA.xhtml">registry.khronos.org</a>
	 */
	public void glWindowPos2iMESA(int x, int y) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glWindowPos2iMESA(\{x}, \{y})");
		try {
		    functions.glWindowPos2iMESA().invokeExact(x, y);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glWindowPos2iv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glWindowPos2iv.xhtml">registry.khronos.org</a>
	 */
	public void glWindowPos2iv(ConstPointer<Integer> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glWindowPos2iv(\{v})");
		try {
		    functions.glWindowPos2iv().invokeExact(v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glWindowPos2ivARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glWindowPos2ivARB.xhtml">registry.khronos.org</a>
	 */
	public void glWindowPos2ivARB(ConstPointer<Integer> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glWindowPos2ivARB(\{v})");
		try {
		    functions.glWindowPos2ivARB().invokeExact(v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glWindowPos2ivMESA">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glWindowPos2ivMESA.xhtml">registry.khronos.org</a>
	 */
	public void glWindowPos2ivMESA(ConstPointer<Integer> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glWindowPos2ivMESA(\{v})");
		try {
		    functions.glWindowPos2ivMESA().invokeExact(v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glWindowPos2s">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glWindowPos2s.xhtml">registry.khronos.org</a>
	 */
	public void glWindowPos2s(short x, short y) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glWindowPos2s(\{x}, \{y})");
		try {
		    functions.glWindowPos2s().invokeExact(x, y);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glWindowPos2sARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glWindowPos2sARB.xhtml">registry.khronos.org</a>
	 */
	public void glWindowPos2sARB(short x, short y) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glWindowPos2sARB(\{x}, \{y})");
		try {
		    functions.glWindowPos2sARB().invokeExact(x, y);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glWindowPos2sMESA">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glWindowPos2sMESA.xhtml">registry.khronos.org</a>
	 */
	public void glWindowPos2sMESA(short x, short y) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glWindowPos2sMESA(\{x}, \{y})");
		try {
		    functions.glWindowPos2sMESA().invokeExact(x, y);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glWindowPos2sv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glWindowPos2sv.xhtml">registry.khronos.org</a>
	 */
	public void glWindowPos2sv(ConstPointer<Short> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glWindowPos2sv(\{v})");
		try {
		    functions.glWindowPos2sv().invokeExact(v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glWindowPos2svARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glWindowPos2svARB.xhtml">registry.khronos.org</a>
	 */
	public void glWindowPos2svARB(ConstPointer<Short> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glWindowPos2svARB(\{v})");
		try {
		    functions.glWindowPos2svARB().invokeExact(v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glWindowPos2svMESA">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glWindowPos2svMESA.xhtml">registry.khronos.org</a>
	 */
	public void glWindowPos2svMESA(ConstPointer<Short> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glWindowPos2svMESA(\{v})");
		try {
		    functions.glWindowPos2svMESA().invokeExact(v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glWindowPos3d">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glWindowPos3d.xhtml">registry.khronos.org</a>
	 */
	public void glWindowPos3d(double x, double y, double z) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glWindowPos3d(\{x}, \{y}, \{z})");
		try {
		    functions.glWindowPos3d().invokeExact(x, y, z);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glWindowPos3dARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glWindowPos3dARB.xhtml">registry.khronos.org</a>
	 */
	public void glWindowPos3dARB(double x, double y, double z) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glWindowPos3dARB(\{x}, \{y}, \{z})");
		try {
		    functions.glWindowPos3dARB().invokeExact(x, y, z);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glWindowPos3dMESA">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glWindowPos3dMESA.xhtml">registry.khronos.org</a>
	 */
	public void glWindowPos3dMESA(double x, double y, double z) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glWindowPos3dMESA(\{x}, \{y}, \{z})");
		try {
		    functions.glWindowPos3dMESA().invokeExact(x, y, z);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glWindowPos3dv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glWindowPos3dv.xhtml">registry.khronos.org</a>
	 */
	public void glWindowPos3dv(ConstPointer<Double> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glWindowPos3dv(\{v})");
		try {
		    functions.glWindowPos3dv().invokeExact(v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glWindowPos3dvARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glWindowPos3dvARB.xhtml">registry.khronos.org</a>
	 */
	public void glWindowPos3dvARB(ConstPointer<Double> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glWindowPos3dvARB(\{v})");
		try {
		    functions.glWindowPos3dvARB().invokeExact(v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glWindowPos3dvMESA">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glWindowPos3dvMESA.xhtml">registry.khronos.org</a>
	 */
	public void glWindowPos3dvMESA(ConstPointer<Double> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glWindowPos3dvMESA(\{v})");
		try {
		    functions.glWindowPos3dvMESA().invokeExact(v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glWindowPos3f">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glWindowPos3f.xhtml">registry.khronos.org</a>
	 */
	public void glWindowPos3f(float x, float y, float z) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glWindowPos3f(\{x}, \{y}, \{z})");
		try {
		    functions.glWindowPos3f().invokeExact(x, y, z);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glWindowPos3fARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glWindowPos3fARB.xhtml">registry.khronos.org</a>
	 */
	public void glWindowPos3fARB(float x, float y, float z) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glWindowPos3fARB(\{x}, \{y}, \{z})");
		try {
		    functions.glWindowPos3fARB().invokeExact(x, y, z);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glWindowPos3fMESA">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glWindowPos3fMESA.xhtml">registry.khronos.org</a>
	 */
	public void glWindowPos3fMESA(float x, float y, float z) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glWindowPos3fMESA(\{x}, \{y}, \{z})");
		try {
		    functions.glWindowPos3fMESA().invokeExact(x, y, z);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glWindowPos3fv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glWindowPos3fv.xhtml">registry.khronos.org</a>
	 */
	public void glWindowPos3fv(ConstPointer<Float> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glWindowPos3fv(\{v})");
		try {
		    functions.glWindowPos3fv().invokeExact(v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glWindowPos3fvARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glWindowPos3fvARB.xhtml">registry.khronos.org</a>
	 */
	public void glWindowPos3fvARB(ConstPointer<Float> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glWindowPos3fvARB(\{v})");
		try {
		    functions.glWindowPos3fvARB().invokeExact(v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glWindowPos3fvMESA">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glWindowPos3fvMESA.xhtml">registry.khronos.org</a>
	 */
	public void glWindowPos3fvMESA(ConstPointer<Float> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glWindowPos3fvMESA(\{v})");
		try {
		    functions.glWindowPos3fvMESA().invokeExact(v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glWindowPos3i">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glWindowPos3i.xhtml">registry.khronos.org</a>
	 */
	public void glWindowPos3i(int x, int y, int z) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glWindowPos3i(\{x}, \{y}, \{z})");
		try {
		    functions.glWindowPos3i().invokeExact(x, y, z);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glWindowPos3iARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glWindowPos3iARB.xhtml">registry.khronos.org</a>
	 */
	public void glWindowPos3iARB(int x, int y, int z) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glWindowPos3iARB(\{x}, \{y}, \{z})");
		try {
		    functions.glWindowPos3iARB().invokeExact(x, y, z);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glWindowPos3iMESA">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glWindowPos3iMESA.xhtml">registry.khronos.org</a>
	 */
	public void glWindowPos3iMESA(int x, int y, int z) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glWindowPos3iMESA(\{x}, \{y}, \{z})");
		try {
		    functions.glWindowPos3iMESA().invokeExact(x, y, z);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glWindowPos3iv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glWindowPos3iv.xhtml">registry.khronos.org</a>
	 */
	public void glWindowPos3iv(ConstPointer<Integer> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glWindowPos3iv(\{v})");
		try {
		    functions.glWindowPos3iv().invokeExact(v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glWindowPos3ivARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glWindowPos3ivARB.xhtml">registry.khronos.org</a>
	 */
	public void glWindowPos3ivARB(ConstPointer<Integer> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glWindowPos3ivARB(\{v})");
		try {
		    functions.glWindowPos3ivARB().invokeExact(v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glWindowPos3ivMESA">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glWindowPos3ivMESA.xhtml">registry.khronos.org</a>
	 */
	public void glWindowPos3ivMESA(ConstPointer<Integer> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glWindowPos3ivMESA(\{v})");
		try {
		    functions.glWindowPos3ivMESA().invokeExact(v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glWindowPos3s">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glWindowPos3s.xhtml">registry.khronos.org</a>
	 */
	public void glWindowPos3s(short x, short y, short z) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glWindowPos3s(\{x}, \{y}, \{z})");
		try {
		    functions.glWindowPos3s().invokeExact(x, y, z);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glWindowPos3sARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glWindowPos3sARB.xhtml">registry.khronos.org</a>
	 */
	public void glWindowPos3sARB(short x, short y, short z) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glWindowPos3sARB(\{x}, \{y}, \{z})");
		try {
		    functions.glWindowPos3sARB().invokeExact(x, y, z);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glWindowPos3sMESA">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glWindowPos3sMESA.xhtml">registry.khronos.org</a>
	 */
	public void glWindowPos3sMESA(short x, short y, short z) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glWindowPos3sMESA(\{x}, \{y}, \{z})");
		try {
		    functions.glWindowPos3sMESA().invokeExact(x, y, z);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glWindowPos3sv">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glWindowPos3sv.xhtml">registry.khronos.org</a>
	 */
	public void glWindowPos3sv(ConstPointer<Short> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glWindowPos3sv(\{v})");
		try {
		    functions.glWindowPos3sv().invokeExact(v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glWindowPos3svARB">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glWindowPos3svARB.xhtml">registry.khronos.org</a>
	 */
	public void glWindowPos3svARB(ConstPointer<Short> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glWindowPos3svARB(\{v})");
		try {
		    functions.glWindowPos3svARB().invokeExact(v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glWindowPos3svMESA">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glWindowPos3svMESA.xhtml">registry.khronos.org</a>
	 */
	public void glWindowPos3svMESA(ConstPointer<Short> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glWindowPos3svMESA(\{v})");
		try {
		    functions.glWindowPos3svMESA().invokeExact(v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glWindowPos4dMESA">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glWindowPos4dMESA.xhtml">registry.khronos.org</a>
	 */
	public void glWindowPos4dMESA(double x, double y, double z, double w) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glWindowPos4dMESA(\{x}, \{y}, \{z}, \{w})");
		try {
		    functions.glWindowPos4dMESA().invokeExact(x, y, z, w);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glWindowPos4dvMESA">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glWindowPos4dvMESA.xhtml">registry.khronos.org</a>
	 */
	public void glWindowPos4dvMESA(ConstPointer<Double> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glWindowPos4dvMESA(\{v})");
		try {
		    functions.glWindowPos4dvMESA().invokeExact(v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glWindowPos4fMESA">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glWindowPos4fMESA.xhtml">registry.khronos.org</a>
	 */
	public void glWindowPos4fMESA(float x, float y, float z, float w) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glWindowPos4fMESA(\{x}, \{y}, \{z}, \{w})");
		try {
		    functions.glWindowPos4fMESA().invokeExact(x, y, z, w);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glWindowPos4fvMESA">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glWindowPos4fvMESA.xhtml">registry.khronos.org</a>
	 */
	public void glWindowPos4fvMESA(ConstPointer<Float> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glWindowPos4fvMESA(\{v})");
		try {
		    functions.glWindowPos4fvMESA().invokeExact(v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glWindowPos4iMESA">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glWindowPos4iMESA.xhtml">registry.khronos.org</a>
	 */
	public void glWindowPos4iMESA(int x, int y, int z, int w) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glWindowPos4iMESA(\{x}, \{y}, \{z}, \{w})");
		try {
		    functions.glWindowPos4iMESA().invokeExact(x, y, z, w);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glWindowPos4ivMESA">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glWindowPos4ivMESA.xhtml">registry.khronos.org</a>
	 */
	public void glWindowPos4ivMESA(ConstPointer<Integer> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glWindowPos4ivMESA(\{v})");
		try {
		    functions.glWindowPos4ivMESA().invokeExact(v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glWindowPos4sMESA">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glWindowPos4sMESA.xhtml">registry.khronos.org</a>
	 */
	public void glWindowPos4sMESA(short x, short y, short z, short w) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glWindowPos4sMESA(\{x}, \{y}, \{z}, \{w})");
		try {
		    functions.glWindowPos4sMESA().invokeExact(x, y, z, w);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glWindowPos4svMESA">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glWindowPos4svMESA.xhtml">registry.khronos.org</a>
	 */
	public void glWindowPos4svMESA(ConstPointer<Short> v) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glWindowPos4svMESA(\{v})");
		try {
		    functions.glWindowPos4svMESA().invokeExact(v);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glWindowRectanglesEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glWindowRectanglesEXT.xhtml">registry.khronos.org</a>
	 */
	public void glWindowRectanglesEXT(GLEnum mode, int count, ConstPointer<Integer> box) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glWindowRectanglesEXT(\{mode}, \{count}, \{box})");
		try {
		    functions.glWindowRectanglesEXT().invokeExact(mode, count, box);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were mode=\{mode}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glWriteMaskEXT">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glWriteMaskEXT.xhtml">registry.khronos.org</a>
	 */
	public void glWriteMaskEXT(int res, int in, GLVertexShaderWriteMaskEXT outX, GLVertexShaderWriteMaskEXT outY, GLVertexShaderWriteMaskEXT outZ, GLVertexShaderWriteMaskEXT outW) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glWriteMaskEXT(\{res}, \{in}, \{outX}, \{outY}, \{outZ}, \{outW})");
		try {
		    functions.glWriteMaskEXT().invokeExact(res, in, outX, outY, outZ, outW);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were outX=\{outX}, outY=\{outY}, outZ=\{outZ}, outW=\{outW}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glDrawVkImageNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glDrawVkImageNV.xhtml">registry.khronos.org</a>
	 */
	public void glDrawVkImageNV(long vkImage, int sampler, float x0, float y0, float x1, float y1, float z, float s0, float t0, float s1, float t1) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glDrawVkImageNV(\{vkImage}, \{sampler}, \{x0}, \{y0}, \{x1}, \{y1}, \{z}, \{s0}, \{t0}, \{s1}, \{t1})");
		try {
		    functions.glDrawVkImageNV().invokeExact(vkImage, sampler, x0, y0, x1, y1, z, s0, t0, s1, t1);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetVkProcAddrNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetVkProcAddrNV.xhtml">registry.khronos.org</a>
	 */
	public int glGetVkProcAddrNV(ConstPointer<Byte> name) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetVkProcAddrNV(\{name})");
		try {
		    var r = (int)functions.glGetVkProcAddrNV().invokeExact(name);
			checkError();
			return r;
		} catch (Throwable e) { throw throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glWaitVkSemaphoreNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glWaitVkSemaphoreNV.xhtml">registry.khronos.org</a>
	 */
	public void glWaitVkSemaphoreNV(long vkSemaphore) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glWaitVkSemaphoreNV(\{vkSemaphore})");
		try {
		    functions.glWaitVkSemaphoreNV().invokeExact(vkSemaphore);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glSignalVkSemaphoreNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glSignalVkSemaphoreNV.xhtml">registry.khronos.org</a>
	 */
	public void glSignalVkSemaphoreNV(long vkSemaphore) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glSignalVkSemaphoreNV(\{vkSemaphore})");
		try {
		    functions.glSignalVkSemaphoreNV().invokeExact(vkSemaphore);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glSignalVkFenceNV">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glSignalVkFenceNV.xhtml">registry.khronos.org</a>
	 */
	public void glSignalVkFenceNV(long vkFence) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glSignalVkFenceNV(\{vkFence})");
		try {
		    functions.glSignalVkFenceNV().invokeExact(vkFence);
			checkError();
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glFramebufferParameteriMESA">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glFramebufferParameteriMESA.xhtml">registry.khronos.org</a>
	 */
	public void glFramebufferParameteriMESA(GLFramebufferTarget target, GLFramebufferParameterName pname, int param) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glFramebufferParameteriMESA(\{target}, \{pname}, \{param})");
		try {
		    functions.glFramebufferParameteriMESA().invokeExact(target, pname, param);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	/**
	 *
	 * @see <a href="https://docs.gl/gl4/glGetFramebufferParameterivMESA">docs.gl</a>
	 * @see <a href="https://registry.khronos.org/OpenGL-Refpages/gl4/html/glGetFramebufferParameterivMESA.xhtml">registry.khronos.org</a>
	 */
	public void glGetFramebufferParameterivMESA(GLFramebufferTarget target, GLFramebufferAttachmentParameterName pname, MutPointer<Integer> params) {
	    if (LOG_COMMANDS) logCmd(() -> STR."glGetFramebufferParameterivMESA(\{target}, \{pname}, \{params})");
		try {
		    functions.glGetFramebufferParameterivMESA().invokeExact(target, pname, params);
			checkError();
		} catch (GLException e) {
			if (e.code == GLErrorCode.INVALID_ENUM) {
				throw e.withAdditionalMessage(STR."The enums passed to the command were target=\{target}, pname=\{pname}.");
			}
			throw e;
		} catch (Throwable e) { throwUnchecked(e); }
	}

	public void checkError() {
		if (!ERROR_CHECKS) return;
		var error = glGetError();
		if (error == GLErrorCode.NO_ERROR) return;
		throw new GLException(error);
	}

    public int glGetAttribLocation(int program, String name) {
		if (LOG_COMMANDS) logCmd(() -> STR."glGetAttribLocation(\{program}, \{name})");
		try (var arena = Arena.ofConfined()) {
			return glGetAttribLocation(program, ConstPointer.allocateCString(arena, name));
		}
	}

    public int glGetUniformLocation(int program, String name) {
		if (LOG_COMMANDS) logCmd(() -> STR."glGetUniformLocation(\{program}, \{name})");
		try (var arena = Arena.ofConfined()) {
			return glGetUniformLocation(program, ConstPointer.allocateCString(arena, name));
		}
	}

    private static void logCmd(StringSupplier msg) {
    	LOGGER.log(msg, COMMAND_TAGS);
    }
}